% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the semantics of the assertion
  language.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `assert_semantics`;;

new_parent `assert_syntax`;;
new_parent `bindings`;;


loadf `autoload`;;

autoload_all_defs_and_thms
 `assert_syntax bindings finite_sets more_finite_sets`;;


loadf `semantics_tactics`;;


% First we define << as a lexicographic ordering between lists of numbers. %


let LEXI_DEF = 
    new_infix_list_rec_definition (`LEXI_DEF`,
      "($<< NIL es'         = (~(NULL es')))  /\
       ($<< (CONS e es) es' =
                   ((es' = []) => F
                 |  ((e < (HD es')) \/
                     ((e = (HD es')) /\
                      ($<< es (TL es'))))))");;


let LEXI = prove_thm
 (`LEXI`,
  "([] << [] = F) /\
   (!es e. [] << (CONS e es) = T) /\
   (!es e. (CONS e es) << [] = F) /\
   (!e1 es1 e2 es2. (CONS e1 es1) << (CONS e2 es2) =
                      ((e1 < e2) \/
                       ((e1 = e2) /\
                        (es1 << es2))))",
    REWRITE_TAC[LEXI_DEF;HD;TL;NULL;NOT_CONS_NIL]
   );;

%
g "([] << [] = F) /\
   (!es e. [] << (CONS e es) = T) /\
   (!es e. (CONS e es) << [] = F) /\
   (!e1 es1 e2 es2. (CONS e1 es1) << (CONS e2 es2) =
                      ((e1 < e2) \/
                       ((e1 = e2) /\
                        (es1 << es2))))";;
e(REWRITE_TAC[LEXI_DEF;HD;TL]);;
e(REWRITE_TAC[NULL]);;
e(REWRITE_TAC[NOT_CONS_NIL]);;
%


% ===================================================================== %
% STATES:                                                               %
%   We define states as mappings from variables to numeric values.      %
%   These are nonnegative integers, in {0,1,2,... }.                    %
% ===================================================================== %

new_type_abbrev(`state`, ":var -> num");;


let V_DEF =
    new_rec_def `V_DEF` vexp_Axiom
      "(V (ANUM n) (s:state) = n)    /\
       (V (AVAR x) s         = s x)  /\
       (V (APLUS e1 e2) s    = (V e1 s) + (V e2 s))  /\
       (V (AMINUS e1 e2) s   = (V e1 s) - (V e2 s))  /\
       (V (AMULT e1 e2) s    = (V e1 s) * (V e2 s))";;

let VS_DEF = 
    new_list_rec_definition (`VS_DEF`,
      "(VS NIL s         = (NIL))  /\
       (VS (CONS e es) s = (CONS (V e s) (VS es s)))");;

let A_DEF =
    new_rec_def `A_DEF` aexp_Axiom
      "(A (ATRUE) s           = T)    /\
       (A (AFALSE) s          = F)    /\
       (A (AEQ e1 e2)      s  = (V e1 s) = (V e2 s))           /\
       (A (ALESS e1 e2)    s  = (V e1 s) < (V e2 s))           /\
       (A (ALLESS es1 es2) s  = (VS es1 s) << (VS es2 s))      /\
       (A (AAND b1 b2)     s  = (A b1 s) /\ (A b2 s))          /\
       (A (AOR  b1 b2)     s  = (A b1 s) \/ (A b2 s))          /\
       (A (ANOT b)         s  = ~(A b s))                      /\
       (A (AIMP b1 b2)     s  = (A b1 s) ==> (A b2 s))         /\
       (A (AEQB b1 b2)     s  = (A b1 s) = (A b2 s))           /\
       (A (ACOND b1 b2 b3) s  = (A b1 s => A b2 s | A b3 s))   /\
       (A (ACLOSE b)       s  = (!s1. A b s1))                 /\
       (A (AFORALL x b)    s = (!n. A b (BND n x s)))          /\
       (A (AEXISTS x b)    s = (?n. A b (BND n x s))) ";;


let LENGTH_VS =
 prove_thm
  (`LENGTH_VS`,
   "!vs s. LENGTH (VS vs s) = LENGTH vs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[VS_DEF;LENGTH]
  );;

% LENGTH_VS
g "!vs s. LENGTH (VS vs s) = LENGTH vs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[VS_DEF;LENGTH]);;

 e(ASM_REWRITE_TAC[VS_DEF;LENGTH]);;
%





% ======================================================= %
% We define free and bound vars on assertion expressions. %
% ======================================================= %

let FVv_DEF = 
    new_recursive_definition false vexp_Axiom `FVv_DEF`
      "(FVv (ANUM n)       = EMPTY)           /\
       (FVv (AVAR x)       = INSERT x EMPTY)  /\
       (FVv (APLUS e1 e2)  = FVv e1 UNION FVv e2)  /\
       (FVv (AMINUS e1 e2) = FVv e1 UNION FVv e2)  /\
       (FVv (AMULT e1 e2)  = FVv e1 UNION FVv e2)";;

let FVvs_DEF = 
    new_list_rec_definition (`FVvs_DEF`,
      "(FVvs NIL         = EMPTY)  /\
       (FVvs (CONS v vs) = FVv v UNION FVvs vs)");;

let FVa_DEF =
    new_recursive_definition false aexp_Axiom `FVa_DEF`
      "(FVa (ATRUE)          = EMPTY)    /\
       (FVa (AFALSE)         = EMPTY)    /\
       (FVa (AEQ e1 e2)      = FVv e1 UNION FVv e2)  /\
       (FVa (ALESS e1 e2)    = FVv e1 UNION FVv e2)  /\
       (FVa (ALLESS es1 es2) = FVvs es1 UNION FVvs es2)  /\
       (FVa (AAND b1 b2)     = FVa b1 UNION FVa b2)  /\
       (FVa (AOR  b1 b2)     = FVa b1 UNION FVa b2)  /\
       (FVa (ANOT b)         = FVa b)    /\
       (FVa (AIMP b1 b2)     = FVa b1 UNION FVa b2)  /\
       (FVa (AEQB b1 b2)     = FVa b1 UNION FVa b2)  /\
       (FVa (ACOND b1 b2 b3) = FVa b1 UNION FVa b2 UNION FVa b3)  /\
       (FVa (ACLOSE b)       = EMPTY)    /\
       (FVa (AFORALL x b)    = FVa b DELETE x)       /\
       (FVa (AEXISTS x b)    = FVa b DELETE x)";;


% =========================================================== %
% We now prove several independence theorems for assertion    %
% language numeric and boolean expressions.                   %
% These state the essential unchangedness of the semantics of %
% each construct, under modifications of the state which are  %
% outside the free variables of the construct.                %
% =========================================================== %


% Old name: NOT_FREE_V_BND %

let V_BND_THM =
 prove_thm
  (`V_BND_THM`,
   "!v n x st. ~(x IN FVv v) ==> (V v (BND n x st) = V v st)",
   INDUCT_THEN vexp_induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[FVv_DEF;V_DEF]
    THEN REWRITE_TAC[IN;IN_UNION;DE_MORGAN_THM]
    THEN REPEAT STRIP_TAC
    THENL [MATCH_MP_TAC BND_NOT_MATCH
           THEN POP_ASSUM (ACCEPT_TAC o NOT_EQ_SYM);
           RES_TAC;
           RES_TAC;
           RES_TAC
          ]
    THEN ASM_REWRITE_TAC[]
  );;

% V_BND_THM:
g "!v n x st. ~(x IN FVv v) ==> (V v (BND n x st) = V v st)";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;
 e(REWRITE_TAC[FVv_DEF;V_DEF]);;

 e(REWRITE_TAC[FVv_DEF;IN;V_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC BND_NOT_MATCH);;
 e(POP_ASSUM (ACCEPT_TAC o NOT_EQ_SYM));;

 e(ASM_REWRITE_TAC[FVv_DEF;IN_UNION;DE_MORGAN_THM;V_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[FVv_DEF;IN_UNION;DE_MORGAN_THM;V_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[FVv_DEF;IN_UNION;DE_MORGAN_THM;V_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let VS_BND_THM =
 prove_thm
  (`VS_BND_THM`,
   "!vs n x st. ~(x IN FVvs vs) ==> (VS vs (BND n x st) = VS vs st)",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVvs_DEF;IN_UNION;DE_MORGAN_THM;VS_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM V_BND_THM
   THEN RES_THEN REWRITE_THM
  );;

% VS_BND_THM:
g "!vs n x st. ~(x IN FVvs vs) ==> (VS vs (BND n x st) = VS vs st)";;
e(LIST_INDUCT_TAC);;
 e(REWRITE_TAC[VS_DEF]);;

 e(REWRITE_TAC[FVvs_DEF;IN_UNION;DE_MORGAN_THM;VS_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
 e(RES_THEN REWRITE_THM);;
%


% Old name: NOT_FREE_A_BND %

let A_BND_THM =
 prove_thm
  (`A_BND_THM`,
   "!a n x st. ~(x IN FVa a) ==>
                (A a (BND n x st) = A a st)",

   INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVa_DEF;A_DEF]
   THEN REWRITE_TAC[IN_UNION;IN_DELETE;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC V_BND_THM
   THEN IMP_RES_TAC VS_BND_THM
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[BND_OVERRIDE]
   THEN ASM_CASES_TAC "v = (x:var)"
   THEN ASM_REWRITE_TAC[BND_OVERRIDE]
   THEN IMP_RES_TAC (INST_TYPE [":num",":*"] BND_COMMUTE)
   THEN ASM_REWRITE_TAC[]
  );;

% A_BND_THM:
g "!a n x st. ~(x IN FVa a) ==>
              (A a (BND n x st) = A a st)";;
e(INDUCT_THEN aexp_induct ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;A_DEF]
  THEN REWRITE_TAC[IN_UNION;IN_DELETE;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC
  THEN IMP_RES_TAC V_BND_THM
  THEN IMP_RES_TAC VS_BND_THM
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[]);;

 e(ASM_CASES_TAC "v = (x:var)");;
  e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

  e(IMP_RES_TAC (INST_TYPE [":num",":*"] BND_COMMUTE));;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[BND_OVERRIDE]);;

 e(ASM_CASES_TAC "v = (x:var)");;
  e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

  e(IMP_RES_TAC (INST_TYPE [":num",":*"] BND_COMMUTE));;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[BND_OVERRIDE]);;
%




let V_BNDsL_THM =
 prove_thm
  (`V_BNDsL_THM`,
   "!ms xs v s.
     (DISJOINT (SL xs) (FVv v)) ==>
       (V v (BNDsL ms xs s) = V v s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM V_BND_THM
  );;

% V_BNDsL_THM:
g "!ms xs v s.
     (DISJOINT (SL xs) (FVv v)) ==>
       (V v (BNDsL ms xs s) = V v s)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(RES_TAC);;
e(POP_ASSUM REWRITE_THM);;
e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
%

let V_BNDsR_THM =
 prove_thm
  (`V_BNDsR_THM`,
   "!ms xs v s.
     (DISJOINT (SL xs) (FVv v)) ==>
       (V v (BNDsR ms xs s) = V v s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM V_BND_THM
   THEN RES_THEN REWRITE_THM
  );;

% V_BNDsR_THM:
g "!ms xs v s.
     (DISJOINT (SL xs) (FVv v)) ==>
       (V v (BNDsR ms xs s) = V v s)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
e(RES_THEN REWRITE_THM);;
%




let VS_BNDsL_THM =
 prove_thm
  (`VS_BNDsL_THM`,
   "!vs ms xs s.
     (DISJOINT (SL xs) (FVvs vs)) ==>
       (VS vs (BNDsL ms xs s) = VS vs s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVvs_DEF;VS_DEF]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_UNION]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM V_BNDsL_THM
  );;

% VS_BNDsL_THM:
g "!vs ms xs s.
     (DISJOINT (SL xs) (FVvs vs)) ==>
       (VS vs (BNDsL ms xs s) = VS vs s)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;

 e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[DISJOINT_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REPEAT GEN_TAC
   THEN STRIP_TAC);;
e(RES_THEN REWRITE_THM);;
e(IMP_RES_THEN REWRITE_THM V_BNDsL_THM);;
%

let VS_BNDsR_THM =
 prove_thm
  (`VS_BNDsR_THM`,
   "!vs ms xs s.
     (DISJOINT (SL xs) (FVvs vs)) ==>
       (VS vs (BNDsR ms xs s) = VS vs s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVvs_DEF;VS_DEF]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_UNION]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM V_BNDsR_THM
  );;

% VS_BNDsR_THM:
g "!vs ms xs s.
     (DISJOINT (SL xs) (FVvs vs)) ==>
       (VS vs (BNDsR ms xs s) = VS vs s)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;

 e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[DISJOINT_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REPEAT GEN_TAC
   THEN STRIP_TAC);;
e(RES_THEN REWRITE_THM);;
e(IMP_RES_THEN REWRITE_THM V_BNDsR_THM);;
%


let A_BNDsL_THM =
 prove_thm
  (`A_BNDsL_THM`,
   "!ms xs a s.
     (DISJOINT (SL xs) (FVa a)) ==>
       (A a (BNDsL ms xs s) = A a s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM A_BND_THM
  );;

% A_BNDsL_THM:
g "!ms xs a s.
     (DISJOINT (SL xs) (FVa a)) ==>
       (A a (BNDsL ms xs s) = A a s)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(RES_TAC);;
e(POP_TAC THEN POP_TAC);;
e(POP_ASSUM REWRITE_THM);;
e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
%

let A_BNDsR_THM =
 prove_thm
  (`A_BNDsR_THM`,
   "!ms xs a s.
     (DISJOINT (SL xs) (FVa a)) ==>
       (A a (BNDsR ms xs s) = A a s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM A_BND_THM
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% A_BNDsR_THM:
g "!ms xs a s.
     (DISJOINT (SL xs) (FVa a)) ==>
       (A a (BNDsR ms xs s) = A a s)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%


% The effect of evaluating an assertion-language expression is the
  same in two different states s1 and s2 *if* the states are the same
  on the free variables of the expression.  This is captured in the
  next few theorems.
%



let OR_IMP_AND =
 TAC_PROOF
  (([],
    "(a \/ b ==> c) = ((a ==> c) /\ (b ==> c))"),
   EQ_TAC
   THENL
    [ REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[];

      REPEAT STRIP_TAC
      THEN RES_TAC
    ]
  );;

% OR_IMP_AND
g "(a \/ b ==> c) = ((a ==> c) /\ (b ==> c))";;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%


let V_EQ_THM =
 prove_thm
  (`V_EQ_THM`,
   "!v s1 s2.
      (!x. x IN FVv v ==> (s1 x = s2 x)) ==>
      (V v s1 = V v s2)",
   INDUCT_THEN vexp_induct ASSUME_TAC
    THEN REWRITE_TAC[FVv_DEF;V_DEF]
    THEN REWRITE_TAC[IN;IN_UNION;OR_IMP_AND]
    THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
    THEN REPEAT STRIP_TAC
    THENL [FIRST_ASSUM MATCH_MP_TAC
           THEN REFL_TAC;
           RES_TAC;
           RES_TAC;
           RES_TAC
          ]
    THEN ASM_REWRITE_TAC[]
  );;

%< V_EQ_THM:
g "!v xs s1 s2.
      (FVv v SUBSET SL xs) /\ (MAP s1 xs = MAP s2 xs) ==>
      (V v s1 = V v s2)";;

g "!v s1 s2.
      (!x. x IN FVv v ==> (s1 x = s2 x)) ==>
      (V v s1 = V v s2)";;

e(INDUCT_THEN vexp_induct ASSUME_TAC);;

 e(REWRITE_TAC[FVv_DEF;V_DEF]);;

 e(REWRITE_TAC[FVv_DEF;IN;V_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(REFL_TAC);;

 e(REWRITE_TAC[FVv_DEF;IN_UNION;OR_IMP_AND;V_DEF]);;
 e(CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV));;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[FVv_DEF;IN_UNION;OR_IMP_AND;V_DEF]);;
 e(CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV));;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[FVv_DEF;IN_UNION;OR_IMP_AND;V_DEF]);;
 e(CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV));;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


let VS_EQ_THM =
 prove_thm
  (`VS_EQ_THM`,
   "!vs s1 s2.
      (!x. x IN FVvs vs ==> (s1 x = s2 x)) ==>
      (VS vs s1 = VS vs s2)",
   LIST_INDUCT_TAC
    THEN REWRITE_TAC[FVvs_DEF;VS_DEF]
    THEN REWRITE_TAC[IN_UNION;OR_IMP_AND]
    THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC V_EQ_THM
    THEN RES_TAC
    THEN ASM_REWRITE_TAC[]
  );;

%< VS_EQ_THM:
g "!vs s1 s2.
      (!x. x IN FVvs vs ==> (s1 x = s2 x)) ==>
      (VS vs s1 = VS vs s2)";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;

 e(REWRITE_TAC[FVvs_DEF;IN_UNION;OR_IMP_AND;VS_DEF]);;
 e(CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV));;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC V_EQ_THM);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%




let A_EQ_THM =
 prove_thm
  (`A_EQ_THM`,
   "!a s1 s2.
      (!x. x IN FVa a ==> (s1 x = s2 x)) ==>
      (A a s1 = A a s2)",
   INDUCT_THEN aexp_induct ASSUME_TAC
    THEN REWRITE_TAC[FVa_DEF;A_DEF]
    THEN REWRITE_TAC[IN_UNION;OR_IMP_AND]
    THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
    THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC V_EQ_THM
    THEN IMP_RES_TAC VS_EQ_THM
    THEN RES_TAC
    THEN ASM_REWRITE_TAC[]
    THEN ( FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC )
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC
    THEN REWRITE_TAC[BND]
    THEN COND_CASES_TAC
    THEN REWRITE_TAC[]
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN ASM_REWRITE_TAC[IN_DELETE]
  );;

%< A_EQ_THM:
g "!a s1 s2.
      (!x. x IN FVa a ==> (s1 x = s2 x)) ==>
      (A a s1 = A a s2)";;

e(INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVa_DEF;A_DEF]
   THEN REWRITE_TAC[IN_UNION;OR_IMP_AND]
   THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
   THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC V_EQ_THM
   THEN IMP_RES_TAC VS_EQ_THM
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]);;

 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BND]);;
 e(COND_CASES_TAC);;

  e(REWRITE_TAC[]);;

  e(REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[IN_DELETE]);;

 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BND]);;
 e(COND_CASES_TAC);;

  e(REWRITE_TAC[]);;

  e(REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[IN_DELETE]);;
>%



close_theory();;

