% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definition of the `BND` operator to add a 
  binding to a mapping, and the associated `BNDsL` and `BNDsR` operators
  for adding a list of bindings to a mapping, either left-to-right or
  right-to-left.  It also contains the definitions of the `SL` operator
  to convert a list into a set, and the `DL` predicate, which is true
  when a list contains no duplicates.  Along with these definitions are
  many theorems concerning these operators, especially concerning the
  binding operators.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path( (library_pathname() ^ `/finite_sets/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `bindings`;;

new_parent `more_finite_sets`;;

loadf `load_finite_sets`;;
load_finite_sets();;

% Set up autoloading from ancestor theories of bindings %

loadf `autoload`;;

autoload_all_defs_and_thms
     `more_finite_sets`;;

loadf `semantics_tactics`;;


% ===================================================================== %
% In order to define the semantics of assignment, we must define the    %
% incremental modification of states.                                   %
% ===================================================================== %

% BND : *   ->   **   ->   (** -> *)    ->   (** -> *)
      value    index     old dictionary    new dictionary

     BND v x s = s[v/x]
%

let BND_DEF = new_definition
 (`BND_DEF`,
  "BND (v:*) (x:**) s  =  \y. ((y=x) => v | s y)");;

% Five useful theorems about BND %

let BND =
 prove_thm
  (`BND`,
   "!(v:*) (x:**) s y. BND v x s y = (y=x) => v | s y",
   REWRITE_TAC[BND_DEF]
   THEN BETA_TAC
   THEN REWRITE_TAC[]);;

let BND_MATCH =
 prove_thm
  (`BND_MATCH`,
   "!(e:*) (x:**) s. BND e x s x = e",
   REWRITE_TAC[BND]
   THEN REWRITE_TAC[]);;

let BND_NOT_MATCH =
 prove_thm
  (`BND_NOT_MATCH`,
   "!(e:*) (x:**) s y. ~(y = x) ==> (BND e x s y = (s y))",
   REWRITE_TAC[BND]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]);;

let BND_OVERRIDE = prove_thm
   (`BND_OVERRIDE`,
    "!(n:*) m (x:**) s. BND n x (BND m x s) = BND n x s",
    REPEAT STRIP_TAC
    THEN EXT_TAC "z:**"
    THEN GEN_TAC
    THEN REWRITE_TAC[BND]
    THEN COND_CASES_TAC
    THEN REWRITE_TAC[]
   );;

% BND_OVERRIDE:
g "!(n:*) m (x:**) s. BND n x (BND m x s) = BND n x s";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:**");;
e(GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;
 e(REWRITE_TAC[]);;
 e(REWRITE_TAC[]);;
%

let BND_COMMUTE = prove_thm
   (`BND_COMMUTE`,
    "!(n:*) m (x:**) y s.
      ~(y = x) ==> (BND n y (BND m x s) = BND m x (BND n y s))",
    REPEAT STRIP_TAC
    THEN EXT_TAC "z:**"
    THEN GEN_TAC
    THEN REWRITE_TAC[BND]
    THEN COND_CASES_TAC
    THEN ASM_REWRITE_TAC[]
   );;

% BND_COMMUTE:
g "!(n:*) m (x:**) y s.
    ~(y = x) ==> (BND n y (BND m x s) = BND m x (BND n y s))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:**");;
e(GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
%

let BND_I = prove_thm
   (`BND_I`,
    "!x:*. BND x x I = I",
    GEN_TAC
    THEN EXT_TAC "z:*"
    THEN GEN_TAC
    THEN REWRITE_TAC[BND;I_THM]
    THEN COND_CASES_TAC
    THENL
      [  ASM_REWRITE_TAC[];
         REFL_TAC
      ]
   );;

% BND_I
g "!x:*. BND x x I = I";;
e(GEN_TAC);;
e(EXT_TAC "y:*");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;I_THM]);;
e(COND_CASES_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(REFL_TAC);;
%

let o_BND =
 prove_thm
  (`o_BND`,
   "!y x (f:**->***) (s:*->**).
      (f o (BND y x s)) = BND (f y) x (f o s)",
   REPEAT GEN_TAC
   THEN EXT_TAC "z:*"
   THEN GEN_TAC
   THEN REWRITE_TAC[o_THM;BND]
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[]
  );;

% o_BND
g "!y x (f:**->***) (s:*->**).
      (f o (BND y x s)) = BND (f y) x (f o s)";;
e(REPEAT GEN_TAC);;
e(EXT_TAC "z:*");;
e(GEN_TAC);;
e(REWRITE_TAC[o_THM;BND]);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
%


let BND_o_ONE_ONE =
 prove_thm
  (`BND_o_ONE_ONE`,
   "!(ss:*->**) n v (s1:**->***).
      ONE_ONE ss ==>
      ((BND n (ss v) s1) o ss = BND n v (s1 o ss))",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN EXT_TAC "u:*"
   THEN GEN_TAC
   THEN ASM_REWRITE_TAC[BND;o_THM]
  );;

% BND_o_ONE_ONE
g "!(ss:*->**) n v (s1:**->***).
      ONE_ONE ss ==>
      ((BND n (ss v) s1) o ss = BND n v (s1 o ss))";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:*");;
e(GEN_TAC);;
e(ASM_REWRITE_TAC[BND;o_THM]);;
%

let BND_o_biject1 =
 prove_thm
  (`BND_o_biject1`,
   "!v n (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BND n v s1) o (\y. @z. ss z = y) =
        BND n (ss v) (s1 o (\y. @z. ss z = y)))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "u:*"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND;o_THM]
   THEN BETA_TAC
   THEN IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE
   THEN COND_CASES_TAC
   THENL
     [  POP_ASSUM REWRITE_THM;
        POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
     ]
  );;

% BND_o_biject
g "!v n (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BND n v s1) o (\y. @z. ss z = y) =
        BND n (ss v) (s1 o (\y. @z. ss z = y)))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:*");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;o_THM]);;
e(BETA_TAC);;
e(IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE);;
e(COND_CASES_TAC);;

 e(POP_ASSUM REWRITE_THM);;

 e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
%

let BND_o_biject2 =
 prove_thm
  (`BND_o_biject2`,
   "!v n (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BND n v s1) o ss =
        BND n (@z. ss z = v) (s1 o ss))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "u:*"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND;o_THM]
   THEN IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1
  );;

% BND_o_biject2
g "!v n (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BND n v s1) o ss =
        BND n (@z. ss z = v) (s1 o ss))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:*");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;o_THM]);;
e(IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1);;
%


let IMAGE_BND_ABSORPTION =
 prove_thm
  (`IMAGE_BND_ABSORPTION`,
   "!s x w (f:**->*). ~x IN s ==> (IMAGE (BND w x f) s = IMAGE f s)",
   SET_INDUCT_TAC
   THEN REWRITE_TAC[IN;IMAGE]
   THEN REWRITE_TAC[DE_MORGAN_THM]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC BND_NOT_MATCH
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% IMAGE_BND_ABSORPTION
g "!s x w (f:**->*). ~x IN s ==> (IMAGE (BND w x f) s = IMAGE f s)";;
e(SET_INDUCT_TAC);;
e(REWRITE_TAC[IN;IMAGE]);;

e(REWRITE_TAC[IN;IMAGE]);;
e(REWRITE_TAC[DE_MORGAN_THM]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC BND_NOT_MATCH);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let IMAGE_BND_DELETE =
 prove_thm
  (`IMAGE_BND_DELETE`,
   "!s w x (f:**->*). IMAGE (BND w x f) (s DELETE x) = IMAGE f (s DELETE x)",
   REPEAT GEN_TAC
   THEN MATCH_MP_TAC IMAGE_BND_ABSORPTION
   THEN REWRITE_TAC[IN_DELETE]
  );;

% IMAGE_BND_DELETE
g "!s w x (f:**->*). IMAGE (BND w x f) (s DELETE x) = IMAGE f (s DELETE x)";;
e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC IMAGE_BND_ABSORPTION);;
e(REWRITE_TAC[IN_DELETE]);;
%


% ===================================================================== %
% In order to define the semantics of procedure calls, we must define   %
% the binding of multiple values sequentially.  There are two orderings %
% for the sequencing, left-to-right or right-to-left; both are useful.  %
% ===================================================================== %

% BNDsL : (* list)  ->  (** list)  ->   (** -> *)    ->   (** -> *)
  BNDsR : (* list)  ->  (** list)  ->   (** -> *)    ->   (** -> *)
         value list    index list     old dictionary    new dictionary

     BNDsL [v1;...;vm] [x1;...;xm] s = s[v1/x1;...;vm/xm]
     BNDsR [v1;...;vm] [x1;...;xm] s = s[vm/xm;...;v1/x1]
%

let BNDsL_DEF = new_list_rec_definition
 (`BNDsL_DEF`,
  "(BNDsL NIL xs (s:** -> *) = s) /\
   (BNDsL (CONS v vs) xs s =
      ((xs = []) => s
                 |  BNDsL vs (TL xs) (BND v (HD xs) s)))");;

let BNDsR_DEF = new_list_rec_definition
 (`BNDsR_DEF`,
  "(BNDsR NIL xs (s:** -> *) = s) /\
   (BNDsR (CONS v vs) xs s =
      ((xs = []) => s
                 |  BND v (HD xs) (BNDsR vs (TL xs) s)))");;


let BNDsL = prove_thm
   (`BNDsL`,
    "(!xs s. BNDsL [] xs (s:** -> *) = s) /\
     (!vs s. BNDsL vs [] (s:** -> *) = s) /\
     (!v vs x xs s. BNDsL (CONS v vs) (CONS x xs) (s:** -> *) =
                    BNDsL vs xs (BND v x s))",
    REWRITE_TAC[BNDsL_DEF;HD;TL;NOT_CONS_NIL]
    THEN LIST_INDUCT_TAC
    THEN REWRITE_TAC[BNDsL_DEF]
   );;

%
g "(!xs s. BNDsL [] xs (s:** -> *) = s) /\
     (!vs s. BNDsL vs [] (s:** -> *) = s) /\
     (!v vs x xs s. BNDsL (CONS v vs) (CONS x xs) (s:** -> *) =
                    BNDsL vs xs (BND v x s))";;
e(REWRITE_TAC[BNDsL_DEF;HD;TL;NOT_CONS_NIL]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL_DEF]);;

 e(REWRITE_TAC[BNDsL_DEF]);;
%

let BNDsR = prove_thm
   (`BNDsR`,
    "(!xs s. BNDsR [] xs (s:** -> *) = s) /\
     (!vs s. BNDsR vs [] (s:** -> *) = s) /\
     (!v vs x xs s. BNDsR (CONS v vs) (CONS x xs) (s:** -> *) =
                    BND v x (BNDsR vs xs s))",
    REWRITE_TAC[BNDsR_DEF;HD;TL;NOT_CONS_NIL]
    THEN LIST_INDUCT_TAC
    THEN REWRITE_TAC[BNDsR_DEF]
   );;

%
g "(!xs s. BNDsR [] xs (s:** -> *) = s) /\
   (!vs s. BNDsR vs [] (s:** -> *) = s) /\
   (!v vs x xs s. BNDsR (CONS v vs) (CONS x xs) (s:** -> *) =
                  BND v x (BNDsR vs xs s))";;
e(REWRITE_TAC[BNDsR_DEF;HD;TL;NOT_CONS_NIL]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR_DEF]);;

 e(REWRITE_TAC[BNDsR_DEF]);;
%

let o_BNDsL =
 prove_thm
  (`o_BNDsL`,
   "!ys xs (f:**->***) (s:*->**).
      (f o (BNDsL ys xs s)) = BNDsL (MAP f ys) xs (f o s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN ASM_REWRITE_TAC[o_BND]
  );;

% o_BNDsL
g "!ys xs (f:**->***) (s:*->**).
      (f o (BNDsL ys xs s)) = BNDsL (MAP f ys) xs (f o s)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;
 e(ASM_REWRITE_TAC[o_BND]);;
%

let o_BNDsR =
 prove_thm
  (`o_BNDsR`,
   "!ys xs (f:**->***) (s:*->**).
      (f o (BNDsR ys xs s)) = BNDsR (MAP f ys) xs (f o s)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR;MAP]
   THEN ASM_REWRITE_TAC[o_BND]
  );;

% o_BNDsR
g "!ys xs (f:**->***) (s:*->**).
      (f o (BNDsR ys xs s)) = BNDsR (MAP f ys) xs (f o s)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;
 e(ASM_REWRITE_TAC[o_BND]);;
%



let BNDsL_o_biject =
 prove_thm
  (`BNDsL_o_biject`,
   "!vs ns (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsL ns vs s1) o ss = BNDsL ns (MAP (\y. @z. ss z = y) vs) (s1 o ss))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN RES2_THEN REWRITE_THM
   THEN IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_o_ONE_ONE
   THEN BETA_TAC
   THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE
  );;

% BNDsL_o_biject
g "!vs ns (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsL ns vs s1) o ss = BNDsL ns (MAP (\y. @z. ss z = y) vs) (s1 o ss))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;
 e(RES2_THEN REWRITE_THM);;
 e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_o_ONE_ONE);;
 e(BETA_TAC);;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;
%

let BNDsL_o_biject1 =
 prove_thm
  (`BNDsL_o_biject1`,
   "!vs ns (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsL ns vs s1) o (\y. @z. ss z = y) =
        BNDsL ns (MAP ss vs) (s1 o (\y. @z. ss z = y)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN RES2_THEN REWRITE_THM
   THEN IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_o_biject1
  );;

% BNDsL_o_biject1
g "!vs ns (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsL ns vs s1) o (\y. @z. ss z = y) =
        BNDsL ns (MAP ss vs) (s1 o (\y. @z. ss z = y)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsL;MAP]);;
 e(RES2_THEN REWRITE_THM);;
 e(IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_o_biject1);;
%




let BNDsR_o_biject =
 prove_thm
  (`BNDsR_o_biject`,
   "!vs ns (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsR ns vs s1) o ss = BNDsR ns (MAP (\y. @z. ss z = y) vs) (s1 o ss))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[BNDsR;MAP]
   THEN IMP_RES2_THEN REWRITE_THM BND_o_biject2
   THEN BETA_TAC
   THEN RES2_THEN REWRITE_THM
  );;

% BNDsR_o_biject
g "!vs ns (ss:*->**) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsR ns vs s1) o ss = BNDsR ns (MAP (\y. @z. ss z = y) vs) (s1 o ss))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;
 e(IMP_RES2_THEN REWRITE_THM BND_o_biject2);;
 e(BETA_TAC);;
 e(RES2_THEN REWRITE_THM);;
%

let BNDsR_o_biject1 =
 prove_thm
  (`BNDsR_o_biject1`,
   "!vs ns (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsR ns vs s1) o (\y. @z. ss z = y) =
        BNDsR ns (MAP ss vs) (s1 o (\y. @z. ss z = y)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[BNDsR;MAP]
   THEN IMP_RES2_THEN REWRITE_THM BND_o_biject1
   THEN RES2_THEN REWRITE_THM
  );;

% BNDsR_o_biject1
g "!vs ns (ss:**->*) (s1:**->***).
      ONE_ONE ss /\ ONTO ss ==>
      ((BNDsR ns vs s1) o (\y. @z. ss z = y) =
        BNDsR ns (MAP ss vs) (s1 o (\y. @z. ss z = y)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[BNDsR;MAP]);;
 e(IMP_RES2_THEN REWRITE_THM BND_o_biject1);;
 e(RES2_THEN REWRITE_THM);;
%


% We will want to frequently discuss the elements of a list  %
% as if they were a set; the following function translates.  %

let SET_LIST = 
    new_list_rec_definition (`SET_LIST`,
      "(SL NIL         = {}:(*)set)  /\
       (SL (CONS x xs) = x INSERT (SL xs))");;

let SL_EMPTY =
 prove_thm
  (`SL_EMPTY`,
   "!l:(*)list. (SL l = {}) = (l = [])",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST]
   THEN REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) NOT_EMPTY_INSERT;NOT_CONS_NIL]
  );;

% SL_EMPTY:
g "!l:(*)list. (SL l = {}) = (l = [])";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST]);;

  e(REWRITE_TAC[SET_LIST]);;
  e(REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) NOT_EMPTY_INSERT;NOT_CONS_NIL]);;
%

let SL_APPEND =
 prove_thm
  (`SL_APPEND`,
   "!a b : (*)list . SL (APPEND a b) = SL a UNION SL b",
   LIST_INDUCT_TAC
   THENL
     [ REWRITE_TAC[APPEND;SET_LIST;UNION];

       ASM_REWRITE_TAC[APPEND;SET_LIST;EXTENSION;IN;IN_UNION]
       THEN REPEAT GEN_TAC
       THEN EQ_TAC
       THEN STRIP_TAC
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% SL_APPEND:
g "!a b : (*)list . SL (APPEND a b) = SL a UNION SL b";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[UNION]);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(ASM_REWRITE_TAC[EXTENSION]);;
 e(REWRITE_TAC[IN;IN_UNION]);;
 e(REPEAT GEN_TAC);;
 e(EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]);;
%


let SL_MAP =
 prove_thm
  (`SL_MAP`,
   "!l (f:*->**). SL (MAP f l) = IMAGE f (SL l)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[MAP;SET_LIST;IMAGE]
  );;

% SL_MAP
g "!l (f:*->**). SL (MAP f l) = IMAGE f (SL l)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;SET_LIST;IMAGE]);;

 e(ASM_REWRITE_TAC[MAP;SET_LIST;IMAGE]);;
%


let ALL_IN_SL =
 prove_thm
  (`ALL_IN_SL`,
   "!x y P. ALL_EL P x /\ (y:*) IN SL x ==> P y",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[ALL_EL;SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[];
        RES_TAC
     ]
  );;

%
g "!x y P. ALL_EL P x /\ (y:*) IN SL x ==> P y";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[ALL_EL;SET_LIST;IN]);;

 e(REWRITE_TAC[ALL_EL;SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;

  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
%



let DISJOINT_APPEND =
 prove_thm
  (`DISJOINT_APPEND`,
   "!x y z : (*)list.
     (DISJOINT (SL (APPEND x y)) (SL z) =
      (DISJOINT (SL x) (SL z) /\ DISJOINT (SL y) (SL z))) /\
     (DISJOINT (SL x) (SL (APPEND y z)) =
      (DISJOINT (SL x) (SL y) /\ DISJOINT (SL x) (SL z)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;APPEND;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN ASM_REWRITE_TAC[CONJ_ASSOC;SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% DISJOINT_APPEND
g "!x y z : (*)list.
     (DISJOINT (SL (APPEND x y)) (SL z) =
      (DISJOINT (SL x) (SL z) /\ DISJOINT (SL y) (SL z))) /\
     (DISJOINT (SL x) (SL (APPEND y z)) =
      (DISJOINT (SL x) (SL y) /\ DISJOINT (SL x) (SL z)))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;APPEND;DISJOINT_EMPTY]);;

  e(REWRITE_TAC[SET_LIST;APPEND;DISJOINT_INSERT]);;
  e(ASM_REWRITE_TAC[CONJ_ASSOC;SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(REPEAT STRIP_TAC);;
  e(EQ_TAC
    THEN STRIP_TAC
    THEN ASM_REWRITE_TAC[]);;
%

let MAP_EXTENSION =
 prove_thm
  (`MAP_EXTENSION`,
   "!xs (f:*->**) g.
       (MAP f xs = MAP g xs) =
       (!x. x IN SL xs ==> (f x = g x))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP]
   THEN ASM_REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL
     [  ALL_TAC;
        RES_TAC;
        FIRST_ASSUM MATCH_MP_TAC;
        FIRST_ASSUM MATCH_MP_TAC
     ]
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_EXTENSION
g "!xs (f:*->**) g.
       (MAP f xs = MAP g xs) =
       (!x. x IN SL xs ==> (f x = g x))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;
  e(ASM_REWRITE_TAC[CONS_11]);;
  e(REPEAT GEN_TAC);;
  e(EQ_TAC THEN REPEAT STRIP_TAC);;

    e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%

let MAP_EXTENSION_SUBSET =
 prove_thm
  (`MAP_EXTENSION_SUBSET`,
   "!xs s (f:*->**) g.
     (MAP f xs = MAP g xs) /\ s SUBSET SL xs ==>
     (!x. x IN s ==> (f x = g x))",
   REWRITE_TAC[SUBSET_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC MAP_EXTENSION
  );;

% MAP_EXTENSION_SUBSET
g "!xs (f:*->**) g.
     (MAP f xs = MAP g xs) /\ s SUBSET SL xs ==>
     (!x. x IN s ==> (f x = g x))";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC MAP_EXTENSION);;
%



let SET_LIST_DECOMPOSITION =
 prove_thm
  (`SET_LIST_DECOMPOSITION`,
   "!l (e:*).
       (e IN SL l) ==>
       (?l'. ((SL l) = INSERT e (SL l')) /\ (LENGTH l = SUC(LENGTH l')))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;LENGTH]
   THEN REWRITE_TAC[ADD1;EQ_MONO_ADD_EQ]
   THEN REPEAT STRIP_TAC
   THENL
    [ EXISTS_TAC "l:(*)list"
      THEN ASM_REWRITE_TAC[];

      RES_TAC
      THEN EXISTS_TAC "(CONS (h:*) l')"
      THEN ASM_REWRITE_TAC[SET_LIST;LENGTH]
      THEN REWRITE_TAC[EXTENSION;IN]
      THEN GEN_TAC
      THEN EQ_TAC
      THEN STRIP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% SET_LIST_DECOMPOSITION
g "!l (e:*).
       (e IN SL l) ==>
       (?l'. ((SL l) = INSERT e (SL l')) /\ (LENGTH l = SUC(LENGTH l')))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN;LENGTH]);;
 e(REWRITE_TAC[ADD1;EQ_MONO_ADD_EQ]);;
 e(REPEAT STRIP_TAC);;

  e(EXISTS_TAC "l:(*)list");;
  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
  e(EXISTS_TAC "(CONS (h:*) l')");;
  e(ASM_REWRITE_TAC[SET_LIST;LENGTH]);;
  e(REWRITE_TAC[EXTENSION;IN]);;
  e(GEN_TAC);;
  e(EQ_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
%



% --------------------------------------------------------------- %
% To prove some theorems, we will sometimes need to restrict the  %
% lists of variables to be mutually disjoint, that is, containing %
% no duplicate variables.                                         %
% --------------------------------------------------------------- %

let DISJOINT_LIST = 
    new_list_rec_definition (`DISJOINT_LIST`,
      "(DL (NIL) = T)  /\
       (DL (CONS (x:*) l) = (~x IN SL l) /\ DL l)");;


let DL_MAP =
 prove_thm
  (`DL_MAP`,
  "!xs (f:*->**). DL (MAP f xs) ==> DL xs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;DISJOINT_LIST]
   THEN REPEAT STRIP_TAC
   THENL
     [  FIRST_ASSUM MATCH_MP_TAC
        THEN REWRITE_TAC[SL_MAP;IN_IMAGE]
        THEN EXISTS_TAC "h:*"
        THEN ASM_REWRITE_TAC[];

        RES_THEN REWRITE_THM
     ]
  );;

% DL_MAP
g "!xs (f:*->**). DL (MAP f xs) ==> DL xs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;DISJOINT_LIST]);;

 e(REWRITE_TAC[MAP;DISJOINT_LIST]);;
 e(REPEAT STRIP_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
  e(EXISTS_TAC "h:*");;
  e(ASM_REWRITE_TAC[]);;

  e(RES_THEN REWRITE_THM);;
%


let DL_MAP_ONE_ONE =
 prove_thm
  (`DL_MAP_ONE_ONE`,
  "!xs (f:*->**). ONE_ONE f ==> (DL (MAP f xs) = DL xs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;DISJOINT_LIST]
   THEN REPEAT STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN REWRITE_ALL_TAC[ONE_ONE]
   THEN REWRITE_TAC[SL_MAP;IN_IMAGE]
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THENL
     [  EXISTS_TAC "h:*"
        THEN ASM_REWRITE_TAC[];

        FIRST_ASSUM (REWRITE_THM o SYM o SYM)
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% DL_MAP_ONE_ONE
g "!xs (f:*->**). ONE_ONE f ==> (DL (MAP f xs) = DL xs)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;DISJOINT_LIST]);;

 e(REWRITE_TAC[MAP;DISJOINT_LIST]);;
 e(REPEAT STRIP_TAC);;
 e(RES_THEN REWRITE_THM);;
 e(REWRITE_ALL_TAC[ONE_ONE]);;
 e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
 e(EQ_TAC);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "h:*");;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
  e(ASM_REWRITE_TAC[]);;
%


let DL_APPEND =
 prove_thm
  (`DL_APPEND`,
   "!x y:(*)list. DL (APPEND x y) = (DL x /\ DL y /\ (DISJOINT (SL x) (SL y)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;APPEND]
   THEN REWRITE_TAC[DISJOINT_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!x y:(*)list. DL (APPEND x y) = (DL x /\ DL y /\ (DISJOINT (SL x) (SL y)))";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;APPEND]);;
e(REWRITE_TAC[DISJOINT_LIST;DISJOINT_EMPTY]);;

e(REWRITE_TAC[SET_LIST;APPEND]);;
e(REWRITE_TAC[DISJOINT_LIST;DISJOINT_INSERT]);;
e(ASM_REWRITE_TAC[SL_APPEND]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let DL_APPEND_IMP_DISJOINT =
 prove_thm
  (`DL_APPEND_IMP_DISJOINT`,
   "!xs ys : (*)list.
      DL (APPEND xs ys) ==>
      DISJOINT (SL xs) (SL ys)",
   REWRITE_TAC[DL_APPEND]
   THEN REPEAT STRIP_TAC
  );;

% DL_APPEND_IMP_DISJOINT
g "!xs ys : (*)list.
      DL (APPEND xs ys) ==>
      DISJOINT (SL xs) (SL ys)";;

e(REWRITE_TAC[DL_APPEND]);;
e(REPEAT STRIP_TAC);;
%



let IN_DISJOINT_LIST =
 prove_thm
  (`IN_DISJOINT_LIST`,
   "!u v (z:**).
     DL (APPEND u v)
     ==>
     ((z IN SL u  ==>  ~(z IN SL v)) /\
      (z IN SL v  ==>  ~(z IN SL u))) ",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;SET_LIST;IN]
   THEN REWRITE_TAC[DISJOINT_LIST;SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN CONJ_TAC
   THEN STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[];

        RES_TAC;

        RES_TAC
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC IN_NOT_IN
        THEN EXISTS_TAC "SL (v:(**)list)"
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!u v (z:**).
     DL (APPEND u v)
     ==>
     ((z IN SL u  ==>  ~(z IN SL v)) /\
      (z IN SL v  ==>  ~(z IN SL u))) ";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[IN]);;

 e(REWRITE_TAC[APPEND;SET_LIST;IN]);;
 e(REWRITE_TAC[DISJOINT_LIST;SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(CONJ_TAC THEN STRIP_TAC);;

   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;

   e(RES_TAC);;
   e(ASM_REWRITE_TAC[]);;
   e(MATCH_MP_TAC IN_NOT_IN);;
   e(EXISTS_TAC "SL (v:(**)list)");;
   e(ASM_REWRITE_TAC[]);;
%

let IN_DISJOINT_LIST3 =
 prove_thm
  (`IN_DISJOINT_LIST3`,
   "!u v w (z:**).
     DL (APPEND (APPEND u v) w)
     ==>
     ((z IN SL u  ==>  ~(z IN SL v) /\ ~(z IN SL w)) /\
      (z IN SL v  ==>  ~(z IN SL u) /\ ~(z IN SL w)) /\
      (z IN SL w  ==>  ~(z IN SL u) /\ ~(z IN SL v))) ",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;SET_LIST;IN]
   THEN REWRITE_TAC[IN_DISJOINT_LIST]
   THEN REWRITE_TAC[DISJOINT_LIST;SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN CONJ_TAC
   THENL
     [  STRIP_TAC
        THENL [ ALL_TAC; RES_TAC ]
        THEN ASM_REWRITE_TAC[];

        CONJ_TAC
        THEN STRIP_TAC
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC IN_NOT_IN
        THENL [ EXISTS_TAC "SL (v:(**)list)";
                EXISTS_TAC "SL (w:(**)list)" ]
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% IN_DISJOINT_LIST3
g "!u v w (z:**).
     DL (APPEND (APPEND u v) w)
     ==>
     ((z IN SL u  ==>  ~(z IN SL v) /\ ~(z IN SL w)) /\
      (z IN SL v  ==>  ~(z IN SL u) /\ ~(z IN SL w)) /\
      (z IN SL w  ==>  ~(z IN SL u) /\ ~(z IN SL v))) ";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[IN]);;
 e(REWRITE_TAC[IN_DISJOINT_LIST]);;

 e(REWRITE_TAC[APPEND;SET_LIST;IN]);;
 e(REWRITE_TAC[IN_DISJOINT_LIST]);;
 e(REWRITE_TAC[DISJOINT_LIST;SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(CONJ_TAC);;

   e(STRIP_TAC);;

     e(ASM_REWRITE_TAC[]);;

     e(RES_TAC);;
     e(ASM_REWRITE_TAC[]);;

   e(CONJ_TAC);;

     e(STRIP_TAC);;
     e(RES_TAC);;
     e(ASM_REWRITE_TAC[]);;
     e(MATCH_MP_TAC IN_NOT_IN);;
     e(EXISTS_TAC "SL (v:(**)list)");;
     e(ASM_REWRITE_TAC[]);;

     e(STRIP_TAC);;
     e(RES_TAC);;
     e(ASM_REWRITE_TAC[]);;
     e(MATCH_MP_TAC IN_NOT_IN);;
     e(EXISTS_TAC "SL (w:(**)list)");;
     e(ASM_REWRITE_TAC[]);;
%


let DL_SUBSET_LENGTH_EQ =
 prove_thm
  (`DL_SUBSET_LENGTH_EQ`,
   "!l1 l2:(*)list.
       DL l1 /\ (SL l1 SUBSET SL l2) /\ 
       (LENGTH l1 = LENGTH l2) ==>
       (SL l1 = SL l2)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]
   THENL
    [ ONCE_REWRITE_TAC[EQ_SYM_EQ]
      THEN REWRITE_TAC[LENGTH_NIL]
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[SET_LIST];

      REPEAT STRIP_TAC
      THEN IMP_RES_TAC SET_LIST_DECOMPOSITION
      THEN UNDISCH_TAC "SUC(LENGTH (l1:(*)list)) = LENGTH (l2:(*)list)"
      THEN POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[ADD1;EQ_MONO_ADD_EQ]
      THEN DISCH_TAC
      THEN UNDISCH_TAC "(SL (l1:(*)list)) SUBSET (SL (l2:(*)list))"
      THEN ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]
      THEN IMP_RES_THEN REWRITE_THM DELETE_NON_ELEMENT
      THEN DISCH_TAC
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% DL_SUBSET_LENGTH_EQ
g "!l1 l2:(*)list.
       DL l1 /\ (SL l1 SUBSET SL l2) /\ 
       (LENGTH l1 = LENGTH l2) ==>
       (SL l1 = SL l2)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]);;
 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(REWRITE_TAC[LENGTH_NIL]);;
 e(REPEAT STRIP_TAC);;
 e(ASM_REWRITE_TAC[SET_LIST]);;

 e(REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC SET_LIST_DECOMPOSITION);;
 e(UNDISCH_TAC "SUC(LENGTH (l1:(*)list)) = LENGTH (l2:(*)list)");;
 e(POP_ASSUM REWRITE_THM);;
 e(REWRITE_TAC[ADD1;EQ_MONO_ADD_EQ]);;
 e(DISCH_TAC);;
 e(UNDISCH_TAC "(SL (l1:(*)list)) SUBSET (SL (l2:(*)list))");;
 e(ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]);;
 e(IMP_RES_THEN REWRITE_THM DELETE_NON_ELEMENT);;
 e(DISCH_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let DL_SUBSET_LESS_EQ_LENGTH =
 prove_thm
  (`DL_SUBSET_LESS_EQ_LENGTH`,
   "!l1 l2:(*)list.
       DL l1 /\ (SL l1 SUBSET SL l2) ==> (LENGTH l1 <= LENGTH l2)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]
   THENL
    [ REWRITE_TAC[LESS_OR_EQ]
      THEN ONCE_REWRITE_TAC[DISJ_SYM]
      THEN REWRITE_TAC[LESS_0_CASES];

      REPEAT STRIP_TAC
      THEN IMP_RES_TAC SET_LIST_DECOMPOSITION
      THEN UNDISCH_TAC "(SL (l1:(*)list)) SUBSET (SL (l2:(*)list))"
      THEN ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]
      THEN IMP_RES_THEN REWRITE_THM DELETE_NON_ELEMENT
      THEN DISCH_TAC
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[LESS_EQ_MONO]
    ]
  );;

% DL_SUBSET_LESS_EQ_LENGTH
g "!l1 l2:(*)list.
       DL l1 /\ (SL l1 SUBSET SL l2) ==> (LENGTH l1 <= LENGTH l2)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]);;
 e(REWRITE_TAC[LESS_OR_EQ]);;
 e(ONCE_REWRITE_TAC[DISJ_SYM]);;
 e(REWRITE_TAC[LESS_0_CASES]);;

 e(REWRITE_TAC[DISJOINT_LIST;SET_LIST;SUBSET;LENGTH]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC SET_LIST_DECOMPOSITION);;
 e(UNDISCH_TAC "(SL (l1:(*)list)) SUBSET (SL (l2:(*)list))");;
 e(ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]);;
 e(IMP_RES_THEN REWRITE_THM DELETE_NON_ELEMENT);;
 e(DISCH_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[LESS_EQ_MONO]);;
%


let MAP_BND_ABSORPTION =
 prove_thm
  (`MAP_BND_ABSORPTION`,
   "!xs y v (s:**->*).
        (~y IN SL xs) ==>
        ((MAP (BND v y s) xs) = (MAP s xs))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;MAP;IN;DE_MORGAN_THM;CONS_11]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN FIRST_ASSUM (ASSUME_TAC o GEN_ALL o MATCH_MP BND_NOT_MATCH o NOT_EQ_SYM)
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_BND_ABSORPTION
g "!xs y v (s:**->*).
        (~y IN SL xs) ==>
        ((MAP (BND v y s) xs) = (MAP s xs))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[SET_LIST;MAP;IN;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(REWRITE_TAC[CONS_11]);;
 e(RES_TAC);;
 e(FIRST_ASSUM (ASSUME_TAC o GEN_ALL o MATCH_MP BND_NOT_MATCH o NOT_EQ_SYM));;
 e(ASM_REWRITE_TAC[]);;
%


let BNDsL_NOT_MATCH =
 prove_thm
  (`BNDsL_NOT_MATCH`,
   "!ys ns x (s:**->*). ~(x IN SL ys) ==>
        (BNDsL ns ys s x = s x)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL]
   THEN ASM_REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BND_NOT_MATCH
   THEN ASM_REWRITE_TAC[]
  );;

% BNDsL_NOT_MATCH:
g "!ys ns x (s:**->*). ~(x IN SL ys) ==>
        (BNDsL ns ys s x = s x)";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;
e(ASM_REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC BND_NOT_MATCH);;
e(ASM_REWRITE_TAC[]);;
set_flag(`show_types`,false);;
%

let BNDsR_NOT_MATCH =
 prove_thm
  (`BNDsR_NOT_MATCH`,
   "!ys ns x (s:**->*). ~(x IN SL ys) ==>
        (BNDsR ns ys s x = s x)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR]
   THEN REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BND_NOT_MATCH
   THEN ASM_REWRITE_TAC[]
  );;

% BNDsR_NOT_MATCH:
g "!ys ns x (s:**->*). ~(x IN SL ys) ==>
        (BNDsR ns ys s x = s x)";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;
e(REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC BND_NOT_MATCH);;
e(ASM_REWRITE_TAC[]);;
%


let BNDsL_MAP_MATCH =
 prove_thm
  (`BNDsL_MAP_MATCH`,
   "!ys (f:**->*) x s. (x IN SL ys) ==>
        (BNDsL (MAP f ys) ys s x = f x)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[MAP;BNDsL]
   THEN REPEAT STRIP_TAC
   THENL
     [ ASM_CASES_TAC "(x:**) IN (SL ys)"
       THENL
         [ RES_THEN REWRITE_THM;

           IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
           THEN ASM_REWRITE_TAC[BND_MATCH]
         ];

      RES_THEN REWRITE_THM
    ]
  );;

% BNDsL_MAP_MATCH:
g "!ys (f:**->*) x s. (x IN SL ys) ==>
        (BNDsL (MAP f ys) ys s x = f x)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN]);;

e(REWRITE_TAC[SET_LIST;IN]);;
e(REWRITE_TAC[MAP;BNDsL]);;
e(REPEAT STRIP_TAC);;

 e(ASM_CASES_TAC "(x:**) IN (SL ys)");;

  e(RES_THEN REWRITE_THM);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

 e(RES_THEN REWRITE_THM);;
%

let BNDsR_MAP_MATCH =
 prove_thm
  (`BNDsR_MAP_MATCH`,
   "!ys (f:**->*) x s. (x IN SL ys) ==>
        (BNDsR (MAP f ys) ys s x = f x)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[MAP;BNDsR]
   THEN REPEAT STRIP_TAC
   THENL
     [ ASM_REWRITE_TAC[BND_MATCH];

       ASM_CASES_TAC "(x:**) = h"
       THENL
         [ ASM_REWRITE_TAC[BND_MATCH];

           IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
           THEN RES_THEN REWRITE_THM
         ]
    ]
  );;

% BNDsR_MAP_MATCH:
g "!ys (f:**->*) x s. (x IN SL ys) ==>
        (BNDsR (MAP f ys) ys s x = f x)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN]);;

e(REWRITE_TAC[SET_LIST;IN]);;
e(REWRITE_TAC[MAP;BNDsR]);;
e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[BND_MATCH]);;

 e(ASM_CASES_TAC "(x:**) = h");;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(RES_THEN REWRITE_THM);;
%


let BND_BNDsL_COMMUTE =
 prove_thm
  (`BND_BNDsL_COMMUTE`,
   "!ys ns (m:*) (x:**) s. ~(x IN SL ys) ==>
        (BND m x (BNDsL ns ys s) = BNDsL ns ys (BND m x s))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL]
   THEN REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

% BND_BNDsL_COMMUTE:
g "!ys ns (m:*) (x:**) s. ~(x IN SL ys) ==>
        (BND m x (BNDsL ns ys s) = BNDsL ns ys (BND m x s))";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;
e(REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC BND_COMMUTE);;
e(ASM_REWRITE_TAC[]);;
%

let BND_BNDsR_COMMUTE =
 prove_thm
  (`BND_BNDsR_COMMUTE`,
   "!ys ns (m:*) (x:**) s. ~(x IN SL ys) ==>
        (BND m x (BNDsR ns ys s) = BNDsR ns ys (BND m x s))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR]
   THEN REWRITE_TAC[SET_LIST;HD;TL;IN;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

% BND_BNDsR_COMMUTE:
g "!ys ns (m:*) (x:**) s. ~(x IN SL ys) ==>
        (BND m x (BNDsR ns ys s) = BNDsR ns ys (BND m x s))";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;
e(REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC BND_COMMUTE);;
e(ASM_REWRITE_TAC[]);;
%



let BNDsL_BNDsL_COMMUTE =
 prove_thm
  (`BNDsL_BNDsL_COMMUTE`,
   "!xs ms ys ns (s:**->*).
      DISJOINT (SL xs) (SL ys)
      ==>
      (BNDsL ms xs (BNDsL ns ys s)  =  BNDsL ns ys (BNDsL ms xs s))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;BNDsL]
   THEN REWRITE_TAC[DISJOINT_INSERT]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE
   THEN RES_THEN REWRITE_THM
  );;

% BNDsL_BNDsL_COMMUTE
g "!xs ms ys ns (s:**->*).
      DISJOINT (SL xs) (SL ys)
      ==>
      (BNDsL ms xs (BNDsL ns ys s)  =  BNDsL ns ys (BNDsL ms xs s))";;

e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;BNDsL]);;

  e(REWRITE_TAC[SET_LIST;BNDsL]);;

  e(REWRITE_TAC[SET_LIST;BNDsL]);;
  e(REWRITE_TAC[DISJOINT_INSERT]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE);;
  e(RES_THEN REWRITE_THM);;
%

let BNDsR_BNDsR_COMMUTE =
 prove_thm
  (`BNDsR_BNDsR_COMMUTE`,
   "!xs ms ys ns (s:**->*).
      DISJOINT (SL xs) (SL ys)
      ==>
      (BNDsR ms xs (BNDsR ns ys s)  =  BNDsR ns ys (BNDsR ms xs s))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;BNDsR]
   THEN REWRITE_TAC[DISJOINT_INSERT]
   THEN REPEAT STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE
  );;

% BNDsR_BNDsR_COMMUTE
g "!xs ms ys ns (s:**->*).
      DISJOINT (SL xs) (SL ys)
      ==>
      (BNDsR ms xs (BNDsR ns ys s)  =  BNDsR ns ys (BNDsR ms xs s))";;

e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;BNDsR]);;

  e(REWRITE_TAC[SET_LIST;BNDsR]);;

  e(REWRITE_TAC[SET_LIST;BNDsR]);;
  e(REWRITE_TAC[DISJOINT_INSERT]);;
  e(REPEAT STRIP_TAC);;
  e(RES_THEN REWRITE_THM);;
  e(IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE);;
%


let BNDsL_BNDsR_COMMUTE =
 prove_thm
  (`BNDsL_BNDsR_COMMUTE`,
   "!u x v y (s:**->*).
        (DISJOINT(SL x)(SL y)) ==>
        (BNDsL u x (BNDsR v y s) = BNDsR v y (BNDsL u x s))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL]
   THEN REWRITE_TAC[SET_LIST;DISJOINT_INSERT]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC BND_BNDsR_COMMUTE
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% BNDsL_BNDsR_COMMUTE:
g "!u x v y (s:**->*).
        (DISJOINT(SL x)(SL y)) ==>
        (BNDsL u x (BNDsR v y s) = BNDsR v y (BNDsL u x s))";;

e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL]);;

 e(REWRITE_TAC[BNDsL]);;

 e(REWRITE_TAC[BNDsL]);;
 e(REWRITE_TAC[SET_LIST;DISJOINT_INSERT]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC BND_BNDsR_COMMUTE);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let BNDsR_BNDsL_COMMUTE =
 prove_thm
  (`BNDsR_BNDsL_COMMUTE`,
   "!u x v y (s:**->*).
        (DISJOINT(SL x)(SL y)) ==>
        (BNDsR u x (BNDsL v y s) = BNDsL v y (BNDsR u x s))",
   ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE
  );;

% BNDsR_BNDsL_COMMUTE:
g "!u x v y (s:**->*).
        (DISJOINT(SL x)(SL y)) ==>
        (BNDsR u x (BNDsL v y s) = BNDsL v y (BNDsR u x s))";;

e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;
%


let BNDsL_BND_ABSORPTION =
 prove_thm
  (`BNDsL_BND_ABSORPTION`,
   "!xs ys (m:*) (x:**) s.
        (LENGTH ys = LENGTH xs) ==> (x IN SL xs) ==>
        (BNDsL ys xs (BND m x s) = BNDsL ys xs s)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;BNDsL]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
     [ ASM_REWRITE_TAC[BND_OVERRIDE];

       RES_TAC
       THEN ASM_CASES_TAC "h = (x:**)"
       THENL
         [ ASM_REWRITE_TAC[BND_OVERRIDE];

           IMP_RES_TAC BND_COMMUTE
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

% BNDsL_BND_ABSORPTION:
g "!xs ys (m:*) (x:**) s.
        (LENGTH ys = LENGTH xs) ==> (x IN SL xs) ==>
        (BNDsL ys xs (BND m x s) = BNDsL ys xs s)";;
e(LENGTH_LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN;BNDsL]);;

e(REWRITE_TAC[SET_LIST;IN;BNDsL]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;

  e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

  e(RES_TAC);;
  e(ASM_CASES_TAC "h = (x:**)");;

    e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

    e(IMP_RES_TAC BND_COMMUTE);;
    e(ASM_REWRITE_TAC[]);;
%

let BNDsR_BND_ABSORPTION =
 prove_thm
  (`BNDsR_BND_ABSORPTION`,
   "!ys xs (m:*) (x:**) s.
        (LENGTH xs = LENGTH ys) ==> (x IN SL xs) ==>
        (BNDsR ys xs (BND m x s) = BNDsR ys xs s)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsR]
   THEN STRIP_TAC
   THENL
     [ ASM_CASES_TAC "(x:**) IN SL l'"
       THENL
         [ RES_TAC
           THEN ASM_REWRITE_TAC[];

           IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsR_COMMUTE
           THEN ASM_REWRITE_TAC[BND_OVERRIDE]
         ];

       RES_TAC
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% BNDsR_BND_ABSORPTION:
g "!ys xs (m:*) (x:**) s.
        (LENGTH xs = LENGTH ys) ==> (x IN SL xs) ==>
        (BNDsR ys xs (BND m x s) = BNDsR ys xs s)";;
e(LENGTH_LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN]);;

e(REWRITE_TAC[SET_LIST;IN]);;
e(REWRITE_TAC[BNDsR]);;
e(STRIP_TAC);;

  e(ASM_CASES_TAC "(x:**) IN SL l'");;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsR_COMMUTE);;
    e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let BNDsL_BND_CASES =
 prove_thm
  (`BNDsL_BND_CASES`,
   "!xs ys (m:*) (x:**) s.
        ((BNDsL ys xs (BND m x s) = BNDsL ys xs s) \/
         (BNDsL ys xs (BND m x s) = BND m x (BNDsL ys xs s)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(h:**) = x"
   THENL
     [ ASM_REWRITE_TAC[BND_OVERRIDE];

       IMP_RES_THEN REWRITE_THM BND_COMMUTE
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% BNDsL_BND_CASES:
g "!xs ys (m:*) (x:**) s.
        ((BNDsL ys xs (BND m x s) = BNDsL ys xs s) \/
         (BNDsL ys xs (BND m x s) = BND m x (BNDsL ys xs s)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;

e(REWRITE_TAC[BNDsL]);;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(h:**) = x");;

  e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

  e(IMP_RES_THEN REWRITE_THM BND_COMMUTE);;
  e(ASM_REWRITE_TAC[]);;
%

let BNDsR_BND_CASES =
 prove_thm
  (`BNDsR_BND_CASES`,
   "!xs ys (m:*) (x:**) s.
        ((BNDsR ys xs (BND m x s) = BNDsR ys xs s) \/
         (BNDsR ys xs (BND m x s) = BND m x (BNDsR ys xs s)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR]
   THEN REPEAT GEN_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
   THEN ASM_CASES_TAC "(h:**) = x"
   THENL
     [ ASM_REWRITE_TAC[BND_OVERRIDE];

       IMP_RES_THEN REWRITE_THM BND_COMMUTE
     ]
  );;

% BNDsR_BND_CASES:
g "!xs ys (m:*) (x:**) s.
        ((BNDsR ys xs (BND m x s) = BNDsR ys xs s) \/
         (BNDsR ys xs (BND m x s) = BND m x (BNDsR ys xs s)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;

e(REWRITE_TAC[BNDsR]);;
e(REPEAT GEN_TAC);;
e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL));;

  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(ASM_CASES_TAC "(h:**) = x");;

   e(ASM_REWRITE_TAC[BND_OVERRIDE]);;

   e(IMP_RES_THEN REWRITE_THM BND_COMMUTE);;
%


let MAP_BNDsL_IDENT =
 prove_thm
  (`MAP_BNDsL_IDENT`,
   "!ns xs (s:**->*). (LENGTH xs = LENGTH ns) ==> DL xs ==>
        (MAP (BNDsL ns xs s) xs = ns)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_LIST;BNDsL;MAP]
   THEN STRIP_TAC
   THEN IMP_RES_TAC BNDsL_NOT_MATCH
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[CONS_11;BND_MATCH]
  );;

% MAP_BNDsL_IDENT
g "!ns xs (s:**->*). (LENGTH xs = LENGTH ns) ==> DL xs ==>
        (MAP (BNDsL ns xs s) xs = ns)";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[DISJOINT_LIST;BNDsL;MAP]);;
 e(STRIP_TAC);;
 e(REWRITE_TAC[CONS_11]);;
 e(IMP_RES_TAC BNDsL_NOT_MATCH);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;
%

let MAP_BNDsR_IDENT =
 prove_thm
  (`MAP_BNDsR_IDENT`,
   "!ns xs (s:**->*). (LENGTH xs = LENGTH ns) ==> DL xs ==>
        (MAP (BNDsR ns xs s) xs = ns)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_LIST;BNDsR;MAP]
   THEN STRIP_TAC
   THEN IMP_RES_TAC MAP_BND_ABSORPTION
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[CONS_11;BND_MATCH]
  );;

% MAP_BNDsR_IDENT
g "!ns xs (s:**->*). (LENGTH xs = LENGTH ns) ==> DL xs ==>
        (MAP (BNDsR ns xs s) xs = ns)";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[DISJOINT_LIST;BNDsR;MAP]);;
 e(STRIP_TAC);;
 e(REWRITE_TAC[CONS_11]);;
 e(REWRITE_TAC[BND_MATCH]);;
 e(IMP_RES_TAC MAP_BND_ABSORPTION);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let MAP_BNDsL_NOT_MATCH =
 prove_thm
  (`MAP_BNDsL_NOT_MATCH`,
   "!ys xs (ss:**->*).
              (DISJOINT (SL zs) (SL xs)) ==>
              (MAP (BNDsL ys xs ss) zs = MAP ss zs)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL]
   THEN REWRITE_TAC[SET_LIST;DISJOINT_INSERT2]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC MAP_BND_ABSORPTION
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_BNDsL_NOT_MATCH:
g "!ys xs (ss:**->*).
              (DISJOINT (SL zs) (SL xs)) ==>
              (MAP (BNDsL ys xs ss) zs = MAP ss zs)";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[BNDsL]);;

  e(REWRITE_TAC[BNDsL]);;

  e(REWRITE_TAC[BNDsL]);;
  e(REWRITE_TAC[SET_LIST]);;
  e(REWRITE_TAC[DISJOINT_INSERT2]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(IMP_RES_TAC MAP_BND_ABSORPTION);;
  e(ASM_REWRITE_TAC[]);;
%

let MAP_BNDsR_NOT_MATCH =
 prove_thm
  (`MAP_BNDsR_NOT_MATCH`,
   "!ys xs (ss:**->*).
              (DISJOINT (SL zs) (SL xs)) ==>
              (MAP (BNDsR ys xs ss) zs = MAP ss zs)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR]
   THEN REWRITE_TAC[SET_LIST;DISJOINT_INSERT2]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC MAP_BND_ABSORPTION
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_BNDsR_NOT_MATCH:
g "!ys xs (ss:**->*).
              (DISJOINT (SL zs) (SL xs)) ==>
              (MAP (BNDsR ys xs ss) zs = MAP ss zs)";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[BNDsR]);;

  e(REWRITE_TAC[BNDsR]);;

  e(REWRITE_TAC[BNDsR]);;
  e(REWRITE_TAC[SET_LIST]);;
  e(REWRITE_TAC[DISJOINT_INSERT2]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(IMP_RES_TAC MAP_BND_ABSORPTION);;
  e(ASM_REWRITE_TAC[]);;
%

let BNDsL_MATCH_EQ =
 prove_thm
  (`BNDsL_MATCH_EQ`,
   "!ys xs (s1:**->*) s2.
     (LENGTH xs = LENGTH ys) ==> (z IN (SL xs)) ==>
        (BNDsL ys xs s1 z = BNDsL ys xs s2 z)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsL]
   THEN STRIP_TAC
   THENL
     [  ASM_CASES_TAC "(z:**) IN SL l'"
        THENL
          [  RES_TAC
             THEN POP_ASSUM (ACCEPT_TAC o
                   SPECL ["BND h h' (s1:**->*)";"BND h h' (s2:**->*)"]);

             IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
             THEN ASM_REWRITE_TAC[BND_MATCH]
          ];

        RES_TAC
        THEN POP_ASSUM (ACCEPT_TAC o
              SPECL ["BND h h' (s1:**->*)";"BND h h' (s2:**->*)"])
     ]
  );;

% BNDsL_MATCH_EQ:
g "!ys xs (s1:**->*) s2.
     (LENGTH xs = LENGTH ys) ==> (z IN (SL xs)) ==>
        (BNDsL ys xs s1 z = BNDsL ys xs s2 z)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsL]);;
  e(STRIP_TAC);;

    e(ASM_CASES_TAC "(z:**) IN SL l'");;

      e(RES_TAC);;
      e(POP_ASSUM (ACCEPT_TAC o
                   SPECL ["BND h h' (s1:**->*)";"BND h h' (s2:**->*)"]));;

      e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
      e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(RES_TAC);;
    e(POP_ASSUM (ACCEPT_TAC o
                 SPECL ["BND h h' (s1:**->*)";"BND h h' (s2:**->*)"]));;
%

let BNDsR_MATCH_EQ =
 prove_thm
  (`BNDsR_MATCH_EQ`,
   "!ys xs (s1:**->*) s2.
     (LENGTH xs = LENGTH ys) ==> (z IN (SL xs)) ==>
        (BNDsR ys xs s1 z = BNDsR ys xs s2 z)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsR]
   THEN STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        ASM_CASES_TAC "(z:**) = h'"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% BNDsR_MATCH_EQ:
g "!ys xs (s1:**->*) s2.
     (LENGTH xs = LENGTH ys) ==> (z IN (SL xs)) ==>
        (BNDsR ys xs s1 z = BNDsR ys xs s2 z)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsR]);;
  e(STRIP_TAC);;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(ASM_CASES_TAC "(z:**) = h'");;

      e(ASM_REWRITE_TAC[BND_MATCH]);;

      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
      e(ASM_REWRITE_TAC[]);;
%


let BNDsL_MEMBER =
 prove_thm
  (`BNDsL_MEMBER`,
   "!ys xs (s:**->*) z. 
        (LENGTH xs = LENGTH ys) ==>
         (z IN SL xs) ==> (BNDsL ys xs s z IN SL ys)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsL]
   THEN REPEAT STRIP_TAC
   THENL
     [  ASM_CASES_TAC "(z:**) IN SL l'"
        THENL
          [  RES_TAC
             THEN ASM_REWRITE_TAC[];

             IMP_RES_TAC BNDsL_NOT_MATCH
             THEN ASM_REWRITE_TAC[BND_MATCH]
          ];

        RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% BNDsL_MEMBER
g "!ys xs (s:**->*) z. 
        (LENGTH xs = LENGTH ys) ==>
         (z IN SL xs) ==> (BNDsL ys xs s z IN SL ys)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsL]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_CASES_TAC "(z:**) IN SL l'");;

      e(RES_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(IMP_RES_TAC BNDsL_NOT_MATCH);;
      e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%

let BNDsR_MEMBER =
 prove_thm
  (`BNDsR_MEMBER`,
   "!ys xs (s:**->*) z. 
        (LENGTH xs = LENGTH ys) ==>
         (z IN SL xs) ==> (BNDsR ys xs s z IN SL ys)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsR]
   THEN REPEAT STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        ASM_CASES_TAC "(z:**) = h'"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH];

             IMP_RES_TAC BND_NOT_MATCH
             THEN RES_TAC
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% BNDsR_MEMBER
g "!ys xs (s:**->*) z. 
        (LENGTH xs = LENGTH ys) ==>
         (z IN SL xs) ==> (BNDsR ys xs s z IN SL ys)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsR]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(ASM_CASES_TAC "(z:**) = h'");;

      e(ASM_REWRITE_TAC[BND_MATCH]);;

      e(IMP_RES_TAC BND_NOT_MATCH);;
      e(RES_TAC);;
      e(ASM_REWRITE_TAC[]);;
%


% BNDsL_APPEND_ELEMENTS ?
g "!z w y x v s.
     (LENGTH w = LENGTH z) ==>
     v IN SL w ==>
     BNDsL (APPEND y z) (APPEND x w) (s:**->*) v IN SL z";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN]);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsL]);;
  e(ASM_CASES_TAC "(v:**) IN SL l'");;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_TAC IN_UNION);;
    e(POP_TAC);;
    e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[(SYM o SPEC_ALL) SL_APPEND]
                            o SPEC "SL (w:(**)list)"));;
    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
    e(STRIP_TAC);;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%

let BNDsR_APPEND_MEMBER =
 prove_thm
  (`BNDsR_APPEND_MEMBER`,
   "!y x z w v s.
     (LENGTH x = LENGTH y) ==>
     v IN SL x ==>
     BNDsR (APPEND y z) (APPEND x w) (s:**->*) v IN SL y",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;SET_LIST;IN]
   THEN REWRITE_TAC[BNDsR]
   THEN ASM_CASES_TAC "(v:**) = h'"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH
        THEN STRIP_TAC
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% BNDsR_APPEND_MEMBER
g "!y x z w v s.
     (LENGTH x = LENGTH y) ==>
     v IN SL x ==>
     BNDsR (APPEND y z) (APPEND x w) (s:**->*) v IN SL y";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN]);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN]);;
  e(REWRITE_TAC[BNDsR]);;
  e(ASM_CASES_TAC "(v:**) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
    e(STRIP_TAC);;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let IDENT_BND =
 prove_thm
  (`IDENT_BND`,
   "!x (s:**->*). BND (s x) x s = s",
   REPEAT STRIP_TAC
   THEN EXT_TAC "y:**"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND]
   THEN COND_CASES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% IDENT_BND
g "!x (s:**->*). BND (s x) x s = s";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "y:**");;
e(GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;

e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[]);;
%

let IDENT_BNDsL =
 prove_thm
  (`IDENT_BNDsL`,
   "!x (s:**->*). BNDsL (MAP s x) x s = s",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;BNDsL]
   THEN ASM_REWRITE_TAC[CONS_11;IDENT_BND]
  );;

let IDENT_BNDsR =
 prove_thm
  (`IDENT_BNDsR`,
   "!x (s:**->*). BNDsR (MAP s x) x s = s",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;BNDsR]
   THEN ASM_REWRITE_TAC[CONS_11;IDENT_BND]
  );;

% IDENT_BNDsL
g "!x (s:**->*). BNDsL (MAP s x) x s = s";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP;BNDsL]);;

e(REWRITE_TAC[MAP;BNDsL]);;
e(ASM_REWRITE_TAC[CONS_11;IDENT_BND]);;
%

let BNDsL_MAP_BND =
 prove_thm
  (`BNDsL_MAP_BND`,
   "!x (s:**->*) n y.
     y IN SL x ==>
     (BNDsL (MAP (BND n y s) x) x s = BND n y s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP;BNDsL]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(h:**) = y"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH;IDENT_BNDsL];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
        THEN STRIP_TAC
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[IDENT_BND]
     ]
  );;

% BNDsL_MAP_BND
g "!x (s:**->*) n y.
     y IN SL x ==>
     (BNDsL (MAP (BND n y s) x) x s = BND n y s)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN;MAP;BNDsL]);;

e(REWRITE_TAC[SET_LIST;IN;MAP;BNDsL]);;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(h:**) = y");;

  e(ASM_REWRITE_TAC[BND_MATCH;IDENT_BNDsL]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
  e(STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[IDENT_BND]);;
%

let BNDsR_MAP_BND =
 prove_thm
  (`BNDsR_MAP_BND`,
   "!x (s:**->*) n y.
     y IN SL x ==>
     (BNDsR (MAP (BND n y s) x) x s = BND n y s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP;BNDsR]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:**) IN SL x"
   THENL
     [  RES_TAC
        THEN ASM_REWRITE_TAC[IDENT_BND];

        ASM_REWRITE_TAC[]
        THEN DISCH_THEN (REWRITE_THM o SYM)
        THEN IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE
        THEN REWRITE_TAC[BND_MATCH;IDENT_BNDsR]
     ]
  );;

% BNDsR_MAP_BND
g "!x (s:**->*) n y.
     y IN SL x ==>
     (BNDsR (MAP (BND n y s) x) x s = BND n y s)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;IN;MAP;BNDsR]);;

e(REWRITE_TAC[SET_LIST;IN;MAP;BNDsR]);;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:**) IN SL x");;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[IDENT_BND]);;

  e(ASM_REWRITE_TAC[]);;
  e(DISCH_THEN (REWRITE_THM o SYM));;
  e(IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE);;
  e(REWRITE_TAC[BND_MATCH;IDENT_BNDsR]);;
%

let BNDsL_MAP_BND_COMMUTE =
 prove_thm
  (`BNDsL_MAP_BND_COMMUTE`,
   "!x (s1:**->*) s2 n y.
     y IN SL x ==>
     (BNDsL (MAP (BND n y s1) x) x s2 = BND n y (BNDsL (MAP s1 x) x s2))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:**"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:**) = y"
   THENL
     [  IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH
        THEN REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN ASM_CASES_TAC "(z:**) IN SL x"
        THENL
          [  IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
             THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH;

             IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
          ]
     ]
  );;

% BNDsL_MAP_BND_COMMUTE
g "!x (s1:**->*) s2 n y.
     y IN SL x ==>
     (BNDsL (MAP (BND n y s1) x) x s2 = BND n y (BNDsL (MAP s1 x) x s2))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:**");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:**) = y");;

  e(IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH);;
  e(REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(ASM_CASES_TAC "(z:**) IN SL x");;

    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
%

let BNDsR_MAP_BND_COMMUTE =
 prove_thm
  (`BNDsR_MAP_BND_COMMUTE`,
   "!x (s1:**->*) s2 n y.
     y IN SL x ==>
     (BNDsR (MAP (BND n y s1) x) x s2 = BND n y (BNDsR (MAP s1 x) x s2))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:**"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:**) = y"
   THENL
     [  IMP_RES_THEN ASM_REWRITE_THM BNDsR_MAP_MATCH
        THEN REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN ASM_CASES_TAC "(z:**) IN SL x"
        THENL
          [  IMP_RES_THEN REWRITE_THM BNDsR_MAP_MATCH
             THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH;

             IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
          ]
     ]
  );;

% BNDsR_MAP_BND_COMMUTE
g "!x (s1:**->*) s2 n y.
     y IN SL x ==>
     (BNDsR (MAP (BND n y s1) x) x s2 = BND n y (BNDsR (MAP s1 x) x s2))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:**");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:**) = y");;

  e(IMP_RES_THEN ASM_REWRITE_THM BNDsR_MAP_MATCH);;
  e(REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(ASM_CASES_TAC "(z:**) IN SL x");;

    e(IMP_RES_THEN REWRITE_THM BNDsR_MAP_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;

    e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
%

let BNDsL_MAP_BNDsL_COMMUTE =
 prove_thm
  (`BNDsL_MAP_BNDsL_COMMUTE`,
   "!y n x (s1:**->*) s2.
     SL y SUBSET SL x ==>
     (BNDsL (MAP (BNDsL n y s1) x) x s2 = BNDsL n y (BNDsL (MAP s1 x) x s2))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsL_MAP_BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

% BNDsL_MAP_BNDsL_COMMUTE
g "!y n x (s1:**->*) s2.
     SL y SUBSET SL x ==>
     (BNDsL (MAP (BNDsL n y s1) x) x s2 = BNDsL n y (BNDsL (MAP s1 x) x s2))";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;SUBSET;BNDsL]);;

  e(REWRITE_TAC[SET_LIST;SUBSET;BNDsL]);;

  e(REWRITE_TAC[SET_LIST;SUBSET;BNDsL]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(IMP_RES_TAC BNDsL_MAP_BND_COMMUTE);;
  e(ASM_REWRITE_TAC[]);;
%

let BNDsL_MAP_BNDsR_COMMUTE =
 prove_thm
  (`BNDsL_MAP_BNDsR_COMMUTE`,
   "!y n x (s1:**->*) s2.
     SL y SUBSET SL x ==>
     (BNDsL (MAP (BNDsR n y s1) x) x s2 = BNDsR n y (BNDsL (MAP s1 x) x s2))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsL_MAP_BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

let BNDsR_MAP_BNDsL_COMMUTE =
 prove_thm
  (`BNDsR_MAP_BNDsL_COMMUTE`,
   "!y n x (s1:**->*) s2.
     SL y SUBSET SL x ==>
     (BNDsR (MAP (BNDsL n y s1) x) x s2 = BNDsL n y (BNDsR (MAP s1 x) x s2))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsR_MAP_BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

let BNDsR_MAP_BNDsR_COMMUTE =
 prove_thm
  (`BNDsR_MAP_BNDsR_COMMUTE`,
   "!y n x (s1:**->*) s2.
     SL y SUBSET SL x ==>
     (BNDsR (MAP (BNDsR n y s1) x) x s2 = BNDsR n y (BNDsR (MAP s1 x) x s2))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsR_MAP_BND_COMMUTE
   THEN ASM_REWRITE_TAC[]
  );;

let MAP_BND_EQUAL =
 prove_thm
  (`MAP_BND_EQUAL`,
   "!x (s1:**->*) s2 n y.
     (MAP s1 x = MAP s2 x) ==>
     (MAP (BND n y s1) x = MAP (BND n y s2) x)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN ASM_CASES_TAC "(h:**) = y"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH
     ]
  );;

% MAP_BND_EQUAL
g "!x (s1:**->*) s2 n y.
     (MAP s1 x = MAP s2 x) ==>
     (MAP (BND n y s1) x = MAP (BND n y s2) x)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP]);;

e(REWRITE_TAC[MAP]);;
e(REWRITE_TAC[CONS_11]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(RES_THEN REWRITE_THM);;
e(ASM_CASES_TAC "(h:**) = y");;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
%

let BNDsL_BNDsL_ABSORPTION =
 prove_thm
  (`BNDsL_BNDsL_ABSORPTION`,
   "!ys xs zs (s:**->*).
     (LENGTH xs = LENGTH ys) ==>
     (BNDsL ys xs (BNDsL zs xs s) = BNDsL ys xs s)",
   REPEAT STRIP_TAC
   THEN EXT_TAC "v:**"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(v:**) IN SL xs"
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MATCH_EQ)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
     ]
  );;

% BNDsL_BNDsL_ABSORPTION
g "!ys xs zs (s:**->*).
     (LENGTH xs = LENGTH ys) ==>
     (BNDsL ys xs (BNDsL zs xs s) = BNDsL ys xs s)";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "v:**");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(v:**) IN SL xs");;

  e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MATCH_EQ));;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
%

let BNDsR_BNDsR_ABSORPTION =
 prove_thm
  (`BNDsR_BNDsR_ABSORPTION`,
   "!ys xs zs (s:**->*).
     (LENGTH xs = LENGTH ys) ==>
     (BNDsR ys xs (BNDsR zs xs s) = BNDsR ys xs s)",
   REPEAT STRIP_TAC
   THEN EXT_TAC "v:**"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(v:**) IN SL xs"
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MATCH_EQ)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
     ]
  );;

% BNDsR_BNDsR_ABSORPTION
g "!ys xs zs (s:**->*).
     (LENGTH xs = LENGTH ys) ==>
     (BNDsR ys xs (BNDsR zs xs s) = BNDsR ys xs s)";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "v:**");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(v:**) IN SL xs");;

  e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MATCH_EQ));;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
%

let MAP_BND_EQ_MAP =
 prove_thm
  (`MAP_BND_EQ_MAP`,
   "!x y (s1:**->*) s2.
     (MAP s2 x = MAP s1 x) ==> (MAP (BND (s1 y) y s2) x = MAP s1 x)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN ASM_CASES_TAC "(h:**) = y"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH
     ]
  );;

% MAP_BND_EQ_MAP
g "!x y (s1:**->*) s2.
     (MAP s2 x = MAP s1 x) ==> (MAP (BND (s1 y) y s2) x = MAP s1 x)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP]);;

e(REWRITE_TAC[MAP]);;
e(REWRITE_TAC[CONS_11]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(RES_THEN REWRITE_THM);;
e(ASM_CASES_TAC "(h:**) = y");;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
%

let MAP_BNDsL_MAP =
 prove_thm
  (`MAP_BNDsL_MAP`,
   "!x (s1:**->*) s2.
     MAP (BNDsL (MAP s1 x) x s2) x = MAP s1 x",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN ASM_REWRITE_TAC[BNDsL;CONS_11]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(h:**) IN SL x"
   THENL
     [  IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

        IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN REWRITE_TAC[BND_MATCH]
     ]
  );;

% MAP_BNDsL_MAP
g "!x (s1:**->*) s2.
     MAP (BNDsL (MAP s1 x) x s2) x = MAP s1 x";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP]);;

e(REWRITE_TAC[MAP]);;
e(ASM_REWRITE_TAC[BNDsL;CONS_11]);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "(h:**) IN SL x");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(REWRITE_TAC[BND_MATCH]);;
%

let MAP_BNDsR_MAP =
 prove_thm
  (`MAP_BNDsR_MAP`,
   "!x (s1:**->*) s2.
     MAP (BNDsR (MAP s1 x) x s2) x = MAP s1 x",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN ASM_REWRITE_TAC[BNDsR;CONS_11;BND_MATCH]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC MAP_BND_EQ_MAP
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_BNDsR_MAP
g "!x (s1:**->*) s2.
     MAP (BNDsR (MAP s1 x) x s2) x = MAP s1 x";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP]);;

e(REWRITE_TAC[MAP]);;
e(REWRITE_TAC[BNDsR;CONS_11;BND_MATCH]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC MAP_BND_EQ_MAP);;
e(ASM_REWRITE_TAC[]);;
%

let MAP_BNDsL_MAP_SUBSET =
 prove_thm
  (`MAP_BNDsL_MAP_SUBSET`,
   "!y x (s1:**->*) s2.
     (SL y SUBSET SL x) ==>
     (MAP (BNDsL (MAP s1 x) x s2) y = MAP s1 y)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;MAP]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsL_MAP_MATCH
   THEN ASM_REWRITE_TAC[CONS_11]
  );;

% MAP_BNDsL_MAP_SUBSET
g "!y x (s1:**->*) s2.
     (SL y SUBSET SL x) ==>
     (MAP (BNDsL (MAP s1 x) x s2) y = MAP s1 y)";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[SET_LIST;SUBSET;MAP]);;

e(REWRITE_TAC[SET_LIST;SUBSET;MAP]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC BNDsL_MAP_MATCH);;
e(ASM_REWRITE_TAC[CONS_11]);;
%

let MAP_BNDsR_MAP_SUBSET =
 prove_thm
  (`MAP_BNDsR_MAP_SUBSET`,
   "!y x (s1:**->*) s2.
     (SL y SUBSET SL x) ==>
     (MAP (BNDsR (MAP s1 x) x s2) y = MAP s1 y)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;MAP]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BNDsR_MAP_MATCH
   THEN ASM_REWRITE_TAC[CONS_11]
  );;



let MAP_IDENT_IMAGE_IDENT =
 prove_thm
  (`MAP_IDENT_IMAGE_IDENT`,
   "!(x:(*)list) (f:*->*).
         (MAP f x = x) ==> (IMAGE f (SL x) = SL x)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IMAGE]
   THEN REWRITE_TAC[MAP;CONS_11]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_IDENT_IMAGE_IDENT
g "!(x:(*)list) (f:*->*).
         (MAP f x = x) ==> (IMAGE f (SL x) = SL x)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IMAGE]);;

  e(REWRITE_TAC[SET_LIST;IMAGE]);;
  e(REWRITE_TAC[MAP;CONS_11]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%



let MAP_IDENT_ELEMENTS =
 prove_thm
  (`MAP_IDENT_ELEMENTS`,
   "!xs (f:*->*).
       (MAP f xs = xs) =
       (!x. x IN SL xs ==> (f x = x))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP]
   THEN ASM_REWRITE_TAC[CONS_11]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL
     [  ALL_TAC;
        RES_TAC;
        FIRST_ASSUM MATCH_MP_TAC;
        FIRST_ASSUM MATCH_MP_TAC
     ]
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_IDENT_ELEMENTS
g "!xs (f:*->*).
       (MAP f xs = xs) =
       (!x. x IN SL xs ==> (f x = x))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;
  e(ASM_REWRITE_TAC[CONS_11]);;
  e(REPEAT GEN_TAC);;
  e(EQ_TAC THEN REPEAT STRIP_TAC);;

    e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%

let MAP_IDENT_ELEMENTS_SUBSET =
 prove_thm
  (`MAP_IDENT_ELEMENTS_SUBSET`,
   "!xs gs (f:*->*).
      (SL gs SUBSET SL xs) /\
      (MAP f xs = xs) ==>
      (MAP f gs = gs)",
   REWRITE_TAC[SUBSET_DEF;MAP_IDENT_ELEMENTS]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

% MAP_IDENT_ELEMENTS_SUBSET:
g "!xs gs (f:*->*).
      (SL gs SUBSET SL xs) /\
      (MAP f xs = xs) ==>
      (MAP f gs = gs)";;
e(REWRITE_TAC[SUBSET_DEF;MAP_IDENT_ELEMENTS]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
%



let BNDsR_TRANS =
 prove_thm
  (`BNDsR_TRANS`,
   "!x y z (v:*).
     (LENGTH y = LENGTH x) ==> (LENGTH z = LENGTH y) /\
     DL y /\ (v IN SL y ==> v IN SL x) ==>
     (((BNDsR z y I) o (BNDsR y x I)) v = (BNDsR z x I) v)",
   REWRITE_TAC[o_THM]
   THEN LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[LENGTH;LENGTH_NIL;SET_LIST;IN]
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[BNDsR;I_THM];

      REWRITE_TAC[LENGTH;LENGTH_CONS;SET_LIST;IN]
      THEN REPEAT GEN_TAC THEN STRIP_TAC
      THEN ASM_REWRITE_TAC[LENGTH;LENGTH_CONS;SET_LIST;IN]
      THEN REWRITE_TAC[DISJOINT_LIST]
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[BNDsR]
      THEN ASM_CASES_TAC "(v:*) = h"
      THENL
       [ ASM_REWRITE_TAC[BND_MATCH];

         IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         THEN POP_ASSUM REWRITE_ALL_THM
         THEN ASM_CASES_TAC "(v:*) IN SL x"
         THENL
          [ UNASSUME_THEN (ASSUME_TAC o SYM)
                   "LENGTH (l':(*)list) = LENGTH (x:(*)list)"
            THEN IMP_RES_TAC BNDsR_MEMBER
            THEN POP_TAC
            THEN POP_ASSUM (ASSUME_TAC o SPEC "I:*->*")
            THEN IMP_RES_TAC IN_NOT_IN
            THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
            THEN POP_ASSUM (\th. REWRITE_ALL_TAC[th;DE_MORGAN_THM])
            THEN POP_ASSUM STRIP_ASSUME_TAC
            THEN REWRITE_TAC[I_THM]
            THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
            THEN REWRITE_TAC[I_THM]
          ]
       ]
    ]
  );;

% BNDsR_TRANS
g "!x y z (v:*).
     (LENGTH y = LENGTH x) ==> (LENGTH z = LENGTH y) /\
     DL y /\ (v IN SL y ==> v IN SL x) ==>
     (((BNDsR z y I) o (BNDsR y x I)) v = (BNDsR z x I) v)";;
e(REWRITE_TAC[o_THM]);;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[LENGTH;LENGTH_NIL;SET_LIST;IN]);;
  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[BNDsR;I_THM]);;

  e(REWRITE_TAC[LENGTH;LENGTH_CONS;SET_LIST;IN]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(ASM_REWRITE_TAC[LENGTH;LENGTH_CONS;SET_LIST;IN]);;
  e(REWRITE_TAC[DISJOINT_LIST]);;
  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[BNDsR]);;
  e(ASM_CASES_TAC "(v:*) = h");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(POP_ASSUM REWRITE_ALL_THM);;
    e(ASM_CASES_TAC "(v:*) IN SL x");;

      e(UNASSUME_THEN (ASSUME_TAC o SYM)
                 "LENGTH (l':(*)list) = LENGTH (x:(*)list)");;
      e(IMP_RES_TAC BNDsR_MEMBER);;
      e(POP_TAC);;
      e(POP_ASSUM (ASSUME_TAC o SPEC "I:*->*"));;
      e(IMP_RES_TAC IN_NOT_IN);;
      e(POP_TAC);;
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
      e(ASM_REWRITE_TAC[]);;

      e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
      e(POP_ASSUM (\th. REWRITE_ALL_TAC[th;DE_MORGAN_THM]));;
      e(POP_ASSUM STRIP_ASSUME_TAC);;
      e(REWRITE_TAC[I_THM]);;
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
      e(REWRITE_TAC[I_THM]);;
%

let BNDsL_MAP_BNDsL_MAP_COMMUTE =
 prove_thm
  (`BNDsL_MAP_BNDsL_MAP_COMMUTE`,
   "!y x (s1:**->*) s2 s3.
     BNDsL (MAP (BNDsL (MAP s1 y) y s2) x) x s3 =
     BNDsL (MAP (BNDsL (MAP s1 x) x s3) y) y (BNDsL (MAP s2 x) x s3)",
   REPEAT GEN_TAC
   THEN EXT_TAC "z:**"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:**) IN SL x"
   THENL
     [  IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
        THEN ASM_CASES_TAC "(z:**) IN SL y"
        THENL
          [  IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

             IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
             THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
          ];

        IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN ASM_CASES_TAC "(z:**) IN SL y"
        THENL
          [  IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
             THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH;

             IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
          ]
     ]
  );;

% BNDsL_MAP_BNDsL_MAP_COMMUTE
g "!y x (s1:**->*) s2 s3.
     BNDsL (MAP (BNDsL (MAP s1 y) y s2) x) x s3 =
     BNDsL (MAP (BNDsL (MAP s1 x) x s3) y) y (BNDsL (MAP s2 x) x s3)";;
e(REPEAT GEN_TAC);;
e(EXT_TAC "z:**");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:**) IN SL x");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
 e(ASM_CASES_TAC "(z:**) IN SL y");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(ASM_CASES_TAC "(z:**) IN SL y");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
%



let IN_MAP_ONE_ONE =
 prove_thm
  (`IN_MAP_ONE_ONE`,
   "!xs (ss:*->**) x.
        ONE_ONE ss ==>
        ((ss x) IN SL(MAP ss xs) = x IN SL xs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC ONE_ONE
   THEN ASM_REWRITE_TAC[]
  );;

% IN_MAP_ONE_ONE
g "!xs (ss:*->**) x.
        ONE_ONE ss ==>
        ((ss x) IN SL(MAP ss xs) = x IN SL xs)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;SET_LIST;IN]);;

 e(REWRITE_TAC[MAP;SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(IMP_RES_TAC ONE_ONE);;
 e(ASM_REWRITE_TAC[]);;
%



let BNDsL_MAP_ONE_ONE =
 prove_thm
  (`BNDsL_MAP_ONE_ONE`,
   "!xs ns ss (s:**->*) x.
        (LENGTH ns = LENGTH xs) ==>
        ONE_ONE ss /\ (x IN SL xs) ==>
        (BNDsL ns (MAP ss xs) s (ss x) = BNDsL ns xs s x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN REPEAT STRIP_TAC
   THENL
    [ ASM_CASES_TAC "(x:**) IN SL xs"
      THENL
       [ RES_TAC
         THEN POP_ASSUM REWRITE_THM
         THEN UNASSUME_THEN (ASSUME_TAC o SYM)
                   "LENGTH (l':(*)list) = LENGTH (xs:(**)list)"
         THEN IMP_RES_TAC BNDsL_MATCH_EQ
         THEN POP_ASSUM (REWRITE_THM o
                 SPECL["BND h'((ss:**->**) h)(s:**->*)";"BND h' h(s:**->*)"]);

         IMP_RES_TAC IN_MAP_ONE_ONE
         THEN POP_TAC
         THEN POP_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)
         THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
         THEN ASM_REWRITE_TAC[BND_MATCH]
       ];

      RES_TAC
      THEN POP_ASSUM REWRITE_THM
      THEN UNASSUME_THEN (ASSUME_TAC o SYM)
                "LENGTH (l':(*)list) = LENGTH (xs:(**)list)"
      THEN IMP_RES_TAC BNDsL_MATCH_EQ
      THEN POP_ASSUM (REWRITE_THM o
              SPECL["BND h'((ss:**->**) h)(s:**->*)";"BND h' h(s:**->*)"])
    ]
  );;

% BNDsL_MAP_ONE_ONE
g "!xs ns ss (s:**->*) x.
        (LENGTH ns = LENGTH xs) ==>
        ONE_ONE ss /\ (x IN SL xs) ==>
        (BNDsL ns (MAP ss xs) s (ss x) = BNDsL ns xs s x)";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REWRITE_TAC[BNDsL;MAP]);;
 e(REPEAT STRIP_TAC);;

  e(ASM_CASES_TAC "(x:**) IN SL xs");;

   e(RES_TAC);;
   e(POP_ASSUM REWRITE_THM);;
   e(UNASSUME_THEN (ASSUME_TAC o SYM)
           "LENGTH (l':(*)list) = LENGTH (xs:(**)list)");;
   e(IMP_RES_TAC BNDsL_MATCH_EQ);;
   e(POP_ASSUM (REWRITE_THM o
        SPECL["BND h'((ss:**->**) h)(s:**->*)";"BND h' h(s:**->*)"]));;

   e(IMP_RES_TAC IN_MAP_ONE_ONE);;
   e(POP_TAC);;
   e(POP_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC_ALL));;
   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(RES_TAC);;
  e(POP_ASSUM REWRITE_THM);;
  e(UNASSUME_THEN (ASSUME_TAC o SYM)
           "LENGTH (l':(*)list) = LENGTH (xs:(**)list)");;
  e(IMP_RES_TAC BNDsL_MATCH_EQ);;
  e(POP_ASSUM (REWRITE_THM o
        SPECL["BND h'((ss:**->**) h)(s:**->*)";"BND h' h(s:**->*)"]));;
%



close_theory();;

