% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains theorems about the semantics of commands.  These
  are important theorems, and have substantial proofs, especially
  C_EXTENSION and c_subst_biject.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `cmd_semantics`;;

new_parent `well_formed`;;


% --------------------------------------------------------------------- %
% Need the finite_sets, and ind_defs libraries.                         %
% --------------------------------------------------------------------- %
%
loadt `load_string`;;
load_string();;
%

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of cmd_semantics %

loadf `autoload`;;

autoload_all_defs_and_thms
  `variables more_finite_sets bindings variants
   assert_syntax assert_semantics prog_syntax prog_semantics
   substitutions var_substitutions prog_substitutions
   free_variables translations progress well_formed`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;



% =================================================================== %
% The following theorems express ways to verify the truth of command  %
% semantics tuples based on previously known, similar command         %
% semantics tuples.  These theorems are some of the most deep and     %
% subtle in the Sunrise system, and require considerable computation. %
%                                                                     %
% The primary theorems presented here are C_EXTENSION, c_subst_biject %
% and c_subst_THM, and C_BND_THM with C_BNDsL_THM and C_BNDsR_THM.    %
% =================================================================== %

let PROC_STATE_TRANSFORM =
 prove_thm
  (`PROC_STATE_TRANSFORM`,
   "!r p vars vals glbs pre post calls rec c x vals' x' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (x' = APPEND (APPEND xs vals') glbs) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL vals'"
   THENL
    [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

      IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
      THEN IMP_RES_TAC FVc_proc_subst
      THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "(x':(var)list) =
                                            APPEND(APPEND xs vals')glbs")])
      THEN IMP_RES_TAC WFc_proc_subst
      THEN ASM_CASES_TAC "(z:var) IN SL x"
      THENL
       [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
         THEN IMP_RES_TAC C_FREE_ONLY_VARIATION
         THEN POP_ASSUM (CONV_TAC o ONCE_DEPTH_CONV o RATOR_CONV o REWR_CONV o SYM)
         THEN ASM_CASES_TAC "(z:var) IN SL x'"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
          ];

         IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
         THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
         THENL
          [ REWRITE_TAC[ASSUME "(x':(var)list) = APPEND(APPEND xs vals')glbs";
                 SL_APPEND;IN_UNION;DE_MORGAN_THM]
            THEN ASM_REWRITE_TAC[]
            THEN IMP_RES_TAC NOT_IN_SUBSET
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
          ]
       ]
    ]
  );;

%<
g "!r p vars vals glbs pre post calls rec c x vals' x' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (x' = APPEND (APPEND xs vals') glbs) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL vals'");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(IMP_RES_TAC FVc_proc_subst);;
 e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "(x':(var)list) =
                                            APPEND(APPEND xs vals')glbs")]));;
 e(IMP_RES_TAC WFc_proc_subst);;
 e(ASM_CASES_TAC "(z:var) IN SL x");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
  e(IMP_RES_TAC C_FREE_ONLY_VARIATION);;
  e(POP_ASSUM (CONV_TAC o ONCE_DEPTH_CONV o RATOR_CONV o REWR_CONV o SYM));;
  e(ASM_CASES_TAC "(z:var) IN SL x'");;

   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(REWRITE_TAC[ASSUME "(x':(var)list) = APPEND(APPEND xs vals')glbs";
                 SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
   e(IMP_RES_TAC NOT_IN_SUBSET);;
   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
>%



let PROC_STATE_TRANSFORM_THM =
 prove_thm
  (`PROC_STATE_TRANSFORM_THM`,
   "!r p vars vals glbs pre post calls rec c x vals' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      let x' = APPEND (APPEND xs vals') glbs in
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))",
   REPEAT STRIP_TAC
   THEN CONV_TAC (ONCE_DEPTH_CONV let_CONV)
   THEN MATCH_MP_TAC (SPEC_ALL PROC_STATE_TRANSFORM)
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!r p vars vals glbs pre post calls rec c x vals' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      let x' = APPEND (APPEND xs vals') glbs in
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))";;
e(REPEAT STRIP_TAC);;
e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
e(MATCH_MP_TAC (SPEC_ALL PROC_STATE_TRANSFORM));;
e(ASM_REWRITE_TAC[]);;
>%




let EXTENSION_MAP_THM =
 prove_thm
  (`EXTENSION_MAP_THM`,
   "!vals xs glbs vals' es s1 ns s2 x.
    (vals' = variants vals(SL(APPEND xs glbs))) /\
    ES es s1 ns s2 /\
    (LENGTH vals = LENGTH es) /\
    DL(APPEND xs glbs) /\
    (SL xs SUBSET SL x) /\
    (SL glbs SUBSET SL x) ==>
    (MAP(BNDsL ns vals' s2)(APPEND(APPEND xs vals')glbs) =
     MAP(BNDsL ns vals'(BNDsL(MAP s2 x)x s1'))(APPEND(APPEND xs vals')glbs))",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[MAP_APPEND]
   THEN UNASSUME_THEN (STRIP_ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o
                   REWRITE_RULE[DL_APPEND]) "DL(APPEND xs glbs :(var)list)"
   THEN IMP_RES_THEN (\th1. 
      let th3 = (SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"]
                       LENGTH_variants) in
      FIRST_ASSUM (\th2.  % LENGTH vals = LENGTH ns %
          ASSUME_TAC (REWRITE_RULE[SYM(ASSUME
                        "vals' = variants vals(SL(APPEND xs glbs))")]
                      (TRANS th3 (TRANS th2 th1))))) ES_LENGTH_THM
   THEN (ASSUME_TAC o
     REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
     SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"])
                      DL_variants
   THEN (STRIP_ASSUME_TAC o
         REWRITE_RULE[DISJOINT_UNION] o
         ONCE_REWRITE_RULE[DISJOINT_SYM] o
         REWRITE_RULE[SL_APPEND] o
         REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
         SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"])
                      DISJOINT_variants
   THEN IMP_RES_TAC MAP_BNDsL_IDENT
   THEN IMP_RES_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
   THEN IMP_RES_THEN REWRITE_THM MAP_BNDsL_MAP_SUBSET
   THEN ASM_REWRITE_TAC[]
  );;

%< EXTENSION_MAP_THM
g "!vals xs glbs vals' es s1 ns s2 x.
    (vals' = variants vals(SL(APPEND xs glbs))) /\
    ES es s1 ns s2 /\
    (LENGTH vals = LENGTH es) /\
    DL(APPEND xs glbs) /\
    (SL xs SUBSET SL x) /\
    (SL glbs SUBSET SL x) ==>
    (MAP(BNDsL ns vals' s2)(APPEND(APPEND xs vals')glbs) =
     MAP(BNDsL ns vals'(BNDsL(MAP s2 x)x s1'))(APPEND(APPEND xs vals')glbs))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[MAP_APPEND]);;
e(UNASSUME_THEN (STRIP_ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o
                   REWRITE_RULE[DL_APPEND]) "DL(APPEND xs glbs :(var)list)");;
e(IMP_RES_THEN (\th1. 
      let th3 = (SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"]
                       LENGTH_variants) in
      FIRST_ASSUM (\th2.  % LENGTH vals = LENGTH ns %
          ASSUME_TAC (REWRITE_RULE[SYM(ASSUME
                        "vals' = variants vals(SL(APPEND xs glbs))")]
                      (TRANS th3 (TRANS th2 th1))))) ES_LENGTH_THM);;
e((ASSUME_TAC o
   REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
   SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"])
                      DL_variants);;
e((STRIP_ASSUME_TAC o
   REWRITE_RULE[DISJOINT_UNION] o
   ONCE_REWRITE_RULE[DISJOINT_SYM] o
   REWRITE_RULE[SL_APPEND] o
   REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
   SPECL ["vals: var list";"SL (APPEND xs glbs) :var set"])
                      DISJOINT_variants);;
e(IMP_RES_TAC MAP_BNDsL_IDENT);;
e(IMP_RES_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;
e(IMP_RES_THEN REWRITE_THM MAP_BNDsL_MAP_SUBSET);;
e(ASM_REWRITE_TAC[]);;
>%



% ========================================================== %
% Here is the C_EXTENSION theorem we could not prove before. %
% ========================================================== %


let C_EXTENSION_LEMMA =
 TAC_PROOF
  (([],
    "!c r s1 s2.
      C c r s1 s2 ==>
      (!x s1' g.
        WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
        (MAP s1 x = MAP s1' x) ==>
        C c r s1' (BNDsL (MAP s2 x) x s1'))"
   ),
   CMD_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc;FVc_DEF;FVxs_SL]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN ASM_REWRITE_TAC[SUBSET;UNION_SUBSET;o_THM]
   THEN REPEAT STRIP_TAC
   THENL

     [  % SKIP: %
        ASM_REWRITE_TAC[IDENT_BNDsL]
        THEN SKIP_TAC;

        % ASSIGN: %
        IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE
        THEN ASSIGN_TAC
        THEN IMP_RES_TAC E_EXTENSION;

        % SEQ: %
        SEQ_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % IF_T: %
        IF_T_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             IMP_RES_TAC B_EXTENSION
          ];

        % IF_F: %
        IF_F_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             IMP_RES_TAC B_EXTENSION
          ];

        % WHILE_T: %
        WHILE_T_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN EXISTS_TAC "BNDsL(MAP (s3:state) x) x s1'"
        THEN REPEAT CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s4:state) x";"x:(var)list";
                                  "MAP (s3:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[WFc;FVc_DEF;UNION_SUBSET;MAP_BNDsL_MAP]
             THEN EXISTS_TAC "v:vexp"
             THEN EXISTS_TAC "x':var"
             THEN ASM_REWRITE_TAC[];

             IMP_RES_TAC B_EXTENSION
          ];

        % WHILE_F: %
        WHILE_F_TAC
        THEN IMP_RES_TAC B_EXTENSION;

        % CALL: %
        ONCE_REWRITE_TAC[BNDsL_MAP_BNDsL_MAP_COMMUTE]
        THEN REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL PROC_STATE_TRANSFORM_THM)
        THENL
         [  ASM_REWRITE_TAC[];

            CONV_TAC (ONCE_DEPTH_CONV let_CONV)
            THEN CALL_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:(var)list"
            THEN EXISTS_TAC "vals:(var)list"
            THEN EXISTS_TAC "ns:(num)list"
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
            THEN CONJ_TAC
            THENL
              [  REWRITE_TAC[SYM(ASSUME
                       "(vals':var list) = variants vals(SL(APPEND xs glbs))")]
                 THEN FIRST_ASSUM MATCH_MP_TAC
                 THEN EXISTS_TAC "calls <|g ((APPEND xs vals') //v (APPEND vars vals))"
                 THEN IMP_RES_TAC FVc_proc_subst
                 THEN IMP_RES_TAC WFc_proc_subst
                 THEN USE_IMP_THEN ASM_REWRITE_THM (SPEC_ALL EXTENSION_MAP_THM)
                 THEN ASM_REWRITE_TAC[];

                 IMP_RES_TAC ES_EXTENSION
              ]
         ]
     ]
  );;

%< C_EXTENSION_LEMMA
g "!c r s1 s2.
      C c r s1 s2 ==>
      (!x s1' g.
        WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
        (MAP s1 x = MAP s1' x) ==>
        C c r s1' (BNDsL (MAP s2 x) x s1'))";;
e(CMD_STR_INDUCT_TAC);;

% SKIP: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[IDENT_BNDsL]);;
e(SKIP_TAC);;

% ASSIGN: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE);;
e(ASSIGN_TAC);;
e(IMP_RES_TAC E_EXTENSION);;

% SEQ: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(SEQ_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% IF_T: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IF_T_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(IMP_RES_TAC B_EXTENSION);;

% IF_F: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IF_F_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(IMP_RES_TAC B_EXTENSION);;

% WHILE_T: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(WHILE_T_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(EXISTS_TAC "BNDsL(MAP (s3:state) x) x s1'");;
e(REPEAT CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s4:state) x";"x:(var)list";
                         "MAP (s3:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[WFc;FVc_DEF;UNION_SUBSET;MAP_BNDsL_MAP]);;
  e(EXISTS_TAC "v:vexp");;
  e(EXISTS_TAC "x':var");;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC B_EXTENSION);;

% WHILE_F: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(WHILE_F_TAC);;
e(IMP_RES_TAC B_EXTENSION);;

% CALL: %
e(REWRITE_TAC[WFc;FVc_DEF;FVxs_SL]);;
e(ASM_REWRITE_TAC[SUBSET;UNION_SUBSET;o_THM]);;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[BNDsL_MAP_BNDsL_MAP_COMMUTE]);;
e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
e(USE_IMP_THEN REWRITE_THM (SPEC_ALL PROC_STATE_TRANSFORM_THM));;

 e(ASM_REWRITE_TAC[]);;

e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
e(CALL_TAC);;
e(EXISTS_TAC "c:cmd" THEN
  EXISTS_TAC "vars:(var)list" THEN
  EXISTS_TAC "vals:(var)list" THEN
  EXISTS_TAC "ns:(num)list" THEN
  EXISTS_TAC "glbs:(var)list" THEN
  EXISTS_TAC "pre:aexp" THEN
  EXISTS_TAC "post:aexp" THEN
  EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "rec:aexp");;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

  e(REWRITE_TAC[SYM(ASSUME
         "(vals':var list) = variants vals(SL(APPEND xs glbs))")]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "calls <|g ((APPEND xs vals') //v (APPEND vars vals))");;
  e(IMP_RES_TAC FVc_proc_subst);;
  e(IMP_RES_TAC WFc_proc_subst);;
  e(USE_IMP_THEN ASM_REWRITE_THM (SPEC_ALL EXTENSION_MAP_THM));;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC ES_EXTENSION);;

C_EXTENSION_LEMMA = 
|- !c r s1 s2.
    C c r s1 s2 ==>
    (!x s1' g.
      WFc c g r /\
      WFenv_syntax r /\
      (FVc c r) SUBSET (SL x) /\
      (MAP s1 x = MAP s1' x) ==>
      C c r s1'(BNDsL(MAP s2 x)x s1'))
>%

let C_EXTENSION =
 prove_thm
  (`C_EXTENSION`,
   "!c g r s1 s2 x s1'.
      WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) /\ C c r s1 s2  ==>
      C c r s1' (BNDsL (MAP s2 x) x s1')",
   REPEAT STRIP_TAC
   THEN POP_ASSUM (MATCH_MP_TAC o SPEC_ALL o MATCH_MP C_EXTENSION_LEMMA)
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!c g r s1 s2 x s1'.
      WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) /\ C c r s1 s2  ==>
      C c r s1' (BNDsL (MAP s2 x) x s1')";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (ASSUME_TAC o MATCH_MP C_EXTENSION_LEMMA));;
e(POP_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[]);;
%




let WFcsubst_THM =
 prove_thm
  (`WFcsubst_THM`,
  "(!r ss. WFcsubst (SKIP)  r ss = ((!x. WFx(ss x) = WFx x) /\
                                     ONE_ONE ss /\ ONTO ss)) /\
   (!r ss. WFcsubst (ABORT) r ss = ((!x. WFx(ss x) = WFx x) /\
                                     ONE_ONE ss /\ ONTO ss)) /\
   (!x e r ss. WFcsubst (ASSIGN x e) r ss =
                   ((!x. WFx(ss x) = WFx x) /\ ONE_ONE ss /\ ONTO ss)) /\
   (!c1 c2 r ss. WFcsubst (SEQ c1 c2) r ss =
                   (WFcsubst c1 r ss /\ WFcsubst c2 r ss)) /\
   (!b c1 c2 r ss. WFcsubst (IF b c1 c2) r ss =
                   (WFcsubst c1 r ss /\ WFcsubst c2 r ss)) /\
   (!a pr b c r ss. WFcsubst (WHILE a pr b c) r ss =
                   (WFcsubst c r ss)) /\
   (!p xs es r ss. WFcsubst (CALL p xs es) r ss =
                   ((!x. WFx(ss x) = WFx x) /\ ONE_ONE ss /\ ONTO ss /\
                    (!x. x IN (SL ((FST o SND o SND o r) p)) ==>
                                (ss x = x))))",
   REPEAT CONJ_TAC
   THEN REWRITE_TAC[WFcsubst;GVc]
   THEN REWRITE_TAC[IN;IN_UNION]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

%< WFcsubst_THM
g "(!r ss. WFcsubst (SKIP)  r ss = ((!x. WFx(ss x) = WFx x) /\
                                     ONE_ONE ss /\ ONTO ss)) /\
   (!r ss. WFcsubst (ABORT) r ss = ((!x. WFx(ss x) = WFx x) /\
                                     ONE_ONE ss /\ ONTO ss)) /\
   (!x e r ss. WFcsubst (ASSIGN x e) r ss =
                   ((!x. WFx(ss x) = WFx x) /\ ONE_ONE ss /\ ONTO ss)) /\
   (!c1 c2 r ss. WFcsubst (SEQ c1 c2) r ss =
                   (WFcsubst c1 r ss /\ WFcsubst c2 r ss)) /\
   (!b c1 c2 r ss. WFcsubst (IF b c1 c2) r ss =
                   (WFcsubst c1 r ss /\ WFcsubst c2 r ss)) /\
   (!a pr b c r ss. WFcsubst (WHILE a pr b c) r ss =
                   (WFcsubst c r ss)) /\
   (!p xs es r ss. WFcsubst (CALL p xs es) r ss =
                   ((!x. WFx(ss x) = WFx x) /\ ONE_ONE ss /\ ONTO ss /\
                    (!x. x IN (SL ((FST o SND o SND o r) p)) ==>
                                (ss x = x))))";;
e(REPEAT CONJ_TAC
  THEN REWRITE_TAC[WFcsubst;GVc]
  THEN REWRITE_TAC[IN;IN_UNION]
  THEN REPEAT STRIP_TAC
  THEN EQ_TAC
  THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[]
  THEN REPEAT STRIP_TAC
  THEN RES_TAC);;

e(REPEAT CONJ_TAC);;
e(REWRITE_TAC[WFcsubst;GVc]);;

 e(REWRITE_TAC[IN]);;

 e(REWRITE_TAC[IN]);;

 e(REWRITE_TAC[IN]);;

 e(REWRITE_TAC[IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;

   e(RES_TAC);;

   e(RES_TAC);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;

   e(RES_TAC);;

   e(RES_TAC);;
>%





let c_subst_STATE_BNDsL_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' ns s2 s3 vals0 xs l0 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3  /\
     (vals' = variants vals (SL (APPEND xs glbs)))  /\
     (MAP ss vals0 = vals')  /\
     (xs = l0 <|xs ss)  /\
     WFxs l0  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss  /\
     (!x. WFx (ss x) = WFx x)
     ==>
     ((BNDsL (MAP s2 vals') vals0 (s3 o ss)) =
      (BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)))"
   ),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFxs_subst
   THEN POP_TAC
   THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "xs = l0 <|xs ss")])
   THEN SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
            ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
      THEN IMP_RES_TAC DL_APPEND_xs_subst
      THEN ASM_REWRITE_TAC[LENGTH_xs_subst];

      SUBGOAL_THEN
         "DL (APPEND (APPEND l0 glbs) vals0: (var)list)"
               ASSUME_TAC
      THENL
       [ MATCH_MP_TAC (INST_TYPE [":var",":**"] DL_MAP)
         THEN EXISTS_TAC "ss:vsubst"
         THEN REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
         THEN UNASSUME_THEN (ASSUME_TAC o SYM o REWRITE_RULE[xs_subst_MAP])
                   "xs = l0 <|xs ss"
         THEN ASM_REWRITE_TAC[MAP_APPEND]
         THEN ONCE_REWRITE_TAC[DL_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN REWRITE_TAC[DL_variants;DISJOINT_variants]
         THEN IMP_RES_TAC DL_MAP_ONE_ONE
         THEN POP_TAC
         THEN POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                           ASSUME "MAP (ss:vsubst) l0 = xs";
                           ASSUME "MAP (ss:vsubst) glbs = glbs"]);

         IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                               C_FREE_ONLY_IMPACT)
         THEN REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'");MAP_MAP_o]
         THEN EXT_TAC "z:var"
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(z:var) IN SL vals0"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
            THEN IMP_RES_TAC IN_DISJOINT_LIST
            THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)"
            THENL
             [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN POP_ASSUM (STRIP_ASSUME_TAC o
                               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM])
               THEN REWRITE_TAC[o_THM]
               THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
               THENL
                [ MATCH_MP_TAC NOT_IN_SUBSET
                  THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs): var set"
                  THEN CONJ_TAC
                  THENL
                   [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                     THEN REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
                     THEN REWRITE_TAC[ASSUME "xs = l0 <|xs ss";
                                      SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]
                     THEN ONCE_REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) glbs = glbs")]
                     THEN REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]
                     THEN IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst;

                     MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                     THEN IMP_RES_TAC DL_APPEND_xs_subst
                     THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                   ];

                  USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
                  THEN REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]
                  THEN REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]
                  THEN IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst
                ]
             ]
          ]
       ]
    ]
  );;

%< c_subst_STATE_BNDsL_LEMMA
g "!r p vars vals glbs pre post calls rec c vals' ns s2 s3 vals0 xs l0 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r
       (BNDsL ns vals' s2) s3  /\
     (vals' = variants vals (SL (APPEND xs glbs)))  /\
     (MAP ss vals0 = vals')  /\
     (xs = l0 <|xs ss)  /\
     WFxs l0  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss  /\
     (!x. WFx (ss x) = WFx x)
     ==>
     ((BNDsL (MAP s2 vals') vals0 (s3 o ss)) =
      (BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFxs_subst);;
e(POP_TAC);;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "xs = l0 <|xs ss")]));;
e(SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
       ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
  e(IMP_RES_TAC DL_APPEND_xs_subst);;
  e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND l0 glbs) vals0: (var)list)"
            ASSUME_TAC);;

 e(MATCH_MP_TAC (INST_TYPE [":var",":**"] DL_MAP));;
 e(EXISTS_TAC "ss:vsubst");;
let st2 = get_state();;
set_state(st2);;
 e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM o REWRITE_RULE[xs_subst_MAP]) "xs = l0 <|xs ss");;
 e(ASM_REWRITE_TAC[MAP_APPEND]);;
 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[DL_variants;DISJOINT_variants]);;
 e(IMP_RES_TAC DL_MAP_ONE_ONE);;
 e(POP_TAC);;
 e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                           ASSUME "MAP (ss:vsubst) l0 = xs";
                           ASSUME "MAP (ss:vsubst) glbs = glbs"]));;

e(IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                               C_FREE_ONLY_IMPACT));;
e(REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'");MAP_MAP_o]);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL vals0");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
 e(IMP_RES_TAC IN_DISJOINT_LIST);;
 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM]));;
  e(REWRITE_TAC[o_THM]);;
  e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs): var set");;
   e(CONJ_TAC);;

    e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
    e(REWRITE_TAC[ASSUME "xs = l0 <|xs ss";
                  SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]);;
    e(ONCE_REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) glbs = glbs")]);;
    e(REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]);;
    e(IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst);;

    e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
    e(IMP_RES_TAC DL_APPEND_xs_subst);;
    e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;

   e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]);;
   e(REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]);;
   e(IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst);;
>%





let c_subst_PROC_STATE_LEMMA =
 prove_thm
  (`c_subst_PROC_STATE_LEMMA`,
  "!vals'' vals vals' xs glbs r vars pre post calls rec c l0 (s2:state) s3 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     (vals'  = variants vals (SL (APPEND xs glbs)))  /\
     (vals'' = variants vals (SL (APPEND l0 glbs)))  /\
     (xs = l0 <|xs ss)  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss
     ==>
     ((BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)) =
      (BNDsL (MAP (s2 o ss) vals'') vals''
         (BNDsL (MAP (s3 o (((APPEND xs vals') //v (APPEND vars vals)) o
                            (\y. @z. ((APPEND l0 vals'') //v (APPEND vars vals))z = y)))
                     (APPEND (APPEND l0 glbs) vals''))
                (APPEND (APPEND l0 glbs) vals'')
            (BNDsL ns vals'' (s2 o ss)))))",

   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "DL (APPEND (APPEND l0 glbs) vals'': (var)list)"
            ASSUME_TAC
   THENL
    [ ONCE_REWRITE_TAC[DL_APPEND]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants];

      SUBGOAL_THEN
           "DL (APPEND (APPEND xs glbs) vals': (var)list)"
                  ASSUME_TAC
      THENL
       [ REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
         THEN ONCE_REWRITE_TAC[DL_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]
         THEN IMP_RES_TAC DL_MAP_ONE_ONE
         THEN POP_TAC
         THEN POP_TAC
         THEN POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                            ASSUME "MAP (ss:vsubst) glbs = glbs"]);

         SUBGOAL_THEN
            "DL (APPEND (APPEND vars vals) glbs: (var)list) /\ DL vars /\ DL vals"
                  STRIP_ASSUME_TAC
         THENL
          [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_ALL_TAC[DL_APPEND]
            THEN ASM_REWRITE_TAC[];

            EXT_TAC "z:var"
            THEN GEN_TAC
            THEN ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)"
            THENL
             [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
               THEN IMP_RES_TAC IN_DISJOINT_LIST
               THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
               THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH
               THENL
                [ ONCE_REWRITE_TAC[SL_APPEND]
                  THEN ASM_REWRITE_TAC[IN_UNION];

                  REWRITE_TAC[o_THM]
                  THEN (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND] o ASSUME)
                            "DL(APPEND l0 glbs: var list)"
                  THEN (STRIP_ASSUME_TAC o REWRITE_RULE[SL_APPEND;IN_UNION] o ASSUME)
                            "(z:var) IN SL(APPEND l0 glbs)"
                  THENL
                   [ USE_IMP_THEN REWRITE_THM NULL_APPEND_vsubst_INVERSE
                     THENL
                      [ ASM_REWRITE_TAC[];

                        USE_IMP_THEN REWRITE_THM
                           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
                        THENL
                         [ REWRITE_ALL_TAC[DL_APPEND]
                           THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                           THEN IMP_RES_TAC vsubst_ELEMENTS_INVERSE;

                           USE_IMP_THEN REWRITE_THM vsubst_REVERSE_INVERSE
                           THENL
                            [ ASM_REWRITE_TAC[];

                              USE_IMP_THEN REWRITE_THM vsubst_TRANS_ELEMENTS1
                              THENL
                               [ REWRITE_ALL_TAC[DL_APPEND]
                                 THEN ASM_REWRITE_TAC[LENGTH_xs_subst];

                                 ASM_REWRITE_TAC[xs_subst_MAP]
                                 THEN USE_IMP_THEN REWRITE_THM vsubst_MAP_ONE_ONE
                                 THEN ASM_REWRITE_TAC[]
                               ]
                            ]
                         ]
                      ];

                     USE_IMP_THEN REWRITE_THM IDENT_vsubst_INVERSE
                     THENL
                      [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                        THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3;

                        USE_IMP_THEN REWRITE_THM IDENT_vsubst
                        THENL
                         [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                           THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3;

                           RES_THEN REWRITE_THM
                         ]
                      ]
                   ]
                ];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN POP_ASSUM (STRIP_ASSUME_TAC o
                               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM])
               THEN ASM_CASES_TAC "(z:var) IN SL vals''"
               THENL
                [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

                  IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
                  THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
                  THENL
                   [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                     THEN ASM_REWRITE_TAC[];

                     IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
                   ]
                ]
             ]
          ]
       ]
    ]
  );;

%<
g "!vals'' vals vals' xs glbs r vars pre post calls rec c l0 (s2:state) s3 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     (vals'  = variants vals (SL (APPEND xs glbs)))  /\
     (vals'' = variants vals (SL (APPEND l0 glbs)))  /\
     (xs = l0 <|xs ss)  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss
     ==>
     ((BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)) =
      (BNDsL (MAP (s2 o ss) vals'') vals''
         (BNDsL (MAP (s3 o (((APPEND xs vals') //v (APPEND vars vals)) o
                            (\y. @z. ((APPEND l0 vals'') //v (APPEND vars vals))z = y)))
                     (APPEND (APPEND l0 glbs) vals''))
                (APPEND (APPEND l0 glbs) vals'')
            (BNDsL ns vals'' (s2 o ss)))))";;

e(REPEAT STRIP_TAC);;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND l0 glbs) vals'': (var)list)"
            ASSUME_TAC);;

 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants]);;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND xs glbs) vals': (var)list)"
            ASSUME_TAC);;

 e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]);;
 e(IMP_RES_TAC DL_MAP_ONE_ONE);;
 e(POP_TAC);;
 e(POP_TAC);;
 e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                            ASSUME "MAP (ss:vsubst) glbs = glbs"]));;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND vars vals) glbs: (var)list) /\ DL vars /\ DL vals"
            STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_ALL_TAC[DL_APPEND]);;
 e(ASM_REWRITE_TAC[]);;

e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
 e(IMP_RES_TAC IN_DISJOINT_LIST);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(ONCE_REWRITE_TAC[SL_APPEND]);;
  e(ASM_REWRITE_TAC[IN_UNION]);;

  e(REWRITE_TAC[o_THM]);;
  e((STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND] o ASSUME)
         "DL(APPEND l0 glbs: var list)");;
  e((STRIP_ASSUME_TAC o REWRITE_RULE[SL_APPEND;IN_UNION] o ASSUME)
         "(z:var) IN SL(APPEND l0 glbs)");;

   e(USE_IMP_THEN REWRITE_THM NULL_APPEND_vsubst_INVERSE);;

    e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

     e(REWRITE_ALL_TAC[DL_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
     e(IMP_RES_TAC vsubst_ELEMENTS_INVERSE);;

     e(USE_IMP_THEN REWRITE_THM vsubst_REVERSE_INVERSE);;

      e(ASM_REWRITE_TAC[]);;

      e(USE_IMP_THEN REWRITE_THM vsubst_TRANS_ELEMENTS1);;

       e(REWRITE_ALL_TAC[DL_APPEND]);;
       e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;

       e(ASM_REWRITE_TAC[xs_subst_MAP]);;
       e(USE_IMP_THEN REWRITE_THM vsubst_MAP_ONE_ONE);;
       e(ASM_REWRITE_TAC[]);;

   e(USE_IMP_THEN REWRITE_THM IDENT_vsubst_INVERSE);;

    e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;

    e(USE_IMP_THEN REWRITE_THM IDENT_vsubst);;

     e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
     e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;

     e(RES_THEN REWRITE_THM);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(POP_ASSUM (STRIP_ASSUME_TAC o
              REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM]));;
 e(ASM_CASES_TAC "(z:var) IN SL vals''");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
>%





let MAP_EXTENSION_STATE_LEMMA =
 prove_thm
  (`MAP_EXTENSION_STATE_LEMMA`,
  "!vals'' xs vals' vars vals glbs pre post calls rec c l0 es ns (s2:state) ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     (vals'  = variants vals (SL (APPEND xs glbs)))  /\
     (vals'' = variants vals (SL (APPEND l0 glbs)))  /\
     ES es s1 ns s2  /\
     (xs = l0 <|xs ss)  /\
     (es = l1 <|es ss)  /\
     (LENGTH vars = LENGTH l0)  /\
     (LENGTH vals = LENGTH l1)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss
     ==>
     (MAP ((BNDsL ns vals' s2) o
           (((APPEND xs vals') //v (APPEND vars vals)) o
            (\y. @z. ((APPEND l0 vals'') //v (APPEND vars vals))z = y)))
          (APPEND (APPEND l0 glbs) vals'') =
      MAP (BNDsL ns vals'' (s2 o ss))
          (APPEND (APPEND l0 glbs) vals''))",
   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "DL (APPEND (APPEND l0 glbs) vals'': (var)list)"
            ASSUME_TAC
   THENL
    [ ONCE_REWRITE_TAC[DL_APPEND]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants];

      SUBGOAL_THEN
           "DL (APPEND (APPEND xs glbs) vals': (var)list)"
                  ASSUME_TAC
      THENL
       [ REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
         THEN ONCE_REWRITE_TAC[DL_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]
         THEN IMP_RES_TAC DL_MAP_ONE_ONE
         THEN POP_TAC
         THEN POP_TAC
         THEN POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                            ASSUME "MAP (ss:vsubst) glbs = glbs"]);

         SUBGOAL_THEN
              "DL (APPEND (APPEND vars vals) glbs: (var)list)
               /\ DL vars /\ DL vals"
                    STRIP_ASSUME_TAC
         THENL
          [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_ALL_TAC[DL_APPEND]
            THEN ASM_REWRITE_TAC[];

            (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND] o ASSUME)
                     "DL(APPEND l0 glbs: var list)"
            THEN REWRITE_TAC[MAP_APPEND;MAP_o;o_THM]
            THEN USE_IMP_THEN REWRITE_THM APPEND_11
                 THEN REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]
            THEN USE_IMP_THEN REWRITE_THM APPEND_11
                 THEN REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]
            THEN REPEAT CONJ_TAC % begins 3 subgoals %
            THENL
             [ USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst_INVERSE
               THENL
                [ ASM_REWRITE_TAC[SUBSET_REFL];

                  USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE
                  THENL
                   [ ASM_REWRITE_TAC[];

                     USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
                     THENL
                      [ ASM_REWRITE_TAC[LENGTH_xs_subst;SUBSET_REFL]
                        THEN REWRITE_TAC[xs_subst_MAP]
                        THEN IMP_RES_THEN ASM_REWRITE_THM DL_MAP_ONE_ONE;

                        USE_IMP_THEN REWRITE_THM
                            (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
                        THENL
                         [ ASM_REWRITE_TAC[LENGTH_xs_subst]
                           THEN REWRITE_TAC[xs_subst_MAP]
                           THEN IMP_RES_THEN ASM_REWRITE_THM DL_MAP_ONE_ONE;

                           USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                           THENL
                            [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                              THEN ASM_REWRITE_TAC[];

                              USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                              THENL
                               [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                                 THEN ASM_REWRITE_TAC[];

                                 REWRITE_TAC[MAP_o;o_THM]
                                 THEN ASM_REWRITE_TAC[xs_subst_MAP]
                               ]  % finished 1 goal of 3 %
                            ]
                         ]
                      ]
                   ]
                ];

               USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst_INVERSE
               THENL
                [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                  THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND];

                  USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
                  THENL
                   [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                     THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND];

                     USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                     THENL
                      [ ONCE_REWRITE_TAC[DISJOINT_SYM]
                        THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND];

                        USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                        THENL
                         [ ONCE_REWRITE_TAC[DISJOINT_SYM]
                           THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND];

                           REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
                           THEN ASM_REWRITE_TAC[MAP_o;o_THM]
                         ]  % finished 2nd of 3 goals %
                      ]
                   ]
                ];

               USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2_INVERSE
               THENL
                [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                  THEN ASM_REWRITE_TAC[LENGTH_variants;SUBSET_REFL];

                  USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE
                  THENL
                   [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants];

                     USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2
                     THENL
                      [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
                        THEN ASM_REWRITE_TAC[LENGTH_xs_subst;LENGTH_variants;
                                                SUBSET_REFL];

                        USE_IMP_THEN REWRITE_THM
                            (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
                        THENL
                         [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants];

                           USE_IMP_THEN REWRITE_THM
                               (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT)
                           THENL
                            [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
                              THEN IMP_RES_THEN (REWRITE_THM o SYM) ES_LENGTH_THM
                              THEN ASM_REWRITE_TAC[LENGTH_es_subst];

                              USE_IMP_THEN REWRITE_THM
                                  (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT)
                              THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
                              THEN IMP_RES_THEN (REWRITE_THM o SYM) ES_LENGTH_THM
                              THEN ASM_REWRITE_TAC[LENGTH_es_subst]
                                  % finished 3rd goal of 3 %
                            ]
                         ]
                      ]
                   ]
                ]
             ]
          ]
       ]
    ]
  );;

%< MAP_EXTENSION_STATE_LEMMA
g "!vals'' xs vals' vars vals glbs pre post calls rec c l0 es ns (s2:state) ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     (vals'  = variants vals (SL (APPEND xs glbs)))  /\
     (vals'' = variants vals (SL (APPEND l0 glbs)))  /\
     ES es s1 ns s2  /\
     (xs = l0 <|xs ss)  /\
     (es = l1 <|es ss)  /\
     (LENGTH vars = LENGTH l0)  /\
     (LENGTH vals = LENGTH l1)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss
     ==>
     (MAP ((BNDsL ns vals' s2) o
           (((APPEND xs vals') //v (APPEND vars vals)) o
            (\y. @z. ((APPEND l0 vals'') //v (APPEND vars vals))z = y)))
          (APPEND (APPEND l0 glbs) vals'') =
      MAP (BNDsL ns vals'' (s2 o ss))
          (APPEND (APPEND l0 glbs) vals''))";;
e(REPEAT STRIP_TAC);;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND l0 glbs) vals'': (var)list)"
            ASSUME_TAC);;

 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants]);;

 e(SUBGOAL_THEN
      "DL (APPEND (APPEND xs glbs) vals': (var)list)"
            ASSUME_TAC);;

  e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
  e(ONCE_REWRITE_TAC[DL_APPEND]);;
  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]);;
  e(IMP_RES_TAC DL_MAP_ONE_ONE);;
  e(POP_TAC);;
  e(POP_TAC);;
  e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                             ASSUME "MAP (ss:vsubst) glbs = glbs"]));;

  e(SUBGOAL_THEN
      "DL (APPEND (APPEND vars vals) glbs: (var)list) /\ DL vars /\ DL vals"
            STRIP_ASSUME_TAC);;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_ALL_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[]);;

   e((STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND] o ASSUME)
           "DL(APPEND l0 glbs: var list)");;
   e(REWRITE_TAC[MAP_APPEND;MAP_o;o_THM]);;
   e(USE_IMP_THEN REWRITE_THM APPEND_11
     THEN REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;
   e(USE_IMP_THEN REWRITE_THM APPEND_11
     THEN REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;
   e(REPEAT CONJ_TAC);; % 3 subgoals %

    e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst_INVERSE);;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE);;

      e(ASM_REWRITE_TAC[]);;

      e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

       e(ASM_REWRITE_TAC[LENGTH_xs_subst;SUBSET_REFL]);;
       e(REWRITE_TAC[xs_subst_MAP]);;
       e(IMP_RES_THEN ASM_REWRITE_THM DL_MAP_ONE_ONE);;

       e(USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

        e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
        e(REWRITE_TAC[xs_subst_MAP]);;
        e(IMP_RES_THEN ASM_REWRITE_THM DL_MAP_ONE_ONE);;

        e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

         e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
         e(ASM_REWRITE_TAC[]);;

         e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

          e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
          e(ASM_REWRITE_TAC[]);;

          e(REWRITE_TAC[MAP_o;o_THM]);;
          e(ASM_REWRITE_TAC[xs_subst_MAP]);; % finished 1 goal of 3 %

    e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst_INVERSE);;

     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

     e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

      e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
      e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

      e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

       e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
       e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

       e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

        e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
        e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

        e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
        e(ASM_REWRITE_TAC[MAP_o;o_THM]);;  % finished 2nd of 3 goals %

    e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2_INVERSE);;

     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_variants;SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE);;

      e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

      e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2);;

       e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
       e(ASM_REWRITE_TAC[LENGTH_xs_subst;LENGTH_variants;SUBSET_REFL]);;

       e(USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

        e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

        e(USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT));;

         e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;
         e(IMP_RES_THEN (REWRITE_THM o SYM) ES_LENGTH_THM);;
         e(ASM_REWRITE_TAC[LENGTH_es_subst]);;

         e(USE_IMP_THEN REWRITE_THM
                (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT));;
         e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;
         e(IMP_RES_THEN (REWRITE_THM o SYM) ES_LENGTH_THM);;
         e(ASM_REWRITE_TAC[LENGTH_es_subst]);; % end of 3rd of 3 subgoals %
>%



let SL_APPEND_SYM23 =
 TAC_PROOF
  (([],
    "!a b (c:(*)list). SL(APPEND(APPEND a b) c) = SL(APPEND(APPEND a c) b)"
   ),
   REWRITE_TAC[EXTENSION]
   THEN REWRITE_TAC[SL_APPEND;IN_UNION]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% SL_APPEND_SYM23
g "!a b (c:(*)list). SL(APPEND(APPEND a b) c) = SL(APPEND(APPEND a c) b)";;
e(REWRITE_TAC[EXTENSION]);;
e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC
  THEN REPEAT STRIP_TAC
  THEN ASM_REWRITE_TAC[]
);;
%




let WFcsubst_subst_LEMMA =
 prove_thm
  (`WFcsubst_subst_LEMMA`,
    "!r p vars vals glbs pre post calls rec c vals' xs l0 ss.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (xs = l0 <|xs ss) /\
      WFxs l0 /\
      (LENGTH vars = LENGTH l0) /\
      DL(APPEND l0 glbs) /\
      (!x. WFx(ss x) = WFx x) /\
      ONE_ONE ss /\ ONTO ss /\
      (!x. x IN SL glbs ==> (ss x = x))
      ==>
      WFcsubst
       (c <|c
        ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
         (APPEND vars vals)))
       r
       (((APPEND xs vals') //v (APPEND vars vals)) o
        (\y. @z.
           ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
            (APPEND vars vals))
           z = y))",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[WFcsubst]
   THEN REPEAT CONJ_TAC
   THENL %4 subgoals%
    [ MATCH_MP_TAC o_OF_WFx_KEEPS_WFx
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN CONJ_TAC
      THENL
       [ MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx
         THEN REWRITE_ALL_TAC[WFxs_APPEND]
         THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;
                              LENGTH_variants;LENGTH_xs_subst]
         THEN USE_IMP_THEN ASM_REWRITE_THM WFxs_subst
         THEN ASM_REWRITE_TAC[];

         MATCH_MP_TAC SUBST_INVERSE_OF_WFx_KEEPS_WFx
         THEN REWRITE_ALL_TAC[WFxs_APPEND]
         THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;
                              LENGTH_variants;LENGTH_xs_subst]
       ];

      MATCH_MP_TAC o_ONE_ONE
      THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_INVERSE_ONE_ONE];

      MATCH_MP_TAC o_ONTO
      THEN REWRITE_TAC[vsubst_ONTO;vsubst_INVERSE_ONTO];

      REWRITE_TAC[GVc_subst;o_THM]
      THEN SUBGOAL_THEN
              "DL (APPEND (APPEND l0 glbs)
                          (variants vals(SL(APPEND l0 glbs))))"
                    ASSUME_TAC
      THENL
       [ ONCE_REWRITE_TAC[DL_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants];

         SUBGOAL_THEN
             "DL (APPEND (APPEND xs glbs) vals': (var)list)"
                   ASSUME_TAC
         THENL
          [ REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
            THEN ONCE_REWRITE_TAC[DL_APPEND]
            THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
            THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]
            THEN IMP_RES_TAC DL_MAP_ONE_ONE
            THEN POP_TAC
            THEN POP_TAC
            THEN POP_ASSUM (REWRITE_THM o
                            REWRITE_RULE[MAP_APPEND;
                                 ASSUME "MAP (ss:vsubst) glbs = glbs"]);

            SUBGOAL_THEN
                "DL (APPEND (APPEND vars vals) glbs: (var)list)"
                       ASSUME_TAC
            THENL
             [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX;

               REPEAT STRIP_TAC
               THEN SUBGOAL_THEN "(x:var) IN SL glbs" ASSUME_TAC
               THENL
                [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                  THEN IMP_RES_TAC SUBSET_DEF;

                  IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST3
                  THEN USE_IMP_THEN REWRITE_THM IDENT_vsubst_INVERSE
                  THENL
                   [ ONCE_REWRITE_TAC[SL_APPEND]
                     THEN ASM_REWRITE_TAC[IN_UNION];

                     USE_IMP_THEN REWRITE_THM IDENT_vsubst
                     THEN REWRITE_TAC[SL_APPEND;IN_UNION]
                     THEN ASM_REWRITE_TAC[]
                   ]
                ]
             ]
          ]
       ]
    ]
  );;

%<
g "!r p vars vals glbs pre post calls rec c vals' xs l0 ss.
   WFenv_syntax r /\
   (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
   (vals' = variants vals (SL (APPEND xs glbs))) /\
   (xs = l0 <|xs ss) /\
   WFxs l0 /\
   (LENGTH vars = LENGTH l0) /\
   DL(APPEND l0 glbs) /\
   (!x. WFx(ss x) = WFx x) /\
   ONE_ONE ss /\ ONTO ss /\
   (!x. x IN SL glbs ==> (ss x = x))
   ==>
   WFcsubst
    (c <|c
     ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
      (APPEND vars vals)))
    r
    (((APPEND xs vals') //v (APPEND vars vals)) o
     (\y. @z.
        ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
         (APPEND vars vals))
        z = y))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[WFcsubst]);;
e(REPEAT CONJ_TAC);; %4 subgoals%

 e(MATCH_MP_TAC o_OF_WFx_KEEPS_WFx);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;
                    LENGTH_variants;LENGTH_xs_subst]);;
  e(USE_IMP_THEN ASM_REWRITE_THM WFxs_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(MATCH_MP_TAC SUBST_INVERSE_OF_WFx_KEEPS_WFx);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;
                    LENGTH_variants;LENGTH_xs_subst]);;

 e(MATCH_MP_TAC o_ONE_ONE);;
 e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_INVERSE_ONE_ONE]);;

 e(MATCH_MP_TAC o_ONTO);;
 e(REWRITE_TAC[vsubst_ONTO;vsubst_INVERSE_ONTO]);;

 e(REWRITE_TAC[GVc_subst;o_THM]);;
 e(SUBGOAL_THEN
       "DL (APPEND (APPEND l0 glbs) (variants vals(SL(APPEND l0 glbs))))"
             ASSUME_TAC);;

  e(ONCE_REWRITE_TAC[DL_APPEND]);;
  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants]);;

  e(SUBGOAL_THEN
       "DL (APPEND (APPEND xs glbs) vals': (var)list)"
             ASSUME_TAC);;

   e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;xs_subst_MAP]);;
   e(IMP_RES_TAC DL_MAP_ONE_ONE);;
   e(POP_TAC);;
   e(POP_TAC);;
   e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                              ASSUME "MAP (ss:vsubst) glbs = glbs"]));;

   e(SUBGOAL_THEN
       "DL (APPEND (APPEND vars vals) glbs: (var)list)"
              ASSUME_TAC);;

    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

    e(REPEAT STRIP_TAC);;
    e(SUBGOAL_THEN "(x:var) IN SL glbs" ASSUME_TAC);;

     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(IMP_RES_TAC SUBSET_DEF);;

     e(IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST3);;
     e(USE_IMP_THEN REWRITE_THM IDENT_vsubst_INVERSE);;

      e(ONCE_REWRITE_TAC[SL_APPEND]);;
      e(ASM_REWRITE_TAC[IN_UNION]);;

      e(USE_IMP_THEN REWRITE_THM IDENT_vsubst);;
      e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
      e(ASM_REWRITE_TAC[]);;
>%






let WFc_WHILE =
 prove_thm
  (`WFc_WHILE`,
   "!a v x b c g r.
     WFc (WHILE a (ALESS v (AVAR x)) b c) g r =
     WFb b /\ WFc c g r /\ ~WFx x /\ ~x IN ((FVa a) UNION (FVv v))
           /\ (!p. ~x IN FVa (g p))",
   REWRITE_TAC[WFc]
   THEN REWRITE_TAC[aexp_one_one;vexp_one_one]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN EXISTS_TAC "v:vexp"
   THEN EXISTS_TAC "x:var"
   THEN ASM_REWRITE_TAC[]
  );;

% WFc_WHILE
g "!a v x b c g r.
     WFc (WHILE a (ALESS v (AVAR x)) b c) g r =
     WFb b /\ WFc c g r /\ ~WFx x /\ ~x IN ((FVa a) UNION (FVv v))
           /\ (!p. ~x IN FVa (g p))";;
e(REWRITE_TAC[WFc]);;
e(REWRITE_TAC[aexp_one_one;vexp_one_one]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(EXISTS_TAC "v:vexp");;
 e(EXISTS_TAC "x:var");;
 e(ASM_REWRITE_TAC[]);;
%


let c_subst_LEMMA1 =
 TAC_PROOF
  (([],
    "!c' r s1 s2. 
      C c' r s1 s2 ==>
      (!c ss g. (c' = (c <|c ss))
            /\ WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss ==>
        C c r (s1 o ss) (s2 o ss))"
   ),
   CMD_STR_INDUCT_TAC
   THENL
        % SKIP: %
     [  GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN SKIP_TAC;

        % ASSIGN: %
        GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_THEN ASM_REWRITE_THM BND_o_ONE_ONE_THM
        THEN ASSIGN_TAC
        THEN IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) e_subst_biject
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "e = e' <|e ss"
        THEN ASM_REWRITE_TAC[];

        % SEQ: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN SEQ_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[];

        % IF_T: %
        GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IF_T_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN RES_TAC
        THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
                    "WFcsubst c r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % IF_F: %
        GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IF_F_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN RES_TAC
        THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
                    "WFcsubst c r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % WHILE_T: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN WHILE_T_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN EXISTS_TAC "(s3:state) o (ss:vsubst)"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN RES_TAC
        THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
                      "WFcsubst c'' r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[WFc_WHILE;c_subst;cmd_one_one;WFcsubst_THM];

        % WHILE_F: %
        GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN WHILE_F_TAC
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
                      "WFcsubst c'' r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % CALL: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = s"
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
                       "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN IMP_RES2_THEN REWRITE_THM BNDsL_o_biject
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL c_subst_STATE_BNDsL_LEMMA)
        THENL
         [ ASM_REWRITE_TAC[MAP_MAP_o]
           THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
           THEN REWRITE_TAC[MAP_I];

           USE_IMP_THEN REWRITE_THM
               ((SPEC_ALL o SPECL["variants vals (SL (APPEND l0 glbs))";
                                  "vals: var list"])
                     c_subst_PROC_STATE_LEMMA)
           THENL
            [ ASM_REWRITE_TAC[];

              CALL_TAC
              THEN EXISTS_TAC "c:cmd"
              THEN EXISTS_TAC "vars:var list"
              THEN EXISTS_TAC "vals:var list"
              THEN EXISTS_TAC "ns:num list"
              THEN EXISTS_TAC "glbs:var list"
              THEN EXISTS_TAC "pre:aexp"
              THEN EXISTS_TAC "post:aexp"
              THEN EXISTS_TAC "calls:prog_env"
              THEN EXISTS_TAC "rec:aexp"
              THEN ASM_REWRITE_TAC[]
              THEN CONJ_TAC
              THENL
               [ MATCH_MP_TAC C_EXTENSION
                 THEN REWRITE_TAC (map (SYM o ASSUME)
                        ["vals' = variants vals(SL(APPEND xs glbs))";
                         "xs = l0 <|xs (ys //v xs')"])
                 THEN EXISTS_TAC "(calls <|g
   ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))"
                 THEN EXISTS_TAC
    "(BNDsL ns vals' (s2:state)) o
         (((APPEND xs vals') //v (APPEND vars vals)) o
          (\y. @z. ((APPEND l0 (variants vals(SL(APPEND l0 glbs)))) //v
                    (APPEND vars vals))z = y))"
                 THEN REPEAT CONJ_TAC
                 THENL %5 subgoals%
                    % WFc (c <|c ss2) g r %
                  [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                    THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                    THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                    THEN ASM_REWRITE_TAC[];

                    % WFenv r %
                    ASM_REWRITE_TAC[];

                    % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
                    ONCE_REWRITE_TAC[SL_APPEND_SYM23]
                    THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                    THEN ASM_REWRITE_TAC[];

                    % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
                      MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
                    MATCH_MP_TAC (SPEC_ALL MAP_EXTENSION_STATE_LEMMA)
                    THEN ASM_REWRITE_TAC[];

                    % 5th of 5 subgoals: %
                    % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
                    REWRITE_TAC[SYM(ASSUME "xs = l0 <|xs ss");
                                SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
                    THEN FIRST_ASSUM MATCH_MP_TAC
                    THEN EXISTS_TAC "(calls <|g
   ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))"
                    THEN REPEAT CONJ_TAC
                    THENL  % 5 subgoals: (of this 5th subgoal!) %
                     [ REWRITE_TAC[SYM(SPEC_ALL c_subst_o)]
                       THEN REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                       THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                       THEN ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[];

                       MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA)
                       THEN ASM_REWRITE_TAC[]
                     ]
                  ];

                 UNASSUME_THEN (ASSUME_TAC o SYM) "es = l1 <|es ss"
                 THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL)
                                   es_subst_biject
               ]
            ]
         ]
     ]
  );;

%< c_subst_LEMMA1:
g "!c' r s1 s2. 
      C c' r s1 s2 ==>
      (!c ss g. (c' = (c <|c ss))
            /\ WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss ==>
        C c r (s1 o ss) (s2 o ss))";;
e(CMD_STR_INDUCT_TAC);;

% SKIP: %
 e(GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(SKIP_TAC);;

% ASSIGN: %
 e(GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN ASM_REWRITE_THM BND_o_ONE_ONE_THM);;
 e(ASSIGN_TAC);;
 e(IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) e_subst_biject);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "e = e' <|e ss");;
 e(ASM_REWRITE_TAC[]);;

% SEQ: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(SEQ_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

% IF_T: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IF_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(RES_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
        "WFcsubst c r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% IF_F: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IF_F_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(RES_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
        "WFcsubst c r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% WHILE_T: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(WHILE_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(EXISTS_TAC "(s3:state) o (ss:vsubst)");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(RES_TAC);;
 e((STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
        "WFcsubst c'' r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[WFc_WHILE;c_subst;cmd_one_one;WFcsubst_THM]);;

% WHILE_F: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(WHILE_F_TAC);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e((STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
        "WFcsubst c'' r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% CALL: %
let st1 = get_state();;
set_state(st1);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = s");;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
             "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(IMP_RES2_THEN REWRITE_THM BNDsL_o_biject);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL c_subst_STATE_BNDsL_LEMMA));;

  e(ASM_REWRITE_TAC[MAP_MAP_o]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[MAP_I]);;

 e(USE_IMP_THEN REWRITE_THM
    ((SPEC_ALL o SPECL["variants vals (SL (APPEND l0 glbs))";"vals: var list"])
          c_subst_PROC_STATE_LEMMA));;

  e(ASM_REWRITE_TAC[]);;

let st2 = get_state();;
set_state(st2);;
 e(CALL_TAC);;
 e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:var list"
    THEN EXISTS_TAC "vals:var list"
    THEN EXISTS_TAC "ns:num list"
    THEN EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC C_EXTENSION);;
  e(REWRITE_TAC (map (SYM o ASSUME)
                 ["vals' = variants vals(SL(APPEND xs glbs))";
                  "xs = l0 <|xs (ys //v xs')"]));;
  e(EXISTS_TAC "(calls <|g
     ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))");;
  e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND l0 (variants vals(SL(APPEND l0 glbs)))) //v
                           (APPEND vars vals))z = y))");;
let st3 = get_state();;
set_state(st3);;
  e(REPEAT CONJ_TAC);; %5 subgoals%

   % WFc (c <|c ss2) r %
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   % WFenv r %
   e(ASM_REWRITE_TAC[]);;

   % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
   e(ONCE_REWRITE_TAC[SL_APPEND_SYM23]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
   e(MATCH_MP_TAC (SPEC_ALL MAP_EXTENSION_STATE_LEMMA));;
   e(ASM_REWRITE_TAC[]);;

   % 5th of 5 subgoals: %
   % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
let st4 = get_state();;
set_state(st4);;
   e(REWRITE_TAC[SYM(ASSUME "xs = l0 <|xs ss");
                 SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "(calls <|g
                    ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
                     (APPEND vars vals)))");;
   e(REPEAT CONJ_TAC);;  % 4 subgoals: %

    e(REWRITE_TAC[SYM(SPEC_ALL c_subst_o)]);;
    e(REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
    e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;

    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA));;
    e(ASM_REWRITE_TAC[]);;

  e(UNASSUME_THEN (ASSUME_TAC o SYM) "es = l1 <|es ss");;
  e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) es_subst_biject);;

>%

%
save_thm(`c_subst_LEMMA1`,c_subst_LEMMA1);;
%





let o_23_ASSOC =
 TAC_PROOF
  (([],
    "!(f:*4->*5) (g:*3->*4) (h:*2->*3) (i:*1->*2).
       ((f o g) o (h o i)) = (f o (g o h) o i)"
   ),
   REWRITE_TAC[o_ASSOC]
  );;

% o_23_ASSOC
g "!(f:*4->*5) (g:*3->*4) (h:*2->*3) (i:*1->*2).
       ((f o g) o (h o i)) = (f o (g o h) o i)";;
e(REWRITE_TAC[o_ASSOC]);;
%





let c_subst_LEMMA2 =
 TAC_PROOF
  (([],
    "!c r s1' s2'. 
      C c r s1' s2' ==>
      (!s1 s2 ss g.
           (s1' = (s1 o ss)) /\
           (s2' = (s2 o ss)) /\
           WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss  ==>
       C (c <|c ss) r s1 s2)"
   ),
   ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN CMD_STR_INDUCT_TAC
   THEN REWRITE_TAC[c_subst;WFc;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFcsubst_IMP_biject
   THEN IMP_RES2_THEN REWRITE_ALL_THM o_EQ_INVERSE
   THEN ASM_REWRITE_TAC[]
   THEN IMP_RES_TAC o_INVERSE
   THENL

        % SKIP: %
     [  SKIP_TAC;

        % ASSIGN: %
        IMP_RES2_THEN REWRITE_THM BND_o_biject1
        THEN ASSIGN_TAC
        THEN IMP_RES2_THEN REWRITE_THM e_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % SEQ: %
        SEQ_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst) z = y)"
        THEN CONJ_TAC
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % IF_T: %
        IF_T_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % IF_F: %
        IF_F_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % WHILE_T: %
        WHILE_T_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN EXISTS_TAC "(s3:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN CONJ_TAC
        THENL
         [ FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
           THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

           REWRITE_TAC[(SYM o ASSUME)"pr = ALESS v (AVAR x)"]
           THEN REWRITE_ALL_TAC[c_subst]
           THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
           THEN ASM_REWRITE_TAC[WFc_WHILE;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
           THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
         ];

        % WHILE_F: %
        WHILE_F_TAC
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % CALL: %
        REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
        THEN IMP_RES2_THEN REWRITE_THM BNDsL_o_biject1
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
                    "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN USE_IMP_THEN REWRITE_THM
               (SPECL["xs:var list";"xs <|xs ss"]
                  (GENL["xs:var list";"l0:var list";"vals0:var list";"ss:vsubst"]
                     (SPEC_ALL c_subst_STATE_BNDsL_LEMMA)))
        THENL
         [ ASM_REWRITE_TAC[MAP_MAP_o;SYM(SPEC_ALL xs_subst_o);
                           MAP_I;xs_subst_I;LENGTH_xs_subst]
           THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
           THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
           THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
           THEN BETA_TAC
           THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE
           THEN IMP_RES_TAC WFx_biject_INVERSE;

           USE_IMP_THEN REWRITE_THM
              ((SPEC_ALL o SPECL["variants vals (SL (APPEND (xs <|xs ss) glbs))";
                                 "vals: var list";"vals': var list";"xs: var list"])
                    c_subst_PROC_STATE_LEMMA)
           THENL
            [ ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;LENGTH_xs_subst]
              THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
              THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
              THEN BETA_TAC
              THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE;

              CALL_TAC
              THEN EXISTS_TAC "c:cmd"
              THEN EXISTS_TAC "vars:var list"
              THEN EXISTS_TAC "vals:var list"
              THEN EXISTS_TAC "ns:num list"
              THEN EXISTS_TAC "glbs:var list"
              THEN EXISTS_TAC "pre:aexp"
              THEN EXISTS_TAC "post:aexp"
              THEN EXISTS_TAC "calls:prog_env"
              THEN EXISTS_TAC "rec:aexp"
              THEN ASM_REWRITE_TAC[]
              THEN REWRITE_TAC[SYM(ASSUME
                          "vals' = variants vals(SL(APPEND xs glbs))")]
              THEN CONJ_TAC
              THENL
               [ MATCH_MP_TAC C_EXTENSION
                 THEN EXISTS_TAC "(calls <|g
   ((APPEND(xs <|xs ss)(variants vals(SL(APPEND(xs <|xs ss)glbs)))) //v
    (APPEND vars vals)))"
                 THEN EXISTS_TAC
               "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND (xs <|xs ss)
                            (variants vals(SL(APPEND (xs <|xs ss) glbs)))) //v
                           (APPEND vars vals))z = y))"
                 THEN REPEAT CONJ_TAC %5 subgoals%
                 THENL
               % WFc (c <|c ss2) g r %
                  [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                    THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                    THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                    THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                    THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst;

               % WFenv_syntax r %
                    ASM_REWRITE_TAC[];

               % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
                    ONCE_REWRITE_TAC[SL_APPEND_SYM23]
                    THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                    THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                    THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
                    MATCH_MP_TAC
                         (SPECL["es <|es ss"]
                            (GENL["l1:exp list";"xs:var list";"l0:var list";
                                  "vals'':var list";"ss:vsubst"]
                               (SPEC_ALL MAP_EXTENSION_STATE_LEMMA)))
                    THEN ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;
                                         SYM(SPEC_ALL es_subst_o);es_subst_I;
                                         LENGTH_xs_subst;LENGTH_es_subst]
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
                    THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
                    THEN BETA_TAC
                    THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE;

               % 5th of 5 subgoals: %
               % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
                    FIRST_ASSUM
                     (MATCH_MP_TAC o
                      REWRITE_RULE[SYM(SPEC_ALL c_subst_o);
                                   SYM(SPEC_ALL o_ASSOC);vsubst_INVERSE;I_o_ID] o
                      SPEC "calls <|g (APPEND xs vals' //v APPEND vars vals)" o
                      SPEC "(APPEND (xs <|xs ss)
                                    (variants vals(SL(APPEND (xs <|xs ss) glbs))) //v
                             APPEND vars vals) o
                            (\y. @z. (APPEND xs vals' //v APPEND vars vals)z = y)"  o
                      SPECL["s1':state";"s2':state"])
                    THEN REPEAT CONJ_TAC
                    THENL  % 5 subgoals: %
                     [ ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[o_23_ASSOC]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[o_23_ASSOC]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                       THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                       THEN ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[]
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA)
                       THEN ASM_REWRITE_TAC[]
                     ]
                  ];

                 IMP_RES2_THEN REWRITE_THM es_subst_biject
                 THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
               ]
            ]
         ]
     ]
  );;

%< c_subst_LEMMA2:
g "!c r s1' s2'. 
      C c r s1' s2' ==>
      (!s1 s2 ss g.
           (s1' = (s1 o ss)) /\
           (s2' = (s2 o ss)) /\
           WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss  ==>
       C (c <|c ss) r s1 s2)";;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]
  THEN CMD_STR_INDUCT_TAC
  THEN REWRITE_TAC[c_subst;WFc;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN IMP_RES_TAC WFcsubst_IMP_biject
  THEN IMP_RES2_THEN REWRITE_ALL_THM o_EQ_INVERSE
  THEN ASM_REWRITE_TAC[]
 );;

% SKIP: %
 e(SKIP_TAC);;

% ASSIGN: %
 e(IMP_RES2_THEN REWRITE_THM BND_o_biject1);;
 e(ASSIGN_TAC);;
 e(IMP_RES2_THEN REWRITE_THM e_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% SEQ: %
 e(SEQ_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst) z = y)");;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% IF_T: %
 e(IF_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% IF_F: %
 e(IF_F_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% WHILE_T: %
 e(WHILE_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(EXISTS_TAC "(s3:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

  e(REWRITE_TAC[(SYM o ASSUME) "pr = ALESS v (AVAR x)"]);;
  e(REWRITE_ALL_TAC[c_subst]);;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[WFc_WHILE;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% WHILE_F: %
 e(WHILE_F_TAC);;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% CALL: %
let st1 = get_state();;
set_state st1;;
 e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
 e(IMP_RES2_THEN REWRITE_THM BNDsL_o_biject1);;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
             "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(USE_IMP_THEN REWRITE_THM
     (SPECL["xs:var list";"xs <|xs ss"]
        (GENL["xs:var list";"l0:var list";"vals0:var list";"ss:vsubst"]
           (SPEC_ALL c_subst_STATE_BNDsL_LEMMA))));;

  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[MAP_MAP_o;SYM(SPEC_ALL xs_subst_o);
                    MAP_I;xs_subst_I;LENGTH_xs_subst]);;
  e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
  e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
  e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
  e(BETA_TAC);;
  e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;
  e(IMP_RES_TAC WFx_biject_INVERSE);;

let st3 = get_state();;
set_state(st3);;
 e(USE_IMP_THEN REWRITE_THM
    ((SPEC_ALL o SPECL["variants vals (SL (APPEND (xs <|xs ss) glbs))";
                       "vals: var list";"vals': var list";"xs: var list"])
          c_subst_PROC_STATE_LEMMA));;

  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;LENGTH_xs_subst]);;
  e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
  e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
  e(BETA_TAC);;
  e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;

 e(CALL_TAC);;
 e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:var list"
    THEN EXISTS_TAC "vals:var list"
    THEN EXISTS_TAC "ns:num list"
    THEN EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC C_EXTENSION);;
  e(EXISTS_TAC "(calls <|g
      ((APPEND(xs <|xs ss)(variants vals(SL(APPEND(xs <|xs ss)glbs)))) //v
       (APPEND vars vals)))");;
  e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND (xs <|xs ss)
                            (variants vals(SL(APPEND (xs <|xs ss) glbs)))) //v
                           (APPEND vars vals))z = y))");;
let st5 = get_state();;
set_state(st5);;
  e(REPEAT CONJ_TAC);; %5 subgoals%

   % WFc (c <|c ss2) r %
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
   e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;

   % WFenv r %
   e(ASM_REWRITE_TAC[]);;

   % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
   e(ONCE_REWRITE_TAC[SL_APPEND_SYM23]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
   e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
   e(MATCH_MP_TAC
     (SPECL["es <|es ss"]
        (GENL["l1:exp list";"xs:var list";"l0:var list";
              "vals'':var list";"ss:vsubst"]
           (SPEC_ALL MAP_EXTENSION_STATE_LEMMA))));;
   e(IMP_RES_TAC o_INVERSE);;
   e(ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;
                     SYM(SPEC_ALL es_subst_o);es_subst_I;
                     LENGTH_xs_subst;LENGTH_es_subst]);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
   e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
   e(BETA_TAC);;
   e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;

   % 5th of 5 subgoals: %
   % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
let st6 = get_state();;
set_state(st6);;
   e(FIRST_ASSUM
      (MATCH_MP_TAC o pthm o
       REWRITE_RULE[SYM(SPEC_ALL c_subst_o);
                    SYM(SPEC_ALL o_ASSOC);vsubst_INVERSE;I_o_ID] o pthm o
       SPEC "calls <|g (APPEND xs vals' //v APPEND vars vals)" o
       SPEC "(APPEND (xs <|xs ss)
                     (variants vals(SL(APPEND (xs <|xs ss) glbs))) //v
              APPEND vars vals) o
             (\y. @z. (APPEND xs vals' //v APPEND vars vals)z = y)"  o pthm o
       SPECL["s1':state";"s2':state"]));;
   e(REPEAT CONJ_TAC);;  % 5 subgoals: %

    e(ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[o_23_ASSOC]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[o_23_ASSOC]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
    e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA));;
    e(ASM_REWRITE_TAC[]);;

  e(IMP_RES2_THEN REWRITE_THM es_subst_biject);;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

>%


%
save_thm(`c_subst_LEMMA2`,c_subst_LEMMA2);;
%





let c_subst_biject =
 prove_thm
  (`c_subst_biject`,
   "!c g r ss s1 s2. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r ss ==>
     (C (c <|c ss) r s1 s2 = C c r (s1 o ss) (s2 o ss))",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [ DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_subst_LEMMA1))
       THEN ASM_REWRITE_TAC[];

       DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_subst_LEMMA2))
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% c_subst_biject
g "!c g r ss s1 s2. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r ss ==>
     (C (c <|c ss) r s1 s2 = C c r (s1 o ss) (s2 o ss))";;

e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

  e(DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_subst_LEMMA1)));;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_subst_LEMMA2)));;
  e(ASM_REWRITE_TAC[]);;
%



let c_subst_THM =
 prove_thm
  (`c_subst_THM`,
   "!c g r ys xs s1 s2. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) ==>
     (C (c <|c (ys //v xs)) r s1 s2 =
      C c r (s1 o (ys //v xs)) (s2 o (ys //v xs)))",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL c_subst_biject)
   THEN ASM_REWRITE_TAC[]
  );;

% c_subst_THM
g "!c g r ys xs s1 s2. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) ==>
     (C (c <|c (ys //v xs)) r s1 s2 =
      C c r (s1 o (ys //v xs)) (s2 o (ys //v xs)))";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL c_subst_biject));;
e(ASM_REWRITE_TAC[]);;
%





let MAP_WFsubst_logicals =
 prove_thm
  (`MAP_WFsubst_logicals`,
   "!xs ss. WFsubst ss /\ WFxs xs ==>
     (MAP ss (logicals xs) = logicals (MAP ss xs))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;logicals_THM;MAP]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN REWRITE_ALL_TAC[WFsubst]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_WFsubst_logicals
g "!xs ss. WFsubst ss /\ WFxs xs ==>
     (MAP ss (logicals xs) = logicals (MAP ss xs))";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[WFxs;logicals_THM;MAP]);;

e(REWRITE_TAC[WFxs;logicals_THM;MAP]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(REWRITE_ALL_TAC[WFsubst]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let mk_eqs_vsubst =
 prove_thm
  (`mk_eqs_vsubst`,
   "!y x ss. ((mk_eqs y x) <|av ss = mk_eqs (MAP ss y) (MAP ss x))",
   DOUBLE_LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[mk_eqs_THM;MAP;a_vsubst;v_vsubst]
  );;

% mk_eqs_vsubst
g "!y x ss. ((mk_eqs y x) <|av ss = mk_eqs (MAP ss y) (MAP ss x))";;
e(DOUBLE_LIST_INDUCT_TAC);;
e(REWRITE_TAC[mk_eqs_THM;MAP;a_vsubst]);;

e(REWRITE_TAC[mk_eqs_THM;MAP;a_vsubst]);;

e(REWRITE_TAC[mk_eqs_THM;MAP;a_vsubst]);;
e(ASM_REWRITE_TAC[v_vsubst]);;
%




let PCSPEC_biject =
 prove_thm
  (`PCSPEC_biject`,
   "!ss x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (PCSPEC(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                  (pre <|av ss),
                  c    <|c  ss,
                  post <|av ss,
                  r) =
      PCSPEC(AAND (mk_eqs (logicals x) x) pre, c, post, r))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFcsubst_IMP_biject
   THEN REWRITE_TAC[PCSPEC;A_DEF]
   THEN IMP_RES2_THEN (REWRITE_THM o SYM) MAP_WFsubst_logicals
   THEN REWRITE_TAC[(SYM o SPEC_ALL) mk_eqs_vsubst]
   THEN IMP_RES_THEN REWRITE_THM a_vsubst_ONE_ONE
   THEN IMP_RES_TAC c_subst_biject
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN EQ_TAC
   THENL
     [  REPEAT STRIP_TAC
        THEN IMP_RES_TAC o_INVERSE
        THEN POP_ASSUM (\th1. POP_ASSUM (\th2.
                FIRST_ASSUM (MATCH_MP_TAC o
                     REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
                         SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)";
                               "(s2:state) o (\y. @z. (ss:vsubst)z = y)"])))
        THEN IMP_RES_TAC o_INVERSE
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "(s1:state) o (ss:vsubst)"
        THEN ASM_REWRITE_TAC[]
     ]
  );;


let TCSPEC_biject =
 prove_thm
  (`TCSPEC_biject`,
   "!ss x pre c g r.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (TCSPEC(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                  (pre <|av ss),
                  c    <|c  ss,
                  r) =
      TCSPEC(AAND (mk_eqs (logicals x) x) pre, c, r))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFcsubst_IMP_biject
   THEN REWRITE_TAC[TCSPEC;A_DEF]
   THEN IMP_RES2_THEN (REWRITE_THM o SYM) MAP_WFsubst_logicals
   THEN REWRITE_TAC[(SYM o SPEC_ALL) mk_eqs_vsubst]
   THEN IMP_RES_THEN REWRITE_THM a_vsubst_ONE_ONE
   THEN IMP_RES_TAC c_subst_biject
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN EQ_TAC
   THENL
     [  REPEAT STRIP_TAC
        THEN IMP_RES_TAC o_INVERSE
        THEN POP_ASSUM (\th1. POP_ASSUM (\th2.
                FIRST_ASSUM (IMP_RES_TAC o
                     REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
                         SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)"])))
        THEN POP_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN ASM_REWRITE_TAC[];

        REPEAT STRIP_TAC
        THEN RES_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES_TAC o_INVERSE
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
     ]
  );;


let CSPEC_biject =
 prove_thm
  (`CSPEC_biject`,
   "!ss x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (CSPEC(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                 (pre <|av ss),
                 c    <|c  ss,
                 post <|av ss,
                 r) =
      CSPEC(AAND (mk_eqs (logicals x) x) pre, c, post, r))",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[CSPEC]
   THEN MATCH_MP_TAC CONJ_11
   THEN CONJ_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL PCSPEC_biject)
        THEN ASM_REWRITE_TAC[];

        MATCH_MP_TAC (SPEC_ALL TCSPEC_biject)
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% CSPEC_biject
g "!ss x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (CSPEC(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                 (pre <|av ss),
                 c    <|c  ss,
                 post <|av ss,
                 r) =
      CSPEC(AAND (mk_eqs (logicals x) x) pre, c, post, r))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFcsubst_IMP_biject);;
e(REWRITE_TAC[CSPEC;A_DEF]);;
e(IMP_RES2_THEN (REWRITE_THM o SYM) MAP_WFsubst_logicals);;
e(REWRITE_TAC[(SYM o SPEC_ALL) mk_eqs_vsubst]);;
e(IMP_RES_THEN REWRITE_THM a_vsubst_ONE_ONE);;
e(IMP_RES_TAC c_subst_biject);;
e(POP_TAC THEN POP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;

   e(IMP_RES_TAC o_INVERSE);;
   e(POP_ASSUM (\th1. POP_ASSUM (\th2.
     FIRST_ASSUM (MATCH_MP_TAC o
         REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
         SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)";
               "(s2:state) o (\y. @z. (ss:vsubst)z = y)"]))));;
   e(IMP_RES_TAC o_INVERSE);;
   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

   e(IMP_RES_TAC o_INVERSE);;
   e(POP_ASSUM (\th1. POP_ASSUM (\th2.
     FIRST_ASSUM (IMP_RES_TAC o
         REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
         SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)"]))));;
   e(POP_TAC);;
   e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
   e(ASM_REWRITE_TAC[]);;

  e(REPEAT STRIP_TAC);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "(s1:state) o (ss:vsubst)");;
   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;
   e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
   e(IMP_RES_TAC o_INVERSE);;
   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
%


let CSPEC_vsubst =
 prove_thm
  (`CSPEC_vsubst`,
   "!ys xs x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) /\
     WFsubst (ys //v xs) /\ WFxs x ==>
     (CSPEC(AAND (mk_eqs (logicals (MAP (ys //v xs) x))
                         (MAP (ys //v xs) x))
                 (pre <|av (ys //v xs)),
                 c    <|c  (ys //v xs),
                 post <|av (ys //v xs),
                 r) =
      CSPEC(AAND (mk_eqs (logicals x) x) pre, c, post, r))",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL CSPEC_biject)
   THEN ASM_REWRITE_TAC[]
  );;

% CSPEC_vsubst
g "!ys xs x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) /\
     WFsubst (ys //v xs) /\ WFxs x ==>
     (CSPEC(AAND (mk_eqs (logicals (MAP (ys //v xs) x))
                         (MAP (ys //v xs) x))
                 (pre <|av (ys //v xs)),
                 c    <|c  (ys //v xs),
                 post <|av (ys //v xs),
                 r) =
      CSPEC(AAND (mk_eqs (logicals x) x) pre, c, post, r))";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL CSPEC_biject));;
e(ASM_REWRITE_TAC[]);;
%





let A_mk_eqs_NOT_WFx_WFx =
 prove_thm
  (`A_mk_eqs_NOT_WFx_WFx`,
   "!x y s1. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ NOT_WFxs y ==>
             A (mk_eqs y x) (s1 <|s (x // y))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;mk_eqs_THM;A_DEF]
   THEN REWRITE_TAC[DISJOINT_LIST;WFxs;NOT_WFxs;o_THM;V_DEF;var_list_THM]
   THEN REWRITE_TAC[LENGTH;INV_SUC_EQ]
   THEN REPEAT STRIP_TAC
   THENL
     [
       REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]
       THEN IMP_RES_TAC WFx_NOT_WFx
       THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
       THEN REWRITE_TAC[s_subst_o;o_THM]
       THEN IMP_RES_TAC WFx_NOT_WFxs
       THEN IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
       THEN REWRITE_TAC[I_THM];

       REWRITE_TAC[s_subst_BND;V_DEF]
       THEN USE_IMP_THEN REWRITE_THM A_BND_THM
       THENL
         [
           USE_IMP_THEN REWRITE_THM FVa_mk_eqs
           THENL
             [  ASM_REWRITE_TAC[];

                REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
                THEN IMP_RES_TAC WFxs_NOT_WFx
                THEN ASM_REWRITE_TAC[]
             ];

           FIRST_ASSUM MATCH_MP_TAC
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

% A_mk_eqs_NOT_WFx_WFx
g "!x y s1. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ NOT_WFxs y ==>
             A (mk_eqs y x) (s1 <|s (x // y))";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[logicals_THM;mk_eqs_THM;A_DEF]);;

e(REWRITE_TAC[logicals_THM;mk_eqs_THM;A_DEF]);;

e(REWRITE_TAC[logicals_THM;mk_eqs_THM;A_DEF]);;
e(REWRITE_TAC[DISJOINT_LIST;WFxs;NOT_WFxs;o_THM;V_DEF;var_list_THM]);;
e(REWRITE_TAC[LENGTH;INV_SUC_EQ]);;
e(REPEAT STRIP_TAC);;

  e(REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]);;
  e(IMP_RES_TAC WFx_NOT_WFx);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(REWRITE_TAC[s_subst_o;o_THM]);;
  e(IMP_RES_TAC WFx_NOT_WFxs);;
  e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
  e(REWRITE_TAC[I_THM]);;

  e(REWRITE_TAC[s_subst_BND;V_DEF]);;
  e(USE_IMP_THEN REWRITE_THM A_BND_THM);;

    e(USE_IMP_THEN REWRITE_THM FVa_mk_eqs);;

      e(ASM_REWRITE_TAC[]);;

      e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
      e(IMP_RES_TAC WFxs_NOT_WFx);;
      e(ASM_REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let WF_proc_free_vars =
 TAC_PROOF
  (([], "!r p vars vals glbs pre post calls rec c x.
      WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      ~x IN (SL vars) /\ ~x IN (SL vals) /\ ~x IN (SL glbs) ==>
      ~x IN (FVc c r)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (REWRITE_RULE[SL_APPEND;SUBSET_DEF;IN_UNION]
                        WELL_FORMED_PROCEDURE_SYNTAX)
   THEN RES_TAC
  );;

%
g "!r p vars vals glbs pre post calls rec c x.
      WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      ~x IN (SL vars) /\ ~x IN (SL vals) /\ ~x IN (SL glbs) ==>
      ~x IN (FVc c r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (REWRITE_RULE[SL_APPEND;SUBSET_DEF;IN_UNION]
                WELL_FORMED_PROCEDURE_SYNTAX));;
e(RES_TAC);;
%


% ============================================================== %
% Here is the C_BND_THM that we could not prove earlier.  It is  %
% actually rather subtle, relying on a proof involving two cases %
% whether or not a variable z which is perturbing the states s1  %
% and s2 is one of the variants of the value formal parameters.  %
% ============================================================== %




let C_BND_THM =
 prove_thm
  (`C_BND_THM`,
   "!c r s1 s2.
     C c r s1 s2 ==>
     (!g x n. WFc c g r /\ WFenv_syntax r /\ (~x IN FVc c r) ==>
       C c r (BND n x s1) (BND n x s2))",
   let exists_tac = 
          EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:(var)list"
     THEN EXISTS_TAC "vals:(var)list"
     THEN EXISTS_TAC "ns:(num)list"
     THEN EXISTS_TAC "glbs:(var)list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp"
   in
   CMD_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc;FVc_DEF]
   THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THENL
     [
        SKIP_TAC;

        IMP_RES_TAC E_BND_THM
        THEN IMP_RES_THEN REWRITE_THM BND_COMMUTE
        THEN ASSIGN_TAC
        THEN ASM_REWRITE_TAC[];

        RES_TAC
        THEN SEQ_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN RES_TAC
        THEN IF_T_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN RES_TAC
        THEN IF_F_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN ASM_REWRITE_TAC[];

        UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')"
        THEN REWRITE_ALL_TAC[WFc_WHILE;FVc_DEF;IN_UNION;DE_MORGAN_THM]
        THEN IMP_RES_TAC B_BND_THM
        THEN WHILE_T_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN EXISTS_TAC "BND n x (s3:state)"
        THEN ASM_REWRITE_TAC[]
        THEN CONJ_TAC
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN WHILE_F_TAC
        THEN ASM_REWRITE_TAC[];

        UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                      "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN UNASSUME_THEN ASSUME_TAC
                      "LENGTH (vals:(var)list) = LENGTH (es:(exp)list)"
        THEN IMP_RES_TAC ES_LENGTH_THM
        THEN POP_ASSUM (\en. POP_ASSUM (\ve. ASSUME_TAC (SYM (TRANS ve en))))
        THEN (ASSUME_TAC o SYM o
              REWRITE_RULE[SYM(ASSUME
                      "vals' = variants vals(SL(APPEND xs glbs))")] o
              SPECL["vals:var list";"SL(APPEND xs glbs: var list)"])
                           LENGTH_variants
        THEN POP_ASSUM (\vv'. POP_ASSUM (\nv. ASSUME_TAC (TRANS nv vv')))
        THEN IMP_RES_TAC ES_BND_THM
        THEN ASM_CASES_TAC "(x:var) IN SL vals'"
        THENL
          [
            IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BNDsL_MAP_BND_COMMUTE
            THEN CALL_TAC
            THEN exists_tac
            THEN IMP_RES2_THEN REWRITE_THM BNDsL_BND_ABSORPTION
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE
            THEN IMP_RES_THEN (SUBST1_TAC o SYM o ISPECL["n:num";"s2:state"])
                              MAP_BND_ABSORPTION
            THEN CALL_TAC
            THEN exists_tac
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
            THEN IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsL_COMMUTE
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "(calls <|g ((APPEND xs vals') //v (APPEND vars vals)))"
            THEN REPEAT CONJ_TAC
            THENL  % 3 subgoals %
              [  MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
                 THEN ASM_REWRITE_TAC[];

                 ASM_REWRITE_TAC[];

                 MATCH_MP_TAC NOT_IN_SUBSET
                 THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):var set"
                 THEN REWRITE_ALL_TAC[FVxs_SL]
                 THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                 THEN ASM_REWRITE_TAC[]
                 THEN REWRITE_TAC[SYM(SPEC_ALL SL_APPEND);
                       SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
                 THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                 THEN ASM_REWRITE_TAC[]
              ]
          ] 
      ]
  );;

%< C_BND_THM
g "!c r s1 s2.
     C c r s1 s2 ==>
     (!g x n. WFc c g r /\ WFenv_syntax r /\ (~x IN FVc c r) ==>
       C c r (BND n x s1) (BND n x s2))";;
e(CMD_STR_INDUCT_TAC
  THEN REWRITE_TAC[WFc;FVc_DEF]
  THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(SKIP_TAC);;

 e(IMP_RES_TAC E_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM BND_COMMUTE);;
 e(ASSIGN_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(SEQ_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IF_T_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IF_F_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')");;
 e(REWRITE_ALL_TAC[WFc_WHILE;FVc_DEF;IN_UNION;DE_MORGAN_THM]);;
 e(IMP_RES_TAC B_BND_THM);;
 e(WHILE_T_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(EXISTS_TAC "BND n x (s3:state)");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(WHILE_F_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
      "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(UNASSUME_THEN ASSUME_TAC "LENGTH (vals:(var)list) = LENGTH (es:(exp)list)");;
 e(IMP_RES_TAC ES_LENGTH_THM);;
 e(POP_ASSUM (\en. POP_ASSUM (\ve. ASSUME_TAC (SYM (TRANS ve en)))));;
 e((ASSUME_TAC o SYM o
    REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
    SPECL["vals:var list";"SL(APPEND xs glbs: var list)"])
                 LENGTH_variants);;
 e(POP_ASSUM (\vv'. POP_ASSUM (\nv. ASSUME_TAC (TRANS nv vv'))));;
 e(ASM_CASES_TAC "(x:var) IN SL vals'");;

  e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BNDsL_MAP_BND_COMMUTE);;
  e(CALL_TAC);;
  e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:(var)list"
    THEN EXISTS_TAC "vals:(var)list"
    THEN EXISTS_TAC "ns:(num)list"
    THEN EXISTS_TAC "glbs:(var)list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
  e(IMP_RES2_THEN REWRITE_THM BNDsL_BND_ABSORPTION);;
  e(IMP_RES_TAC ES_BND_THM);;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE);;
  e(IMP_RES_THEN (SUBST1_TAC o SYM o ISPECL["n:num";"s2:state"])
                 MAP_BND_ABSORPTION);;
  e(CALL_TAC);;
  e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:(var)list"
    THEN EXISTS_TAC "vals:(var)list"
    THEN EXISTS_TAC "ns:(num)list"
    THEN EXISTS_TAC "glbs:(var)list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
  e(IMP_RES_TAC ES_BND_THM);;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
  e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsL_COMMUTE);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "(calls <|g ((APPEND xs vals') //v (APPEND vars vals)))");;
  e(REPEAT CONJ_TAC);;

   e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;

   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):var set");;
   e(REWRITE_ALL_TAC[FVxs_SL]);;
   e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_TAC[SYM(SPEC_ALL SL_APPEND);
            SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;
>%






let C_BNDsL_THM =
 prove_thm
  (`C_BNDsL_THM`,
   "!ns xs c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
      C c r s1 s2 ==> C c r (BNDsL ns xs s1) (BNDsL ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC C_BND_THM
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h:num")
   THEN RES_TAC
  );;

% C_BNDsL_THM:
g "!ns xs c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
      C c r s1 s2 ==> C c r (BNDsL ns xs s1) (BNDsL ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT STRIP_TAC);;
e(IMP_RES_TAC C_BND_THM);;
e(POP_ASSUM (ASSUME_TAC o SPEC "h:num"));;
e(RES_TAC);;
%

let C_BNDsR_THM =
 prove_thm
  (`C_BNDsR_THM`,
   "!ns xs c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
     C c r s1 s2 ==> C c r (BNDsR ns xs s1) (BNDsR ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC C_BND_THM
   THEN ASM_REWRITE_TAC[]
  );;

% C_BNDsR_THM:
g "!ns xs c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
     C c r s1 s2 ==> C c r (BNDsR ns xs s1) (BNDsR ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC C_BND_THM);;
e(ASM_REWRITE_TAC[]);;
%




close_theory ();;

