% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% An initial version of a semantics for a small language;
  First given in a structural operational style;
  then expressed as an axiomatic semantics, which is
  proven from the structural operational semantics,
  then the weakest derivable precondition functions
  are defined and proven sound.

  This file contains example programs for this simple programming
  language, and code to apply the verification condition generator
  to them.

  By Peter V. Homeier, November 4, 1991.
%

set_flag(`timing`,true);;

set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Load the theory.							%
% --------------------------------------------------------------------- %
load_theory `vcg`;;


% --------------------------------------------------------------------- %
% Need the string and finite_sets libraries.                            %
% --------------------------------------------------------------------- %
loadt `load_string`;;
load_string();;

loadt `load_finite_sets`;;
load_finite_sets();;

%<
% Set up autoloading from all theories of vcg %

loadf `autoload`;;

autoload_all_defs_and_thms
     `string bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules semantic_stages
      stage_semantics termination vcg`;;
>%


loadf `semantics_tactics`;;

% ----------------------- %

% load parser to translate strings containing concrete program text into
  the corresponding abstract syntax terms in the HOL object language
%
loadf `vcg_parser`;;


% load pretty-printer to output abstract syntax terms as the corresponding
  programming language text.
%
loadf `vcg_prettyp`;;

% ----------------------- %

%
save `vcg-hol`;;
compilef `vcg_tactics`;;
Run time: 831.2s
Garbage collection time: 31.6s
Intermediate theorems generated: 7456
%
% Under 2 minutes (115.5 sec) to load if compiled: %
loadf `vcg_tactics`;;

set_flag(`print_vcg`,true);;
assignable_print_vcg_term := print_vcg_term;;




let SUBTRACT_CONV =
  (\tm. let (mns,[a;b]) = strip_comb tm in
        (if not (mns = "$-") or not is_const a or not is_const b then fail);
        ALL_CONV tm) THENC
  REWRITE_CONV[SUB_0] THENC
  REPEATC
    ((RATOR_CONV o RAND_CONV) num_CONV THENC
     RAND_CONV num_CONV THENC
     PURE_REWRITE_CONV[SUB_MONO_EQ;SUB_0]);;

%
SUBTRACT_CONV "12 - 2";;
SUBTRACT_CONV "101 - 10";;
%

let LESS_CONV =
  (\tm. let (lss,[a;b]) = strip_comb tm in
        (if not (lss = "$<") or not is_const a or not is_const b then fail);
        ALL_CONV tm) THENC
  REWRITE_CONV[NOT_LESS_0] THENC
  REPEATC
    (RAND_CONV num_CONV THENC
     PURE_REWRITE_CONV[LESS_0] THENC
     (RATOR_CONV o RAND_CONV) num_CONV THENC
     PURE_REWRITE_CONV[LESS_MONO_EQ;NOT_LESS_0]) THENC
  ((RAND_CONV num_CONV THENC
     PURE_REWRITE_CONV[LESS_0]) ORELSEC ALL_CONV);;

%
LESS_CONV "12 < 2";;
LESS_CONV "101 < 10";;
LESS_CONV "2 < 12";;
LESS_CONV "10 < 101";;
LESS_CONV "101 < 101";;
LESS_CONV "0 < 0";;
%



% --------------------------------------------------------------------- %
% Define the first example program: quotient-remainder.			%
% --------------------------------------------------------------------- %

g [[ program
        procedure quotient_remainder (var q,r; val x,y);
           pre  0 < y;
           post ^x = q * ^y + r /\ r < ^y;

           r := x;
           q := 0;
           assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
             with r < ^r
           while ~(r < y) do
              r := r - y;
              q := ++q
           od
        end procedure;

        quotient_remainder(q,r;7,3)
     end program
     [ q = 2 /\ r = 1 ]
  ]];;

e(VCG_TAC);;

% 920.1s = 15 min, 20.1 sec %

%
OK..
For procedure `quotient_remainder`,

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r}
       q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y
      {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "SEQ" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y; q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "WHILE" rule, we have
   [[ {^x = q * ^y + r /\ 0 < y /\ ^y = y}
       assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
         with r < ^r
       while ~(r < y) do
          r := r - y; q := ++q
       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + r /\ 0 < y /\ ^y = y}
       q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x
      {^x = 0 * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By precondition strengthening, we have
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with additional verification condition
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y ==>
       ^x = 0 * ^y + x /\ 0 < y /\ ^y = y} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure quotient_remainder:

By the call graph progress from procedure quotient_remainder to quotient_remainder, we have
   [[ {0 < y /\ true} quotient_remainder-<>->quotient_remainder {false} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(0 < 3 /\ true) /\
       (!q r x1 y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)}
       quotient_remainder(q,r;7,3)
      {q = 2 /\ r = 1} ]]

By precondition strengthening, we have
   [[ {true} quotient_remainder(q,r;7,3) {q = 2 /\ r = 1} ]]
with additional verification condition
   [[ {true ==> (0 < 3 /\ true) /\
                (!q r x1 y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)} ]]

4 subgoals
"0 < 3 /\
 (!q r x1 y1. (7 = (q * 3) + r) /\ r < 3 ==> (q = 2) /\ (r = 1))"

"!^x q ^y r y.
  ((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ r < y ==>
  (^x = (q * ^y) + r) /\ r < ^y"

"!^x q ^y r y ^r.
  (((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ ~r < y) /\ (r = ^r) ==>
  ((^x = ((q + 1) * ^y) + (r - y)) /\ 0 < y /\ (^y = y)) /\ (r - y) < ^r"

"!^q q ^r r ^x x ^y y.
  ((^q = q) /\ (^r = r) /\ (^x = x) /\ (^y = y)) /\ 0 < y ==>
  (^x = (0 * ^y) + x) /\ 0 < y /\ (^y = y)"

() : void
Run time: 827.5s
Garbage collection time: 92.6s
Intermediate theorems generated: 53837

%

%<

% 1st subgoal (head of quotient_remainder: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(REWRITE_TAC[MULT_CLAUSES]);;
e(REWRITE_TAC[ADD_CLAUSES]);;

or%

e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN REWRITE_TAC[MULT_CLAUSES]
  THEN REWRITE_TAC[ADD_CLAUSES]);;

% 2nd subgoal (loop invariant): %

%e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(POP_ASSUM (REWRITE_THM o SYM));;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS]));;
e(REWRITE_TAC[SUB_RIGHT_LESS]);;
e(IMP_RES2_THEN REWRITE_THM LESS_LESS_EQ_TRANS);;
e(IMP_RES_THEN (REWRITE_THM o ONCE_REWRITE_RULE[ADD_SYM] o
                  MATCH_MP LESS_ADD_NONZERO o NOT_EQ_SYM)
               LESS_NOT_EQ);;
e(ASM_REWRITE_TAC[]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1;MULT_CLAUSES]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]);;
e(CONV_TAC ((RAND_CONV o RAND_CONV) (REWR_CONV ADD_SYM)));;
e(IMP_RES_THEN REWRITE_THM SUB_ADD);;%

e(REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN POP_ASSUM (REWRITE_THM o SYM)
  THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS])
  THEN REWRITE_TAC[SUB_RIGHT_LESS]
  THEN IMP_RES2_THEN REWRITE_THM LESS_LESS_EQ_TRANS
  THEN IMP_RES_THEN (REWRITE_THM o ONCE_REWRITE_RULE[ADD_SYM] o
                  MATCH_MP LESS_ADD_NONZERO o NOT_EQ_SYM)
               LESS_NOT_EQ
  THEN ASM_REWRITE_TAC[]
  THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD1;MULT_CLAUSES]
  THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]
  THEN CONV_TAC ((RAND_CONV o RAND_CONV) (REWR_CONV ADD_SYM))
  THEN IMP_RES_THEN REWRITE_THM SUB_ADD);;

% 3rd subgoal (loop exit): %

%e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[]);;

% 4th subgoal (main body): %

%e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV));;
e(REWRITE_TAC[]);;
e(REPEAT GEN_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (q * 3) + r) /\ r < 3","r:num")));;
e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE));;
e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + r) /\ r < 3","q:num")));;
e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE));;
e(SUBGOAL_THEN "(7 = (2 * 3) + 1) /\ 1 < 3" ASSUME_TAC);;

 e(CONV_TAC (DEPTH_CONV LESS_CONV));;
 e(REWRITE_TAC[num_CONV "7";num_CONV "1"]);;
 e(REWRITE_TAC[ADD_CLAUSES;INV_SUC_EQ]);;
 e(REWRITE_TAC[num_CONV "2"]);;
 e(REWRITE_TAC[MULT_CLAUSES]);;
 e(CONV_TAC (DEPTH_CONV ADD_CONV));;
 e(REWRITE_TAC[]);;

 e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (2 * 3) + r) /\ r < 3","1:num")));;
 e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE));;
 e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + 1) /\ 1 < 3","2:num")));;
 e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE));;%

e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV)
  THEN REWRITE_TAC[]
  THEN REPEAT GEN_TAC
  THEN DISCH_TAC
  THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (q * 3) + r) /\ r < 3","r:num"))
  THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE)
  THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + r) /\ r < 3","q:num"))
  THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE)
  THEN SUBGOAL_THEN "(7 = (2 * 3) + 1) /\ 1 < 3" ASSUME_TAC
  THENL
   [ 
     CONV_TAC (DEPTH_CONV LESS_CONV)
     THEN REWRITE_TAC[num_CONV "7";num_CONV "1"]
     THEN REWRITE_TAC[ADD_CLAUSES;INV_SUC_EQ]
     THEN REWRITE_TAC[num_CONV "2"]
     THEN REWRITE_TAC[MULT_CLAUSES]
     THEN CONV_TAC (DEPTH_CONV ADD_CONV)
     THEN REWRITE_TAC[];

     FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (2 * 3) + r) /\ r < 3","1:num"))
     THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE)
     THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + 1) /\ 1 < 3","2:num"))
     THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE)
  ]);;

let quotient_remainder = top_thm();;
print_vcg_thm quotient_remainder;;
%
|- [[ program
         procedure quotient_remainder(q,r;x,y);
            global ;
            pre  0 < y;
            post ^x = q * ^y + r /\ r < ^y;
            recurses with false;

            r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                              with r < ^r
                            while ~(r < y) do
                               r := r - y; q := ++q
                            od
         end procedure;

         quotient_remainder(q,r;7,3)
      end program
      [q = 2 /\ r = 1] ]]
%

>%



% --------------------------------------------------------------------- %
% Define the second example program: McCarthy's "91 function".          %
% --------------------------------------------------------------------- %

g [[ program
        procedure p91(var x; val y);
           pre  true;
           post 100 < ^y  =>  x = ^y - 10  |  x = 91;
           calls p91 with 101 - y < 101 - ^y;
           recurses  with 101 - y < ^z;

           if 100 < y then x := y - 10
           else
              p91(x; y + 11);
              p91(x; x)
           fi
        end procedure;

        p91(a; 77)

     end program
     [ a = 91 ]
  ]];;

e(VCG_TAC);;

% 1401.2s = 23 min, 21.2 sec %

%<
#e(VCG_TAC);;
OK..
For procedure `p91`,

By the "ASSIGN" rule, we have
   [[ {(100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)}
       x := y - 10
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "CALL" rule, we have
   [[ {(true /\ 101 - x < 101 - ^y) /\
       (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                (100 < ^y => x1 = ^y - 10 | x1 = 91))}
       p91(x;x)
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "CALL" rule, we have
   [[ {(true /\ 101 - (y + 11) < 101 - ^y) /\
       (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91)))}
       p91(x;y + 11)
      {(true /\ 101 - x < 101 - ^y) /\
       (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                (100 < ^y => x1 = ^y - 10 | x1 = 91))} ]]

By the "SEQ" rule, we have
   [[ {(true /\ 101 - (y + 11) < 101 - ^y) /\
       (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91)))}
       p91(x;y + 11); p91(x;x)
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "IF" rule, we have
   [[ {(100 < y => (100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)
           | (true /\ 101 - (y + 11) < 101 - ^y) /\
             (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
                     (true /\ 101 - x < 101 - ^y) /\
                     (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                              (100 < ^y => x1 = ^y - 10 | x1 = 91))))}
       if 100 < y then x := y - 10 else p91(x;y + 11); p91(x;x) fi
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By precondition strengthening, we have
   [[ {(^x = x /\ ^y = y /\ true) /\ true}
       if 100 < y then x := y - 10 else p91(x;y + 11); p91(x;x) fi
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]
with additional verification condition
   [[ {(^x = x /\ ^y = y /\ true) /\ true ==>
       (100 < y => (100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)
           | (true /\ 101 - (y + 11) < 101 - ^y) /\
             (!x y1.
               (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91))))} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure p91:

By the call graph progress from procedure p91 to p91, we have
   [[ {true /\ (!x y1. 101 - y1 < 101 - y ==> 101 - y1 < ^z)}
      p91-<>->p91
      {101 - y < ^z} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ 101 - y = ^z ==>
       (!x y1. 101 - y1 < 101 - y ==> 101 - y1 < ^z)} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(true /\ true) /\
       (!a y1. (100 < 77 => a = 77 - 10 | a = 91) ==> a = 91)}
       p91(a;77)
      {a = 91} ]]

By precondition strengthening, we have
   [[ {true} p91(a;77) {a = 91} ]]
with additional verification condition
   [[ {true ==> (true /\ true) /\
                (!a y1. (100 < 77 => a = 77 - 10 | a = 91) ==> a = 91)} ]]

3 subgoals
"!a y1. (100 < 77 => (a = 77 - 10) | (a = 91)) ==> (a = 91)"

"!y ^z.
  (101 - y = ^z) ==> (!x y1. (101 - y1) < (101 - y) ==> (101 - y1) < ^z)"

"!^x x ^y y.
  (^x = x) /\ (^y = y) ==>
  (100 < y => 
   (100 < ^y => (y - 10 = ^y - 10) | (y - 10 = 91)) | 
   ((101 - (y + 11)) < (101 - ^y) /\
    (!x' y1.
      (100 < (y + 11) => (x' = (y + 11) - 10) | (x' = 91)) ==>
      (101 - x') < (101 - ^y) /\
      (!x1 y1'.
        (100 < x' => (x1 = x' - 10) | (x1 = 91)) ==>
        (100 < ^y => (x1 = ^y - 10) | (x1 = 91))))))"

() : void
Run time: 1280.3s
Garbage collection time: 120.9s
Intermediate theorems generated: 76317

>%

let st2 = get_state();;
set_state st2;;



let SUB_ADD_LESS_SUB =
 TAC_PROOF
  (([],
    "!a b c. b < a /\ 0 < c ==> (a - (b + c)) < a - b"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[SUB_PLUS]
   THEN MP_TAC (SPECL["a - b";"c:num"] SUB_LESS_EQ)
   THEN REWRITE_TAC[LESS_OR_EQ]
   THEN REWRITE_TAC[SUB_EQ_EQ_0]
   THEN REWRITE_TAC[SUB_EQ_0]
   THEN IMP_RES_THEN (REWRITE_THM o NOT_EQ_SYM) LESS_NOT_EQ
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)NOT_LESS]
  );;

%
g "!a b c. b < a /\ 0 < c ==> (a - (b + c)) < a - b";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[SUB_PLUS]);;
e(MP_TAC (SPECL["a - b";"c:num"] SUB_LESS_EQ));;
e(REWRITE_TAC[LESS_OR_EQ]);;
e(REWRITE_TAC[SUB_EQ_EQ_0]);;
e(REWRITE_TAC[SUB_EQ_0]);;
e(IMP_RES_THEN (REWRITE_THM o NOT_EQ_SYM) LESS_NOT_EQ);;
e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)NOT_LESS]);;
%

%<

% There are four cases to consider:
     y < 90, 90 <= y < 100, y = 100, and y > 100.
%

%
e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
 e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS]));;
 e(USE_THEN REWRITE_THM SUB_ADD_LESS_SUB);;

  e(CONV_TAC (DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[]);;
  e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
  e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;

  e(GEN_TAC);;
  e(COND_CASES_TAC);;

   e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11");
                                          LESS_MONO_ADD_EQ]));;
   e(SUBGOAL_THEN "10 <= 11" (REWRITE_THM o MATCH_MP LESS_EQ_ADD_SUB));;

    e(REWRITE_TAC[LESS_OR_EQ]);;
    e(CONV_TAC (DEPTH_CONV LESS_CONV));;
    e(REWRITE_TAC[]);;

    e(CONV_TAC (DEPTH_CONV SUBTRACT_CONV));;
    e(DISCH_THEN REWRITE_THM);;
    e(USE_THEN REWRITE_THM SUB_ADD_LESS_SUB);;

     e(CONV_TAC (DEPTH_CONV LESS_CONV));;
     e(REWRITE_TAC[]);;
     e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
     e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;

     e(GEN_TAC);;
     e(UNDISCH_LAST_TAC);;
     e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[LESS_OR_EQ]));;

      e(IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[ADD1]) LESS_SUC_NOT);;

      e(POP_ASSUM REWRITE_THM);;
      e(CONV_TAC (DEPTH_CONV ADD_CONV));;
      e(CONV_TAC (DEPTH_CONV LESS_CONV));;
      e(CONV_TAC (DEPTH_CONV SUBTRACT_CONV));;
      e(REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[NOT_LESS]);;
  e(DISCH_THEN (MP_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11")]));;
  e(REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]);;
  e(DISCH_TAC);;
  e(DISCH_THEN REWRITE_THM);;
  e(CONV_TAC (DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[]);;
  e(SUBGOAL_THEN "91 = y + (91 - y)" ONCE_REWRITE_THM);;

   e(ONCE_REWRITE_TAC[ADD_SYM]);;
   e(USE_IMP_THEN REWRITE_THM SUB_ADD);;
   e(IMP_RES_TAC LESS_EQ_TRANS);;
   e(POP_TAC THEN POP_TAC THEN POP_TAC);;
   e(POP_ASSUM MATCH_MP_TAC);;
   e(ONCE_REWRITE_TAC[LESS_OR_EQ]);;
   e(CONV_TAC (DEPTH_CONV LESS_CONV));;
   e(REWRITE_TAC[]);;

   e(MATCH_MP_TAC SUB_ADD_LESS_SUB);;
   e(CONV_TAC (RATOR_CONV (ONCE_DEPTH_CONV num_CONV)));;
   e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[(SYM o SPEC_ALL)SUB_LESS_0]);;
   e(IMP_RES_TAC LESS_TRANS);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[LESS_EQ]);;
   e(REWRITE_TAC[LESS_OR_EQ]);;
   e(DISJ2_TAC);;
   e(REWRITE_TAC(map (SYM o num_CONV) ["90";"91"]));;

e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;

e(CONV_TAC (DEPTH_CONV LESS_CONV));;
e(REWRITE_TAC[]);;
%




e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN COND_CASES_TAC
  THENL
   [ REWRITE_TAC[];

     REWRITE_TAC[]
     THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS])
     THEN USE_THEN REWRITE_THM SUB_ADD_LESS_SUB
     THENL
      [ CONV_TAC (DEPTH_CONV LESS_CONV)
        THEN REWRITE_TAC[]
        THEN CONV_TAC (ONCE_DEPTH_CONV num_CONV)
        THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC;

        GEN_TAC
        THEN COND_CASES_TAC
        THENL
         [ POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11");
                                          LESS_MONO_ADD_EQ])
           THEN SUBGOAL_THEN "10 <= 11" (REWRITE_THM o MATCH_MP LESS_EQ_ADD_SUB)
           THENL
            [ REWRITE_TAC[LESS_OR_EQ]
              THEN CONV_TAC (DEPTH_CONV LESS_CONV)
              THEN REWRITE_TAC[];

              CONV_TAC (DEPTH_CONV SUBTRACT_CONV)
              THEN DISCH_THEN REWRITE_THM
              THEN USE_THEN REWRITE_THM SUB_ADD_LESS_SUB
              THENL
               [ CONV_TAC (DEPTH_CONV LESS_CONV)
                 THEN REWRITE_TAC[]
                 THEN CONV_TAC (ONCE_DEPTH_CONV num_CONV)
                 THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC;

                 GEN_TAC
                 THEN UNDISCH_LAST_TAC
                 THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[LESS_OR_EQ])
                 THENL
                  [ IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[ADD1]) LESS_SUC_NOT;

                    POP_ASSUM REWRITE_THM
                    THEN CONV_TAC (DEPTH_CONV ADD_CONV)
                    THEN CONV_TAC (DEPTH_CONV LESS_CONV)
                    THEN CONV_TAC (DEPTH_CONV SUBTRACT_CONV)
                    THEN REWRITE_TAC[]
                  ]
               ]
            ];

           UNDISCH_LAST_TAC
           THEN REWRITE_TAC[NOT_LESS]
           THEN DISCH_THEN (MP_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11")])
           THEN REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]
           THEN DISCH_TAC
           THEN DISCH_THEN REWRITE_THM
           THEN CONV_TAC (DEPTH_CONV LESS_CONV)
           THEN REWRITE_TAC[]
           THEN SUBGOAL_THEN "91 = y + (91 - y)" ONCE_REWRITE_THM
           THENL
            [ ONCE_REWRITE_TAC[ADD_SYM]
              THEN USE_IMP_THEN REWRITE_THM SUB_ADD
              THEN IMP_RES_TAC LESS_EQ_TRANS
              THEN POP_TAC THEN POP_TAC THEN POP_TAC
              THEN POP_ASSUM MATCH_MP_TAC
              THEN ONCE_REWRITE_TAC[LESS_OR_EQ]
              THEN CONV_TAC (DEPTH_CONV LESS_CONV)
              THEN REWRITE_TAC[];

              MATCH_MP_TAC SUB_ADD_LESS_SUB
              THEN CONV_TAC (RATOR_CONV (ONCE_DEPTH_CONV num_CONV))
              THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC
              THEN POP_ASSUM REWRITE_THM
              THEN REWRITE_TAC[(SYM o SPEC_ALL)SUB_LESS_0]
              THEN IMP_RES_TAC LESS_TRANS
              THEN FIRST_ASSUM MATCH_MP_TAC
              THEN REWRITE_TAC[LESS_EQ]
              THEN REWRITE_TAC[LESS_OR_EQ]
              THEN DISJ2_TAC
              THEN REWRITE_TAC(map (SYM o num_CONV) ["90";"91"])
            ]
         ]
      ]
   ]

e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM);;

e(CONV_TAC (DEPTH_CONV LESS_CONV)
  THEN REWRITE_TAC[]);;




let p91 = top_thm();;
print_vcg_thm p91;;
%
|- [[ program procedure p91(x;y);
                 global ;
                 pre  true;
                 post (100 < ^y => x = ^y - 10 | x = 91);
                 calls p91 with 101 - y < 101 - ^y;
                 recurses with 101 - y < ^z;

                 if 100 < y
                    then x := y - 10
                    else p91(x;y + 11); p91(x;x)
                 fi
              end procedure; p91(a;77) end program
      [a = 91] ]]
%

>%






% --------------------------------------------------------------------- %
% Define the third example program: odd and even numbers.		%
% --------------------------------------------------------------------- %

g [[ program
        procedure odd(var a; val n);
           pre  true;
           post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
           calls odd  with n < ^n;
           calls even with n < ^n;
           recurses   with n < ^n;

           if n = 0 then a:=0
           else if n = 1 then even(a; n-1)
                         else odd (a; n-2)
                fi
           fi
        end procedure;

        procedure even(var a; val n);
           pre  true;
           post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
           calls even with n < ^n;
           calls odd  with n < ^n;
           recurses   with n < ^n;

           if n = 0 then a:=1
           else if n = 1 then odd (a; n-1)
                         else even(a; n-2)
                fi
           fi
        end procedure;

        odd(a; 5)

     end program
     [ a = 1 ]
  ]];;
%print_vcg_term (snd (top_goal()));;%

e(VCG_TAC);;

% 4197.5s = 1 hour, 9 min, 57.5 sec %

%<
OK..
For procedure `odd`,

By the "ASSIGN" rule, we have
   [[ {(?b. ^n = 2 * b + 0) /\ 0 < 2 /\ n = ^n}
       a := 0
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "CALL" rule, we have
   [[ {(true /\ n - 1 < ^n) /\
       (!a n2.
         (?b. (n - 1) + 1 = 2 * b + a) /\ a < 2 /\ n2 = n - 1 ==>
         (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)}
       even(a;n - 1)
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "CALL" rule, we have
   [[ {(true /\ n - 2 < ^n) /\
       (!a n2. (?b. n - 2 = 2 * b + a) /\ a < 2 /\ n2 = n - 2 ==>
               (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)}
       odd(a;n - 2)
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "IF" rule, we have
   [[ {(n = 1
          => (true /\ n - 1 < ^n) /\
             (!a n2.
               (?b. (n - 1) + 1 = 2 * b + a) /\
               a < 2 /\
               n2 = n - 1 ==> (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)
           | (true /\ n - 2 < ^n) /\
             (!a n2.
               (?b. n - 2 = 2 * b + a) /\ a < 2 /\ n2 = n - 2 ==>
               (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n))}
       if n = 1 then even(a;n - 1) else odd(a;n - 2) fi
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "IF" rule, we have
   [[ {(n = 0 => (?b. ^n = 2 * b + 0) /\ 0 < 2 /\ n = ^n
           | (n = 1 => (true /\ n - 1 < ^n) /\
                       (!a n2. (?b. (n - 1) + 1 = 2 * b + a) /\
                               a < 2 /\
                               n2 = n - 1 ==>
                               (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)
                 | (true /\ n - 2 < ^n) /\
                   (!a n2. (?b. n - 2 = 2 * b + a) /\
                           a < 2 /\
                           n2 = n - 2 ==>
                           (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)))}
       if n = 0
          then a := 0
          else if n = 1 then even(a;n - 1) else odd(a;n - 2) fi
       fi
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By precondition strengthening, we have
   [[ {(^a = a /\ ^n = n /\ true) /\ true}
       if n = 0
          then a := 0
          else if n = 1 then even(a;n - 1) else odd(a;n - 2) fi
       fi
      {(?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n} ]]
with additional verification condition
   [[ {(^a = a /\ ^n = n /\ true) /\ true ==>
       (n = 0 => (?b. ^n = 2 * b + 0) /\ 0 < 2 /\ n = ^n
           | (n = 1
                => (true /\ n - 1 < ^n) /\
                   (!a n2. (?b. (n - 1) + 1 = 2 * b + a) /\
                           a < 2 /\
                           n2 = n - 1 ==>
                           (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)
                 | (true /\ n - 2 < ^n) /\
                   (!a n2. (?b. n - 2 = 2 * b + a) /\
                           a < 2 /\
                           n2 = n - 2 ==>
                           (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n)))} ]]

For procedure `even`,

By the "ASSIGN" rule, we have
   [[ {(?b. ^n + 1 = 2 * b + 1) /\ 1 < 2 /\ n = ^n}
       a := 1
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "CALL" rule, we have
   [[ {(true /\ n - 1 < ^n) /\
       (!a n2. (?b. n - 1 = 2 * b + a) /\ a < 2 /\ n2 = n - 1 ==>
               (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)}
       odd(a;n - 1)
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "CALL" rule, we have
   [[ {(true /\ n - 2 < ^n) /\
       (!a n2.
         (?b. (n - 2) + 1 = 2 * b + a) /\ a < 2 /\ n2 = n - 2 ==>
         (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)}
       even(a;n - 2)
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "IF" rule, we have
   [[ {(n = 1 => (true /\ n - 1 < ^n) /\
                 (!a n2.
                   (?b. n - 1 = 2 * b + a) /\ a < 2 /\ n2 = n - 1 ==>
                   (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)
           | (true /\ n - 2 < ^n) /\
             (!a n2. (?b. (n - 2) + 1 = 2 * b + a) /\
                     a < 2 /\
                     n2 = n - 2 ==>
                     (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n))}
       if n = 1 then odd(a;n - 1) else even(a;n - 2) fi
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By the "IF" rule, we have
   [[ {(n = 0 => (?b. ^n + 1 = 2 * b + 1) /\ 1 < 2 /\ n = ^n
           | (n = 1 => (true /\ n - 1 < ^n) /\
                       (!a n2.
                         (?b. n - 1 = 2 * b + a) /\
                         a < 2 /\
                         n2 = n - 1 ==>
                         (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)
                 | (true /\ n - 2 < ^n) /\
                   (!a n2.
                     (?b. (n - 2) + 1 = 2 * b + a) /\
                     a < 2 /\
                     n2 = n - 2 ==>
                     (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)))}
       if n = 0
          then a := 1
          else if n = 1 then odd(a;n - 1) else even(a;n - 2) fi
       fi
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]

By precondition strengthening, we have
   [[ {(^a = a /\ ^n = n /\ true) /\ true}
       if n = 0
          then a := 1
          else if n = 1 then odd(a;n - 1) else even(a;n - 2) fi
       fi
      {(?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n} ]]
with additional verification condition
   [[ {(^a = a /\ ^n = n /\ true) /\ true ==>
       (n = 0 => (?b. ^n + 1 = 2 * b + 1) /\ 1 < 2 /\ n = ^n
           | (n = 1 => (true /\ n - 1 < ^n) /\
                       (!a n2.
                         (?b. n - 1 = 2 * b + a) /\
                         a < 2 /\
                         n2 = n - 1 ==>
                         (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)
                 | (true /\ n - 2 < ^n) /\
                   (!a n2.
                     (?b. (n - 2) + 1 = 2 * b + a) /\
                     a < 2 /\
                     n2 = n - 2 ==>
                     (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n)))} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure even:

By the call graph progress from procedure even to even, we have
   [[ {true /\ (!a n1. n1 < n ==> n1 < ^n)} even-<>->even {n < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==> (!a n1. n1 < n ==> n1 < ^n)} ]]

By the call graph progress from procedure odd to even, we have
   [[ {true /\ (!a n1. n1 < n ==> n1 < ^n)} odd-<>->even {n < ^n} ]]

By the call graph progress from procedure even to odd, we have
   [[ {true /\ (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))}
      even-<>->odd
      {!a n1. n1 < n ==> n1 < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==>
       (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))} ]]

By the call graph progress from procedure odd to odd, we have
   [[ {true /\ (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))}
      odd-<>->odd
      {!a n1. n1 < n ==> n1 < ^n} ]]

Generating the diversion verification condition
   [[ {(!a n1. n1 < n ==> n1 < ^n) ==>
       (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))} ]]

Traversing the call graph back from the procedure odd:

By the call graph progress from procedure even to odd, we have
   [[ {true /\ (!a n1. n1 < n ==> n1 < ^n)} even-<>->odd {n < ^n} ]]

By the call graph progress from procedure even to even, we have
   [[ {true /\ (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))}
      even-<>->even
      {!a n1. n1 < n ==> n1 < ^n} ]]

Generating the diversion verification condition
   [[ {(!a n1. n1 < n ==> n1 < ^n) ==>
       (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))} ]]

By the call graph progress from procedure odd to even, we have
   [[ {true /\ (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))}
      odd-<>->even
      {!a n1. n1 < n ==> n1 < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==>
       (!a n1. n1 < n ==> (!a n2. n2 < n1 ==> n2 < ^n))} ]]

By the call graph progress from procedure odd to odd, we have
   [[ {true /\ (!a n1. n1 < n ==> n1 < ^n)} odd-<>->odd {n < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==> (!a n1. n1 < n ==> n1 < ^n)} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(true /\ true) /\
       (!a n1. (?b. 5 = 2 * b + a) /\ a < 2 /\ n1 = 5 ==> a = 1)}
       odd(a;5)
      {a = 1} ]]

By precondition strengthening, we have
   [[ {true} odd(a;5) {a = 1} ]]
with additional verification condition
   [[ {true ==>
       (true /\ true) /\
       (!a n1. (?b. 5 = 2 * b + a) /\ a < 2 /\ n1 = 5 ==> a = 1)} ]]

9 subgoals
"!a n1. (?b. 5 = (2 * b) + a) /\ a < 2 /\ (n1 = 5) ==> (a = 1)"

"!n ^n. (n = ^n) ==> (!a n1. n1 < n ==> n1 < ^n)"

"!n ^n. (n = ^n) ==> (!a n1. n1 < n ==> (!a' n2. n2 < n1 ==> n2 < ^n))"

"!n ^n.
  (!a n1. n1 < n ==> n1 < ^n) ==>
  (!a n1. n1 < n ==> (!a' n2. n2 < n1 ==> n2 < ^n))"

"!n ^n.
  (!a n1. n1 < n ==> n1 < ^n) ==>
  (!a n1. n1 < n ==> (!a' n2. n2 < n1 ==> n2 < ^n))"

"!n ^n. (n = ^n) ==> (!a n1. n1 < n ==> (!a' n2. n2 < n1 ==> n2 < ^n))"

"!n ^n. (n = ^n) ==> (!a n1. n1 < n ==> n1 < ^n)"

"!^a a ^n n.
  (^a = a) /\ (^n = n) ==>
  ((n = 0) => 
   ((?b. ^n + 1 = (2 * b) + 1) /\ 1 < 2 /\ (n = ^n)) | 
   ((n = 1) => 
    ((n - 1) < ^n /\
     (!a' n2.
       (?b. n - 1 = (2 * b) + a') /\ a' < 2 /\ (n2 = n - 1) ==>
       (?b. ^n + 1 = (2 * b) + a') /\ a' < 2 /\ (n = ^n))) | 
    ((n - 2) < ^n /\
     (!a' n2.
       (?b. (n - 2) + 1 = (2 * b) + a') /\ a' < 2 /\ (n2 = n - 2) ==>
       (?b. ^n + 1 = (2 * b) + a') /\ a' < 2 /\ (n = ^n)))))"

"!^a a ^n n.
  (^a = a) /\ (^n = n) ==>
  ((n = 0) => 
   ((?b. ^n = (2 * b) + 0) /\ 0 < 2 /\ (n = ^n)) | 
   ((n = 1) => 
    ((n - 1) < ^n /\
     (!a' n2.
       (?b. (n - 1) + 1 = (2 * b) + a') /\ a' < 2 /\ (n2 = n - 1) ==>
       (?b. ^n = (2 * b) + a') /\ a' < 2 /\ (n = ^n))) | 
    ((n - 2) < ^n /\
     (!a' n2.
       (?b. n - 2 = (2 * b) + a') /\ a' < 2 /\ (n2 = n - 2) ==>
       (?b. ^n = (2 * b) + a') /\ a' < 2 /\ (n = ^n)))))"

() : void
Run time: 3772.4s
Garbage collection time: 425.1s
Intermediate theorems generated: 310178

>%

%< Tactics to solve these 9 verification conditions:

let st3 = get_state();;
set_state st3;;

% 1st subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(COND_CASES_TAC);;

 e(CONV_TAC (DEPTH_CONV LESS_CONV));;
 e(ASM_REWRITE_TAC[]);;
 e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
 e(REWRITE_TAC[MULT_CLAUSES]);;
 e(EXISTS_TAC "0");;
 e(REWRITE_TAC[ADD_CLAUSES]);;

 e(MP_TAC (SPEC "n:num" LESS_0_CASES));;
 e(FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
 e(DISCH_TAC);;
 e(COND_CASES_TAC);;

  e(ASM_REWRITE_TAC[SUB_EQUAL_0]);;
  e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[ADD]);;
  e(REPEAT STRIP_TAC);;
  e(EXISTS_TAC "b:num");;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

 e(MP_TAC (SPECL["1";"n:num"] LESS_CASES));;
 e(REWRITE_TAC[LESS_OR_EQ]);;
 e(FIRST_ASSUM REWRITE_THM);;
 e(IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[SYM(num_CONV "1")]) LESS_SUC_NOT);;
 e(DISCH_TAC);;
 e(FIRST_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(num_CONV "2")] o
                   MATCH_MP LESS_OR));;
 e(IMP_RES_THEN REWRITE_THM LESS_EQ_SUB_LESS);;
 e(CONJ_TAC);;

  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
  e(REWRITE_TAC[LESS_ADD_SUC]);;

  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "b + 1");;
  e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1]);;
  e(REWRITE_TAC[theorem `arithmetic` `MULT_SUC`]);;
  e(POP_TAC);;
  e(REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]);;
  e(FIRST_ASSUM (REWRITE_THM o SYM));;
  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(IMP_RES_THEN REWRITE_THM SUB_ADD);;

% 2nd subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(COND_CASES_TAC);;

 e(CONV_TAC (DEPTH_CONV LESS_CONV));;
 e(ASM_REWRITE_TAC[]);;
 e(EXISTS_TAC "0");;
 e(REWRITE_TAC[MULT_CLAUSES]);;

 e(MP_TAC (SPEC "n:num" LESS_0_CASES));;
 e(FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
 e(DISCH_TAC);;
 e(COND_CASES_TAC);;

  e(ASM_REWRITE_TAC[SUB_EQUAL_0]);;
  e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "SUC b");;
  e(CONV_TAC (ONCE_DEPTH_CONV ADD_CONV));;
  e(REWRITE_TAC[theorem `arithmetic` `MULT_SUC`]);;
  e(REWRITE_TAC[(SYM o SPEC_ALL) ADD_ASSOC]);;
  e(POP_TAC);;
  e(FIRST_ASSUM (REWRITE_THM o SYM));;
  e(REWRITE_TAC[ADD_CLAUSES]);;

 e(MP_TAC (SPECL["1";"n:num"] LESS_CASES));;
 e(REWRITE_TAC[LESS_OR_EQ]);;
 e(FIRST_ASSUM REWRITE_THM);;
 e(IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[SYM(num_CONV "1")]) LESS_SUC_NOT);;
 e(DISCH_TAC);;
 e(FIRST_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(num_CONV "2")] o
                   MATCH_MP LESS_OR));;
 e(IMP_RES_THEN REWRITE_THM LESS_EQ_SUB_LESS);;
 e(CONJ_TAC);;

  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
  e(REWRITE_TAC[LESS_ADD_SUC]);;

  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "SUC b");;
  e(REWRITE_TAC[theorem `arithmetic` `MULT_SUC`]);;
  e(POP_TAC);;
  e(REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]);;
  e(FIRST_ASSUM (REWRITE_THM o SYM));;
  e(REWRITE_TAC[ADD_ASSOC]);;
  e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1;INV_SUC_EQ]);;
  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(IMP_RES_THEN REWRITE_THM SUB_ADD);;

% 3rd subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;

% 4th subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC LESS_TRANS);;

% 5th subgoal: %

e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC LESS_TRANS);;

% 6th subgoal: %

e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC LESS_TRANS);;

% 7th subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC LESS_TRANS);;

% 8th subgoal: %

e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;

% 9th subgoal (main body): %

e(REPEAT STRIP_TAC);;
e(POP_TAC);;
e(STRIP_ASSUME_TAC (SPEC "a:num" LESS_0_CASES));;

 e(UNDISCH_TAC "5 = (2 * b) + a");;
 e(POP_ASSUM (REWRITE_THM o SYM));;
 e(REWRITE_TAC[num_CONV "2"]);;
 e(REWRITE_TAC[MULT_CLAUSES;ADD_CLAUSES]);;
 e(REWRITE_TAC[num_CONV "5"]);;
 e(REWRITE_TAC[SYM(ADD_CONV "2+2")]);;
 e(REWRITE_TAC[NOT_ODD_EQ_EVEN]);;

 e(MP_TAC (SPECL["a:num";"1"] LESS_CASES));;
 e(REWRITE_TAC[LESS_OR_EQ]);;
 e(IMP_RES_TAC LESS_SUC_NOT);;
 e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[num_CONV "2";LESS_MONO_EQ]));;
 e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[SYM(num_CONV "1")]));;
 e(DISCH_THEN REWRITE_THM);;

let odd_even = top_thm();;
print_vcg_thm odd_even;;
%
|- [[ program
         procedure odd(a;n);
            global ;
            pre  true;
            post (?b. ^n = 2 * b + a) /\ a < 2 /\ n = ^n;
            calls odd with n < ^n;
            calls even with n < ^n;
            recurses with n < ^n;

            if n = 0
               then a := 0
               else if n = 1 then even(a;n - 1) else odd(a;n - 2) fi
            fi
         end procedure;
         procedure even(a;n);
            global ;
            pre  true;
            post (?b. ^n + 1 = 2 * b + a) /\ a < 2 /\ n = ^n;
            calls even with n < ^n;
            calls odd with n < ^n;
            recurses with n < ^n;

            if n = 0
               then a := 1
               else if n = 1 then odd(a;n - 1) else even(a;n - 2) fi
            fi
         end procedure;

         odd(a;5)
      end program
      [a = 1] ]]()
: void
Run time: 50.5s
Garbage collection time: 14.7s
%

>%




% --------------------------------------------------------------------- %
% Define the fourth example program: product (from Pandya and Joseph).  %
% --------------------------------------------------------------------- %

g [[ program
        procedure quotient_remainder (var q,r; val x,y);
           pre  0 < y;
           post ^x = q * ^y + r /\ r < ^y;

           r := x;
           q := 0;
           assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
             with r < ^r
           while ~(r < y) do
              r := r - y;
              q := ++q
           od
        end procedure;

        procedure product (;);
           global x,y,z,a,b,q,r;
           pre  z + x * y = a * b;
           post z = a * b;
           calls quotient_remainder with true;
           calls evenproduct with y = ^y;
           calls oddproduct  with y = ^y;
           recurses with y < ^y;

           quotient_remainder(q,r;y,2);
           if r = 0
           then evenproduct(;)
           else oddproduct(;)
           fi
        end procedure;

        procedure oddproduct (;);
           global x,y,z,a,b,q,r;
           pre  z + x * y = a * b /\ y = 2*q + 1;
           post z = a * b;
           calls evenproduct with y < ^y;
           recurses with y < ^y;

           y := y - 1;
           z := z + x;
           evenproduct(;)
        end procedure;

        procedure evenproduct (;);
           global x,y,z,a,b,q,r;
           pre  z + x * y = a * b /\ y = 2*q;
           post z = a * b;
           calls product with y < ^y;
           recurses with y < ^y;

           if y = 0
           then
              skip
           else
              x := 2 * x;
              y := q;
              product(;)
           fi
        end procedure;

        a := 7;
        b := 11;
        x := a;
        y := b;
        z := 0;
        product(;)
     end program
     [ z = 77 ]
  ]];;

e(VCG_TAC);;

%
OK..
For procedure `quotient_remainder`,

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r}
       q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y
      {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "SEQ" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y; q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "WHILE" rule, we have
   [[ {^x = q * ^y + r /\ 0 < y /\ ^y = y}
       assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
         with r < ^r
       while ~(r < y) do
          r := r - y; q := ++q
       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + r /\ 0 < y /\ ^y = y}
       q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x
      {^x = 0 * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By precondition strengthening, we have
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with additional verification condition
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y ==>
       ^x = 0 * ^y + x /\ 0 < y /\ ^y = y} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure quotient_remainder:

By the call graph progress from procedure quotient_remainder to quotient_remainder, we have
   [[ {0 < y /\ true} quotient_remainder-<>->quotient_remainder {false} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(0 < 3 /\ true) /\
       (!q. !r. !x1. !y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)}
       quotient_remainder(q,r;7,3)
      {q = 2 /\ r = 1} ]]

By precondition strengthening, we have
   [[ {true} quotient_remainder(q,r;7,3) {q = 2 /\ r = 1} ]]
with additional verification condition
   [[ {true ==>
       (0 < 3 /\ true) /\
       (!q. !r. !x1. !y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)} ]]

4 subgoals
"0 < 3 /\
 (!q r x1 y1. (7 = (q * 3) + r) /\ r < 3 ==> (q = 2) /\ (r = 1))"

"!^x q ^y r y.
  ((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ r < y ==>
  (^x = (q * ^y) + r) /\ r < ^y"

"!^x q ^y r y ^r.
  (((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ ~r < y) /\ (r = ^r) ==>
  ((^x = ((q + 1) * ^y) + (r - y)) /\ 0 < y /\ (^y = y)) /\ (r - y) < ^r"

"!^q q ^r r ^x x ^y y.
  ((^q = q) /\ (^r = r) /\ (^x = x) /\ (^y = y)) /\ 0 < y ==>
  (^x = (0 * ^y) + x) /\ 0 < y /\ (^y = y)"

() : void
Run time: 881.5s
Garbage collection time: 92.0s
Intermediate theorems generated: 53837

%





% --------------------------------------------------------------------- %
% Define the fifth example program: Bicycling				%
% --------------------------------------------------------------------- %

g [[ program

        procedure pedal (;val n,m);
           pre  true;
           post true;
           calls pedal with n < ^n /\ m = ^m;
           calls coast with n < ^n /\ m < ^m;
           recurses    with n < ^n;

           if 0 < n then
              if 0 < m then
                 coast(;n - 1,m - 1)
              else skip
              fi;
              pedal(;n - 1,m)
           else skip
           fi
        end procedure;

        procedure coast (;val n,m);
           pre  true;
           post true;
           calls pedal with n = ^n /\ m = ^m;
           calls coast with n = ^n /\ m < ^m;
           recurses    with m < ^m;

           pedal(;n,m);
           if 0 < m then
              coast(;n,m - 1)
           else skip
           fi
        end procedure;

        pedal(;7,12)

     end program
     [ true ]
  ]];;

e(VCG_TAC);;

% 3075.6s = 51 min, 15.6 sec %

%<
OK..
For procedure `pedal`,

By the "CALL" rule, we have
   [[ {(true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true)}
       pedal(;n - 1,m)
      {true} ]]

By the "CALL" rule, we have
   [[ {(true /\ n - 1 < ^n /\ m - 1 < ^m) /\
       (!n1 m1.
         true ==>
         (true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true))}
       coast(;n - 1,m - 1)
      {(true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true)} ]]

By the "SKIP" rule, we have
   [[ {(true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true)}
       skip
      {(true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true)} ]]

By the "IF" rule, we have
   [[ {(0 < m => (true /\ n - 1 < ^n /\ m - 1 < ^m) /\
                 (!n1 m1. true ==> (true /\ n - 1 < ^n /\ m = ^m) /\
                                   (!n m. true ==> true))
           | (true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true))}
       if 0 < m then coast(;n - 1,m - 1) else skip fi
      {(true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true)} ]]

By the "SEQ" rule, we have
   [[ {(0 < m => (true /\ n - 1 < ^n /\ m - 1 < ^m) /\
                 (!n1 m1. true ==> (true /\ n - 1 < ^n /\ m = ^m) /\
                                   (!n m. true ==> true))
           | (true /\ n - 1 < ^n /\ m = ^m) /\ (!n m. true ==> true))}
       if 0 < m then coast(;n - 1,m - 1) else skip fi; pedal(;n - 1,m)
      {true} ]]

By the "SKIP" rule, we have
   [[ {true} skip {true} ]]

By the "IF" rule, we have
   [[ {(0 < n
          => (0 < m
                => (true /\ n - 1 < ^n /\ m - 1 < ^m) /\
                   (!n1 m1. true ==> (true /\ n - 1 < ^n /\ m = ^m) /\
                                     (!n m. true ==> true))
                 | (true /\ n - 1 < ^n /\ m = ^m) /\
                   (!n m. true ==> true)) | true)}
       if 0 < n then if 0 < m then coast(;n - 1,m - 1) else skip fi;
                     pedal(;n - 1,m) else skip fi
      {true} ]]

By precondition strengthening, we have
   [[ {(^n = n /\ ^m = m /\ true) /\ true}
       if 0 < n then if 0 < m then coast(;n - 1,m - 1) else skip fi;
                     pedal(;n - 1,m) else skip fi
      {true} ]]
with additional verification condition
   [[ {(^n = n /\ ^m = m /\ true) /\ true ==>
       (0 < n => (0 < m => (true /\ n - 1 < ^n /\ m - 1 < ^m) /\
                           (!n1 m1.
                             true ==> (true /\ n - 1 < ^n /\ m = ^m) /\
                                      (!n m. true ==> true))
                     | (true /\ n - 1 < ^n /\ m = ^m) /\
                       (!n m. true ==> true)) | true)} ]]

For procedure `coast`,

By the "CALL" rule, we have
   [[ {(true /\ n = ^n /\ m - 1 < ^m) /\ (!n m. true ==> true)}
       coast(;n,m - 1)
      {true} ]]

By the "SKIP" rule, we have
   [[ {true} skip {true} ]]

By the "IF" rule, we have
   [[ {(0 < m => (true /\ n = ^n /\ m - 1 < ^m) /\ (!n m. true ==> true)
           | true)} if 0 < m then coast(;n,m - 1) else skip fi {true} ]]

By the "CALL" rule, we have
   [[ {(true /\ n = ^n /\ m = ^m) /\
       (!n1 m1. true ==> (0 < m => (true /\ n = ^n /\ m - 1 < ^m) /\
                                   (!n m. true ==> true) | true))}
       pedal(;n,m)
      {(0 < m => (true /\ n = ^n /\ m - 1 < ^m) /\ (!n m. true ==> true)
           | true)} ]]

By the "SEQ" rule, we have
   [[ {(true /\ n = ^n /\ m = ^m) /\
       (!n1 m1. true ==> (0 < m => (true /\ n = ^n /\ m - 1 < ^m) /\
                                   (!n m. true ==> true) | true))}
       pedal(;n,m); if 0 < m then coast(;n,m - 1) else skip fi
      {true} ]]

By precondition strengthening, we have
   [[ {(^n = n /\ ^m = m /\ true) /\ true}
       pedal(;n,m); if 0 < m then coast(;n,m - 1) else skip fi
      {true} ]]
with additional verification condition
   [[ {(^n = n /\ ^m = m /\ true) /\ true ==>
       (true /\ n = ^n /\ m = ^m) /\
       (!n1 m1. true ==> (0 < m => (true /\ n = ^n /\ m - 1 < ^m) /\
                                   (!n m. true ==> true) | true))} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure coast:

By the call graph progress from procedure coast to coast, we have
   [[ {true /\ (!n1 m1. n1 = n /\ m1 < m ==> m1 < ^m)}
      coast-<>->coast
      {m < ^m} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ m = ^m ==> (!n1 m1. n1 = n /\ m1 < m ==> m1 < ^m)} ]]

By the call graph progress from procedure pedal to coast, we have
   [[ {true /\ (!n1 m1. n1 < n /\ m1 < m ==> m1 < ^m)}
      pedal-<>->coast
      {m < ^m} ]]

By the call graph progress from procedure coast to pedal, we have
   [[ {true /\ (!n1 m1. n1 = n /\ m1 = m ==>
                        (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))}
      coast-<>->pedal
      {!n1 m1. n1 < n /\ m1 < m ==> m1 < ^m} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ m = ^m ==>
       (!n1 m1. n1 = n /\ m1 = m ==>
                (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))} ]]

By the call graph progress from procedure pedal to pedal, we have
   [[ {true /\ (!n1 m1. n1 < n /\ m1 = m ==>
                        (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))}
      pedal-<>->pedal
      {!n1 m1. n1 < n /\ m1 < m ==> m1 < ^m} ]]

Generating the diversion verification condition
   [[ {(!n1 m1. n1 < n /\ m1 < m ==> m1 < ^m) ==>
       (!n1 m1. n1 < n /\ m1 = m ==>
                (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))} ]]

Traversing the call graph back from the procedure pedal:

By the call graph progress from procedure coast to pedal, we have
   [[ {true /\ (!n1 m1. n1 = n /\ m1 = m ==> n1 < ^n)}
      coast-<>->pedal
      {n < ^n} ]]

By the call graph progress from procedure coast to coast, we have
   [[ {true /\ (!n1 m1. n1 = n /\ m1 < m ==>
                        (!n2 m2. n2 = n1 /\ m2 = m1 ==> n2 < ^n))}
      coast-<>->coast
      {!n1 m1. n1 = n /\ m1 = m ==> n1 < ^n} ]]

Generating the diversion verification condition
   [[ {(!n1 m1. n1 = n /\ m1 = m ==> n1 < ^n) ==>
       (!n1 m1. n1 = n /\ m1 < m ==>
                (!n2 m2. n2 = n1 /\ m2 = m1 ==> n2 < ^n))} ]]

By the call graph progress from procedure pedal to coast, we have
   [[ {true /\ (!n1 m1. n1 < n /\ m1 < m ==>
                        (!n2 m2. n2 = n1 /\ m2 = m1 ==> n2 < ^n))}
      pedal-<>->coast
      {!n1 m1. n1 = n /\ m1 = m ==> n1 < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==>
       (!n1 m1. n1 < n /\ m1 < m ==>
                (!n2 m2. n2 = n1 /\ m2 = m1 ==> n2 < ^n))} ]]

By the call graph progress from procedure pedal to pedal, we have
   [[ {true /\ (!n1 m1. n1 < n /\ m1 = m ==> n1 < ^n)}
      pedal-<>->pedal
      {n < ^n} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ n = ^n ==> (!n1 m1. n1 < n /\ m1 = m ==> n1 < ^n)} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(true /\ true) /\ (!n m. true ==> true)} pedal(;7,12) {true} ]]

By precondition strengthening, we have
   [[ {true} pedal(;7,12) {true} ]]
with additional verification condition
   [[ {true ==> (true /\ true) /\ (!n m. true ==> true)} ]]

8 subgoals
"!m ^m n. (m = ^m) ==> (!n1 m1. (n1 = n) /\ m1 < m ==> m1 < ^m)"

"!m ^m n.
  (m = ^m) ==>
  (!n1 m1.
    (n1 = n) /\ (m1 = m) ==> (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))"

"!n m ^m.
  (!n1 m1. n1 < n /\ m1 < m ==> m1 < ^m) ==>
  (!n1 m1.
    n1 < n /\ (m1 = m) ==> (!n2 m2. n2 < n1 /\ m2 < m1 ==> m2 < ^m))"

"!n m ^n.
  (!n1 m1. (n1 = n) /\ (m1 = m) ==> n1 < ^n) ==>
  (!n1 m1.
    (n1 = n) /\ m1 < m ==> (!n2 m2. (n2 = n1) /\ (m2 = m1) ==> n2 < ^n))"

"!n ^n m.
  (n = ^n) ==>
  (!n1 m1.
    n1 < n /\ m1 < m ==> (!n2 m2. (n2 = n1) /\ (m2 = m1) ==> n2 < ^n))"

"!n ^n m. (n = ^n) ==> (!n1 m1. n1 < n /\ (m1 = m) ==> n1 < ^n)"

"!^n n ^m m.
  (^n = n) /\ (^m = m) ==>
  ((n = ^n) /\ (m = ^m)) /\
  (!n1 m1. (0 < m => ((n = ^n) /\ (m - 1) < ^m) | T))"

"!^n n ^m m.
  (^n = n) /\ (^m = m) ==>
  (0 < n => 
   (0 < m => 
    (((n - 1) < ^n /\ (m - 1) < ^m) /\
     (!n1 m1. (n - 1) < ^n /\ (m = ^m))) | 
    ((n - 1) < ^n /\ (m = ^m))) | 
   T)"

() : void
Run time: 2760.1s
Garbage collection time: 315.5s
Intermediate theorems generated: 224220

>%

%< These verification conditions are solved by the following tactics:

let st5 = get_state();;
set_state st5;;

% 1st subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(REWRITE_TAC[SUB_RIGHT_LESS]);;
e(REWRITE_TAC[num_CONV "1";ADD_CLAUSES;LESS_SUC_REFL]);;
e(COND_CASES_TAC);;

 e(COND_CASES_TAC);;

  e(REWRITE_TAC[]);;

  e(REWRITE_TAC[]);;

 e(COND_CASES_TAC);;

  e(REWRITE_TAC[]);;

  e(REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN REWRITE_TAC[SUB_RIGHT_LESS]
  THEN REWRITE_TAC[num_CONV "1";ADD_CLAUSES;LESS_SUC_REFL]
  THEN COND_CASES_TAC
  THEN COND_CASES_TAC
  THEN REWRITE_TAC[]);;

% 2nd subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(REWRITE_TAC[SUB_RIGHT_LESS]);;
e(REWRITE_TAC[num_CONV "1";ADD_CLAUSES;LESS_SUC_REFL]);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN REWRITE_TAC[SUB_RIGHT_LESS]
  THEN REWRITE_TAC[num_CONV "1";ADD_CLAUSES;LESS_SUC_REFL]
  THEN COND_CASES_TAC
  THEN REWRITE_TAC[]);;

% 3rd subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN (REWRITE_THM o SYM)
  THEN REPEAT STRIP_TAC);;

% 4th subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN (REWRITE_THM o SYM)
  THEN REPEAT STRIP_TAC
  THEN ASM_REWRITE_TAC[]);;

% 5th subgoal: %

%e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "m:num");;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT STRIP_TAC
  THEN FIRST_ASSUM MATCH_MP_TAC
  THEN EXISTS_TAC "m:num"
  THEN ASM_REWRITE_TAC[]);;

% 6th subgoal: %

%e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "n1:num");;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT STRIP_TAC
  THEN FIRST_ASSUM MATCH_MP_TAC
  THEN EXISTS_TAC "n1:num"
  THEN FIRST_ASSUM (REWRITE_THM o SYM)
  THEN ASM_REWRITE_TAC[]);;

% 7th subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN (REWRITE_THM o SYM)
  THEN REPEAT STRIP_TAC
  THEN FIRST_ASSUM (REWRITE_THM o SYM)
  THEN ASM_REWRITE_TAC[]);;

% 8th subgoal: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;%

e(REPEAT GEN_TAC
  THEN DISCH_THEN (REWRITE_THM o SYM)
  THEN REPEAT STRIP_TAC);;

let cycling = top_thm();;
print_vcg_thm cycling;;
%
|- [[ program
         procedure pedal(;n,m);
            global ;
            pre  true;
            post true;
            calls pedal with n < ^n /\ m = ^m;
            calls coast with n < ^n /\ m < ^m;
            recurses with n < ^n;

            if 0 < n
               then if 0 < m then coast(;n - 1,m - 1) else skip fi;
                    pedal(;n - 1,m)
               else skip
            fi
         end procedure;
         procedure coast(;n,m);
            global ;
            pre  true;
            post true;
            calls pedal with n = ^n /\ m = ^m;
            calls coast with n = ^n /\ m < ^m;
            recurses with m < ^m;

            pedal(;n,m); if 0 < m then coast(;n,m - 1) else skip fi
         end procedure;

         pedal(;7,12)
      end program
      [true] ]]
%
>%