% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the functions to compute the
  sets of free variables of phrases of the programming language.  It
  also contains several theorems concerning the independence of phrases
  under modification of variables not free in the phrase.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `free_variables`;;

new_parent `prog_semantics`;;


% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %
load_library `ind_defs`;;

loadf `load_finite_sets`;;
load_finite_sets();;


% Set up autoloading from ancestor theories of free_variables: %

loadf `autoload`;;

autoload_all_defs_and_thms
 `finite_sets more_finite_sets bindings variants
  assert_syntax assert_semantics prog_syntax
  substitutions var_substitutions prog_substitutions
  prog_semantics`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;




% ========================================================== %
% We now define the free variable functions for expressions, %
% expression lists, boolean expressions, and commands.       %
% ========================================================== %


let FVxs_DEF = 
    new_list_rec_definition (`FVxs_DEF`,
      "(FVxs (NIL) = EMPTY:(var)set)  /\
       (FVxs (CONS x xs) = x INSERT FVxs xs)");;

let FVe_DEF = 
    new_recursive_definition false exp_Axiom `FVe_DEF`
      "(FVe (NUM n)       = EMPTY)           /\
       (FVe (PVAR x)      = INSERT x EMPTY)  /\
       (FVe (INC x)       = INSERT x EMPTY)  /\
       (FVe (PLUS e1 e2)  = FVe e1 UNION FVe e2)  /\
       (FVe (MINUS e1 e2) = FVe e1 UNION FVe e2)  /\
       (FVe (MULT e1 e2)  = FVe e1 UNION FVe e2)";;

let FVes_DEF = 
    new_list_rec_definition (`FVes_DEF`,
      "(FVes (NIL) = EMPTY)  /\
       (FVes (CONS e es) = FVe e UNION FVes es)");;

let FVb_DEF = 
    new_recursive_definition false bexp_Axiom `FVb_DEF`
      "(FVb (EQ  e1 e2)     = FVe e1 UNION FVe e2)  /\
       (FVb (LESS e1 e2)    = FVe e1 UNION FVe e2)  /\
       (FVb (LLESS es1 es2) = FVes es1 UNION FVes es2)  /\
       (FVb (AND b1 b2)     = FVb b1 UNION FVb b2)  /\
       (FVb (OR  b1 b2)     = FVb b1 UNION FVb b2)  /\
       (FVb (NOT b)         = FVb b)";;

let FVc_DEF = 
    new_recursive_definition false cmd_Axiom `FVc_DEF`
      "(FVc (SKIP) (d:env) = EMPTY)  /\
       (FVc (ABORT) d = EMPTY)  /\
       (FVc (ASSIGN x e) d = INSERT x (FVe e))  /\
       (FVc (SEQ c1 c2) d = FVc c1 d UNION FVc c2 d)  /\
       (FVc (IF b c1 c2) d = FVb b UNION FVc c1 d UNION FVc c2 d)  /\
       (FVc (WHILE a pr b c) d = FVb b UNION FVc c d)  /\
       (FVc (CALL p xs es) d =
              (FVxs xs) UNION (FVes es) UNION SL ((FST o SND o SND o d) p))";;



% =========================================================== %
% We now define the function that yields the global variables %
% of a construct; these are presently the global variables    %
% of any procedure called from within the construct.          %
% =========================================================== %


let GVc = 
    new_recursive_definition false cmd_Axiom `GVc`
      "(GVc (SKIP)     (r:env) = {})    /\
       (GVc (ABORT)          r = {})    /\
       (GVc (ASSIGN x e)     r = {})    /\
       (GVc (SEQ c1 c2)      r = GVc c1 r UNION GVc c2 r)  /\
       (GVc (IF b c1 c2)     r = GVc c1 r UNION GVc c2 r)  /\
       (GVc (WHILE a pr b c) r = GVc c r)  /\
       (GVc (CALL p xs es)   r = SL ((FST o SND o SND o r) p))";;


let GVc_SUBSET_FVc =
 prove_thm
  (`GVc_SUBSET_FVc`,
   "!c r. GVc c r SUBSET FVc c r",
   REWRITE_TAC[SUBSET_DEF]
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[GVc;FVc_DEF]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[IN;IN_UNION]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% GVc_SUBSET_FVc:
g "!c r. GVc c r SUBSET FVc c r";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(INDUCT_THEN cmd_induct ASSUME_TAC);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;

   e(RES_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;
   e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;

   e(RES_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;
   e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[GVc;FVc_DEF;IN;IN_UNION]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[GVc;FVc_DEF]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[IN_UNION]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%



% =========================================================== %
% We now prove several independence theorems for expressions, %
% expression lists, boolean expressions, and commands.        %
% These state the essential unchangedness of the semantics of %
% each construct, under modifications of the state which are  %
% outside the free variables of the construct.                %
% =========================================================== %



let E_BND_THM =
 prove_thm
  (`E_BND_THM`,
   "!e s1 n s2.
    E e s1 n s2 ==> (!x m. (~x IN FVe e) ==> E e (BND m x s1) n (BND m x s2))",
   EXP_INDUCT_TAC
   THEN REWRITE_TAC[FVe_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THENL
     [ NUM_TAC;

       POP_ASSUM (SUBST1_TAC o SYM o SPEC_ALL
             o (MATCH_MP (ISPEC "m:num" BND_NOT_MATCH))
             o NOT_EQ_SYM)
       THEN PVAR_TAC;

       FIRST_ASSUM (REWRITE_THM o GEN_ALL
             o (MATCH_MP (INST_TYPE [":num",":*"] BND_COMMUTE)))
       THEN INC_TAC
       THEN FIRST_ASSUM MATCH_MP_TAC
       THEN ASM_REWRITE_TAC[FVe_DEF;IN];

       RES_TAC
       THEN PLUS_TAC
       THEN EXISTS_TAC "BND m' x (s2:state)"
       THEN ASM_REWRITE_TAC[];

       RES_TAC
       THEN MINUS_TAC
       THEN EXISTS_TAC "BND m' x (s2:state)"
       THEN ASM_REWRITE_TAC[];

       RES_TAC
       THEN MULT_TAC
       THEN EXISTS_TAC "BND m' x (s2:state)"
       THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!e s1 n s2.
    E e s1 n s2 ==> (!x m. (~x IN FVe e) ==> E e (BND m x s1) n (BND m x s2))";;
e(EXP_INDUCT_TAC
  THEN REWRITE_TAC[FVe_DEF;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(NUM_TAC);;

 e(POP_ASSUM (SUBST1_TAC o pthm o SYM o pthm o SPEC_ALL
             o (MATCH_MP (ISPEC "m:num" BND_NOT_MATCH))
             o NOT_EQ_SYM));;
 e(PVAR_TAC);;

 e(FIRST_ASSUM (REWRITE_THM o pthm o GEN_ALL
             o (MATCH_MP (INST_TYPE [":num",":*"] BND_COMMUTE))));;
 e(INC_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[FVe_DEF;IN]);;

 e(RES_TAC);;
 e(PLUS_TAC);;
 e(EXISTS_TAC "BND m' x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(MINUS_TAC);;
 e(EXISTS_TAC "BND m' x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(MULT_TAC);;
 e(EXISTS_TAC "BND m' x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;
%


let ES_BND_THM =
 prove_thm
  (`ES_BND_THM`,
   "!es s1 ns s2.
     ES es s1 ns s2 ==>
     (!x m. (~x IN FVes es) ==>
        ES es (BND m x s1) ns (BND m x s2))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVes_DEF;ES_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REPEAT GEN_TAC
   THEN ASM_REWRITE_TAC[]
   THEN STRIP_TAC
   THEN IMP_RES_TAC E_BND_THM
   THEN RES_TAC
   THEN EXISTS_TAC "BND m x (s2':state)"
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!es s1 ns s2.
    ES es s1 ns s2 ==>
    (!x m. (~x IN FVes es) ==>
       ES es (BND m x s1) ns (BND m x s2))";;
e(LIST_INDUCT_TAC
  THEN REWRITE_TAC[FVes_DEF;ES_DEF;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT GEN_TAC THEN STRIP_TAC
  THEN REPEAT GEN_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[]);;
 e(STRIP_TAC);;
 e(IMP_RES_TAC E_BND_THM);;
 e(RES_TAC);;
 e(EXISTS_TAC "BND m x (s2':state)");;
 e(ASM_REWRITE_TAC[]);;
%

let ES_LENGTH_THM =
 prove_thm
  (`ES_LENGTH_THM`,
   "!es s1 ns s2.
     ES es s1 ns s2 ==>
     (LENGTH es = LENGTH ns)",
   LIST_INDUCT_TAC
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS;ES_DEF]
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC "HD (ns:(num)list)"
   THEN EXISTS_TAC "TL (ns:(num)list)"
   THEN RES_TAC
   THEN IMP_RES_TAC CONS
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!es s1 ns s2.
    ES es s1 ns s2 ==>
    (LENGTH es = LENGTH ns)";;
e(LIST_INDUCT_TAC
  THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
  THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS;ES_DEF]
  THEN REPEAT STRIP_TAC);;

 e(EXISTS_TAC "HD (ns:(num)list)");;
 e(EXISTS_TAC "TL (ns:(num)list)");;
 e(RES_TAC);;
 e(IMP_RES_TAC CONS);;
 e(ASM_REWRITE_TAC[]);;
%


let B_BND_THM =
 prove_thm
  (`B_BND_THM`,
   "!b s1 t s2.
    B b s1 t s2 ==> (!x m. (~x IN FVb b) ==> B b (BND m x s1) t (BND m x s2))",
   BEXP_INDUCT_TAC
   THEN REWRITE_TAC[FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC E_BND_THM
   THEN IMP_RES_TAC ES_BND_THM
   THEN RES_TAC
   THEN BEXP_TAC
   THENL
     [ EXISTS_TAC "BND m' x (s2:state)";

       EXISTS_TAC "BND m' x (s2:state)";

       EXISTS_TAC "BND m x (s2:state)";

       EXISTS_TAC "BND m x (s2:state)";

       EXISTS_TAC "BND m x (s2:state)";

       ALL_TAC
     ]
   THEN ASM_REWRITE_TAC[]
  );;


%
g "!b s1 t s2.
    B b s1 t s2 ==> (!x m. (~x IN FVb b) ==> B b (BND m x s1) t (BND m x s2))";;
e(BEXP_INDUCT_TAC
  THEN REWRITE_TAC[FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(IMP_RES_TAC E_BND_THM);;
 e(BEQ_TAC);;
 e(EXISTS_TAC "BND m' x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC E_BND_THM);;
 e(LESS_TAC);;
 e(EXISTS_TAC "BND m' x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC ES_BND_THM);;
 e(LLESS_TAC);;
 e(EXISTS_TAC "BND m x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(AND_TAC);;
 e(EXISTS_TAC "BND m x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(OR_TAC);;
 e(EXISTS_TAC "BND m x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(NOT_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


% ===================================================================== %
% The proof of the corresponding C_BND_THM must wait for the definition %
% of well-formedness, as it is only true for well-formed commands.      %
% ===================================================================== %



% ============================================ %
% Independence theorems on lists of variables: %
% ============================================ %



let E_BNDsL_THM =
 prove_thm
  (`E_BNDsL_THM`,
   "!xs ms e s1 n s2.
     (DISJOINT (SL xs) (FVe e)) /\ E e s1 n s2 ==>
     E e (BNDsL ms xs s1) n (BNDsL ms xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[HD;TL;DISJOINT_INSERT]
   THEN IMP_RES_TAC E_BND_THM
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h':num")
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% E_BNDsL_THM:
g "!xs ms e s1 n s2.
     (DISJOINT (SL xs) (FVe e)) /\ E e s1 n s2 ==>
     E e (BNDsL ms xs s1) n (BNDsL ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]);;

  e(REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]);;
  e(REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]);;
  e(REPEAT GEN_TAC
    THEN STRIP_TAC
    THEN ASM_REWRITE_TAC[]);;
  e(IMP_RES_TAC E_BND_THM);;
  e(POP_ASSUM (ASSUME_TAC o SPEC "h':num"));;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let E_BNDsR_THM =
 prove_thm
  (`E_BNDsR_THM`,
   "!xs ms e s1 n s2.
    (DISJOINT (SL xs) (FVe e)) /\ E e s1 n s2 ==>
     E e (BNDsR ms xs s1) n (BNDsR ms xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN RES_TAC
   THEN POP_ASSUM (ASSUME_TAC o SPEC "ms:(num)list")
   THEN IMP_RES_TAC E_BND_THM
   THEN ASM_REWRITE_TAC[]
  );;

% E_BNDsR_THM:
g "!xs ms e s1 n s2.
    (DISJOINT (SL xs) (FVe e)) /\ E e s1 n s2 ==>
     E e (BNDsR ms xs s1) n (BNDsR ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[]);;
e(RES_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC "ms:(num)list"));;
e(IMP_RES_TAC E_BND_THM);;
e(ASM_REWRITE_TAC[]);;
%


let ES_BNDsL_THM =
 prove_thm
  (`ES_BNDsL_THM`,
   "!xs ms es s1 ns s2.
    (DISJOINT (SL xs) (FVes es)) /\ ES es s1 ns s2 ==>
    ES es (BNDsL ms xs s1) ns (BNDsL ms xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_TAC ES_BND_THM
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h':num")
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% ES_BNDsL_THM:
g "!xs ms es s1 ns s2.
    (DISJOINT (SL xs) (FVes es)) /\ ES es s1 ns s2 ==>
    ES es (BNDsL ms xs s1) ns (BNDsL ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(IMP_RES_TAC ES_BND_THM);;
e(POP_ASSUM (ASSUME_TAC o SPEC "h':num"));;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let ES_BNDsR_THM =
 prove_thm
  (`ES_BNDsR_THM`,
   "!xs ms es s1 ns s2.
    (DISJOINT (SL xs) (FVes es)) /\ ES es s1 ns s2 ==>
     ES es (BNDsR ms xs s1) ns (BNDsR ms xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN POP_ASSUM (ASSUME_TAC o SPEC "ms:(num)list")
   THEN IMP_RES_TAC ES_BND_THM
   THEN ASM_REWRITE_TAC[]
  );;

% ES_BNDsR_THM:
g "!xs ms es s1 ns s2.
    (DISJOINT (SL xs) (FVes es)) /\ ES es s1 ns s2 ==>
     ES es (BNDsR ms xs s1) ns (BNDsR ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(RES_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC "ms:(num)list"));;
e(IMP_RES_TAC ES_BND_THM);;
e(ASM_REWRITE_TAC[]);;
%


let B_BNDsL_THM =
 prove_thm
  (`B_BNDsL_THM`,
   "!xs ms b s1 t s2.
    DISJOINT (SL xs) (FVb b) /\ B b s1 t s2 ==>
    B b (BNDsL ms xs s1) t (BNDsL ms xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_TAC B_BND_THM
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h':num")
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% B_BNDsL_THM:
g "!xs ms b s1 t s2.
    DISJOINT (SL xs) (FVb b) /\ B b s1 t s2 ==>
    B b (BNDsL ms xs s1) t (BNDsL ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(IMP_RES_TAC B_BND_THM);;
e(POP_ASSUM (ASSUME_TAC o SPEC "h':num"));;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let B_BNDsR_THM =
 prove_thm
  (`B_BNDsR_THM`,
   "!ms xs b s1 t s2.
    (LENGTH xs = LENGTH ms) ==> DISJOINT (SL xs) (FVb b) ==>
      B b s1 t s2 ==> B b (BNDsR ms xs s1) t (BNDsR ms xs s2)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[BNDsR;SET_LIST;DISJOINT_INSERT;HD;TL]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC B_BND_THM
   THEN ASM_REWRITE_TAC[]
  );;

% B_BNDsR_THM:
g "!xs ms b s1 t s2.
    DISJOINT (SL xs) (FVb b) /\ B b s1 t s2 ==>
    B b (BNDsR ms xs s1) t (BNDsR ms xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT GEN_TAC
  THEN STRIP_TAC);;
e(RES_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC "ms:(num)list"));;
e(IMP_RES_TAC B_BND_THM);;
e(ASM_REWRITE_TAC[]);;
%




% ===================================================================== %
% The following are "extension" theorems in that they enable us to      %
% extend the set of known semantic tuples from ones that are given, to  %
% ones whose states differ ONLY on their free variables.  The BND       %
% theorems above are simple versions of this, where there is only one   %
% different variable; the BNDsL, BNDsR theorems above extend this to    %
% finite lists, but the "extension" theorems work even for infinitely   %
% different states, as long as they are equal on the free variables.    %
% ===================================================================== %


let E_EXTENSION =
 prove_thm
  (`E_EXTENSION`,
   "!e s1 n s2.
     E e s1 n s2 ==>
     (!x s1'.
       (FVe e SUBSET SL x) /\
       (MAP s1 x = MAP s1' x) ==>
       E e s1' n (BNDsL (MAP s2 x) x s1'))",
   EXP_STR_INDUCT_TAC
   THEN REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THENL
     [  % NUM: %
        ASM_REWRITE_TAC[IDENT_BNDsL]
        THEN NUM_TAC;

        % PVAR: %
        ASM_REWRITE_TAC[IDENT_BNDsL]
        THEN REWRITE_ALL_TAC[MAP_EXTENSION]
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[]
        THEN PVAR_TAC;

        % INC: %
        REWRITE_ALL_TAC[FVe_DEF;SUBSET]
        THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE
        THEN INC_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[];

        % PLUS: %
        PLUS_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % MINUS: %
        MINUS_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % MULT: %
        MULT_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ]
     ]
  );;

%< 
g "!e s1 n s2.
    E e s1 n s2 ==>
    (!x s1'.
      (FVe e SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) ==>
      E e s1' n (BNDsL (MAP s2 x) x s1'))";;
e(EXP_STR_INDUCT_TAC);;

% NUM: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[IDENT_BNDsL]);;
e(NUM_TAC);;

% PVAR: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[IDENT_BNDsL]);;
e(REWRITE_ALL_TAC[MAP_EXTENSION]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
e(PVAR_TAC);;

% INC: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_ALL_TAC[FVe_DEF;SUBSET]);;
e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE);;
e(INC_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(ASM_REWRITE_TAC[]);;

% PLUS: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(PLUS_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% MINUS: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(MINUS_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% MULT: %
e(REWRITE_TAC[FVe_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(MULT_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

>%


let ES_EXTENSION =
 prove_thm
  (`ES_EXTENSION`,
   "!es s1 ns s2.
     ES es s1 ns s2 ==>
     (!x s1'.
       (FVes es SUBSET SL x) /\
       (MAP s1 x = MAP s1' x) ==>
       ES es s1' ns (BNDsL (MAP s2 x) x s1'))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVes_DEF;SUBSET;UNION_SUBSET;ES_DEF]
   THENL
     [  % NIL: %
        REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[IDENT_BNDsL];

        % CONS: %
        REPEAT STRIP_TAC
        THEN RES_TAC
        THEN EXISTS_TAC "BNDsL (MAP (s2':state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  IMP_RES_TAC E_EXTENSION;

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s2:state) x";"x:(var)list";
                                  "MAP (s2':state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (IMP_RES_THEN IMP_RES_TAC)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ]
     ]
  );;

%< 
g "!es s1 ns s2.
    ES es s1 ns s2 ==>
    (!x s1'.
      (FVes es SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) ==>
      ES es s1' ns (BNDsL (MAP s2 x) x s1'))";;
e(LIST_INDUCT_TAC);;

% NIL: %
e(REWRITE_TAC[FVes_DEF;SUBSET;UNION_SUBSET]);;
e(REWRITE_TAC[ES_DEF]);;
e(REPEAT STRIP_TAC);;
   e(ASM_REWRITE_TAC[IDENT_BNDsL]);;

   e(ASM_REWRITE_TAC[IDENT_BNDsL]);;

% CONS: %
e(REWRITE_TAC[FVes_DEF;SUBSET;UNION_SUBSET]);;
e(REWRITE_TAC[ES_DEF]);;
e(REPEAT STRIP_TAC);;
  e(RES_TAC);;

  e(EXISTS_TAC "BNDsL (MAP (s2':state) x) x s1'");;
  e(CONJ_TAC);;

    e(IMP_RES_TAC E_EXTENSION);;

    e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                   ISPECL ["MAP (s2:state) x";"x:(var)list";
                           "MAP (s2':state) x";"s1':state"])
                  BNDsL_BNDsL_ABSORPTION));;
    e(FIRST_ASSUM (IMP_RES_THEN (IMP_RES_TAC)));;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

>%


let B_EXTENSION =
 prove_thm
  (`B_EXTENSION`,
   "!b s1 t s2.
     B b s1 t s2 ==>
     (!x s1'.
       (FVb b SUBSET SL x) /\
       (MAP s1 x = MAP s1' x) ==>
       B b s1' t (BNDsL (MAP s2 x) x s1'))",
   BEXP_STR_INDUCT_TAC
   THEN REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THENL

     [  % EQ: %
        BEQ_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  IMP_RES_TAC E_EXTENSION;

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN IMP_RES_TAC E_EXTENSION
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % LESS: %
        LESS_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  IMP_RES_TAC E_EXTENSION;

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN IMP_RES_TAC E_EXTENSION
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % LLESS: %
        LLESS_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  IMP_RES_TAC ES_EXTENSION;

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN IMP_RES_TAC ES_EXTENSION
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % AND: %
        AND_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % OR: %
        OR_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % NOT: %
        NOT_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%< 
g "!b s1 t s2.
    B b s1 t s2 ==>
    (!x s1'.
      (FVb b SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) ==>
      B b s1' t (BNDsL (MAP s2 x) x s1'))";;
e(BEXP_STR_INDUCT_TAC);;

% EQ: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(BEQ_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(IMP_RES_TAC E_EXTENSION);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(IMP_RES_TAC E_EXTENSION);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% LESS: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(LESS_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(IMP_RES_TAC E_EXTENSION);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(IMP_RES_TAC E_EXTENSION);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% LLESS: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(LLESS_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(IMP_RES_TAC ES_EXTENSION);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(IMP_RES_TAC ES_EXTENSION);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% AND: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(AND_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% OR: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(OR_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% NOT: %
e(REWRITE_TAC[FVb_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(NOT_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(ASM_REWRITE_TAC[]);;

>%

% ===================================================================== %
% The proof of the corresponding C_EXTENSION theorem must wait for the  %
% definition of well-formedness, as it is only true for well-formed     %
% commands.                                                             %
% ===================================================================== %




% ============================================================ %
% We prove several theorems about the free variables of        %
% substituted expressions.                                     %
% ============================================================ %

let FVxs_subst =
 prove_thm
  (`FVxs_subst`,
   "!xs ss. FVxs (xs <|xs ss) = IMAGE ss (FVxs xs)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[xs_subst;FVxs_DEF;IMAGE;IMAGE_UNION]
  );;

let FVe_subst =
 prove_thm
  (`FVe_subst`,
   "!e ss. FVe (e <|e ss) = IMAGE ss (FVe e)",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[e_subst;FVe_DEF;IMAGE;IMAGE_UNION]
  );;

let FVes_subst =
 prove_thm
  (`FVes_subst`,
   "!es ss. FVes (es <|es ss) = IMAGE ss (FVes es)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[es_subst;FVes_DEF;FVe_subst;IMAGE;IMAGE_UNION]
  );;

let FVb_subst =
 prove_thm
  (`FVb_subst`,
   "!b ss. FVb (b <|b ss) = IMAGE ss (FVb b)",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[b_subst;FVb_DEF;FVe_subst;FVes_subst;IMAGE;IMAGE_UNION]
  );;

let FVc_subst =
 prove_thm
  (`FVc_subst`,
   "!c r ss glbs.
       (GVc c r SUBSET SL glbs) /\ (MAP ss glbs = glbs) ==>
       (FVc (c <|c ss) r = IMAGE ss (FVc c r))",
   INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[GVc;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[c_subst;FVc_DEF;FVxs_subst;FVe_subst;FVes_subst;FVb_subst;
                       IMAGE;IMAGE_UNION]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[IMAGE_UNION]
   THEN IMP_RES_TAC MAP_IDENT_ELEMENTS_SUBSET
   THEN IMP_RES_TAC MAP_IDENT_IMAGE_IDENT
   THEN ASM_REWRITE_TAC[]
  );;

% FVc_subst
g "!c r ss glbs.
       (GVc c r SUBSET SL glbs) /\ (MAP ss glbs = glbs) ==>
       (FVc (c <|c ss) r = IMAGE ss (FVc c r))";;
e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[GVc;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[c_subst;FVc_DEF;FVxs_subst;FVe_subst;FVes_subst;FVb_subst;
                       IMAGE;IMAGE_UNION]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[IMAGE_UNION]);;
e(IMP_RES_TAC MAP_IDENT_ELEMENTS_SUBSET);;
e(IMP_RES_TAC MAP_IDENT_IMAGE_IDENT);;
e(ASM_REWRITE_TAC[]);;
%


let GVc_subst =
 prove_thm
  (`GVc_subst`,
   "!c r ss. GVc (c <|c ss) r = GVc c r",
   INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[c_subst;GVc;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% GVc_subst
g "!c r ss. GVc (c <|c ss) r = GVc c r";;
e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[c_subst;GVc;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[]);;
%




% ============================================================ %
% We now prove several substitution theorems for programming   %
% language expressions, expression lists, boolean expressions, %
% and commands.   These state the invariance of the semantics  %
% of these phrases under replacement of their free variables.  %
% ============================================================ %


% Program language rewriting and simplification tools. %

let E_REWRITE_RULE = REWRITE_RULE [ exp_distinct; exp_one_one]
                     o ONCE_REWRITE_RULE [E_cases] ;;
let B_REWRITE_RULE = REWRITE_RULE [bexp_distinct;bexp_one_one]
                     o ONCE_REWRITE_RULE [B_cases] ;;
let C_REWRITE_RULE = REWRITE_RULE [ cmd_distinct; cmd_one_one]
                     o ONCE_REWRITE_RULE [C_cases] ;;




let e_subst_biject =
 prove_thm
  (`e_subst_biject`,
   "!e s1 n s2 ss.
    ONE_ONE ss /\ ONTO ss ==>
    (E (e <|e ss) s1 n s2 = E e (s1 o ss) n (s2 o ss))",

   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[e_subst]
   THEN REPEAT STRIP_TAC
   THENL
       % NUM: %
     [ EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN NUM_TAC;

            IMP_RES_TAC o_ONTO_IMP_11
            THEN ASM_REWRITE_TAC[]
            THEN NUM_TAC
         ];

       % PVAR: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  POP_ASSUM REWRITE_THM
            THEN POP_ASSUM REWRITE_THM
            THEN POP_ASSUM (REWRITE_THM o SYM)
            THEN REWRITE_TAC[SYM (ISPECL ["s1:state";"ss:vsubst";"v:var"] o_THM)]
            THEN PVAR_TAC;

            IMP_RES_TAC o_ONTO_IMP_11
            THEN ASM_REWRITE_TAC[o_THM]
            THEN PVAR_TAC
         ];

       % INC: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[SYM (ASSUME "(ss:vsubst)v = x")]
            THEN IMP_RES_THEN REWRITE_THM BND_o_ONE_ONE
            THEN INC_TAC
            THEN POP_ASSUM (STRIP_ASSUME_TAC o E_REWRITE_RULE)
            THEN POP_ASSUM REWRITE_THM
            THEN POP_ASSUM REWRITE_THM
            THEN POP_ASSUM (REWRITE_THM o SYM)
            THEN REWRITE_TAC[SYM (ASSUME "(ss:vsubst)v = x")]
            THEN REWRITE_TAC[SYM (ISPECL ["s1:state";"ss:vsubst";"v:var"] o_THM)]
            THEN PVAR_TAC;

            POP_ASSUM (STRIP_ASSUME_TAC o E_REWRITE_RULE)
            THEN POP_ASSUM REWRITE_ALL_THM
            THEN POP_ASSUM REWRITE_ALL_THM
            THEN POP_ASSUM (REWRITE_ALL_THM o SYM)
            THEN IMP_RES_THEN (REWRITE_ALL_THM o SYM o SPEC_ALL) BND_o_ONE_ONE
            THEN IMP_RES_THEN REWRITE_ALL_THM o_ONTO_11
            THEN ASM_REWRITE_TAC[]
            THEN INC_TAC
            THEN REWRITE_TAC[o_THM]
            THEN PVAR_TAC
         ];

       % PLUS: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN PLUS_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                      ["(e:exp)=e1";"(e':exp)=e2"]
            THEN ASM_REWRITE_TAC[]
            THEN PLUS_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN RES_THEN (IMP_RES_THEN REWRITE_THM)
            THEN REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % MINUS: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN MINUS_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                      ["(e:exp)=e1";"(e':exp)=e2"]
            THEN ASM_REWRITE_TAC[]
            THEN MINUS_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN RES_THEN (IMP_RES_THEN REWRITE_THM)
            THEN REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % MULT: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN MULT_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                      ["(e:exp)=e1";"(e':exp)=e2"]
            THEN ASM_REWRITE_TAC[]
            THEN MULT_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN RES_THEN (IMP_RES_THEN REWRITE_THM)
            THEN REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ]
     ]
  );;

%< e_subst_biject
g "!e s1 n s2 ss.
    ONE_ONE ss /\ ONTO ss ==>
    (E (e <|e ss) s1 n s2 = E e (s1 o ss) n (s2 o ss))";;

e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC);;

% NUM: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(NUM_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(IMP_RES_TAC o_ONTO_IMP_11);;
  e(ASM_REWRITE_TAC[]);;
  e(NUM_TAC);;

% PVAR: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(POP_ASSUM REWRITE_THM);;
  e(POP_ASSUM REWRITE_THM);;
  e(POP_ASSUM (REWRITE_THM o SYM));;
  e(REWRITE_TAC[SYM (ISPECL ["s1:state";"ss:vsubst";"v:var"] o_THM)]);;
  e(PVAR_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(IMP_RES_TAC o_ONTO_IMP_11);;
  e(ASM_REWRITE_TAC[o_THM]);;
  e(PVAR_TAC);;

% INC: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[SYM (ASSUME "(ss:vsubst)v = x")]);;
  e(IMP_RES_THEN REWRITE_THM BND_o_ONE_ONE);;
  e(INC_TAC);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(POP_ASSUM (\th. REWRITE_TAC[th]));;
  e(POP_ASSUM (\th. REWRITE_TAC[th]));;
  e(POP_ASSUM (\th. REWRITE_TAC[SYM th]));;
  e(REWRITE_TAC[SYM (ASSUME "(ss:vsubst)v = x")]);;
  e(REWRITE_TAC[SYM (ISPECL ["s1:state";"ss:vsubst";"v:var"] o_THM)]);;
  e(PVAR_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(POP_ASSUM (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(POP_ASSUM REWRITE_ALL_THM);;
  e(POP_ASSUM REWRITE_ALL_THM);;
  e(POP_ASSUM (REWRITE_ALL_THM o SYM));;
  e(IMP_RES_THEN (REWRITE_ALL_THM o SYM o SPEC_ALL) BND_o_ONE_ONE);;
  e(IMP_RES_THEN REWRITE_ALL_THM o_ONTO_11);;
  e(ASM_REWRITE_TAC[]);;
  e(INC_TAC);;
  e(REWRITE_TAC[o_THM]);;
  e(PVAR_TAC);;

% PLUS: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(PLUS_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(e:exp)=e1";"(e':exp)=e2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(PLUS_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(RES_THEN (IMP_RES_THEN REWRITE_THM));;
  e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% MINUS: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(MINUS_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(e:exp)=e1";"(e':exp)=e2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(MINUS_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(RES_THEN (IMP_RES_THEN REWRITE_THM));;
  e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% MULT: %
 e(REWRITE_TAC[e_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(MULT_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o E_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(e:exp)=e1";"(e':exp)=e2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(MULT_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(RES_THEN (IMP_RES_THEN REWRITE_THM));;
  e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;
>%



let e_subst_THM =
 prove_thm
  (`e_subst_THM`,
   "!e s1 n s2 ys xs. 
    E (e <|e (ys //v xs)) s1 n s2 =
    E e (s1 o (ys //v xs)) n (s2 o (ys //v xs))",

   REPEAT GEN_TAC
   THEN MATCH_MP_TAC e_subst_biject
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
  );;

%
g "!e s1 n s2 ys xs. 
    E (e <|e (ys //v xs)) s1 n s2 =
    E e (s1 o (ys //v xs)) n (s2 o (ys //v xs))";;

e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC e_subst_biject);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
%



let es_subst_biject =
 prove_thm
  (`es_subst_biject`,
   "!es s1 ns s2 ss.
     ONE_ONE ss /\ ONTO ss ==>
     (ES (es <|es ss) s1 ns s2 = ES es (s1 o ss) ns (s2 o ss))",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[es_subst;ES_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM o_ONTO_11
   THEN RES_THEN (IMP_RES_THEN REWRITE_THM)
   THEN IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) e_subst_biject
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THENL
     [ EXISTS_TAC "(s2':state) o (ss:vsubst)"
       THEN ASM_REWRITE_TAC[];

       EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
       THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
       THEN IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE
       THEN ASM_REWRITE_TAC[I_o_ID]
     ]
  );;

%< es_subst_biject
g "!es s1 ns s2 ss.
     ONE_ONE ss /\ ONTO ss ==>
     (ES (es <|es ss) s1 ns s2 = ES es (s1 o ss) ns (s2 o ss))";;

e(LIST_INDUCT_TAC);;

% NULL: %
 e(REWRITE_TAC[es_subst;ES_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN REWRITE_THM o_ONTO_11);;

% CONS: %
 e(REWRITE_TAC[es_subst;ES_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(RES_THEN (IMP_RES_THEN REWRITE_THM));;
 e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) e_subst_biject);;
 e(EQ_TAC);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;
>%



let es_subst_THM =
 prove_thm
  (`es_subst_THM`,
   "!es s1 ns s2 ys xs. 
     ES (es <|es (ys //v xs)) s1 ns s2 =
     ES es (s1 o (ys //v xs)) ns (s2 o (ys //v xs))",

   REPEAT GEN_TAC
   THEN MATCH_MP_TAC es_subst_biject
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
  );;

%
g "!es s1 ns s2 ys xs. 
     ES (es <|es (ys //v xs)) s1 ns s2 =
     ES es (s1 o (ys //v xs)) ns (s2 o (ys //v xs))";;

e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC es_subst_biject);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
%




let b_subst_biject =
 prove_thm
  (`b_subst_biject`,
   "!b s1 t s2 ss.
     ONE_ONE ss /\ ONTO ss ==>
     (B (b <|b ss) s1 t s2 = B b (s1 o ss) t (s2 o ss))",

   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[b_subst]
   THEN REPEAT STRIP_TAC
   THENL
       % EQ: %
     [ EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN BEQ_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN IMP_RES2_THEN REWRITE_THM
                     (ONCE_REWRITE_RULE[EQ_SYM_EQ] e_subst_biject)
            THEN ASM_REWRITE_TAC[];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(e0:exp)=e1'";"(e1:exp)=e2"]
            THEN ASM_REWRITE_TAC[]
            THEN BEQ_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN IMP_RES2_THEN REWRITE_THM e_subst_biject
            THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % LESS: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN LESS_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN IMP_RES2_THEN REWRITE_THM
                     (ONCE_REWRITE_RULE[EQ_SYM_EQ] e_subst_biject)
            THEN ASM_REWRITE_TAC[];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(e0:exp)=e1'";"(e1:exp)=e2"]
            THEN ASM_REWRITE_TAC[]
            THEN LESS_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN IMP_RES2_THEN REWRITE_THM e_subst_biject
            THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % LLESS: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN LLESS_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN IMP_RES2_THEN REWRITE_THM
                     (ONCE_REWRITE_RULE[EQ_SYM_EQ] es_subst_biject)
            THEN ASM_REWRITE_TAC[];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(l0:exps)=es1";"(l1:exps)=es2"]
            THEN ASM_REWRITE_TAC[]
            THEN LLESS_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN IMP_RES2_THEN REWRITE_THM es_subst_biject
            THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % AND: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN AND_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(b:bexp)=b1";"(b':bexp)=b2"]
            THEN ASM_REWRITE_TAC[]
            THEN AND_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN RES2_THEN REWRITE_THM
            THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % OR: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN OR_TAC
            THEN EXISTS_TAC "(s2':state) o (ss:vsubst)"
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(b:bexp)=b1";"(b':bexp)=b2"]
            THEN ASM_REWRITE_TAC[]
            THEN OR_TAC
            THEN EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)"
            THEN RES2_THEN REWRITE_THM
            THEN REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN ASM_REWRITE_TAC[I_o_ID]
         ];

       % NOT: %
       EQ_TAC
       THEN DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE)
       THENL
         [  ASM_REWRITE_TAC[]
            THEN NOT_TAC
            THEN RES_TAC
            THEN EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC)
            THEN FILTER_ASM_REWRITE_TAC ($not o is_forall) [];

            MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
                     ["(b:bexp)=b'"]
            THEN ASM_REWRITE_TAC[]
            THEN NOT_TAC
            THEN RES2_THEN REWRITE_THM
            THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

%< b_subst_biject
g "!b s1 t s2 ss.
     ONE_ONE ss /\ ONTO ss ==>
     (B (b <|b ss) s1 t s2 = B b (s1 o ss) t (s2 o ss))";;

e(INDUCT_THEN bexp_induct STRIP_ASSUME_TAC);;

% EQ: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(BEQ_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(IMP_RES2_THEN REWRITE_THM
          (ONCE_REWRITE_RULE[EQ_SYM_EQ] e_subst_biject));;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(e0:exp)=e1'";"(e1:exp)=e2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(BEQ_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(IMP_RES2_THEN REWRITE_THM e_subst_biject);;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% LESS: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(LESS_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(IMP_RES2_THEN REWRITE_THM
          (ONCE_REWRITE_RULE[EQ_SYM_EQ] e_subst_biject));;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(e0:exp)=e1'";"(e1:exp)=e2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(LESS_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(IMP_RES2_THEN REWRITE_THM e_subst_biject);;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% LLESS: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(LLESS_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(IMP_RES2_THEN REWRITE_THM
          (ONCE_REWRITE_RULE[EQ_SYM_EQ] es_subst_biject));;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(l0:exps)=es1";"(l1:exps)=es2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(LLESS_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(IMP_RES2_THEN REWRITE_THM es_subst_biject);;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% AND: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(AND_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(b:bexp)=b1";"(b':bexp)=b2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(AND_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(RES2_THEN REWRITE_THM);;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% OR: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(OR_TAC);;
  e(EXISTS_TAC "(s2':state) o (ss:vsubst)");;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(b:bexp)=b1";"(b':bexp)=b2"]);;
  e(ASM_REWRITE_TAC[]);;
  e(OR_TAC);;
  e(EXISTS_TAC "(s2':state) o (\y. @z. (ss:vsubst)z = y)");;
  e(RES2_THEN REWRITE_THM);;
  e(REWRITE_TAC[SYM (SPEC_ALL o_ASSOC)]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(ASM_REWRITE_TAC[I_o_ID]);;

% NOT: %
 e(REWRITE_TAC[b_subst]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(ASM_REWRITE_TAC[]);;
  e(NOT_TAC);;
  e(RES_TAC);;
  e(EVERY_ASSUM (\th. (REWRITE_THM o SYM o SPEC_ALL) th ? ALL_TAC));;
  e(FILTER_ASM_REWRITE_TAC ($not o is_forall) []);;

  e(DISCH_THEN (STRIP_ASSUME_TAC o B_REWRITE_RULE));;
  e(MAP_EVERY (UNASSUME_THEN (REWRITE_ALL_THM o SYM))
               ["(b:bexp)=b'"]);;
  e(ASM_REWRITE_TAC[]);;
  e(NOT_TAC);;
  e(RES2_THEN REWRITE_THM);;
  e(ASM_REWRITE_TAC[]);;
>%



let b_subst_THM =
 prove_thm
  (`b_subst_THM`,
   "!b s1 t s2 ys xs. 
     B (b <|b (ys //v xs)) s1 t s2 =
     B b (s1 o (ys //v xs)) t (s2 o (ys //v xs))",

   REPEAT GEN_TAC
   THEN MATCH_MP_TAC b_subst_biject
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
  );;

%
g "!b s1 t s2 ys xs. 
     B (b <|b (ys //v xs)) s1 t s2 =
     B b (s1 o (ys //v xs)) t (s2 o (ys //v xs))";;

e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC b_subst_biject);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
%


% ============================================================= %
% The corresponding c_subst_biject and c_subst_THM must wait    %
% for the definition of well-formedness for their proofs.       %
% ============================================================= %





% ========================= %
% FREE ONLY IMPACT THEOREMS %
% ========================= %

% =================================================================== %
% The purpose of these theorems is to state that the only impact that %
% executing a phrase of the programming language has is to alter the  %
% values of the free variables of that phrase; all other variables    %
% are unaffected.                                                     %
% =================================================================== %



let E_FREE_ONLY_IMPACT =
 prove_thm
  (`E_FREE_ONLY_IMPACT`,
   "!e s1 n s2.
     E e s1 n s2 ==>
      (!z. (~z IN FVe e) ==>
           (s2 z = s1 z))",
   EXP_INDUCT_TAC
   THEN REWRITE_ALL_TAC[FVe_DEF;IN]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN (IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         ORELSE ALL_TAC)
   THEN RES_THEN REWRITE_THM
  );;

% E_FREE_ONLY_IMPACT:
g "!e s1 n s2.
    E e s1 n s2 ==>
     (!z. (~z IN FVe e) ==>
          (s2 z = s1 z))";;
e(EXP_INDUCT_TAC);;

e(REWRITE_ALL_TAC[FVe_DEF;IN]);;

e(REWRITE_ALL_TAC[FVe_DEF;IN]);;

e(REWRITE_ALL_TAC[FVe_DEF;IN]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
e(RES_THEN REWRITE_THM);;

e(REWRITE_ALL_TAC[FVe_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_THEN REWRITE_THM);;
%


let ES_FREE_ONLY_IMPACT =
 prove_thm
  (`ES_FREE_ONLY_IMPACT`,
   "!es s1 ns s2.
     ES es s1 ns s2 ==>
      (!z. (~z IN FVes es) ==>
           (s2 z = s1 z))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[ES_DEF;FVes_DEF;IN]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC E_FREE_ONLY_IMPACT
   THEN ASM_REWRITE_TAC[]
  );;

% ES_FREE_ONLY_IMPACT:
g "!es s1 ns s2.
    ES es s1 ns s2 ==>
     (!z. (~z IN FVes es) ==>
          (s2 z = s1 z))";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[ES_DEF;FVes_DEF;IN]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[ES_DEF;FVes_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC E_FREE_ONLY_IMPACT);;
e(ASM_REWRITE_TAC[]);;
%


let B_FREE_ONLY_IMPACT =
 prove_thm
  (`B_FREE_ONLY_IMPACT`,
   "!b s1 t s2.
     B b s1 t s2 ==>
      (!z. (~z IN FVb b) ==>
           (s2 z = s1 z))",
   BEXP_INDUCT_TAC
   THEN REWRITE_TAC[FVb_DEF;IN]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC E_FREE_ONLY_IMPACT
   THEN IMP_RES_TAC ES_FREE_ONLY_IMPACT
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% B_FREE_ONLY_IMPACT:
g "!b s1 t s2.
    B b s1 t s2 ==>
     (!z. (~z IN FVb b) ==>
          (s2 z = s1 z))";;
e(BEXP_INDUCT_TAC);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC E_FREE_ONLY_IMPACT);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC E_FREE_ONLY_IMPACT);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC ES_FREE_ONLY_IMPACT);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[FVb_DEF;IN]);;
e(ASM_REWRITE_TAC[]);;
%


let FVxs_SL =
 prove_thm
  (`FVxs_SL`,
   "!xs. FVxs xs = SL xs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[FVxs_DEF;SET_LIST]
  );;

% FVxs_SL
g "!xs. FVxs xs = SL xs";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[FVxs_DEF;SET_LIST]);;

  e(ASM_REWRITE_TAC[FVxs_DEF;SET_LIST]);;
%


let NOT_IN_APPEND =
 prove_thm
  (`NOT_IN_APPEND`,
   "!xs ys (z:*).
     ~z IN (SL xs) /\ ~z IN (SL ys) ==> ~z IN (SL (APPEND xs ys))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% NOT_IN_APPEND
g "!xs ys (z:*).
     ~z IN (SL xs) /\ ~z IN (SL ys) ==> ~z IN (SL (APPEND xs ys))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN;DE_MORGAN_THM]);;

  e(REWRITE_TAC[APPEND;SET_LIST;IN;DE_MORGAN_THM]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%




let xs_subst_EQ =
 prove_thm
  (`xs_subst_EQ`,
   "!xs ss1 ss2.
      (!x. x IN FVxs xs ==> (ss1 x = ss2 x)) ==>
      (xs <|xs ss1 = xs <|xs ss2)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVxs_DEF;IN;IN_UNION;xs_subst]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
   THEN REWRITE_TAC[CONS_11]
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% xs_subst_EQ
g "!xs ss1 ss2.
      (!x. x IN FVxs xs ==> (ss1 x = ss2 x)) ==>
      (xs <|xs ss1 = xs <|xs ss2)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[FVxs_DEF;IN;IN_UNION;xs_subst]);;

 e(REWRITE_TAC[FVxs_DEF;IN;IN_UNION;xs_subst]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REWRITE_TAC[]);;

  e(REWRITE_TAC[CONS_11]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;
%




let e_subst_EQ =
 prove_thm
  (`e_subst_EQ`,
   "!e (ss1:vsubst) ss2.
     (!x. x IN (FVe e) ==> (ss1 x = ss2 x)) ==>
     ((e <|e ss1) = (e <|e ss2))",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]
   THEN REPEAT STRIP_TAC
   THENL
    [ FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THEN REWRITE_TAC[];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THEN REWRITE_TAC[];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "e' <|e ss1 = e' <|e ss2" REWRITE_THM
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "e' <|e ss1 = e' <|e ss2" REWRITE_THM
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "e' <|e ss1 = e' <|e ss2" REWRITE_THM
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< e_subst_EQ
g "!e (ss1:vsubst) ss2.
     (!x. x IN (FVe e) ==> (ss1 x = ss2 x)) ==>
     ((e <|e ss1) = (e <|e ss2))";;

e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC);;

% NUM: %
 e(REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]);;

% PVAR: %
 e(REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
 e(REWRITE_TAC[]);;

% INC: %
 e(REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
 e(REWRITE_TAC[]);;

% PLUS: %
 e(REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "e' <|e ss1 = e' <|e ss2" REWRITE_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% MINUS: %
 e(REWRITE_TAC[FVe_DEF;e_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "e' <|e ss1 = e' <|e ss2" REWRITE_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;
>%


let es_subst_EQ =
 prove_thm
  (`es_subst_EQ`,
   "!es (ss1:vsubst) ss2.
     (!x. x IN (FVes es) ==> (ss1 x = ss2 x)) ==>
     ((es <|es ss1) = (es <|es ss2))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVes_DEF;es_subst;IN;IN_UNION]
   THEN REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM e_subst_EQ
   THENL
    [ REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< es_subst_EQ
g "!es (ss1:vsubst) ss2.
     (!x. x IN (FVes es) ==> (ss1 x = ss2 x)) ==>
     ((es <|es ss1) = (es <|es ss2))";;

e(LIST_INDUCT_TAC);;

% NIL: %
 e(REWRITE_TAC[FVes_DEF;es_subst;IN;IN_UNION]);;

% CONS: %
 e(REWRITE_TAC[FVes_DEF;es_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM e_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;
>%



let b_subst_EQ =
 prove_thm
  (`b_subst_EQ`,
   "!b (ss1:vsubst) ss2.
     (!x. x IN (FVb b) ==> (ss1 x = ss2 x)) ==>
     ((b <|b ss1) = (b <|b ss2))",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]
   THEN REPEAT STRIP_TAC
   THENL
    [ USE_IMP_THEN REWRITE_THM e_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "e1 <|e ss1 = e1 <|e ss2" REWRITE_THM
         THEN MATCH_MP_TAC e_subst_EQ
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      USE_IMP_THEN REWRITE_THM e_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "e1 <|e ss1 = e1 <|e ss2" REWRITE_THM
         THEN MATCH_MP_TAC e_subst_EQ
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      USE_IMP_THEN REWRITE_THM es_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "l1 <|es ss1 = l1 <|es ss2" REWRITE_THM
         THEN MATCH_MP_TAC es_subst_EQ
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "b' <|b ss1 = b' <|b ss2" REWRITE_THM
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         SUBGOAL_THEN "b' <|b ss1 = b' <|b ss2" REWRITE_THM
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< b_subst_EQ
g "!b (ss1:vsubst) ss2.
     (!x. x IN (FVb b) ==> (ss1 x = ss2 x)) ==>
     ((b <|b ss1) = (b <|b ss2))";;

e(INDUCT_THEN bexp_induct STRIP_ASSUME_TAC);;

% EQ: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM e_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "e1 <|e ss1 = e1 <|e ss2" REWRITE_THM);;
  e(MATCH_MP_TAC e_subst_EQ);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% LESS: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM e_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "e1 <|e ss1 = e1 <|e ss2" REWRITE_THM);;
  e(MATCH_MP_TAC e_subst_EQ);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% LLESS: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM es_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "l1 <|es ss1 = l1 <|es ss2" REWRITE_THM);;
  e(MATCH_MP_TAC es_subst_EQ);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% AND: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "b' <|b ss1 = b' <|b ss2" REWRITE_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% OR: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBGOAL_THEN "b' <|b ss1 = b' <|b ss2" REWRITE_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% NOT: %
 e(REWRITE_TAC[FVb_DEF;b_subst;IN;IN_UNION]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%



let c_subst_EQ_LEMMA =
 TAC_PROOF
  (([],
    "!c1 r s1 s2.
     C c1 r s1 s2 ==>
     (!c (ss1:vsubst) ss2.
        (c <|c ss2 = c1) /\
        (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
        C (c <|c ss1) r s1 s2)"),
   CMD_STR_INDUCT_TAC
   THENL
    [
% SKIP: %
      GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN SKIP_TAC;

% ASSIGN: %
      GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM e_subst_EQ
         THENL
          [ REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            ASM_REWRITE_TAC[]
            THEN ASSIGN_TAC
            THEN ASM_REWRITE_TAC[]
          ]
       ];

% SEQ: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN SEQ_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "ss2:vsubst"
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[];

% IF_T: %
      GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         IF_T_TAC
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "ss2:vsubst"
         THEN ASM_REWRITE_TAC[]
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

% IF_F: %
      GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         IF_F_TAC
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "ss2:vsubst"
         THEN ASM_REWRITE_TAC[]
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

% WHILE_T: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN WHILE_T_TAC
      THEN EXISTS_TAC "s2:state"
      THEN EXISTS_TAC "s3:state"
      THEN REWRITE_TAC[ONCE_REWRITE_RULE[EQ_SYM_EQ] c_subst]
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[]
         THEN CONJ_TAC
         THENL
          [ FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "ss2:vsubst"
            THEN ASM_REWRITE_TAC[]
            THEN REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "ss2:vsubst"
            THEN ASM_REWRITE_TAC[c_subst;cmd_one_one;FVc_DEF;IN_UNION]
          ]
       ];

% WHILE_F: %
      GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN WHILE_F_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[]
       ];

% CALL: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN UNASSUME_THEN REWRITE_ALL_THM "(s:string) = p"
      THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
                  "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
      THEN USE_IMP_THEN REWRITE_THM xs_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM es_subst_EQ
         THENL
          [ REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            ASM_REWRITE_TAC[]
            THEN CALL_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:(var)list"
            THEN EXISTS_TAC "vals:(var)list"
            THEN EXISTS_TAC "ns:(num)list"
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "(APPEND xs vals') //v (APPEND vars vals)"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

%<
g "!c1 r s1 s2.
     C c1 r s1 s2 ==>
     (!c (ss1:vsubst) ss2.
        (c <|c ss2 = c1) /\
        (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
        C (c <|c ss1) r s1 s2)";;
e(CMD_STR_INDUCT_TAC);;

% SKIP: %
 e(GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(SKIP_TAC);;

% ASSIGN: %
 e(GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM e_subst_EQ);;

   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;
   e(ASSIGN_TAC);;
   e(ASM_REWRITE_TAC[]);;

% SEQ: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(SEQ_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% IF_T: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(IF_T_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% IF_F: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(IF_F_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% WHILE_T: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(WHILE_T_TAC);;
 e(EXISTS_TAC "s2:state"
    THEN EXISTS_TAC "s3:state");;
 e(REWRITE_TAC[ONCE_REWRITE_RULE[EQ_SYM_EQ] c_subst]);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(CONJ_TAC);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "ss2:vsubst");;
   e(ASM_REWRITE_TAC[]);;
   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "ss2:vsubst");;
   e(ASM_REWRITE_TAC[c_subst;cmd_one_one;FVc_DEF;IN_UNION]);;

% WHILE_F: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(WHILE_F_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

% CALL: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(UNASSUME_THEN REWRITE_ALL_THM "(s:string) = p");;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
          "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
let st3 = get_state();;
set_state(st3);;
 e(USE_IMP_THEN REWRITE_THM xs_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM es_subst_EQ);;

   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:(var)list"
     THEN EXISTS_TAC "vals:(var)list"
     THEN EXISTS_TAC "ns:(num)list"
     THEN EXISTS_TAC "glbs:(var)list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "(APPEND xs vals') //v (APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

>%


let c_subst_IMP =
    REWRITE_RULE[AND_IMP_INTRO]
           (CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
                        c_subst_EQ_LEMMA);;


let c_subst_EQ =
 prove_thm
  (`c_subst_EQ`,
   "!ss2 ss1 c r s1 s2.
      (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
      (C (c <|c ss1) r s1 s2 = C (c <|c ss2) r s1 s2)",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN DISCH_TAC
   THEN MATCH_MP_TAC c_subst_IMP
   THENL
    [ EXISTS_TAC "c <|c ss1"
      THEN EXISTS_TAC "ss1:vsubst"
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      EXISTS_TAC "c <|c ss2"
      THEN EXISTS_TAC "ss2:vsubst"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< c_subst_EQ
g "!ss2 ss1 c r s1 s2.
      (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
      (C (c <|c ss1) r s1 s2 = C (c <|c ss2) r s1 s2)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(MATCH_MP_TAC c_subst_IMP);;
 e(EXISTS_TAC "c <|c ss1"
   THEN EXISTS_TAC "ss1:vsubst");;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(DISCH_TAC);;
 e(MATCH_MP_TAC c_subst_IMP);;
 e(EXISTS_TAC "c <|c ss2"
   THEN EXISTS_TAC "ss2:vsubst");;
 e(ASM_REWRITE_TAC[]);;
>%





% ================================================================== %
% We cannot prove the C_FREE_ONLY_IMPACT theorem until we define     %
% the well-formedness predicate for commands.                        %
% ================================================================== %




close_theory();;

