% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains an extension of the `finite_sets` theory, and 
  provides several theorems which are related to finite sets, as well
  as theorems about one-to-one functions and onto functions (bijections).
  Some additional theorems of general utility are also proved here.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `more_finite_sets`;;


% --------------------------------------------------------------------- %
% Need the finite_sets library.                                         %
% --------------------------------------------------------------------- %
load_library `finite_sets`;;


loadf `semantics_tactics`;;



let IN_NOT_IN = prove_thm
   (`IN_NOT_IN`,
    "!s (x:*) y. (x IN s) /\ ~(y IN s) ==> ~(x = y)",
    REPEAT STRIP_TAC
    THEN POP_ASSUM REWRITE_ALL_THM
    THEN RES_TAC
   );;

% IN_NOT_IN
g "!s (x:*) y. (x IN s) /\ ~(y IN s) ==> ~(x = y)";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM REWRITE_ALL_THM);;
e(RES_TAC);;
%


let NOT_IN_SUBSET =
 prove_thm
  (`NOT_IN_SUBSET`,
   "!s t (x:*). ~(x IN s) /\ (t SUBSET s) ==> ~(x IN t)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC SUBSET_DEF
   THEN RES_TAC
  );;

% NOT_IN_SUBSET
g "!s t (x:*). ~(x IN s) /\ (t SUBSET s) ==> ~(x IN t)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC SUBSET_DEF);;
e(RES_TAC);;
%



let IN_DISJOINT_IMP =
 prove_thm
  (`IN_DISJOINT_IMP`,
   "!s t (x:*). DISJOINT s t ==>
       ((x IN s ==> ~(x IN t)) /\
        (x IN t ==> ~(x IN s)))",
   SET_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_EMPTY;NOT_IN_EMPTY;IN_INSERT]
   THEN REWRITE_TAC[DISJOINT_INSERT;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC
   THENL
     [ POP_ASSUM REWRITE_ALL_THM
       THEN ASM_REWRITE_TAC[];

       RES_TAC;

       RES_TAC
       THEN IMP_RES_TAC IN_NOT_IN
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% IN_DISJOINT_IMP
g "!s t (x:*). DISJOINT s t ==>
       ((x IN s ==> ~(x IN t)) /\
        (x IN t ==> ~(x IN s)))";;
e(SET_INDUCT_TAC);;

 e(REWRITE_TAC[DISJOINT_EMPTY;NOT_IN_EMPTY]);;

 e(REWRITE_TAC[DISJOINT_EMPTY;IN_INSERT]);;
 e(REWRITE_TAC[DISJOINT_INSERT;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;

  e(RES_TAC);;
  e(IMP_RES_TAC IN_NOT_IN);;
  e(ASM_REWRITE_TAC[]);;
%


let CONJ_11 =
 prove_thm
  (`CONJ_11`,
   "!a b c d. (a = b) /\ (c = d) ==> (a /\ c = b /\ d)",
   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% CONJ_11
g "!a b c d. (a = b) /\ (c = d) ==> (a /\ c = b /\ d)";;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%



let MAP_I =
 prove_thm
  (`MAP_I`,
   "!lst:(*)list. MAP I lst = lst",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[MAP;I_THM]
  );;

% MAP_I
g "!lst:(*)list. MAP I lst = lst";;
e(LIST_INDUCT_TAC);;

 e(ASM_REWRITE_TAC[MAP;I_THM]);;
%

let APPEND_11 =
 prove_thm
  (`APPEND_11`,
   "!l1 l2 l3 l4:(*)list.
      (LENGTH l3 = LENGTH l1) /\ (LENGTH l4 = LENGTH l2) ==>
      ((APPEND l1 l2 = APPEND l3 l4) = ((l1 = l3) /\ (l2 = l4)))",
   PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REPEAT STRIP_TAC
   THEN PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN IMP_RES_TAC APPEND_LENGTH_EQ
  );;

% APPEND_11
g "!l1 l2 l3 l4:(*)list.
      (LENGTH l3 = LENGTH l1) /\ (LENGTH l4 = LENGTH l2) ==>
      ((APPEND l1 l2 = APPEND l3 l4) = ((l1 = l3) /\ (l2 = l4)))";;
e(PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(REPEAT STRIP_TAC);;
e(PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(IMP_RES_TAC APPEND_LENGTH_EQ);;
%


let NULL_APPEND =
 prove_thm
  (`NULL_APPEND`,
   "!l1 l2:(*) list.
      NULL (APPEND l1 l2) =  NULL l1 /\ NULL l2",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NULL;APPEND]
  );;

%< NULL_APPEND
g "!l1 l2:(*) list.
      NULL (APPEND l1 l2) =  NULL l1 /\ NULL l2";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[NULL;APPEND]);;
>%


let ONE_ONE =
 prove_thm
  (`ONE_ONE`,
   "!f:*->**.
      ONE_ONE f =
      (!x1 x2. (f x1 = f x2) = (x1 = x2))",
   REWRITE_TAC[ONE_ONE_DEF]
   THEN GEN_TAC
   THEN EQ_TAC
   THENL
     [  REPEAT STRIP_TAC
        THEN EQ_TAC
        THEN DISCH_TAC
        THENL
          [  RES_TAC;
             ASM_REWRITE_TAC[]
          ];

        DISCH_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%< ONE_ONE
Place this theorem in theory `more_finite_sets`

set_flag(`show_types`,false);;

g "!f:*->**.
    ONE_ONE f =
    (!x1 x2. (f x1 = f x2) = (x1 = x2))";;
e(REWRITE_TAC[ONE_ONE_DEF]);;
e(GEN_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;
 e(EQ_TAC);;

  e(DISCH_TAC);;
  e(RES_TAC);;

  e(DISCH_TAC);;
  e(ASM_REWRITE_TAC[]);;

 e(DISCH_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%




let ONTO_INVERSE =
 prove_thm
  (`ONTO_INVERSE`,
   "!(ss:*->**) y.
      ONTO ss  ==> (ss(@z. ss z = y) = y)",
   REWRITE_TAC[ONTO_DEF]
   THEN REPEAT STRIP_TAC
   THEN CONV_TAC SELECT_CONV
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN ASM_REWRITE_TAC[]
  );;

% ONTO_INVERSE
g "!(ss:*->**) y.
      ONTO ss  ==> (ss(@z. ss z = y) = y)";;
e(REWRITE_TAC[ONE_ONE;ONTO_DEF]);;
e(REPEAT STRIP_TAC);;
e(CONV_TAC SELECT_CONV);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(ASM_REWRITE_TAC[]);;
%

let o_ONTO_11 =
 prove_thm
  (`o_ONTO_11`,
   "!(ss:*->**) (s1:**->***) s2.
      ONTO ss ==> ((s2 o ss = s1 o ss) = (s2 = s1))",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [  DISCH_TAC
        THEN EXT_TAC "y:**"
        THEN GEN_TAC
        THEN POP_ASSUM (\th. ASSUME_TAC (AP_THM th "@z. (ss:*->**) z = y"))
        THEN UNDISCH_LAST_TAC
        THEN IMP_RES_TAC ONTO_INVERSE
        THEN ASM_REWRITE_TAC[o_THM];

        DISCH_THEN REWRITE_THM
     ]
  );;

% o_ONTO_11
g "!(ss:*->**) (s1:**->***) s2.
      ONTO ss ==> ((s2 o ss = s1 o ss) = (s2 = s1))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(EXT_TAC "y:**");;
 e(GEN_TAC);;
 e(POP_ASSUM (\th. ASSUME_TAC (AP_THM th "@z. (ss:*->**) z = y")));;
 e(UNDISCH_LAST_TAC);;
 e(REWRITE_TAC[o_THM]);;
 e(IMP_RES_TAC ONTO_INVERSE);;
 e(ASM_REWRITE_TAC[]);;

 e(DISCH_THEN REWRITE_THM);;
%

let o_ONTO_IMP_11 =
 prove_thm
  (`o_ONTO_IMP_11`,
   "!(ss:*->**) (s1:**->***) s2.
      ONTO ss /\ (s2 o ss = s1 o ss) ==> (s2 = s1)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC o_ONTO_11
  );;

% o_ONTO_IMP_11
g "!(ss:*->**) (s1:**->***) s2.
      ONTO ss /\ (s2 o ss = s1 o ss) ==> (s2 = s1)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC o_ONTO_11);;
%

let o_INVERSE =
 prove_thm
  (`o_INVERSE`,
   "!(ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      (ss o (\y. @z. ss z = y) = I) /\
      ((\y. @z. ss z = y) o ss = I)",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THENL [ EXT_TAC "x:**"; EXT_TAC "x:*" ]
   THEN GEN_TAC
   THEN REWRITE_TAC[o_THM;I_THM]
   THEN BETA_TAC
   THENL
     [  IMP_RES_THEN REWRITE_THM ONTO_INVERSE;
        ASM_REWRITE_TAC[SELECT_REFL]
     ]
  );;

% o_INVERSE
g "!(ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      (ss o (\y. @z. ss z = y) = I) /\
      ((\y. @z. ss z = y) o ss = I)";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;

 e(EXT_TAC "x:**");;
 e(GEN_TAC);;
 e(REWRITE_TAC[o_THM;I_THM]);;
 e(BETA_TAC);;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;

 e(EXT_TAC "x:*");;
 e(GEN_TAC);;
 e(REWRITE_TAC[o_THM;I_THM]);;
 e(BETA_TAC);;
 e(ASM_REWRITE_TAC[SELECT_REFL]);;
%

let o_EQ_INVERSE =
 prove_thm
  (`o_EQ_INVERSE`,
   "!(ss:*->**) (s1:**->***) s2.
      ONE_ONE ss /\ ONTO ss ==>
      ((s1 o ss = s2) = (s1 = s2 o (\y. @z. ss z = y)))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC o_ONTO_11
   THEN EQ_TAC
   THENL [ DISCH_THEN (REWRITE_THM o SYM);
           DISCH_THEN REWRITE_THM ]
   THEN REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
   THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
   THEN REWRITE_TAC[I_o_ID]
  );;

% o_EQ_INVERSE
g "!(ss:*->**) (s1:**->***) s2.
      ONE_ONE ss /\ ONTO ss ==>
      ((s1 o ss = s2) = (s1 = s2 o (\y. @z. ss z = y)))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC o_ONTO_11);;
e(EQ_TAC);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
 e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
 e(REWRITE_TAC[I_o_ID]);;

 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
 e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
 e(REWRITE_TAC[I_o_ID]);;
%

let o_ONTO =
 prove_thm
  (`o_ONTO`,
   "!(ss1:**->***) (ss2:*->**).
      ONTO ss1 /\ ONTO ss2 ==>
      ONTO (ss1 o ss2)",
   REWRITE_TAC[ONTO_DEF]
   THEN REPEAT STRIP_TAC
   THEN UNDISCH_ALL_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o SPEC_ALL)
   THEN DISCH_THEN (STRIP_ASSUME_TAC o SPEC "x:**")
   THEN EXISTS_TAC "x':*"
   THEN ASM_REWRITE_TAC[o_THM]
  );;

% o_ONTO
g "!(ss1:**->***) (ss2:*->**).
      ONTO ss1 /\ ONTO ss2 ==>
      ONTO (ss1 o ss2)";;
e(REWRITE_TAC[ONTO_DEF]);;
e(REPEAT STRIP_TAC);;
e(UNDISCH_ALL_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o SPEC_ALL));;
e(DISCH_THEN (STRIP_ASSUME_TAC o SPEC "x:**"));;
e(EXISTS_TAC "x':*");;
e(ASM_REWRITE_TAC[o_THM]);;
%

let o_ONE_ONE =
 prove_thm
  (`o_ONE_ONE`,
   "!(ss1:**->***) (ss2:*->**).
      ONE_ONE ss1 /\ ONE_ONE ss2 ==>
      ONE_ONE (ss1 o ss2)",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[o_THM]
  );;

% o_ONE_ONE
g "!(ss1:**->***) (ss2:*->**).
      ONE_ONE ss1 /\ ONE_ONE ss2 ==>
      ONE_ONE (ss1 o ss2)";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[o_THM]);;
%


let biject_EQ_INVERSE =
 prove_thm
  (`biject_EQ_INVERSE`,
   "!y x (ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      (((@z. ss z = y) = x) = (ss x = y))",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN DISCH_THEN (REWRITE_THM o SYM)
   THENL
     [ IMP_RES_THEN REWRITE_THM ONTO_INVERSE;

       ASM_REWRITE_TAC[SELECT_REFL]
     ]
  );;

% biject_EQ_INVERSE
g "!y x (ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      (((@z. ss z = y) = x) = (ss x = y))";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(ASM_REWRITE_TAC[SELECT_REFL]);;
%

let biject_EQ_INVERSE1 =
 prove_thm
  (`biject_EQ_INVERSE1`,
   "!y x (ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      ((x = (@z. ss z = y)) = (ss x = y))",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [ DISCH_THEN REWRITE_THM
       THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE;

       DISCH_THEN (REWRITE_THM o SYM)
       THEN ASM_REWRITE_TAC[SELECT_REFL]
     ]
  );;

% biject_EQ_INVERSE1
g "!y x (ss:*->**).
      ONE_ONE ss /\ ONTO ss ==>
      ((x = (@z. ss z = y)) = (ss x = y))";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN REWRITE_THM);;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(ASM_REWRITE_TAC[SELECT_REFL]);;
%

let INVERSE_biject =
 prove_thm
  (`INVERSE_biject`,
   "!ss:*->**.
       ONE_ONE ss /\ ONTO ss ==>
       ONE_ONE (\y. @z. ss z = y) /\ ONTO (\y. @z. ss z = y)",
   REPEAT STRIP_TAC
   THENL
     [ REWRITE_TAC[ONE_ONE]
       THEN BETA_TAC
       THEN IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1
       THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE;

       REWRITE_TAC[ONTO_DEF]
       THEN GEN_TAC
       THEN BETA_TAC
       THEN IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1
       THEN EXISTS_TAC "(ss:*->**) y"
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% INVERSE_biject
g "!ss:*->**.
       ONE_ONE ss /\ ONTO ss ==>
       ONE_ONE (\y. @z. ss z = y) /\ ONTO (\y. @z. ss z = y)";;
e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[ONE_ONE]);;
 e(BETA_TAC);;
 e(IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1);;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;

 e(REWRITE_TAC[ONTO_DEF]);;
 e(GEN_TAC);;
 e(BETA_TAC);;
 e(IMP_RES2_THEN REWRITE_THM biject_EQ_INVERSE1);;
 e(EXISTS_TAC "(ss:*->**) y");;
 e(ASM_REWRITE_TAC[]);;
%


let IN_IDENT_subst =
 prove_thm
  (`IN_IDENT_subst`,
   "!s ss (z:*).
      (!x. (x IN s) ==> (ss x = x)) /\ ONE_ONE ss ==>
      ((ss z IN s) = (z IN s))",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THENL
     [  UNDISCH_LAST_TAC
        THEN ASM_REWRITE_TAC[]
        THEN DISCH_THEN REWRITE_ALL_THM
        THEN ASM_REWRITE_TAC[];

        ASM_REWRITE_TAC[]
     ]
  );;

% IN_IDENT_subst
g "!s ss (z:*).
      (!x. (x IN s) ==> (ss x = x)) /\ ONE_ONE ss ==>
      ((ss z IN s) = (z IN s))";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(RES_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_THEN REWRITE_ALL_THM);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%




let MEMBER_IMP_POS_CARD = prove_thm
   (`MEMBER_IMP_POS_CARD`,
    "!(x:*) s. x IN s ==> (0 < CARD s)",
    REPEAT GEN_TAC
    THEN STRIP_TAC
    THEN MATCH_MP_TAC (DISJ_IMP (SPEC_ALL LESS_0_CASES))
    THEN PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]
    THEN PURE_ONCE_REWRITE_TAC[CARD_EQ_0]
    THEN PURE_ONCE_REWRITE_TAC[SYM(SPEC_ALL MEMBER_NOT_EMPTY)]
    THEN EXISTS_TAC "x:*"
    THEN ASM_REWRITE_TAC[]
   );;

%
Theorem MEMBER_IMP_POS_CARD:
g "!(x:*) s. x IN s ==> (0 < CARD s)";;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(MATCH_MP_TAC (DISJ_IMP (SPEC_ALL LESS_0_CASES)));;
e(PURE_ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(PURE_ONCE_REWRITE_TAC[CARD_EQ_0]);;
e(PURE_ONCE_REWRITE_TAC[SYM(SPEC_ALL MEMBER_NOT_EMPTY)]);;
e(EXISTS_TAC "x:*");;
e(ASM_REWRITE_TAC[]);;

or,
e(REPEAT GEN_TAC);;
e(DISCH_THEN ( ACCEPT_TAC
               o (MP (DISJ_IMP (SPEC "CARD (s:(*)set)" LESS_0_CASES)))
               o NOT_EQ_SYM
               o (MP (CONTRAPOS (SPEC "s:(*)set" CARD_EQ_0)))
               o (EQ_MP (SYM (SPEC "s:(*)set" NONEMPTY_MEMBER)))
               o (EXISTS ("?x:*. x IN s","x:*")) ));;
or,
e(STRIP_TAC);;
e(MATCH_MP_TAC (DISJ_IMP (SPEC "CARD (s:(*)set)" LESS_0_CASES)));;
e(PURE_ONCE_REWRITE_TAC [EQ_SYM_EQ]);;
e(MATCH_MP_TAC (CONTRAPOS (SPEC "s:(*)set" CARD_EQ_0)));;
e(REWRITE_TAC [NONEMPTY_MEMBER]);;
e(EXISTS_TAC "x:*");;
e(ASM_REWRITE_TAC[]);;
%


let SUB1_SUC = prove_thm
   (`SUB1_SUC`,
    "!m. (0 < m) ==> (SUC (m - 1) = m)",
    INDUCT_TAC
    THENL [REWRITE_TAC[NOT_LESS_0];
           REWRITE_TAC[LESS_0;SUC_SUB1]
          ]
   );;

%
Theorem SUB1_SUC:
g "!m. (0 < m) ==> (SUC (m - 1) = m)";;
e(INDUCT_TAC);;
e(REWRITE_TAC[NOT_LESS_0]);;

e(REWRITE_TAC[LESS_0;SUC_SUB1]);;
%



let UNION_INTER =
 prove_thm
  (`UNION_INTER`,
   "!s1 s2 s3 :(*)set.
     (s1 UNION s2) INTER s3 =
     ((s1 INTER s3) UNION (s2 INTER s3))",
   ONCE_REWRITE_TAC[INTER_COMM]
   THEN REWRITE_TAC[UNION_OVER_INTER]
  );;

% UNION_INTER
g "!s1 s2 s3 :(*)set.
     (s1 UNION s2) INTER s3 =
     ((s1 INTER s3) UNION (s2 INTER s3))";;
e(ONCE_REWRITE_TAC[INTER_COMM]);;
e(REWRITE_TAC[UNION_OVER_INTER]);;
%


let UNION_THM =
 prove_thm
  (`UNION_THM`,
   "!s t (x:*).
      (EMPTY UNION t = t) /\
      (s UNION EMPTY = s) /\
      ((x INSERT s) UNION t = x INSERT (s UNION t)) /\
      (s UNION (x INSERT t) = x INSERT (s UNION t))",
   REWRITE_TAC[UNION_EMPTY;INSERT_UNION_EQ]
   THEN ONCE_REWRITE_TAC[UNION_COMM]
   THEN REWRITE_TAC[INSERT_UNION_EQ]
  );;

% UNION_THM
g "!s t (x:*).
      (EMPTY UNION t = t) /\
      (s UNION EMPTY = s) /\
      ((x INSERT s) UNION t = x INSERT (s UNION t)) /\
      (s UNION (x INSERT t) = x INSERT (s UNION t))";;
e(REWRITE_TAC[UNION_EMPTY;INSERT_UNION_EQ]);;
e(ONCE_REWRITE_TAC[UNION_COMM]);;
e(REWRITE_TAC[INSERT_UNION_EQ]);;
%



let UNION_SUBSET =
 prove_thm
  (`UNION_SUBSET`,
   "!(s:(*)set) t u.
     ((s UNION t) SUBSET u) = ((s SUBSET u) /\ (t SUBSET u))",
   REWRITE_TAC[SUBSET_DEF;IN_UNION]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% UNION_SUBSET
g "!(s:(*)set) t u.
     ((s UNION t) SUBSET u) = ((s SUBSET u) /\ (t SUBSET u))";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(REWRITE_TAC[IN_UNION]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%



let UNION_DIFF =
 prove_thm
  (`UNION_DIFF`,
   "!s t r:(*)set. (s UNION t) DIFF r = ((s DIFF r) UNION (t DIFF r))",
   REWRITE_TAC[EXTENSION;IN_UNION;IN_DIFF]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< UNION_DIFF
g "!s t r:(*)set. (s UNION t) DIFF r = ((s DIFF r) UNION (t DIFF r))";;
e(REWRITE_TAC[EXTENSION]);;
e(REWRITE_TAC[IN_UNION;IN_DIFF]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(STRIP_TAC THEN ASM_REWRITE_TAC[]);;

  e(STRIP_TAC THEN ASM_REWRITE_TAC[]);;
>%



% Theorems about DISJOINT: %


let DISJOINT_INSERT2 = prove_thm
   (`DISJOINT_INSERT2`,
    "!s t (x:*). DISJOINT s (INSERT x t) = (DISJOINT s t /\ ~(x IN s))",
    ONCE_REWRITE_TAC[DISJOINT_SYM]
    THEN REWRITE_TAC[DISJOINT_INSERT]
   );;

% DISJOINT_INSERT2:
g "!s t (x:*). DISJOINT s (INSERT x t) = (DISJOINT s t /\ ~(x IN s))";;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_INSERT]);;
%

let DISJOINT_UNION2 = prove_thm
   (`DISJOINT_UNION2`,
    "!s t r:(*)set. DISJOINT s (t UNION r) = (DISJOINT s t /\ DISJOINT s r)",
    ONCE_REWRITE_TAC[DISJOINT_SYM]
    THEN REWRITE_TAC[DISJOINT_UNION]
   );;

% DISJOINT_UNION2:
g "!s t r:(*)set. DISJOINT s (t UNION r) = (DISJOINT s t /\ DISJOINT s r)";;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_UNION]);;
%


let DISJOINT_SUBSET =
 prove_thm
  (`DISJOINT_SUBSET`,
   "!s t r:(*)set. s SUBSET t /\ DISJOINT t r ==> DISJOINT s r",
   REWRITE_TAC[SUBSET_DEF;IN_DISJOINT]
   THEN CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV)
   THEN REWRITE_TAC[DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC "x:*")
   THENL
    [ FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)
      THEN ASM_REWRITE_TAC[];

      ASM_REWRITE_TAC[]
    ]
  );;

% DISJOINT_SUBSET
g "!s t r:(*)set. s SUBSET t /\ DISJOINT t r ==> DISJOINT s r";;
e(REWRITE_TAC[SUBSET_DEF;IN_DISJOINT]);;
e(CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV));;
e(REWRITE_TAC[DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC "x:*"));;

 e(FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[]);;
%


let IN = save_thm(`IN`, CONJ NOT_IN_EMPTY IN_INSERT);;

let UNION = save_thm(`UNION`, CONJ UNION_EMPTY INSERT_UNION_EQ);;

let INTER = save_thm(`INTER`, CONJ INTER_EMPTY INSERT_INTER);;

let DELETE = save_thm(`DELETE`, CONJ EMPTY_DELETE DELETE_INSERT);;

let SUBSET = save_thm(`SUBSET`, CONJ EMPTY_SUBSET INSERT_SUBSET);;

let IMAGE = save_thm(`IMAGE`, CONJ IMAGE_EMPTY IMAGE_INSERT);;

let DIFF = save_thm(`DIFF`, CONJ DIFF_EMPTY DIFF_INSERT);;


% ======================================================================== %
% UNION_SET is a function which takes a set of sets and produces the union %
% of all of them, that is, if S = {Si} then UNION_SET S = S1 U ... U Sn.   %
% ======================================================================== %


let UNION_SET_EXISTS =
    TAC_PROOF
    (([], "!s:((*)set)set. ?t. !x. x IN t = ?si. si IN s /\ x IN si"),
     SET_INDUCT_TAC THENL
     [EXISTS_TAC "EMPTY:(*)set" THEN REWRITE_TAC [IN];
      FIRST_ASSUM (CHOOSE_THEN STRIP_ASSUME_TAC) THEN
      EXISTS_TAC "(e:(*)set) UNION t" THEN
      ASM_REWRITE_TAC [IN;IN_UNION] THEN GEN_TAC THEN
      EQ_TAC THEN ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
      REPEAT STRIP_TAC THENL
      [EXISTS_TAC "e :(*)set" THEN ASM_REWRITE_TAC [];
       EXISTS_TAC "si:(*)set" THEN ASM_REWRITE_TAC [];
       ASM_REWRITE_TAC [];
       DISJ2_TAC THEN EXISTS_TAC "si:(*)set" THEN ASM_REWRITE_TAC []]]);;

%
g("!s:((*)set)set. ?t. !x. x IN t = ?si. si IN s /\ x IN si");;

e(SET_INDUCT_TAC);;

 e(EXISTS_TAC "EMPTY:(*)set");;
 e(REWRITE_TAC [IN]);;

 e(FIRST_ASSUM (CHOOSE_THEN STRIP_ASSUME_TAC));;
 e(EXISTS_TAC "(e:(*)set) UNION t");;
 e(ASM_REWRITE_TAC [IN;IN_UNION]);;
 e(GEN_TAC);;
 e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;

   e(EXISTS_TAC "e :(*)set" THEN ASM_REWRITE_TAC []);;
   e(EXISTS_TAC "si:(*)set" THEN ASM_REWRITE_TAC []);;

  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(REPEAT STRIP_TAC);;

   e(ASM_REWRITE_TAC []);;

   e(DISJ2_TAC);;
   e(EXISTS_TAC "si:(*)set");;
   e(ASM_REWRITE_TAC []);;
%


let IN_UNION_SET =
    let th1 = CONV_RULE SKOLEM_CONV UNION_SET_EXISTS in
    new_specification `IN_UNION_SET` [`constant`,`UNION_SET`] th1;;


let UNION_SET_EMPTY_LEMMA = TAC_PROOF
   (([], "UNION_SET EMPTY = EMPTY :(*)set"),
    REWRITE_TAC [EXTENSION; IN_UNION_SET; IN]
   );;

let UNION_SET_INSERT_LEMMA = TAC_PROOF
    (([],  "! (si:(*)set) (s:((*)set)set) .
            UNION_SET (INSERT si s) = si UNION (UNION_SET s)"),
    REWRITE_TAC [EXTENSION; IN_UNION_SET; IN_UNION; IN]
    THEN REPEAT GEN_TAC
    THEN EQ_TAC
    THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
    THEN REPEAT STRIP_TAC
    THENL [ ALL_TAC;

            DISJ2_TAC
            THEN EXISTS_TAC "si':(*)set";

            EXISTS_TAC "si:(*)set";

            EXISTS_TAC "si':(*)set"
          ]
    THEN ASM_REWRITE_TAC []
   );;

%
g "! (si:(*)set) (s:((*)set)set) .
            UNION_SET (INSERT si s) = si UNION (UNION_SET s)";;
e(REWRITE_TAC [EXTENSION; IN_UNION_SET; IN_UNION; IN]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(REPEAT STRIP_TAC);;

  e(ASM_REWRITE_TAC []);;

  e(DISJ2_TAC);;
  e(EXISTS_TAC "si':(*)set");;
  e(ASM_REWRITE_TAC []);;

 e(REPEAT STRIP_TAC);;

  e(EXISTS_TAC "si:(*)set");;
  e(ASM_REWRITE_TAC []);;

  e(EXISTS_TAC "si':(*)set");;
  e(ASM_REWRITE_TAC []);;
%


let UNION_SET = save_thm (`UNION_SET`,
                          CONJ UNION_SET_EMPTY_LEMMA UNION_SET_INSERT_LEMMA);;



let UNION_SET_UNION =
 prove_thm
  (`UNION_SET_UNION`,
   "!s t : ((*)set)set.
      UNION_SET (s UNION t) =
                     UNION_SET s UNION UNION_SET t",
   REWRITE_TAC[EXTENSION;IN_UNION;IN_UNION_SET]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL [ DISJ1_TAC; DISJ2_TAC; ALL_TAC; ALL_TAC ]
   THEN EXISTS_TAC "si:(*)set"
   THEN ASM_REWRITE_TAC[]
  );;

% UNION_SET_UNION:
g "!s t : ((*)set)set.
      UNION_SET (s UNION t) =
                     UNION_SET s UNION UNION_SET t";;
e(REWRITE_TAC[EXTENSION]);;
e(REWRITE_TAC[IN_UNION;IN_UNION_SET]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

   e(DISJ1_TAC);;
   e(EXISTS_TAC "si:(*)set");;
   e(ASM_REWRITE_TAC[]);;

   e(DISJ2_TAC);;
   e(EXISTS_TAC "si:(*)set");;
   e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;

   e(EXISTS_TAC "si:(*)set");;
   e(ASM_REWRITE_TAC[]);;

   e(EXISTS_TAC "si:(*)set");;
   e(ASM_REWRITE_TAC[]);;
%



close_theory();;

