% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the semantics of the programming
  language.  These definitions are relations, whose definitions are
  expressed as sets of rules to inductively define the tuples of the
  relations, using Tom Melham's inductive rule definition library.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `prog_semantics`;;

new_parent `prog_substitutions`;;


% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %
load_library `ind_defs`;;


loadf `semantics_tactics`;;
loadf `new_ind_def`;;



% ===================================================================== %
% STRUCTURAL OPERATIONAL SEMANTICS                                      %
% ===================================================================== %


% ============================================================== %
% Next we define the semantics of expressions and commands.      %
% This is a structural operational semantics definition,         %
% using Tom Melham's rule induction tools (extended by PVH).     %
% ============================================================== %


let E_rules,E_ind,E_cases,E_str_induct =
    let E = "E : exp -> state -> num -> state -> bool" in
    new_ind_def `E`

    ("^E e s1 n s2", []) 
    
     [ [				             
       % -------------------------------------------- % ],
                       "^E (NUM n) s n s"             ;


       [				             
       % -------------------------------------------- % ],
                    "^E (PVAR x) s (s x) s"           ;


       [             "^E (PVAR x) s1 n s2"
       % -------------------------------------------- % ],
            "^E (INC x) s1 (n+1) (BND (n+1) x s2)"    ;


       [     "^E e1 s1 n s2";     "^E e2 s2 m s3"
       % -------------------------------------------- % ],
                "^E (PLUS e1 e2) s1 (n+m) s3"         ;


       [     "^E e1 s1 n s2";     "^E e2 s2 m s3"
       % -------------------------------------------- % ],
                "^E (MINUS e1 e2) s1 (n-m) s3"        ;


       [     "^E e1 s1 n s2";     "^E e2 s2 m s3"
       % -------------------------------------------- % ],
                "^E (MULT e1 e2) s1 (n*m) s3"         ];;


let ES_DEF =
    new_list_rec_definition
      (`ES_DEF`,
      "(!s1 vs s2.       ES NIL s1 vs s2 =
           ((vs = NIL) /\ (s2 = s1))) /\
       (!e es s1 vs s3.  ES (CONS e es) s1 vs s3 =
           (~(NULL vs) /\ ?s2. E e s1 (HD vs) s2 /\ ES es s2 (TL vs) s3))");;


let B_rules,B_ind,B_cases,B_str_induct =
    let B = "B : bexp -> state -> bool -> state -> bool" in
    new_ind_def `B`

    ("^B b s1 t s2", []) 
    
     [ [     "E e1 s1 n s2";      "E e2 s2 m s3"    
       % -------------------------------------------- % ],
              "^B (EQ e1 e2) s1 ((n:num)=m) s3"       ;


       [     "E e1 s1 n s2";      "E e2 s2 m s3"    
       % -------------------------------------------- % ],
                "^B (LESS e1 e2) s1 (n<m) s3"         ;


       [    "ES es1 s1 ns s2";   "ES es2 s2 ms s3"    
       % -------------------------------------------- % ],
             "^B (LLESS es1 es2) s1 (ns<<ms) s3"      ;


       [    "^B b1 s1 t1 s2";     "^B b2 s2 t2 s3"    
       % -------------------------------------------- % ],
              "^B (AND b1 b2) s1 (t1 /\ t2) s3"       ;


       [    "^B b1 s1 t1 s2";     "^B b2 s2 t2 s3"    
       % -------------------------------------------- % ],
               "^B (OR b1 b2) s1 (t1 \/ t2) s3"       ;


       [                 "^B b s1 t s2"    
       % -------------------------------------------- % ],
                    "^B (NOT b) s1 (~t) s2"           ];;


let C_rules,C_ind,C_cases,C_str_induct =
    let C = "C : cmd -> env -> state -> state -> bool" in
    new_ind_def `C`

    ("^C c r s1 s2", []) 
    
     [ [                                             
       % -------------------------------------------- % ],
                       "^C (SKIP) r s s"              ;


% Note: no rule for ABORT, since there are no elements in that relation. %

    
       [                 "E e s1 n s2"
       % -------------------------------------------- % ],
              "^C (ASSIGN x e) r s1 (BND n x s2)"     ;


       [      "^C c1 r s1 s2";    "^C c2 r s2 s3"    
       % -------------------------------------------- % ],
                    "^C (SEQ c1 c2) r s1 s3"          ;


       [                "^C c1 r s2 s3";
       % -------------------------------------------- % "B b s1 T s2" ],
                   "^C (IF b c1 c2) r s1 s3"          ;


       [                "^C c2 r s2 s3";
       % -------------------------------------------- % "B b s1 F s2" ],
                   "^C (IF b c1 c2) r s1 s3"          ;


       [ "^C c r s2 s3";  "^C (WHILE a pr b c) r s3 s4";
       % -------------------------------------------- % "B b s1 T s2" ],
                  "^C (WHILE a pr b c) r s1 s4"       ;


       [                                             
       % -------------------------------------------- % "B b s1 F s2" ],
                  "^C (WHILE a pr b c) r s1 s2"       ;


       ["^C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
            r (BNDsL ns vals' s2) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
        "vals' = variants vals (SL (APPEND xs glbs))";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^C (CALL p xs es) r s1 (BNDsL (MAP s2 vals') vals' s3)" ];;

%<
There are different possible alternatives here for the CALL rule.

       ["^C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
            r (BNDsL ns vals' s2) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
        "vals' = variants vals (SL (APPEND xs glbs))";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^C (CALL p xs es) r s1 (BNDsL (MAP s2 vals') vals' s3)" ];;

       ["^C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
            r (BNDsL ns vals' s2) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
        "vals' = variants vals (SL (APPEND xs glbs))";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^C (CALL p xs es) r s1
           (BNDsL (MAP s3 (APPEND xs glbs)) (APPEND xs glbs) s2)" ];;

The above two are call-by-name, which since we are dealing with only
simple variables, is equivalent to call-by-reference.

An alternative mechanism for passing parameters is call-by-value-return,
as in the next two possible versions:

       ["^C c r (BNDsL ns vals (BNDsL (MAP s2 xs) vars s2)) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^C (CALL p xs es) r s1
           (BNDsL (MAP s3 (APPEND vars glbs)) (APPEND xs glbs) s2)" ];;


       ["^C c r (BNDsL ns vals (BNDsL (MAP s2 xs) vars s2)) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^C (CALL p xs es) r s1
           (BNDsL (MAP s3 vars) xs
             (BNDsL (MAP s2 (APPEND vars vals)) (APPEND vars vals) s3))" ];;


>%

let D_rules,D_ind,D_cases,D_str_induct =
    let D = "D : decl -> env -> env -> bool" in
    new_ind_def `D`

    ("^D d r1 r2", []) 
    
     [ [				             
       % -------------------------------------------- % ],
         "^D (PROC p vars vals glbs pre post calls rec c) r
               (BND (vars,vals,glbs,pre,post,calls,rec,c) p r)"  ;


       [       "^D d1 r1 r2";     "^D d2 r2 r3"
       % -------------------------------------------- % ],
                    "^D (DSEQ d1 d2) r1 r3"           ;


       [                                             
       % -------------------------------------------- % ],
                       "^D (DEMPTY) r r"              ];;


let EMPTY_ENV =
    new_def `EMPTY_ENV`
    "r0 = (\p. [], [], [], AFALSE, ATRUE, (\p. AFALSE), AFALSE, ABORT) : env";;

let INITIAL_STATE =
    new_def `INITIAL_STATE`
    "s0 = (\x. 0) : state";;

let P_rules,P_ind,P_cases,P_str_induct =
    let P = "P : prog -> state -> bool" in
    new_ind_def `P`

    ("^P p s", []) 
    
     [ [       "D d r0 r1";      "C c r1 s0 s1"
       % -------------------------------------------- % ],
                      "^P (PROG d c) s1"              ];;



% --------------------------------------------------------------------- %
% Tactics for the rules.						%
% --------------------------------------------------------------------- %

let EXP_TAC = FIRST (map RULE_TAC E_rules);;
let [NUM_TAC;PVAR_TAC;INC_TAC;
     PLUS_TAC;MINUS_TAC;MULT_TAC] = map RULE_TAC E_rules;;

let EXPS_TAC = LIST_INDUCT_TAC;;

let BEXP_TAC = FIRST (map RULE_TAC B_rules);;
let [BEQ_TAC;LESS_TAC;LLESS_TAC;AND_TAC;OR_TAC;NOT_TAC] = map RULE_TAC B_rules;;

let CMD_TAC = FIRST (map RULE_TAC C_rules);;
let [SKIP_TAC;ASSIGN_TAC;SEQ_TAC;IF_T_TAC;IF_F_TAC;
     WHILE_T_TAC;WHILE_F_TAC;CALL_TAC] = map RULE_TAC C_rules;;

let DECL_TAC = FIRST (map RULE_TAC D_rules);;
let [PROC_TAC;DSEQ_TAC;DEMPTY_TAC] = map RULE_TAC D_rules;;

let [PROG_TAC] = map RULE_TAC P_rules;;

% --------------------------------------------------------------------- %
% Rule induction tactics.                                               %
% --------------------------------------------------------------------- %

let  EXP_INDUCT_TAC = RULE_INDUCT_THEN E_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let BEXP_INDUCT_TAC = RULE_INDUCT_THEN B_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let  CMD_INDUCT_TAC = RULE_INDUCT_THEN C_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let DECL_INDUCT_TAC = RULE_INDUCT_THEN D_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let PROG_INDUCT_TAC = RULE_INDUCT_THEN P_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;



let decl_Axiom  = theorem `prog_syntax` `decl_Axiom`;;
let decl_induct = theorem `prog_syntax` `decl_induct`;;

let mkenv =
    new_rec_def `mkenv` decl_Axiom
      "(mkenv (PROC p vars vals glbs pre post calls rec c) r =
                 BND (vars,vals,glbs,pre,post,calls,rec,c) p r)  /\
       (mkenv (DSEQ d1 d2) r =
                 mkenv d2 (mkenv d1 r))  /\
       (mkenv (DEMPTY) r =
                 r)";;


let D_mkenv =
 prove_thm
  (`D_mkenv`,
   "!d r. D d r (mkenv d r)",
   INDUCT_THEN decl_induct ASSUME_TAC
   THEN REWRITE_TAC[mkenv]
   THEN REPEAT GEN_TAC
   THENL
     [  PROC_TAC;

        DSEQ_TAC
        THEN EXISTS_TAC "mkenv d r"
        THEN ASM_REWRITE_TAC[];

        DEMPTY_TAC
     ]
  );;

%
g "!d r. D d r (mkenv d r)";;
e(INDUCT_THEN decl_induct ASSUME_TAC);;

 e(REWRITE_TAC[mkenv]);;
 e(REPEAT GEN_TAC);;
 e(PROC_TAC);;

 e(REWRITE_TAC[mkenv]);;
 e(REPEAT GEN_TAC);;
 e(DSEQ_TAC);;
 e(EXISTS_TAC "mkenv d r");;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[mkenv]);;
 e(REPEAT GEN_TAC);;
 e(DEMPTY_TAC);;
%



let D_IMP_mkenv =
 prove_thm
  (`D_IMP_mkenv`,
   "!d r1 r2. D d r1 r2 ==> (r2 = mkenv d r1)",
   DECL_INDUCT_TAC
   THEN ASM_REWRITE_TAC[mkenv]
  );;

% D_IMP_mkenv
g "!d r1 r2. D d r1 r2 ==> (r2 = mkenv d r1)";;
e(DECL_INDUCT_TAC);;

 e(REWRITE_TAC[mkenv]);;

 e(REWRITE_TAC[mkenv]);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[mkenv]);;
%


let D_EQ_mkenv =
 prove_thm
  (`D_EQ_mkenv`,
   "!d r1 r2. D d r1 r2 = (r2 = mkenv d r1)",
   REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [  REWRITE_TAC[D_IMP_mkenv];

        STRIP_TAC
        THEN ASM_REWRITE_TAC[D_mkenv]
     ]
  );;

% D_EQ_mkenv
g "!d r1 r2. D d r1 r2 = (r2 = mkenv d r1)";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(REWRITE_TAC[D_IMP_mkenv]);;

 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[D_mkenv]);;
%





close_theory();;

