% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains definitions of functions to substitute variables
  for variables within programming-language expressions.  This is both
  for one particular variable for another, and for a list of such
  variable-for-variable substitutions.  Using the same substitutions
  as defined for variable-for-variable substitutions, operators are
  defined to apply such substitutions to programming-language expressions.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `prog_substitutions`;;

new_parent `prog_syntax`;;
new_parent `var_substitutions`;;



% Set up autoloading from ancestor theories of prog_substitutions: %

loadf `autoload`;;

autoload_all_defs_and_thms
 `finite_sets more_finite_sets bindings variants
  assert_syntax assert_semantics prog_syntax
  substitutions var_substitutions`;;

loadf `semantics_tactics`;;



% ===================================================================== %
% New Type Definitions and Abbreviations                                %
% ===================================================================== %
% ENVIRONMENTS:                                                         %
%   We define environments as mappings from identifiers to              %
%   procedure definitions, that is, to tuples of the form               %
%   < <list of formal variable parameters>,                             %
%     <list of formal value parameters>,                                %
%     <list of possibly read/written global variables ("globals" list)> %
%     <precondition>,                                                   %
%     <postcondition (possibly involving logical ("pre") variables )>,  %
%     <progress environment: function yielding progress expressions     %
%         for each procedure called from the body>                      %
%     <recursion progress condition: progress achieved between          %
%         recursive calls>                                              %
%     <the command which is the body of the procedure>                  %
%   >                                                                   %
% ===================================================================== %

new_type_abbrev(`env`,
   ":string -> (vars # vars # vars # aexp # aexp # prog_env # aexp # cmd)");;



% ============================================================ %
% We now define the application of a variable-substitution     %
% for programming language variable lists, expressions,        %
% expression lists, boolean expressions, and commands.         %
% ============================================================ %

let xs_subst = 
    new_infix_list_rec_definition (`xs_subst`,
      "($<|xs (NIL) (ss:vsubst) = NIL)  /\
       ($<|xs (CONS x xs) ss = CONS (ss x) ($<|xs xs ss))");;

let xs_subst_MAP =
 prove_thm
  (`xs_subst_MAP`,
   "!xs ss. xs <|xs ss = MAP ss xs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[xs_subst;MAP]
  );;

let LENGTH_xs_subst =
 prove_thm
  (`LENGTH_xs_subst`,
   "!xs ss. LENGTH (xs <|xs ss) = LENGTH xs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[xs_subst;LENGTH]
  );;

let e_subst = 
    new_recursive_definition true exp_Axiom `e_subst`
      "($<|e (NUM n) ss       = NUM n)           /\
       ($<|e (PVAR x) ss      = PVAR (ss x))     /\
       ($<|e (INC x) ss       = INC (ss x))      /\
       ($<|e (PLUS e1 e2) ss  = PLUS  ($<|e e1 ss) ($<|e e2 ss))  /\
       ($<|e (MINUS e1 e2) ss = MINUS ($<|e e1 ss) ($<|e e2 ss))  /\
       ($<|e (MULT e1 e2) ss  = MULT  ($<|e e1 ss) ($<|e e2 ss))";;

let es_subst = 
    new_infix_list_rec_definition (`es_subst`,
      "($<|es (NIL) ss = NIL)  /\
       ($<|es (CONS e es) ss = CONS (e <|e ss) ($<|es es ss))");;

let LENGTH_es_subst =
 prove_thm
  (`LENGTH_es_subst`,
   "!es ss. LENGTH (es <|es ss) = LENGTH es",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[es_subst;LENGTH]
  );;

let b_subst = 
    new_recursive_definition true bexp_Axiom `b_subst`
      "($<|b (EQ  e1 e2) ss     = EQ    (e1 <|e ss) (e2 <|e ss))      /\
       ($<|b (LESS e1 e2) ss    = LESS  (e1 <|e ss) (e2 <|e ss))      /\
       ($<|b (LLESS es1 es2) ss = LLESS (es1 <|es ss) (es2 <|es ss))  /\
       ($<|b (AND b1 b2) ss     = AND   ($<|b b1 ss) ($<|b b2 ss))    /\
       ($<|b (OR  b1 b2) ss     = OR    ($<|b b1 ss) ($<|b b2 ss))    /\
       ($<|b (NOT b) ss         = NOT   ($<|b b ss))";;

let c_subst = 
    new_recursive_definition true cmd_Axiom `c_subst`
      "($<|c (SKIP) ss = SKIP)  /\
       ($<|c (ABORT) ss = ABORT)  /\
       ($<|c (ASSIGN x e) ss = ASSIGN (ss x) (e <|e ss))  /\
       ($<|c (SEQ c1 c2) ss = SEQ ($<|c c1 ss) ($<|c c2 ss))  /\
       ($<|c (IF b c1 c2) ss = IF (b <|b ss) ($<|c c1 ss) ($<|c c2 ss))  /\
       ($<|c (WHILE a pr b c) ss =
              WHILE (a <|av ss) (pr <|av ss) (b <|b ss) ($<|c c ss))  /\
       ($<|c (CALL p xs es) ss = CALL p (xs <|xs ss) (es <|es ss) )";;

let env_subst =
    new_infix_definition (`env_subst`,
      "$<|env (d:env) ss =
       \p. let (vars,vals,glbs,pre,post,calls,rec,c) = d p in
             (MAP ss vars,
              MAP ss vals,
              MAP ss glbs,
              pre  <|av ss,
              post <|av ss,
              (\p. (calls p) <|av ss),
              rec <|av ss,
              c <|c ss
             ) ");;

let env_subst_THM =
 prove_thm
  (`env_subst_THM`,
   "!d p vars vals glbs pre post calls rec c ss.
     (d p = (vars, vals, glbs, pre, post, calls, rec, c)) ==>
     ((d <|env ss) p = (MAP ss vars, MAP ss vals, MAP ss glbs,
                        pre <|av ss, post <|av ss,
                        (\p. (calls p) <|av ss), rec <|av ss,
                        c <|c ss))",
   REWRITE_TAC[env_subst]
   THEN REPEAT STRIP_TAC
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REFL_TAC
  );;

% env_subst_THM
g "!d p vars vals glbs pre post calls rec c ss.
     (d p = (vars, vals, glbs, pre, post, calls, rec, c)) ==>
     ((d <|env ss) p = (MAP ss vars, MAP ss vals, MAP ss glbs,
                        pre <|av ss, post <|av ss,
                        (\p. (calls p) <|av ss), rec <|av ss,
                        c <|c ss))";;
e(REWRITE_TAC[env_subst]);;
e(REPEAT STRIP_TAC);;
e(BETA_TAC);;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REFL_TAC);;
%


let g_subst =
    new_infix_definition (`g_subst`,
      "$<|g (g:prog_env) ss =
       (\p. (g p) <|av ss)");;



let xs_subst_o =
 prove_thm
  (`xs_subst_o`,
   "!xs ss1 ss2. (xs <|xs (ss2 o ss1)) = ((xs <|xs ss1) <|xs ss2)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[xs_subst;o_THM]
  );;

% xs_subst_o
g "!xs ss1 ss2. (xs <|xs (ss2 o ss1)) = ((xs <|xs ss1) <|xs ss2)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst]);;

 e(ASM_REWRITE_TAC[xs_subst;o_THM]);;
%

let e_subst_o =
 prove_thm
  (`e_subst_o`,
   "!e ss1 ss2. (e <|e (ss2 o ss1)) = ((e <|e ss1) <|e ss2)",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[e_subst;o_THM]
  );;

% e_subst_o
g "!e ss1 ss2. (e <|e (ss2 o ss1)) = ((e <|e ss1) <|e ss2)";;
e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC);;

 e(REWRITE_TAC[e_subst]);;

 e(REWRITE_TAC[e_subst;o_THM]);;

 e(REWRITE_TAC[e_subst;o_THM]);;

 e(ASM_REWRITE_TAC[e_subst;o_THM]);;

 e(ASM_REWRITE_TAC[e_subst;o_THM]);;
%

let es_subst_o =
 prove_thm
  (`es_subst_o`,
   "!es ss1 ss2. (es <|es (ss2 o ss1)) = ((es <|es ss1) <|es ss2)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[es_subst;e_subst_o]
  );;

% es_subst_o
g "!es ss1 ss2. (es <|es (ss2 o ss1)) = ((es <|es ss1) <|es ss2)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[es_subst]);;

 e(ASM_REWRITE_TAC[es_subst;e_subst_o]);;
%

let b_subst_o =
 prove_thm
  (`b_subst_o`,
   "!b ss1 ss2. (b <|b (ss2 o ss1)) = ((b <|b ss1) <|b ss2)",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[b_subst;e_subst_o;es_subst_o]
  );;

let c_subst_o =
 prove_thm
  (`c_subst_o`,
   "!c ss1 ss2. (c <|c (ss2 o ss1)) = ((c <|c ss1) <|c ss2)",
   INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[c_subst;o_THM;a_vsubst_o;xs_subst_o;
                        e_subst_o;es_subst_o;b_subst_o]
  );;

%
g "!c ss1 ss2. (c <|c (ss2 o ss1)) = ((c <|c ss1) <|c ss2)";;
e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[c_subst;o_THM;a_vsubst_o;xs_subst_o;
                        e_subst_o;es_subst_o;b_subst_o]);;
%


let ENV_TUPLE =
 prove_thm
  (`ENV_TUPLE`,
   "!x : * # ** # *** # **** # ***** # ****** # ******* # ********.
    x =  FST x,
        (FST o SND) x,
        (FST o SND o SND) x,
        (FST o SND o SND o SND) x,
        (FST o SND o SND o SND o SND) x,
        (FST o SND o SND o SND o SND o SND) x,
        (FST o SND o SND o SND o SND o SND o SND) x,
        (SND o SND o SND o SND o SND o SND o SND) x",
   REWRITE_TAC[o_THM]
  );;


% ENV_TUPLE:
g "!x : * # ** # *** # **** # ***** # ****** # ******* # ********.
    x =  FST x,
        (FST o SND) x,
        (FST o SND o SND) x,
        (FST o SND o SND o SND) x,
        (FST o SND o SND o SND o SND) x,
        (FST o SND o SND o SND o SND o SND) x,
        (FST o SND o SND o SND o SND o SND o SND) x,
        (SND o SND o SND o SND o SND o SND o SND) x";;
e(REWRITE_TAC[o_THM]);;
%




let env_subst_o =
 prove_thm
  (`env_subst_o`,
   "!r ss1 ss2. (r <|env (ss2 o ss1)) = ((r <|env ss1) <|env ss2)",
   REPEAT STRIP_TAC
   THEN EXT_TAC "p:string"
   THEN GEN_TAC
   THEN REWRITE_TAC[env_subst]
   THEN BETA_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN BETA_TAC
   THEN REWRITE_TAC[o_THM;MAP_o;a_vsubst_o;c_subst_o]
  );;

% env_subst_o
g "!r ss1 ss2. (r <|env (ss2 o ss1)) = ((r <|env ss1) <|env ss2)";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "p:string");;
e(GEN_TAC);;
e(REWRITE_TAC[env_subst]);;
e(BETA_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(BETA_TAC);;
e(REWRITE_TAC[o_THM;MAP_o;a_vsubst_o;c_subst_o]);;
%


let g_subst_o =
 prove_thm
  (`g_subst_o`,
   "!g ss1 ss2. (g <|g (ss2 o ss1)) = ((g <|g ss1) <|g ss2)",
   REPEAT GEN_TAC
   THEN EXT_TAC "p:string"
   THEN GEN_TAC
   THEN REWRITE_TAC[g_subst]
   THEN BETA_TAC
   THEN REWRITE_TAC[a_vsubst_o]
  );;

% g_subst_o
g "!g ss1 ss2. (g <|g (ss2 o ss1)) = ((g <|g ss1) <|g ss2)";;
e(REPEAT GEN_TAC);;
e(EXT_TAC "p:string");;
e(GEN_TAC);;
e(REWRITE_TAC[g_subst]);;
e(BETA_TAC);;
e(REWRITE_TAC[a_vsubst_o]);;
%


let xs_subst_I =
 prove_thm
  (`xs_subst_I`,
   "!xs. (xs <|xs I) = xs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[xs_subst;I_THM]
  );;

let e_subst_I =
 prove_thm
  (`e_subst_I`,
   "!e. (e <|e I) = e",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[e_subst;I_THM]
  );;

% e_subst_I
g "!e. (e <|e I) = e";;
e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC
  THEN ASM_REWRITE_TAC[e_subst;I_THM]);;
%

let es_subst_I =
 prove_thm
  (`es_subst_I`,
   "!es. (es <|es I) = es",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[es_subst;I_THM;e_subst_I]
  );;

let b_subst_I =
 prove_thm
  (`b_subst_I`,
   "!b. (b <|b I) = b",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[b_subst;e_subst_I;es_subst_I]
  );;

let c_subst_I =
 prove_thm
  (`c_subst_I`,
   "!c. (c <|c I) = c",
   INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[c_subst;I_THM;a_vsubst_I;
                        xs_subst_I;e_subst_I;es_subst_I;b_subst_I]
  );;

let env_subst_I =
 prove_thm
  (`env_subst_I`,
   "!r. (r <|env I) = r",
   GEN_TAC
   THEN EXT_TAC "p:string"
   THEN GEN_TAC
   THEN REWRITE_TAC[env_subst]
   THEN BETA_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN BETA_TAC
   THEN REWRITE_TAC[I_THM;MAP_I;ETA_AX;a_vsubst_I;c_subst_I]
  );;

% env_subst_I
g "!r. (r <|env I) = r";;
e(GEN_TAC);;
e(EXT_TAC "p:string");;
e(GEN_TAC);;
e(REWRITE_TAC[env_subst]);;
e(BETA_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(BETA_TAC);;
e(REWRITE_TAC[I_THM;MAP_I;ETA_AX;a_vsubst_I;c_subst_I]);;
%

let g_subst_I =
 prove_thm
  (`g_subst_I`,
   "!g. (g <|g I) = g",
   GEN_TAC
   THEN EXT_TAC "p:string"
   THEN GEN_TAC
   THEN REWRITE_TAC[g_subst]
   THEN BETA_TAC
   THEN REWRITE_TAC[a_vsubst_I]
  );;

% g_subst_I
g "!g. (g <|g I) = g";;
e(GEN_TAC);;
e(EXT_TAC "p:string");;
e(GEN_TAC);;
e(REWRITE_TAC[g_subst]);;
e(BETA_TAC);;
e(REWRITE_TAC[a_vsubst_I]);;
%

let env_subst_EQ_INVERSE =
 prove_thm
  (`env_subst_EQ_INVERSE`,
   "!d1 d2 ys xs.
     (d1 = d2 <|env (ys //v xs)) =
     (d2 = d1 <|env (\y. @z. (ys //v xs)z = y))",
   REPEAT GEN_TAC
   THEN EQ_TAC
   THEN DISCH_THEN REWRITE_THM
   THEN REWRITE_TAC[(SYM o SPEC_ALL) env_subst_o;vsubst_INVERSE;env_subst_I]
  );;

% env_subst_EQ_INVERSE
g "!d1 d2 ys xs.
     (d1 = d2 <|env (ys //v xs)) =
     (d2 = d1 <|env (\y. @z. (ys //v xs)z = y))";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) env_subst_o;vsubst_INVERSE;env_subst_I]);;

 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) env_subst_o;vsubst_INVERSE;env_subst_I]);;
%



let AP_xs_subst_IN =
 prove_thm
  (`AP_xs_subst_IN`,
   "!xs ss x.
      (x IN SL xs) ==> (ss x IN SL(xs <|xs ss))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% AP_xs_subst_IN
g "!xs ss x.
      (x IN SL xs) ==> (ss x IN SL(xs <|xs ss))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;

  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let IN_AP_xs_subst =
 prove_thm
  (`IN_AP_xs_subst`,
   "!xs ss x.
      ONE_ONE ss /\ (ss x IN SL(xs <|xs ss)) ==> (x IN SL xs)",
   REWRITE_TAC[ONE_ONE_DEF]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% IN_AP_xs_subst
g "!xs ss x.
      ONE_ONE ss /\ (ss x IN SL(xs <|xs ss)) ==> (x IN SL xs)";;
e(REWRITE_TAC[ONE_ONE_DEF]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let IN_xs_subst =
 prove_thm
  (`IN_xs_subst`,
   "!xs ss x.
      ONE_ONE ss ==> ((ss x IN SL(xs <|xs ss)) = (x IN SL xs))",
   REWRITE_TAC[ONE_ONE]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% IN_xs_subst
g "!xs ss x.
      ONE_ONE ss ==> ((ss x IN SL(xs <|xs ss)) = (x IN SL xs))";;
e(REWRITE_TAC[ONE_ONE]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;

 e(REWRITE_TAC[xs_subst;SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let DL_APPEND_xs_subst =
 prove_thm
  (`DL_APPEND_xs_subst`,
   "!xs ys ss.
      (!x. x IN (SL ys) ==> (ss x = x)) /\ ONE_ONE ss ==>
      (DL(APPEND(xs <|xs ss) ys) = DL(APPEND xs ys))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;APPEND]
   THEN REWRITE_TAC[DISJOINT_LIST]
   THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN POP_TAC THEN POP_TAC
   THEN IMP_RES_TAC IN_xs_subst
   THEN POP_TAC THEN POP_TAC
   THEN IMP_RES_TAC IN_IDENT_subst
   THEN POP_TAC THEN POP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% DL_APPEND_xs_subst
g "!xs ys ss.
      (!x. x IN (SL ys) ==> (ss x = x)) /\ ONE_ONE ss ==>
      (DL(APPEND(xs <|xs ss) ys) = DL(APPEND xs ys))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst;APPEND]);;

 e(REWRITE_TAC[xs_subst;APPEND]);;
 e(REWRITE_TAC[DISJOINT_LIST]);;
 e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(POP_TAC THEN POP_TAC);;
 e(IMP_RES_TAC IN_xs_subst);;
 e(POP_TAC THEN POP_TAC);;
 e(IMP_RES_TAC IN_IDENT_subst);;
 e(POP_TAC THEN POP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let DISJOINT_xs_subst =
 prove_thm
  (`DISJOINT_xs_subst`,
   "!x y ss.
     ONE_ONE ss ==>
     (DISJOINT (SL (x <|xs ss)) (SL (y <|xs ss)) =
      DISJOINT (SL x) (SL y)) ",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;SET_LIST]
   THEN REWRITE_TAC[DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN REPEAT STRIP_TAC
   THEN RES_THEN REWRITE_THM
   THEN IMP_RES_THEN REWRITE_THM IN_xs_subst
  );;

% DISJOINT_xs_subst
g "!x y ss.
     ONE_ONE ss ==>
     (DISJOINT (SL (x <|xs ss)) (SL (y <|xs ss)) =
      DISJOINT (SL x) (SL y)) ";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[xs_subst;SET_LIST;DISJOINT_EMPTY]);;

 e(REWRITE_TAC[xs_subst;SET_LIST;DISJOINT_INSERT]);;
 e(REPEAT STRIP_TAC);;
 e(RES_THEN REWRITE_THM);;
 e(IMP_RES_THEN REWRITE_THM IN_xs_subst);;
%



close_theory();;

