% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the semantic progress relations
  C_calls and M_calls.  It also contains the definition of the partial
  and total correctness specifications, as well as procedure entrance
  specifications, and the proofs of theorems about these.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `progress`;;

new_parent `translations`;;


% --------------------------------------------------------------------- %
% Need the string, finite_sets, and ind_defs libraries.                 %
% --------------------------------------------------------------------- %
loadt `load_string`;;
load_string();;

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of progress %

loadf `autoload`;;

autoload_all_defs_and_thms
  `variables more_finite_sets bindings variants
   assert_syntax assert_semantics prog_syntax prog_semantics
   substitutions var_substitutions prog_substitutions
   free_variables translations `;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `new_ind_def`;;




% =================================================================== %
% Define a relation between states at the beginning of a command and  %
% just within a procedure called directly from that command.          %
% =================================================================== %

let C_calls_rules,C_calls_ind,C_calls_cases,C_calls_str_induct =
    let C_calls = "C_calls : cmd -> env -> state -> string -> state -> bool" in
    new_ind_def `C_calls`

    ("^C_calls c r s1 p s2", []) 


% Note: no rule for SKIP, since there are no elements in that relation. %

% Note: no rule for ABORT, since there are no elements in that relation. %

% Note: no rule for ASSIGN, since there are no elements in that relation. %

    
     [ [           "^C_calls c1 r s1 p s2"           
       % -------------------------------------------- % ],
               "^C_calls (SEQ c1 c2) r s1 p s2"       ;

    
       [            "^C_calls c2 r s2 p s3";
       % -------------------------------------------- % "C c1 r s1 s2" ],
               "^C_calls (SEQ c1 c2) r s1 p s3"       ;


       [            "^C_calls c1 r s2 p s3";
       % -------------------------------------------- % "B b s1 T s2" ],
               "^C_calls (IF b c1 c2) r s1 p s3"      ;


       [            "^C_calls c2 r s2 p s3";
       % -------------------------------------------- % "B b s1 F s2" ],
               "^C_calls (IF b c1 c2) r s1 p s3"      ;


       [            "^C_calls c r s2 p s3";
       % -------------------------------------------- % "B b s1 T s2" ],
             "^C_calls (WHILE a pr b c) r s1 p s3"    ;


       [     "^C_calls (WHILE a pr b c) r s3 p s4";     "B b s1 T s2";
       % -------------------------------------------- % "C c r s2 s3" ],
             "^C_calls (WHILE a pr b c) r s1 p s4"    ;


       ["ES es s1 ns s2";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
        "vals' = variants vals (SL (APPEND xs glbs))"
       % -------------------------------------------- % ],
       "^C_calls (CALL p xs es) r s1 p
                 ((BNDsL ns vals' s2) <|s
                  ((APPEND xs vals') // (APPEND vars vals)))" ];;




% =================================================================== %
% Define a relation between states at the entry of a procedure, and   %
% at the entry of a procedure called from the first procedure,        %
% possibly with multiple other procedure calls between.               %
% =================================================================== %

let M_calls_rules,M_calls_ind,M_calls_cases,M_calls_str_induct =
    let M_calls = "M_calls : string -> state -> (string)list
                             -> string -> state -> env -> bool" in
    new_ind_def `M_calls`

    ("^M_calls p1 s1 ps p2 s2 r", []) 

    
     [ [ "(r:env) p1 = (vars, vals, glbs, pre, post, calls, rec, c)";
                     "C_calls c r s1 p2 s2"          
       % -------------------------------------------- % ],
                  "^M_calls p1 s1 [] p2 s2 r"         ;

    
       [         "^M_calls p1 s1 ps1 p2 s2 r";
                 "^M_calls p2 s2 ps2 p3 s3 r"
       % -------------------------------------------- % ],
        "^M_calls p1 s1 (APPEND ps1 (CONS p2 ps2)) p3 s3 r" ];;



% --------------------------------------------------------------------- %
% Tactics for the rules.						%
% --------------------------------------------------------------------- %

let C_CALLS_TAC = FIRST (map RULE_TAC C_calls_rules);;
let [SEQ1_CALLS_TAC;SEQ2_CALLS_TAC;IF_T_CALLS_TAC;IF_F_CALLS_TAC;
     WHILE1_CALLS_TAC;WHILE2_CALLS_TAC;CALL_CALLS_TAC] =
    map RULE_TAC C_calls_rules;;

let M_CALLS_TAC = FIRST (map RULE_TAC M_calls_rules);;
let [SINGLE_CALL_TAC;MULTIPLE_CALLS_TAC] = map RULE_TAC M_calls_rules;;

% --------------------------------------------------------------------- %
% Rule induction tactics.                                               %
% --------------------------------------------------------------------- %

let C_CALLS_INDUCT_TAC = RULE_INDUCT_THEN C_calls_ind
                           STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let M_CALLS_INDUCT_TAC = RULE_INDUCT_THEN M_calls_ind
                           STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;


let mk_str_ind_tac str_ind_thm = RULE_INDUCT_THEN str_ind_thm
                                 STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;

let C_CALLS_STR_INDUCT_TAC = mk_str_ind_tac C_calls_str_induct;;
let M_CALLS_STR_INDUCT_TAC = mk_str_ind_tac M_calls_str_induct;;




let NULL_EQ_NIL =
 TAC_PROOF
  (([],
    "!l:(*)list. NULL l = (l = [])"),
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NULL]
   THEN REWRITE_TAC[NOT_CONS_NIL]
  );;

% NULL_EQ_NIL
g "!l:(*)list. NULL l = (l = [])";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[NULL]);;

 e(REWRITE_TAC[NULL]);;
 e(REWRITE_TAC[NOT_CONS_NIL]);;
%


let M_calls_IMP_C_calls_LEMMA =
 TAC_PROOF
  (([],
    "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!vars vals glbs pre post calls rec c.
        (ps = []) /\
        (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
        C_calls c r s1 p2 s2)"),
   M_CALLS_STR_INDUCT_TAC
   THENL
    [ REPEAT GEN_TAC
      THEN ASM_REWRITE_TAC[PAIR_EQ]
      THEN DISCH_THEN (REWRITE_ALL_TAC o map SYM o CONJUNCTS)
      THEN ASM_REWRITE_TAC[];

      REWRITE_TAC[(SYM o SPEC_ALL)NULL_EQ_NIL]
      THEN REWRITE_TAC[NULL_APPEND;NULL]
    ]
  );;

%< M_calls_IMP_C_calls_LEMMA
g "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!vars vals glbs pre post calls rec c.
        (ps = []) /\
        (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
        C_calls c r s1 p2 s2)";;
e(M_CALLS_STR_INDUCT_TAC);;

 e(REPEAT GEN_TAC);;
 e(ASM_REWRITE_TAC[PAIR_EQ]);;
 e(DISCH_THEN (REWRITE_ALL_TAC o map SYM o CONJUNCTS));;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[(SYM o SPEC_ALL)NULL_EQ_NIL]);;
 e(REWRITE_TAC[NULL_APPEND]);;
 e(REWRITE_TAC[NULL]);;
>%


let M_calls_IMP_C_calls =
 prove_thm
  (`M_calls_IMP_C_calls`,
   "!p1 s1 p2 s2 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      M_calls p1 s1 [] p2 s2 r ==>
      C_calls c r s1 p2 s2",
   REPEAT STRIP_TAC
   THEN ASSUME_TAC (REFL "[]:string list")
   THEN IMP_RES_TAC M_calls_IMP_C_calls_LEMMA
  );;

%< M_calls_IMP_C_calls
g "!p1 s1 p2 s2 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      M_calls p1 s1 [] p2 s2 r ==>
      C_calls c r s1 p2 s2";;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (REFL "[]:string list"));;
e(IMP_RES_TAC M_calls_IMP_C_calls_LEMMA);;
>%


let C_calls_M_calls =
 prove_thm
  (`C_calls_M_calls`,
   "!p1 s1 p2 s2 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (C_calls c r s1 p2 s2 = M_calls p1 s1 [] p2 s2 r)",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THENL
    [ SINGLE_CALL_TAC
      THEN EXISTS_TAC "vars:var list"
      THEN EXISTS_TAC "vals:var list"
      THEN EXISTS_TAC "glbs:var list"
      THEN EXISTS_TAC "pre:aexp"
      THEN EXISTS_TAC "post:aexp"
      THEN EXISTS_TAC "calls:prog_env"
      THEN EXISTS_TAC "rec:aexp"
      THEN EXISTS_TAC "c:cmd"
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC M_calls_IMP_C_calls
    ]
  );;

%< C_calls_M_calls
g "!p1 s1 p2 s2 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (C_calls c r s1 p2 s2 = M_calls p1 s1 [] p2 s2 r)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(SINGLE_CALL_TAC);;
 e(EXISTS_TAC "vars:var list"
   THEN EXISTS_TAC "vals:var list"
   THEN EXISTS_TAC "glbs:var list"
   THEN EXISTS_TAC "pre:aexp"
   THEN EXISTS_TAC "post:aexp"
   THEN EXISTS_TAC "calls:prog_env"
   THEN EXISTS_TAC "rec:aexp"
   THEN EXISTS_TAC "c:cmd");;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(IMP_RES_TAC M_calls_IMP_C_calls);;
>%


let M_calls_CONS_LEMMA =
 TAC_PROOF
  (([],
    "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!p3 ps'.
        (ps = CONS p3 ps') ==>
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps' p2 s2 r))"),
   M_CALLS_STR_INDUCT_TAC
   THENL
    [ REWRITE_TAC[NOT_NIL_CONS];

      REPEAT STRIP_TAC
      THEN STRIP_ASSUME_TAC (ISPEC "ps1:string list" list_CASES)
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN REWRITE_ALL_TAC[APPEND;CONS_11]
         THEN POP_ASSUM REWRITE_ALL_THM
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[];

         POP_ASSUM REWRITE_ALL_THM
         THEN REWRITE_ALL_TAC[APPEND;CONS_11]
         THEN POP_ASSUM STRIP_ASSUME_TAC
         THEN POP_ASSUM (ASSUME_TAC o SYM)
         THEN SUBGOAL_THEN "(h = (p3':string)) /\ ((t:string list) = t)"
                (\th1. FIRST_ASSUM (\th2. STRIP_ASSUME_TAC (MATCH_MP th2 th1)))
         THENL
          [ ASM_REWRITE_TAC[];

            EXISTS_TAC "s3':state"
            THEN ASM_REWRITE_TAC[]
            THEN MULTIPLE_CALLS_TAC
            THEN EXISTS_TAC "s2:state"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

%< M_calls_CONS_LEMMA
g "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!p3 ps'.
        (ps = CONS p3 ps') ==>
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps' p2 s2 r))";;
e(M_CALLS_STR_INDUCT_TAC);;

 e(REWRITE_TAC[NOT_NIL_CONS]);;

 e(REPEAT STRIP_TAC);;
 e(STRIP_ASSUME_TAC (ISPEC "ps1:string list" list_CASES));;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(REWRITE_ALL_TAC[APPEND;CONS_11]);;
  e(POP_ASSUM REWRITE_ALL_THM);;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(REWRITE_ALL_TAC[APPEND;CONS_11]);;
  e(POP_ASSUM STRIP_ASSUME_TAC);;
  e(POP_ASSUM (ASSUME_TAC o SYM));;
  e(SUBGOAL_THEN "(h = (p3':string)) /\ ((t:string list) = t)"
             (\th1. FIRST_ASSUM (\th2. STRIP_ASSUME_TAC (MATCH_MP th2 th1))));;

   e(ASM_REWRITE_TAC[]);;

   e(EXISTS_TAC "s3':state");;
   e(ASM_REWRITE_TAC[]);;
   e(MULTIPLE_CALLS_TAC);;
   e(EXISTS_TAC "s2:state");;
   e(ASM_REWRITE_TAC[]);;
>%


let M_calls_CONS_IMP =
 prove_thm
  (`M_calls_CONS_IMP`,
   "!p1 s1 p3 ps p2 s2 r.
      (M_calls p1 s1 (CONS p3 ps) p2 s2 r ==>
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps p2 s2 r))",

   REPEAT STRIP_TAC
   THEN ASSUME_TAC (REFL "CONS (p3:string) ps")
   THEN IMP_RES_TAC M_calls_CONS_LEMMA
   THEN EXISTS_TAC "s3:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< M_calls_CONS_IMP
g "!p1 s1 p3 ps p2 s2 r.
      (M_calls p1 s1 (CONS p3 ps) p2 s2 r ==>
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps p2 s2 r))";;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (REFL "CONS (p3:string) ps"));;
e(IMP_RES_TAC M_calls_CONS_LEMMA);;
e(EXISTS_TAC "s3:state");;
e(ASM_REWRITE_TAC[]);;
>%


let M_calls_CONS =
 prove_thm
  (`M_calls_CONS`,
   "!p1 s1 p3 ps p2 s2 r.
      M_calls p1 s1 (CONS p3 ps) p2 s2 r =
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps p2 s2 r)",

   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
    [ DISCH_TAC
      THEN IMP_RES_TAC M_calls_CONS_IMP
      THEN EXISTS_TAC "s3:state"
      THEN ASM_REWRITE_TAC[];

      STRIP_TAC
      THEN SUBGOAL_THEN "CONS (p3:string) ps = APPEND [] (CONS p3 ps)"
                  ONCE_REWRITE_THM
      THENL
       [ REWRITE_TAC[APPEND];

         MULTIPLE_CALLS_TAC
         THEN EXISTS_TAC "s3:state"
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< M_calls_CONS
g "!p1 s1 p3 ps p2 s2 r.
      M_calls p1 s1 (CONS p3 ps) p2 s2 r =
        (?s3. M_calls p1 s1 [] p3 s3 r /\ M_calls p3 s3 ps p2 s2 r)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(IMP_RES_TAC M_calls_CONS_IMP);;
 e(EXISTS_TAC "s3:state");;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(SUBGOAL_THEN "CONS (p3:string) ps = APPEND [] (CONS p3 ps)" ONCE_REWRITE_THM);;

  e(REWRITE_TAC[APPEND]);;

  e(MULTIPLE_CALLS_TAC);;
  e(EXISTS_TAC "s3:state");;
  e(ASM_REWRITE_TAC[]);;
>%



let M_calls_APPEND_IMP =
 prove_thm
  (`M_calls_APPEND_IMP`,
   "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r ==>
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)",

   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[APPEND;M_calls_CONS_IMP];

      REWRITE_TAC[APPEND]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC M_calls_CONS_IMP
      THEN RES_TAC
      THEN EXISTS_TAC "s3':state"
      THEN ASM_REWRITE_TAC[]
      THEN REWRITE_TAC[M_calls_CONS]
      THEN EXISTS_TAC "s3:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< M_calls_APPEND_IMP
g "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r ==>
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;M_calls_CONS_IMP]);;

 e(REWRITE_TAC[APPEND]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC M_calls_CONS_IMP);;
 e(RES_TAC);;
 e(EXISTS_TAC "s3':state");;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[M_calls_CONS]);;
 e(EXISTS_TAC "s3:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let M_calls_APPEND_LEMMA =
 TAC_PROOF
  (([],
    "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r =
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)"),
   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[APPEND;M_calls_CONS];

      REWRITE_TAC[APPEND]
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[M_calls_CONS]
      THEN EQ_TAC
      THEN STRIP_TAC
      THEN EXISTS_TAC "s3':state"
      THEN ASM_REWRITE_TAC[]
      THEN EXISTS_TAC "s3:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< M_calls_APPEND_LEMMA
g "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r =
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;M_calls_CONS]);;

 e(REWRITE_TAC[APPEND]);;
 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[M_calls_CONS]);;
 e(ASM_REWRITE_TAC[]);;
 e(EQ_TAC);;

   e(STRIP_TAC);;
   e(EXISTS_TAC "s3':state");;
   e(ASM_REWRITE_TAC[]);;
   e(EXISTS_TAC "s3:state");;
   e(ASM_REWRITE_TAC[]);;

   e(STRIP_TAC);;
   e(EXISTS_TAC "s3':state");;
   e(ASM_REWRITE_TAC[]);;
   e(EXISTS_TAC "s3:state");;
   e(ASM_REWRITE_TAC[]);;
>%


let M_calls_APPEND =
 prove_thm
  (`M_calls_APPEND`,
   "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r =
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)",
   REWRITE_TAC[M_calls_APPEND_LEMMA]
  );;

%< M_calls_APPEND
g "!ps1 p3 ps2 p1 s1 p2 s2 r.
      M_calls p1 s1 (APPEND ps1 (CONS p3 ps2)) p2 s2 r =
        (?s3. M_calls p1 s1 ps1 p3 s3 r /\ M_calls p3 s3 ps2 p2 s2 r)";;
e(REWRITE_TAC[M_calls_APPEND_LEMMA]);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(MATCH_MP_TAC M_calls_APPEND_IMP);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(IMP_RES_TAC DL_M_calls_path);;
 e(USE_IMP_THEN REWRITE_THM M_calls_APPEND_LEMMA);;

  e(ONCE_REWRITE_TAC[DL_APPEND]);;
  e(ASM_REWRITE_TAC[]);;

  e(EXISTS_TAC "s3:state");;
  e(ASM_REWRITE_TAC[]);;
>%




% =================================================================== %
% Define the basic predicates that specify Hoare-axiom relationships. %
% =================================================================== %


% {p} (p is true, in any state) is represented by CLOSE(p) %

let CLOSE =
  new_def
  `CLOSE`
  "CLOSE(p) = !s. A p s";;


% {p}e{q} is represented by ESPEC(p,e,q) %

let ESPEC = new_definition
 (`ESPEC`,
  "ESPEC(p,e,q) =
      (!s1 n s2. A p s1 /\ E e s1 n s2 ==> A q s2)  /\
      (!s1.      A p s1 ==> (?n s2. E e s1 n s2))");;


% {p}es{q} is represented by ES_SPEC(p,es,q) %

let ES_SPEC = new_definition
 (`ES_SPEC`,
  "ES_SPEC(p,es,q) =
      (!s1 ns s2. A p s1 /\ ES es s1 ns s2 ==> A q s2)  /\
      (!s1.       A p s1 ==> (?ns s2. ES es s1 ns s2))");;


% {p}b{q} is represented by BSPEC(p,b,q) %

let BSPEC = new_definition
 (`BSPEC`,
  "BSPEC(p,b,q) =
      (!s1 t s2. A p s1 /\ B b s1 t s2 ==> A q s2)  /\
      (!s1.      A p s1 ==> (?t s2. B b s1 t s2))");;


% {p}c{q}/r is represented by CSPEC(p,c,q,r) %

let PCSPEC = new_definition
 (`PCSPEC`,
  "PCSPEC(p,c,q,r) =
      (!s1 s2. A p s1 /\ C c r s1 s2 ==> A q s2)");;

let TCSPEC = new_definition
 (`TCSPEC`,
  "TCSPEC(p,c,r) =
      (!s1.    A p s1 ==> (?s2. C c r s1 s2))");;

let CSPEC = new_definition
 (`CSPEC`,
  "CSPEC(p,c,q,r) =
      PCSPEC(p,c,q,r)  /\
      TCSPEC(p,c,r)");;


let CSPEC_THM =
 prove_thm
  (`CSPEC_THM`,
   "!p c q r.
       CSPEC(p,c,q,r) =
       (!s1 s2. A p s1 /\ C c r s1 s2 ==> A q s2) /\
       (!s1. A p s1 ==> (?s2. C c r s1 s2))",
   REWRITE_TAC[CSPEC;PCSPEC;TCSPEC]
  );;

% CSPEC_THM
g "!p c q r.
       CSPEC(p,c,q,r) =
       (!s1 s2. A p s1 /\ C c r s1 s2 ==> A q s2) /\
       (!s1. A p s1 ==> (?s2. C c r s1 s2))";;
e(REWRITE_TAC[CSPEC_THM;PCSPEC_THM;TCSPEC]);;
%


% pi{q} is represented by PSPEC(pi,q)   (pi is a program) %

let PSPEC = new_definition
 (`PSPEC`,
  "PSPEC(pi,q) =
      (!s. P pi s ==> A q s)  /\
      (?s. P pi s)");;



% =================================================================== %
% Define the basic predicates that specify progress relationships in  %
% a Hoare-rule style.                                                 %
% =================================================================== %


% {a}c->p{q}/r is represented by CPROGRESS(a,c,p,q,r) %

let CPROGRESS = new_definition
 (`CPROGRESS`,
  "CPROGRESS(a,c,p,q,r) =
      (!s1 s2. A a s1 /\ C_calls c r s1 p s2 ==> A q s2)");;


% {a}c->{pre}/r is represented by PRE_PROGRESS(a,c,r) %

let PRE_PROGRESS = new_definition
 (`PRE_PROGRESS`,
  "PRE_PROGRESS(a,c,r) =
      (!p. let (vars,vals,glbs,pre,post,calls,rec,c') = r p in
           CPROGRESS(a,c,p,pre,r))");;


% {a}c->{calls}/r is represented by BODY_PROGRESS(a,c,calls,r) %

let BODY_PROGRESS = new_definition
 (`BODY_PROGRESS`,
  "BODY_PROGRESS(a,c,calls,r) =
      (!p. CPROGRESS(a,c,p,calls p,r))");;


% {a1}p1-ps->p2{a2}/r is represented by PATH_PROGRESS(a1,p1,ps,p2,a2,r) %

let PATH_PROGRESS = new_definition
 (`PATH_PROGRESS`,
  "PATH_PROGRESS(a1,p1,ps,p2,a2,r) =
      (!s1 s2. A a1 s1 /\ M_calls p1 s1 ps p2 s2 r ==> A a2 s2)");;


% {a1}<p>{a2}/r is represented by REC_PROGRESS(a1,p,a2,r) %

let REC_PROGRESS = new_definition
 (`REC_PROGRESS`,
  "REC_PROGRESS(a1,p,a2,r) =
      (!ps. PATH_PROGRESS(a1,p,ps,p,a2,r))");;





let CPROGRESS_PATH_PROGRESS =
 prove_thm
  (`CPROGRESS_PATH_PROGRESS`,
   "!a p1 p2 q r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (CPROGRESS(a, c, p2, q, r) =
       PATH_PROGRESS(a, p1, [], p2, q, r))",
   REWRITE_TAC[CPROGRESS;PATH_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM C_calls_M_calls
  );;

%< CPROGRESS_PATH_PROGRESS
g "!a p1 p2 q r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (CPROGRESS(a, c, p2, q, r) =
       PATH_PROGRESS(a, p1, [], p2, q, r))";;
e(REWRITE_TAC[CPROGRESS;PATH_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM C_calls_M_calls);;
>%


let BODY_PROGRESS_PATH_PROGRESS =
 prove_thm
  (`BODY_PROGRESS_PATH_PROGRESS`,
   "!a p1 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (BODY_PROGRESS(a, c, calls, r) =
       (!p2. PATH_PROGRESS(a, p1, [], p2, calls p2, r)))",
   REWRITE_TAC[BODY_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM CPROGRESS_PATH_PROGRESS
  );;

%< BODY_PROGRESS_PATH_PROGRESS
g "!a p1 r vars vals glbs pre post calls rec c.
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      (BODY_PROGRESS(a, c, calls, r) =
       (!p2. PATH_PROGRESS(a, p1, [], p2, calls p2, r)))";;
e(REWRITE_TAC[BODY_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM CPROGRESS_PATH_PROGRESS);;
>%



let PATH_PROGRESS_CONS =
 prove_thm
  (`PATH_PROGRESS_CONS`,
   "!a1 a2 a3 p1 p2 p3 ps r.
      PATH_PROGRESS(a1,p1,[],p2,a2,r) /\
      PATH_PROGRESS(a2,p2,ps,p3,a3,r) ==>
      PATH_PROGRESS(a1,p1,CONS p2 ps,p3,a3,r)",
   REWRITE_TAC[PATH_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC M_calls_CONS_IMP
   THEN RES_TAC
   THEN RES_TAC
  );;

%< PATH_PROGRESS_CONS
g "!a1 a2 a3 p1 p2 p3 ps r.
      PATH_PROGRESS(a1,p1,[],p2,a2,r) /\
      PATH_PROGRESS(a2,p2,ps,p3,a3,r) ==>
      PATH_PROGRESS(a1,p1,CONS p2 ps,p3,a3,r)";;
e(REWRITE_TAC[PATH_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC M_calls_CONS_IMP);;
e(RES_TAC);;
e(RES_TAC);;
>%



let PATH_PROGRESS_APPEND =
 prove_thm
  (`PATH_PROGRESS_APPEND`,
   "!a1 a2 a3 p1 p2 p3 ps1 ps2 r.
      PATH_PROGRESS(a1,p1,ps1,p2,a2,r) /\
      PATH_PROGRESS(a2,p2,ps2,p3,a3,r) ==>
      PATH_PROGRESS(a1,p1,APPEND ps1 (CONS p2 ps2),p3,a3,r)",
   REWRITE_TAC[PATH_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC M_calls_APPEND_IMP
   THEN RES_TAC
   THEN RES_TAC
  );;

%< PATH_PROGRESS_APPEND
g "!a1 a2 a3 p1 p2 p3 ps1 ps2 r.
      PATH_PROGRESS(a1,p1,ps1,p2,a2,r) /\
      PATH_PROGRESS(a2,p2,ps2,p3,a3,r) ==>
      PATH_PROGRESS(a1,p1,APPEND ps1 (CONS p2 ps2),p3,a3,r)";;
e(REWRITE_TAC[PATH_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC M_calls_APPEND_IMP);;
e(RES_TAC);;
e(RES_TAC);;
>%




close_theory ();;

