% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the proofs of theorems in the form of
  Floyd/Hoare-style programming language axioms and rules of inference,
  for the procedure entrance logic.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `progress_rules`;;

new_parent `hoare_rules`;;


% --------------------------------------------------------------------- %
% Need the string, finite_sets, and ind_defs libraries.                 %
% --------------------------------------------------------------------- %
loadt `load_string`;;
load_string();;

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of progress_rules %

loadf `autoload`;;

autoload_all_defs_and_thms
  `more_finite_sets bindings variables variants
   assert_syntax assert_semantics prog_syntax prog_semantics
   substitutions var_substitutions prog_substitutions
   prog_semantics free_variables translations progress
   well_formed cmd_semantics hoare_rules`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `progress_tactics`;;



% ==================================================== %
% Now for the C_calls_BND_THM and M_calls_BND_THM,     %
% BNDsL and BNDsR list versions, and theorems about    %
% the impact of a C_calls or M_calls state transition  %
% being limited to program variables.                  %
%                                                      %
% These theorems depend on well-formed commands and    %
% environments, and so could not be included in the    %
% `progress` theory earlier.                           %
% ==================================================== %



% Put the following theorem in theory `substitutions`. %

let BND_s_subst_var_list_COMMUTE =
 TAC_PROOF
  (([],
    "!n x s ys xs.
      ~(x IN SL ys) /\ ~(x IN SL xs) ==>
      (BND n x (s <|s (ys // xs)) = (BND n x s) <|s (ys // xs))"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE
   THEN IMP_RES_THEN REWRITE_THM MAP_BND_ABSORPTION
  );;

%< BND_s_subst_var_list_COMMUTE
g "!n x s ys xs.
      ~(x IN SL ys) /\ ~(x IN SL xs) ==>
      (BND n x (s <|s (ys // xs)) = (BND n x s) <|s (ys // xs))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(IMP_RES_THEN REWRITE_THM BND_BNDsR_COMMUTE);;
e(IMP_RES_THEN REWRITE_THM MAP_BND_ABSORPTION);;
>%


let C_calls_BND_LEMMA =
 TAC_PROOF
  (([],
    "!c r s1 p s2.
      C_calls c r s1 p s2 ==>
      (!g x n. WFenv_syntax r /\ WFc c g r /\ ~(WFx x) ==>
        C_calls c r (BND n x s1) p (BND n x s2))"),
   C_CALLS_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc]
   THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THENL  % 7 subgoals %
    [ RES_TAC
      THEN SEQ1_CALLS_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC (AND2IMP WFc_NOT_WFx)
      THEN IMP_RES_TAC (AND2IMP C_BND_THM)
      THEN RES_TAC
      THEN SEQ2_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_BND_THM
      THEN RES_TAC
      THEN IF_T_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_BND_THM
      THEN RES_TAC
      THEN IF_F_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[];

      UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')"
      THEN IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_BND_THM
      THEN RES_TAC
      THEN WHILE1_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[];

      UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')"
      THEN IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFx)
      THEN IMP_RES_TAC B_BND_THM
      THEN IMP_RES_TAC (AND2IMP C_BND_THM)
      THEN WHILE2_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s3:state)"
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[]
      THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM];

      UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
      THEN USE_IMP_THEN REWRITE_THM BND_s_subst_var_list_COMMUTE
      THENL
       [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
         THEN REWRITE_ALL_TAC[WFxs_APPEND]
         THEN CONJ_TAC
         THEN MATCH_MP_TAC WFxs_NOT_WFx
         THEN ASM_REWRITE_TAC[WFxs_APPEND;WFxs_variants];

         USE_IMP_THEN REWRITE_THM BND_BNDsL_COMMUTE
         THENL
          [ MATCH_MP_TAC WFxs_NOT_WFx
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN REWRITE_ALL_TAC[WFxs_APPEND]
            THEN ASM_REWRITE_TAC[WFxs_variants];

            CALL_CALLS_TAC
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN EXISTS_TAC "c:cmd"
            THEN IMP_RES_TAC WFes_NOT_WFx
            THEN IMP_RES_TAC ES_BND_THM
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

%< C_calls_BND_LEMMA
g "!c r s1 p s2.
      C_calls c r s1 p s2 ==>
      (!g x n. WFenv_syntax r /\ WFc c g r /\ ~(WFx x) ==>
        C_calls c r (BND n x s1) p (BND n x s2))";;
e(C_CALLS_STR_INDUCT_TAC
  THEN REWRITE_TAC[WFc]
  THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(SEQ1_CALLS_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC (AND2IMP WFc_NOT_WFx));;
 e(IMP_RES_TAC (AND2IMP C_BND_THM));;
 e(RES_TAC);;
 e(SEQ2_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IF_T_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IF_F_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')");;
 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(WHILE1_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')");;
 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC (AND2IMP WFc_NOT_WFx));;
 e(IMP_RES_TAC B_BND_THM);;
 e(IMP_RES_TAC (AND2IMP C_BND_THM));;
 e(WHILE2_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s3:state)");;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;

let st2 = get_state();;
set_state st2;;
 e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
      "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(USE_IMP_THEN REWRITE_THM BND_s_subst_var_list_COMMUTE);;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(CONJ_TAC);;

   e(MATCH_MP_TAC WFxs_NOT_WFx);;
   e(ASM_REWRITE_TAC[WFxs_APPEND;WFxs_variants]);;

   e(MATCH_MP_TAC WFxs_NOT_WFx);;
   e(ASM_REWRITE_TAC[WFxs_APPEND]);;

  e(USE_IMP_THEN REWRITE_THM BND_BNDsL_COMMUTE);;

   e(MATCH_MP_TAC WFxs_NOT_WFx);;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants]);;

   e(CALL_CALLS_TAC);;
   e(EXISTS_TAC "glbs:(var)list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp"
     THEN EXISTS_TAC "c:cmd");;
   e(IMP_RES_TAC WFes_NOT_WFx);;
   e(IMP_RES_TAC ES_BND_THM);;
   e(ASM_REWRITE_TAC[]);;
>%


let C_calls_BND_THM =
 prove_thm
  (`C_calls_BND_THM`,
   "!n x c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\ C_calls c r s1 p s2 /\ ~WFx x ==>
      C_calls c r (BND n x s1) p (BND n x s2)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC C_calls_BND_LEMMA
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_BND_THM:
g "!n x c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\ C_calls c r s1 p s2 /\ ~WFx x ==>
      C_calls c r (BND n x s1) p (BND n x s2)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC C_calls_BND_LEMMA);;
e(ASM_REWRITE_TAC[]);;
>%

let C_calls_BNDsL_THM =
 prove_thm
  (`C_calls_BNDsL_THM`,
   "!ns xs c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\
      C_calls c r s1 p s2 /\ NOT_WFxs xs ==>
      C_calls c r (BNDsL ns xs s1) p (BNDsL ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC C_calls_BND_THM
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_BNDsL_THM:
g "!ns xs c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\
      C_calls c r s1 p s2 /\ NOT_WFxs xs ==>
      C_calls c r (BNDsL ns xs s1) p (BNDsL ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[NOT_WFxs;BNDsL]
  THEN REPEAT STRIP_TAC);;
e(IMP_RES_TAC C_calls_BND_THM);;
e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[]);;
>%

let C_calls_BNDsR_THM =
 prove_thm
  (`C_calls_BNDsR_THM`,
   "!ns xs c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\
      C_calls c r s1 p s2 /\ NOT_WFxs xs ==>
      C_calls c r (BNDsR ns xs s1) p (BNDsR ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL C_calls_BND_THM)
   THEN ASM_REWRITE_TAC[]
   THEN RES_TAC
  );;

%< C_calls_BNDsR_THM:
g "!ns xs c g r s1 p s2.
      WFenv_syntax r /\ WFc c g r /\
      C_calls c r s1 p s2 /\ NOT_WFxs xs ==>
      C_calls c r (BNDsR ns xs s1) p (BNDsR ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[NOT_WFxs;BNDsR]
  THEN REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL C_calls_BND_THM));;
e(ASM_REWRITE_TAC[]);;
e(RES_TAC);;
>%



let M_calls_BND_LEMMA =
 TAC_PROOF
  (([],
    "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!x n. WFenv_syntax r /\ ~(WFx x) ==>
        M_calls p1 (BND n x s1) ps p2 (BND n x s2) r)"),
   M_CALLS_STR_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN IMP_RES_TAC (AND2IMP C_calls_BND_THM)
      THEN POP_ASSUM (MP_TAC o SPEC_ALL)
      THEN IMP_RES_THEN REWRITE_THM C_calls_M_calls;

      RES_TAC
      THEN MULTIPLE_CALLS_TAC
      THEN EXISTS_TAC "BND n x (s2:state)"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< M_calls_BND_LEMMA
g "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!x n. WFenv_syntax r /\ ~(WFx x) ==>
        M_calls p1 (BND n x s1) ps p2 (BND n x s2) r)";;

e(M_CALLS_STR_INDUCT_TAC
  THEN REPEAT STRIP_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(IMP_RES_TAC (AND2IMP C_calls_BND_THM));;
 e(POP_ASSUM (MP_TAC o SPEC_ALL));;
 e(IMP_RES_THEN REWRITE_THM C_calls_M_calls);;

 e(RES_TAC);;
 e(MULTIPLE_CALLS_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;
>%

let M_calls_BND_THM =
 prove_thm
  (`M_calls_BND_THM`,
   "!x n p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ ~(WFx x) ==>
     M_calls p1 (BND n x s1) ps p2 (BND n x s2) r",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC M_calls_BND_LEMMA
   THEN ASM_REWRITE_TAC[]
  );;

%< M_calls_BND_THM
g "!x n p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ ~(WFx x) ==>
     M_calls p1 (BND n x s1) ps p2 (BND n x s2) r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC M_calls_BND_LEMMA);;
e(ASM_REWRITE_TAC[]);;
>%

let M_calls_BNDsL_THM =
 prove_thm
  (`M_calls_BNDsL_THM`,
   "!xs ns p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ NOT_WFxs xs ==>
     M_calls p1 (BNDsL ns xs s1) ps p2 (BNDsL ns xs s2) r",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP M_calls_BND_THM)
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< M_calls_BNDsL_THM:
g "!xs ns p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ NOT_WFxs xs ==>
     M_calls p1 (BNDsL ns xs s1) ps p2 (BNDsL ns xs s2) r";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[NOT_WFxs;BNDsL]
  THEN REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP M_calls_BND_THM));;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(ASM_REWRITE_TAC[]);;
>%

let M_calls_BNDsR_THM =
 prove_thm
  (`M_calls_BNDsR_THM`,
   "!ns xs p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ NOT_WFxs xs ==>
     M_calls p1 (BNDsR ns xs s1) ps p2 (BNDsR ns xs s2) r",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC M_calls_BND_THM
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< M_calls_BNDsR_THM:
g "!ns xs p1 s1 ps p2 s2 r.
     WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ NOT_WFxs xs ==>
     M_calls p1 (BNDsR ns xs s1) ps p2 (BNDsR ns xs s2) r";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[NOT_WFxs;BNDsR]
  THEN REPEAT STRIP_TAC);;
e(MATCH_MP_TAC M_calls_BND_THM);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
>%



let C_calls_PROG_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!c r s1 p s2.
     C_calls c r s1 p s2 ==>
     (!g x. WFenv_syntax r /\ WFc c g r /\ ~(WFx x) ==>
            (s2 x = s1 x))"),
   C_CALLS_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc]
   THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THENL  % 7 subgoals %
    [ RES_TAC;

      IMP_RES_TAC (AND2IMP WFc_NOT_WFx)
      THEN IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT)
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC WFb_NOT_WFx
      THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFx)
      THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT)
      THEN ASSUM_LIST (REWRITE_TAC o mapfilter SYM)
      THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM];

      UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN REWRITE_TAC[s_subst_o;o_THM]
      THEN USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH
      THENL
       [ MATCH_MP_TAC WFxs_NOT_WFx
         THEN ASM_REWRITE_TAC[WFxs_APPEND];

         REWRITE_TAC[I_THM]
         THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
         THENL
          [ MATCH_MP_TAC WFxs_NOT_WFx
            THEN ASM_REWRITE_TAC[WFxs_variants];

            IMP_RES_TAC WFes_NOT_WFx
            THEN IMP_RES_TAC ES_FREE_ONLY_IMPACT
          ]
       ]
    ]
  );;

%< C_calls_PROG_ONLY_IMPACT_LEMMA
g "!c r s1 p s2.
     C_calls c r s1 p s2 ==>
     (!g x. WFenv_syntax r /\ WFc c g r /\ ~(WFx x) ==>
            (s2 x = s1 x))";;
e(C_CALLS_STR_INDUCT_TAC
  THEN REWRITE_TAC[WFc]
  THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(RES_TAC);;

 e(IMP_RES_TAC (AND2IMP WFc_NOT_WFx));;
 e(IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT));;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(IMP_RES_TAC (AND2IMP WFc_NOT_WFx));;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT));;
 e(ASSUM_LIST (REWRITE_TAC o mapfilter SYM));;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;

 e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
      "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(REWRITE_TAC[s_subst_o;o_THM]);;
 e(USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH);;

  e(MATCH_MP_TAC WFxs_NOT_WFx
    THEN ASM_REWRITE_TAC[WFxs_APPEND]);;

  e(REWRITE_TAC[I_THM]);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(MATCH_MP_TAC WFxs_NOT_WFx);;
   e(ASM_REWRITE_TAC[WFxs_variants]);;

   e(IMP_RES_TAC WFes_NOT_WFx);;
   e(IMP_RES_TAC ES_FREE_ONLY_IMPACT);;
>%

let C_calls_PROG_ONLY_IMPACT =
 prove_thm
  (`C_calls_PROG_ONLY_IMPACT`,
   "!c g r s1 p s2 x.
     WFenv_syntax r /\ WFc c g r /\ C_calls c r s1 p s2 /\ ~(WFx x) ==>
            (s2 x = s1 x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC C_calls_PROG_ONLY_IMPACT_LEMMA
  );;


let M_calls_PROG_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!x. WFenv_syntax r /\ ~(WFx x) ==>
           (s2 x = s1 x))"),
   M_CALLS_STR_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN IMP_RES_TAC (AND2IMP C_calls_PROG_ONLY_IMPACT);

      RES_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< M_calls_PROG_ONLY_IMPACT_LEMMA
g "!p1 s1 ps p2 s2 r.
      M_calls p1 s1 ps p2 s2 r ==>
      (!x. WFenv_syntax r /\ ~(WFx x) ==>
           (s2 x = s1 x))";;

e(M_CALLS_STR_INDUCT_TAC
  THEN REPEAT STRIP_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(IMP_RES_TAC C_calls_PROG_ONLY_IMPACT);;

 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%

let M_calls_PROG_ONLY_IMPACT =
 prove_thm
  (`M_calls_PROG_ONLY_IMPACT`,
   "!p1 s1 ps p2 s2 r x.
      WFenv_syntax r /\ M_calls p1 s1 ps p2 s2 r /\ ~(WFx x) ==>
           (s2 x = s1 x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC M_calls_PROG_ONLY_IMPACT_LEMMA
  );;




% ============================================================== %
% ============================================================== %
% Prove as theorems an axiomatic semantics for the progress      %
% achieved across commands which issue calls to procedures.      %
% ============================================================== %
% ============================================================== %


let PRE_STRENGTH_CPROGRESS =
 prove_thm
  (`PRE_STRENGTH_CPROGRESS`,
   "!a0 a1 a2 c p r.
     CLOSE(AIMP a0 a1) /\ CPROGRESS(a1, c, p, a2, r) ==>
     CPROGRESS(a0, c, p, a2, r)",
   REWRITE_TAC[CLOSE;CPROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

% PRE_STRENGTH_CPROGRESS
g "!a0 a1 a2 c p r.
     CLOSE(AIMP a0 a1) /\ CPROGRESS(a1, c, p, a2, r) ==>
     CPROGRESS(a0, c, p, a2, r)";;
e(REWRITE_TAC[CLOSE;CPROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
%


let PRE_STRENGTH_PRE_PROGRESS =
 prove_thm
  (`PRE_STRENGTH_PRE_PROGRESS`,
   "!a0 a1 c r.
     CLOSE(AIMP a0 a1) /\ PRE_PROGRESS(a1, c, r) ==>
     PRE_PROGRESS(a0, c, r)",
   REWRITE_TAC[PRE_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (MP_TAC o SPEC_ALL)
   THEN ONCE_REWRITE_TAC[ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN DISCH_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_CPROGRESS
  );;

% PRE_STRENGTH_PRE_PROGRESS
g "!a0 a1 c r.
     CLOSE(AIMP a0 a1) /\ PRE_PROGRESS(a1, c, r) ==>
     PRE_PROGRESS(a0, c, r)";;
e(REWRITE_TAC[PRE_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (MP_TAC o SPEC_ALL));;
e(ONCE_REWRITE_TAC[ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(DISCH_TAC);;
e(IMP_RES_TAC PRE_STRENGTH_CPROGRESS);;
%


let PRE_STRENGTH_BODY_PROGRESS =
 prove_thm
  (`PRE_STRENGTH_BODY_PROGRESS`,
   "!a0 a1 c calls r.
     CLOSE(AIMP a0 a1) /\ BODY_PROGRESS(a1, c, calls, r) ==>
     BODY_PROGRESS(a0, c, calls, r)",
   REWRITE_TAC[BODY_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC PRE_STRENGTH_CPROGRESS
  );;

% PRE_STRENGTH_BODY_PROGRESS
g "!a0 a1 c calls r.
     CLOSE(AIMP a0 a1) /\ BODY_PROGRESS(a1, c, calls, r) ==>
     BODY_PROGRESS(a0, c, calls, r)";;
e(REWRITE_TAC[BODY_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC PRE_STRENGTH_CPROGRESS);;
%


let PRE_STRENGTH_PATH_PROGRESS =
 prove_thm
  (`PRE_STRENGTH_PATH_PROGRESS`,
   "!a0 a1 a2 p1 p2 ps r.
     CLOSE(AIMP a0 a1) /\ PATH_PROGRESS(a1, p1, ps, p2, a2, r) ==>
     PATH_PROGRESS(a0, p1, ps, p2, a2, r)",
   REWRITE_TAC[CLOSE;PATH_PROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

% PRE_STRENGTH_PATH_PROGRESS
g "!a0 a1 a2 p1 p2 ps r.
     CLOSE(AIMP a0 a1) /\ PATH_PROGRESS(a1, p1, ps, p2, a2, r) ==>
     PATH_PROGRESS(a0, p1, ps, p2, a2, r)";;
e(REWRITE_TAC[CLOSE;PATH_PROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
%


let PRE_STRENGTH_REC_PROGRESS =
 prove_thm
  (`PRE_STRENGTH_REC_PROGRESS`,
   "!a0 a1 a2 p r.
     CLOSE(AIMP a0 a1) /\ REC_PROGRESS(a1,p,a2,r) ==>
     REC_PROGRESS(a0,p,a2,r)",
   REWRITE_TAC[REC_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC PRE_STRENGTH_PATH_PROGRESS
  );;

% PRE_STRENGTH_REC_PROGRESS
g "!a0 a1 a2 p r.
     CLOSE(AIMP a0 a1) /\ REC_PROGRESS(a1,p,a2,r) ==>
     REC_PROGRESS(a0,p,a2,r)";;
e(REWRITE_TAC[REC_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC PRE_STRENGTH_PATH_PROGRESS);;
%


let POST_WEAK_CPROGRESS =
 prove_thm
  (`POST_WEAK_CPROGRESS`,
   "!a1 a2 a3 c p r.
     CPROGRESS(a1, c, p, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     CPROGRESS(a1, c, p, a3, r)",
   REWRITE_TAC[CLOSE;CPROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

% POST_WEAK_CPROGRESS
g "!a1 a2 a3 c p r.
     CPROGRESS(a1, c, p, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     CPROGRESS(a1, c, p, a3, r)";;
e(REWRITE_TAC[CLOSE;CPROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
%


let POST_WEAK_BODY_PROGRESS =
 prove_thm
  (`POST_WEAK_BODY_PROGRESS`,
   "!a c calls calls' r.
     BODY_PROGRESS(a, c, calls, r) /\
     (!p. CLOSE(AIMP (calls p) (calls' p))) ==>
     BODY_PROGRESS(a, c, calls', r)",
   REWRITE_TAC[BODY_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC POST_WEAK_CPROGRESS
   THEN EXISTS_TAC "(calls:prog_env) p"
   THEN ASM_REWRITE_TAC[]
  );;

% POST_WEAK_BODY_PROGRESS
g "!a c calls calls' r.
     BODY_PROGRESS(a, c, calls, r) /\
     (!p. CLOSE(AIMP (calls p) (calls' p))) ==>
     BODY_PROGRESS(a, c, calls', r)";;
e(REWRITE_TAC[BODY_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC POST_WEAK_CPROGRESS);;
e(EXISTS_TAC "(calls:prog_env) p");;
e(ASM_REWRITE_TAC[]);;
%


let POST_WEAK_PATH_PROGRESS =
 prove_thm
  (`POST_WEAK_PATH_PROGRESS`,
   "!a1 a2 a3 p1 p2 ps r.
     PATH_PROGRESS(a1, p1, ps, p2, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     PATH_PROGRESS(a1, p1, ps, p2, a3, r)",
   REWRITE_TAC[CLOSE;PATH_PROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

% POST_WEAK_PATH_PROGRESS
g "!a1 a2 a3 p1 p2 ps r.
     PATH_PROGRESS(a1, p1, ps, p2, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     PATH_PROGRESS(a1, p1, ps, p2, a3, r)";;
e(REWRITE_TAC[CLOSE;PATH_PROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
%


let POST_WEAK_REC_PROGRESS =
 prove_thm
  (`POST_WEAK_REC_PROGRESS`,
   "!a1 a2 a3 p r.
     REC_PROGRESS(a1, p, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     REC_PROGRESS(a1, p, a3, r)",
   REWRITE_TAC[REC_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (ASSUME_TAC o SPEC "ps:string list")
   THEN IMP_RES_TAC POST_WEAK_PATH_PROGRESS
  );;

% POST_WEAK_REC_PROGRESS
g "!a1 a2 a3 p r.
     REC_PROGRESS(a1, p, a2, r) /\ CLOSE(AIMP a2 a3) ==>
     REC_PROGRESS(a1, p, a3, r)";;
e(REWRITE_TAC[REC_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (ASSUME_TAC o SPEC "ps:string list"));;
e(IMP_RES_TAC POST_WEAK_PATH_PROGRESS);;
%


let POST_AAND_CPROGRESS =
 prove_thm
  (`POST_AAND_CPROGRESS`,
   "!a1 a2 a3 c p r.
     CPROGRESS(a1, c, p, a2, r) /\ CPROGRESS(a1, c, p, a3, r) ==>
     CPROGRESS(a1, c, p, AAND a2 a3, r)",
   REWRITE_TAC[CPROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% POST_AAND_CPROGRESS
g "!a1 a2 a3 c p r.
     CPROGRESS(a1, c, p, a2, r) /\ CPROGRESS(a1, c, p, a3, r) ==>
     CPROGRESS(a1, c, p, AAND a2 a3, r)";;
e(REWRITE_TAC[CPROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;

 e(RES_TAC);;
%

let POST_AAND_BODY_PROGRESS =
 prove_thm
  (`POST_AAND_BODY_PROGRESS`,
   "!a c calls1 calls2 r.
     BODY_PROGRESS(a, c, calls1, r) /\ BODY_PROGRESS(a, c, calls2, r) ==>
     BODY_PROGRESS(a, c, (\p. AAND (calls1 p) (calls2 p)), r)",
   REWRITE_TAC[BODY_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN BETA_TAC
   THEN MATCH_MP_TAC POST_AAND_CPROGRESS
   THEN ASM_REWRITE_TAC[]
  );;

% POST_AAND_BODY_PROGRESS
g "!a c calls1 calls2 r.
     BODY_PROGRESS(a, c, calls1, r) /\ BODY_PROGRESS(a, c, calls2, r) ==>
     BODY_PROGRESS(a, c, (\p. AAND (calls1 p) (calls2 p)), r)";;
e(REWRITE_TAC[BODY_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(BETA_TAC);;
e(MATCH_MP_TAC POST_AAND_CPROGRESS);;
e(ASM_REWRITE_TAC[]);;
%

let POST_AAND_PATH_PROGRESS =
 prove_thm
  (`POST_AAND_PATH_PROGRESS`,
   "!a1 a2 a3 p1 ps p2 r.
     PATH_PROGRESS(a1, p1, ps, p2, a2, r) /\
     PATH_PROGRESS(a1, p1, ps, p2, a3, r) ==>
     PATH_PROGRESS(a1, p1, ps, p2, AAND a2 a3, r)",
   REWRITE_TAC[PATH_PROGRESS;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% POST_AAND_PATH_PROGRESS
g "!a1 a2 a3 p1 ps p2 r.
     PATH_PROGRESS(a1, p1, ps, p2, a2, r) /\
     PATH_PROGRESS(a1, p1, ps, p2, a3, r) ==>
     PATH_PROGRESS(a1, p1, ps, p2, AAND a2 a3, r)";;
e(REWRITE_TAC[PATH_PROGRESS;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;

 e(RES_TAC);;
%

let POST_AAND_REC_PROGRESS =
 prove_thm
  (`POST_AAND_REC_PROGRESS`,
   "!a1 a2 a3 p r.
     REC_PROGRESS(a1, p, a2, r) /\
     REC_PROGRESS(a1, p, a3, r) ==>
     REC_PROGRESS(a1, p, AAND a2 a3, r)",
   REWRITE_TAC[REC_PROGRESS]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC POST_AAND_PATH_PROGRESS
   THEN ASM_REWRITE_TAC[]
  );;

% POST_AAND_REC_PROGRESS
g "!a1 a2 a3 p r.
     REC_PROGRESS(a1, p, a2, r) /\
     REC_PROGRESS(a1, p, a3, r) ==>
     REC_PROGRESS(a1, p, AAND a2 a3, r)";;
e(REWRITE_TAC[REC_PROGRESS]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC POST_AAND_PATH_PROGRESS);;
e(ASM_REWRITE_TAC[]);;
%


let FALSE_CPROGRESS =
 prove_thm
  (`FALSE_CPROGRESS`,
   "!c p a r. CPROGRESS(AFALSE, c, p, a, r)",
   REWRITE_TAC[CPROGRESS;A_DEF]
  );;

% FALSE_CPROGRESS
g "!c p a r. CPROGRESS(AFALSE, c, p, a, r)";;
e(REWRITE_TAC[CPROGRESS;A_DEF]);;
%


let FALSE_PRE_PROGRESS =
 prove_thm
  (`FALSE_PRE_PROGRESS`,
   "!c r. PRE_PROGRESS(AFALSE, c, r)",
   REWRITE_TAC[PRE_PROGRESS]
   THEN ONCE_REWRITE_TAC[ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[FALSE_CPROGRESS]
  );;

% FALSE_PRE_PROGRESS
g "!c r. PRE_PROGRESS(AFALSE, c, r)";;
e(REWRITE_TAC[PRE_PROGRESS]);;
e(ONCE_REWRITE_TAC[ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[FALSE_CPROGRESS]);;
%


let FALSE_BODY_PROGRESS =
 prove_thm
  (`FALSE_BODY_PROGRESS`,
   "!c calls r. BODY_PROGRESS(AFALSE, c, calls, r)",
   REWRITE_TAC[BODY_PROGRESS;FALSE_CPROGRESS]
  );;

% FALSE_BODY_PROGRESS
g "!c calls r. BODY_PROGRESS(AFALSE, c, calls, r)";;
e(REWRITE_TAC[BODY_PROGRESS;FALSE_CPROGRESS]);;
%


let FALSE_PATH_PROGRESS =
 prove_thm
  (`FALSE_PATH_PROGRESS`,
   "!p1 ps p2 a r. PATH_PROGRESS(AFALSE, p1, ps, p2, a, r)",
   REWRITE_TAC[PATH_PROGRESS;A_DEF]
  );;

% FALSE_PATH_PROGRESS
g "!p1 ps p2 a r. PATH_PROGRESS(AFALSE, p1, ps, p2, a, r)";;
e(REWRITE_TAC[PATH_PROGRESS;A_DEF]);;
%


let FALSE_REC_PROGRESS =
 prove_thm
  (`FALSE_REC_PROGRESS`,
   "!p a r. REC_PROGRESS(AFALSE, p, a, r)",
   REWRITE_TAC[REC_PROGRESS;FALSE_PATH_PROGRESS]
  );;

% FALSE_REC_PROGRESS
g "!p a r. REC_PROGRESS(AFALSE, p, a, r)";;
e(REWRITE_TAC[REC_PROGRESS;FALSE_PATH_PROGRESS]);;
%



% ============================================================== %
% Prove the progress of each kind of command, and show what must %
% be true within the entry of each procedure directly called.    %
% ============================================================== %


let SKIP_PROGRESS =
 prove_thm
  (`SKIP_PROGRESS`,
   "!a1 a2 p r. CPROGRESS(a1, SKIP, p, a2, r)",
   REWRITE_TAC[CPROGRESS]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct]
  );;

% SKIP_PROGRESS
g "!a1 a2 p r. CPROGRESS(a1, SKIP, p, a2, r)";;
e(REWRITE_TAC[CPROGRESS]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
%


let ABORT_PROGRESS =
 prove_thm
  (`ABORT_PROGRESS`,
   "!a1 a2 p r. CPROGRESS(a1, ABORT, p, a2, r)",
   REWRITE_TAC[CPROGRESS]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct]
  );;

% ABORT_PROGRESS
g "!a1 a2 p r. CPROGRESS(a1, ABORT, p, a2, r)";;
e(REWRITE_TAC[CPROGRESS]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
%


let ASSIGN_PROGRESS =
 prove_thm
  (`ASSIGN_PROGRESS`,
   "!a1 a2 x e p r. CPROGRESS(a1, ASSIGN x e, p, a2, r)",
   REWRITE_TAC[CPROGRESS]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct]
  );;

% ASSIGN_PROGRESS
g "!a1 a2 x e p r. CPROGRESS(a1, ASSIGN x e, p, a2, r)";;
e(REWRITE_TAC[CPROGRESS]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
%


let SEQ_PROGRESS =
 prove_thm
  (`SEQ_PROGRESS`,
   "!a1 a2 a3 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           PCSPEC(a2, c1, a3, r) /\
           CPROGRESS(a3, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(AAND a1 a2, SEQ c1 c2, p, q, r)",

   REWRITE_TAC[CPROGRESS;PCSPEC;A_DEF]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REPEAT GEN_TAC THEN STRIP_TAC  % 2 subgoals %
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'"
   THEN RES_TAC
   THEN RES_TAC
  );;

%< SEQ_PROGRESS
g "!a1 a2 a3 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           PCSPEC(a2, c1, a3, r) /\
           CPROGRESS(a3, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(AAND a1 a2, SEQ c1 c2, p, q, r)";;

e(REWRITE_TAC[CPROGRESS;PCSPEC;A_DEF]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
% 2 subgoals %

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(RES_TAC);;

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(RES_TAC);;
 e(RES_TAC);;

>%


let SEQ_PROGRESS1 =
 prove_thm
  (`SEQ_PROGRESS1`,
   "!a1 a2 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           PCSPEC(a1, c1, a2, r) /\
           CPROGRESS(a2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(a1, SEQ c1 c2, p, q, r)",

   REWRITE_TAC[CPROGRESS;PCSPEC;A_DEF]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REPEAT GEN_TAC THEN STRIP_TAC  % 2 subgoals %
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'"
   THEN RES_TAC
   THEN RES_TAC
  );;

%< SEQ_PROGRESS1
g "!a1 a2 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           PCSPEC(a1, c1, a2, r) /\
           CPROGRESS(a2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(a1, SEQ c1 c2, p, q, r)";;

e(REWRITE_TAC[CPROGRESS;PCSPEC;A_DEF]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
% 2 subgoals %

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(RES_TAC);;

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(RES_TAC);;
 e(RES_TAC);;

>%


let IF_PROGRESS =
 prove_thm
  (`IF_PROGRESS`,
   "!a1 a2 q1 q2 b c1 c2 p q r.

           BSPEC(a1, b, q1) /\
           CPROGRESS(q1, c1, p, q, r) /\
           BSPEC(a2, b, q2) /\
           CPROGRESS(q2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(ACOND (AB b) a1 a2, IF b c1 c2, p, q, r)",

   REWRITE_TAC[CPROGRESS;BSPEC]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[A_DEF]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REPEAT GEN_TAC THEN STRIP_TAC  % 2 subgoals %
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'"
   THEN FIRST_ASSUM (REWRITE_ALL_THM o REWRITE_RULE[]
                     o CONV_RULE (REWR_CONV AB_THM))
   THEN RES_TAC
   THEN RES_TAC
  );;

%< IF_PROGRESS
g "!a1 a2 q1 q2 b c1 c2 p q r.

           BSPEC(a1, b, q1) /\
           CPROGRESS(q1, c1, p, q, r) /\
           BSPEC(a2, b, q2) /\
           CPROGRESS(q2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(ACOND (AB b) a1 a2, IF b c1 c2, p, q, r)";;

e(REWRITE_TAC[CPROGRESS;BSPEC]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(REWRITE_TAC[A_DEF]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
% 2 subgoals %

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(FIRST_ASSUM (REWRITE_ALL_THM o REWRITE_RULE[]
                o CONV_RULE (REWR_CONV AB_THM)));;
 e(RES_TAC);;
 e(RES_TAC);;

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(FIRST_ASSUM (REWRITE_ALL_THM o REWRITE_RULE[]
                o CONV_RULE (REWR_CONV AB_THM)));;
 e(RES_TAC);;
 e(RES_TAC);;

>%


let IF_PROGRESS1 =
 prove_thm
  (`IF_PROGRESS1`,
   "!b a1 a2 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           CPROGRESS(a2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(ACOND (AB b) (ab_pre b a1) (ab_pre b a2),
                     IF b c1 c2, p, q, r)",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC IF_PROGRESS
   THEN EXISTS_TAC "a1:aexp"
   THEN EXISTS_TAC "a2:aexp"
   THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;

%< IF_PROGRESS1
g "!b a1 a2 c1 c2 p q r.

           CPROGRESS(a1, c1, p, q, r) /\
           CPROGRESS(a2, c2, p, q, r)
       % -------------------------------------------- % ==>
           CPROGRESS(ACOND (AB b) (ab_pre b a1) (ab_pre b a2),
                     IF b c1 c2, p, q, r)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC IF_PROGRESS);;
e(EXISTS_TAC "a1:aexp");;
e(EXISTS_TAC "a2:aexp");;
e(ASM_REWRITE_TAC[ab_pre_THM]);;
>%


let  C_calls_STR_INDUCT_TAC = RULE_INDUCT_THEN C_calls_str_induct
                                   STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;


let WHILE_PROGRESS_LEMMA =
 prove_thm
  (`WHILE_PROGRESS_LEMMA`,
   "!w r s1 p s2.
     C_calls w r s1 p s2 ==>
     (!a pr b c a1.
           BSPEC(AAND a (AB b), b, a1) /\
           CPROGRESS(a1, c, p, q, r) /\
           PCSPEC(a1, c, a, r) /\
           (w = WHILE a pr b c) /\ A a s1
       % -------------------------------------------- % ==>
           A q s2)",

   C_calls_STR_INDUCT_TAC
   THEN REWRITE_ALL_TAC[cmd_distinct;cmd_one_one]
   THENL
    [ REWRITE_TAC[CPROGRESS;BSPEC;A_DEF]
      THEN REPEAT GEN_TAC THEN STRIP_TAC
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(pr:aexp) = pr'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
      THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[]
                         o CONV_RULE (REWR_CONV AB_THM))
      THEN RES_TAC
      THEN RES_TAC;

      REPEAT GEN_TAC THEN STRIP_TAC
      THEN REWRITE_ALL_TAC[cmd_one_one]
      THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[]
      THEN UNDISCH_TAC "BSPEC(AAND a' (AB b'),b',a1)"
      THEN UNDISCH_TAC "CPROGRESS(a1,c',p,q,r)"
      THEN UNDISCH_TAC "PCSPEC(a1,c',a',r)"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(pr:aexp) = pr'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
      THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[]
                         o CONV_RULE (REWR_CONV AB_THM))
      THEN POP_TAC
      THEN REWRITE_TAC[CPROGRESS;BSPEC;PCSPEC;A_DEF]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "s1:state"
      THEN EXISTS_TAC "T"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< WHILE_PROGRESS_LEMMA

g "!w r s1 p s2.
     C_calls w r s1 p s2 ==>
     (!a pr b c a1.
           BSPEC(AAND a (AB b), b, a1) /\
           CPROGRESS(a1, c, p, q, r) /\
           PCSPEC(a1, c, a, r) /\
           (w = WHILE a pr b c) /\ A a s1
       % -------------------------------------------- % ==>
           A q s2)";;

e(C_calls_STR_INDUCT_TAC
  THEN REWRITE_TAC[cmd_distinct;cmd_one_one]);;
% 2 subgoals %

 e(REWRITE_TAC[CPROGRESS;BSPEC;A_DEF]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(pr:aexp) = pr'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
 e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[]
                o CONV_RULE (REWR_CONV AB_THM)));;
 e(RES_TAC);;
 e(RES_TAC);;

 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(REWRITE_ALL_TAC[cmd_one_one]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;
 e(UNDISCH_TAC "BSPEC(AAND a'(AB b'),b',a1)");;
 e(UNDISCH_TAC "CPROGRESS(a1,c',p,q,r)");;
 e(UNDISCH_TAC "PCSPEC(a1,c',a',r)");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(pr:aexp) = pr'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
 e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[]
                o CONV_RULE (REWR_CONV AB_THM)));;
 e(POP_TAC);;
 e(REWRITE_TAC[CPROGRESS;BSPEC;PCSPEC;A_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "s1:state"
   THEN EXISTS_TAC "T");;
 e(ASM_REWRITE_TAC[]);;

>%



let WHILE_PROGRESS =
 prove_thm
  (`WHILE_PROGRESS`,
   "!a a1 pr b c p q r.

           BSPEC(AAND a (AB b), b, a1) /\
           CPROGRESS(a1, c, p, q, r) /\
           PCSPEC(a1, c, a, r)
       % -------------------------------------------- % ==>
           CPROGRESS(a, WHILE a pr b c, p, q, r)",

   REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[CPROGRESS]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN POP_ASSUM (MATCH_MP_TAC o MATCH_MP WHILE_PROGRESS_LEMMA)
   THEN EXISTS_TAC "a:aexp"
   THEN EXISTS_TAC "pr:aexp"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "a1:aexp"
   THEN ASM_REWRITE_TAC[]
  );;

%< WHILE_PROGRESS
g "!a a1 pr b c p q r.

           BSPEC(AAND a (AB b), b, a1) /\
           CPROGRESS(a1, c, p, q, r) /\
           PCSPEC(a1, c, a, r)
       % -------------------------------------------- % ==>
           CPROGRESS(a, WHILE a pr b c, p, q, r)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(REWRITE_TAC[CPROGRESS]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(POP_ASSUM (MATCH_MP_TAC o MATCH_MP WHILE_PROGRESS_LEMMA));;
e(EXISTS_TAC "a:aexp"
  THEN EXISTS_TAC "pr:aexp"
  THEN EXISTS_TAC "b:bexp"
  THEN EXISTS_TAC "c:cmd"
  THEN EXISTS_TAC "a1:aexp");;
e(ASM_REWRITE_TAC[]);;
>%



let BND_SAME_COMMUTE =
 TAC_PROOF
  (([],
    "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)"),
   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:**) = x"
   THENL
    [ ASM_REWRITE_TAC[];

      IMP_RES_TAC BND_COMMUTE
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% BND_SAME_COMMUTE
g "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)";;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:**) = x");;

 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC BND_COMMUTE);;
 e(ASM_REWRITE_TAC[]);;
%




%<
let CPROGRESS_CHANGE_VAR_LEMMA =
 prove_thm
  (`CPROGRESS_CHANGE_VAR_LEMMA`,
   "!a v x b c a0 p q r.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa a0 UNION FVa q))
      ==>
      ( CPROGRESS(a0,                  c, p, q,                  r) =
        CPROGRESS(a0 <|a (AVAR y / x), c, p, q <|a (AVAR y / x), r) )",

   REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM MP_TAC
   THEN DISCH_THEN (\th.
            SUBGOAL_THEN "~WFx y" ASSUME_TAC
            THENL [ ASM_REWRITE_TAC[th;WFx_variant]; ALL_TAC ]
            THEN MP_TAC th)
   THEN DISCH_THEN (\th.
          (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
           SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                UNION FVa a0 UNION FVa q"])
          variant_not_in_set)
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN REWRITE_TAC[CPROGRESS]
   THEN REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]
   THEN EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
    [ FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "BND ((s2:state) y) x (s1:state)"
      THEN IMP_RES_TAC C_calls_BND_THM
      THEN IMP_RES_TAC C_calls_PROG_ONLY_IMPACT
      THEN ASM_REWRITE_TAC[];

      FIRST_ASSUM (MP_TAC o
                   SPECL["BND (s1 x) y (s1:state)";
                         "BND (s2 x) y (s2:state)"])
      THEN REWRITE_TAC[BND_MATCH]
      THEN ONCE_REWRITE_TAC[BND_SAME_COMMUTE]
      THEN REWRITE_TAC[IDENT_BND]
      THEN IMP_RES_THEN REWRITE_THM A_BND_THM
      THEN DISCH_THEN MATCH_MP_TAC
      THEN IMP_RES_TAC C_calls_BND_THM
      THEN IMP_RES_TAC C_calls_PROG_ONLY_IMPACT
      THEN ASM_REWRITE_TAC[]
    ]
  );;
>%

%< CPROGRESS_CHANGE_VAR_LEMMA
g "!a v x b c a0 p q r.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa a0 UNION FVa q))
      ==>
      ( CPROGRESS(a0,                  c, p, q,                  r) =
        CPROGRESS(a0 <|a (AVAR y / x), c, p, q <|a (AVAR y / x), r) )";;
e(REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM MP_TAC);;
e(DISCH_THEN (\th.
    SUBGOAL_THEN "~WFx y" ASSUME_TAC
    THENL [ ASM_REWRITE_TAC[th;WFx_variant]; ALL_TAC ]
    THEN MP_TAC th));;
e(DISCH_THEN (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa a0 UNION FVa q"])
  variant_not_in_set));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(REWRITE_TAC[CPROGRESS]);;
e(REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]);;
e(EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BND ((s2:state) y) x (s1:state)");;
 e(IMP_RES_TAC C_calls_BND_THM);;
 e(IMP_RES_TAC C_calls_PROG_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
 e(FIRST_ASSUM (MP_TAC o
                SPECL["BND (s1 x) y (s1:state)";
                      "BND (s2 x) y (s2:state)"]));;
 e(REWRITE_TAC[BND_MATCH]);;
 e(ONCE_REWRITE_TAC[BND_SAME_COMMUTE]);;
 e(REWRITE_TAC[IDENT_BND]);;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(DISCH_THEN MATCH_MP_TAC);;
 e(IMP_RES_TAC C_calls_BND_THM);;
 e(IMP_RES_TAC C_calls_PROG_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;
>%


let IMP_CONJ_EQ =
 TAC_PROOF
  (([],
    "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))"),
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% IMP_CONJ_EQ
g "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%


let WHILE_PROGRESS_LEMMA1 =
 prove_thm
  (`WHILE_PROGRESS_LEMMA1`,
   "!a v x b c g r a0 p q n.
    (\n.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      ~(x IN FVa q) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
      CPROGRESS(a0, c, p, q, r) /\
      PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r) ==>
      CPROGRESS(AAND a (AEQ v (ANUM n)), WHILE a (ALESS v (AVAR x)) b c,
                p, q, r))
    n",

   REPEAT GEN_TAC
   THEN SPEC_TAC("n:num","n:num")
   THEN MATCH_MP_TAC NUM_WELL_FOUNDED
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th))
   THEN ASM_REWRITE_TAC[]
   THEN DISCH_THEN (ASSUME_TAC o REWRITE_RULE[CPROGRESS])
   THEN UNDISCH_ALL_TAC
   THEN CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o
                     REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM])
   THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[AND_IMP_INTRO]
   THEN REWRITE_TAC[BSPEC;CPROGRESS;PCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o
                   REWRITE_RULE[cmd_distinct; cmd_one_one] o
                   CONV_RULE (REWR_CONV C_calls_cases) )
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o
                     CONV_RULE (REWR_CONV AB_THM))
   THEN IMP_RES_THEN (SUBST1_TAC o SYM o SPECL["n:num";"s2:state"]) A_BND_THM
   THEN POP_TAC
   THENL
    [ FIRST_ASSUM (MATCH_MP_TAC o SPEC "BND n x (s2':state)")
      THEN IMP_RES_TAC (AND2IMP C_calls_BND_THM)
      THEN ASM_REWRITE_TAC[];

      FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "V v (BND n x s3)"
      THEN EXISTS_TAC "BND n x (s3:state)"
      THEN IMP_RES_TAC (AND2IMP C_calls_BND_THM)
      THEN ASM_REWRITE_TAC[]
      THEN ONCE_REWRITE_TAC[CONJ_SYM]
      THEN SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n x (s3:state) x)" REWRITE_THM
           THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "BND n x (s2':state)"
      THEN IMP_RES_TAC (AND2IMP C_BND_THM)
      THEN ASM_REWRITE_TAC[]
    ]
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPECL["BND n x (s1:state)";"T"])
   THEN IMP_RES_THEN REWRITE_THM A_BND_THM
   THEN IMP_RES_THEN REWRITE_THM V_BND_THM
   THEN IMP_RES_TAC B_BND_THM
   THEN ASM_REWRITE_TAC[BND_MATCH]
  );;

%< WHILE_PROGRESS_LEMMA1
g "!a v x b c g r a0 p q n.
    (\n.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      ~(x IN FVa q) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
      CPROGRESS(a0, c, p, q, r) /\
      PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r) ==>
      CPROGRESS(AAND a (AEQ v (ANUM n)), WHILE a (ALESS v (AVAR x)) b c,
                p, q, r))
    n";;

e(REPEAT GEN_TAC);;
e(SPEC_TAC("n:num","n:num"));;
e(MATCH_MP_TAC NUM_WELL_FOUNDED);;
e(BETA_TAC);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th)));;
e(ASM_REWRITE_TAC[]);;
e(DISCH_THEN (ASSUME_TAC o REWRITE_RULE[CPROGRESS]));;
e(UNDISCH_ALL_TAC);;
e(CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM]));;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(REWRITE_TAC[AND_IMP_INTRO]);;
e(REWRITE_TAC[BSPEC;CPROGRESS;PCSPEC;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (STRIP_ASSUME_TAC o
             REWRITE_RULE[cmd_distinct; cmd_one_one] o
             CONV_RULE (REWR_CONV C_calls_cases) ));;

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
 e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
 e(POP_TAC);;
 e(IMP_RES_THEN (SUBST1_TAC o pthm o SYM o SPECL["n:num";"s2:state"]) A_BND_THM);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC "BND n x (s2':state)"));;
 e(IMP_RES_TAC (AND2IMP C_calls_BND_THM));;
 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPECL["BND n x (s1:state)";"T"]));;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
 e(IMP_RES_TAC B_BND_THM);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;

 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
 e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o
                 CONV_RULE (REWR_CONV AB_THM)));;
 e(POP_TAC);;
 e(IMP_RES_THEN (SUBST1_TAC o pthm o SYM o SPECL["n:num";"s2:state"]) A_BND_THM);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "V v (BND n x s3)");;
 e(EXISTS_TAC "BND n x (s3:state)");;
 e(IMP_RES_TAC (AND2IMP C_calls_BND_THM));;
 e(ASM_REWRITE_TAC[]);;
 e(ONCE_REWRITE_TAC[CONJ_SYM]);;
 e(SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n x (s3:state) x)" REWRITE_THM
     THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BND n x (s2':state)");;
 e(IMP_RES_TAC (AND2IMP C_BND_THM));;
 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BND n x (s1:state)");;
 e(EXISTS_TAC "T");;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
 e(IMP_RES_TAC B_BND_THM);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;
>%



let WHILE_PROGRESS1 =
 prove_thm
  (`WHILE_PROGRESS1`,
   "!a a0 v x b c p g q r.

           WFenv_syntax r /\
           WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
           ~(x IN FVa q) /\
           BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
           CPROGRESS(a0, c, p, q, r) /\
           PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r)
       % -------------------------------------------- % ==>
           CPROGRESS(a, WHILE a (ALESS v (AVAR x)) b c, p, q, r)",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP (CONV_RULE (ONCE_DEPTH_CONV BETA_CONV)
                               WHILE_PROGRESS_LEMMA1))
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[CPROGRESS;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPECL["V v s1";"s1:state"])
   THEN ASM_REWRITE_TAC[]
  );;

%< WHILE_PROGRESS1
g "!a a0 v x b c p g q r.

           WFenv_syntax r /\
           WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
           ~(x IN FVa q) /\
           BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
           CPROGRESS(a0, c, p, q, r) /\
           PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r)
       % -------------------------------------------- % ==>
           CPROGRESS(a, WHILE a (ALESS v (AVAR x)) b c, p, q, r)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP (CONV_RULE (ONCE_DEPTH_CONV BETA_CONV)
                             WHILE_PROGRESS_LEMMA1)));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[CPROGRESS;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (MATCH_MP_TAC o SPECL["V v s1";"s1:state"]));;
e(ASM_REWRITE_TAC[]);;
>%



let CALL_NOT_SAME_PROGRESS =
 prove_thm
  (`CALL_NOT_SAME_PROGRESS`,
   "!a1 a2 p1 xs es p r.
      ~(p1 = p) ==>
      CPROGRESS(a1, CALL p1 xs es, p, a2, r)",
   REWRITE_TAC[CPROGRESS]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN REWRITE_THM
  );;

% CALL_NOT_SAME_PROGRESS
g "!a1 a2 p1 xs es p r.
      ~(p1 = p) ==>
      CPROGRESS(a1, CALL p1 xs es, p, a2, r)";;
e(REWRITE_TAC[CPROGRESS]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
%




let CALL_SAME_PROGRESS =
 prove_thm
  (`CALL_SAME_PROGRESS`,
   "!p xs es q r g vars vals glbs pre post calls rec c vals'.
      WFenv_syntax r /\
      WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs)))
      ==>
      CPROGRESS((q <|a (APPEND xs vals' // APPEND vars vals))
                   <|a (vals' :== es),
                CALL p xs es, p, q, r)",
   REWRITE_TAC[CPROGRESS]
   THEN ONCE_REWRITE_TAC[C_calls_cases]
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REWRITE_TAC[WFc;o_THM]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ASSUM_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN REPEAT GEN_TAC
   THEN REWRITE_TAC[a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(xs:var list) = xs'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es:exp list) = es'"
   THEN UNDISCH_LAST_TAC
   THEN ASM_REWRITE_TAC[PAIR_EQ]
   THEN DISCH_THEN (EVERY o (map (REWRITE_ALL_THM o SYM)) o CONJUNCTS)
   THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
   THEN REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
   THEN ASM_REWRITE_TAC[]
  );;

%< CALL_SAME_PROGRESS
g "!p xs es q r g vars vals glbs pre post calls rec c vals' %y%.
      WFenv_syntax r /\
      WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) %/\
      DL y /\
      ((FVa q) SUBSET
       (SL (APPEND (APPEND (APPEND vars vals) glbs) (logicals y))))%
      ==>
      CPROGRESS((q <|a (APPEND xs vals' // APPEND vars vals))
                   <|a (vals' :== es),
                CALL p xs es, p, q, r)";;
e(REWRITE_TAC[CPROGRESS]);;
e(ONCE_REWRITE_TAC[C_calls_cases]);;
e(REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(UNASSUME_THEN (\th. REWRITE_ASSUM_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[a_subst_THM]);;
e(REPEAT STRIP_TAC);;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(xs:var list) = xs'");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es:exp list) = es'");;
e(UNDISCH_LAST_TAC);;
e(ASM_REWRITE_TAC[PAIR_EQ]);;
e(DISCH_THEN (EVERY o (map (REWRITE_ALL_THM o SYM)) o CONJUNCTS));;
e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
e(ASM_REWRITE_TAC[]);;

<--- Old version, based on //v instead of //, follows. --->

e(SUBGOAL_THEN
    "!s. A (q <|a ((APPEND (APPEND (APPEND xs vals') glbs) (logicals y)) //
                   (APPEND (APPEND (APPEND vars vals) glbs) (logicals y)))) s
       = A (q <|a ((APPEND xs vals') // (APPEND vars vals))) s"
      (REWRITE_THM o SYM o SPEC_ALL));;

 e(GEN_TAC);;
 e(USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants]);;

  e(USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants]);;

 e(USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] FVa_vsubst_subst));;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
  e(UNDISCH_TAC "WFxs(APPEND(APPEND vars vals)glbs)");;
  e(REWRITE_TAC[WFxs_APPEND]);;
  e(STRIP_TAC);;
  e(SUBGOAL_THEN "WFxs vals'" ASSUME_TAC);;

   e(ASM_REWRITE_TAC[WFxs_variants]);;

   e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;DL_variants;DL_logicals;
                     DISJOINT_variants_APPEND]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants_APPEND]);;

let st1 = get_state();;
set_state st1;;

  e(USE_IMP_THEN REWRITE_THM APPEND_IDENTITY_vsubst);;

   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants]);;

   e(USE_IMP_THEN REWRITE_THM APPEND_IDENTITY_vsubst);;

    e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants]);;

    e(REWRITE_TAC[a_vsubst_THM]);;
    e(REPEAT STRIP_TAC);;
    e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(xs:var list) = xs'");;
    e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es:exp list) = es'");;
    e(UNDISCH_LAST_TAC);;
    e(ASM_REWRITE_TAC[PAIR_EQ]);;
    e(DISCH_THEN (EVERY o (map (REWRITE_ALL_THM o SYM)) o CONJUNCTS));;
    e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
    e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(ASM_REWRITE_TAC[]);;
>%




let PATH_PROGRESS_pre_NULL_PATH =
 prove_thm
  (`PATH_PROGRESS_pre_NULL_PATH`,
   "!vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2.
      WFenv_pre r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) ==>
      PATH_PROGRESS(pre, p1, [], p2, pre', r)",
   REPEAT STRIP_TAC
   THEN IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) CPROGRESS_PATH_PROGRESS
   THEN UNDISCH_TAC "WFenv_pre r"
   THEN REWRITE_TAC[WFenv_pre]
   THEN DISCH_THEN (MP_TAC o SPEC "p1:string")
   THEN ASM_REWRITE_TAC[WFproc_pre]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[PRE_PROGRESS]
   THEN DISCH_THEN (MP_TAC o SPEC "p2:string")
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

%< PATH_PROGRESS_pre_NULL_PATH
g "!vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2.
      WFenv_pre r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) ==>
      PATH_PROGRESS(pre, p1, [], p2, pre', r)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) CPROGRESS_PATH_PROGRESS);;
e(UNDISCH_TAC "WFenv_pre r");;
e(REWRITE_TAC[WFenv_pre]);;
e(DISCH_THEN (MP_TAC o SPEC "p1:string"));;
e(ASM_REWRITE_TAC[WFproc_pre]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[PRE_PROGRESS]);;
e(DISCH_THEN (MP_TAC o SPEC "p2:string"));;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
>%




let PATH_PROGRESS_pre =
 prove_thm
  (`PATH_PROGRESS_pre`,
   "!ps vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2.
      WFenv_pre r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) ==>
      PATH_PROGRESS(pre, p1, ps, p2, pre', r)",
   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[PATH_PROGRESS_pre_NULL_PATH];

      REPEAT STRIP_TAC
      THEN MATCH_MP_TAC PATH_PROGRESS_CONS
      THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)"
      THEN CONJ_TAC
      THENL
       [ UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"
         THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
         THEN IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre_NULL_PATH)
         THEN DISCH_TAC
         THEN ASM_REWRITE_TAC[];

         UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c"
         THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
         THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "FST((r:env) h)")
         THEN DISCH_TAC
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< PATH_PROGRESS_pre
g "!ps vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2.
      WFenv_pre r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) ==>
      PATH_PROGRESS(pre, p1, ps, p2, pre', r)";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[PATH_PROGRESS_pre_NULL_PATH]);;

 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC PATH_PROGRESS_CONS);;
 e(EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)");;
 e(CONJ_TAC);;

  e(UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'");;
  e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
  e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre_NULL_PATH));;
  e(DISCH_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c");;
  e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
  e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "FST((r:env) h)"));;
  e(DISCH_TAC);;
  e(ASM_REWRITE_TAC[]);;
>%





let call_progress = 
    new_definition (`call_progress`,
      "call_progress p1 p2 q (r:env) =
          let (vars,vals,glbs,pre,post,calls,rec,c) = r p1 in
          let x = APPEND (APPEND vars vals) glbs in
          let x0 = logicals x in
          let x0' = variants x0 (FVa q) in
          let (vars',vals',glbs',pre',post',calls',rec',c') = r p2 in
          let y = APPEND (APPEND vars' vals') glbs' in
          let a = calls p2 in
             ((a = AFALSE) => ATRUE |
               ((mk_foralls y (AIMP (a <|a (x0' // x0)) q)) <|a (x // x0'))
             )");;


let call_path_progress = 
    new_list_rec_definition (`call_path_progress`,
      "(call_path_progress p1 (NIL) p2 q r =
             call_progress p1 p2 q r)  /\
       (call_path_progress p1 (CONS p ps) p2 q r =
             call_progress p1 p (call_path_progress p ps p2 q r) r)");;



let call_path_progress_APPEND =
 prove_thm
  (`call_path_progress_APPEND`,
   "!ps1 ps2 p1 p2 p3 q r.
      call_path_progress p1 (APPEND (APPEND ps1 [p2]) ps2) p3 q r =
           call_path_progress p1 ps1 p2 (call_path_progress p2 ps2 p3 q r) r",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[APPEND;call_path_progress]
  );;

%< call_path_progress_APPEND
g "!ps1 ps2 p1 p2 p3 q r.
      call_path_progress p1 (APPEND (APPEND ps1 [p2]) ps2) p3 q r =
           call_path_progress p1 ps1 p2 (call_path_progress p2 ps2 p3 q r) r";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;call_path_progress]);;

 e(ASM_REWRITE_TAC[APPEND;call_path_progress]);;

>%



let FVa_call_progress =
 prove_thm
  (`FVa_call_progress`,
   "!r vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' p1 p2 q x y z.
      WFenv_syntax r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (x = APPEND (APPEND vars vals) glbs) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      FVa (call_progress p1 p2 q r) SUBSET (SL (APPEND x (logicals z)))",
   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[call_progress]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[(SYM o ASSUME)"(x:var list) = APPEND (APPEND vars vals) glbs"]
   THEN REWRITE_TAC[(SYM o ASSUME)"(y:var list) = APPEND (APPEND vars' vals') glbs'"]
   THEN UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN POP_ASSUM (MP_TAC o SPEC "p2:string")
   THEN DISCH_THEN (\th1. DISCH_TAC THEN MP_TAC th1)
   THEN FIRST_ASSUM REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[(SYM o ASSUME)"(x:var list) = APPEND (APPEND vars vals) glbs"]
   THEN REWRITE_TAC[(SYM o ASSUME)"(y:var list) = APPEND (APPEND vars' vals') glbs'"]
   THEN DISCH_TAC
   THEN COND_CASES_TAC
   THENL
    [ REWRITE_TAC[FVa_DEF;EMPTY_SUBSET];

      REWRITE_TAC[FVa_DEF;FVa_var_list_subst;FVa_mk_foralls;
                  IMAGE_UNION;UNION_SUBSET;UNION_DIFF]
      THEN REWRITE_ALL_TAC[SUBSET_DEF]
      THEN REWRITE_TAC[SL_APPEND]
      THEN REWRITE_TAC[IN_IMAGE;IN_UNION;IN_DIFF]
      THEN REPEAT STRIP_TAC
      THENL  % 2 subgoals %
       [ DISJ1_TAC
         THEN ASM_REWRITE_TAC[]
         THEN RES_TAC
         THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
         THEN REWRITE_TAC[LENGTH_variants;LENGTH_logicals]
         THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
         THEN REWRITE_TAC[LENGTH_variants;LENGTH_logicals]
         THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[SL_APPEND;IN_UNION])
         THENL
          [ UNDISCH_TAC "~(x'':var) IN (SL y)"
            THEN ASM_REWRITE_TAC[]
            THEN SUBGOAL_THEN "WFx x'''" ASSUME_TAC
            THENL
             [ MATCH_MP_TAC WFxs_IS_WFx
               THEN EXISTS_TAC "y:var list"
               THEN ASM_REWRITE_TAC[]
               THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX;

               USE_IMP_AND_THEN REWRITE_THM BNDsR_NOT_MATCH
                       (IMP_RES_THEN REWRITE_THM WFx_NOT_logicals)
               THEN REWRITE_TAC[I_THM]
               THEN UNDISCH_TAC "(x''':var) IN SL y"
               THEN ASM_REWRITE_TAC[]
               THEN DISCH_THEN REWRITE_THM
             ];

            POP_ASSUM MP_TAC
            THEN ASM_REWRITE_TAC[]
          ];

         DISJ2_TAC
         THEN RES_TAC
         THEN UNDISCH_LAST_TAC
         THEN REWRITE_TAC[SL_APPEND;IN_UNION]
         THEN ASM_REWRITE_TAC[]
         THEN DISCH_TAC
         THEN USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH
         THENL
          [ MP_TAC (SPECL["logicals x";"FVa q"] DISJOINT_variants)
            THEN ASM_REWRITE_TAC[]
            THEN DISCH_TAC
            THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_IMP;

            ASM_REWRITE_TAC[I_THM]
          ]
       ]
    ]
  );;


%< FVa_call_progress

g "!r vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' p1 p2 q x y z.
      WFenv_syntax r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (x = APPEND (APPEND vars vals) glbs) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      FVa (call_progress p1 p2 q r) SUBSET (SL (APPEND x (logicals z)))";;

e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[call_progress]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[(SYM o ASSUME)"(x:var list) = APPEND (APPEND vars vals) glbs"]);;
e(REWRITE_TAC[(SYM o ASSUME)"(y:var list) = APPEND (APPEND vars' vals') glbs'"]);;
e(UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'");;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(POP_ASSUM (MP_TAC o SPEC "p2:string"));;
e(DISCH_THEN (\th1. DISCH_TAC THEN MP_TAC th1));;
e(FIRST_ASSUM REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[(SYM o ASSUME)"(x:var list) = APPEND (APPEND vars vals) glbs"]);;
e(REWRITE_TAC[(SYM o ASSUME)"(y:var list) = APPEND (APPEND vars' vals') glbs'"]);;
e(DISCH_TAC);;
e(COND_CASES_TAC);; % two cases %

 e(REWRITE_TAC[FVa_DEF;EMPTY_SUBSET]);;

let st1 = get_state();;
set_state st1;;
 e(REWRITE_TAC[FVa_DEF;FVa_var_list_subst;FVa_mk_foralls;
                  IMAGE_UNION;UNION_SUBSET;UNION_DIFF]);;
 e(REWRITE_ALL_TAC[SUBSET_DEF]);;
 e(REWRITE_TAC[SL_APPEND]);;
 e(REWRITE_TAC[IN_IMAGE;IN_UNION;IN_DIFF]);;
 e(REPEAT STRIP_TAC);;  % 2 subgoals %

let st2 = get_state();;
set_state st2;;
  e(DISJ1_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(RES_TAC);;
  e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
  e(REWRITE_TAC[LENGTH_variants;LENGTH_logicals]);;
  e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
  e(REWRITE_TAC[LENGTH_variants;LENGTH_logicals]);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[SL_APPEND;IN_UNION]));;

   e(UNDISCH_TAC "~(x'':var) IN (SL y)");;
   e(ASM_REWRITE_TAC[]);;
   e(SUBGOAL_THEN "WFx x'''" ASSUME_TAC);;

    e(MATCH_MP_TAC WFxs_IS_WFx);;
    e(EXISTS_TAC "y:var list");;
    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

    e(USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH);;

     e(IMP_RES_THEN REWRITE_THM WFx_NOT_logicals);;

     e(REWRITE_TAC[I_THM]);;
     e(UNDISCH_TAC "(x''':var) IN SL y");;
     e(ASM_REWRITE_TAC[]);;
     e(DISCH_THEN REWRITE_THM);;

   e(POP_ASSUM MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

let st3 = get_state();;
set_state st3;;
  e(DISJ2_TAC);;
  e(RES_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
  e(ASM_REWRITE_TAC[]);;
  e(DISCH_TAC);;
  e(USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH);;

   e(MP_TAC (SPECL["logicals x";"FVa q"] DISJOINT_variants));;
   e(ASM_REWRITE_TAC[]);;
   e(DISCH_TAC);;
   e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_IMP);;

   e(ASM_REWRITE_TAC[I_THM]);;
>%





let FVa_call_path_progress =
 prove_thm
  (`FVa_call_path_progress`,
   "!ps r vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' p1 p2 q x y z.
      WFenv_syntax r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (x = APPEND (APPEND vars vals) glbs) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      FVa (call_path_progress p1 ps p2 q r) SUBSET (SL (APPEND x (logicals z)))",
   LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[call_path_progress]
   THENL
    [ MATCH_MP_TAC (SPEC_ALL FVa_call_progress)
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC FVa_call_progress
      THEN EXISTS_TAC "vars:var list"
      THEN EXISTS_TAC "vals:var list"
      THEN EXISTS_TAC "glbs:var list"
      THEN EXISTS_TAC "pre:aexp"
      THEN EXISTS_TAC "post:aexp"
      THEN EXISTS_TAC "calls:prog_env"
      THEN EXISTS_TAC "rec:aexp"
      THEN EXISTS_TAC "c:cmd"
      THEN EXISTS_TAC "(FST)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "APPEND (APPEND ((FST)((r:env) h)) ((FST o SND)(r h)))
                              ((FST o SND o SND)(r h))"
      THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "(FST)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "vars':var list"
      THEN EXISTS_TAC "vals':var list"
      THEN EXISTS_TAC "glbs':var list"
      THEN EXISTS_TAC "pre':aexp"
      THEN EXISTS_TAC "post':aexp"
      THEN EXISTS_TAC "calls':prog_env"
      THEN EXISTS_TAC "rec':aexp"
      THEN EXISTS_TAC "c':cmd"
      THEN EXISTS_TAC "APPEND (APPEND vars' vals') glbs' :var list"
      THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
      THEN UNDISCH_LAST_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< FVa_call_path_progress

g "!ps r vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' p1 p2 q x y z.
      WFenv_syntax r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (x = APPEND (APPEND vars vals) glbs) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      FVa (call_path_progress p1 ps p2 q r) SUBSET (SL (APPEND x (logicals z)))";;

e(LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[call_path_progress]);;
 e(MATCH_MP_TAC (SPEC_ALL FVa_call_progress));;
 e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[call_path_progress]);;
 e(MATCH_MP_TAC FVa_call_progress);;
 e(EXISTS_TAC "vars:var list"
      THEN EXISTS_TAC "vals:var list"
      THEN EXISTS_TAC "glbs:var list"
      THEN EXISTS_TAC "pre:aexp"
      THEN EXISTS_TAC "post:aexp"
      THEN EXISTS_TAC "calls:prog_env"
      THEN EXISTS_TAC "rec:aexp"
      THEN EXISTS_TAC "c:cmd"
      THEN EXISTS_TAC "(FST)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "APPEND (APPEND ((FST)((r:env) h)) ((FST o SND)(r h)))
                              ((FST o SND o SND)(r h))");;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "(FST)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) h)"
      THEN EXISTS_TAC "vars':var list"
      THEN EXISTS_TAC "vals':var list"
      THEN EXISTS_TAC "glbs':var list"
      THEN EXISTS_TAC "pre':aexp"
      THEN EXISTS_TAC "post':aexp"
      THEN EXISTS_TAC "calls':prog_env"
      THEN EXISTS_TAC "rec':aexp"
      THEN EXISTS_TAC "c':cmd"
      THEN EXISTS_TAC "APPEND (APPEND vars' vals') glbs' :var list");;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
 e(UNDISCH_LAST_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%




let CPROGRESS_call_progress =
 prove_thm
  (`CPROGRESS_call_progress`,
   "!vars vals glbs pre post calls rec c r p1 p2 q
       vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      CPROGRESS(AAND pre (call_progress p1 p2 q r), c, p2, q, r)",

   REPEAT STRIP_TAC
   THEN UNDISCH_TAC "WFenv_calls r"
   THEN REWRITE_TAC[WFenv_calls;WFproc_calls]
   THEN DISCH_THEN (MP_TAC o SPEC "p1:string")
   THEN ASM_REWRITE_TAC[BODY_PROGRESS;CPROGRESS;A_DEF;call_progress]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN DEFINE_NEW_VAR "x = APPEND (APPEND vars vals) glbs :var list"
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN (REWRITE_THM o SYM o ASSUME)
                      "(y:var list) = APPEND(APPEND vars' vals')glbs'"
   THEN DISCH_TAC
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN ((uncurry $THEN) o (MP_TAC # ASSUME_TAC) o CONJ_PAIR)
   THEN POP_ASSUM (\th1. POP_ASSUM (\th2. ASSUME_TAC th1 THEN ASSUME_TAC th2))
   THEN POP_ASSUM
           (MP_TAC o SPECL["p2:string";
                           "BNDsL (MAP s1 x) (logicals x) (s1:state)";
                           "BNDsL (MAP s1 x) (logicals x) (s2:state)"])
   THEN UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN DISCH_TAC
   THEN USE_IMP_AND_THEN REWRITE_THM A_mk_eqs_THM
           (REWRITE_TAC[LENGTH_logicals])
   THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
           (ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals])
   THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
           (MATCH_MP_TAC WFxs_NOT_logicals
            THEN ASM_REWRITE_TAC[])
   THEN USE_IMP_AND_THEN REWRITE_THM A_BNDsL_THM
           (ONCE_REWRITE_TAC[DISJOINT_SYM]
            THEN MATCH_MP_TAC DISJOINT_SUBSET
            THEN EXISTS_TAC "SL x:var set"
            THEN IMP_RES_THEN ASM_REWRITE_THM WFxs_NOT_logicals)
   THEN USE_AND_THEN REWRITE_THM C_calls_BNDsL_THM
           (EXISTS_TAC "calls:prog_env"
            THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
   THEN COND_CASES_TAC
   THENL
    [ POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[A_DEF]
      THEN DISCH_THEN REWRITE_THM;

      POP_TAC
      THEN ONCE_REWRITE_TAC[a_subst_THM]
      THEN REWRITE_TAC[A_mk_foralls]
      THEN REPEAT STRIP_TAC
      THEN POP_ASSUM (MP_TAC o SPEC "MAP (s2:state) y")
      THEN REWRITE_TAC[LENGTH_MAP]
      THEN RES_TAC
      THEN DEFINE_NEW_VAR "x0' = variants (logicals x) (FVa q)"
      THEN FIRST_ASSUM (REWRITE_THM o SYM)
      THEN REWRITE_TAC[A_DEF;a_subst_THM;s_subst_var_list]
      THEN SUBGOAL_THEN
           "A ((calls:prog_env) p2)
              (BNDsR
                (MAP (BNDsL (MAP s2 y) y (BNDsR (MAP s1 x) x0' s1)) x0')
                (logicals x)
                (BNDsL (MAP s2 y) y (BNDsR (MAP s1 x) x0' s1))) =
            A (calls p2)
              (BNDsL (MAP s1 x) (logicals x) s2)"
           REWRITE_THM
      THENL
       [ MATCH_MP_TAC A_EQ_THM
         THEN MATCH_MP_TAC MAP_EXTENSION_SUBSET
         THEN EXISTS_TAC "APPEND y (logicals x):var list"
         THEN CONJ_TAC
         THENL
          [ REWRITE_TAC[MAP_APPEND]
            THEN USE_IMP_AND_THEN REWRITE_THM APPEND_11
                    (REWRITE_TAC[LENGTH_MAP])
            THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
            THEN USE_IMP_AND_THEN REWRITE_THM
                            (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
            THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
            THEN USE_IMP_AND_THEN REWRITE_THM
                            (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                     LENGTH_logicals;DL_logicals])
            THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (ONCE_REWRITE_TAC[DISJOINT_SYM]
                     THEN MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
            THEN USE_IMP_AND_THEN REWRITE_THM
                            (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                     LENGTH_logicals;DL_variants])
            THEN USE_IMP_THEN REWRITE_THM
                        (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT)
            THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_logicals;DL_logicals];

            FIRST_ASSUM (MP_TAC o SPEC "p2:string")
            THEN REWRITE_TAC[ASSUME
                       "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"]
            THEN CONV_TAC (ONCE_DEPTH_CONV let_CONV)
            THEN ASM_REWRITE_TAC[]
          ];

         REWRITE_TAC[ASSUME "A((calls:prog_env) p2)(BNDsL(MAP s1 x)(logicals x)s2)"]
         THEN SUBGOAL_THEN
                    "A q (BNDsL(MAP s2 y)y(BNDsR(MAP s1 x)x0' s1)) = A q s2"
                 REWRITE_THM
         THEN MATCH_MP_TAC A_EQ_THM
         THEN REPEAT STRIP_TAC
         THEN ASM_CASES_TAC "(x':var) IN SL y"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH
            THENL
             [ REWRITE_TAC[ASSUME "x0' = variants(logicals x)(FVa q)"]
               THEN ASSUME_TAC (SPECL["logicals x";"FVa q"] DISJOINT_variants)
               THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_IMP;

               IMP_RES2_THEN MP_TAC SUBSET_DEF
               THEN REWRITE_TAC[SL_APPEND;IN_UNION]
               THEN ASM_REWRITE_TAC[]
               THEN DISCH_THEN (ASSUME_TAC o MATCH_MP logicals_IS_NOT_WFx)
               THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
               THEN IMP_RES_TAC (AND2IMP C_calls_PROG_ONLY_IMPACT)
               THEN ASM_REWRITE_TAC[]
             ]
          ]
       ]
    ]
  );;

%< CPROGRESS_call_progress
g "!vars vals glbs pre post calls rec c r p1 p2 q
       vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      CPROGRESS(AAND pre (call_progress p1 p2 q r), c, p2, q, r)";;

e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "WFenv_calls r");;
e(REWRITE_TAC[WFenv_calls;WFproc_calls]);;
e(DISCH_THEN (MP_TAC o SPEC "p1:string"));;
e(ASM_REWRITE_TAC[BODY_PROGRESS;CPROGRESS;A_DEF;call_progress]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(DEFINE_NEW_VAR "x = APPEND (APPEND vars vals) glbs :var list");;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e((REWRITE_THM o SYM o ASSUME)
                "(y:var list) = APPEND(APPEND vars' vals')glbs'");;
e(DISCH_TAC);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN ((uncurry $THEN) o (MP_TAC # ASSUME_TAC) o CONJ_PAIR));;
e(POP_ASSUM (\th1. (POP_ASSUM (\th2. ASSUME_TAC th1 THEN ASSUME_TAC th2))));;
e(POP_ASSUM
     (MP_TAC o SPECL["p2:string";
                     "BNDsL (MAP s1 x) (logicals x) (s1:state)";
                     "BNDsL (MAP s1 x) (logicals x) (s2:state)"]));;
e(UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'");;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(DISCH_TAC);;
e(USE_IMP_THEN REWRITE_THM A_mk_eqs_THM);;

  e(REWRITE_TAC[LENGTH_logicals]);;

 e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT));;

  e(ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals]);;

 e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

  e(MATCH_MP_TAC WFxs_NOT_logicals);;
  e(ASM_REWRITE_TAC[]);;

 e(USE_IMP_THEN REWRITE_THM A_BNDsL_THM);;

  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(MATCH_MP_TAC DISJOINT_SUBSET);;
  e(EXISTS_TAC "SL x:var set");;
  e(IMP_RES_THEN ASM_REWRITE_THM WFxs_NOT_logicals);;

 e(USE_THEN REWRITE_THM C_calls_BNDsL_THM);;

  e(EXISTS_TAC "calls:prog_env");;
  e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

e(COND_CASES_TAC);; % two cases %

 e(POP_ASSUM REWRITE_THM);;
 e(REWRITE_TAC[A_DEF]);;
 e(DISCH_THEN REWRITE_THM);;

 e(POP_TAC);;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REWRITE_TAC[A_mk_foralls]);;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (MP_TAC o SPEC "MAP (s2:state) y"));;
 e(REWRITE_TAC[LENGTH_MAP]);;
 e(RES_TAC);;
 e(DEFINE_NEW_VAR "x0' = variants (logicals x) (FVa q)");;
 e(FIRST_ASSUM (REWRITE_THM o SYM));;
 e(REWRITE_TAC[A_DEF;a_subst_THM;s_subst_var_list]);;
 e(SUBGOAL_THEN
     "A ((calls:prog_env) p2)
        (BNDsR
          (MAP (BNDsL (MAP s2 y) y (BNDsR (MAP s1 x) x0' s1)) x0')
          (logicals x)
          (BNDsL (MAP s2 y) y (BNDsR (MAP s1 x) x0' s1))) =
      A (calls p2)
        (BNDsL (MAP s1 x) (logicals x) s2)"
     REWRITE_THM);;

  e(MATCH_MP_TAC A_EQ_THM);;
  e(MATCH_MP_TAC MAP_EXTENSION_SUBSET);;
  e(EXISTS_TAC "APPEND y (logicals x):var list");;
  e(CONJ_TAC);;

   e(REWRITE_TAC[MAP_APPEND]);;
   e(USE_IMP_THEN REWRITE_THM APPEND_11);;

    e(REWRITE_TAC[LENGTH_MAP]);;

   e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

    e(MATCH_MP_TAC WFxs_NOT_WFxs);;
    e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT));;

    e(ASM_REWRITE_TAC[LENGTH_MAP]);;
    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

   e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

    e(MATCH_MP_TAC WFxs_NOT_WFxs);;
    e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT));;

    e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;LENGTH_logicals;
                      DL_logicals]);;

   e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

    e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
    e(MATCH_MP_TAC WFxs_NOT_WFxs);;
    e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;
    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT));;

    e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;LENGTH_logicals;DL_variants]);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsL_IDENT));;
   e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_logicals;DL_logicals]);;

  e(FIRST_ASSUM (MP_TAC o SPEC "p2:string"));;
  e(REWRITE_TAC[ASSUME "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"]);;
  e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[ASSUME "A((calls:prog_env) p2)(BNDsL(MAP s1 x)(logicals x)s2)"]);;
let st1 = get_state();;
set_state st1;;
 e(SUBGOAL_THEN
     "A q (BNDsL(MAP s2 y)y(BNDsR(MAP s1 x)x0' s1)) = A q s2"
     REWRITE_THM);;
 e(MATCH_MP_TAC A_EQ_THM);;
 e(REPEAT STRIP_TAC);;
 e(ASM_CASES_TAC "(x':var) IN SL y");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(USE_IMP_THEN REWRITE_THM BNDsR_NOT_MATCH);;

   e(REWRITE_TAC[ASSUME "x0' = variants(logicals x)(FVa q)"]);;
   e(ASSUME_TAC (SPECL["logicals x";"FVa q"] DISJOINT_variants));;
   e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_IMP);;

   e(IMP_RES2_THEN MP_TAC SUBSET_DEF);;
   e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
   e(ASM_REWRITE_TAC[]);;
   e(DISCH_THEN (ASSUME_TAC o MATCH_MP logicals_IS_NOT_WFx));;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(IMP_RES_TAC (AND2IMP C_calls_PROG_ONLY_IMPACT));;
   e(ASM_REWRITE_TAC[]);;

>%





let PATH_PROGRESS_call_progress =
 prove_thm
  (`PATH_PROGRESS_call_progress`,
   "!vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2 q y z.
      WFenv_syntax r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      PATH_PROGRESS(AAND pre (call_progress p1 p2 q r), p1, [], p2, q, r)",
   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN
     "CPROGRESS(AAND pre (call_progress p1 p2 q r),c,p2,q,r) =
      PATH_PROGRESS(AAND pre (call_progress p1 p2 q r),p1,[],p2,q,r)"
    (REWRITE_THM o SYM)
   THENL
    [ MATCH_MP_TAC CPROGRESS_PATH_PROGRESS
      THEN EXISTS_TAC "vars:var list"
      THEN EXISTS_TAC "vals:var list"
      THEN EXISTS_TAC "glbs:var list"
      THEN EXISTS_TAC "pre:aexp"
      THEN EXISTS_TAC "post:aexp"
      THEN EXISTS_TAC "calls:prog_env"
      THEN EXISTS_TAC "rec:aexp"
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC (AND2IMP CPROGRESS_call_progress)
    ]
  );;

%< PATH_PROGRESS_call_progress
g "!vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2 q y z.
      WFenv_syntax r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      PATH_PROGRESS(AAND pre (call_progress p1 p2 q r), p1, [], p2, q, r)";;
e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN
     "CPROGRESS(AAND pre (call_progress p1 p2 q r),c,p2,q,r) =
      PATH_PROGRESS(AAND pre (call_progress p1 p2 q r),p1,[],p2,q,r)"
    (REWRITE_THM o SYM));;

 e(MATCH_MP_TAC CPROGRESS_PATH_PROGRESS);;
 e(EXISTS_TAC "vars:var list"
    THEN EXISTS_TAC "vals:var list"
    THEN EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC (AND2IMP CPROGRESS_call_progress));;
>%




let PATH_PROGRESS_call_path_progress =
 prove_thm
  (`PATH_PROGRESS_call_path_progress`,
   "!ps vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2 q y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      PATH_PROGRESS(AAND pre (call_path_progress p1 ps p2 q r),
                    p1, ps, p2, q, r)",

   LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[call_path_progress]
   THENL
    [ MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_call_progress)
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC PATH_PROGRESS_CONS
      THEN EXISTS_TAC "AAND ((FST o SND o SND o SND)(r h))
                            (call_path_progress h ps p2 q r)"
      THEN CONJ_TAC
      THENL
       [ MATCH_MP_TAC POST_AAND_PATH_PROGRESS
         THEN CONJ_TAC
         THENL
          [ MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS
            THEN EXISTS_TAC "pre:aexp"
            THEN CONJ_TAC
            THENL
             [ REWRITE_TAC[CLOSE;A_DEF]
               THEN REPEAT STRIP_TAC;

               UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"
               THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
               THEN IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre_NULL_PATH)
               THEN DISCH_TAC
               THEN ASM_REWRITE_TAC[]
             ];

            UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c"
            THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
            THEN DEFINE_NEW_VAR "(x:var list) =
                     APPEND (APPEND (FST((r:env) h)) ((FST o SND)(r h)))
                            ((FST o (SND o SND))(r h))"
            THEN IMP_RES_TAC (AND2IMP FVa_call_path_progress)
            THEN POP_TAC THEN POP_TAC
            THEN POP_ASSUM (\th. POP_TAC THEN ASSUME_TAC (SPEC_ALL th))
            THEN DISCH_TAC
            THEN UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'"
            THEN IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress)
            THEN DISCH_TAC
            THEN ASM_REWRITE_TAC[]
          ];

         UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c"
         THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
         THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "FST((r:env) h)")
         THEN DISCH_TAC
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< PATH_PROGRESS_call_path_progress
g "!ps vars vals glbs pre post calls rec c
       vars' vals' glbs' pre' post' calls' rec' c' r p1 p2 q y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p1 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (r p2 = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET (SL (APPEND y (logicals z)))) ==>
      PATH_PROGRESS(AAND pre (call_path_progress p1 ps p2 q r),
                    p1, ps, p2, q, r)";;

e(LIST_INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[call_path_progress]);;
 e(MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_call_progress));;
 e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[call_path_progress]);;
 e(MATCH_MP_TAC PATH_PROGRESS_CONS);;
 e(EXISTS_TAC "AAND ((FST o SND o SND o SND)(r h))
                    (call_path_progress h ps p2 q r)");;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC POST_AAND_PATH_PROGRESS);;
  e(CONJ_TAC);;

   e(MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS);;
   e(EXISTS_TAC "pre:aexp");;
   e(CONJ_TAC);;

    e(REWRITE_TAC[CLOSE;A_DEF]);;
    e(REPEAT STRIP_TAC);;

    e(UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'");;
    e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
    e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre_NULL_PATH));;
    e(DISCH_TAC);;
    e(ASM_REWRITE_TAC[]);;

   e(UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c");;
   e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
   e(DEFINE_NEW_VAR "(x:var list) =
                     APPEND (APPEND (FST((r:env) h)) ((FST o SND)(r h)))
                            ((FST o (SND o SND))(r h))");;
   e(IMP_RES_TAC (AND2IMP FVa_call_path_progress));;
   e(POP_TAC THEN POP_TAC);;
   e(POP_ASSUM (\th. POP_TAC THEN ASSUME_TAC (SPEC_ALL th)));;
   e(DISCH_TAC);;
   e(UNDISCH_TAC "(r:env) p2 = vars',vals',glbs',pre',post',calls',rec',c'");;
   e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress));;
   e(DISCH_TAC);;
   e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;

  e(UNDISCH_TAC "(r:env) p1 = vars,vals,glbs,pre,post,calls,rec,c");;
  e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
  e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "FST((r:env) h)"));;
  e(DISCH_TAC);;
  e(ASM_REWRITE_TAC[]);;
>%



close_theory ();;

