% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of semantic stages, where each stage
  allows one more depth of procedure call than the stage before.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `semantic_stages`;;

new_parent `progress_rules`;;


% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of semantic_stages %

loadf `autoload`;;

autoload_all_defs_and_thms
     `bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `progress_tactics`;;
loadf `new_ind_def`;;



% =============================================================== %
% Define a staged version of the semantic relation for commands.  %
% =============================================================== %


let Ck_rules,Ck_ind,Ck_cases,Ck_str_induct =
    let Ck = "Ck : cmd -> env -> num -> state -> state -> bool" in
    new_ind_def `Ck`

    ("^Ck c r k s1 s2", []) 
    
     [ [                                             
       % -------------------------------------------- % ],
                     "^Ck (SKIP) r k s s"             ;


% Note: no rule for ABORT, since there are no elements in that relation. %

    
       [                 "E e s1 n s2"
       % -------------------------------------------- % ],
            "^Ck (ASSIGN x e) r k s1 (BND n x s2)"    ;


       [  "^Ck c1 r k1 s1 s2";    "^Ck c2 r k2 s2 s3";    
       % -------------------------------------------- % "k1 <= k";"k2 <= k"],
                  "^Ck (SEQ c1 c2) r k s1 s3"         ;


       [              "^Ck c1 r k s2 s3";
       % -------------------------------------------- % "B b s1 T s2" ],
                 "^Ck (IF b c1 c2) r k s1 s3"         ;


       [              "^Ck c2 r k s2 s3";
       % -------------------------------------------- % "B b s1 F s2" ],
                 "^Ck (IF b c1 c2) r k s1 s3"         ;


       [               "^Ck c r k1 s2 s3";
               "^Ck (WHILE a pr b c) r k2 s3 s4";       "k1 <= k";"k2 <= k";
       % -------------------------------------------- % "B b s1 T s2" ],
               "^Ck (WHILE a pr b c) r k s1 s4"       ;


       [                                             
       % -------------------------------------------- % "B b s1 F s2" ],
               "^Ck (WHILE a pr b c) r k s1 s2"       ;


       ["^Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
            r k (BNDsL ns vals' s2) s3";     
        "(r:env) p = (vars, vals, glbs, pre, post, calls, rec, c)";
        "vals' = variants vals (SL (APPEND xs glbs))";
       % -------------------------------------------- % "ES es s1 ns s2" ],
       "^Ck (CALL p xs es) r (k+1) s1 (BNDsL (MAP s2 vals') vals' s3)" ];;


% --------------------------------------------------------------------- %
% Tactics for the rules.						%
% --------------------------------------------------------------------- %

let CMDk_TAC = FIRST (map RULE_TAC Ck_rules);;
let [SKIPk_TAC;ASSIGNk_TAC;SEQk_TAC;IFk_T_TAC;IFk_F_TAC;
     WHILEk_T_TAC;WHILEk_F_TAC;CALLk_TAC] = map RULE_TAC Ck_rules;;

% --------------------------------------------------------------------- %
% Rule induction tactics.                                               %
% --------------------------------------------------------------------- %

let  CMDk_INDUCT_TAC = RULE_INDUCT_THEN Ck_ind
                              STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;

let  CMDk_STR_INDUCT_TAC = mk_str_ind_tac Ck_str_induct;;



let C_IMP_Ck =
 prove_thm
  (`C_IMP_Ck`,
   "!c r s1 s2. C c r s1 s2 ==> ?k. Ck c r k s1 s2",
   CMD_INDUCT_TAC
   THEN REPEAT GEN_TAC
   THENL
     [  EXISTS_TAC "0"
        THEN SKIPk_TAC;

        EXISTS_TAC "0"
        THEN ASSIGNk_TAC
        THEN ASM_REWRITE_TAC[];

        EXISTS_TAC "k + k'"
        THEN SEQk_TAC
        THEN ASM_REWRITE_TAC[]
        THEN EXISTS_TAC "k:num"
        THEN EXISTS_TAC "s2:state"
        THEN EXISTS_TAC "k':num"
        THEN ASM_REWRITE_TAC[LESS_EQ_ADD]
        THEN ONCE_REWRITE_TAC[ADD_SYM]
        THEN REWRITE_TAC[LESS_EQ_ADD];

        EXISTS_TAC "k:num"
        THEN IFk_T_TAC
        THEN EXISTS_TAC "s2:state"
        THEN ASM_REWRITE_TAC[];

        EXISTS_TAC "k:num"
        THEN IFk_F_TAC
        THEN EXISTS_TAC "s2:state"
        THEN ASM_REWRITE_TAC[];

        EXISTS_TAC "k + k'"
        THEN WHILEk_T_TAC
        THEN EXISTS_TAC "k:num"
        THEN EXISTS_TAC "s2:state"
        THEN EXISTS_TAC "s3:state"
        THEN EXISTS_TAC "k':num"
        THEN ASM_REWRITE_TAC[LESS_EQ_ADD]
        THEN ONCE_REWRITE_TAC[ADD_SYM]
        THEN REWRITE_TAC[LESS_EQ_ADD];

        EXISTS_TAC "k:num"
        THEN WHILEk_F_TAC
        THEN ASM_REWRITE_TAC[];

        EXISTS_TAC "k + 1"
        THEN CALLk_TAC
        THEN EXISTS_TAC "c:cmd"
        THEN EXISTS_TAC "vars:(var)list"
        THEN EXISTS_TAC "vals:(var)list"
        THEN EXISTS_TAC "ns:(num)list"
        THEN EXISTS_TAC "glbs:(var)list"
        THEN EXISTS_TAC "pre:aexp"
        THEN EXISTS_TAC "post:aexp"
        THEN EXISTS_TAC "calls:prog_env"
        THEN EXISTS_TAC "rec:aexp"
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%< C_IMP_Ck
g "!c r s1 s2. C c r s1 s2 ==> ?k. Ck c r k s1 s2";;
e(CMD_INDUCT_TAC);;

% SKIP: %
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "0");;
e(SKIPk_TAC);;

% ASSIGN: %
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "0");;
e(ASSIGNk_TAC);;
e(ASM_REWRITE_TAC[]);;

% SEQ: %
e(EXISTS_TAC "k + k'");;
e(SEQk_TAC);;
e(ASM_REWRITE_TAC[]);;
e(EXISTS_TAC "k:num" THEN
  EXISTS_TAC "s2:state" THEN
  EXISTS_TAC "k':num");;
e(ASM_REWRITE_TAC[LESS_EQ_ADD]);;
e(ONCE_REWRITE_TAC[ADD_SYM]);;
e(REWRITE_TAC[LESS_EQ_ADD]);;

% IF_T: %
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "k:num");;
e(IFk_T_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;

% IF_F: %
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "k:num");;
e(IFk_F_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;

% WHILE_T: %
e(EXISTS_TAC "k + k'");;
e(WHILEk_T_TAC);;
e(EXISTS_TAC "k:num" THEN
  EXISTS_TAC "s2:state" THEN
  EXISTS_TAC "s3:state" THEN
  EXISTS_TAC "k':num");;
e(ASM_REWRITE_TAC[LESS_EQ_ADD]);;
e(ONCE_REWRITE_TAC[ADD_SYM]);;
e(REWRITE_TAC[LESS_EQ_ADD]);;

% WHILE_F: %
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "k:num");;
e(WHILEk_F_TAC);;
e(ASM_REWRITE_TAC[]);;

% CALL: %
e(EXISTS_TAC "k + 1");;
e(CALLk_TAC);;
e(EXISTS_TAC "c:cmd" THEN
  EXISTS_TAC "vars:(var)list" THEN
  EXISTS_TAC "vals:(var)list" THEN
  EXISTS_TAC "ns:(num)list" THEN
  EXISTS_TAC "glbs:(var)list" THEN
  EXISTS_TAC "pre:aexp" THEN
  EXISTS_TAC "post:aexp" THEN
  EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "rec:aexp");;
e(ASM_REWRITE_TAC[]);;

>%


let Ck_IMP_C =
 prove_thm
  (`Ck_IMP_C`,
   "!c r k s1 s2. Ck c r k s1 s2 ==> C c r s1 s2",
   CMDk_INDUCT_TAC
   THENL
     [  SKIP_TAC;

        ASSIGN_TAC
        THEN ASM_REWRITE_TAC[];

        SEQ_TAC
        THEN EXISTS_TAC "s2:state"
        THEN ASM_REWRITE_TAC[];

        IF_T_TAC
        THEN EXISTS_TAC "s2:state"
        THEN ASM_REWRITE_TAC[];

        IF_F_TAC
        THEN EXISTS_TAC "s2:state"
        THEN ASM_REWRITE_TAC[];

        WHILE_T_TAC
        THEN EXISTS_TAC "s2:state"
        THEN EXISTS_TAC "s3:state"
        THEN ASM_REWRITE_TAC[];

        WHILE_F_TAC
        THEN ASM_REWRITE_TAC[];

        CALL_TAC
        THEN EXISTS_TAC "c:cmd"
        THEN EXISTS_TAC "vars:(var)list"
        THEN EXISTS_TAC "vals:(var)list"
        THEN EXISTS_TAC "ns:(num)list"
        THEN EXISTS_TAC "glbs:(var)list"
        THEN EXISTS_TAC "pre:aexp"
        THEN EXISTS_TAC "post:aexp"
        THEN EXISTS_TAC "calls:prog_env"
        THEN EXISTS_TAC "rec:aexp"
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%< Ck_IMP_C
g "!c r k s1 s2. Ck c r k s1 s2 ==> C c r s1 s2";;
e(CMDk_INDUCT_TAC);;

% SKIP: %
e(SKIP_TAC);;

% ASSIGN: %
e(ASSIGN_TAC);;
e(ASM_REWRITE_TAC[]);;

% SEQ: %
e(SEQ_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;

% IF_T: %
e(IF_T_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;

% IF_F: %
e(IF_F_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;

% WHILE_T: %
e(WHILE_T_TAC);;
e(EXISTS_TAC "s2:state"
  THEN EXISTS_TAC "s3:state");;
e(ASM_REWRITE_TAC[]);;

% WHILE_F: %
e(WHILE_F_TAC);;
e(ASM_REWRITE_TAC[]);;

% CALL: %
e(CALL_TAC);;
e(EXISTS_TAC "c:cmd" THEN
  EXISTS_TAC "vars:(var)list" THEN
  EXISTS_TAC "vals:(var)list" THEN
  EXISTS_TAC "ns:(num)list" THEN
  EXISTS_TAC "glbs:(var)list" THEN
  EXISTS_TAC "pre:aexp" THEN
  EXISTS_TAC "post:aexp" THEN
  EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "rec:aexp");;
e(ASM_REWRITE_TAC[]);;

>%


let C_Ck =
 prove_thm
  (`C_Ck`,
   "!c r s1 s2. C c r s1 s2 = ?k. Ck c r k s1 s2",
   REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [ REWRITE_TAC[C_IMP_Ck];

       STRIP_TAC
       THEN UNDISCH_ALL_TAC
       THEN REWRITE_TAC[Ck_IMP_C]
     ]
  );;

%
g "!c r s1 s2. C c r s1 s2 = ?k. Ck c r k s1 s2";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(REWRITE_TAC[C_IMP_Ck]);;

  e(STRIP_TAC);;
  e(UNDISCH_ALL_TAC);;
  e(REWRITE_TAC[C_Ck_LEMMA2]);;
%

let GREATER_IMP_POSITIVE =
 TAC_PROOF
  (([], "!k k'. (k + 1) <= k' ==> 0 < k'"),
   REWRITE_TAC[(SYM o SPEC_ALL)ADD1]
   THEN REPEAT GEN_TAC
   THEN DISCH_TAC
   THEN ASSUME_TAC (ONCE_REWRITE_RULE[(SYM o SPEC_ALL)LESS_EQ_MONO]
                          (SPEC "k:num" ZERO_LESS_EQ))
   THEN IMP_RES_TAC LESS_EQ_TRANS
   THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC
   THEN REWRITE_ALL_TAC[LESS_MONO_EQ]
   THEN ASM_REWRITE_TAC[]
  );;

let POSITIVE_IMP_DEC_INC =
 TAC_PROOF
  (([], "!k. 0 < k ==> ((k - 1) + 1 = k)"),
   GEN_TAC
   THEN DISCH_TAC
   THEN IMP_RES_TAC SUB1_SUC
   THEN POP_ASSUM (ACCEPT_TAC o REWRITE_RULE[ADD1])
  );;

%
g "!k k'. (k + 1) <= k' ==> 0 < k'";;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1]);;
e(REPEAT GEN_TAC);;
e(DISCH_TAC);;
e(ASSUME_TAC (ONCE_REWRITE_RULE[(SYM o SPEC_ALL)LESS_EQ_MONO]
                          (SPEC "k:num" ZERO_LESS_EQ)));;
e(IMP_RES_TAC LESS_EQ_TRANS);;
e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;
e(REWRITE_ALL_TAC[LESS_MONO_EQ]);;
e(IMP_RES_TAC SUB1_SUC);;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[ADD1]));;
e(ASM_REWRITE_TAC[]);;
%


let ALL_GREATER_Ck =
 prove_thm
  (`ALL_GREATER_Ck`,
   "!c r k s1 s2. Ck c r k s1 s2 ==> !k'. (k <= k') ==> Ck c r k' s1 s2",
   CMDk_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THENL
     [ SKIPk_TAC;

       ASSIGNk_TAC
       THEN ASM_REWRITE_TAC[];

       SEQk_TAC
       THEN IMP_RES_TAC LESS_EQ_TRANS
       THEN RES_TAC
       THEN EXISTS_TAC "k':num"
       THEN EXISTS_TAC "s2:state"
       THEN EXISTS_TAC "k':num"
       THEN ASM_REWRITE_TAC[LESS_EQ_REFL];

       IFk_T_TAC
       THEN EXISTS_TAC "s2:state"
       THEN RES_TAC
       THEN ASM_REWRITE_TAC[];

       IFk_F_TAC
       THEN EXISTS_TAC "s2:state"
       THEN RES_TAC
       THEN ASM_REWRITE_TAC[];

       WHILEk_T_TAC
       THEN IMP_RES_TAC LESS_EQ_TRANS
       THEN RES_TAC
       THEN EXISTS_TAC "k':num"
       THEN EXISTS_TAC "s2:state"
       THEN EXISTS_TAC "s3:state"
       THEN EXISTS_TAC "k':num"
       THEN ASM_REWRITE_TAC[LESS_EQ_REFL];

       WHILEk_F_TAC
       THEN ASM_REWRITE_TAC[];

       IMP_RES_TAC GREATER_IMP_POSITIVE
       THEN IMP_RES_TAC POSITIVE_IMP_DEC_INC
       THEN POP_ASSUM (SUBST1_TAC o SYM)
       THEN CALLk_TAC
       THEN EXISTS_TAC "c:cmd"
       THEN EXISTS_TAC "vars:(var)list"
       THEN EXISTS_TAC "vals:(var)list"
       THEN EXISTS_TAC "ns:(num)list"
       THEN EXISTS_TAC "glbs:(var)list"
       THEN EXISTS_TAC "pre:aexp"
       THEN EXISTS_TAC "post:aexp"
       THEN EXISTS_TAC "calls:prog_env"
       THEN EXISTS_TAC "rec:aexp"
       THEN UNASSUME_THEN (ASSUME_TAC o SYM)
               "vals' = variants vals(SL(APPEND xs glbs))"
       THEN ASM_REWRITE_TAC[]
       THEN FIRST_ASSUM MATCH_MP_TAC
       THEN ONCE_REWRITE_TAC[(SYM o SPEC_ALL)LESS_EQ_MONO]
       THEN IMP_RES_TAC POSITIVE_IMP_DEC_INC
       THEN ASM_REWRITE_TAC[ADD1]
     ]
  );;

%< ALL_GREATER_Ck
g "!c r k s1 s2. Ck c r k s1 s2 ==> !k'. (k <= k') ==> Ck c r k' s1 s2";;
e(CMDk_INDUCT_TAC);;

% SKIP: %
e(REPEAT STRIP_TAC);;
e(SKIPk_TAC);;

% ASSIGN: %
e(REPEAT STRIP_TAC);;
e(ASSIGNk_TAC);;
e(ASM_REWRITE_TAC[]);;

% SEQ: %
e(REPEAT STRIP_TAC);;
e(SEQk_TAC);;
e(IMP_RES_TAC LESS_EQ_TRANS);;
e(RES_TAC);;
e(EXISTS_TAC "k':num" THEN
  EXISTS_TAC "s2:state" THEN
  EXISTS_TAC "k':num");;
e(ASM_REWRITE_TAC[LESS_EQ_REFL]);;

% IF_T: %
e(REPEAT STRIP_TAC);;
e(IFk_T_TAC);;
e(EXISTS_TAC "s2:state");;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

% IF_F: %
e(REPEAT STRIP_TAC);;
e(IFk_F_TAC);;
e(EXISTS_TAC "s2:state");;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

% WHILE_T: %
e(REPEAT STRIP_TAC);;
e(WHILEk_T_TAC);;
e(IMP_RES_TAC LESS_EQ_TRANS);;
e(RES_TAC);;
e(EXISTS_TAC "k':num" THEN
  EXISTS_TAC "s2:state" THEN
  EXISTS_TAC "s3:state" THEN
  EXISTS_TAC "k':num");;
e(ASM_REWRITE_TAC[LESS_EQ_REFL]);;

% WHILE_F: %
e(REPEAT STRIP_TAC);;
e(WHILEk_F_TAC);;
e(ASM_REWRITE_TAC[]);;

% CALL: %
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC GREATER_IMP_POSITIVE);;
e(IMP_RES_TAC POSITIVE_IMP_DEC_INC);;
e(POP_ASSUM (SUBST1_TAC o SYM));;
e(CALLk_TAC);;
e(EXISTS_TAC "c:cmd" THEN
  EXISTS_TAC "vars:(var)list" THEN
  EXISTS_TAC "vals:(var)list" THEN
  EXISTS_TAC "ns:(num)list" THEN
  EXISTS_TAC "glbs:(var)list" THEN
  EXISTS_TAC "pre:aexp" THEN
  EXISTS_TAC "post:aexp" THEN
  EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "rec:aexp");;
e(UNASSUME_THEN (ASSUME_TAC o SYM)
           "vals' = variants vals(SL(APPEND xs glbs))");;
e(ASM_REWRITE_TAC[]);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL)LESS_EQ_MONO]);;
e(IMP_RES_TAC POSITIVE_IMP_DEC_INC);;
e(ASM_REWRITE_TAC[ADD1]);;
>%


let Ck_CALL_IMP_NOT_0 =
 TAC_PROOF
  (([],
    "!c r k s1 s2. Ck c r k s1 s2 ==>
                   !p xs es. (c = CALL p xs es) ==> ~(k = 0)"
   ),
   CMDk_INDUCT_TAC
   THEN REWRITE_TAC[cmd_distinct;cmd_one_one]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD1;NOT_SUC]
  );;

% Ck_CALL_IMP_NOT_0
g "!c r k s1 s2. Ck c r k s1 s2 ==>
                 !p xs es. (c = CALL p xs es) ==> ~(k = 0)";;
e(CMDk_INDUCT_TAC
  THEN REWRITE_TAC[cmd_distinct;cmd_one_one]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1;NOT_SUC]);;
%

let NOT_Ck_CALL_0 =
 prove_thm
  (`NOT_Ck_CALL_0`,
   "!p xs es r s1 s2. ~(Ck (CALL p xs es) r 0 s1 s2)",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC Ck_CALL_IMP_NOT_0
   THEN POP_ASSUM (ACCEPT_TAC o REWRITE_RULE[]
                   o SPECL["p:string";"xs:(var)list";"es:(exp)list"])
  );;

% NOT_Ck_CALL_0
g "!p xs es r s1 s2. ~(Ck (CALL p xs es) r 0 s1 s2)";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC Ck_CALL_IMP_NOT_0);;
e(POP_ASSUM (ACCEPT_TAC o REWRITE_RULE[]
             o SPECL["p:string";"xs:(var)list";"es:(exp)list"]));;
%


% {p}c{q}/r,k is represented by PPCSPECk(p,c,q,r,k) %

let PCSPECk = new_definition
 (`PCSPECk`,
  "PCSPECk(p,c,q,r,k) = !(s1 s2:state). A p s1 /\ Ck c r k s1 s2 ==> A q s2");;


let PCSPEC_PCSPECk =
 prove_thm
  (`PCSPEC_PCSPECk`,
   "!p c q r. PCSPEC(p,c,q,r) = (!k. PCSPECk(p,c,q,r,k))",
   REWRITE_TAC[PCSPEC;PCSPECk]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [ REPEAT STRIP_TAC
       THEN FIRST_ASSUM MATCH_MP_TAC
       THEN EXISTS_TAC "s1:state"
       THEN ASM_REWRITE_TAC[C_Ck]
       THEN EXISTS_TAC "k:num"
       THEN ASM_REWRITE_TAC[];

       REWRITE_TAC[C_Ck]
       THEN REPEAT STRIP_TAC
       THEN RES_TAC
     ]
  );;

% PCSPEC_PCSPECk
g "!p c q r. PCSPEC(p,c,q,r) = (!k. PCSPECk(p,c,q,r,k))";;
e(REWRITE_TAC[PCSPEC;PCSPECk]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "s1:state");;
  e(ASM_REWRITE_TAC[C_Ck]);;
  e(EXISTS_TAC "k:num");;
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[C_Ck]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
%


let ALL_LESS_PCSPECk =
 prove_thm
  (`ALL_LESS_PCSPECk`,
   "!p c q r k k'. PCSPECk(p,c,q,r,k) /\ (k' <= k) ==> PCSPECk(p,c,q,r,k')",
   REWRITE_TAC[PCSPECk]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "s1:state"
   THEN ASM_REWRITE_TAC[]
   THEN IMP_RES_TAC ALL_GREATER_Ck
  );;

% ALL_LESS_PCSPECk
g "!p c q r k k'. PCSPECk(p,c,q,r,k) /\ (k' <= k) ==> PCSPECk(p,c,q,r,k')";;
e(REWRITE_TAC[PCSPECk]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "s1:state");;
e(ASM_REWRITE_TAC[]);;
e(IMP_RES_TAC ALL_GREATER_Ck);;
%


let PCSPECk_CALL_0 =
 prove_thm
  (`PCSPECk_CALL_0`,
   "!p s xs es q r. PCSPECk(p,CALL s xs es,q,r,0)",
   REWRITE_TAC[PCSPECk;NOT_Ck_CALL_0]
  );;

% PCSPECk_CALL_0
g "!p s es q r. PCSPECk(p,CALL s es,q,r,0)";;
e(REWRITE_TAC[PCSPECk]);;
e(REWRITE_TAC[NOT_Ck_CALL_0]);;
%


let WFprock_partial = 
    new_definition (`WFprock_partial`,
      "WFprock_partial (vars,vals,glbs,pre,post,calls:prog_env,rec:aexp,c)
                       (r:env) k =
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            PCSPECk(AAND (mk_eqs x0 x) pre, c, post, r, k)");;

let WFprock = 
    new_definition (`WFprock`,
      "WFprock t r k =
            WFproc_syntax t r  /\
            WFprock_partial t r k");;

let WFenvk_partial = 
    new_definition (`WFenvk_partial`,
      "WFenvk_partial r k =
         !p. WFprock_partial (r p) r k");;

let WFenvk = 
    new_definition (`WFenvk`,
      "WFenvk r k =
         !p. WFprock (r p) r k");;

let WFenvk_aspects =
 prove_thm
  (`WFenvk_aspects`,
   "!r. WFenvk r k = 
          WFenv_syntax r /\
          WFenvk_partial r k",
   REWRITE_TAC[WFenvk;WFprock]
   THEN REWRITE_TAC[WFenv_syntax;WFenvk_partial]
   THEN CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV)
   THEN REWRITE_TAC[]
  );;

% WFenvk_aspects
g "!r k. WFenvk r k = 
          WFenv_syntax r /\
          WFenvk_partial r k";;
e(REWRITE_TAC[WFenvk;WFprock]);;
e(REWRITE_TAC[WFenv_syntax;WFenvk_partial]);;
e(CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV));;
e(REWRITE_TAC[]);;
%


let WFenvk_WFenv_syntax =
 prove_thm
  (`WFenvk_WFenv_syntax`,
   "!r k. WFenvk r k ==> WFenv_syntax r",
   REWRITE_TAC[WFenvk_aspects]
   THEN REPEAT STRIP_TAC
  );;

%
g "!r k. WFenvk r k ==> WFenv_syntax r";;
e(REWRITE_TAC[WFenvk_aspects]);;
e(REPEAT STRIP_TAC);;
%


let WFenvp_WFenvk =
 prove_thm
  (`WFenvp_WFenvk`,
   "!r. WFenvp r = (!k. WFenvk r k)",
   REWRITE_TAC[WFenvp;WFenvk]
   THEN REWRITE_TAC[WFprocp;WFprock]
   THEN REWRITE_TAC[WFproc_partial]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN REWRITE_TAC[WFprock_partial]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[PCSPEC_PCSPECk]
   THEN CONV_TAC (DEPTH_CONV RIGHT_AND_FORALL_CONV)
   THEN GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFenvp_WFenvk
g "!r. WFenvp r = (!k. WFenvk r k)";;
e(REWRITE_TAC[WFenvp;WFenvk]);;
e(REWRITE_TAC[WFprocp;WFprock]);;
e(REWRITE_TAC[WFproc_partial]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(REWRITE_TAC[WFprock_partial]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[PCSPEC_PCSPECk]);;
e(CONV_TAC (DEPTH_CONV RIGHT_AND_FORALL_CONV));;
e(GEN_TAC);;
e(EQ_TAC THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);;
%


let ALL_LESS_WFenvk =
 prove_thm
  (`ALL_LESS_WFenvk`,
   "!r k k'. WFenvk r k /\ (k' <= k) ==> WFenvk r k'",
   REWRITE_TAC[WFenvk;WFprock]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN REWRITE_TAC[WFprock_partial]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC ALL_LESS_PCSPECk
   THEN EXISTS_TAC "k:num"
   THEN ASM_REWRITE_TAC[]
  );;

% ALL_LESS_WFenvk
g "!r k k'. WFenvk r k /\ (k' <= k) ==> WFenvk r k'";;
e(REWRITE_TAC[WFenvk]);;
e(REWRITE_TAC[WFprock]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(REWRITE_TAC[WFprock_partial]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(MATCH_MP_TAC ALL_LESS_PCSPECk);;
 e(EXISTS_TAC "k:num");;
 e(ASM_REWRITE_TAC[]);;
%


% We will want to prove, eventually, 

   |- !c q r k. WFc c r /\ WFenvk r k ==>
                let (p,h) = vcg1 c q p in
                (ALL_EL close h ==> PCSPECk(p,c,q,r,k+1))

in order to prove

   |- !p c q r k. WFc c r /\ WFenvk r k ==>
                  (ALL_EL close (vcgc p c q r) ==> PCSPECk(p,c,q,r,k+1))

in order to prove

   |- !d r k. r = mkenv d r0 /\ WFd d r /\ WFenvk r k ==>
              (ALL_EL close (vcgd d r) ==> WFenvk r (k+1))

and then

   |- !d r. r = mkenv d r0 /\ WFd d r ==>
            (ALL_EL close (vcgd d r) ==> WFenvp r)


But first we need to prove all of the superstructure needed to deal
with WFenvk, so that we will have the tools to build the needed
critical procedure call rule.

%



% from theory free_variables: %
let c_substk_EQ_LEMMA =
 TAC_PROOF
  (([],
    "!c1 r k s1 s2.
      Ck c1 r k s1 s2 ==>
      (!c (ss1:vsubst) ss2.
        (c <|c ss2 = c1) /\
        (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
        Ck (c <|c ss1) r k s1 s2)"),
   CMDk_STR_INDUCT_TAC
   THENL
    [
% SKIP: %
      GEN_TAC THEN GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN SKIPk_TAC;

% ASSIGN: %
      GEN_TAC THEN GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM e_subst_EQ
         THENL
          [ REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            ASM_REWRITE_TAC[]
            THEN ASSIGNk_TAC
            THEN ASM_REWRITE_TAC[]
          ]
       ];

% SEQ: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN SEQk_TAC
      THEN EXISTS_TAC "k1:num"
      THEN EXISTS_TAC "s2:state"
      THEN EXISTS_TAC "k2:num"
      THEN ASM_REWRITE_TAC[]
      THEN CONJ_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "ss2:vsubst"
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[];

% IF_T: %
      GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         IFk_T_TAC
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "ss2:vsubst"
         THEN ASM_REWRITE_TAC[]
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

% IF_F: %
      GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         IFk_F_TAC
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "ss2:vsubst"
         THEN ASM_REWRITE_TAC[]
         THEN REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

% WHILE_T: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN WHILEk_T_TAC
      THEN EXISTS_TAC "k1:num"
      THEN EXISTS_TAC "s2:state"
      THEN EXISTS_TAC "s3:state"
      THEN EXISTS_TAC "k2:num"
      THEN REWRITE_TAC[ONCE_REWRITE_RULE[EQ_SYM_EQ] c_subst]
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[]
         THEN CONJ_TAC
         THENL
          [ FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "ss2:vsubst"
            THEN ASM_REWRITE_TAC[]
            THEN REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "ss2:vsubst"
            THEN ASM_REWRITE_TAC[c_subst;cmd_one_one;FVc_DEF;IN_UNION]
          ]
       ];

% WHILE_F: %
      GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC
      THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN WHILEk_F_TAC
      THEN USE_IMP_THEN REWRITE_THM b_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[]
       ];

% CALL: %
      INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
      THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]
      THEN REPEAT STRIP_TAC
      THEN UNASSUME_THEN REWRITE_ALL_THM "(s:string) = p"
      THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
                  "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
      THEN USE_IMP_THEN REWRITE_THM xs_subst_EQ
      THENL
       [ REPEAT STRIP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM es_subst_EQ
         THENL
          [ REPEAT STRIP_TAC
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN ASM_REWRITE_TAC[];

            ASM_REWRITE_TAC[]
            THEN CALLk_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:(var)list"
            THEN EXISTS_TAC "vals:(var)list"
            THEN EXISTS_TAC "ns:(num)list"
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "(APPEND xs vals') //v (APPEND vars vals)"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

%<
g "!c1 r k s1 s2.
     Ck c1 r k s1 s2 ==>
     (!c (ss1:vsubst) ss2.
        (c <|c ss2 = c1) /\
        (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
        Ck (c <|c ss1) r k s1 s2)";;
e(CMDk_STR_INDUCT_TAC);;

% SKIP: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(SKIPk_TAC);;

% ASSIGN: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

  e(REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM e_subst_EQ);;

   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;
   e(ASSIGNk_TAC);;
   e(ASM_REWRITE_TAC[]);;

% SEQ: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(SEQk_TAC);;
 e(EXISTS_TAC "k1:num"
   THEN EXISTS_TAC "s2:state"
   THEN EXISTS_TAC "k2:num");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% IF_T: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(IFk_T_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% IF_F: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(IFk_F_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "ss2:vsubst");;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

% WHILE_T: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
let st1 = get_state();;
set_state(st1);;
 e(WHILEk_T_TAC);;
 e(EXISTS_TAC "k1:num"
    THEN EXISTS_TAC "s2:state"
    THEN EXISTS_TAC "s3:state"
    THEN EXISTS_TAC "k2:num");;
 e(REWRITE_TAC[ONCE_REWRITE_RULE[EQ_SYM_EQ] c_subst]);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(CONJ_TAC);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "ss2:vsubst");;
   e(ASM_REWRITE_TAC[]);;
   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "ss2:vsubst");;
   e(ASM_REWRITE_TAC[c_subst;cmd_one_one;FVc_DEF;IN_UNION]);;

% WHILE_F: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(WHILEk_F_TAC);;
 e(USE_IMP_THEN REWRITE_THM b_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

let st2 = get_state();;
set_state(st2);;

% CALL: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVc_DEF;IN;IN_UNION;c_subst;cmd_one_one;cmd_distinct]);;
 e(REPEAT STRIP_TAC);;
 e(UNASSUME_THEN REWRITE_ALL_THM "(s:string) = p");;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
          "(r:env) p = vars,vals,glbs,pre,post,c");;
let st3 = get_state();;
set_state(st3);;
 e(USE_IMP_THEN REWRITE_THM xs_subst_EQ);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM es_subst_EQ);;

   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;
   e(CALLk_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:(var)list"
     THEN EXISTS_TAC "vals:(var)list"
     THEN EXISTS_TAC "ns:(num)list"
     THEN EXISTS_TAC "glbs:(var)list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp");;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "(APPEND xs vals') //v (APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

>%


let c_substk_IMP =
    REWRITE_RULE[AND_IMP_INTRO]
           (CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
                        c_substk_EQ_LEMMA);;


let c_substk_EQ =
 prove_thm
  (`c_substk_EQ`,
   "!ss2 ss1 c r k s1 s2.
      (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
      (Ck (c <|c ss1) r k s1 s2 = Ck (c <|c ss2) r k s1 s2)",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN DISCH_TAC
   THEN MATCH_MP_TAC c_substk_IMP
   THENL
    [ EXISTS_TAC "c <|c ss1"
      THEN EXISTS_TAC "ss1:vsubst"
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      EXISTS_TAC "c <|c ss2"
      THEN EXISTS_TAC "ss2:vsubst"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< c_substk_EQ
g "!ss2 ss1 c r k s1 s2.
      (!x. x IN (FVc c r) ==> (ss1 x = ss2 x)) ==>
      (Ck (c <|c ss1) r k s1 s2 = Ck (c <|c ss2) r k s1 s2)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(MATCH_MP_TAC c_substk_IMP);;
 e(EXISTS_TAC "c <|c ss1"
   THEN EXISTS_TAC "ss1:vsubst");;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(DISCH_TAC);;
 e(MATCH_MP_TAC c_substk_IMP);;
 e(EXISTS_TAC "c <|c ss2"
   THEN EXISTS_TAC "ss2:vsubst");;
 e(ASM_REWRITE_TAC[]);;
>%




% ============================================ %
% Variants of command semantics theorems:      %
% ============================================ %







let C_PROC_FREE_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' xs (s2:state) s3 z.
      (WFenv_syntax r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL(APPEND xs glbs))) /\
      (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
      (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
           (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r /\
       WFenv_syntax r ==>
        (!z.
          ~z IN (FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r) ==>
          (s3 z = BNDsL ns vals' s2 z))) /\
      ~(z IN (SL xs)) /\ ~(z IN (SL glbs))
      ==>
      (BNDsL (MAP s2 vals') vals' s3 z = s2 z)"
   ),
   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(z:var) IN (SL vals')"
   THENL
     [ IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH;

       IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
       THEN FIRST_ASSUM (SUBST1_TAC o SYM
                         o ISPECL["ns:(num)list";"s2:state"]
                         o MATCH_MP BNDsL_NOT_MATCH)
       THEN SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                              (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
                  ASSUME_TAC
       THENL
         [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
           THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
           THEN ASM_REWRITE_TAC[];

           RES_TAC
           THEN POP_ASSUM MATCH_MP_TAC
           THEN IMP_RES_TAC FVc_proc_subst
           THEN MATCH_MP_TAC NOT_IN_SUBSET
           THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):(var)set"
           THEN POP_ASSUM REWRITE_THM
           THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

%< C_PROC_FREE_ONLY_IMPACT_LEMMA
g "!r p vars vals glbs pre post calls rec c vals' xs (s2:state) s3 z.
      (WFenv_syntax r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL(APPEND xs glbs))) /\
      (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
      (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
           (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r /\
       WFenv_syntax r ==>
        (!z.
          ~z IN (FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r) ==>
          (s3 z = BNDsL ns vals' s2 z))) /\
      ~(z IN (SL xs)) /\ ~(z IN (SL glbs))
      ==>
      (BNDsL (MAP s2 vals') vals' s3 z = s2 z)";;
 e(REPEAT STRIP_TAC);;
 e(ASM_CASES_TAC "(z:var) IN (SL vals')");;

  e(IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(FIRST_ASSUM (SUBST1_TAC o SYM
                 o ISPECL["ns:(num)list";"s2:state"]
                 o MATCH_MP BNDsL_NOT_MATCH));;
  e(SUBGOAL_THEN "WFc(c <|c ((APPEND xs vals') //v (APPEND vars vals)))r"
         ASSUME_TAC);;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;
   e(POP_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_TAC FVc_proc_subst);;
   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):(var)set");;
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[SL_APPEND]);;
   e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
>%





let Ck_FREE_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!c r k s1 s2.
      Ck c r k s1 s2 ==>
      (!g. WFc c g r /\ WFenv_syntax r ==>
           (!z. (~z IN FVc c r) ==>
                (s2 z = s1 z)))"
   ),
   CMDk_INDUCT_TAC
   THEN REWRITE_ALL_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN ASM_REWRITE_TAC[FVxs_SL;o_THM]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC (INST_TYPE [":num",":*"] BND_NOT_MATCH)
      THEN IMP_RES_TAC E_FREE_ONLY_IMPACT
      THEN ASM_REWRITE_TAC[];

      RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN FIRST_ASSUM (\th. (SPEC "g:prog_env" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC_ALL)
      THEN SUBGOAL_THEN  "?v x. (pr = ALESS v (AVAR x)) /\ ~WFx x /\
                                (~x IN (FVa a) /\ ~x IN (FVv v)) /\
                                (!p. ~x IN (FVa ((g:prog_env) p)))"
               REWRITE_THM
      THENL
       [ EXISTS_TAC "v:vexp"
         THEN EXISTS_TAC "x:var"
         THEN ASM_REWRITE_TAC[aexp_one_one;vexp_one_one];

         ASM_REWRITE_TAC[]
         THEN DISCH_TAC
         THEN RES_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_TAC B_FREE_ONLY_IMPACT;

      REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
      THEN IMP_RES_TAC ES_FREE_ONLY_IMPACT
      THEN POP_ASSUM (REWRITE_THM o SYM)
      THEN MATCH_MP_TAC (SPEC_ALL C_PROC_FREE_ONLY_IMPACT_LEMMA)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< C_FREE_ONLY_IMPACT_LEMMA:
g "!c r k s1 s2.
      Ck c r k s1 s2 ==>
      (!g. WFc c g r /\ WFenv_syntax r ==>
           (!z. (~z IN FVc c r) ==>
                (s2 z = s1 z)))";;
e(CMDk_INDUCT_TAC);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (INST_TYPE [":num",":*"] BND_NOT_MATCH));;
 e(IMP_RES_TAC E_FREE_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_ALL_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(FIRST_ASSUM (\th. SPEC "g:prog_env" th; UNDISCH_TAC (concl th)));;
 e(DISCH_THEN (MP_TAC o SPEC_ALL));;
 e(SUBGOAL_THEN  "?v x. (pr = ALESS v (AVAR x)) /\ ~WFx x /\
                        (~x IN (FVa a) /\ ~x IN (FVv v)) /\
                        (!p. ~x IN (FVa ((g:prog_env) p)))"
       REWRITE_THM);;

  e(EXISTS_TAC "v:vexp"
     THEN EXISTS_TAC "x:var");;
  e(ASM_REWRITE_TAC[aexp_one_one;vexp_one_one]);;

  e(ASM_REWRITE_TAC[]);;
  e(DISCH_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(ASM_REWRITE_TAC[FVxs_SL;o_THM]);;
 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
 e(IMP_RES_TAC ES_FREE_ONLY_IMPACT);;
 e(POP_ASSUM (REWRITE_THM o SYM));;
 e(MATCH_MP_TAC (SPEC_ALL C_PROC_FREE_ONLY_IMPACT_LEMMA));;
 e(ASM_REWRITE_TAC[]);;
>%


let Ck_FREE_ONLY_IMPACT =
 prove_thm
  (`Ck_FREE_ONLY_IMPACT`,
   "!c r k s1 s2 g z.
      WFc c g r /\ WFenv_syntax r /\
      Ck c r k s1 s2 /\
      (~z IN FVc c r) ==>
      (s2 z = s1 z)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC Ck_FREE_ONLY_IMPACT_LEMMA
  );;

%< Ck_FREE_ONLY_IMPACT:
g "!c r k s1 s2 g z.
      WFc c g r /\ WFenv_syntax r /\
      Ck c r k s1 s2 /\
      (~z IN FVc c r) ==>
      (s2 z = s1 z)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC Ck_FREE_ONLY_IMPACT_LEMMA);;
>%


let Ck_FREE_ONLY_VARIATION =
 prove_thm
  (`Ck_FREE_ONLY_VARIATION`,
   "!x c g r k s1 s2.
     Ck c r k s1 s2 /\
     WFenv_syntax r /\ WFc c g r /\ FVc c r SUBSET SL x ==>
     (BNDsL (MAP s2 x) x s1 = s2)",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL x"
   THENL
     [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

       IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
       THEN IMP_RES_TAC NOT_IN_SUBSET
       THEN IMP_RES_TAC Ck_FREE_ONLY_IMPACT
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% Ck_FREE_ONLY_VARIATION:
g "!x c g r k s1 s2.
     Ck c r k s1 s2 /\
     WFenv_syntax r /\ WFc c g r /\ FVc c r SUBSET SL x ==>
     (BNDsL (MAP s2 x) x s1 = s2)";;

e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL x");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(IMP_RES_TAC NOT_IN_SUBSET);;
 e(IMP_RES_TAC Ck_FREE_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;
%


let BNDsL_MAP_state_SYM_STAGE =
 prove_thm
  (`BNDsL_MAP_state_SYM_STAGE`,
   "!vars vals glbs c g r k s1 s2.
      WFenv_syntax r /\ WFc c g r /\
      DL (APPEND (APPEND vars vals) glbs) /\
      (FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs)) /\
      Ck c r k s1 s2
      ==>
      (BNDsL (MAP s2 (APPEND vars glbs)) (APPEND vars glbs) s1  =
       BNDsL (MAP s1 vals) vals s2)",
   REWRITE_TAC[SUBSET_DEF]
   THEN REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
   THENL
     [  UNDISCH_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
        THEN STRIP_TAC
        THEN IMP_RES_TAC IN_DISJOINT_LIST3
        THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
        THENL [ USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH;
                USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH;
                USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH ]
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION];

        FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "z:var")
        THEN UNDISCH_TAC "~(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
        THEN STRIP_TAC
        THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN IMP_RES_TAC Ck_FREE_ONLY_IMPACT
        THEN ASM_REWRITE_TAC[]
        THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
     ]
  );;

%< BNDsL_MAP_state_SYM_STAGE
g "!vars vals glbs c g r k s1 s2.
      WFenv_syntax r /\ WFc c g r /\
      DL (APPEND (APPEND vars vals) glbs) /\
      (FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs)) /\
      Ck c r k s1 s2
      ==>
      (BNDsL (MAP s2 (APPEND vars glbs)) (APPEND vars glbs) s1  =
       BNDsL (MAP s1 vals) vals s2)";;

e(REWRITE_TAC[SUBSET_DEF]);;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)");;

  e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
  e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
  e(STRIP_TAC);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

  e(FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "z:var"));;
  e(UNDISCH_TAC "~(z:var) IN SL (APPEND (APPEND vars vals) glbs)");;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(STRIP_TAC);;
  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(IMP_RES_TAC Ck_FREE_ONLY_IMPACT);;
  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;
>%


% =========================== %
% From file cmd_semantics.ml: %
% =========================== %



let PROC_STATE_TRANSFORM_STAGE =
 TAC_PROOF
  (([],
    "!r k p vars vals glbs pre post calls rec c x vals' x' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (x' = APPEND (APPEND xs vals') glbs) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))"
   ),
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL vals'"
   THENL
    [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

      IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
      THEN IMP_RES_TAC FVc_proc_subst
      THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "(x':(var)list) =
                                            APPEND(APPEND xs vals')glbs")])
      THEN IMP_RES_TAC WFc_proc_subst
      THEN ASM_CASES_TAC "(z:var) IN SL x"
      THENL
       [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
         THEN IMP_RES_TAC Ck_FREE_ONLY_VARIATION
         THEN POP_ASSUM (CONV_TAC o ONCE_DEPTH_CONV o RATOR_CONV o REWR_CONV o SYM)
         THEN ASM_CASES_TAC "(z:var) IN SL x'"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
          ];

         IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
         THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
         THENL
          [ REWRITE_TAC[ASSUME "(x':(var)list) = APPEND(APPEND xs vals')glbs";
                 SL_APPEND;IN_UNION;DE_MORGAN_THM]
            THEN ASM_REWRITE_TAC[]
            THEN IMP_RES_TAC NOT_IN_SUBSET
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
          ]
       ]
    ]
  );;

%<
g "!r k p vars vals glbs pre post calls rec c x vals' x' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (x' = APPEND (APPEND xs vals') glbs) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL vals'");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(IMP_RES_TAC FVc_proc_subst);;
 e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "(x':(var)list) =
                                            APPEND(APPEND xs vals')glbs")]));;
 e(IMP_RES_TAC WFc_proc_subst);;
 e(ASM_CASES_TAC "(z:var) IN SL x");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
  e(IMP_RES_TAC Ck_FREE_ONLY_VARIATION);;
  e(POP_ASSUM (CONV_TAC o ONCE_DEPTH_CONV o RATOR_CONV o REWR_CONV o SYM));;
  e(ASM_CASES_TAC "(z:var) IN SL x'");;

   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(REWRITE_TAC[ASSUME "(x':(var)list) = APPEND(APPEND xs vals')glbs";
                 SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
   e(IMP_RES_TAC NOT_IN_SUBSET);;
   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
>%




let PROC_STATE_TRANSFORM_THM_STAGE =
 TAC_PROOF
  (([],
    "!r k p vars vals glbs pre post calls rec c x vals' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      let x' = APPEND (APPEND xs vals') glbs in
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))"
   ),
   REPEAT STRIP_TAC
   THEN CONV_TAC (ONCE_DEPTH_CONV let_CONV)
   THEN MATCH_MP_TAC (SPEC_ALL PROC_STATE_TRANSFORM_STAGE)
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!r k p vars vals glbs pre post calls rec c x vals' xs s2 s3 (s1':state).
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (SL xs SUBSET SL x) /\ (SL glbs SUBSET SL x) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3
     ==>
     (BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x) x s1') =
      let x' = APPEND (APPEND xs vals') glbs in
      BNDsL (MAP (BNDsL (MAP s2 x) x s1') vals') vals' (BNDsL (MAP s3 x') x'
        (BNDsL ns vals' (BNDsL (MAP s2 x) x s1'))))";;
e(REPEAT STRIP_TAC);;
e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
e(MATCH_MP_TAC (SPEC_ALL PROC_STATE_TRANSFORM_STAGE));;
e(ASM_REWRITE_TAC[]);;
>%



% =========================================================== %
% Here is the Ck_EXTENSION theorem.                           %
% =========================================================== %


let Ck_EXTENSION_LEMMA =
 TAC_PROOF
  (([],
    "!c r k s1 s2.
      Ck c r k s1 s2 ==>
      (!x s1' g.
        WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
        (MAP s1 x = MAP s1' x) ==>
        Ck c r k s1' (BNDsL (MAP s2 x) x s1'))"
   ),
   CMDk_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc;FVc_DEF;FVxs_SL]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN ASM_REWRITE_TAC[SUBSET;UNION_SUBSET;o_THM]
   THEN REPEAT STRIP_TAC
   THENL

     [  % SKIP: %
        ASM_REWRITE_TAC[IDENT_BNDsL]
        THEN SKIPk_TAC;

        % ASSIGN: %
        IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE
        THEN ASSIGNk_TAC
        THEN IMP_RES_TAC E_EXTENSION;

        % SEQ: %
        SEQk_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN EXISTS_TAC "k2:num"
        THEN CONJ_TAC
        THENL
          [  FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN IMP_RES_TAC ALL_LESS_WFenvk
             THEN ASM_REWRITE_TAC[];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN ASM_REWRITE_TAC[]
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN IMP_RES_TAC ALL_LESS_WFenvk
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP]
          ];

        % IF_T: %
        IFk_T_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             IMP_RES_TAC B_EXTENSION
          ];

        % IF_F: %
        IFk_F_TAC
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             IMP_RES_TAC B_EXTENSION
          ];

        % WHILE_T: %
        WHILEk_T_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'"
        THEN EXISTS_TAC "BNDsL(MAP (s3:state) x) x s1'"
        THEN EXISTS_TAC "k2:num"
        THEN REPEAT CONJ_TAC
        THENL
          [  SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s3:state) x";"x:(var)list";
                                  "MAP (s2:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN IMP_RES_TAC ALL_LESS_WFenvk
             THEN ASM_REWRITE_TAC[MAP_BNDsL_MAP];

             SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                          ISPECL ["MAP (s4:state) x";"x:(var)list";
                                  "MAP (s3:state) x";"s1':state"])
                         BNDsL_BNDsL_ABSORPTION)
             THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
             THEN IMP_RES_TAC ALL_LESS_WFenvk
             THEN ASM_REWRITE_TAC[WFc;FVc_DEF;UNION_SUBSET;MAP_BNDsL_MAP]
             THEN EXISTS_TAC "v:vexp"
             THEN EXISTS_TAC "x':var"
             THEN ASM_REWRITE_TAC[];

             ASM_REWRITE_TAC[];

             ASM_REWRITE_TAC[];

             IMP_RES_TAC B_EXTENSION
          ];

        % WHILE_F: %
        WHILEk_F_TAC
        THEN IMP_RES_TAC B_EXTENSION;

        % CALL: %
        ONCE_REWRITE_TAC[BNDsL_MAP_BNDsL_MAP_COMMUTE]
        THEN REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
        THEN ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD)
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL PROC_STATE_TRANSFORM_THM_STAGE)
        THENL
          [  ASM_REWRITE_TAC[];

             CONV_TAC (ONCE_DEPTH_CONV let_CONV)
             THEN CALLk_TAC
             THEN EXISTS_TAC "c:cmd"
             THEN EXISTS_TAC "vars:(var)list"
             THEN EXISTS_TAC "vals:(var)list"
             THEN EXISTS_TAC "ns:(num)list"
             THEN EXISTS_TAC "glbs:(var)list"
             THEN EXISTS_TAC "pre:aexp"
             THEN EXISTS_TAC "post:aexp"
             THEN EXISTS_TAC "calls:prog_env"
             THEN EXISTS_TAC "rec:aexp"
             THEN ASM_REWRITE_TAC[]
             THEN REWRITE_TAC[(SYM o ASSUME)
                         "vals' = variants vals(SL(APPEND xs glbs))"]
             THEN CONJ_TAC
             THENL
               [  FIRST_ASSUM MATCH_MP_TAC
                  THEN EXISTS_TAC "calls <|g ((APPEND xs vals') //v (APPEND vars vals))"
                  THEN IMP_RES_TAC FVc_proc_subst
                  THEN IMP_RES_TAC WFc_proc_subst
                  THEN USE_IMP_THEN ASM_REWRITE_THM (SPEC_ALL EXTENSION_MAP_THM)
                  THEN ASM_REWRITE_TAC[];

                  IMP_RES_TAC ES_EXTENSION
               ]
          ]
     ]
  );;

%< Ck_EXTENSION_LEMMA
g "!c r k s1 s2.
      Ck c r k s1 s2 ==>
      (!x s1' g.
        WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
        (MAP s1 x = MAP s1' x) ==>
        Ck c r k s1' (BNDsL (MAP s2 x) x s1'))";;
e(CMDk_STR_INDUCT_TAC);;

% SKIP: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[IDENT_BNDsL]);;
e(SKIPk_TAC);;

% ASSIGN: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_BND_COMMUTE);;
e(ASSIGNk_TAC);;
e(IMP_RES_TAC E_EXTENSION);;

% SEQ: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(SEQk_TAC);;
e(EXISTS_TAC "k1:num");;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(EXISTS_TAC "k2:num");;
e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC ALL_LESS_WFenvk);;
  e(ASM_REWRITE_TAC[]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC ALL_LESS_WFenvk);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

% IF_T: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IFk_T_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(IMP_RES_TAC B_EXTENSION);;

% IF_F: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(IFk_F_TAC);;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(IMP_RES_TAC B_EXTENSION);;

% WHILE_T: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(WHILEk_T_TAC);;
e(EXISTS_TAC "k1:num");;
e(EXISTS_TAC "BNDsL(MAP (s2:state) x) x s1'");;
e(EXISTS_TAC "BNDsL(MAP (s3:state) x) x s1'");;
e(EXISTS_TAC "k2:num");;
e(REPEAT CONJ_TAC);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s3:state) x";"x:(var)list";
                         "MAP (s2:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC ALL_LESS_WFenvk);;
  e(ASM_REWRITE_TAC[MAP_BNDsL_MAP]);;

  e(SUBST1_TAC ((SYM o REWRITE_RULE[LENGTH_MAP] o
                 ISPECL ["MAP (s4:state) x";"x:(var)list";
                         "MAP (s3:state) x";"s1':state"])
                BNDsL_BNDsL_ABSORPTION));;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC ALL_LESS_WFenvk);;
  e(ASM_REWRITE_TAC[WFc;FVc_DEF;UNION_SUBSET;MAP_BNDsL_MAP]);;
  e(EXISTS_TAC "v:vexp");;
  e(EXISTS_TAC "x':var");;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC B_EXTENSION);;

% WHILE_F: %
e(REWRITE_TAC[WFc;FVc_DEF;SUBSET;UNION_SUBSET]);;
e(REPEAT STRIP_TAC);;
e(WHILEk_F_TAC);;
e(IMP_RES_TAC B_EXTENSION);;

% CALL: %
e(REWRITE_TAC[WFc;FVc_DEF;FVxs_SL]);;
e(ASM_REWRITE_TAC[SUBSET;UNION_SUBSET;o_THM]);;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[BNDsL_MAP_BNDsL_MAP_COMMUTE]);;
e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
e(ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD));;
e(IMP_RES_TAC ALL_LESS_WFenvk);;
e(USE_IMP_THEN REWRITE_THM (SPEC_ALL PROC_STATE_TRANSFORM_THM_STAGE));;

 e(ASM_REWRITE_TAC[]);;

e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
e(CALLk_TAC);;
e(EXISTS_TAC "c:cmd" THEN
  EXISTS_TAC "vars:(var)list" THEN
  EXISTS_TAC "vals:(var)list" THEN
  EXISTS_TAC "ns:(num)list" THEN
  EXISTS_TAC "glbs:(var)list" THEN
  EXISTS_TAC "pre:aexp" THEN
  EXISTS_TAC "post:aexp" THEN
  EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "rec:aexp");;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

  e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "calls <|g ((APPEND xs vals') //v (APPEND vars vals))");;
  e(IMP_RES_TAC FVc_proc_subst);; 
  e(IMP_RES_TAC WFc_proc_subst);;
  e(USE_IMP_THEN ASM_REWRITE_THM (SPEC_ALL EXTENSION_MAP_THM));;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC ES_EXTENSION);;

C_EXTENSION_LEMMA = 
|- !c r s1 s2.
    C c r s1 s2 ==>
    (!x s1'.
      WFc c r /\
      WFenv r /\
      (FVc c r) SUBSET (SL x) /\
      (MAP s1 x = MAP s1' x) ==>
      C c r s1'(BNDsL(MAP s2 x)x s1'))
>%

let Ck_EXTENSION =
 prove_thm
  (`Ck_EXTENSION`,
   "!c g r k s1 s2 x s1'.
      WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) /\ Ck c r k s1 s2  ==>
      Ck c r k s1' (BNDsL (MAP s2 x) x s1')",
   REPEAT STRIP_TAC
   THEN POP_ASSUM (MATCH_MP_TAC o SPEC_ALL o MATCH_MP Ck_EXTENSION_LEMMA)
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!c g r k s1 s2 x s1'.
      WFc c g r /\ WFenv_syntax r /\ (FVc c r SUBSET SL x) /\
      (MAP s1 x = MAP s1' x) /\ Ck c r k s1 s2  ==>
      Ck c r k s1' (BNDsL (MAP s2 x) x s1')";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (ASSUME_TAC o MATCH_MP Ck_EXTENSION_LEMMA));;
e(POP_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[]);;
%




let c_subst_STATE_BNDsL_LEMMA_STAGE =
 TAC_PROOF
  (([],
    "!r k p vars vals glbs pre post calls rec c vals' ns s2 s3 vals0 xs l0 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3  /\
     (vals' = variants vals (SL (APPEND xs glbs)))  /\
     (MAP ss vals0 = vals')  /\
     (xs = l0 <|xs ss)  /\
     WFxs l0  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss  /\
     (!x. WFx (ss x) = WFx x)
     ==>
     ((BNDsL (MAP s2 vals') vals0 (s3 o ss)) =
      (BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)))"
   ),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFxs_subst
   THEN POP_TAC
   THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "xs = l0 <|xs ss")])
   THEN SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
            ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
      THEN IMP_RES_TAC DL_APPEND_xs_subst
      THEN ASM_REWRITE_TAC[LENGTH_xs_subst];

      SUBGOAL_THEN
         "DL (APPEND (APPEND l0 glbs) vals0: (var)list)"
               ASSUME_TAC
      THENL
       [ MATCH_MP_TAC (INST_TYPE [":var",":**"] DL_MAP)
         THEN EXISTS_TAC "ss:vsubst"
         THEN REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
         THEN UNASSUME_THEN (ASSUME_TAC o SYM o REWRITE_RULE[xs_subst_MAP])
                   "xs = l0 <|xs ss"
         THEN ASM_REWRITE_TAC[MAP_APPEND]
         THEN ONCE_REWRITE_TAC[DL_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN REWRITE_TAC[DL_variants;DISJOINT_variants]
         THEN IMP_RES_TAC DL_MAP_ONE_ONE
         THEN POP_TAC
         THEN POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                           ASSUME "MAP (ss:vsubst) l0 = xs";
                           ASSUME "MAP (ss:vsubst) glbs = glbs"]);

         IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT)
         THEN REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'");MAP_MAP_o]
         THEN EXT_TAC "z:var"
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(z:var) IN SL vals0"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
            THEN IMP_RES_TAC IN_DISJOINT_LIST
            THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)"
            THENL
             [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN POP_ASSUM (STRIP_ASSUME_TAC o
                               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM])
               THEN REWRITE_TAC[o_THM]
               THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
               THENL
                [ MATCH_MP_TAC NOT_IN_SUBSET
                  THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs): var set"
                  THEN CONJ_TAC
                  THENL
                   [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                     THEN REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]
                     THEN REWRITE_TAC[ASSUME "xs = l0 <|xs ss";
                                      SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]
                     THEN ONCE_REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) glbs = glbs")]
                     THEN REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]
                     THEN IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst;

                     MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                     THEN IMP_RES_TAC DL_APPEND_xs_subst
                     THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                   ];

                  USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
                  THEN REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]
                  THEN REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]
                  THEN IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst
                ]
             ]
          ]
       ]
    ]
  );;

% c_subst_STATE_BNDsL_LEMMA_STAGE
g "!r k p vars vals glbs pre post calls rec c vals' ns s2 s3 vals0 xs l0 ss.
     WFenv_syntax r  /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c)  /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k
       (BNDsL ns vals' s2) s3  /\
     (vals' = variants vals (SL (APPEND xs glbs)))  /\
     (MAP ss vals0 = vals')  /\
     (xs = l0 <|xs ss)  /\
     WFxs l0  /\
     (LENGTH vars = LENGTH l0)  /\
     DL (APPEND l0 glbs)  /\
     (!x. x IN SL glbs ==> (ss x = x))  /\
     ONE_ONE ss  /\
     (!x. WFx (ss x) = WFx x)
     ==>
     ((BNDsL (MAP s2 vals') vals0 (s3 o ss)) =
      (BNDsL (MAP (s3 o ss) (APPEND l0 glbs)) (APPEND l0 glbs) (s2 o ss)))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFxs_subst);;
e(POP_TAC);;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ASSUME "xs = l0 <|xs ss")]));;
e(SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
       ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
  e(IMP_RES_TAC DL_APPEND_xs_subst);;
  e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;

e(SUBGOAL_THEN
      "DL (APPEND (APPEND l0 glbs) vals0: (var)list)"
            ASSUME_TAC);;

 e(MATCH_MP_TAC (INST_TYPE [":var",":**"] DL_MAP));;
 e(EXISTS_TAC "ss:vsubst");;
 e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM o REWRITE_RULE[xs_subst_MAP])
        "xs = l0 <|xs ss");;
 e(ASM_REWRITE_TAC[MAP_APPEND]);;
 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[DL_variants;DISJOINT_variants]);;
 e(IMP_RES_TAC DL_MAP_ONE_ONE);;
 e(POP_TAC);;
 e(POP_ASSUM (REWRITE_THM o REWRITE_RULE[MAP_APPEND;
                           ASSUME "MAP (ss:vsubst) l0 = xs";
                           ASSUME "MAP (ss:vsubst) glbs = glbs"]));;

e(IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT));;
e(REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'");MAP_MAP_o]);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL vals0");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
 e(IMP_RES_TAC IN_DISJOINT_LIST);;
 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(ASM_CASES_TAC "(z:var) IN SL (APPEND l0 glbs)");;

  e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM]));;
  e(REWRITE_TAC[o_THM]);;
  e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs): var set");;
   e(CONJ_TAC);;

    e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(REWRITE_ALL_TAC[SYM(SPEC_ALL MAP_IDENT_ELEMENTS)]);;
    e(REWRITE_TAC[ASSUME "xs = l0 <|xs ss";
                  SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]);;
    e(ONCE_REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) glbs = glbs")]);;
    e(REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]);;
    e(IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst);;

    e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
    e(IMP_RES_TAC DL_APPEND_xs_subst);;
    e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;

   e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(REWRITE_TAC[SYM(ASSUME "MAP (ss:vsubst) vals0 = vals'")]);;
   e(REWRITE_TAC[SYM(SPEC_ALL xs_subst_MAP)]);;
   e(IMP_RES_THEN ASM_REWRITE_THM IN_xs_subst);;
%






let SL_APPEND_SYM23 =
 TAC_PROOF
  (([],
    "!a b (c:(*)list). SL(APPEND(APPEND a b) c) = SL(APPEND(APPEND a c) b)"
   ),
   REWRITE_TAC[EXTENSION]
   THEN REWRITE_TAC[SL_APPEND;IN_UNION]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% SL_APPEND_SYM23 (as in mk_cmd_semantics.ml)
g "!a b (c:(*)list). SL(APPEND(APPEND a b) c) = SL(APPEND(APPEND a c) b)";;
e(REWRITE_TAC[EXTENSION]);;
e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC
  THEN REPEAT STRIP_TAC
  THEN ASM_REWRITE_TAC[]
);;
%





let c_substk_LEMMA1 =
 TAC_PROOF
  (([],
    "!c' r k s1 s2. 
      Ck c' r k s1 s2 ==>
      (!c ss g. (c' = (c <|c ss))
            /\ WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss ==>
        Ck c r k (s1 o ss) (s2 o ss))"
   ),
   CMDk_STR_INDUCT_TAC
   THENL
        % SKIP: %
     [  GEN_TAC THEN GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN SKIPk_TAC;

        % ASSIGN: %
        GEN_TAC THEN GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_THEN ASM_REWRITE_THM BND_o_ONE_ONE_THM
        THEN ASSIGNk_TAC
        THEN IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) e_subst_biject
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "e = e' <|e ss"
        THEN ASM_REWRITE_TAC[];

        % SEQ: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN SEQk_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN EXISTS_TAC "k2:num"
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[];

        % IF_T: %
        GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IFk_T_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN RES_TAC
        THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
                    "WFcsubst c r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % IF_F: %
        GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN IFk_F_TAC
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN RES_TAC
        THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
                    "WFcsubst c r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % WHILE_T: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN WHILEk_T_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "(s2:state) o (ss:vsubst)"
        THEN EXISTS_TAC "(s3:state) o (ss:vsubst)"
        THEN EXISTS_TAC "k2:num"
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN RES_TAC
        THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
                      "WFcsubst c'' r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[WFc_WHILE;c_subst;cmd_one_one;WFcsubst_THM];

        % WHILE_F: %
        GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC
        THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN WHILEk_F_TAC
        THEN UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss"
        THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
                      "WFcsubst c'' r ss"
        THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject;

        % CALL: %
        INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
        THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]
        THEN REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
        THEN REPEAT STRIP_TAC
        THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = s"
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
                       "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN IMP_RES2_THEN REWRITE_THM BNDsL_o_biject
        THEN ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD)
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL c_subst_STATE_BNDsL_LEMMA_STAGE)
        THENL
         [ ASM_REWRITE_TAC[MAP_MAP_o]
           THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
           THEN REWRITE_TAC[MAP_I];

           USE_IMP_THEN REWRITE_THM
               ((SPEC_ALL o SPECL["variants vals (SL (APPEND l0 glbs))";
                                  "vals: var list"])
                     c_subst_PROC_STATE_LEMMA)
           THENL
            [ ASM_REWRITE_TAC[];

              CALLk_TAC
              THEN EXISTS_TAC "c:cmd"
              THEN EXISTS_TAC "vars:var list"
              THEN EXISTS_TAC "vals:var list"
              THEN EXISTS_TAC "ns:num list"
              THEN EXISTS_TAC "glbs:var list"
              THEN EXISTS_TAC "pre:aexp"
              THEN EXISTS_TAC "post:aexp"
              THEN EXISTS_TAC "calls:prog_env"
              THEN EXISTS_TAC "rec:aexp"
              THEN ASM_REWRITE_TAC[]
              THEN CONJ_TAC
              THENL
               [ MATCH_MP_TAC Ck_EXTENSION
                 THEN REWRITE_TAC (map (SYM o ASSUME)
                        ["vals' = variants vals(SL(APPEND xs glbs))";
                         "xs = l0 <|xs ss"])
                 THEN EXISTS_TAC "(calls <|g
   ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))"
                 THEN EXISTS_TAC
    "(BNDsL ns vals' (s2:state)) o
         (((APPEND xs vals') //v (APPEND vars vals)) o
          (\y. @z. ((APPEND l0 (variants vals(SL(APPEND l0 glbs)))) //v
                    (APPEND vars vals))z = y))"
                 THEN REPEAT CONJ_TAC
                 THENL %5 subgoals%
                    % WFc (c <|c ss2) r %
                  [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                    THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                    THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                    THEN ASM_REWRITE_TAC[];

                    % WFenv r %
                    ASM_REWRITE_TAC[];

                    % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
                    ONCE_REWRITE_TAC[SL_APPEND_SYM23]
                    THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                    THEN ASM_REWRITE_TAC[];

                    % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
                      MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
                    MATCH_MP_TAC (SPEC_ALL MAP_EXTENSION_STATE_LEMMA)
                    THEN ASM_REWRITE_TAC[];

                    % 5th of 5 subgoals: %
                    % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
                    REWRITE_TAC[SYM(ASSUME "xs = l0 <|xs ss");
                                SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
                    THEN FIRST_ASSUM MATCH_MP_TAC
                    THEN EXISTS_TAC "(calls <|g
   ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))"
                    THEN REPEAT CONJ_TAC
                    THENL  % 5 subgoals: (of this 5th subgoal!) %
                     [ REWRITE_TAC[SYM(SPEC_ALL c_subst_o)]
                       THEN REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                       THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                       THEN ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[];

                       MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA)
                       THEN ASM_REWRITE_TAC[]
                     ]
                  ];

                 UNASSUME_THEN (ASSUME_TAC o SYM) "es = l1 <|es ss"
                 THEN IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL)
                                   es_subst_biject
               ]
            ]
         ]
     ]
  );;

%< c_substk_LEMMA1:
g "!c' r k s1 s2. 
      Ck c' r k s1 s2 ==>
      (!c ss g. (c' = (c <|c ss))
            /\ WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss ==>
        Ck c r k (s1 o ss) (s2 o ss))";;
e(CMDk_STR_INDUCT_TAC);;

% SKIP: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(SKIPk_TAC);;

% ASSIGN: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN ASM_REWRITE_THM BND_o_ONE_ONE_THM);;
 e(ASSIGNk_TAC);;
 e(IMP_RES2_THEN (REWRITE_THM o SYM o SPEC_ALL) e_subst_biject);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "e = e' <|e ss");;
 e(ASM_REWRITE_TAC[]);;

% SEQ: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(SEQk_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(EXISTS_TAC "k2:num");;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

% IF_T: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IFk_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(RES_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
        "WFcsubst c r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% IF_F: %
 e(GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(IFk_F_TAC);;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(RES_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst])
        "WFcsubst c r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% WHILE_T: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(WHILEk_T_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "(s2:state) o (ss:vsubst)");;
 e(EXISTS_TAC "(s3:state) o (ss:vsubst)");;
 e(EXISTS_TAC "k2:num");;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(RES_TAC);;
 e((STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
        "WFcsubst c'' r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[WFc_WHILE;c_subst;cmd_one_one;WFcsubst_THM]);;

% WHILE_F: %
 e(GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN GEN_TAC);;
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(WHILEk_F_TAC);;
 e(UNASSUME_THEN (ASSUME_TAC o SYM) "b = b' <|b ss");;
 e((STRIP_ASSUME_TAC o REWRITE_RULE[WFcsubst] o ASSUME)
        "WFcsubst c'' r ss");;
 e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) b_subst_biject);;

% CALL: %
 e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[WFc;c_subst;cmd_one_one;cmd_distinct]);;
 e(REWRITE_TAC[WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
 e(REPEAT STRIP_TAC);;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = s");;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
             "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(IMP_RES2_THEN REWRITE_THM BNDsL_o_biject);;
 e(ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD));;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL c_subst_STATE_BNDsL_LEMMA_STAGE));;

  e(ASM_REWRITE_TAC[MAP_MAP_o]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[MAP_I]);;

 e(USE_IMP_THEN REWRITE_THM
    ((SPEC_ALL o SPECL["variants vals (SL (APPEND l0 glbs))";"vals: var list"])
          c_subst_PROC_STATE_LEMMA));;

  e(ASM_REWRITE_TAC[]);;

 e(CALLk_TAC);;
 e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:var list"
    THEN EXISTS_TAC "vals:var list"
    THEN EXISTS_TAC "ns:num list"
    THEN EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC Ck_EXTENSION);;
  e(REWRITE_TAC (map (SYM o ASSUME)
                 ["vals' = variants vals(SL(APPEND xs glbs))";
                  "xs = l0 <|xs ss"]));;
  e(EXISTS_TAC "(calls <|g
     ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v (APPEND vars vals)))");;
  e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND l0 (variants vals(SL(APPEND l0 glbs)))) //v
                           (APPEND vars vals))z = y))");;
  e(REPEAT CONJ_TAC);; %5 subgoals%

   % WFc (c <|c ss2) r %
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   % WFenvk r k %
   e(ASM_REWRITE_TAC[]);;

   % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
   e(ONCE_REWRITE_TAC[SL_APPEND_SYM23]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
   e(MATCH_MP_TAC (SPEC_ALL MAP_EXTENSION_STATE_LEMMA));;
   e(ASM_REWRITE_TAC[]);;

   % 5th of 5 subgoals: %
   % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
   e(REWRITE_TAC[SYM(ASSUME "xs = l0 <|xs ss");
                 SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "(calls <|g
                    ((APPEND l0(variants vals(SL(APPEND l0 glbs)))) //v
                     (APPEND vars vals)))");;
   e(REPEAT CONJ_TAC);;  % 4 subgoals: %

    e(REWRITE_TAC[SYM(SPEC_ALL c_subst_o)]);;
    e(REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
    e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;

    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA));;
    e(ASM_REWRITE_TAC[]);;

  e(UNASSUME_THEN (ASSUME_TAC o SYM) "es = l1 <|es ss");;
  e(IMP_RES2_THEN (ASM_REWRITE_THM o SYM o SPEC_ALL) es_subst_biject);;

>%





let WFcsubst_IMP_biject =
 TAC_PROOF
  (([],
    "!c r ss. WFcsubst c r ss ==> ONE_ONE ss /\ ONTO ss"
   ),
   REWRITE_TAC[WFcsubst]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFcsubst_IMP_biject
g "!c r ss. WFcsubst c r ss ==> ONE_ONE ss /\ ONTO ss";;
e(REWRITE_TAC[WFcsubst]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%


let o_23_ASSOC =
 TAC_PROOF
  (([],
    "!(f:*4->*5) (g:*3->*4) (h:*2->*3) (i:*1->*2).
       ((f o g) o (h o i)) = (f o (g o h) o i)"
   ),
   REWRITE_TAC[o_ASSOC]
  );;

% o_23_ASSOC
g "!(f:*4->*5) (g:*3->*4) (h:*2->*3) (i:*1->*2).
       ((f o g) o (h o i)) = (f o (g o h) o i)";;
e(REWRITE_TAC[o_ASSOC]);;
%





let c_substk_LEMMA2 =
 TAC_PROOF
  (([],
    "!c r k s1' s2'. 
     Ck c r k s1' s2' ==>
     (!s1 s2 ss g.
         (s1' = (s1 o ss)) /\
         (s2' = (s2 o ss)) /\
         WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss  ==>
      Ck (c <|c ss) r k s1 s2)"
   ),
   ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN CMDk_STR_INDUCT_TAC
   THEN REWRITE_TAC[c_subst;WFc;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFcsubst_IMP_biject
   THEN IMP_RES2_THEN REWRITE_ALL_THM o_EQ_INVERSE
   THEN ASM_REWRITE_TAC[]
   THEN IMP_RES_TAC o_INVERSE
   THENL

        % SKIP: %
     [  SKIPk_TAC;

        % ASSIGN: %
        IMP_RES2_THEN REWRITE_THM BND_o_biject1
        THEN ASSIGNk_TAC
        THEN IMP_RES2_THEN REWRITE_THM e_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % SEQ: %
        SEQk_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst) z = y)"
        THEN EXISTS_TAC "k2:num"
        THEN ASM_REWRITE_TAC[]
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN CONJ_TAC
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % IF_T: %
        IFk_T_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % IF_F: %
        IFk_F_TAC
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % WHILE_T: %
        WHILEk_T_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN EXISTS_TAC "(s3:state) o (\y. @z. (ss:vsubst)z = y)"
        THEN EXISTS_TAC "k2:num"
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN CONJ_TAC
        THENL
         [ FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
           THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

           REWRITE_TAC[(SYM o ASSUME)"pr = ALESS v (AVAR x)"]
           THEN REWRITE_ALL_TAC[c_subst]
           THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
           THEN ASM_REWRITE_TAC[WFc_WHILE;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
           THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
         ];

        % WHILE_F: %
        WHILEk_F_TAC
        THEN IMP_RES2_THEN REWRITE_THM b_subst_biject
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID];

        % CALL: %
        REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]
        THEN IMP_RES2_THEN REWRITE_THM BNDsL_o_biject1
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
                    "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD)
        THEN IMP_RES_TAC ALL_LESS_WFenvk
        THEN USE_IMP_THEN REWRITE_THM
               (SPECL["xs:var list";"xs <|xs ss"]
                  (GENL["xs:var list";"l0:var list";"vals0:var list";"ss:vsubst"]
                     (SPEC_ALL c_subst_STATE_BNDsL_LEMMA_STAGE)))
        THENL
         [ ASM_REWRITE_TAC[MAP_MAP_o;SYM(SPEC_ALL xs_subst_o);
                           MAP_I;xs_subst_I;LENGTH_xs_subst]
           THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
           THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
           THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
           THEN BETA_TAC
           THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE
           THEN IMP_RES_TAC WFx_biject_INVERSE;

           USE_IMP_THEN REWRITE_THM
              ((SPEC_ALL o SPECL["variants vals (SL (APPEND (xs <|xs ss) glbs))";
                                 "vals: var list";"vals': var list";"xs: var list"])
                    c_subst_PROC_STATE_LEMMA)
           THENL
            [ ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);
                                   xs_subst_I;LENGTH_xs_subst]
              THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
              THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
              THEN BETA_TAC
              THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE;

              CALLk_TAC
              THEN EXISTS_TAC "c:cmd"
              THEN EXISTS_TAC "vars:var list"
              THEN EXISTS_TAC "vals:var list"
              THEN EXISTS_TAC "ns:num list"
              THEN EXISTS_TAC "glbs:var list"
              THEN EXISTS_TAC "pre:aexp"
              THEN EXISTS_TAC "post:aexp"
              THEN EXISTS_TAC "calls:prog_env"
              THEN EXISTS_TAC "rec:aexp"
              THEN ASM_REWRITE_TAC[]
              THEN REWRITE_TAC[SYM(ASSUME
                          "vals' = variants vals(SL(APPEND xs glbs))")]
              THEN CONJ_TAC
              THENL
               [ MATCH_MP_TAC Ck_EXTENSION
                 THEN EXISTS_TAC "(calls <|g
   ((APPEND(xs <|xs ss)(variants vals(SL(APPEND(xs <|xs ss)glbs)))) //v
    (APPEND vars vals)))"
                 THEN EXISTS_TAC
               "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND (xs <|xs ss)
                            (variants vals(SL(APPEND (xs <|xs ss) glbs)))) //v
                           (APPEND vars vals))z = y))"
                 THEN REPEAT CONJ_TAC %5 subgoals%
                 THENL
               % WFc (c <|c ss2) r %
                  [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                    THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                    THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                    THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                    THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst;

               % WFenv r %
                    ASM_REWRITE_TAC[];

               % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
                    ONCE_REWRITE_TAC[SL_APPEND_SYM23]
                    THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                    THEN ASM_REWRITE_TAC[LENGTH_xs_subst]
                    THEN IMP_RES2_THEN REWRITE_THM WFxs_subst
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
                    MATCH_MP_TAC
                         (SPECL["es <|es ss"]
                            (GENL["l1:exp list";"xs:var list";"l0:var list";
                                  "vals'':var list";"ss:vsubst"]
                               (SPEC_ALL MAP_EXTENSION_STATE_LEMMA)))
                    THEN ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;
                                         SYM(SPEC_ALL es_subst_o);es_subst_I;
                                         LENGTH_xs_subst;LENGTH_es_subst]
                    THEN IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst
                    THEN IMP_RES2_THEN REWRITE_THM INVERSE_biject
                    THEN BETA_TAC
                    THEN IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE;

               % 5th of 5 subgoals: %
               % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
                    FIRST_ASSUM
                     (MATCH_MP_TAC o
                      REWRITE_RULE[SYM(SPEC_ALL c_subst_o);
                                   SYM(SPEC_ALL o_ASSOC);vsubst_INVERSE;I_o_ID] o
                      SPEC "calls <|g (APPEND xs vals' //v APPEND vars vals)" o
                      SPEC "(APPEND (xs <|xs ss)
                                    (variants vals(SL(APPEND (xs <|xs ss) glbs))) //v
                             APPEND vars vals) o
                            (\y. @z. (APPEND xs vals' //v APPEND vars vals)z = y)"  o
                      SPECL["s1':state";"s2':state"])
                    THEN REPEAT CONJ_TAC
                    THENL  % 5 subgoals: %
                     [ ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[o_23_ASSOC]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]
                       THEN REWRITE_TAC[o_23_ASSOC]
                       THEN REWRITE_TAC[vsubst_INVERSE;I_o_ID];

                       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                       THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
                       THEN ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[]
                       THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA)
                       THEN ASM_REWRITE_TAC[]
                     ]
                  ];

                 IMP_RES2_THEN REWRITE_THM es_subst_biject
                 THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]
               ]
            ]
         ]
     ]
  );;

%< c_substk_LEMMA2:
g "!c r k s1' s2'. 
     Ck c r k s1' s2' ==>
     (!s1 s2 ss g.
         (s1' = (s1 o ss)) /\
         (s2' = (s2 o ss)) /\
         WFc c g r /\ WFenv_syntax r /\ WFcsubst c r ss  ==>
      Ck (c <|c ss) r k s1 s2)";;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(CMDk_STR_INDUCT_TAC
  THEN REWRITE_TAC[c_subst;WFc;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN IMP_RES_TAC WFcsubst_IMP_biject
  THEN IMP_RES2_THEN REWRITE_ALL_THM o_EQ_INVERSE
  THEN ASM_REWRITE_TAC[]
 );;

% SKIP: %
 e(SKIPk_TAC);;

% ASSIGN: %
 e(IMP_RES2_THEN REWRITE_THM BND_o_biject1);;
 e(ASSIGNk_TAC);;
 e(IMP_RES2_THEN REWRITE_THM e_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% SEQ: %
 e(SEQk_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst) z = y)");;
 e(EXISTS_TAC "k2:num");;
 e(ASM_REWRITE_TAC[]);;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% IF_T: %
 e(IFk_T_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% IF_F: %
 e(IFk_F_TAC);;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% WHILE_T: %
 e(WHILEk_T_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "(s2:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(EXISTS_TAC "(s3:state) o (\y. @z. (ss:vsubst)z = y)");;
 e(EXISTS_TAC "k2:num");;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

  e(REWRITE_TAC[(SYM o ASSUME) "pr = ALESS v (AVAR x)"]);;
  e(REWRITE_ALL_TAC[c_subst]);;
  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[WFc_WHILE;WFcsubst_THM;GVc;SUBSET;UNION_SUBSET]);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% WHILE_F: %
 e(WHILEk_F_TAC);;
 e(IMP_RES2_THEN REWRITE_THM b_subst_biject);;
 e(IMP_RES_TAC o_INVERSE);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

% CALL: %
 e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
 e(IMP_RES2_THEN REWRITE_THM BNDsL_o_biject1);;
 e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[o_THM;th] THEN ASSUME_TAC th)
             "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD));;
 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(USE_IMP_THEN REWRITE_THM
     (SPECL["xs:var list";"xs <|xs ss"]
        (GENL["xs:var list";"l0:var list";"vals0:var list";"ss:vsubst"]
           (SPEC_ALL c_subst_STATE_BNDsL_LEMMA_STAGE))));;

  e(ASM_REWRITE_TAC[MAP_MAP_o;SYM(SPEC_ALL xs_subst_o);
                    MAP_I;xs_subst_I;LENGTH_xs_subst]);;
  e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
  e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
  e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
  e(BETA_TAC);;
  e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;
  e(IMP_RES_TAC WFx_biject_INVERSE);;

 e(USE_IMP_THEN REWRITE_THM
    ((SPEC_ALL o SPECL["variants vals (SL (APPEND (xs <|xs ss) glbs))";
                       "vals: var list";"vals': var list";"xs: var list"])
          c_subst_PROC_STATE_LEMMA));;

  e(ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;LENGTH_xs_subst]);;
  e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
  e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
  e(BETA_TAC);;
  e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;

 e(CALLk_TAC);;
 e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:var list"
    THEN EXISTS_TAC "vals:var list"
    THEN EXISTS_TAC "ns:num list"
    THEN EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC Ck_EXTENSION);;
  e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o
                (((APPEND xs vals') //v
                  (APPEND vars vals)) o
                 (\y. @z. ((APPEND (xs <|xs ss)
                            (variants vals(SL(APPEND (xs <|xs ss) glbs)))) //v
                           (APPEND vars vals))z = y))");;
  e(REPEAT CONJ_TAC);; %5 subgoals%

   % WFc (c <|c ss2) r %
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
   e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;

   % WFenvk r k %
   e(ASM_REWRITE_TAC[]);;

   % FVc (c <|c ss2) r SUBSET (SL (APPEND (APPEND l0 glbs) vals'')) %
   e(ONCE_REWRITE_TAC[SL_APPEND_SYM23]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[LENGTH_xs_subst]);;
   e(IMP_RES2_THEN REWRITE_THM WFxs_subst);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;

   % MAP ((BNDsL ns vals' s2) o ss2)  (APPEND (APPEND l0 glbs) vals'') =
     MAP (BNDsL ns vals'' (s2 o ss1)) (APPEND (APPEND l0 glbs) vals'')   %
   e(MATCH_MP_TAC
     (SPECL["es <|es ss"]
        (GENL["l1:exp list";"xs:var list";"l0:var list";
              "vals'':var list";"ss:vsubst"]
           (SPEC_ALL MAP_EXTENSION_STATE_LEMMA))));;
   e(ASM_REWRITE_TAC[SYM(SPEC_ALL xs_subst_o);xs_subst_I;
                     SYM(SPEC_ALL es_subst_o);es_subst_I;
                     LENGTH_xs_subst;LENGTH_es_subst]);;
   e(IMP_RES2_THEN ASM_REWRITE_THM DL_APPEND_xs_subst);;
   e(IMP_RES2_THEN REWRITE_THM INVERSE_biject);;
   e(BETA_TAC);;
   e(IMP_RES2_THEN ASM_REWRITE_THM biject_EQ_INVERSE);;

   % 5th of 5 subgoals: %
   % C (c <|c ss1?) r ((BNDsL ns vals' s2) o ss2) (s3 o ss2) %
   e(FIRST_ASSUM
      (MATCH_MP_TAC o
       GEN_ALL  o
       REWRITE_RULE[SYM(SPEC_ALL c_subst_o);
                    SYM(SPEC_ALL o_ASSOC);vsubst_INVERSE;I_o_ID] o
       SPEC "calls <|g (APPEND xs vals' //v APPEND vars vals)" o
       SPEC "(APPEND (xs <|xs ss)
                     (variants vals(SL(APPEND (xs <|xs ss) glbs))) //v
              APPEND vars vals) o
             (\y. @z. (APPEND xs vals' //v APPEND vars vals)z = y)"  o
       SPECL["s1':state";"s2':state"]));;
   e(REPEAT CONJ_TAC);;  % 5 subgoals: %

    e(ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[o_23_ASSOC]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(ONCE_REWRITE_TAC[SYM(SPEC_ALL o_ASSOC)]);;
    e(REWRITE_TAC[o_23_ASSOC]);;
    e(REWRITE_TAC[vsubst_INVERSE;I_o_ID]);;

    e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
    e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;
    e(MATCH_MP_TAC (SPEC_ALL WFcsubst_subst_LEMMA));;
    e(ASM_REWRITE_TAC[]);;

  e(IMP_RES2_THEN REWRITE_THM es_subst_biject);;
  e(IMP_RES_TAC o_INVERSE);;
  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;I_o_ID]);;

>%



let c_substk_biject =
 prove_thm
  (`c_substk_biject`,
   "!c g r k s1 s2 ss. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r ss ==>
     (Ck (c <|c ss) r k s1 s2 = Ck c r k (s1 o ss) (s2 o ss))",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [ DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_substk_LEMMA1))
       THEN ASM_REWRITE_TAC[];

       DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_substk_LEMMA2))
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% c_substk_biject
g "!c g r k s1 s2 ss. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r ss ==>
     (Ck (c <|c ss) r k s1 s2 = Ck c r k (s1 o ss) (s2 o ss))";;

e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

  e(DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_substk_LEMMA1)));;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (MATCH_MP_TAC o SPEC_ALL o (MATCH_MP c_substk_LEMMA2)));;
  e(ASM_REWRITE_TAC[]);;
%




let c_substk_THM =
 prove_thm
  (`c_substk_THM`,
   "!c g r k s1 s2 ys xs. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) ==>
     (Ck (c <|c (ys //v xs)) r k s1 s2 =
      Ck c r k (s1 o (ys //v xs)) (s2 o (ys //v xs)))",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL c_substk_biject)
   THEN ASM_REWRITE_TAC[]
  );;

% c_substk_THM
g "!c g r k s1 s2 ys xs. 
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) ==>
     (Ck (c <|c (ys //v xs)) r k s1 s2 =
      Ck c r k (s1 o (ys //v xs)) (s2 o (ys //v xs)))";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL c_substk_biject));;
e(ASM_REWRITE_TAC[]);;
%




let PCSPECk_biject =
 prove_thm
  (`PCSPECk_biject`,
   "!ss x pre post c g r k.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (PCSPECk(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                   (pre <|av ss),
                   c    <|c  ss,
                   post <|av ss,
                   r, k) =
      PCSPECk(AAND (mk_eqs (logicals x) x) pre, c, post, r, k))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFcsubst_IMP_biject
   THEN REWRITE_TAC[PCSPECk;A_DEF]
   THEN IMP_RES2_THEN (REWRITE_THM o SYM) MAP_WFsubst_logicals
   THEN REWRITE_TAC[(SYM o SPEC_ALL) mk_eqs_vsubst]
   THEN IMP_RES_THEN REWRITE_THM a_vsubst_ONE_ONE
   THEN IMP_RES_TAC c_substk_biject
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN EQ_TAC
   THENL
     [  REPEAT STRIP_TAC
        THEN IMP_RES_TAC o_INVERSE
        THEN POP_ASSUM (\th1. POP_ASSUM (\th2.
             FIRST_ASSUM (MATCH_MP_TAC o
                  REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
                      SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)";
                            "(s2:state) o (\y. @z. (ss:vsubst)z = y)"])))
        THEN ASM_REWRITE_TAC[];

        REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "(s1:state) o (ss:vsubst)"
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% PCSPECk_biject
g "!ss x pre post c g r k.
     WFenv_syntax r /\ WFc c g r /\
     WFcsubst c r ss /\ WFsubst ss /\ WFxs x ==>
     (PCSPECk(AAND (mk_eqs (logicals (MAP ss x)) (MAP ss x))
                   (pre <|av ss),
                   c    <|c  ss,
                   post <|av ss,
                   r, k) =
      PCSPECk(AAND (mk_eqs (logicals x) x) pre, c, post, r, k))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFcsubst_IMP_biject);;
e(REWRITE_TAC[PCSPECk;A_DEF]);;
e(IMP_RES2_THEN (REWRITE_THM o SYM) MAP_WFsubst_logicals);;
e(REWRITE_TAC[(SYM o SPEC_ALL) mk_eqs_vsubst]);;
e(IMP_RES_THEN REWRITE_THM a_vsubst_ONE_ONE);;
e(IMP_RES_TAC c_substk_biject);;
e(ASM_REWRITE_TAC[]);;
e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC o_INVERSE);;
  e(POP_ASSUM (\th1. POP_ASSUM (\th2.
    FIRST_ASSUM (MATCH_MP_TAC o
         REWRITE_RULE[(SYM o SPEC_ALL) o_ASSOC;th1;th2;I_o_ID] o
         SPECL["(s1:state) o (\y. @z. (ss:vsubst)z = y)";
               "(s2:state) o (\y. @z. (ss:vsubst)z = y)"]))));;
  e(ASM_REWRITE_TAC[]);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "(s1:state) o (ss:vsubst)");;
  e(ASM_REWRITE_TAC[]);;
%



let PCSPECk_vsubst =
 prove_thm
  (`PCSPECk_vsubst`,
   "!ys xs x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) /\
     WFsubst (ys //v xs) /\ WFxs x ==>
     (PCSPECk(AAND (mk_eqs (logicals (MAP (ys //v xs) x))
                          (MAP (ys //v xs) x))
                   (pre <|av (ys //v xs)),
                   c    <|c  (ys //v xs),
                   post <|av (ys //v xs),
                   r, k) =
      PCSPECk(AAND (mk_eqs (logicals x) x) pre, c, post, r, k))",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL PCSPECk_biject)
   THEN ASM_REWRITE_TAC[]
  );;

% PCSPECk_vsubst
g "!ys xs x pre post c g r.
     WFenv_syntax r /\ WFc c g r /\ WFcsubst c r (ys //v xs) /\
     WFsubst (ys //v xs) /\ WFxs x ==>
     (PCSPECk(AAND (mk_eqs (logicals (MAP (ys //v xs) x))
                          (MAP (ys //v xs) x))
                   (pre <|av (ys //v xs)),
                   c    <|c  (ys //v xs),
                   post <|av (ys //v xs),
                   r, k) =
      PCSPECk(AAND (mk_eqs (logicals x) x) pre, c, post, r, k))";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL PCSPECk_biject));;
e(ASM_REWRITE_TAC[]);;
%



let Ck_BND_THM =
 prove_thm
  (`Ck_BND_THM`,
   "!c r k s1 s2.
     Ck c r k s1 s2 ==>
     (!g x n. WFc c g r /\ WFenv_syntax r /\ (~x IN FVc c r) ==>
       Ck c r k (BND n x s1) (BND n x s2))",
   CMDk_STR_INDUCT_TAC
   THEN REWRITE_TAC[WFc;FVc_DEF]
   THEN CONV_TAC (TOP_DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THENL
     [
        SKIPk_TAC;

        IMP_RES_TAC E_BND_THM
        THEN IMP_RES_THEN REWRITE_THM BND_COMMUTE
        THEN ASSIGNk_TAC
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC ALL_LESS_WFenvk
        THEN RES_TAC
        THEN SEQk_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN EXISTS_TAC "k2:num"
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN RES_TAC
        THEN IFk_T_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN RES_TAC
        THEN IFk_F_TAC
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN ASM_REWRITE_TAC[];

        UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')"
        THEN REWRITE_ALL_TAC[WFc_WHILE;FVc_DEF;IN_UNION;DE_MORGAN_THM]
        THEN IMP_RES_TAC B_BND_THM
        THEN WHILEk_T_TAC
        THEN EXISTS_TAC "k1:num"
        THEN EXISTS_TAC "BND n x (s2:state)"
        THEN EXISTS_TAC "BND n x (s3:state)"
        THEN EXISTS_TAC "k2:num"
        THEN ASM_REWRITE_TAC[]
        THEN CONJ_TAC
        THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC B_BND_THM
        THEN WHILEk_F_TAC
        THEN ASM_REWRITE_TAC[];

        UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
        THEN UNASSUME_THEN ASSUME_TAC
                "LENGTH (vals:(var)list) = LENGTH (es:(exp)list)"
        THEN IMP_RES_TAC ES_LENGTH_THM
        THEN POP_ASSUM (\en. POP_ASSUM (\ve. ASSUME_TAC (SYM (TRANS ve en))))
        THEN (ASSUME_TAC o SYM o
                REWRITE_RULE[SYM(ASSUME
                             "vals' = variants vals(SL(APPEND xs glbs))")] o
                SPECL["vals:var list";"SL(APPEND xs glbs: var list)"])
                             LENGTH_variants
        THEN POP_ASSUM (\vv'. POP_ASSUM (\nv. ASSUME_TAC (TRANS nv vv')))
        THEN IMP_RES_TAC ES_BND_THM
        THEN ASM_CASES_TAC "(x:var) IN SL vals'"
        THENL
          [
            IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BNDsL_MAP_BND_COMMUTE
            THEN CALLk_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:(var)list"
            THEN EXISTS_TAC "vals:(var)list"
            THEN EXISTS_TAC "ns:(num)list"
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN IMP_RES_TAC BNDsL_BND_ABSORPTION
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN CHECK_TAC;

            IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE
            THEN IMP_RES_THEN (SUBST1_TAC o SYM o ISPECL["n:num";"s2:state"])
                              MAP_BND_ABSORPTION
            THEN CALLk_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:(var)list"
            THEN EXISTS_TAC "vals:(var)list"
            THEN EXISTS_TAC "ns:(num)list"
            THEN EXISTS_TAC "glbs:(var)list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[SYM(ASSUME
                             "vals' = variants vals(SL(APPEND xs glbs))")]
            THEN IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsL_COMMUTE
            THEN FIRST_ASSUM MATCH_MP_TAC
            THEN EXISTS_TAC "(calls <|g ((APPEND xs vals') //v (APPEND vars vals)))"
            THEN ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD)
            THEN IMP_RES_TAC ALL_LESS_WFenvk
            THEN REPEAT CONJ_TAC
            THENL  % 3 subgoals %
              [  MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
                 THEN ASM_REWRITE_TAC[];

                 ASM_REWRITE_TAC[];

                 MATCH_MP_TAC NOT_IN_SUBSET
                 THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):var set"
                 THEN REWRITE_ALL_TAC[FVxs_SL]
                 THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
                 THEN ASM_REWRITE_TAC[]
                 THEN REWRITE_TAC[SYM(SPEC_ALL SL_APPEND);
                                  SYM(ASSUME "vals' =
                                       variants vals(SL(APPEND xs glbs))")]
                 THEN MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
                 THEN ASM_REWRITE_TAC[]
              ]
          ] 
      ]
  );;

%< Ck_BND_THM
g "!c r k s1 s2.
     Ck c r k s1 s2 ==>
     (!g x n. WFc c g r /\ WFenv_syntax r /\ (~x IN FVc c r) ==>
       Ck c r k (BND n x s1) (BND n x s2))";;
e(CMDk_STR_INDUCT_TAC
  THEN REWRITE_TAC[WFc;FVc_DEF]
  THEN REWRITE_TAC[o_THM;IN;IN_UNION;DE_MORGAN_THM]
  THEN REPEAT STRIP_TAC);;

 e(SKIPk_TAC);;

 e(IMP_RES_TAC E_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM BND_COMMUTE);;
 e(ASSIGNk_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC ALL_LESS_WFenvk);;
 e(RES_TAC);;
 e(SEQk_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(EXISTS_TAC "k2:num");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IFk_T_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(RES_TAC);;
 e(IFk_F_TAC);;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN REWRITE_ALL_THM "pr = ALESS v (AVAR x')");;
 e(REWRITE_ALL_TAC[WFc_WHILE;FVc_DEF;IN_UNION;DE_MORGAN_THM]);;
 e(IMP_RES_TAC B_BND_THM);;
 e(WHILEk_T_TAC);;
 e(EXISTS_TAC "k1:num");;
 e(EXISTS_TAC "BND n x (s2:state)");;
 e(EXISTS_TAC "BND n x (s3:state)");;
 e(EXISTS_TAC "k2:num");;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
  e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC B_BND_THM);;
 e(WHILEk_F_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
      "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
 e(UNASSUME_THEN ASSUME_TAC "LENGTH (vals:(var)list) = LENGTH (es:(exp)list)");;
 e(IMP_RES_TAC ES_LENGTH_THM);;
 e(POP_ASSUM (\en. POP_ASSUM (\ve. ASSUME_TAC (SYM (TRANS ve en)))));;
 e((ASSUME_TAC o SYM o
    REWRITE_RULE[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")] o
    SPECL["vals:var list";"SL(APPEND xs glbs: var list)"])
                 LENGTH_variants);;
 e(POP_ASSUM (\vv'. POP_ASSUM (\nv. ASSUME_TAC (TRANS nv vv'))));;
 e(IMP_RES_TAC ES_BND_THM);;
 e(ASM_CASES_TAC "(x:var) IN SL vals'");;

  e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BNDsL_MAP_BND_COMMUTE);;
  e(CALLk_TAC);;
  e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:(var)list"
    THEN EXISTS_TAC "vals:(var)list"
    THEN EXISTS_TAC "ns:(num)list"
    THEN EXISTS_TAC "glbs:(var)list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
  e(IMP_RES2_THEN REWRITE_THM BNDsL_BND_ABSORPTION);;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
  e(CHECK_TAC);;

  e(IMP_RES_THEN REWRITE_THM BND_BNDsL_COMMUTE);;
  e(IMP_RES_THEN (SUBST1_TAC o SYM o ISPECL["n:num";"s2:state"])
                 MAP_BND_ABSORPTION);;
  e(CALLk_TAC);;
  e(EXISTS_TAC "c:cmd"
    THEN EXISTS_TAC "vars:(var)list"
    THEN EXISTS_TAC "vals:(var)list"
    THEN EXISTS_TAC "ns:(num)list"
    THEN EXISTS_TAC "glbs:(var)list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp");;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
  e(IMP_RES_THEN (REWRITE_THM o SYM o SPEC_ALL) BND_BNDsL_COMMUTE);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "(calls <|g ((APPEND xs vals') //v (APPEND vars vals)))");;
  e(ASSUME_TAC (SPECL["k:num";"1"] LESS_EQ_ADD));;
  e(IMP_RES_TAC ALL_LESS_WFenvk);;
  e(REPEAT CONJ_TAC);;  % 3 subgoals %

   e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;

   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):var set");;
   e(REWRITE_ALL_TAC[FVxs_SL]);;
   e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_TAC[SYM(SPEC_ALL SL_APPEND);
            SYM(ASSUME "vals' = variants vals(SL(APPEND xs glbs))")]);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
   e(ASM_REWRITE_TAC[]);;
>%



let Ck_BNDsL_THM =
 prove_thm
  (`Ck_BNDsL_THM`,
   "!ns xs c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
      Ck c r k s1 s2 ==> Ck c r k (BNDsL ns xs s1) (BNDsL ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC Ck_BND_THM
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h:num")
   THEN RES_TAC
  );;

% Ck_BNDsL_THM:
g "!ns xs c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
      Ck c r k s1 s2 ==> Ck c r k (BNDsL ns xs s1) (BNDsL ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsL]
  THEN REPEAT STRIP_TAC);;
e(IMP_RES_TAC Ck_BND_THM);;
e(POP_ASSUM (ASSUME_TAC o SPEC "h:num"));;
e(RES_TAC);;
%

let Ck_BNDsR_THM =
 prove_thm
  (`Ck_BNDsR_THM`,
   "!ns xs c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
     Ck c r k s1 s2 ==> Ck c r k (BNDsR ns xs s1) (BNDsR ns xs s2)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC Ck_BND_THM
   THEN ASM_REWRITE_TAC[]
  );;

% Ck_BNDsR_THM:
g "!ns xs c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\ DISJOINT (SL xs) (FVc c r) /\
     Ck c r k s1 s2 ==> Ck c r k (BNDsR ns xs s1) (BNDsR ns xs s2)";;
e(DOUBLE_LIST_INDUCT_TAC
  THEN REWRITE_TAC[SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT;BNDsR]
  THEN REPEAT STRIP_TAC);;
e(RES_TAC);;
e(IMP_RES_TAC Ck_BND_THM);;
e(ASM_REWRITE_TAC[]);;
%




close_theory ();;

