% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains definitions of functions to substitute assertion-
  language expressions for variables within assertion-language expressions.
  This is both for a particular expression for a particular variable,
  and for simultaneous substitutions corresponding to a change in state.
  In fact, the function to apply simultaneous substitutions is defined
  first, and this is used to define the singular substitution function.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `substitutions`;;

new_parent `assert_semantics`;;
new_parent `variants`;;


% --------------------------------------------------------------------- %
% Need the ind_defs and finite_sets libraries                           %
% --------------------------------------------------------------------- %
%load_library `ind_defs`;;%
loadf `load_finite_sets`;;
load_finite_sets();;


% Set up autoloading from ancestor theories of substitutions: %

loadf `autoload`;;

autoload_all_defs_and_thms
 `assert_syntax assert_semantics bindings
  variants more_finite_sets`;;

loadf `semantics_tactics`;;




% ==================================================================== %
% SIMULTANEOUS SUBSTITUTIONS:                                          %
%   We conceptualize simultaneous substitutions as transformations of  %
%   a subject assertion-language expression by simultaneously and in   %
%   parallel, replacing each variable in the subject expression with   %
%   another expression, depending on the variable.  This is actually   %
%   an infinite number of substitutions; all variables are substituted,%
%   although for only a finite number of them will the new expression  %
%   be any change.  These simultaneous substitutions are represented   %
%   as mappings from identifiers to numeric-valued assertion           %
%   expressions.  Originally, all variables are mapped to themselves.  %
%                                                                      %
%   A substitution can also be considered to be applied to a program   %
%   state, in which case it is interpreted as an infinite parallel     %
%   assignment.  For each variable, its "substitution expression" is   %
%   evaluated in the current state to yield a value; these values are  %
%   simultaneously assigned to their original variables to yield the   %
%   resulting state.                                                   %
% ==================================================================== %

new_type_abbrev(`subst`, ":var -> vexp");;

% The basic identity symbolic state maps all variables to themselves.  %
% It has a particularly simple definition.                             %

let IDENT_SS = new_def
  `IDENT_SS`
  "IDENT_SS = AVAR";;

let IDENT_SS_var =
 prove_thm
  (`IDENT_SS_var`,
   "!x. IDENT_SS x = AVAR x",
   REWRITE_TAC[IDENT_SS]);;

let IDENT_SS_THM =
 prove_thm
  (`IDENT_SS_THM`,
   "!x s. V (IDENT_SS x) s = s x",
   REWRITE_TAC[IDENT_SS;V_DEF]);;

%
g "!x s. V (IDENT_SS x) s = s x";;
e(REWRITE_TAC[IDENT_SS;V_DEF]);;
%


% ===================================================================== %
% We use variants of the left-triangle symbol to indicate substitution. %
% A triangle (a <|a ss) indicates the application of a simultaneous     %
% substitution ss to an expression a.                                   %
% The final letter on the symbol indicates the type of the target being %
% modified by the substitution:  s = state, v = vexp, a = aexp.         %
%                                                                       %
% Later we shall introduce new operators to construct substitutions     %
% out of single substitutions (v / x), assignment (x := e), and so on.  %
% These will then use the above triangles to perform the substitutions. %
% ===================================================================== %


% We define (s <|s ss) as the state resulting from applying the         %
% transformation implicit in the simultaneous subst ss to the state s.  %

new_special_symbol `<|s`;;

let s_subst =
  new_infix_def
  `s_subst`
  "$<|s s ss = \x:var. V (ss x) s";;


let s_subst_IDENT =
 prove_thm
  (`s_subst_IDENT`,
   "!s. s <|s IDENT_SS = s",
   REWRITE_TAC[s_subst;IDENT_SS_THM;ETA_AX]);;

%
g "!s. s <|s IDENT_SS = s";;
e(REWRITE_TAC[s_subst]);;
e(REWRITE_TAC[IDENT_SS_THM]);;
e(REWRITE_TAC[ETA_AX]);;
%

let V_IDENT_subst =
 prove_thm
  (`V_IDENT_subst`,
   "!s ss x. V(IDENT_SS x)(s <|s ss) = V(ss x)s",
   REPEAT GEN_TAC
    THEN REWRITE_TAC[IDENT_SS_THM;s_subst]
    THEN BETA_TAC
    THEN REFL_TAC);;

%
g "!s ss x. V(IDENT_SS x)(s <|s ss) = V(ss x)s";;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[IDENT_SS_THM]);;
e(REWRITE_TAC[s_subst]);;
e(BETA_TAC);;
e(REFL_TAC);;
%


% BND is polymorphic, so it modifies a simulatenous substitution %
% with a new symbolic value for an id.                           %

% BND (a: *  ) (x:string) ( s:string -> *) instantiates to  %
% BND (v:vexp) (x:string) (ss:sym_state  ) , where          %
% BND v x ss  =  \y. ((y=x) => n | s y)                     %

let s_subst_BND =
 prove_thm
  (`s_subst_BND`,
   "!v x ss s. (s <|s BND v x ss) = BND (V v s) x (s <|s ss)",
   REWRITE_TAC[s_subst]
    THEN REPEAT GEN_TAC
    THEN EXT_TAC "y:var"
    THEN GEN_TAC
    THEN REWRITE_TAC[BND]
    THEN BETA_TAC
    THEN COND_CASES_TAC
    THEN REWRITE_TAC[]);;

%
g "!v x ss s. (s <|s BND v x ss) = BND (V v s) x (s <|s ss)";;
e(REWRITE_TAC[s_subst]);;
e(REPEAT GEN_TAC);;
e(EXT_TAC "y:var");;
e(GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(BETA_TAC);;
e(COND_CASES_TAC);;
  e(REWRITE_TAC[]);;
  e(REWRITE_TAC[]);;
%


let s_subst_BNDsL =
 prove_thm
  (`s_subst_BNDsL`,
   "!vs xs ss s. (s <|s BNDsL vs xs ss) = BNDsL (VS vs s) xs (s <|s ss)",
   DOUBLE_LIST_INDUCT_TAC
    THEN REWRITE_TAC[BNDsL;VS_DEF]
    THEN ASM_REWRITE_TAC[s_subst_BND]);;

%
g "!vs xs ss s. (s <|s BNDsL vs xs ss) = BNDsL (VS vs s) xs (s <|s ss)";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[VS_DEF;BNDsL]);;

e(REWRITE_TAC[VS_DEF;BNDsL]);;

e(REWRITE_TAC[VS_DEF;BNDsL]);;
e(ASM_REWRITE_TAC[s_subst_BND]);;
%


let s_subst_BNDsR =
 prove_thm
  (`s_subst_BNDsR`,
   "!vs xs ss s. (s <|s BNDsR vs xs ss) = BNDsR (VS vs s) xs (s <|s ss)",
   DOUBLE_LIST_INDUCT_TAC
    THEN REWRITE_TAC[BNDsR;VS_DEF]
    THEN ASM_REWRITE_TAC[s_subst_BND]);;

%
g "!vs xs ss s. (s <|s BNDsR vs xs ss) = BNDsR (VS vs s) xs (s <|s ss)";;
e(DOUBLE_LIST_INDUCT_TAC);;
e(REWRITE_TAC[VS_DEF;BNDsR]);;

e(REWRITE_TAC[VS_DEF;BNDsR]);;

e(REWRITE_TAC[VS_DEF;BNDsR]);;
e(ASM_REWRITE_TAC[s_subst_BND]);;
%

let NULL_BND_IDENT_SS =
 prove_thm
  (`NULL_BND_IDENT_SS`,
   "!x. BND (AVAR x) x IDENT_SS = IDENT_SS",
   GEN_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND]
   THEN COND_CASES_TAC
   THENL
     [ ASM_REWRITE_TAC[IDENT_SS];
       REFL_TAC
     ]
  );;

%
g "!x. BND (AVAR x) x IDENT_SS = IDENT_SS";;
e(GEN_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;
 e(ASM_REWRITE_TAC[IDENT_SS]);;
 e(REFL_TAC);;
%



% =================================================================== %
% We define application of a substitution to an assertion expression. %
% =================================================================== %

let v_subst =
    new_infix_rec_def `v_subst` vexp_Axiom
      "($<|v (ANUM n) ss       = ANUM n)  /\
       ($<|v (AVAR y) ss       = ss y)    /\
       ($<|v (APLUS  e1 e2) ss = APLUS  ($<|v e1 ss)
                                        ($<|v e2 ss))  /\
       ($<|v (AMINUS e1 e2) ss = AMINUS ($<|v e1 ss)
                                        ($<|v e2 ss))  /\
       ($<|v (AMULT  e1 e2) ss = AMULT  ($<|v e1 ss)
                                        ($<|v e2 ss))";;

let vs_subst =
    new_infix_list_rec_def `vs_subst`
      "($<|vs (NIL) ss        = NIL)  /\
       ($<|vs (CONS e es) ss  = CONS (e <|v ss)
                                     ($<|vs es ss))";;

let a_subst =
    new_infix_rec_def `a_subst` aexp_Axiom
      "($<|a (ATRUE) ss          = ATRUE)     /\
       ($<|a (AFALSE) ss         = AFALSE)    /\
       ($<|a (AEQ e1 e2) ss      = AEQ  (e1 <|v ss)
                                        (e2 <|v ss))  /\
       ($<|a (ALESS e1 e2) ss    = ALESS (e1 <|v ss)
                                         (e2 <|v ss)) /\
       ($<|a (ALLESS es1 es2) ss = ALLESS (es1 <|vs ss)
                                          (es2 <|vs ss)) /\
       ($<|a (AAND b1 b2) ss     = AAND ($<|a b1 ss)
                                        ($<|a b2 ss))  /\
       ($<|a (AOR  b1 b2) ss     = AOR  ($<|a b1 ss)
                                        ($<|a b2 ss))  /\
       ($<|a (ANOT b) ss         = ANOT ($<|a b ss))   /\
       ($<|a (AIMP b1 b2) ss     = AIMP ($<|a b1 ss)
                                        ($<|a b2 ss))  /\
       ($<|a (AEQB b1 b2) ss     = AEQB ($<|a b1 ss)
                                        ($<|a b2 ss))  /\
       ($<|a (ACOND b1 b2 b3) ss = ACOND ($<|a b1 ss)
                                         ($<|a b2 ss)
                                         ($<|a b3 ss)) /\
       ($<|a (ACLOSE b) ss       = ACLOSE b)           /\
       ($<|a (AFORALL y b) ss    = 
              let free = (UNION_SET (IMAGE (FVv o ss) (FVa b DELETE y))) in
              let z = variant y free in
                      (AFORALL z ($<|a b (BND (AVAR z) y ss)))
              )                                        /\
       ($<|a (AEXISTS y b) ss    = 
              let free = (UNION_SET (IMAGE (FVv o ss) (FVa b DELETE y))) in
              let z = variant y free in
                      (AEXISTS z ($<|a b (BND (AVAR z) y ss)))
              )  ";;


let v_subst_IDENT =
 prove_thm
  (`v_subst_IDENT`,
   "!v. v <|v IDENT_SS = v",
   INDUCT_THEN vexp_induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[v_subst;IDENT_SS]);;

%
g "!v. v <|v IDENT_SS = v";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;
 e(REWRITE_TAC[v_subst]);;
 e(REWRITE_TAC[v_subst;IDENT_SS]);;
 e(ASM_REWRITE_TAC[v_subst;IDENT_SS]);;
 e(ASM_REWRITE_TAC[v_subst;IDENT_SS]);;
 e(ASM_REWRITE_TAC[v_subst;IDENT_SS]);;
%

let vs_subst_IDENT =
 prove_thm
  (`vs_subst_IDENT`,
   "!vs. vs <|vs IDENT_SS = vs",
   LIST_INDUCT_TAC
    THEN ASM_REWRITE_TAC[vs_subst;v_subst_IDENT]);;

%
g "!vs. vs <|vs IDENT_SS = vs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vs_subst]);;

 e(REWRITE_TAC[vs_subst]);;
 e(ASM_REWRITE_TAC[v_subst_IDENT]);;
%


let IMAGE_IDENT_SS =
 prove_thm
  (`IMAGE_IDENT_SS`,
   "!s. UNION_SET (IMAGE (FVv o IDENT_SS) s) = s",
   SET_INDUCT_TAC
   THEN REWRITE_TAC[IMAGE;UNION_SET]
   THEN ASM_REWRITE_TAC[]
   THEN ASM_REWRITE_TAC[o_THM;IDENT_SS;FVv_DEF;UNION]
  );;

%
g "!s. UNION_SET (IMAGE (FVv o IDENT_SS) s) = s";;
e(SET_INDUCT_TAC);;
e(REWRITE_TAC[IMAGE;UNION_SET]);;

e(REWRITE_TAC[IMAGE;UNION_SET]);;
e(ASM_REWRITE_TAC[]);;
e(ASM_REWRITE_TAC[o_THM;IDENT_SS;FVv_DEF;UNION]);;
%

let a_subst_IDENT =
 prove_thm
  (`a_subst_IDENT`,
   "!b. b <|a IDENT_SS = b",
   INDUCT_THEN aexp_induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[a_subst;v_subst_IDENT;vs_subst_IDENT;IDENT_SS_var]
    THEN REWRITE_TAC[IMAGE_IDENT_SS]
    THEN CONV_TAC (DEPTH_CONV let_CONV)
    THEN ASM_REWRITE_TAC[variant_DELETE;NULL_BND_IDENT_SS]
);;

%
g "!b. b <|a IDENT_SS = b";;
e(INDUCT_THEN aexp_induct ASSUME_TAC
  THEN ASM_REWRITE_TAC[a_subst;v_subst_IDENT;vs_subst_IDENT;IDENT_SS_var]);;
e(REWRITE_TAC[IMAGE_IDENT_SS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[variant_DELETE]);;
e(REWRITE_TAC[NULL_BND_IDENT_SS]);;
e(ASM_REWRITE_TAC[]);;
%



let v_subst_THM =
 prove_thm
  (`v_subst_THM`,
   "!v ss s. V (v <|v ss) s = V v (s <|s ss)",
   INDUCT_THEN vexp_induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[v_subst;s_subst;V_DEF]
    THEN BETA_TAC
    THEN REWRITE_TAC[]);;

%
g "!v ss s. V (v <|v ss) s = V v (s <|s ss)";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;
 e(REWRITE_TAC[v_subst;V_DEF]);;

 e(REWRITE_TAC[v_subst;s_subst;V_DEF]);;
 e(BETA_TAC);;
 e(REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[v_subst;s_subst;V_DEF]);;
%

let vs_subst_THM =
 prove_thm
  (`vs_subst_THM`,
   "!vs ss s. VS (vs <|vs ss) s = VS vs (s <|s ss)",
   LIST_INDUCT_TAC
    THEN REWRITE_TAC[vs_subst;VS_DEF]
    THEN ASM_REWRITE_TAC[v_subst_THM]);;

%
g "!vs ss s. VS (vs <|vs ss) s = VS vs (s <|s ss)";;
e(LIST_INDUCT_TAC);;
 e(REWRITE_TAC[vs_subst;VS_DEF]);;

 e(REWRITE_TAC[vs_subst;VS_DEF]);;
 e(ASM_REWRITE_TAC[v_subst_THM]);;
%


let FVv_v_subst =
 prove_thm
  (`FVv_v_subst`,
   "!v ss. FVv( v <|v ss ) = UNION_SET (IMAGE (FVv o ss) (FVv v))",
   INDUCT_THEN vexp_induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[v_subst;FVv_DEF;IMAGE;UNION_SET;
                         IMAGE_UNION;UNION_SET_UNION]
    THEN ONCE_REWRITE_TAC[UNION_COMM]
    THEN REWRITE_TAC[o_THM;UNION]
  );;

%
g "!v ss. FVv( v <|v ss ) = UNION_SET (IMAGE (FVv o ss) (FVv v))";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;
 e(REWRITE_TAC[v_subst]);;
 e(REWRITE_TAC[FVv_DEF]);;
 e(REWRITE_TAC[IMAGE]);;
 e(REWRITE_TAC[UNION_SET]);;

 e(REWRITE_TAC[v_subst]);;
 e(REWRITE_TAC[FVv_DEF]);;
 e(REWRITE_TAC[IMAGE]);;
 e(REWRITE_TAC[o_THM]);;
 e(REWRITE_TAC[UNION_SET]);;
 e(ONCE_REWRITE_TAC[UNION_COMM]);;
 e(REWRITE_TAC[UNION]);;

 e(REWRITE_TAC[v_subst]);;
 e(REWRITE_TAC[FVv_DEF]);;
 e(REWRITE_TAC[IMAGE_UNION]);;
 e(REWRITE_TAC[UNION_SET_UNION]);;
 e(ASM_REWRITE_TAC[]);;
%

let FVvs_vs_subst =
 prove_thm
  (`FVvs_vs_subst`,
   "!vs ss. FVvs( vs <|vs ss ) = UNION_SET (IMAGE (FVv o ss) (FVvs vs))",
   LIST_INDUCT_TAC
    THEN REWRITE_TAC[vs_subst;FVvs_DEF;IMAGE;UNION_SET]
    THEN ASM_REWRITE_TAC[FVv_v_subst;IMAGE_UNION;UNION_SET_UNION]
  );;

%
g "!vs ss. FVvs( vs <|vs ss ) = UNION_SET (IMAGE (FVv o ss) (FVvs vs))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vs_subst]);;
 e(REWRITE_TAC[FVvs_DEF]);;
 e(REWRITE_TAC[IMAGE]);;
 e(REWRITE_TAC[UNION_SET]);;

 e(REWRITE_TAC[vs_subst]);;
 e(REWRITE_TAC[FVvs_DEF]);;
 e(REWRITE_TAC[IMAGE]);;
 e(REWRITE_TAC[UNION_SET]);;
 e(ASM_REWRITE_TAC[FVv_v_subst;IMAGE_UNION;UNION_SET_UNION]);;
%



let FVa_a_subst =
 prove_thm
  (`FVa_a_subst`,
   "!a ss. FVa( a <|a ss ) = UNION_SET (IMAGE (FVv o ss) (FVa a))",
   INDUCT_THEN aexp_induct ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_subst;FVa_DEF;FVv_v_subst;FVvs_vs_subst;
                        IMAGE;UNION_SET;IMAGE_UNION;UNION_SET_UNION]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[FVa_DEF]
   THEN ASM_REWRITE_TAC[]
   THEN REPEAT GEN_TAC
   THEN (ASM_CASES_TAC "v IN (FVa a)"
   THENL
     [ IMP_RES_THEN (\th. ONCE_REWRITE_TAC[SYM th]) INSERT_DELETE
       THEN REWRITE_TAC[DELETE;DELETE_DELETE;IMAGE;o_THM;BND_MATCH;FVv_DEF;
                        UNION_SET;UNION;IMAGE_COMPOSE;IMAGE_BND_DELETE]
       THEN REWRITE_TAC[SYM (SPEC_ALL IMAGE_COMPOSE);variant_not_in_set;DELETE];

       IMP_RES_THEN REWRITE_THM DELETE_NON_ELEMENT
       THEN REWRITE_TAC[IMAGE_COMPOSE]
       THEN IMP_RES_THEN REWRITE_THM IMAGE_BND_ABSORPTION
       THEN REWRITE_TAC[SYM (SPEC_ALL IMAGE_COMPOSE)]
     ]
   THEN REWRITE_TAC[SYM (SPEC_ALL DELETE_NON_ELEMENT)]
   THEN REWRITE_TAC[variant_not_in_set])
  );;

%
g "!a ss. FVa( a <|a ss ) = UNION_SET (IMAGE (FVv o ss) (FVa a))";;

e(INDUCT_THEN aexp_induct ASSUME_TAC  THEN
  ASM_REWRITE_TAC[a_subst;FVa_DEF;FVv_v_subst;FVvs_vs_subst;
                  IMAGE;UNION_SET;IMAGE_UNION;UNION_SET_UNION]);;

 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[FVa_DEF]);;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT GEN_TAC);;
 e(ASM_CASES_TAC "v IN (FVa a)");;

  e(IMP_RES_THEN (\th. ONCE_REWRITE_TAC[SYM th]) INSERT_DELETE);;
  e(REWRITE_TAC[DELETE]);;
  e(REWRITE_TAC[DELETE_DELETE]);;
  e(REWRITE_TAC[IMAGE]);;
  e(REWRITE_TAC[o_THM; BND_MATCH]);;
  e(REWRITE_TAC[FVv_DEF]);;
  e(REWRITE_TAC[UNION_SET]);;
  e(REWRITE_TAC[UNION]);;
  e(REWRITE_TAC[IMAGE_COMPOSE]);;
  e(REWRITE_TAC[IMAGE_BND_DELETE]);;
  e(REWRITE_TAC[SYM (SPEC_ALL IMAGE_COMPOSE)]);;
  e(REWRITE_TAC[variant_not_in_set]);;
  e(REWRITE_TAC[DELETE]);;
  e(REWRITE_TAC[SYM (SPEC_ALL DELETE_NON_ELEMENT)]);;
  e(REWRITE_TAC[variant_not_in_set]);;

  e(IMP_RES_THEN (REWRITE_THM) DELETE_NON_ELEMENT);;
  e(REWRITE_TAC[IMAGE_COMPOSE]);;
  e(IMP_RES_THEN REWRITE_THM IMAGE_BND_ABSORPTION);;
  e(REWRITE_TAC[SYM (SPEC_ALL IMAGE_COMPOSE)]);;
  e(REWRITE_TAC[SYM (SPEC_ALL DELETE_NON_ELEMENT)]);;
  e(REWRITE_TAC[variant_not_in_set]);;
%


let NOT_EQ_V_BND =
 prove_thm
  (`NOT_EQ_V_BND`,
   "!y n x st. ~(y = x) ==> (V(AVAR y)(BND n x st) = V(AVAR y)st)",
   REPEAT GEN_TAC
   THEN DISCH_THEN (ASSUME_TAC o NOT_EQ_SYM)
   THEN MATCH_MP_TAC V_BND_THM
   THEN ASM_REWRITE_TAC[FVv_DEF;IN]
  );;

%
g "!y n x st. ~(y = x) ==> (V(AVAR y)(BND n x st) = V(AVAR y)st)";;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (ASSUME_TAC o NOT_EQ_SYM));;
e(MATCH_MP_TAC V_BND_THM);;
e(ASM_REWRITE_TAC[FVv_DEF;IN]);;
%

let EQ_V_BND =
 prove_thm
  (`EQ_V_BND`,
   "!n x st. V(AVAR x)(BND n x st) = n",
   REWRITE_TAC[V_DEF;BND_MATCH]
  );;

%
g "!n x st. V(AVAR x)(BND n x st) = n";;
e(REWRITE_TAC[V_DEF;BND_MATCH]);;
%




let V_BND_s_subst =
 prove_thm
  (`V_BND_s_subst`,
   "!v ss n x st. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVv v))) ==>
                   (V v ((BND n x st) <|s ss) = V v (st <|s ss))",
   REWRITE_TAC[SYM (SPEC_ALL v_subst_THM)]  THEN
   REWRITE_TAC[SYM (SPEC_ALL FVv_v_subst)]  THEN
   REWRITE_TAC[V_BND_THM]
  );;

% V_BND_s_subst:
g "!v ss n x st. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVv v))) ==>
                  (V v ((BND n x st) <|s ss) = V v (st <|s ss))";;
e(REWRITE_TAC[SYM (SPEC_ALL v_subst_THM)]);;
e(REWRITE_TAC[SYM (SPEC_ALL FVv_v_subst)]);;
e(REWRITE_TAC[V_BND_THM]);;
%

let VS_BND_s_subst =
 prove_thm
  (`VS_BND_s_subst`,
   "!vs ss n x s. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVvs vs))) ==>
                   (VS vs ((BND n x s) <|s ss) = VS vs (s <|s ss))",
   REWRITE_TAC[SYM (SPEC_ALL vs_subst_THM)]  THEN
   REWRITE_TAC[SYM (SPEC_ALL FVvs_vs_subst)]  THEN
   REWRITE_TAC[VS_BND_THM]
  );;

% VS_BND_s_subst:
g "!vs ss n x s. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVvs vs))) ==>
                  (VS vs ((BND n x s) <|s ss) = VS vs (s <|s ss))";;
e(REWRITE_TAC[SYM (SPEC_ALL vs_subst_THM)]);;
e(REWRITE_TAC[SYM (SPEC_ALL FVvs_vs_subst)]);;
e(REWRITE_TAC[VS_BND_THM]);;
%




let BND_s_subst =
 prove_thm
  (`BND_s_subst`,
   "!ss n x s. (BND n x (s <|s ss) = s <|s (BND (ANUM n) x ss))",
   REWRITE_TAC[s_subst_BND;V_DEF]
  );;

% BND_s_subst:
g "!ss n x s. (BND n x (s <|s ss) = s <|s (BND (ANUM n) x ss))";;
e(REWRITE_TAC[s_subst_BND]);;
e(REWRITE_TAC[V_DEF]);;
%


let BND_s_subst_AVAR =
 prove_thm
  (`BND_s_subst_AVAR`,
   "!ss n x s. (BND n x s) <|s (BND (AVAR x) x ss) =
               (BND n x ((BND n x s) <|s ss))",
   REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]
  );;

% BND_s_subst_AVAR:
g "!ss n x s. (BND n x s) <|s (BND (AVAR x) x ss) =
              (BND n x ((BND n x s) <|s ss))";;
e(REWRITE_TAC[s_subst_BND]);;
e(REWRITE_TAC[V_DEF]);;
e(REWRITE_TAC[BND_MATCH]);;
%


let IMAGE_FV_DELETE =
 prove_thm
  (`IMAGE_FV_DELETE`,
   "!n s ss (z:var) . UNION_SET (IMAGE (FVv o ss) (s DELETE z)) =
                      UNION_SET (IMAGE (FVv o (BND (ANUM n) z ss)) s)",
   REWRITE_TAC[EXTENSION]
   THEN REWRITE_TAC[IN_UNION_SET;IN_IMAGE;IN_DELETE;o_THM]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [ STRIP_TAC
       THEN EXISTS_TAC "si:(var)set"
       THEN ASM_REWRITE_TAC[]
       THEN EXISTS_TAC "x':var"
       THEN IMP_RES_TAC (INST_TYPE [":vexp",":*"] BND_NOT_MATCH)
       THEN ASM_REWRITE_TAC[];

       STRIP_TAC
       THEN EXISTS_TAC "si:(var)set"
       THEN ASM_REWRITE_TAC[]
       THEN EXISTS_TAC "x':var"
       THEN ASM_CASES_TAC "(x':var) = z"
       THENL
         [ ASSUM_LIST (\[th4;th3;th2;th1].
               CONTR_TAC (REWRITE_RULE[th1;th4;BND_MATCH;FVv_DEF;IN]th3));

           IMP_RES_TAC (INST_TYPE [":vexp",":*"] BND_NOT_MATCH)
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

% IMAGE_FV_DELETE:
g "!n s ss (z:var) . UNION_SET (IMAGE (FVv o ss) (s DELETE z)) =
                     UNION_SET (IMAGE (FVv o (BND (ANUM n) z ss)) s)";;
e(REWRITE_TAC[EXTENSION]);;
e(REWRITE_TAC[IN_UNION_SET;IN_IMAGE;IN_DELETE;o_THM]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(EXISTS_TAC "si:(var)set");;
 e(ASM_REWRITE_TAC[]);;
 e(EXISTS_TAC "x':var");;
 e(IMP_RES_TAC (INST_TYPE [":vexp",":*"] BND_NOT_MATCH));;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(EXISTS_TAC "si:(var)set");;
 e(ASM_REWRITE_TAC[]);;
 e(EXISTS_TAC "x':var");;
 e(ASM_CASES_TAC "(x':var) = z");;

  e(ASSUM_LIST (\[th4;th3;th2;th1].
        CONTR_TAC (REWRITE_RULE[th1;th4;BND_MATCH;FVv_DEF;IN]th3)));;

  e(IMP_RES_TAC (INST_TYPE [":vexp",":*"] BND_NOT_MATCH));;
  e(ASM_REWRITE_TAC[]);;
%


let IN_IMAGE_FV_DELETE =
 prove_thm
  (`IN_IMAGE_FV_DELETE`,
   "!n s ss (z:var) x.
     ~(x IN (UNION_SET (IMAGE (FVv o ss) (s DELETE z)))) ==>
     ~(x IN (UNION_SET (IMAGE (FVv o (BND (ANUM n) z ss)) s)))",
   REWRITE_TAC[SYM (SPEC_ALL IMAGE_FV_DELETE)]
  );;

% IN_IMAGE_FV_DELETE:
g "!n s ss (z:var) x.
   ~(x IN (UNION_SET (IMAGE (FVv o ss) (s DELETE z)))) ==>
   ~(x IN (UNION_SET (IMAGE (FVv o (BND (ANUM n) z ss)) s)))";;
e(REWRITE_TAC[SYM (SPEC_ALL FV_IMAGE_DELETE)]);;
%


let A_BND_s_subst =
 prove_thm
  (`A_BND_s_subst`,
   "!a ss m x s. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVa a))) ==>
                  (A a ((BND m x s) <|s ss) = A a (s <|s ss))",
   INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[A_DEF;FVa_DEF;IMAGE_UNION;UNION_SET_UNION;
                    IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC V_BND_s_subst
   THEN IMP_RES_TAC VS_BND_s_subst
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN IMP_RES_TAC IN_IMAGE_FV_DELETE
   THEN ASSUM_LIST (\[th3;th2;th1].
          STRIP_ASSUME_TAC (GEN_ALL
                (MATCH_MP th1 (SPEC_ALL th3))) )
   THEN ASM_REWRITE_TAC[BND_s_subst]
  );;

% A_BND_s_subst:
g "!a ss m x s. ~(x IN UNION_SET (IMAGE (FVv o ss) (FVa a))) ==>
                 (A a ((BND m x s) <|s ss) = A a (s <|s ss))";;
e(INDUCT_THEN aexp_induct ASSUME_TAC
    THEN REWRITE_TAC[A_DEF;FVa_DEF;IMAGE_UNION;UNION_SET_UNION;
                     IN_UNION;DE_MORGAN_THM]
    THEN REPEAT STRIP_TAC
    THEN IMP_RES_TAC V_BND_s_subst
    THEN IMP_RES_TAC VS_BND_s_subst
   THEN RES_TAC
    THEN ASM_REWRITE_TAC[]);;

e(IMP_RES_TAC IN_IMAGE_FV_DELETE);;
e(ASSUM_LIST (\[th3;th2;th1].
    STRIP_ASSUME_TAC (GEN_ALL
          (MATCH_MP th1 (SPEC_ALL th3))) ));;
e(ASM_REWRITE_TAC[BND_s_subst]);;

e(IMP_RES_TAC IN_IMAGE_FV_DELETE);;
e(ASSUM_LIST (\[th3;th2;th1].
    STRIP_ASSUME_TAC (GEN_ALL
          (MATCH_MP th1 (SPEC_ALL th3))) ));;
e(ASM_REWRITE_TAC[BND_s_subst]);;
%

let A_BND_state_subst =
 prove_thm
  (`A_BND_state_subst`,
   "!v' ss a s n v.
    (!x. (x IN (FVa a DELETE v)) ==> ~(v' IN FVv (ss x))) ==>
    (A a (BND n v (s <|s ss)) = A a ((BND n v' s) <|s (BND (AVAR v') v ss)))",
   REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]
   THEN REWRITE_TAC[BND_s_subst]
   THEN REPEAT STRIP_TAC
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC A_BND_s_subst
   THEN MATCH_MP_TAC IN_IMAGE_FV_DELETE
   THEN REWRITE_TAC[IN_UNION_SET;IN_IMAGE]
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN REWRITE_TAC[DE_MORGAN_THM;o_THM]
   THEN ONCE_REWRITE_TAC[(SYM o SPEC_ALL) IMP_DISJ_THM]
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% A_BND_state_subst
g "!v' ss a s n v.
    (!x. (x IN (FVa a DELETE v)) ==> ~(v' IN FVv (ss x))) ==>
    (A a (BND n v (s <|s ss)) = A a ((BND n v' s) <|s (BND (AVAR v') v ss)))";;
e(REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]);;
e(REWRITE_TAC[BND_s_subst]);;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC A_BND_s_subst);;
e(MATCH_MP_TAC IN_IMAGE_FV_DELETE);;
e(REWRITE_TAC[IN_UNION_SET;IN_IMAGE]);;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(REWRITE_TAC[DE_MORGAN_THM;o_THM]);;
e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL) IMP_DISJ_THM]);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%


let A_variant_subst =
 prove_thm
  (`A_variant_subst`,
   "!a y ss s n.
  A a
    ((BND n (variant y (UNION_SET (IMAGE (FVv o ss) ((FVa a) DELETE y)))) s)
     <|s
     (BND (AVAR (variant y (UNION_SET (IMAGE (FVv o ss) ((FVa a) DELETE y)))))
          y ss)) =
  A a(BND n y (s <|s ss))",
   REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]
   THEN REWRITE_TAC[BND_s_subst]
   THEN REPEAT GEN_TAC
   THEN MATCH_MP_TAC A_BND_s_subst
   THEN MATCH_MP_TAC IN_IMAGE_FV_DELETE
   THEN REWRITE_TAC[variant_not_in_set]
  );;

%
g "!a y ss s n.
  A a
    ((BND n (variant y (UNION_SET (IMAGE (FVv o ss) ((FVa a) DELETE y)))) s)
     <|s
     (BND (AVAR (variant y (UNION_SET (IMAGE (FVv o ss) ((FVa a) DELETE y)))))
          y ss)) =
  A a(BND n y (s <|s ss))";;
e(REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]);;
e(REWRITE_TAC[BND_s_subst]);;
e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC A_BND_s_subst);;
e(MATCH_MP_TAC IN_IMAGE_FV_DELETE);;
e(REWRITE_TAC[variant_not_in_set]);;
%


let a_subst_THM =
 prove_thm
  (`a_subst_THM`,
   "!a ss s. A (a <|a ss) s = A a (s <|s ss)",
   INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[a_subst]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN ASM_REWRITE_TAC[v_subst_THM;vs_subst_THM;A_DEF;A_variant_subst]
  );;

%
g "!a ss s. A (a <|a ss) s = A a (s <|s ss)";;
e(INDUCT_THEN aexp_induct ASSUME_TAC
    THEN REWRITE_TAC[a_subst]
    THEN CONV_TAC (DEPTH_CONV let_CONV)
    THEN ASM_REWRITE_TAC[v_subst_THM;vs_subst_THM;A_DEF;A_variant_subst]
);;
%




% ==================================================== %
%                 Single substitutions                 %
% ==================================================== %

let single_subst = 
    new_infix_def
      `single_subst`
       "$/ (e:vexp) (x:var) = BND e x IDENT_SS";;

let single_subst_THM =
 prove_thm
  (`single_subst_THM`,
   "!e x s. s <|s (e/x) = BND (V e s) x s",
   REWRITE_TAC[single_subst;s_subst_BND;s_subst_IDENT]
  );;

% single_subst_THM:
g "!e x s. s <|s (e/x) = BND (V e s) x s";;
e(REWRITE_TAC[single_subst]);;
e(REWRITE_TAC[s_subst_BND]);;
e(REWRITE_TAC[s_subst_IDENT]);;
%



% ==================================================== %
%              Variable list substitutions             %
% ==================================================== %

% Originally, we defined // as

let var_list_subst = 
    new_infix_list_rec_definition (`var_list_subst`,
      "($// (NIL) xs = IDENT_SS)  /\
       ($// (CONS y ys) xs = BND (AVAR y) (HD xs) ($// ys (TL xs)))");;

but now we use:  %

let var_list_subst = 
    new_infix_definition (`var_list_subst`,
      "$// ys xs = BNDsR (MAP AVAR ys) xs IDENT_SS");;


% Alternatively,

let var_list_subst = 
    new_infix_list_rec_definition (`var_list_subst`,
      "($// ys (NIL) = IDENT_SS)  /\
       ($// ys (CONS x xs) = BND (AVAR (HD ys)) x ($// (TL ys) xs))");;

or now we could use:  

let var_list_subst = 
    new_infix_definition (`var_list_subst`,
      "$// ys xs = BNDsL (MAP AVAR ys) xs IDENT_SS");;
%


let var_list_THM =
 prove_thm
  (`var_list_THM`,
   "!ys xs y x.
     ([] // xs = IDENT_SS) /\
     (ys // [] = IDENT_SS) /\
     ((CONS y ys) // (CONS x xs) = BND (AVAR y) x (ys // xs))",
   REWRITE_TAC[var_list_subst;MAP;BNDsR]
  );;

% var_list_THM
g "!ys xs y x.
     ([] // xs = IDENT_SS) /\
     (ys // [] = IDENT_SS) /\
     ((CONS y ys) // (CONS x xs) = BND (AVAR y) x (ys // xs))";;
e(REWRITE_TAC[var_list_subst]);;
e(REWRITE_TAC[MAP;BNDsR]);;
%


let var_list_IDENT =
 prove_thm
  (`var_list_IDENT`,
   "!xs. (xs // xs) = IDENT_SS",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[var_list_THM]
   THEN GEN_TAC
   THEN EXT_TAC "x:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND]
   THEN COND_CASES_TAC
   THEN ASM_REWRITE_TAC[IDENT_SS]
  );;

%< var_list_IDENT
g "!xs. (xs // xs) = IDENT_SS";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[var_list_THM]);;

  e(ASM_REWRITE_TAC[var_list_THM]);;
  e(GEN_TAC);;
  e(EXT_TAC "x:var");;
  e(GEN_TAC);;
  e(REWRITE_TAC[BND]);;
  e(COND_CASES_TAC);;

    e(REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[IDENT_SS]);;
>%


let VS_MAP_AVAR =
 prove_thm
  (`VS_MAP_AVAR`,
   "!xs s. VS (MAP AVAR xs) s = MAP s xs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;VS_DEF]
   THEN ASM_REWRITE_TAC[V_DEF]
  );;

%
g "!xs s. VS (MAP AVAR xs) s = MAP s xs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;VS_DEF]);;

 e(REWRITE_TAC[MAP;VS_DEF]);;
 e(ASM_REWRITE_TAC[V_DEF]);;
%



let s_subst_var_list =
 prove_thm
  (`s_subst_var_list`,
   "!ys xs s. s <|s (ys // xs) = BNDsR (MAP s ys) xs s",
   REWRITE_TAC[var_list_subst;s_subst_BNDsR;s_subst_IDENT;VS_MAP_AVAR]
  );;

% s_subst_var_list:
g "!ys xs s. s <|s (ys // xs) = BNDsR (MAP s ys) xs s";;

e(REWRITE_TAC[var_list_subst]);;
e(REWRITE_TAC[s_subst_BNDsR]);;
e(REWRITE_TAC[s_subst_IDENT]);;
e(REWRITE_TAC[VS_MAP_AVAR]);;
%


let A_var_list_subst_THM =
 prove_thm
  (`A_var_list_subst_THM`,
   "!y x a s.
    (DISJOINT (SL x) (FVa a)) ==>
      (A a (s <|s (y // x)) = A a s)",
   REWRITE_TAC[s_subst_var_list;A_BNDsR_THM]
  );;

% A_var_list_subst_THM:
g "!y x a s.
    (DISJOINT (SL x) (FVa a)) ==>
      (A a (s <|s (y // x)) = A a s)";;
e(REWRITE_TAC[s_subst_var_list;A_BNDsR_THM]);;
%


let FVv_var_list_subst =
 prove_thm
  (`FVv_var_list_subst`,
   "!x y z.
     FVv ((y // x)z) = INSERT (BNDsR y x I z) EMPTY",
   REWRITE_TAC[var_list_subst]
   THEN ONCE_REWRITE_TAC[(SYM o CONJUNCT2 o ISPEC "IDENT_SS") I_o_ID]
   THEN REWRITE_TAC[IDENT_SS;(SYM o SPEC_ALL) o_BNDsR]
   THEN REWRITE_TAC[o_THM;FVv_DEF]
  );;

% FVv_var_list_subst
g "!x y z.
     FVv ((y // x)z) = INSERT (BNDsR y x I z) EMPTY";;
e(REWRITE_TAC[var_list_subst]);;
e(ONCE_REWRITE_TAC[(SYM o CONJUNCT2 o ISPEC "IDENT_SS") I_o_ID]);;
e(REWRITE_TAC[IDENT_SS;(SYM o SPEC_ALL) o_BNDsR]);;
e(REWRITE_TAC[o_THM;FVv_DEF]);;
%


let FVa_var_list_subst =
 prove_thm
  (`FVa_var_list_subst`,
   "!a y x. FVa (a <|a (y // x)) = IMAGE (BNDsR y x I) (FVa a)",
   REWRITE_TAC[FVa_a_subst]
   THEN REWRITE_TAC[EXTENSION]
   THEN REPEAT GEN_TAC
   THEN REWRITE_TAC[IN_UNION_SET;IN_UNION;IN_IMAGE]
   THEN REWRITE_TAC[o_THM;FVv_var_list_subst]
   THEN EQ_TAC
   THENL
    [ STRIP_TAC
      THEN EXISTS_TAC "x'':var"
      THEN POP_ASSUM MP_TAC
      THEN ASM_REWRITE_TAC[IN];

      STRIP_TAC
      THEN EXISTS_TAC "{x':var}"
      THEN REWRITE_TAC[IN]
      THEN EXISTS_TAC "x'':var"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< FVa_var_list_subst
g "!a y x. FVa (a <|a (y // x)) = IMAGE (BNDsR y x I) (FVa a)";;
e(REWRITE_TAC[FVa_a_subst]);;
e(REWRITE_TAC[EXTENSION]);;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[IN_UNION_SET;IN_UNION;IN_IMAGE]);;
e(REWRITE_TAC[o_THM;FVv_var_list_subst]);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "x'':var");;
  e(POP_ASSUM MP_TAC);;
  e(ASM_REWRITE_TAC[IN]);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "{x':var}");;
  e(REWRITE_TAC[IN]);;
  e(EXISTS_TAC "x'':var");;
  e(ASM_REWRITE_TAC[]);;
>%



% ==================================================================== %
% We now prove that the substitution of lists of variables for lists,  %
% where the lists are each composed of two smaller lists, is equal to  %
% the substitution of just the first lists in each IF the free vars    %
% of the subject aexp are within the first list of the source vars.    %
% ==================================================================== %

let var_NULL_APPEND_subst =
 prove_thm
  (`var_NULL_APPEND_subst`,
   "!y x z w v.
     (LENGTH x = LENGTH y) ==> (v IN SL x)   ==>
     ((APPEND y z // APPEND x w) v = (y // x) v)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;APPEND;IN;var_list_THM]
   THEN ASM_CASES_TAC "(v:var) = h'"
   THEN ASM_REWRITE_TAC[BND_MATCH]
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% var_NULL_APPEND_subst
g "!y x z w v.
     (LENGTH x = LENGTH y) ==> (v IN SL x)   ==>
     ((APPEND y z // APPEND x w) v = (y // x) v)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;APPEND;IN;var_list_THM]);;

  e(REWRITE_TAC[SET_LIST;APPEND;IN;var_list_THM]);;
  e(ASM_CASES_TAC "(v:var) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(REPEAT STRIP_TAC);;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%

let V_NULL_APPEND_subst =
 prove_thm
  (`V_NULL_APPEND_subst`,
   "!v y x z w s.
      (LENGTH x = LENGTH y) /\ (FVv v SUBSET SL x)   ==>
      (V v (s <|s (APPEND y z // APPEND x w)) = V v (s <|s (y // x)))",
   INDUCT_THEN vexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET;v_subst;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC var_NULL_APPEND_subst
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[s_subst]
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% V_NULL_APPEND_subst
g "!v y x z w s.
      (LENGTH x = LENGTH y) /\ (FVv v SUBSET SL x)   ==>
      (V v (s <|s (APPEND y z // APPEND x w)) = V v (s <|s (y // x)))";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;

  e(REWRITE_TAC[FVv_DEF;SUBSET;v_subst;V_DEF]);;

  e(REWRITE_TAC[FVv_DEF;SUBSET;v_subst;V_DEF]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC var_NULL_APPEND_subst);;
  e(REWRITE_TAC[s_subst]);;
  e(BETA_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET;v_subst;V_DEF]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let VS_NULL_APPEND_subst =
 prove_thm
  (`VS_NULL_APPEND_subst`,
   "!vs y x z w s.
      (LENGTH x = LENGTH y) /\ (FVvs vs SUBSET SL x)   ==>
      (VS vs (s <|s (APPEND y z // APPEND x w)) = VS vs (s <|s (y // x)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVvs_DEF;SUBSET;VS_DEF]
   THEN REWRITE_TAC[UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC V_NULL_APPEND_subst
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% VS_NULL_APPEND_subst
g "!vs y x z w s.
      (LENGTH x = LENGTH y) /\ (FVvs vs SUBSET SL x)   ==>
      (VS vs (s <|s (APPEND y z // APPEND x w)) = VS vs (s <|s (y // x)))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[FVvs_DEF;SUBSET;VS_DEF]);;

  e(REWRITE_TAC[FVvs_DEF;SUBSET;VS_DEF]);;
  e(REWRITE_TAC[UNION_SUBSET]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC V_NULL_APPEND_subst);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let A_BND_NULL_APPEND_subst =
 prove_thm
  (`A_BND_NULL_APPEND_subst`,
   "!a n v s x y w z.
     ((FVa a) DELETE v) SUBSET (SL x) /\
     (LENGTH x = LENGTH y)
   ==>
     (A a (BND n v (s <|s (APPEND y z // APPEND x w))) =
      A a ((BND n (variant v (SL y)) s) <|s
           (APPEND (CONS (variant v (SL y)) y) z // APPEND (CONS v x) w)))",
   REWRITE_TAC[APPEND;var_list_THM;SUBSET_DEF]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC A_BND_state_subst
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN REWRITE_TAC[FVv_var_list_subst;IN]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC IN_NOT_IN
   THEN EXISTS_TAC "SL (y:(var)list)"
   THEN IMP_RES_TAC BNDsR_APPEND_MEMBER
   THEN ASM_REWRITE_TAC[variant_not_in_set]
  );;

% A_BND_NULL_APPEND_subst
g "!a n v s x y w z.
     ((FVa a) DELETE v) SUBSET (SL x) /\
     (LENGTH x = LENGTH y)
   ==>
     (A a (BND n v (s <|s (APPEND y z // APPEND x w))) =
      A a ((BND n (variant v (SL y)) s) <|s
           (APPEND (CONS (variant v (SL y)) y) z // APPEND (CONS v x) w)))";;

e(REWRITE_TAC[APPEND;var_list_THM;SUBSET_DEF]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC A_BND_state_subst);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(RES_TAC);;
e(REWRITE_TAC[FVv_var_list_subst;IN]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC IN_NOT_IN);;
e(EXISTS_TAC "SL (y:(var)list)");;
e(IMP_RES_THEN (IMP_RES_THEN REWRITE_THM) BNDsR_APPEND_MEMBER);;
e(REWRITE_TAC[variant_not_in_set]);;
%

let A_BND_var_list_subst =
 prove_thm
  (`A_BND_var_list_subst`,
   "!a n v s x y.
     ((FVa a) DELETE v) SUBSET (SL x) /\
     (LENGTH x = LENGTH y)
   ==>
     (A a (BND n v (s <|s (y // x))) =
      A a ((BND n (variant v (SL y)) s) <|s
           (CONS (variant v (SL y)) y // CONS v x)))",
   REWRITE_TAC[var_list_THM;SUBSET_DEF]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC A_BND_state_subst
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN REWRITE_TAC[FVv_var_list_subst;IN]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC IN_NOT_IN
   THEN EXISTS_TAC "SL (y:(var)list)"
   THEN IMP_RES_TAC BNDsR_MEMBER
   THEN ASM_REWRITE_TAC[variant_not_in_set]
  );;

% A_BND_var_list_subst
g "!a n v s x y.
     ((FVa a) DELETE v) SUBSET (SL x) /\
     (LENGTH x = LENGTH y)
   ==>
     (A a (BND n v (s <|s (y // x))) =
      A a ((BND n (variant v (SL y)) s) <|s
           (CONS (variant v (SL y)) y // CONS v x)))";;

... use v' = variant v (SL y).  Then v' is not in the list y.

e(REWRITE_TAC[var_list_THM;SUBSET_DEF]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC A_BND_state_subst);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(RES_TAC);;
e(REWRITE_TAC[FVv_var_list_subst;IN]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC IN_NOT_IN);;
e(EXISTS_TAC "SL (y:(var)list)");;
e(IMP_RES_TAC BNDsR_MEMBER);;
e(ASM_REWRITE_TAC[variant_not_in_set]);;
%

let A_NULL_APPEND_subst =
 prove_thm
  (`A_NULL_APPEND_subst`,
   "!a y x z w s.
      (LENGTH x = LENGTH y) /\ (FVa a SUBSET SL x)   ==>
      (A (a <|a (APPEND y z // APPEND x w)) s = A (a <|a (y // x)) s)",
   REWRITE_TAC[a_subst_THM]
   THEN INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVa_DEF;SUBSET;UNION_SUBSET;a_subst;A_DEF;VS_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC var_NULL_APPEND_subst
   THEN IMP_RES_TAC V_NULL_APPEND_subst
   THEN IMP_RES_TAC VS_NULL_APPEND_subst
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN (FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC)
   THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
   THEN IMP_RES_TAC A_BND_NULL_APPEND_subst
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN IMP_RES_TAC A_BND_var_list_subst
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM REWRITE_THM
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[LENGTH;SET_LIST;SUBSET_INSERT_DELETE]
  );;

% A_NULL_APPEND_subst
g "!a y x z w s.
      (LENGTH x = LENGTH y) /\ (FVa a SUBSET SL x)   ==>
      (A (a <|a (APPEND y z // APPEND x w)) s = A (a <|a (y // x)) s)";;
e(REWRITE_TAC[a_subst_THM]);;
e(INDUCT_THEN aexp_induct ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;SUBSET;UNION_SUBSET;a_subst;A_DEF;VS_DEF;V_DEF]
  THEN REPEAT STRIP_TAC
  THEN IMP_RES_TAC var_NULL_APPEND_subst
  THEN IMP_RES_TAC V_NULL_APPEND_subst
  THEN IMP_RES_TAC VS_NULL_APPEND_subst
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[]);;

  e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
  e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
  e(IMP_RES_TAC A_BND_NULL_APPEND_subst);;
  e(POP_TAC THEN POP_TAC);;
  e(POP_ASSUM REWRITE_THM);;
  e(IMP_RES_TAC A_BND_var_list_subst);;
  e(POP_TAC THEN POP_TAC);;
  e(POP_ASSUM REWRITE_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[LENGTH;SET_LIST;SUBSET_INSERT_DELETE]);;

%





let APPEND_var_list_subst =
 prove_thm
  (`APPEND_var_list_subst`,
   "!y x w z.
     (LENGTH x = LENGTH y) ==>
     ((APPEND y z // APPEND x w) = BNDsR(MAP AVAR y)x(z // w))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;MAP;BNDsR]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[var_list_THM]
  );;

% APPEND_var_list_subst
g "!y x w z.
     (LENGTH x = LENGTH y) ==>
     ((APPEND y z // APPEND x w) = BNDsR(MAP AVAR y)x(z // w))";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;MAP;BNDsR]);;

  e(REWRITE_TAC[APPEND;MAP;BNDsR]);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[var_list_THM]);;
%

let APPEND_var_list_COMMUTE =
 prove_thm
  (`APPEND_var_list_COMMUTE`,
   "!y x w z.
     (LENGTH x = LENGTH y) /\ (LENGTH w = LENGTH z) /\
     DISJOINT(SL x)(SL w) ==>
     ((APPEND y z // APPEND x w) = (APPEND z y // APPEND w x))",
   REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM APPEND_var_list_subst
   THEN REWRITE_TAC[var_list_subst]
   THEN IMP_RES_THEN REWRITE_THM BNDsR_BNDsR_COMMUTE
  );;

% APPEND_var_list_COMMUTE
g "!y x w z.
     (LENGTH x = LENGTH y) /\ (LENGTH w = LENGTH z) /\
     DISJOINT(SL x)(SL w) ==>
     ((APPEND y z // APPEND x w) = (APPEND z y // APPEND w x))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM APPEND_var_list_subst);;
e(REWRITE_TAC[var_list_subst]);;
e(IMP_RES_THEN REWRITE_THM BNDsR_BNDsR_COMMUTE);;
%

let s_subst_APPEND_var_list =
 prove_thm
  (`s_subst_APPEND_var_list`,
   "!y x w z s.
     (LENGTH x = LENGTH y) ==>
     (s <|s (APPEND y z // APPEND x w) = BNDsR(MAP s y)x (BNDsR(MAP s z)w s))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;MAP;BNDsR]
   THENL
     [  REWRITE_TAC[s_subst_var_list];

        RES_TAC
        THEN ASM_REWRITE_TAC[var_list_THM;s_subst_BND;V_DEF]
     ]
  );;

% s_subst_APPEND_var_list
g "!y x w z s.
     (LENGTH x = LENGTH y) ==>
     (s <|s (APPEND y z // APPEND x w) = BNDsR(MAP s y)x (BNDsR(MAP s z)w s))";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;MAP;BNDsR]);;
  e(REWRITE_TAC[s_subst_var_list]);;

  e(REWRITE_TAC[APPEND;MAP;BNDsR]);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[var_list_THM;s_subst_BND;V_DEF]);;
%

let s_subst_APPEND_var_list_COMPOSE =
 prove_thm
  (`s_subst_APPEND_var_list_COMPOSE`,
   "!y x w z s.
     (LENGTH x = LENGTH y) /\ DISJOINT(SL y)(SL w) ==>
     (s <|s (APPEND y z // APPEND x w) = ((s <|s (z // w)) <|s (y // x)))",
   REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM s_subst_APPEND_var_list
   THEN REWRITE_TAC[s_subst_var_list]
   THEN IMP_RES_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
  );;

% s_subst_APPEND_var_list_COMPOSE
g "!y x w z s.
     (LENGTH x = LENGTH y) /\ DISJOINT(SL y)(SL w) ==>
     (s <|s (APPEND y z // APPEND x w) = ((s <|s (z // w)) <|s (y // x)))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM s_subst_APPEND_var_list);;
e(REWRITE_TAC[s_subst_var_list]);;
e(IMP_RES_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;
%

let a_subst_APPEND_var_list_COMPOSE =
 prove_thm
  (`a_subst_APPEND_var_list_COMPOSE`,
   "!y x w z a s.
     (LENGTH x = LENGTH y) /\ DISJOINT(SL y)(SL w) ==>
     (A  (a <|a (APPEND y z // APPEND x w)) s =
      A ((a <|a (y // x)) <|a (z // w)) s)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC s_subst_APPEND_var_list_COMPOSE
   THEN ASM_REWRITE_TAC[a_subst_THM]
  );;

% a_subst_APPEND_var_list_COMPOSE
g "!y x w z a s.
     (LENGTH x = LENGTH y) /\ DISJOINT(SL y)(SL w) ==>
     (A  (a <|a (APPEND y z // APPEND x w)) s =
      A ((a <|a (y // x)) <|a (z // w)) s)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[a_subst_THM]);;
e(IMP_RES_TAC s_subst_APPEND_var_list_COMPOSE);;
e(ASM_REWRITE_TAC[]);;
%

let s_subst_o =
 prove_thm
  (`s_subst_o`,
   "!ys xs s.
    (s <|s (ys // xs)) = (s o (BNDsR ys xs I))",
   REWRITE_TAC[s_subst_var_list]
   THEN DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;BNDsR;I_o_ID]
   THEN REPEAT GEN_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_REWRITE_TAC[o_THM]
   THEN ASM_CASES_TAC "(z:var) = h'"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN REWRITE_TAC[o_THM]
     ]
  );;

% s_subst_o
g "!ys xs s.
    (s <|s (ys // xs)) = (s o (BNDsR ys xs I))";;
e(REWRITE_TAC[s_subst_var_list]);;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[MAP;BNDsR;I_o_ID]);;

  e(REWRITE_TAC[MAP;BNDsR;I_o_ID]);;

  e(REWRITE_TAC[MAP;BNDsR;I_o_ID]);;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT GEN_TAC);;
  e(EXT_TAC "z:var");;
  e(GEN_TAC);;
  e(ASM_REWRITE_TAC[o_THM]);;
  e(ASM_CASES_TAC "(z:var) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(REWRITE_TAC[o_THM]);;
%


let var_list_TRANS =
 prove_thm
  (`var_list_TRANS`,
   "!x y z v s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ (v IN SL y ==> v IN SL x) ==>
     (((s <|s (z // y)) <|s (y // x)) v = (s <|s (z // x)) v)",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_o;(SYM o SPEC_ALL) o_ASSOC]
   THEN ONCE_REWRITE_TAC[o_THM]
   THEN USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] BNDsR_TRANS)
   THEN ASM_REWRITE_TAC[]
  );;

% var_list_TRANS
g "!x y z v s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ (v IN SL y ==> v IN SL x) ==>
     (((s <|s (z // y)) <|s (y // x)) v = (s <|s (z // x)) v)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_o]);;
e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
e(ONCE_REWRITE_TAC[o_THM]);;
e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] BNDsR_TRANS));;
e(ASM_REWRITE_TAC[]);;
%

let V_var_list_TRANS =
 prove_thm
  (`V_var_list_TRANS`,
   "!v x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVv v INTER SL y) SUBSET SL x) ==>
     (V v ((s <|s (z // y)) <|s (y // x)) = V v (s <|s (z // x)))",
   INDUCT_THEN vexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVv_DEF;V_DEF]
   THEN
     (( REWRITE_TAC[SUBSET_DEF;IN_INTER;IN]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[] o SPEC "v:var")
   THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] var_list_TRANS)
   THEN ASM_REWRITE_TAC[]
      ) ORELSE (
        REWRITE_TAC[UNION_INTER;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
      ))
  );;

%< V_var_list_TRANS
g "!v x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVv v INTER SL y) SUBSET SL x) ==>
     (V v ((s <|s (z // y)) <|s (y // x)) = V v (s <|s (z // x)))";;
e(INDUCT_THEN vexp_induct ASSUME_TAC);;

  e(REWRITE_TAC[FVv_DEF;V_DEF]);;

  e(REWRITE_TAC[FVv_DEF;V_DEF]);;
  e(REWRITE_TAC[SUBSET_DEF;IN_INTER;IN]);;
  e(REPEAT STRIP_TAC);;
  e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[] o SPEC "v:var"));;
  e(IMP_RES_TAC var_list_TRANS);;
  %e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] var_list_TRANS));;%
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[FVv_DEF;V_DEF]);;
  e(REWRITE_TAC[UNION_INTER;UNION_SUBSET]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
>%

let VS_var_list_TRANS =
 prove_thm
  (`VS_var_list_TRANS`,
   "!vs x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVvs vs INTER SL y) SUBSET SL x) ==>
     (VS vs ((s <|s (z // y)) <|s (y // x)) = VS vs (s <|s (z // x)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[FVvs_DEF;VS_DEF]
   THEN REWRITE_TAC[UNION_INTER;UNION_SUBSET;CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
    [ MATCH_MP_TAC V_var_list_TRANS;

      FIRST_ASSUM MATCH_MP_TAC
    ]
   THEN ASM_REWRITE_TAC[]
  );;

%< VS_var_list_TRANS
g "!vs x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVvs vs INTER SL y) SUBSET SL x) ==>
     (VS vs ((s <|s (z // y)) <|s (y // x)) = VS vs (s <|s (z // x)))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;

  e(REWRITE_TAC[FVvs_DEF;VS_DEF]);;
  e(REWRITE_TAC[UNION_INTER;UNION_SUBSET;CONS_11]);;
  e(REPEAT STRIP_TAC);;

   e(MATCH_MP_TAC V_var_list_TRANS);;
   e(ASM_REWRITE_TAC[]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[]);;
>%

let A_BND_state_subst_TRANS =
 prove_thm
  (`A_BND_state_subst_TRANS`,
   "!v1 v2 ss1 ss2 a s n v.
     (!x. (x IN (FVa a DELETE v)) ==> ~(v1 IN FVv (ss1 x))) /\
     (!x y. x IN (FVa a DELETE v) /\ y IN FVv (ss1 x) ==>
          ~(v2 IN FVv (ss2 y))) ==>
     (A a (BND n v ((s <|s ss2) <|s ss1)) =
      A a (((BND n v2 s) <|s (BND (AVAR v2) v1 ss2))
                         <|s (BND (AVAR v1) v ss1)))",
   REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM A_BND_state_subst
   THEN ONCE_REWRITE_TAC[(SYM o SPEC_ALL) a_subst_THM]
   THEN MATCH_MP_TAC A_BND_state_subst
   THEN REWRITE_TAC[FVa_a_subst;IN_DELETE;IN_UNION_SET;IN_IMAGE]
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN UNDISCH_TAC "(x:var) IN si"
   THEN ASM_REWRITE_TAC[o_THM]
   THEN ASM_CASES_TAC "(x':var) = v"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH;FVv_DEF;IN];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN STRIP_TAC
        THEN EXISTS_TAC "x':var"
        THEN ASM_REWRITE_TAC[IN_DELETE]
     ]
  );;

% A_BND_state_subst_TRANS
g "!v1 v2 ss1 ss2 a s n v.
     (!x. (x IN (FVa a DELETE v)) ==> ~(v1 IN FVv (ss1 x))) /\
     (!x y. x IN (FVa a DELETE v) /\ y IN FVv (ss1 x) ==>
          ~(v2 IN FVv (ss2 y))) ==>
     (A a (BND n v ((s <|s ss2) <|s ss1)) =
      A a (((BND n v2 s) <|s (BND (AVAR v2) v1 ss2))
                         <|s (BND (AVAR v1) v ss1)))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM A_BND_state_subst);;
e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL) a_subst_THM]);;
e(MATCH_MP_TAC A_BND_state_subst);;
e(REWRITE_TAC[FVa_a_subst]);;
e(REWRITE_TAC[IN_DELETE;IN_UNION_SET;IN_IMAGE]);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(UNDISCH_TAC "(x:var) IN si");;
e(ASM_REWRITE_TAC[o_THM]);;
e(ASM_CASES_TAC "(x':var) = v");;

  e(ASM_REWRITE_TAC[BND_MATCH;FVv_DEF;IN]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "x':var");;
  e(ASM_REWRITE_TAC[IN_DELETE]);;
%

let A_var_list_TRANS_LEMMA1 =
 TAC_PROOF
  (([], "!a n v s x y z.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\ DL y /\
     (((FVa a) DELETE v) INTER SL y) SUBSET (SL x)
   ==>
     (A a (BND n v ((s <|s (z // y)) <|s (y // x))) =
      A a (((BND n (variant v (FVa a UNION SL z)) s) <|s
            (CONS (variant v (FVa a UNION SL z)) z //
             CONS (variant v (FVa a UNION SL y)) y)
           ) <|s (CONS (variant v (FVa a UNION SL y)) y // CONS v x)
          ))"),

   REWRITE_TAC[var_list_THM;SUBSET_DEF;IN_INTER;IN_DELETE]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC A_BND_state_subst_TRANS
   THEN REWRITE_TAC[FVv_var_list_subst;IN;IN_DELETE]
   THEN CONJ_TAC
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN RES_TAC
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC IN_NOT_IN
   THENL [ EXISTS_TAC "FVa a UNION SL y";
           EXISTS_TAC "FVa a UNION SL z"
         ]
   THEN REWRITE_TAC[variant_not_in_set;IN_UNION]
   THEN ASM_CASES_TAC "(x':var) IN SL x"
   THENL
     [  DISJ2_TAC
        THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
        THEN ASM_REWRITE_TAC[I_THM];

        DISJ2_TAC
        THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
        THEN ASM_REWRITE_TAC[];

        ASM_REWRITE_TAC[]
        THEN IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
        THEN REWRITE_TAC[I_THM]
        THEN IMP_RES_TAC ((CONTRAPOS o ASSUME)
                          "(x':var) IN (SL y) ==> x' IN (SL x)")
        THEN IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
        THEN ASM_REWRITE_TAC[I_THM]
     ]
  );;

% A_var_list_TRANS_LEMMA1
g "!a n v s x y z.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\ DL y /\
     (((FVa a) DELETE v) INTER SL y) SUBSET (SL x)
   ==>
     (A a (BND n v ((s <|s (z // y)) <|s (y // x))) =
      A a (((BND n (variant v (FVa a UNION SL z)) s) <|s
            (CONS (variant v (FVa a UNION SL z)) z //
             CONS (variant v (FVa a UNION SL y)) y)
           ) <|s (CONS (variant v (FVa a UNION SL y)) y // CONS v x)
          ))";;

e(REWRITE_TAC[var_list_THM;SUBSET_DEF;IN_INTER;IN_DELETE]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC A_BND_state_subst_TRANS);;
e(REWRITE_TAC[FVv_var_list_subst;IN;IN_DELETE]);;
e(CONJ_TAC);;

  e(REPEAT GEN_TAC);;
  e(STRIP_TAC);;
  e(RES_TAC);;
  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(MATCH_MP_TAC IN_NOT_IN);;
  e(EXISTS_TAC "FVa a UNION SL y");;
  e(REWRITE_TAC[variant_not_in_set;IN_UNION]);;
  e(ASM_CASES_TAC "(x':var) IN SL x");;

    e(DISJ2_TAC);;
    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
    e(ASM_REWRITE_TAC[I_THM]);;

  e(REPEAT GEN_TAC);;
  e(STRIP_TAC);;
  e(RES_TAC);;
  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(MATCH_MP_TAC IN_NOT_IN);;
  e(EXISTS_TAC "FVa a UNION SL z");;
  e(ASM_REWRITE_TAC[variant_not_in_set;IN_UNION]);;
  e(ASM_CASES_TAC "(x':var) IN SL x");;

    e(DISJ2_TAC);;
    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
    e(ASM_REWRITE_TAC[]);;
    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
    e(REWRITE_TAC[I_THM]);;
    e(IMP_RES_TAC ((CONTRAPOS o ASSUME)
                   "(x':var) IN (SL y) ==> x' IN (SL x)"));;
    e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
    e(ASM_REWRITE_TAC[I_THM]);;
%

let A_var_list_TRANS_LEMMA2 =
 TAC_PROOF
  (([], "!a n v s x z.
     (LENGTH x = LENGTH z)
    ==>
     (A a (BND n v (s <|s (z // x))) =
      A a ((BND n (variant v (FVa a UNION SL z)) s) <|s
            (CONS (variant v (FVa a UNION SL z)) z // CONS v x)))"),
   REWRITE_TAC[var_list_THM]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC A_BND_state_subst
   THEN REWRITE_TAC[IN_DELETE]
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN REWRITE_TAC[FVv_var_list_subst;IN]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC IN_NOT_IN
   THEN EXISTS_TAC "FVa a UNION SL z"
   THEN REWRITE_TAC[variant_not_in_set;IN_UNION]
   THEN ASM_CASES_TAC "(x':var) IN SL x"
   THENL
     [  DISJ2_TAC
        THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER)
        THEN ASM_REWRITE_TAC[];

        IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH
        THEN ASM_REWRITE_TAC[I_THM]
     ]
  );;

% A_var_list_TRANS_LEMMA2
g "!a n v s x z.
     (LENGTH x = LENGTH z)
   ==>
     (A a (BND n v (s <|s (z // x))) =
      A a ((BND n (variant v (FVa a UNION SL z)) s) <|s
            (CONS (variant v (FVa a UNION SL z)) z // CONS v x)))";;

e(REWRITE_TAC[var_list_THM]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC A_BND_state_subst);;
e(REWRITE_TAC[IN_DELETE]);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(REWRITE_TAC[FVv_var_list_subst;IN]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC IN_NOT_IN);;
e(EXISTS_TAC "FVa a UNION SL z");;
e(REWRITE_TAC[variant_not_in_set;IN_UNION]);;
e(ASM_CASES_TAC "(x':var) IN SL x");;

  e(DISJ2_TAC);;
  e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] BNDsR_MEMBER));;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_THEN REWRITE_THM BNDsR_NOT_MATCH);;
  e(ASM_REWRITE_TAC[I_THM]);;
%

let A_var_list_TRANS =
 prove_thm
  (`A_var_list_TRANS`,
   "!a x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVa a INTER SL y) SUBSET SL x) ==>
     (A ((a <|a (y // x)) <|a (z // y)) s = A (a <|a (z // x)) s)",

   let try_imp tac = (tac THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC in

   REWRITE_TAC[a_subst_THM]
   THEN INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[FVa_DEF;A_DEF;UNION_INTER;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN ( try_imp (USE_IMP_THEN REWRITE_THM V_var_list_TRANS) )
   THEN ( try_imp (USE_IMP_THEN REWRITE_THM V_var_list_TRANS) )
   THEN ( try_imp (USE_IMP_THEN REWRITE_THM VS_var_list_TRANS) )
   THEN ( try_imp (USE_IMP_THEN REWRITE_THM VS_var_list_TRANS) )
   THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
   THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
   THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
   THEN ASM_REWRITE_TAC[]

   THEN ( FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC )
   THEN ( USE_IMP_THEN REWRITE_THM A_var_list_TRANS_LEMMA1
          THENL [ ASM_REWRITE_TAC[]; ALL_TAC ] )
   THEN ( USE_IMP_THEN REWRITE_THM A_var_list_TRANS_LEMMA2
          THENL [ ASM_REWRITE_TAC[]; ALL_TAC ] )
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM] o
          SPECL["v:var";"FVa a UNION SL y"])  variant_not_in_set
   THEN ASM_REWRITE_TAC[LENGTH;DISJOINT_LIST;SET_LIST]
   THEN REWRITE_ALL_TAC[SUBSET_DEF;IN_INTER;IN;IN_DELETE]
   THEN GEN_TAC
   THEN ( STRIP_TAC
   THENL
     [  POP_ASSUM REWRITE_ALL_THM
        THEN RES_TAC;

        ASM_CASES_TAC "(x':var) = v"
        THEN ASM_REWRITE_TAC[]
        THEN RES_TAC
     ] )
  );;

%< A_var_list_TRANS
g "!a x y z s.
     (LENGTH x = LENGTH y) /\ (LENGTH y = LENGTH z) /\
     DL y /\ ((FVa a INTER SL y) SUBSET SL x) ==>
     (A ((a <|a (y // x)) <|a (z // y)) s = A (a <|a (z // x)) s)";;
e(REWRITE_TAC[a_subst_THM]);;
e(let try_imp tac = (tac THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC in
  INDUCT_THEN aexp_induct ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;A_DEF;UNION_INTER;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN ( try_imp (USE_IMP_THEN REWRITE_THM V_var_list_TRANS) )
  THEN ( try_imp (USE_IMP_THEN REWRITE_THM V_var_list_TRANS) )
  THEN ( try_imp (USE_IMP_THEN REWRITE_THM VS_var_list_TRANS) )
  THEN ( try_imp (USE_IMP_THEN REWRITE_THM VS_var_list_TRANS) )
  THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
  THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
  THEN ( try_imp (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)) )
  THEN ASM_REWRITE_TAC[]);;

% or,
e(INDUCT_THEN aexp_induct ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;A_DEF;UNION_INTER;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN IMP_RES_TAC V_var_list_TRANS
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[]);;

  e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
    POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
%

  e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
  e(USE_IMP_THEN REWRITE_THM A_var_list_TRANS_LEMMA1
    THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
  e(USE_IMP_THEN REWRITE_THM A_var_list_TRANS_LEMMA2
    THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e((STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM] o
     SPECL["v:var";"FVa a UNION SL y"]) variant_not_in_set);;
  e(ASM_REWRITE_TAC[LENGTH;DISJOINT_LIST;SET_LIST]);;
  e(REWRITE_ALL_TAC[SUBSET_DEF;IN_INTER;IN;IN_DELETE]);;
  e(GEN_TAC);;
  e(STRIP_TAC);;

    e(POP_ASSUM REWRITE_ALL_THM);;
    e(RES_TAC);;

    e(ASM_CASES_TAC "(x':var) = v");;

      e(ASM_REWRITE_TAC[]);;

      e(ASM_REWRITE_TAC[]);;
      e(RES_TAC);;
>%




let var_list_IDENT_APPEND =
 prove_thm
  (`var_list_IDENT_APPEND`,
   "!zs ys xs.
      (LENGTH xs = LENGTH ys) ==>
      (((APPEND ys zs) // (APPEND xs zs)) = (ys // xs))",
   GEN_TAC
   THEN LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% var_list_IDENT_APPEND
g "!zs ys xs.
      (LENGTH xs = LENGTH ys) ==>
      (((APPEND ys zs) // (APPEND xs zs)) = (ys // xs))";;
e(GEN_TAC);;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]);;

 e(REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let var_list_APPEND_CONS_LEMMA =
 TAC_PROOF
  (([],
    "!ys xs ws zs w z.
      (LENGTH xs = LENGTH ys) ==>
      (LENGTH zs = LENGTH ws) /\
      (DISJOINT(SL xs)(SL(CONS z zs))) ==>
      (((APPEND ys (CONS w ws)) // (APPEND xs (CONS z zs))) =
       ((CONS w (APPEND ys ws)) // (CONS z (APPEND xs zs))))"),
   REWRITE_TAC[SET_LIST;DISJOINT_INSERT2]
   THEN LENGTH_LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[APPEND;var_list_IDENT]
   THEN REWRITE_TAC[var_list_THM]
   THEN REWRITE_TAC[SET_LIST;DISJOINT_INSERT;IN;DE_MORGAN_THM]
   THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[APPEND;var_list_THM]
   THEN IMP_RES_THEN REWRITE_THM BND_COMMUTE
  );;

% var_list_APPEND_CONS
g "!ys xs ws zs w z.
      (LENGTH xs = LENGTH ys) ==>
      (LENGTH zs = LENGTH ws) /\
      (DISJOINT(SL xs)(SL(CONS z zs))) ==>
      (((APPEND ys (CONS w ws)) // (APPEND xs (CONS z zs))) =
       ((CONS w (APPEND ys ws)) // (CONS z (APPEND xs zs))))";;
e(REWRITE_TAC[SET_LIST;DISJOINT_INSERT2]);;
e(LENGTH_LIST_INDUCT_TAC);;

 e(ASM_REWRITE_TAC[APPEND;var_list_IDENT]);;

 e(ASM_REWRITE_TAC[APPEND;var_list_IDENT]);;
 e(REWRITE_TAC[var_list_THM]);;
 e(REWRITE_TAC[SET_LIST;DISJOINT_INSERT;IN;DE_MORGAN_THM]);;
 e(STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[APPEND;var_list_THM]);;
 e(IMP_RES_THEN REWRITE_THM BND_COMMUTE);;
%

let var_list_APPEND_CONS =
 save_thm
  (`var_list_APPEND_CONS`,
   (REWRITE_RULE[AND_IMP_INTRO] var_list_APPEND_CONS_LEMMA));;




close_theory();;

