% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definition of the termination semantic relations,
  the termination correctness specifications, and the theorems which are
  the termination logic for Sunrise.  It also contains the critical
  termination argument in several theorems about the termination of the
  defined procedures in the environment.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `termination`;;

new_parent `stage_semantics`;;



% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of termination %

loadf `autoload`;;

autoload_all_defs_and_thms
     `string bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed cmd_semantics
      hoare_rules progress_rules semantic_stages stage_semantics`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `progress_tactics`;;





let C_calls_terminate = new_def
  `C_calls_terminate`
  "C_calls_terminate c r s1 =
       (!p s3. C_calls c r s1 p s3 ==>
               (?s4. let (vars,vals,glbs,pre,post,calls,rec,c') = r p in
                     C c' r s3 s4))";;


let C_calls_terminate_BND_THM =
 prove_thm
  (`C_calls_terminate_BND_THM`,
   "!n x c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ ~WFx x ==>
     C_calls_terminate c r (BND n x s1)",
   REWRITE_TAC[C_calls_terminate]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP C_calls_BND_THM)
   THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[BND_OVERRIDE;IDENT_BND] o
                     SPEC "(s1:state) x")
   THEN RES_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFx)
   THEN IMP_RES_TAC (AND2IMP C_BND_THM)
   THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[BND_OVERRIDE;IDENT_BND] o
                     SPEC "(s3:state) x")
   THEN EXISTS_TAC "BND ((s3:state) x) x (s4:state)"
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_terminate_BND_THM
g "!n x c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ ~WFx x ==>
     C_calls_terminate c r (BND n x s1)";;
e(REWRITE_TAC[C_calls_terminate]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP C_calls_BND_THM));;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[BND_OVERRIDE;IDENT_BND] o
             SPEC "(s1:state) x"));;
e(RES_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(IMP_RES_TAC (AND2IMP WFc_NOT_WFx));;
e(IMP_RES_TAC (AND2IMP C_BND_THM));;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[BND_OVERRIDE;IDENT_BND] o
             SPEC "(s3:state) x"));;
e(EXISTS_TAC "BND ((s3:state) x) x (s4:state)");;
e(ASM_REWRITE_TAC[]);;
>%


let C_calls_terminate_BNDsL_THM =
 prove_thm
  (`C_calls_terminate_BNDsL_THM`,
   "!ns xs c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ NOT_WFxs xs ==>
     C_calls_terminate c r (BNDsL ns xs s1)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL;NOT_WFxs]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM)
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_terminate_BNDsL_THM
g "!ns xs c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ NOT_WFxs xs ==>
     C_calls_terminate c r (BNDsL ns xs s1)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[BNDsL]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[BNDsL;NOT_WFxs]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;
 e(MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM));;
 e(ASM_REWRITE_TAC[]);;
>%


let C_calls_terminate_BNDsR_THM =
 prove_thm
  (`C_calls_terminate_BNDsR_THM`,
   "!ns xs c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ NOT_WFxs xs ==>
     C_calls_terminate c r (BNDsR ns xs s1)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR;NOT_WFxs]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM)
   THEN ASM_REWRITE_TAC[]
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_terminate_BNDsR_THM
g "!ns xs c r g s1.
     WFenv_syntax r /\ WFc c g r /\ C_calls_terminate c r s1 /\ NOT_WFxs xs ==>
     C_calls_terminate c r (BNDsR ns xs s1)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[BNDsR]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[BNDsR;NOT_WFxs]);;
 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM));;
 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;
>%


let SEQ_calls_terminate_1 =
 prove_thm
  (`SEQ_calls_terminate_1`,
   "!c1 c2 r s1.

           C_calls_terminate (SEQ c1 c2) r s1
       % -------------------------------------------- % ==>
           C_calls_terminate c1 r s1",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ1_TAC
   THEN EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "c2:cmd"
   THEN ASM_REWRITE_TAC[]
  );;

%< SEQ_calls_terminate_1
g "!c1 c2 r s1.

           C_calls_terminate (SEQ c1 c2) r s1
       % -------------------------------------------- % ==>
           C_calls_terminate c1 r s1";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ1_TAC);;
e(EXISTS_TAC "c1:cmd");;
e(EXISTS_TAC "c2:cmd");;
e(ASM_REWRITE_TAC[]);;
>%


let SEQ_calls_terminate_2 =
 prove_thm
  (`SEQ_calls_terminate_2`,
   "!c1 c2 r s1 s2.

           C_calls_terminate (SEQ c1 c2) r s1 /\
           C c1 r s1 s2
       % -------------------------------------------- % ==>
           C_calls_terminate c2 r s2",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ2_TAC
   THEN EXISTS_TAC "c2:cmd"
   THEN EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< SEQ_calls_terminate_2
g "!c1 c2 r s1 s2.

           C_calls_terminate (SEQ c1 c2) r s1 /\
           C c1 r s1 s2
       % -------------------------------------------- % ==>
           C_calls_terminate c2 r s2";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ2_TAC);;
e(EXISTS_TAC "c2:cmd");;
e(EXISTS_TAC "c1:cmd");;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let IF_calls_terminate_1 =
 prove_thm
  (`IF_calls_terminate_1`,
   "!b c1 c2 r s1 s2.

           C_calls_terminate (IF b c1 c2) r s1 /\
           B b s1 T s2
       % -------------------------------------------- % ==>
           C_calls_terminate c1 r s2",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ1_TAC
   THEN EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c2:cmd"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< IF_calls_terminate_1
g "!b c1 c2 r s1 s2.

           C_calls_terminate (IF b c1 c2) r s1 /\
           B b s1 T s2
       % -------------------------------------------- % ==>
           C_calls_terminate c1 r s2";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ1_TAC);;
e(EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c2:cmd"
   THEN EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let IF_calls_terminate_2 =
 prove_thm
  (`IF_calls_terminate_2`,
   "!b c1 c2 r s1 s2.

           C_calls_terminate (IF b c1 c2) r s1 /\
           B b s1 F s2
       % -------------------------------------------- % ==>
           C_calls_terminate c2 r s2",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ2_TAC
   THEN EXISTS_TAC "c2:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< IF_calls_terminate_2
g "!b c1 c2 r s1 s2.

           C_calls_terminate (IF b c1 c2) r s1 /\
           B b s1 F s2
       % -------------------------------------------- % ==>
           C_calls_terminate c2 r s2";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ2_TAC);;
e(EXISTS_TAC "c2:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c1:cmd"
   THEN EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let WHILE_calls_terminate_1 =
 prove_thm
  (`WHILE_calls_terminate_1`,
   "!a pr b c r s1 s2.

           C_calls_terminate (WHILE a pr b c) r s1 /\
           B b s1 T s2
       % -------------------------------------------- % ==>
           C_calls_terminate c r s2",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ1_TAC
   THEN EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "a:aexp"
   THEN EXISTS_TAC "pr:aexp"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< WHILE_calls_terminate_1
g "!a pr b c r s1 s2.

           C_calls_terminate (WHILE a pr b c) r s1 /\
           B b s1 T s2
       % -------------------------------------------- % ==>
           C_calls_terminate c r s2";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ1_TAC);;
e(EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "a:aexp"
   THEN EXISTS_TAC "pr:aexp"
   THEN EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let WHILE_calls_terminate_2 =
 prove_thm
  (`WHILE_calls_terminate_2`,
   "!a pr b c r s1 s2 s3.

           C_calls_terminate (WHILE a pr b c) r s1 /\
           B b s1 T s2 /\
           C c r s2 s3
       % -------------------------------------------- % ==>
           C_calls_terminate (WHILE a pr b c) r s3",

   REWRITE_TAC[C_calls_terminate]
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN DISJ2_TAC
   THEN EXISTS_TAC "a:aexp"
   THEN EXISTS_TAC "pr:aexp"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "s3:state"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< WHILE_calls_terminate_2
g "!a pr b c r s1 s2 s3.

           C_calls_terminate (WHILE a pr b c) r s1 /\
           B b s1 T s2 /\
           C c r s2 s3
       % -------------------------------------------- % ==>
           C_calls_terminate (WHILE a pr b c) r s3";;

e(REWRITE_TAC[C_calls_terminate]);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(DISJ2_TAC);;
e(EXISTS_TAC "a:aexp"
   THEN EXISTS_TAC "pr:aexp"
   THEN EXISTS_TAC "b:bexp"
   THEN EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "s3:state"
   THEN EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let C_calls_CALL =
 prove_thm
  (`C_calls_CALL`,
   "!p xs es ns r s1 s2 s3 vars vals glbs pre post calls rec c vals'.

           C_calls (CALL p xs es) r s1 p s3 /\
           ES es s1 ns s2 /\
           (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
           (vals' = variants vals (SL (APPEND xs glbs)))
       % -------------------------------------------- % ==>
           (s3 = ((BNDsL ns vals' s2) <|s
                        (APPEND xs vals' // APPEND vars vals)))",

   REPEAT STRIP_TAC
   THEN UNDISCH_TAC "C_calls (CALL p xs es) r s1 p s3"
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
                    IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                    REWRITE_RULE[cmd_distinct;cmd_one_one] o
                    ONCE_REWRITE_RULE[C_calls_cases])
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[ASSUME "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"]
   THEN DISCH_THEN (REWRITE_ALL_THM o REWRITE_RULE[PAIR_EQ] o SYM)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(xs:var list) = xs'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es:exp list) = es'"
   THEN UNDISCH_TAC "ES es s1 ns' s2'"
   THEN UNDISCH_TAC "ES es s1 ns s2"
   THEN REWRITE_TAC[VES_THM]
   THEN DISCH_THEN REWRITE_ALL_THM
   THEN DISCH_THEN REWRITE_ALL_THM
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_CALL
g "!p xs es ns r s1 s2 s3 vars vals glbs pre post calls rec c vals'.

           C_calls (CALL p xs es) r s1 p s3 /\
           ES es s1 ns s2 /\
           (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
           (vals' = variants vals (SL (APPEND xs glbs)))
       % -------------------------------------------- % ==>
           (s3 = ((BNDsL ns vals' s2) <|s
                        (APPEND xs vals' // APPEND vars vals)))";;

e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "C_calls (CALL p xs es) r s1 p s3");;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              IMP2AND o
              CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
              REWRITE_RULE[cmd_distinct;cmd_one_one] o
              ONCE_REWRITE_RULE[C_calls_cases]));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[ASSUME "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"]);;
e(DISCH_THEN (REWRITE_ALL_THM o pthm o REWRITE_RULE[PAIR_EQ] o SYM));;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(xs:var list) = xs'");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es:exp list) = es'");;
e(UNDISCH_TAC "ES es s1 ns' s2'");;
e(UNDISCH_TAC "ES es s1 ns s2");;
e(REWRITE_TAC[VES_THM]);;
e(DISCH_THEN REWRITE_ALL_THM);;
e(DISCH_THEN REWRITE_ALL_THM);;
e(ASM_REWRITE_TAC[]);;
>%


let CALL_calls_terminate =
 prove_thm
  (`CALL_calls_terminate`,
   "!p xs es g r s1.

           WFenv_syntax r /\ WFc (CALL p xs es) g r /\
           C_calls_terminate (CALL p xs es) r s1
       % -------------------------------------------- % ==>
           (?s2. C (CALL p xs es) r s1 s2)",

   REWRITE_TAC[C_calls_terminate]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC C_calls_CALL_THM
   THEN EXISTS_TAC "g:prog_env"
   THEN EXISTS_TAC "(FST)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
  );;

%< CALL_calls_terminate
g "!p xs es g r s1.

           WFenv_syntax r /\ WFc (CALL p xs es) g r /\
           C_calls_terminate (CALL p xs es) r s1
       % -------------------------------------------- % ==>
           (?s2. C (CALL p xs es) r s1 s2)";;

e(REWRITE_TAC[C_calls_terminate]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC C_calls_CALL_THM);;
e(EXISTS_TAC "g:prog_env"
   THEN EXISTS_TAC "(FST)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)");;
e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
>%





let C_calls_TERM = new_def
  `C_calls_TERM`
  "C_calls_TERM(a,c,r) =
     (!s1. A a s1 /\ C_calls_terminate c r s1 ==> (?s2. C c r s1 s2))";;


let PRE_STRENGTH_C_calls_TERM =
 prove_thm
  (`PRE_STRENGTH_C_calls_TERM`,
   "!a0 a1 c r.
     CLOSE(AIMP a0 a1) /\ C_calls_TERM(a1, c, r) ==>
     C_calls_TERM(a0, c, r)",
   REWRITE_TAC[CLOSE;C_calls_TERM;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< PRE_STRENGTH_C_calls_TERM
g "!a0 a1 c r.
     CLOSE(AIMP a0 a1) /\ C_calls_TERM(a1, c, r) ==>
     C_calls_TERM(a0, c, r)";;
e(REWRITE_TAC[CLOSE;C_calls_TERM;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%



let SKIP_TERM =
 prove_thm
  (`SKIP_TERM`,
   "!a r. C_calls_TERM(a, SKIP, r)",
   REWRITE_TAC[C_calls_TERM]
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC "s1:state"
   THEN SKIP_TAC
  );;

%< SKIP_TERM
g "!a r. C_calls_TERM(a, SKIP, r)";;

e(REWRITE_TAC[C_calls_TERM]);;
e(REPEAT STRIP_TAC);;
e(EXISTS_TAC "s1:state");;
e(SKIP_TAC);;

e(C_CALLS_STR_INDUCT_TAC
  THEN REWRITE_TAC[cmd_distinct;cmd_one_one]);;
>%


let ABORT_TERM =
 prove_thm
  (`ABORT_TERM`,
   "!r. C_calls_TERM(AFALSE, ABORT, r)",
   REWRITE_TAC[C_calls_TERM;A_DEF]
  );;

%< ABORT_TERM
g "!r. C_calls_TERM(AFALSE, ABORT, r)";;

e(REWRITE_TAC[C_calls_TERM]);;
e(REWRITE_TAC[A_DEF]);;
>%


let ASSIGN_TERM =
 prove_thm
  (`ASSIGN_TERM`,
   "!a x e r. C_calls_TERM(a, ASSIGN x e, r)",
   REWRITE_TAC[C_calls_TERM]
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC "s1 <|s (x := e)"
   THEN REWRITE_TAC[assign;s_subst_BND]
   THEN ASSIGN_TAC
   THEN REWRITE_TAC[VE_THM]
  );;

%< ASSIGN_TERM
g "!a x e r. C_calls_TERM(a, ASSIGN x e, r)";;

e(REWRITE_TAC[C_calls_TERM]);;
e(REPEAT STRIP_TAC);;
e(EXISTS_TAC "s1 <|s (x := e)");;
e(REWRITE_TAC[assign;s_subst_BND]);;
e(ASSIGN_TAC);;
e(REWRITE_TAC[VE_THM]);;
>%


let SEQ_TERM =
 prove_thm
  (`SEQ_TERM`,
   "!a1 a2 c1 c2 r.

           C_calls_TERM(a1, c1, r) /\
           PCSPEC(a1, c1, a2, r) /\
           C_calls_TERM(a2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(a1, SEQ c1 c2, r)",

   REWRITE_TAC[C_calls_TERM;PCSPEC]
   THEN REWRITE_TAC[(SYM o SPEC_ALL)AND_IMP_INTRO]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC SEQ_calls_terminate_1
   THEN RES_TAC
   THEN RES_TAC
   THEN POP_TAC
   THEN IMP_RES_TAC SEQ_calls_terminate_2
   THEN RES_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "s2':state"
   THEN SEQ_TAC
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< SEQ_TERM
g "!a1 a2 c1 c2 r.

           C_calls_TERM(a1, c1, r) /\
           PCSPEC(a1, c1, a2, r) /\
           C_calls_TERM(a2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(a1, SEQ c1 c2, r)";;

e(REWRITE_TAC[C_calls_TERM;PCSPEC]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)AND_IMP_INTRO]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC SEQ_calls_terminate_1);;
e(RES_TAC);;
e(RES_TAC);;
e(POP_TAC);;
e(IMP_RES_TAC SEQ_calls_terminate_2);;
e(RES_TAC);;
e(POP_TAC);;
e(EXISTS_TAC "s2':state");;
e(SEQ_TAC);;
e(EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let IF_TERM =
 prove_thm
  (`IF_TERM`,
   "!a1 a2 q1 q2 b c1 c2 r.

           BSPEC(a1, b, q1)  /\  C_calls_TERM(q1, c1, r)  /\
           BSPEC(a2, b, q2)  /\  C_calls_TERM(q2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(ACOND (AB b) a1 a2, IF b c1 c2, r)",

   REWRITE_TAC[C_calls_TERM;BSPEC;A_DEF]
   THEN REWRITE_TAC[(SYM o SPEC_ALL)AND_IMP_INTRO]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "A(AB b)s1"
   THEN POP_ASSUM (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
   THEN RES_TAC
   THEN IMP_RES_TAC AB_THM
   THENL
    [ POP_TAC
      THEN POP_ASSUM REWRITE_ALL_THM
      THEN IMP_RES_TAC IF_calls_terminate_1;

      POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_ASSUM REWRITE_ALL_THM
      THEN UNASSUME_THEN REWRITE_ALL_THM "~A(AB b)s1"
      THEN IMP_RES_TAC IF_calls_terminate_2;
    ]
   THEN RES_TAC
   THEN POP_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "s2':state"
   THENL [ IF_T_TAC; IF_F_TAC ]
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< IF_TERM
g "!a1 a2 q1 q2 b c1 c2 r.

           BSPEC(a1, b, q1)  /\  C_calls_TERM(q1, c1, r)  /\
           BSPEC(a2, b, q2)  /\  C_calls_TERM(q2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(ACOND (AB b) a1 a2, IF b c1 c2, r)";;

e(REWRITE_TAC[C_calls_TERM;BSPEC;A_DEF]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)AND_IMP_INTRO]);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "A(AB b)s1");;

 e(POP_ASSUM (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th));;
 e(RES_TAC);;
 e(IMP_RES_TAC AB_THM);;
 e(POP_TAC);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(IMP_RES_TAC IF_calls_terminate_1);;
 e(RES_TAC);;
 e(POP_TAC);;
 e(POP_TAC);;
 e(EXISTS_TAC "s2':state");;
 e(IF_T_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;

 e(POP_ASSUM (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th));;
 e(RES_TAC);;
 e(IMP_RES_TAC AB_THM);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(UNASSUME_THEN REWRITE_ALL_THM "~A(AB b)s1");;
 e(IMP_RES_TAC IF_calls_terminate_2);;
 e(RES_TAC);;
 e(POP_TAC);;
 e(POP_TAC);;
 e(EXISTS_TAC "s2':state");;
 e(IF_F_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%


let IF_TERM1 =
 prove_thm
  (`IF_TERM1`,
   "!b a1 a2 c1 c2 r.

           C_calls_TERM(a1, c1, r) /\
           C_calls_TERM(a2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(ACOND (AB b) (ab_pre b a1) (ab_pre b a2),
                        IF b c1 c2, r)",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC IF_TERM
   THEN EXISTS_TAC "a1:aexp"
   THEN EXISTS_TAC "a2:aexp"
   THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;

%< IF_TERM1
g "!b a1 a2 c1 c2 r.

           C_calls_TERM(a1, c1, r) /\
           C_calls_TERM(a2, c2, r)
       % -------------------------------------------- % ==>
           C_calls_TERM(ACOND (AB b) (ab_pre b a1) (ab_pre b a2),
                        IF b c1 c2, r)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC IF_TERM);;
e(EXISTS_TAC "a1:aexp");;
e(EXISTS_TAC "a2:aexp");;
e(ASM_REWRITE_TAC[ab_pre_THM]);;
>%





let BND_SAME_COMMUTE =
 TAC_PROOF
  (([],
    "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)"),
   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:**) = x"
   THENL
    [ ASM_REWRITE_TAC[];

      IMP_RES_TAC BND_COMMUTE
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% BND_SAME_COMMUTE
g "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)";;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:**) = x");;

 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC BND_COMMUTE);;
 e(ASM_REWRITE_TAC[]);;
%


let WHILE_TERM_LEMMA =
 prove_thm
  (`WHILE_TERM_LEMMA`,
   "!a v x b c g r a0 n.
    (\n.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
      C_calls_TERM(a0, c, r) /\
      PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r) ==>
      C_calls_TERM(AAND a (AEQ v (ANUM n)), WHILE a (ALESS v (AVAR x)) b c, r))
    n",

   REPEAT GEN_TAC
   THEN SPEC_TAC("n:num","n:num")
   THEN MATCH_MP_TAC NUM_WELL_FOUNDED
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th))
   THEN ASM_REWRITE_TAC[]
   THEN DISCH_THEN (ASSUME_TAC o REWRITE_RULE[C_calls_TERM])
   THEN UNDISCH_ALL_TAC
   THEN CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o
                     REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM])
   THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[AND_IMP_INTRO]
   THEN REWRITE_TAC[BSPEC;C_calls_TERM;PCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "A (AB b) s1"
   THENL
    [ UNASSUME_THEN (MP_TAC o SPEC "BND n x (s1:state)")
               "!s1. (A a s1 /\ A(AB b)s1) /\ (V v s1 = s1 x) ==>
                           (?t s2. B b s1 t s2)"
      THEN IMP_RES_THEN REWRITE_THM A_BND_THM
      THEN IMP_RES_THEN REWRITE_THM V_BND_THM
      THEN ASM_REWRITE_TAC[BND_MATCH]
      THEN STRIP_TAC
      THEN FIRST_ASSUM (MP_TAC o CONJUNCT1 o REWRITE_RULE[AB_THM])
      THEN IMP_RES_THEN REWRITE_THM A_BND_THM
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_THEN REWRITE_ALL_THM
      THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN REWRITE_ALL_TAC[BND_MATCH]
      THEN UNASSUME_THEN (MP_TAC o SPECL["BND n x (s1:state)";"T";"s2:state"])
                      "!s1 t s2. ((A a s1 /\ A(AB b)s1) /\
                                  (V v s1 = s1 x)) /\ B b s1 t s2 ==>
                                  A a0 s2"
      THEN IMP_RES_THEN REWRITE_THM A_BND_THM
      THEN IMP_RES_THEN REWRITE_THM V_BND_THM
      THEN ASM_REWRITE_TAC[BND_MATCH]
      THEN STRIP_TAC
      THEN SUBGOAL_THEN
          "C_calls_terminate (WHILE a (ALESS v (AVAR x)) b c) r (BND n x s1)"
      ASSUME_TAC
      THENL
       [ MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM)
         THEN ASM_REWRITE_TAC[WFc_WHILE];

         ALL_TAC
       ]
      THEN IMP_RES_TAC (AND2IMP WHILE_calls_terminate_1)
      THEN POP_TAC
      THEN IMP_RES_TAC (AND2IMP (ASSUME
              "!s1. A a0 s1 /\ C_calls_terminate c r s1 ==> (?s2. C c r s1 s2)"))
      THEN IMP_RES_TAC (AND2IMP (ASSUME
              "!s1 s2. A a0 s1 /\ C c r s1 s2 ==> A a s2 /\ (V v s2) < (s2 x)"))
      THEN UNDISCH_LAST_TAC
      THEN IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT)
      THEN POP_TAC
      THEN ASM_REWRITE_TAC[]
      THEN STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP WHILE_calls_terminate_2)
      THEN POP_TAC
      THEN UNASSUME_THEN (MP_TAC o SPECL["V v s2'";"s2':state"])
               "!m s1. m < n /\ (A a s1 /\ (V v s1 = m)) /\
                       C_calls_terminate(WHILE a(ALESS v(AVAR x))b c)r s1 ==>
                       (?s2. C(WHILE a(ALESS v(AVAR x))b c)r s1 s2)"
      THEN ASM_REWRITE_TAC[]
      THEN STRIP_TAC
      THEN EXISTS_TAC "BND ((s1:state) x) x (s2'':state)"
      THEN WHILE_T_TAC
      THEN EXISTS_TAC "BND ((s1:state) x) x (s2:state)"
      THEN EXISTS_TAC "BND ((s1:state) x) x (s2':state)"
      THEN IMP_RES_TAC C_BND_THM
      THEN POP_ASSUM REWRITE_THM
      THEN POP_ASSUM REWRITE_THM
      THEN IMP_RES_TAC B_BND_THM
      THEN POP_ASSUM (MP_TAC o SPEC "(s1:state) x")
      THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND];

      SUBGOAL_THEN "B b s1 F (s1 <|s (AB_state b))" ASSUME_TAC
      THENL
       [ ASM_REWRITE_TAC[AB_THM];

         EXISTS_TAC "s1 <|s (AB_state b)"
         THEN WHILE_F_TAC
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< WHILE_TERM_LEMMA
g "!a v x b c g r a0 n.
    (\n.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
      C_calls_TERM(a0, c, r) /\
      PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r) ==>
      C_calls_TERM(AAND a (AEQ v (ANUM n)), WHILE a (ALESS v (AVAR x)) b c, r))
    n";;

e(REPEAT GEN_TAC);;
e(SPEC_TAC("n:num","n:num"));;
e(MATCH_MP_TAC NUM_WELL_FOUNDED);;
e(BETA_TAC);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th)));;
e(ASM_REWRITE_TAC[]);;
e(DISCH_THEN (ASSUME_TAC o REWRITE_RULE[C_calls_TERM]));;
e(UNDISCH_ALL_TAC);;
e(CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM]));;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(REWRITE_TAC[AND_IMP_INTRO]);;
e(REWRITE_TAC[BSPEC;C_calls_TERM;PCSPEC;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "A (AB b) s1");;

 e(UNASSUME_THEN (MP_TAC o SPEC "BND n x (s1:state)")
               "!s1. (A a s1 /\ A(AB b)s1) /\ (V v s1 = s1 x) ==>
                           (?t s2. B b s1 t s2)");;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;
 e(STRIP_TAC);;
 e(FIRST_ASSUM (MP_TAC o CONJUNCT1 o REWRITE_RULE[AB_THM]));;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_THEN REWRITE_ALL_THM);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(REWRITE_ALL_TAC[BND_MATCH]);;
 e(UNASSUME_THEN (MP_TAC o SPECL["BND n x (s1:state)";"T";"s2:state"])
                      "!s1 t s2. ((A a s1 /\ A(AB b)s1) /\
                                  (V v s1 = s1 x)) /\ B b s1 t s2 ==>
                                  A a0 s2");;
 e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
 e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;
 e(STRIP_TAC);;
 e(SUBGOAL_THEN
          "C_calls_terminate (WHILE a (ALESS v (AVAR x)) b c) r (BND n x s1)"
      ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL C_calls_terminate_BND_THM));;
  e(ASM_REWRITE_TAC[WFc_WHILE]);;

 e(IMP_RES_TAC (AND2IMP WHILE_calls_terminate_1));;
 e(POP_TAC);;
 e(IMP_RES_TAC (AND2IMP (ASSUME "!s1. A a0 s1 /\ C_calls_terminate c r s1 ==>
                                      (?s2. C c r s1 s2)")));;
 e(IMP_RES_TAC (AND2IMP (ASSUME "!s1 s2. A a0 s1 /\ C c r s1 s2 ==>
                                         A a s2 /\ (V v s2) < (s2 x)")));;
 e(UNDISCH_LAST_TAC);;
 e(IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT));;
 e(POP_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP WHILE_calls_terminate_2));;
 e(POP_TAC);;
 e(UNASSUME_THEN (MP_TAC o SPECL["V v s2'";"s2':state"])
               "!m s1. m < n /\ (A a s1 /\ (V v s1 = m)) /\
                       C_calls_terminate(WHILE a(ALESS v(AVAR x))b c)r s1 ==>
                       (?s2. C(WHILE a(ALESS v(AVAR x))b c)r s1 s2)");;
 e(ASM_REWRITE_TAC[]);;
 e(STRIP_TAC);;
 e(EXISTS_TAC "BND ((s1:state) x) x (s2'':state)");;
 e(WHILE_T_TAC);;
 e(EXISTS_TAC "BND ((s1:state) x) x (s2:state)");;
 e(EXISTS_TAC "BND ((s1:state) x) x (s2':state)");;
 e(IMP_RES_TAC C_BND_THM);;
 e(POP_ASSUM REWRITE_THM);;
 e(POP_ASSUM REWRITE_THM);;
 e(IMP_RES_TAC B_BND_THM);;
 e(POP_ASSUM (MP_TAC o SPEC "(s1:state) x"));;
 e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;

 e(SUBGOAL_THEN "B b s1 F (s1 <|s (AB_state b))" ASSUME_TAC);;

  e(ASM_REWRITE_TAC[AB_THM]);;

 e(EXISTS_TAC "s1 <|s (AB_state b)");;
 e(WHILE_F_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%



let WHILE_TERM =
 prove_thm
  (`WHILE_TERM`,
   "!a a0 v x b c g r.

           WFenv_syntax r /\
           WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
           BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
           C_calls_TERM(a0, c, r) /\
           PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r)
       % -------------------------------------------- % ==>
           C_calls_TERM(a, WHILE a (ALESS v (AVAR x)) b c, r)",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP (CONV_RULE (ONCE_DEPTH_CONV BETA_CONV)
                               WHILE_TERM_LEMMA))
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[C_calls_TERM;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o IMP2AND o
                     CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                     SPECL["V v s1";"s1:state"])
   THEN ASM_REWRITE_TAC[]
  );;

%< WHILE_TERM
g "!a a0 v x b c g r.

           WFenv_syntax r /\
           WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
           BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, a0) /\
           C_calls_TERM(a0, c, r) /\
           PCSPEC(a0, c, AAND a (ALESS v (AVAR x)), r)
       % -------------------------------------------- % ==>
           C_calls_TERM(a, WHILE a (ALESS v (AVAR x)) b c, r)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP (CONV_RULE (ONCE_DEPTH_CONV BETA_CONV)
                             WHILE_TERM_LEMMA)));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[C_calls_TERM;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (MATCH_MP_TAC o IMP2AND o
               CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
               SPECL["V v s1";"s1:state"]));;
e(ASM_REWRITE_TAC[]);;
>%



let CALL_TERM =
 prove_thm
  (`CALL_TERM`,
   "!a p xs es g r.

           WFenv_syntax r /\ WFc (CALL p xs es) g r
       % -------------------------------------------- % ==>
           C_calls_TERM(a, CALL p xs es, r)",

   REWRITE_TAC[C_calls_TERM]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL CALL_calls_terminate)
   THEN ASM_REWRITE_TAC[]
  );;

%< CALL_TERM
g "!a p xs es g r.

           WFenv_syntax r /\ WFc (CALL p xs es) g r
       % -------------------------------------------- % ==>
           C_calls_TERM(a, CALL p xs es, r)";;

e(REWRITE_TAC[C_calls_TERM]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL CALL_calls_terminate));;
e(ASM_REWRITE_TAC[]);;
>%



% ============================================================== %
% Define a kind of well-formed environment where all procedure   %
% bodies terminate, given that their preconditions are true, and %
% that all other procedures called immediately from this one     %
% terminate.   We will later show that this condition implies    %
% several other "termination" conditions about the environment,  %
% including eventually that all procedure bodies terminate.      %
% ============================================================== %



let WFproc_term = 
    new_definition (`WFproc_term`,
      "WFproc_term p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            C_calls_TERM(pre, c, r)");;

let WFenv_term = 
    new_definition (`WFenv_term`,
      "WFenv_term r =
         !p. WFproc_term p r");;



% ============================================================== %
% We prove that WFenv_term implies that the termination of a     %
% procedure depends only on the termination of its immediate     %
% calls, from states which are reached by M_calls with path [].  %
% ============================================================== %


let M_calls_terminate = new_def
   `M_calls_terminate`
   "M_calls_terminate p1 s1 r =
       (!p2 s2. M_calls p1 s1 [] p2 s2 r ==>
                (?s3. let (vars,vals,glbs,pre,post,calls,rec,c) = r p2 in
                      C c r s2 s3))";;


let C_calls_M_calls_terminate =
 prove_thm
  (`C_calls_M_calls_terminate`,
   "!r p s1 vars vals glbs pre post calls rec c.
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       (C_calls_terminate c r s1 = M_calls_terminate p s1 r)",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[C_calls_terminate;M_calls_terminate]
   THEN IMP_RES_THEN REWRITE_THM C_calls_M_calls
  );;

%< C_calls_M_calls_terminate:
g "!r p s1 vars vals glbs pre post calls rec c.
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       (C_calls_terminate c r s1 = M_calls_terminate p s1 r)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[C_calls_terminate;M_calls_terminate]);;
e(IMP_RES_THEN REWRITE_THM C_calls_M_calls);;
>%



let M_calls_TERM = new_def
  `M_calls_TERM`
  "M_calls_TERM(p,r) =
     let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
     (!s1. A pre s1 /\ M_calls_terminate p s1 r ==>
           (?s2. C c r s1 s2))";;


let C_calls_TERM_M_calls_TERM =
 prove_thm
  (`C_calls_TERM_M_calls_TERM`,
   "!r p vars vals glbs pre post calls rec c.
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       (C_calls_TERM(pre, c, r) = M_calls_TERM(p, r))",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[C_calls_TERM;M_calls_TERM]
   THEN IMP_RES_THEN REWRITE_THM C_calls_M_calls_terminate
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

%< C_calls_TERM_M_calls_TERM:
g "!r p vars vals glbs pre post calls rec c.
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       (C_calls_TERM(pre, c, r) = M_calls_TERM(p, [], r))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[C_calls_TERM;M_calls_TERM]);;
e(IMP_RES_THEN REWRITE_THM C_calls_M_calls_terminate);;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
>%


let WFproc_term_M_calls_TERM =
 prove_thm
  (`WFproc_term_M_calls_TERM`,
   "!p r. WFproc_term p r = M_calls_TERM(p,r)",
   REPEAT GEN_TAC
   THEN REWRITE_TAC[WFproc_term]
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_THEN REWRITE_THM C_calls_TERM_M_calls_TERM
  );;

%< WFproc_term_M_calls_TERM:
g "!p r. WFproc_term p r = M_calls_TERM(p,r)";;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[WFproc_term]);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_THEN REWRITE_THM C_calls_TERM_M_calls_TERM);;
>%



let WFenv_term_M_calls_TERM =
 prove_thm
  (`WFenv_term_M_calls_TERM`,
   "!p r.
      WFenv_term r ==>
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      M_calls_TERM(p,r)",
   REPEAT STRIP_TAC
   THEN MP_TAC (ASSUME "WFenv_term r")
   THEN REWRITE_TAC[WFenv_term]
   THEN DISCH_THEN (MP_TAC o SPEC "p:string")
   THEN REWRITE_TAC[WFproc_term_M_calls_TERM]
  );;

%< WFenv_term_M_calls_TERM
g "!p r.
      WFenv_term r ==>
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      M_calls_TERM(p,r)";;

e(REPEAT STRIP_TAC);;
e(MP_TAC (ASSUME "WFenv_term r"));;
e(REWRITE_TAC[WFenv_term]);;
e(DISCH_THEN (MP_TAC o SPEC "p:string"));;
e(REWRITE_TAC[WFproc_term_M_calls_TERM]);;
>%




let INSERT_UNION_EQ2 =
 TAC_PROOF
  (([],
    "!s t (x:*). s UNION (x INSERT t) = x INSERT (s UNION t)"),
   ONCE_REWRITE_TAC[UNION_COMM]
   THEN REWRITE_TAC[INSERT_UNION_EQ]
  );;

% INSERT_UNION_EQ2
g "!s t (x:*). s UNION (x INSERT t) = x INSERT (s UNION t)";;
e(ONCE_REWRITE_TAC[UNION_COMM]);;
e(REWRITE_TAC[INSERT_UNION_EQ]);;
%



% EXPLORATION SECTION %


% Define n-depth calls. %

let Depth_calls =
    new_prim_rec_def `Depth_calls`
      "(Depth_calls 0 p1 s1 p2 s2 r = (p1 = p2) /\ (s1 = s2)) /\
       (Depth_calls (SUC n) p1 s1 p2 s2 r =
         (?p3 s3. M_calls p1 s1 [] p3 s3 r /\
                  Depth_calls n p3 s3 p2 s2 r))";;



let Depth_calls_END =
 prove_thm
  (`Depth_calls_END`,
   "!n p1 s1 p2 s2 r.
      Depth_calls (SUC n) p1 s1 p2 s2 r =
         (?p3 s3. Depth_calls n p1 s1 p3 s3 r /\
                  M_calls p3 s3 [] p2 s2 r)",
   INDUCT_TAC
   THENL
    [ REWRITE_TAC[Depth_calls]
      THEN REPEAT GEN_TAC
      THEN EQ_TAC
      THENL
       [ STRIP_TAC
         THEN EXISTS_TAC "p1:string"
         THEN EXISTS_TAC "s1:state"
         THEN UNDISCH_TAC "M_calls p1 s1[]p3 s3 r"
         THEN ASM_REWRITE_TAC[];

         STRIP_TAC
         THEN EXISTS_TAC "p2:string"
         THEN EXISTS_TAC "s2:state"
         THEN ASM_REWRITE_TAC[]
       ];

      ONCE_REWRITE_TAC[Depth_calls]
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT GEN_TAC
      THEN EQ_TAC
      THEN STRIP_TAC
      THEN EXISTS_TAC "p3':string"
      THEN EXISTS_TAC "s3':state"
      THEN ASM_REWRITE_TAC[]
      THEN EXISTS_TAC "p3:string"
      THEN EXISTS_TAC "s3:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< Depth_calls_END
g "!n p1 s1 p2 s2 r.
      Depth_calls (SUC n) p1 s1 p2 s2 r =
         (?p3 s3. Depth_calls n p1 s1 p3 s3 r /\
                  M_calls p3 s3 [] p2 s2 r)";;

e(INDUCT_TAC);;

 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT GEN_TAC);;
 e(EQ_TAC);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "p1:string");;
  e(EXISTS_TAC "s1:state");;
  e(UNDISCH_TAC "M_calls p1 s1[]p3 s3 r");;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "p2:string");;
  e(EXISTS_TAC "s2:state");;
  e(ASM_REWRITE_TAC[]);;

 e(ONCE_REWRITE_TAC[Depth_calls]);;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT GEN_TAC);;
 e(EQ_TAC);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "p3':string");;
  e(EXISTS_TAC "s3':state");;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "p3:string");;
  e(EXISTS_TAC "s3:state");;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "p3':string");;
  e(EXISTS_TAC "s3':state");;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "p3:string");;
  e(EXISTS_TAC "s3:state");;
  e(ASM_REWRITE_TAC[]);;
>%



let WFenv_pre_IMP_pre =
 prove_thm
  (`WFenv_pre_IMP_pre`,
   "!r p1 s1 p2 s2.
      WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      M_calls p1 s1 [] p2 s2 r ==>
      A ((FST o SND o SND o SND)(r p2)) s2",
   REPEAT STRIP_TAC
   THEN UNDISCH_TAC "WFenv_pre r"
   THEN REWRITE_TAC[WFenv_pre]
   THEN DISCH_THEN (MP_TAC o SPEC "p1:string")
   THEN REWRITE_TAC[WFproc_pre]
   THEN ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[PRE_PROGRESS;CPROGRESS]
   THEN IMP_RES_THEN REWRITE_THM C_calls_M_calls
   THEN POP_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
   THEN DISCH_THEN IMP_RES_TAC
  );;

%< WFenv_pre_IMP_pre
g "!r p1 s1 p2 s2.
      WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      M_calls p1 s1 [] p2 s2 r ==>
      A ((FST o SND o SND o SND)(r p2)) s2";;

e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "WFenv_pre r");;
e(REWRITE_TAC[WFenv_pre]);;
e(DISCH_THEN (MP_TAC o SPEC "p1:string"));;
e(REWRITE_TAC[WFproc_pre]);;
e(ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[PRE_PROGRESS;CPROGRESS]);;
e(IMP_RES_THEN REWRITE_THM C_calls_M_calls);;
e(POP_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
e(DISCH_THEN IMP_RES_TAC);;
>%


let pre_IMP_IN_all_ps =
 prove_thm
  (`pre_IMP_IN_all_ps`,
   "!r p all_ps.
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      A pre s1 ==>
      (p IN SL all_ps)",
   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(p:string) IN (SL all_ps)"
   THENL
    [ ASM_REWRITE_TAC[];

      RES_TAC
      THEN POP_ASSUM MP_TAC
      THEN POP_TAC
      THEN ASM_REWRITE_TAC[EMPTY_ENV]
      THEN DISCH_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[PAIR_EQ])
      THEN UNDISCH_TAC "A pre s1"
      THEN ASM_REWRITE_TAC[A_DEF]
    ]
  );;

%< pre_IMP_IN_all_ps
g "!r p all_ps.
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      A pre s1 ==>
      (p IN SL all_ps)";;

e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "(p:string) IN (SL all_ps)");;

  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
  e(POP_ASSUM MP_TAC);;
  e(POP_TAC);;
  e(ASM_REWRITE_TAC[EMPTY_ENV]);;
  e(DISCH_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[PAIR_EQ]));;
  e(UNDISCH_TAC "A pre s1");;
  e(ASM_REWRITE_TAC[A_DEF]);;
>%


let Depth_calls_pre =
 prove_thm
  (`Depth_calls_pre`,
   "!n r p1 s1 p2 s2.
      WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r ==>
      A ((FST o SND o SND o SND)(r p2)) s2",
   INDUCT_TAC
   THENL
    [ REWRITE_TAC[Depth_calls]
      THEN REPEAT STRIP_TAC
      THEN POP_ASSUM (REWRITE_THM o SYM)
      THEN POP_ASSUM (REWRITE_THM o SYM)
      THEN ASM_REWRITE_TAC[];

      REWRITE_TAC[Depth_calls]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre)
      THEN RES_TAC
    ]
  );;

%< Depth_calls_pre
g "!n r p1 s1 p2 s2.
      WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r ==>
      A ((FST o SND o SND o SND)(r p2)) s2";;

e(INDUCT_TAC);;

 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (REWRITE_THM o SYM));;
 e(POP_ASSUM (REWRITE_THM o SYM));;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre));;
 e(RES_TAC);;
>%


let Depth_call_IMP_SHALLOW_CALLS =
 prove_thm
  (`Depth_call_IMP_SHALLOW_CALLS`,
   "!n m p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      (m <= n) ==>
      ?p3 s3. Depth_calls m p1 s1 p3 s3 r",
   INDUCT_TAC
   THENL
    [ REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]
      THEN REWRITE_TAC[Depth_calls]
      THEN REPEAT STRIP_TAC
      THEN ASM_REWRITE_TAC[Depth_calls]
      THEN EXISTS_TAC "p2:string"
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[];

      REWRITE_TAC[LESS_OR_EQ]
      THEN REWRITE_TAC[LESS_EQ;LESS_EQ_MONO]
      THEN REWRITE_TAC[Depth_calls_END]
      THEN REPEAT STRIP_TAC
      THENL
       [ FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "m:num")
         THEN EXISTS_TAC "p3':string"
         THEN EXISTS_TAC "s3':state"
         THEN ASM_REWRITE_TAC[];

         POP_ASSUM REWRITE_THM
         THEN EXISTS_TAC "p2:string"
         THEN EXISTS_TAC "s2:state"
         THEN REWRITE_TAC[Depth_calls_END]
         THEN EXISTS_TAC "p3:string"
         THEN EXISTS_TAC "s3:state"
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< Depth_call_IMP_SHALLOW_CALLS
g "!n m p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      (m <= n) ==>
      ?p3 s3. Depth_calls m p1 s1 p3 s3 r";;

e(INDUCT_TAC);;

 e(REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]);;
 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT STRIP_TAC);;
 e(ASM_REWRITE_TAC[Depth_calls]);;
 e(EXISTS_TAC "p2:string");;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[LESS_OR_EQ]);;
 e(REWRITE_TAC[LESS_EQ;LESS_EQ_MONO]);;
 e(REWRITE_TAC[Depth_calls_END]);;
 e(REPEAT STRIP_TAC);;

  e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "m:num"));;
  e(EXISTS_TAC "p3':string");;
  e(EXISTS_TAC "s3':state");;
  e(ASM_REWRITE_TAC[]);;

  e(POP_ASSUM REWRITE_THM);;
  e(EXISTS_TAC "p2:string");;
  e(EXISTS_TAC "s2:state");;
  e(REWRITE_TAC[Depth_calls_END]);;
  e(EXISTS_TAC "p3:string");;
  e(EXISTS_TAC "s3:state");;
  e(ASM_REWRITE_TAC[]);;
>%


let terminates =
    new_def `terminates`
      "terminates p s r =
       let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
             (?s'. C c r s s')";;




let Depth_calls_IMP_terminates =
 prove_thm
  (`Depth_calls_IMP_terminates`,
   "!r.
      WFenv_term r /\ WFenv_pre r ==>
      (!n p1 s1.
        (!m p2 s2.
             (m <= n) /\
             Depth_calls m p1 s1 p2 s2 r ==>
             terminates p2 s2 r) /\
        (let (vars,vals,glbs,pre,post,calls,rec,c) = r p1 in
             A pre s1)
        ==>
        terminates p1 s1 r)",
   GEN_TAC
   THEN STRIP_TAC
   THEN INDUCT_TAC
   THENL
    [ REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]
      THEN REPEAT STRIP_TAC
      THEN UNDISCH_LAST_TAC
      THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN DISCH_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "0"
      THEN REWRITE_TAC[Depth_calls];

      REPEAT STRIP_TAC
      THEN UNDISCH_LAST_TAC
      THEN ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE)
      THEN IMP_RES_TAC WFenv_term_M_calls_TERM
      THEN UNDISCH_LAST_TAC
      THEN REWRITE_TAC[M_calls_TERM]
      THEN REWRITE_TAC[terminates]
      THEN POP_ASSUM ONCE_REWRITE_THM
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REPEAT STRIP_TAC
      THEN POP_ASSUM (\th. POP_ASSUM MATCH_MP_TAC
                           THEN ASSUME_TAC th)
      THEN FIRST_ASSUM REWRITE_THM
      THEN REWRITE_TAC[M_calls_terminate]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM (\th. (SPECL["p1:string";"s1:state"] th;
                              UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPECL["p2:string";"s2:state"])
      THEN REWRITE_TAC[terminates]
      THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p2" ENV_TUPLE]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN DISCH_THEN (MATCH_MP_TAC o IMP2AND)
      THEN IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre)
      THEN FIRST_ASSUM REWRITE_THM
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th)))
      THEN REWRITE_TAC[terminates]
      THEN DISCH_THEN (MP_TAC o ONCE_REWRITE_RULE[ISPEC "(r:env) p2" ENV_TUPLE])
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN DISCH_THEN (MATCH_MP_TAC o IMP2AND)
      THEN EXISTS_TAC "SUC m"
      THEN ASM_REWRITE_TAC[LESS_EQ_MONO]
      THEN REWRITE_TAC[Depth_calls]
      THEN EXISTS_TAC "p2:string"
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< Depth_calls_IMP_terminates
g "!r.
      WFenv_term r /\ WFenv_pre r ==>
      (!n p1 s1.
        (!m p2 s2.
             (m <= n) /\
             Depth_calls m p1 s1 p2 s2 r ==>
             terminates p2 s2 r) /\
        (let (vars,vals,glbs,pre,post,calls,rec,c) = r p1 in
             A pre s1)
        ==>
        terminates p1 s1 r)";;
e(GEN_TAC);;
e(STRIP_TAC);;
e(INDUCT_TAC);;

 e(REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]);;
 e(REPEAT STRIP_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(DISCH_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "0");;
 e(REWRITE_TAC[Depth_calls]);;

 e(REPEAT STRIP_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE));;
 e(IMP_RES_TAC WFenv_term_M_calls_TERM);;
 e(UNDISCH_LAST_TAC);;
 e(REWRITE_TAC[M_calls_TERM]);;
 e(REWRITE_TAC[terminates]);;
 e(POP_ASSUM ONCE_REWRITE_THM);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (\th. POP_ASSUM MATCH_MP_TAC
                   THEN ASSUME_TAC th));;
 e(FIRST_ASSUM REWRITE_THM);;
 e(REWRITE_TAC[M_calls_terminate]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (\th. (SPECL["p1:string";"s1:state"] th;
                      UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o SPECL["p2:string";"s2:state"]));;
 e(REWRITE_TAC[terminates]);;
 e(ONCE_REWRITE_TAC[ISPEC "(r:env) p2" ENV_TUPLE]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(DISCH_THEN (MATCH_MP_TAC o IMP2AND));;
 e(IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre));;
 e(FIRST_ASSUM REWRITE_THM);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th))));;
 e(REWRITE_TAC[terminates]);;
 e(DISCH_THEN (MP_TAC o ONCE_REWRITE_RULE[ISPEC "(r:env) p2" ENV_TUPLE]));;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(DISCH_THEN (MATCH_MP_TAC o IMP2AND));;
 e(EXISTS_TAC "SUC m");;
 e(ASM_REWRITE_TAC[LESS_EQ_MONO]);;
 e(REWRITE_TAC[Depth_calls]);;
 e(EXISTS_TAC "p2:string");;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let NOT_TERMINATES_IMP_DEEPER_CALL =
 prove_thm
  (`NOT_TERMINATES_IMP_DEEPER_CALL`,
   "!n p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      ~terminates p2 s2 r ==>
      ?p3 s3. Depth_calls (SUC n) p1 s1 p3 s3 r /\
              ~terminates p3 s3 r",
   REPEAT STRIP_TAC
   THEN UNDISCH_ALL_TAC
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p2" ENV_TUPLE)
   THEN IMP_RES_TAC WFenv_term_M_calls_TERM
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[M_calls_TERM;terminates]
   THEN POP_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "s2:state")
   THEN UNDISCH_LAST_TAC
   THEN IMP_RES_TAC (AND2IMP Depth_calls_pre)
   THEN FIRST_ASSUM REWRITE_THM
   THEN REWRITE_TAC[M_calls_terminate]
   THEN CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV)
   THEN REWRITE_TAC[NOT_IMP]
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN STRIP_TAC
   THEN EXISTS_TAC "p2':string"
   THEN EXISTS_TAC "s2':state"
   THEN UNDISCH_LAST_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p2'" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN DISCH_THEN REWRITE_THM
   THEN ASM_REWRITE_TAC[Depth_calls_END]
   THEN EXISTS_TAC "p2:string"
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;

%< NOT_TERMINATES_IMP_DEEPER_CALL
g "!n p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      ~terminates p2 s2 r ==>
      ?p3 s3. Depth_calls (SUC n) p1 s1 p3 s3 r /\
              ~terminates p3 s3 r";;
e(REPEAT STRIP_TAC);;
e(UNDISCH_ALL_TAC);;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p2" ENV_TUPLE));;
e(IMP_RES_TAC WFenv_term_M_calls_TERM);;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[M_calls_TERM;terminates]);;
e(POP_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "s2:state"));;
e(UNDISCH_LAST_TAC);;
e(IMP_RES_TAC (AND2IMP Depth_calls_pre));;
e(FIRST_ASSUM REWRITE_THM);;
e(REWRITE_TAC[M_calls_terminate]);;
e(CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV));;
e(REWRITE_TAC[NOT_IMP]);;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(STRIP_TAC);;
e(EXISTS_TAC "p2':string"
   THEN EXISTS_TAC "s2':state");;
e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p2'" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV));;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(DISCH_THEN REWRITE_THM);;
e(ASM_REWRITE_TAC[Depth_calls_END]);;
e(EXISTS_TAC "p2:string"
   THEN EXISTS_TAC "s2:state");;
e(ASM_REWRITE_TAC[]);;
>%


let NOT_TERMINATES_IMP_ALL_DEEPER_CALLS =
 prove_thm
  (`NOT_TERMINATES_IMP_ALL_DEEPER_CALLS`,
   "!m n p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      ~terminates p2 s2 r ==>
      ?p3 s3. Depth_calls (n + m) p1 s1 p3 s3 r /\
              ~terminates p3 s3 r",
   INDUCT_TAC
   THENL
    [ REPEAT STRIP_TAC
      THEN EXISTS_TAC "p2:string"
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[ADD_CLAUSES];

      REPEAT STRIP_TAC
      THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "n:num")
      THEN (IMP_RES_TAC o AND2IMP) NOT_TERMINATES_IMP_DEEPER_CALL
      THEN EXISTS_TAC "p3':string"
      THEN EXISTS_TAC "s3':state"
      THEN ASM_REWRITE_TAC[ADD_CLAUSES]
    ]
  );;

%< NOT_TERMINATES_IMP_ALL_DEEPER_CALLS
g "!m n p1 s1 p2 s2 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      Depth_calls n p1 s1 p2 s2 r /\
      ~terminates p2 s2 r ==>
      ?p3 s3. Depth_calls (n + m) p1 s1 p3 s3 r /\
              ~terminates p3 s3 r";;

e(INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;
 e(EXISTS_TAC "p2:string"
   THEN EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[ADD_CLAUSES]);;

 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "n:num"));;
 e((IMP_RES_TAC o AND2IMP) NOT_TERMINATES_IMP_DEEPER_CALL);;
 e(EXISTS_TAC "p3':string"
   THEN EXISTS_TAC "s3':state");;
 e(ASM_REWRITE_TAC[ADD_CLAUSES]);;
>%


let mk_sequence =
    new_prim_rec_def `mk_sequence`
      "(mk_sequence 0 r p s = (p,s)) /\
       (mk_sequence (SUC i) r p s =
             let (p',s') = (@(p',s'). M_calls p s [] p' s' r /\
                                      ~terminates p' s' r)
             in
             mk_sequence i r p' s')";;



let mk_sequence_PAIR =
 prove_thm
  (`mk_sequence_PAIR`,
   "!i r p s.
      mk_sequence (SUC i) r p s =
             let p' = FST (@(p',s'). M_calls p s [] p' s' r /\
                                     ~terminates p' s' r)
             in
             let s' = SND (@(p',s'). M_calls p s [] p' s' r /\
                                     ~terminates p' s' r)
             in
             mk_sequence i r p' s'",
   REWRITE_TAC[mk_sequence]
   THEN REPEAT GEN_TAC
   THEN CONV_TAC (RATOR_CONV (RAND_CONV (RAND_CONV
            (ONCE_REWRITE_CONV[SYM (ISPEC "x:string # state" PAIR)]))))
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

% mk_sequence_PAIR
g "!i r p s.
      mk_sequence (SUC i) r p s =
             let p' = FST (@(p',s'). M_calls p s [] p' s' r /\
                                     ~terminates p' s' r)
             in
             let s' = SND (@(p',s'). M_calls p s [] p' s' r /\
                                     ~terminates p' s' r)
             in
             mk_sequence i r p' s'";;
e(REWRITE_TAC[mk_sequence]);;
e(REPEAT GEN_TAC);;
e(CONV_TAC (RATOR_CONV (RAND_CONV (RAND_CONV
     (ONCE_REWRITE_CONV[SYM (ISPEC "x:string # state" PAIR)])))));;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
%


let SELECT_M_calls_NOT_TERMINATES =
 prove_thm
  (`SELECT_M_calls_NOT_TERMINATES`,
   "!p s r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p)) s /\
      ~terminates p s r ==>
      (\(p',s'). M_calls p s [] p' s' r /\ ~terminates p' s' r)
      (@(p',s'). M_calls p s [] p' s' r /\ ~terminates p' s' r)",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC SELECT_AX
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]
   THEN CONV_TAC (DEPTH_CONV PAIRED_BETA_CONV)
   THEN UNDISCH_LAST_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN IMP_RES_TAC WFenv_term_M_calls_TERM
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[M_calls_TERM;terminates]
   THEN POP_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "s2:state")
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[M_calls_terminate]
   THEN CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV)
   THEN REWRITE_TAC[NOT_IMP]
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN STRIP_TAC
   THEN EXISTS_TAC "(p2:string),(s2:state)"
   THEN ASM_REWRITE_TAC[]
   THEN UNDISCH_LAST_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p2" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN DISCH_THEN REWRITE_THM
  );;

% SELECT_M_calls_NOT_TERMINATES
g "!p s r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p)) s /\
      ~terminates p s r ==>
      (\(p',s'). M_calls p s [] p' s' r /\ ~terminates p' s' r)
      (@(p',s'). M_calls p s [] p' s' r /\ ~terminates p' s' r)";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC SELECT_AX);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]);;
e(CONV_TAC (DEPTH_CONV PAIRED_BETA_CONV));;
e(UNDISCH_LAST_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(IMP_RES_TAC WFenv_term_M_calls_TERM);;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[M_calls_TERM;terminates]);;
e(POP_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "s2:state"));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[M_calls_terminate]);;
e(CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV));;
e(REWRITE_TAC[NOT_IMP]);;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(STRIP_TAC);;
e(EXISTS_TAC "(p2:string),(s2:state)");;
e(ASM_REWRITE_TAC[]);;
e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p2" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV));;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(DISCH_THEN REWRITE_THM);;
%




let mk_sequence_Depth_calls_NOT_TERMINATES =
 prove_thm
  (`mk_sequence_Depth_calls_NOT_TERMINATES`,
   "!i p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
     ~terminates p1 s1 r ==>
     let (p2,s2) = mk_sequence i r p1 s1 in
     (Depth_calls i p1 s1 p2 s2 r /\
      ~terminates p2 s2 r)",
   ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_TAC
   THENL
    [ REWRITE_TAC[mk_sequence;Depth_calls]
      THEN REPEAT STRIP_TAC;

      REWRITE_TAC[mk_sequence_PAIR]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[Depth_calls]
      THEN REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP (SPEC "p1:string"
             ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
               o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
              SELECT_M_calls_NOT_TERMINATES)))
      THEN IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre)
      THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o
                         SPEC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                                             ~terminates p' s' r)")
      THEN POP_ASSUM REWRITE_THM
      THEN EXISTS_TAC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                                     ~terminates p' s' r)"
      THEN EXISTS_TAC "SND(@(p',s'). M_calls p1 s1[]p' s' r /\
                                     ~terminates p' s' r)"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% mk_sequence_Depth_calls_NOT_TERMINATES
g "!i p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
     ~terminates p1 s1 r ==>
     let (p2,s2) = mk_sequence i r p1 s1 in
     (Depth_calls i p1 s1 p2 s2 r /\
      ~terminates p2 s2 r)";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_TAC);;

 e(REWRITE_TAC[mk_sequence]);;
 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[mk_sequence_PAIR]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[Depth_calls]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP (SPEC "p1:string"
        ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
          o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
         SELECT_M_calls_NOT_TERMINATES))));;
 e(IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre));;
 e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o
                SPEC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                                    ~terminates p' s' r)"));;
 e(POP_ASSUM REWRITE_THM);;
 e(EXISTS_TAC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                             ~terminates p' s' r)");;
 e(EXISTS_TAC "SND(@(p',s'). M_calls p1 s1[]p' s' r /\
                             ~terminates p' s' r)");;
 e(ASM_REWRITE_TAC[]);;
%


let mk_sequence_calls_chain =
 prove_thm
  (`mk_sequence_calls_chain`,
   "!i p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
     ~terminates p1 s1 r ==>
     let (p2,s2) = mk_sequence i r p1 s1 in
     let (p3,s3) = mk_sequence (SUC i) r p1 s1 in
     M_calls p2 s2 [] p3 s3 r",

   ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_TAC
   THENL
    [ REWRITE_TAC[mk_sequence_PAIR]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[mk_sequence]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP
              ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
                o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
               SELECT_M_calls_NOT_TERMINATES));

      ONCE_REWRITE_TAC[mk_sequence_PAIR]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP
              ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
                o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
               SELECT_M_calls_NOT_TERMINATES))
      THEN IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre)
      THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o
                         SPEC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                                             ~terminates p' s' r)")
    ]
  );;

% mk_sequence_calls_chain
g "!i p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
     ~terminates p1 s1 r ==>
     let (p2,s2) = mk_sequence i r p1 s1 in
     let (p3,s3) = mk_sequence (SUC i) r p1 s1 in
     M_calls p2 s2 [] p3 s3 r";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "x:string # state" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_TAC);;

 e(REWRITE_TAC[mk_sequence_PAIR]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[mk_sequence]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP
        ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
          o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
         SELECT_M_calls_NOT_TERMINATES)));;

 e(ONCE_REWRITE_TAC[mk_sequence_PAIR]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP
        ((CONV_RULE (DEPTH_CONV PAIRED_BETA_CONV)
          o ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
         SELECT_M_calls_NOT_TERMINATES)));;
 e(IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre));;
 e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o
                SPEC "FST(@(p',s'). M_calls p1 s1[]p' s' r /\
                                    ~terminates p' s' r)"));;
%




let induct_start =
    new_rec_def `induct_start` aexp_Axiom
      "(induct_start m AFALSE = ATRUE) /\
       (induct_start m (ALESS e1 e2) = (AEQ e1 (ANUM m)))";;


let induct_start_num =
    new_rec_def `induct_start_num` aexp_Axiom
      "(induct_start_num s AFALSE = 0) /\
       (induct_start_num s (ALESS e1 e2) = V e1 s)";;


let induct_start_THM =
 prove_thm
  (`induct_start_THM`,
   "!r p s vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c))  ==>
       A (induct_start (induct_start_num s rec) rec) s",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC)
   THEN ASM_REWRITE_TAC[induct_start_num;induct_start;A_DEF;V_DEF]
  );;

%< induct_start_THM
g "!r p s vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c))  ==>
       A (induct_start (induct_start_num s rec) rec) s";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC));;
e(ASM_REWRITE_TAC[induct_start_num;induct_start;A_DEF;V_DEF]);;
>%


let sequence =
    new_def `sequence`
      "sequence ps sts ns r =
       (!i. M_calls (ps i) (sts i) [] (ps (SUC i)) (sts (SUC i)) r) /\
       (!i. ns i = let (vars,vals,glbs,pre,post,calls,rec,c) = r (ps i) in
                   induct_start_num (sts i) rec)";;


let NOT_Depth_terminates_IMP_sequence =
 prove_thm
  (`NOT_Depth_terminates_IMP_sequence`,
   "!p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      ~(?n.
         (!m p2 s2.
           (m <= n) /\
           Depth_calls m p1 s1 p2 s2 r ==>
           terminates p2 s2 r))
      ==>
      (?ps sts ns.
        sequence ps sts ns r /\
        (ps 0 = p1) /\ (sts 0 = s1))",

   CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV)
   THEN REWRITE_TAC[NOT_IMP;DE_MORGAN_THM;(SYM o SPEC_ALL)CONJ_ASSOC]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o SPEC "0")
   THEN UNDISCH_TAC "m <= 0"
   THEN REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]
   THEN DISCH_THEN REWRITE_ALL_THM
   THEN UNDISCH_TAC "Depth_calls 0 p1 s1 p2 s2 r"
   THEN REWRITE_TAC[Depth_calls]
   THEN DISCH_THEN (REWRITE_ALL_TAC o map SYM o CONJUNCTS)
   THEN EXISTS_TAC "\i. FST(mk_sequence i r p1 s1)"
   THEN EXISTS_TAC "\i. SND(mk_sequence i r p1 s1)"
   THEN EXISTS_TAC "\i. let (vars,vals,glbs,pre,post,calls,rec,c) =
                            (r:env) (FST(mk_sequence i r p1 s1)) in
                        induct_start_num (SND(mk_sequence i r p1 s1)) rec"
   THEN REWRITE_TAC[sequence]
   THEN BETA_TAC
   THEN REWRITE_TAC[]
   THEN CONJ_TAC
   THENL
    [ IMP_RES_TAC (AND2IMP mk_sequence_calls_chain)
      THEN POP_ASSUM (MP_TAC o
              ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)])
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[];

      REWRITE_TAC[mk_sequence]
    ]
  );;

%< NOT_Depth_terminates_IMP_sequence
g "!p1 s1 r.
      WFenv_term r /\ WFenv_pre r /\
      A ((FST o SND o SND o SND)(r p1)) s1 /\
      ~(?n.
         (!m p2 s2.
           (m <= n) /\
           Depth_calls m p1 s1 p2 s2 r ==>
           terminates p2 s2 r))
      ==>
      (?ps sts ns.
        sequence ps sts ns r /\
        (ps 0 = p1) /\ (sts 0 = s1))";;

e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(CONV_TAC (TOP_DEPTH_CONV NOT_FORALL_CONV));;
e(REWRITE_TAC[NOT_IMP;DE_MORGAN_THM;(SYM o SPEC_ALL)CONJ_ASSOC]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(FIRST_ASSUM (STRIP_ASSUME_TAC o SPEC "0"));;
e(UNDISCH_TAC "m <= 0");;
e(REWRITE_TAC[LESS_OR_EQ;NOT_LESS_0]);;
e(DISCH_THEN REWRITE_ALL_THM);;
e(UNDISCH_TAC "Depth_calls 0 p1 s1 p2 s2 r");;
e(REWRITE_TAC[Depth_calls]);;
e(DISCH_THEN (REWRITE_ALL_TAC o map SYM o CONJUNCTS));;
e(EXISTS_TAC "\i. FST(mk_sequence i r p1 s1)");;
e(EXISTS_TAC "\i. SND(mk_sequence i r p1 s1)");;
e(EXISTS_TAC "\i. let (vars,vals,glbs,pre,post,calls,rec,c) =
                      (r:env) (FST(mk_sequence i r p1 s1)) in
                  induct_start_num (SND(mk_sequence i r p1 s1)) rec");;
e(REWRITE_TAC[sequence]);;
e(BETA_TAC);;
e(REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(IMP_RES_TAC (AND2IMP mk_sequence_calls_chain));;
 e(POP_ASSUM (MP_TAC o
       ONCE_REWRITE_RULE[SYM (ISPEC "x:string # state" PAIR)]));;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[mk_sequence]);;
>%




let sequence_M_calls_ADD =
 prove_thm
  (`sequence_M_calls_ADD`,
   "!k i r ps sts ns.
      sequence ps sts ns r ==>
      ?ps'. M_calls (ps i) (sts i) ps' (ps (SUC i + k)) (sts (SUC i + k)) r",

   INDUCT_TAC
   THENL
    [ REWRITE_TAC[sequence]
      THEN REPEAT STRIP_TAC
      THEN EXISTS_TAC "[]:string list"
      THEN ASM_REWRITE_TAC[ADD_CLAUSES];

      REPEAT STRIP_TAC
      THEN RES_TAC
      THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC "SUC i")
      THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD_SUC]
      THEN EXISTS_TAC "CONS (ps (SUC i)) ps' :string list"
      THEN REWRITE_TAC[M_calls_CONS]
      THEN EXISTS_TAC "sts (SUC i) :state"
      THEN REWRITE_ALL_TAC[sequence;ADD_CLAUSES]
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< sequence_M_calls_ADD
g "!k i r ps sts ns.
      sequence ps sts ns r ==>
      ?ps'. M_calls (ps i) (sts i) ps' (ps (SUC i + k)) (sts (SUC i + k)) r";;
e(INDUCT_TAC);;

 e(REWRITE_TAC[sequence]);;
 e(REPEAT STRIP_TAC);;
 e(EXISTS_TAC "[]:string list");;
 e(ASM_REWRITE_TAC[ADD_CLAUSES]);;

 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC "SUC i"));;
 e(REWRITE_TAC[(SYM o SPEC_ALL)ADD_SUC]);;
 e(EXISTS_TAC "CONS (ps (SUC i)) ps' :string list");;
 e(REWRITE_TAC[M_calls_CONS]);;
 e(EXISTS_TAC "sts (SUC i) :state");;
 e(REWRITE_ALL_TAC[sequence;ADD_CLAUSES]);;
 e(ASM_REWRITE_TAC[]);;
>%


let sequence_M_calls =
 prove_thm
  (`sequence_M_calls1`,
   "!r ps sts ns i j.
      sequence ps sts ns r /\
      (i < j) ==>
      ?ps'. M_calls (ps i) (sts i) ps' (ps j) (sts j) r",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC LESS_ADD_1
   THEN POP_ASSUM REWRITE_THM
   THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD1;ADD_CLAUSES]
   THEN REWRITE_TAC[(SYM o SPEC_ALL o CONJUNCT2)ADD]
   THEN IMP_RES_TAC sequence_M_calls_ADD
   THEN ASM_REWRITE_TAC[]
  );;

%< sequence_M_calls
g "!r ps sts ns i j.
      sequence ps sts ns r /\
      (i < j) ==>
      ?ps'. M_calls (ps i) (sts i) ps' (ps j) (sts j) r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC LESS_ADD_1);;
e(POP_ASSUM REWRITE_THM);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1;ADD_CLAUSES]);;
e(REWRITE_TAC[(SYM o SPEC_ALL o CONJUNCT2)ADD]);;
e(IMP_RES_TAC sequence_M_calls_ADD);;
e(ASM_REWRITE_TAC[]);;
>%



let WFenv_rec_REC_PROGRESS =
 prove_thm
  (`WFenv_rec_REC_PROGRESS`,
   "!r p vars vals glbs pre post calls rec c.
       WFenv_rec r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r)",
   REPEAT STRIP_TAC
   THEN UNDISCH_TAC "WFenv_rec r"
   THEN REWRITE_TAC[WFenv_rec]
   THEN DISCH_THEN (MP_TAC o SPEC "p:string")
   THEN ASM_REWRITE_TAC[WFproc_rec]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

%< WFenv_rec_REC_PROGRESS
g "!r p vars vals glbs pre post calls rec c.
       WFenv_rec r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
       REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r)";;
e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "WFenv_rec r");;
e(REWRITE_TAC[WFenv_rec]);;
e(DISCH_THEN (MP_TAC o SPEC "p:string"));;
e(ASM_REWRITE_TAC[WFproc_rec]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
>%



let REC_VAR_NON_INTERFERENCE =
 prove_thm
  (`REC_VAR_NON_INTERFERENCE`,
   "!r p y v vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       ~WFx y /\
       (FVv v SUBSET SL(APPEND (APPEND vars vals) glbs)) ==>
       ~(y IN FVa pre) /\
       ~(y IN FVc c r) /\
       ~(y IN FVv v)",
   REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN IMP_RES_TAC WFxs_NOT_WFx
   THEN IMP_RES_TAC (AND2IMP NOT_IN_SUBSET)
   THEN ASM_REWRITE_TAC[]
  );;

%< REC_VAR_NON_INTERFERENCE
g "!r p y v vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       ~WFx y /\
       (FVv v SUBSET SL(APPEND (APPEND vars vals) glbs)) ==>
       ~(y IN FVa pre) /\
       ~(y IN FVc c r) /\
       ~(y IN FVv v)";;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(IMP_RES_TAC WFxs_NOT_WFx);;
e(IMP_RES_TAC (AND2IMP NOT_IN_SUBSET));;
e(ASM_REWRITE_TAC[]);;
>%


let sequence_pre =
 prove_thm
  (`sequence_pre`,
   "!i r ps sts ns.
      WFenv_pre r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) ==>
      A ((FST o SND o SND o SND)(r (ps i))) (sts i)",
   INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN UNDISCH_TAC "sequence ps sts ns r"
   THEN REWRITE_TAC[sequence]
   THEN DISCH_THEN (ASSUME_TAC o SPEC_ALL o CONJUNCT1)
   THEN IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre)
  );;

%< sequence_pre
g "!i r ps sts ns.
      WFenv_pre r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) ==>
      A ((FST o SND o SND o SND)(r (ps i))) (sts i)";;

e(INDUCT_TAC);;

 e(REPEAT STRIP_TAC);;

 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(UNDISCH_TAC "sequence ps sts ns r");;
 e(REWRITE_TAC[sequence]);;
 e(DISCH_THEN (ASSUME_TAC o SPEC_ALL o CONJUNCT1));;
 e(IMP_RES_TAC (AND2IMP WFenv_pre_IMP_pre));;
>%



let sequence_DECREASING_VALUES =
 prove_thm
  (`sequence_DECREASING_VALUES`,
   "!r ps sts ns p i j vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (ps i = p) /\
      (ps j = p) /\
      (i < j) ==>
      (ns j < ns i)",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP sequence_pre)
   THEN POP_ASSUM (MP_TAC o SPEC_ALL)
   THEN IMP_RES_TAC (AND2IMP sequence_M_calls)
   THEN UNDISCH_LAST_TAC
   THEN ASM_REWRITE_TAC[o_THM]
   THEN REPEAT DISCH_TAC
   THEN IMP_RES_TAC (AND2IMP WFenv_rec_REC_PROGRESS)
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[REC_PROGRESS;PATH_PROGRESS;A_DEF]
   THEN IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC)
   THENL
    [ DISCH_THEN (MP_TAC o SPECL["ps':string list";
                                 "(sts:num->state) i";
                                 "(sts:num->state) j"])
      THEN ASM_REWRITE_TAC[induct_pre;A_DEF];

      DISCH_THEN (MP_TAC o
                  SPECL["ps':string list";
                        "BND (ns i) y ((sts:num->state) i)";
                        "BND (ns (i:num)) y ((sts:num->state) j)"])
      THEN MP_TAC (ASSUME "sequence ps sts ns r")
      THEN REWRITE_TAC[sequence]
      THEN DISCH_THEN (ASSUME_TAC o CONJUNCT2)
      THEN FIRST_ASSUM (MP_TAC o SPEC "i:num")
      THEN POP_ASSUM (MP_TAC o SPEC "j:num")
      THEN ASM_REWRITE_TAC[]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[induct_start_num;induct_pre;A_DEF;V_DEF]
      THEN DISCH_TAC
      THEN DISCH_TAC
      THEN IMP_RES_TAC (AND2IMP REC_VAR_NON_INTERFERENCE)
      THEN IMP_RES_THEN REWRITE_THM A_BND_THM
      THEN IMP_RES_THEN REWRITE_THM V_BND_THM
      THEN REWRITE_TAC[BND_MATCH]
      THEN IMP_RES_TAC (AND2IMP M_calls_BND_THM)
      THEN POP_ASSUM REWRITE_THM
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< sequence_DECREASING_VALUES
g "!r ps sts ns p i j vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (ps i = p) /\
      (ps j = p) /\
      (i < j) ==>
      (ns j < ns i)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP sequence_pre));;
e(POP_ASSUM (MP_TAC o SPEC_ALL));;
e(IMP_RES_TAC (AND2IMP sequence_M_calls));;
e(UNDISCH_LAST_TAC);;
e(ASM_REWRITE_TAC[o_THM]);;
e(REPEAT DISCH_TAC);;
e(IMP_RES_TAC (AND2IMP WFenv_rec_REC_PROGRESS));;
e(UNDISCH_LAST_TAC
    THEN REWRITE_TAC[REC_PROGRESS;PATH_PROGRESS;A_DEF]);;
e(IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC));;

  e(DISCH_THEN (MP_TAC o SPECL["ps':string list";
                               "(sts:num->state) i";
                               "(sts:num->state) j"]));;
  e(ASM_REWRITE_TAC[induct_pre;A_DEF]);;

  e(DISCH_THEN (MP_TAC o
                SPECL["ps':string list";
                      "BND (ns i) y ((sts:num->state) i)";
                      "BND (ns (i:num)) y ((sts:num->state) j)"]));;
  e(MP_TAC (ASSUME "sequence ps sts ns r"));;
  e(REWRITE_TAC[sequence]);;
  e(DISCH_THEN (ASSUME_TAC o CONJUNCT2));;
  e(FIRST_ASSUM (MP_TAC o SPEC "i:num"));;
  e(POP_ASSUM (MP_TAC o SPEC "j:num"));;
  e(ASM_REWRITE_TAC[]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[induct_start_num;induct_pre;A_DEF;V_DEF]);;
  e(DISCH_TAC);;
  e(DISCH_TAC);;
  e(IMP_RES_TAC (AND2IMP REC_VAR_NON_INTERFERENCE)
       THEN IMP_RES_THEN REWRITE_THM A_BND_THM
       THEN IMP_RES_THEN REWRITE_THM V_BND_THM
       THEN REWRITE_TAC[BND_MATCH]
       THEN IMP_RES_TAC (AND2IMP M_calls_BND_THM)
       THEN POP_ASSUM REWRITE_THM
       THEN ASM_REWRITE_TAC[]);;
>%




let WELL_FOUNDED_NUM_TAC :tactic = \(asl,gl).
     (let n = (fst o dest_forall) gl in
        (GEN_TAC
         THEN UNBETA_TAC n
         THEN SPEC_TAC(n,n)
         THEN MATCH_MP_TAC NUM_WELL_FOUNDED
         THEN BETA_TAC
         THEN GEN_TAC
         THEN DISCH_TAC)) (asl,gl)
      ? failwith `WELL_FOUNDED_NUM_TAC`;;





let sequence_OCC_IMP_LIMIT_ON_OCCS =
 prove_thm
  (`sequence_OCC_IMP_LIMIT_ON_OCCS`,
   "!n i p r ps sts ns vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (ps i = p) /\
      (ns i = n) ==>
      (?m. !j. (m < j) ==> ~(ps j = p))",

   WELL_FOUNDED_NUM_TAC
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "?j. (i < j) /\ (ps j = (p:string))"
   THENL
    [ POP_ASSUM STRIP_ASSUME_TAC
      THEN IMP_RES_TAC (AND2IMP sequence_DECREASING_VALUES)
      THEN FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC "(ns:num->num) j")
      THEN UNDISCH_TAC "((ns:num->num) j) < (ns i)"
      THEN REWRITE_TAC[ASSUME "(ns:num->num) i = n"]
      THEN DISCH_THEN (\th1. DISCH_THEN (\th2. MP_TAC (MP th2 th1)))
      THEN DISCH_THEN (MP_TAC o SPEC_ALL o
                       SPECL["j:num";"p:string";"r:env";"ps:num->string";
                             "sts:num->state";"ns:num->num"])
      THEN DISCH_THEN MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[];

      UNDISCH_LAST_TAC
      THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
      THEN REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]
      THEN DISCH_TAC
      THEN EXISTS_TAC "i:num"
      THEN POP_ASSUM REWRITE_THM
    ]
  );;

%< sequence_OCC_IMP_LIMIT_ON_OCCS
g "!n i p r ps sts ns vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (ps i = p) /\
      (ns i = n) ==>
      (?m. !j. (m < j) ==> ~(ps j = p))";;

e(WELL_FOUNDED_NUM_TAC);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "?j. (i < j) /\ (ps j = (p:string))");;

 e(POP_ASSUM STRIP_ASSUME_TAC);;
 e(IMP_RES_TAC (AND2IMP sequence_DECREASING_VALUES));;
 e(FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o SPEC "(ns:num->num) j"));;
 e(UNDISCH_TAC "((ns:num->num) j) < (ns i)");;
 e(REWRITE_TAC[ASSUME "(ns:num->num) i = n"]);;
 e(DISCH_THEN (\th1. DISCH_THEN (\th2. MP_TAC (MP th2 th1))));;
 e(DISCH_THEN (MP_TAC o SPEC_ALL o
               SPECL["j:num";"p:string";"r:env";"ps:num->string";
                     "sts:num->state";"ns:num->num"]));;
 e(DISCH_THEN MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(UNDISCH_LAST_TAC);;
 e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
 e(REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]);;
 e(DISCH_TAC);;
 e(EXISTS_TAC "i:num");;
 e(POP_ASSUM REWRITE_THM);;
>%


let sequence_LIMIT_ON_OCCS =
 prove_thm
  (`sequence_LIMIT_ON_OCCS`,
   "!p r ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (?m. !j. (m < j) ==> ~(ps j = p))",

   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "?i:num. (ps i = (p:string))"
   THENL
    [ POP_ASSUM STRIP_ASSUME_TAC
      THEN ASSUME_TAC (REFL "(ns:num->num) i")
      THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
      THEN IMP_RES_TAC (AND2IMP sequence_OCC_IMP_LIMIT_ON_OCCS)
      THEN EXISTS_TAC "m:num"
      THEN POP_ASSUM REWRITE_THM;

      UNDISCH_LAST_TAC
      THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
      THEN DISCH_THEN REWRITE_THM
    ]
  );;

%< sequence_LIMIT_ON_OCCS
g "!p r ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (?m. !j. (m < j) ==> ~(ps j = p))";;

e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "?i:num. (ps i = (p:string))");;

 e(POP_ASSUM STRIP_ASSUME_TAC);;
 e(ASSUME_TAC (REFL "(ns:num->num) i"));;
 e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
 e(IMP_RES_TAC (AND2IMP sequence_OCC_IMP_LIMIT_ON_OCCS));;
 e(EXISTS_TAC "m:num");;
 e(POP_ASSUM REWRITE_THM);;

 e(UNDISCH_LAST_TAC);;
 e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
 e(DISCH_THEN REWRITE_THM);;
>%


let sequence_IN_all_ps =
 prove_thm
  (`sequence_IN_all_ps`,
   "!all_ps r ps sts ns.
      WFenv_pre r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (!i. ps i IN SL all_ps)",

   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(ps:num->string) i IN SL all_ps"
   THENL
    [ POP_ASSUM REWRITE_THM;

      IMP_RES_TAC (AND2IMP sequence_pre)
      THEN POP_ASSUM (MP_TAC o SPEC_ALL)
      THEN RES_TAC
      THEN POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[EMPTY_ENV;o_THM;A_DEF]
    ]
  );;

%< sequence_IN_all_ps
g "!all_ps r ps sts ns.
      WFenv_pre r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (!i. ps i IN SL all_ps)";;

e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "(ps:num->string) i IN SL all_ps");;

 e(POP_ASSUM REWRITE_THM);;

 e(IMP_RES_TAC (AND2IMP sequence_pre));;
 e(POP_ASSUM (MP_TAC o SPEC_ALL));;
 e(RES_TAC);;
 e(POP_ASSUM REWRITE_THM);;
 e(REWRITE_TAC[EMPTY_ENV;o_THM;A_DEF]);;
>%


let sequence_LIMIT =
 prove_thm
  (`sequence_LIMIT`,
   "!all_ps r ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (?m. !p. (p IN SL all_ps) ==> !j. (m < j) ==> ~(ps j = p))",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "r:env")
   THEN IMP_RES_TAC (AND2IMP sequence_LIMIT_ON_OCCS)
   THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC "h:string" )
   THEN EXISTS_TAC "m' + m"
   THEN GEN_TAC THEN STRIP_TAC
   THENL
    [ GEN_TAC THEN DISCH_TAC
      THEN ASM_REWRITE_TAC[]
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_LAST_TAC
      THEN REWRITE_TAC[LESS_EQ]
      THEN DISCH_TAC
      THEN IMP_RES_TAC LESS_EQ_TRANS
      THEN POP_ASSUM MATCH_MP_TAC
      THEN REWRITE_TAC[LESS_EQ_MONO;LESS_EQ_ADD];

      GEN_TAC THEN DISCH_TAC
      THEN FIRST_ASSUM (IMP_RES_TAC o SPEC "p:string")
      THEN POP_ASSUM MATCH_MP_TAC
      THEN UNDISCH_LAST_TAC
      THEN REWRITE_TAC[LESS_EQ]
      THEN ONCE_REWRITE_TAC[ADD_SYM]
      THEN DISCH_TAC
      THEN IMP_RES_TAC LESS_EQ_TRANS
      THEN POP_ASSUM MATCH_MP_TAC
      THEN REWRITE_TAC[LESS_EQ_MONO;LESS_EQ_ADD]
    ]
  );;

%< sequence_LIMIT
g "!all_ps r ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      sequence ps sts ns r /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      (?m. !p. (p IN SL all_ps) ==> !j. (m < j) ==> ~(ps j = p))";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (IMP_RES_TAC o AND2IMP o SPEC "r:env"));;
 e(IMP_RES_TAC (AND2IMP sequence_LIMIT_ON_OCCS));;
 e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC "h:string" ));;
 e(EXISTS_TAC "m' + m");;
 e(GEN_TAC THEN STRIP_TAC);;

  e(GEN_TAC THEN DISCH_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[LESS_EQ]);;
  e(DISCH_TAC);;
  e(IMP_RES_TAC LESS_EQ_TRANS);;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(REWRITE_TAC[LESS_EQ_MONO;LESS_EQ_ADD]);;

  e(GEN_TAC THEN DISCH_TAC);;
  e(FIRST_ASSUM (IMP_RES_TAC o SPEC "p:string"));;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[LESS_EQ]);;
  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(DISCH_TAC);;
  e(IMP_RES_TAC LESS_EQ_TRANS);;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(REWRITE_TAC[LESS_EQ_MONO;LESS_EQ_ADD]);;
>%


let sequence_CONTRADICTION =
 prove_thm
  (`sequence_CONTRADICTION`,
   "!r all_ps ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      ~sequence ps sts ns r",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP sequence_LIMIT)
   THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC (AND2IMP sequence_IN_all_ps)
   THEN POP_ASSUM (STRIP_ASSUME_TAC o SPEC "SUC m")
   THEN RES_TAC
   THEN POP_ASSUM (MP_TAC o SPEC "SUC m")
   THEN REWRITE_TAC[LESS_SUC_REFL]
  );;

%< sequence_CONTRADICTION
g "!r all_ps ps sts ns.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      A ((FST o SND o SND o SND)(r (ps 0))) (sts 0)  ==>
      ~sequence ps sts ns r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP sequence_LIMIT));;
e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC (AND2IMP sequence_IN_all_ps));;
e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC "SUC m"));;
e(RES_TAC);;
e(POP_ASSUM (MP_TAC o SPEC "SUC m"));;
e(REWRITE_TAC[LESS_SUC_REFL]);;
>%


let PROCEDURE_TERMINATION =
 prove_thm
  (`PROCEDURE_TERMINATION`,
   "!r all_ps p s vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      WFenv_term r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      A pre s  ==>
      terminates p s r",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP Depth_calls_IMP_terminates)
   THEN POP_ASSUM (MATCH_MP_TAC o IMP2AND)
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN ASM_REWRITE_TAC[]
   THEN ASM_CASES_TAC
     "?n. !m p2 s2. m <= n /\ Depth_calls m p s p2 s2 r ==>
                    terminates p2 s2 r"
   THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN MP_TAC (SPECL["p:string";"s:state";"r:env"]
                      NOT_Depth_terminates_IMP_sequence)
   THEN ASM_REWRITE_TAC[o_THM]
   THEN STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP sequence_CONTRADICTION)
   THEN POP_ASSUM (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[o_THM]
   THEN DISCH_THEN (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
  );;

%< PROCEDURE_TERMINATION
g "!r all_ps p s vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      WFenv_pre r /\
      WFenv_rec r /\
      WFenv_term r /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      A pre s  ==>
      terminates p s r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP Depth_calls_IMP_terminates));;
e(POP_ASSUM (MATCH_MP_TAC o IMP2AND));;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(ASM_REWRITE_TAC[]);;
e(ASM_CASES_TAC
     "?n. !m p2 s2. m <= n /\ Depth_calls m p s p2 s2 r ==>
                    terminates p2 s2 r"
   THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
e(MP_TAC (SPECL["p:string";"s:state";"r:env"]
                      NOT_Depth_terminates_IMP_sequence));;
e(ASM_REWRITE_TAC[o_THM]);;
e(STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP sequence_CONTRADICTION));;
e(POP_ASSUM (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[o_THM]);;
e(DISCH_THEN (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[]);;
>%




close_theory ();;

