% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of functions to translate
  programming language expressions into their equivalent assertion
  language expressions and symbolic states.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `translations`;;

new_parent `free_variables`;;


% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of translations: %

loadf `autoload`;;

autoload_all_defs_and_thms
   `more_finite_sets assert_syntax assert_semantics bindings
    prog_syntax prog_semantics substitutions free_variables`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;



% ========================================================= %
% First we need a "translator continuation" function.       %
% ========================================================= %

let trans_cont = new_infix_definition
 (`trans_cont`,
  "$--> (e:**->(* # **)) (k:*->**->***) =
        \s. k (FST (e s)) (SND (e s))");;

let TRANS_CONT_TAC =
    REWRITE_TAC[trans_cont]  THEN
    BETA_TAC  THEN
    PURE_REWRITE_TAC[FST;SND];;

let trans_cont_THM =
 prove_thm
  (`trans_cont_THM`,
   "!e (k:*->**->***) s.
      (e --> k) s = k (FST(e s)) (SND(e s))",
   REWRITE_TAC[trans_cont]
   THEN BETA_TAC
   THEN REWRITE_TAC[]
  );;

% trans_cont_THM
g "!e (k:*->**->***) s.
     (e --> k) s = k (FST(e s)) (SND(e s))";;
e(REWRITE_TAC[trans_cont]);;
e(BETA_TAC);;
e(REWRITE_TAC[]);;
%


% ========================================================= %
% Load the "translator constructor" function.               %
% ========================================================= %

loadf `new_translation`;;


% ========================================================= %
% The function VE translates numeric expressions from the   %
% programming language into the assertion language.         %
% ========================================================= %

let VE1_DEF = 
    new_rec_def `VE1_DEF` exp_Axiom
      "(VE1 (NUM n)  ss = ANUM n, ss)  /\
       (VE1 (PVAR x) ss = ss x, ss)  /\
       (VE1 (INC x)  ss = APLUS (ss x) (ANUM 1),
                         BND (APLUS (ss x) (ANUM 1)) x ss)  /\
       (VE1 (PLUS e1 e2) ss = 
            (VE1 e1 --> \v1. VE1 e2 --> \v2 ss1. APLUS v1 v2, ss1) ss)  /\
       (VE1 (MINUS e1 e2) ss = 
            (VE1 e1 --> \v1. VE1 e2 --> \v2 ss1. AMINUS v1 v2, ss1) ss)  /\
       (VE1 (MULT e1 e2) ss = 
            (VE1 e1 --> \v1. VE1 e2 --> \v2 ss1. AMULT v1 v2, ss1) ss)";;


let VE_DEF = new_definition
 (`VE_DEF`,
  "VE v =  FST (VE1 v IDENT_SS)");;

let VE_state_DEF = new_definition
 (`VE_state_DEF`,
  "VE_state v =  SND (VE1 v IDENT_SS)");;


let VE_LEMMA,VE_THM =
             prove_translation_thm `VE_LEMMA` `VE_THM` `E` `V`
             VE1_DEF [VE_DEF;VE_state_DEF]
             "E e s1 n s2 =    (n = V (VE e) s1)    /\
                               (s2 = s1 <|s (VE_state e))"
             [
             % NUM case %
                         NUM_TAC;
             % PVAR case %
                         PVAR_TAC;
             % INC case %
                         INC_TAC
                    THEN PVAR_TAC;
             % PLUS case %
                         PLUS_TAC
                    THEN EXISTS_TAC "s1 <|s (VE_state e)";
             % MINUS case %
                         MINUS_TAC
                    THEN EXISTS_TAC "s1 <|s (VE_state e)";
             % MULT case %
                         MULT_TAC
                    THEN EXISTS_TAC "s1 <|s (VE_state e)"
             ]
             [];;


% ========================================================== %
% The function VES translates lists of numeric expressions   %
% from the programming language into the assertion language. %
% ========================================================== %

let VES1_DEF = 
    new_list_rec_def `VES1_DEF`
      "(VES1 NIL ss  =  NIL, ss)  /\
       (VES1 (CONS e es) ss  =
            (VE1 e --> \v. VES1 es --> \vs ss1. CONS v vs, ss1) ss)";;


let VES_DEF = new_definition
 (`VES_DEF`,
  "VES es =  FST (VES1 es IDENT_SS)");;

let VES_state_DEF = new_definition
 (`VES_state_DEF`,
  "VES_state es =  SND (VES1 es IDENT_SS)");;




% ====================================================== %
% Here is the key lemma of the translation of lists of   %
% numeric-valued assertion expressions:                  %
% ====================================================== %

let VES_LEMMA =
 prove_thm
  (`VES_LEMMA`,
   "!es s ss.
    (       VS (FST (VES1 es ss)) s
     =      VS (VES es) (s <|s ss) )
  /\   ((s <|s (SND (VES1 es ss)))
     = ((s <|s ss) <|s (VES_state es)) )",

   LIST_INDUCT_TAC
    THEN REPEAT GEN_TAC
    THEN REWRITE_TAC[VES_DEF;VES_state_DEF;VES1_DEF;VS_DEF]
    THEN REWRITE_TAC[trans_cont]
    THEN BETA_TAC
    THEN ASM_REWRITE_TAC[VE_LEMMA;VS_DEF;s_subst_IDENT]
  );;


% The central VES_LEMMA:
g "!es s ss.
    (       VS (FST (VES1 es ss)) s
     =      VS (VES es) (s <|s ss) )
  /\   ((s <|s (SND (VES1 es ss)))
     = ((s <|s ss) <|s (VES_state es)) )";;

e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[VS_DEF;VES_DEF;VES_state_DEF;VES1_DEF]);;
e(REWRITE_TAC[s_subst_IDENT]);;

e(REWRITE_TAC[VES_DEF;VES_state_DEF;VES1_DEF;VS_DEF]);;
e(REWRITE_TAC[trans_cont]);;
e(BETA_TAC);;
e(ASM_REWRITE_TAC[VE_LEMMA;VS_DEF;s_subst_IDENT]);;
%



% =========================================================== %
% Here is the first half of the fundamental theorem on the    %
% translation of lists of numeric-valued expressions.         %
% =========================================================== %

let VES_UNIQUE =
 prove_thm
  (`VES_UNIQUE`,
   "!es s1 ns s2. ES es s1 ns s2 ==>
                      (ns = VS(VES es) s1)
                  /\  (s2 = s1 <|s VES_state es)",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]
   THEN REWRITE_TAC[trans_cont]
   THEN BETA_TAC
   THEN REWRITE_TAC[VS_DEF]
   THEN REWRITE_TAC[VE_LEMMA;VES_LEMMA]
   THEN REWRITE_TAC[VE_THM;s_subst_IDENT]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN (ASSUME_TAC o SYM) CONS
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;


%
g "!es s1 ns s2. ES es s1 ns s2 ==>
                     (ns = VS(VES es) s1)
                 /\  (s2 = s1 <|s VES_state es)";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]);;
e(REWRITE_TAC[VS_DEF;s_subst_IDENT]);;

e(REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]);;
e(REWRITE_TAC[trans_cont] THEN BETA_TAC);;
e(REWRITE_TAC[VS_DEF]);;
e(REWRITE_TAC[VE_LEMMA;VES_LEMMA;s_subst_IDENT]);;
e(REWRITE_TAC[VE_THM]);;
e(REPEAT STRIP_TAC);;

 e(IMP_RES_THEN (ASSUME_TAC o SYM) CONS);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%



% =========================================================== %
% Here is the second half of the fundamental theorem on the   %
% translation of lists of numeric-valued expressions.         %
% =========================================================== %

let VES_EXISTS =
 prove_thm
  (`VES_EXISTS`,
   "!es s1. ES es s1 (VS (VES es) s1) (s1 <|s VES_state es)",

   LIST_INDUCT_TAC
   THEN REPEAT GEN_TAC
   THEN REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]
   THEN REWRITE_TAC[trans_cont]
   THEN BETA_TAC
   THEN REWRITE_TAC[VS_DEF;NULL;HD;TL]
   THEN REWRITE_TAC[VE_LEMMA;VES_LEMMA;s_subst_IDENT]
   THEN REWRITE_TAC[VE_THM]
   THEN EXISTS_TAC "s1 <|s (VE_state h)"
   THEN ASM_REWRITE_TAC[]
  );;


%
Second half of VES translation theorem:

g "!es s1. ES es s1 (VS (VES es) s1) (s1 <|s VES_state es)";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]);;
e(REWRITE_TAC[VS_DEF;s_subst_IDENT]);;

e(REWRITE_TAC[ES_DEF;VES_DEF;VES_state_DEF;VES1_DEF]);;
e(REWRITE_TAC[trans_cont] THEN BETA_TAC);;
e(REWRITE_TAC[VS_DEF]);;
e(REWRITE_TAC[NULL;HD;TL]);;
e(REWRITE_TAC[VE_LEMMA;VES_LEMMA;s_subst_IDENT]);;
e(REWRITE_TAC[VE_THM]);;
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "s1 <|s (VE_state h)");;
e(ASM_REWRITE_TAC[]);;
%



% =========================================================== %
% Now we prove the main VES translation theorem!              %
%   This is essentially a compiler correctness proof.         %
% =========================================================== %

let VES_THM =
 prove_thm
  (`VES_THM`,
   "!es s1 ns s2. ES es s1 ns s2 =
                  (ns = VS(VES es) s1)    /\
                  (s2 = s1 <|s VES_state es)",

   REPEAT GEN_TAC
    THEN EQ_TAC
    THENL [ REWRITE_TAC[VES_UNIQUE];
            STRIP_TAC
            THEN ASM_REWRITE_TAC[VES_EXISTS]
          ]
  );;

%
g "!es s1 ns s2. ES es s1 ns s2 =
                  (ns = VS(VES es) s1)    /\
                  (s2 = s1 <|s VES_state es)";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;
e(REWRITE_TAC[VES_UNIQUE]);;

e(STRIP_TAC);;
e(ASM_REWRITE_TAC[VES_EXISTS]);;
%



let LENGTH_VES1 =
 prove_thm
  (`LENGTH_VES1`,
   "!es ss. LENGTH (FST(VES1 es ss)) = LENGTH es",

   LIST_INDUCT_TAC
    THEN REWRITE_TAC[VES1_DEF]
    THEN REWRITE_TAC[trans_cont]
    THEN BETA_TAC
    THEN ASM_REWRITE_TAC[LENGTH]
  );;

% LENGTH_VES1
g "!es ss. LENGTH (FST(VES1 es ss)) = LENGTH es";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[VES1_DEF]);;
e(REWRITE_TAC[LENGTH]);;

e(REWRITE_TAC[VES1_DEF]);;
e(REWRITE_TAC[trans_cont]);;
e(BETA_TAC);;
e(ASM_REWRITE_TAC[LENGTH]);;
%

let LENGTH_VES =
 prove_thm
  (`LENGTH_VES`,
   "!es. LENGTH (VES es) = LENGTH es",

   REWRITE_TAC[VES_DEF;LENGTH_VES1]
  );;

% LENGTH_VES
g "!es. LENGTH (VES es) = LENGTH es";;
e(REWRITE_TAC[VES_DEF]);;
e(REWRITE_TAC[LENGTH_VES1]);;
%


% ========================================================= %
% The functions AB translates boolean expressions from the  %
% programming language into the assertion language.         %
% ========================================================= %

let AB1_DEF = 
    new_rec_def `AB1_DEF` bexp_Axiom
      "(AB1 (EQ e1 e2) =
            VE1 e1 --> \v1. VE1 e2 --> \v2 ss. AEQ   v1 v2, ss)  /\
       (AB1 (LESS e1 e2) =
            VE1 e1 --> \v1. VE1 e2 --> \v2 ss. ALESS v1 v2, ss)  /\
       (AB1 (LLESS es1 es2) =
            VES1 es1 --> \vs1. VES1 es2 --> \vs2 ss. ALLESS vs1 vs2, ss)  /\
       (AB1 (AND b1 b2) =
            AB1 b1 --> \a1. AB1 b2 --> \a2 ss. AAND  a1 a2, ss)  /\
       (AB1 (OR b1 b2) =
            AB1 b1 --> \a1. AB1 b2 --> \a2 ss. AOR   a1 a2, ss)  /\
       (AB1 (NOT b) =
            AB1 b  --> \a ss. ANOT a, ss)";;


let AB_DEF = new_definition
 (`AB_DEF`,
  "AB b =  FST (AB1 b IDENT_SS)");;

let AB_state_DEF = new_definition
 (`AB_state_DEF`,
  "AB_state b =  SND (AB1 b IDENT_SS)");;


let AB_LEMMA,AB_THM =
             prove_translation_thm `AB_LEMMA` `AB_THM` `B` `A`
             AB1_DEF [AB_DEF;AB_state_DEF]
             "B b s1 t s2 =    (t = A (AB b) s1)    /\
                               (s2 = s1 <|s (AB_state b))"
             [
             % EQ case %
                         BEQ_TAC
                    THEN EXISTS_TAC "s1 <|s (VE_state e0)";
             % LESS case %
                         LESS_TAC
                    THEN EXISTS_TAC "s1 <|s (VE_state e0)";
             % LLESS case %
                         LLESS_TAC
                    THEN EXISTS_TAC "s1 <|s (VES_state l0)";
             % AND case %
                         AND_TAC
                    THEN EXISTS_TAC "s1 <|s (AB_state b)";
             % OR case %
                         OR_TAC
                    THEN EXISTS_TAC "s1 <|s (AB_state b)";
             % NOT case %
                         NOT_TAC
             ]
             [VE_LEMMA;VE_THM;VES_LEMMA;VES_THM];;


% ================================================================== %
% The free variables of a translated expression are the same as the  %
% free variables of the original expression.                         %
% ================================================================== %

let FVv_VE1 =
 prove_thm
  (`FVv_VE1`,
   "!e ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVv  (FST (VE1 e ss))    = FVe e) /\
        (!x. FVv ((SND (VE1 e ss)) x) = {x}))",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[VE1_DEF;trans_cont]
   THEN BETA_TAC
   THEN RES_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[FVv_DEF;FVe_DEF]
   THEN REWRITE_TAC[UNION]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(x:var) = v"
   THENL
    [ ASM_REWRITE_TAC[BND_MATCH;FVv_DEF]
      THEN REWRITE_TAC[UNION];

      IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
      THEN ASM_REWRITE_TAC[FVv_DEF]
    ]
  );;

% FVv_VE1
g "!e ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVv  (FST (VE1 e ss))    = FVe e) /\
        (!x. FVv ((SND (VE1 e ss)) x) = {x}))";;
e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC

  THEN REPEAT GEN_TAC THEN STRIP_TAC
  THEN REWRITE_TAC[VE1_DEF;trans_cont]
  THEN BETA_TAC
  THEN REWRITE_TAC[]);;
e(RES_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[FVv_DEF;FVe_DEF]);;
e(REWRITE_TAC[UNION]);;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(x:var) = v");;

 e(ASM_REWRITE_TAC[BND_MATCH;FVv_DEF]);;
 e(REWRITE_TAC[UNION]);;

 e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
 e(ASM_REWRITE_TAC[FVv_DEF]);;
%

let FVv_VE =
 prove_thm
  (`FVv_VE`,
   "(!e.   FVv  (VE e)          = FVe e) /\
    (!e x. FVv ((VE_state e) x) = {x}  )",
   REWRITE_TAC[VE_DEF;VE_state_DEF]
   THEN SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}"
            (REWRITE_THM o MATCH_MP FVv_VE1)
   THEN REWRITE_TAC[IDENT_SS;FVv_DEF]
  );;

% FVv_VE
g "(!e.   FVv  (VE e)          = FVe e) /\
   (!e x. FVv ((VE_state e) x) = {x}  )";;
e(REWRITE_TAC[VE_DEF;VE_state_DEF]);;
e(SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}" (REWRITE_THM o MATCH_MP FVv_VE1));;
e(REWRITE_TAC[IDENT_SS;FVv_DEF]);;
%

let FVvs_VES1 =
 prove_thm
  (`FVvs_VES1`,
   "!es ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVvs (FST (VES1 es ss))    = FVes es) /\
        (!x. FVv ((SND (VES1 es ss)) x) = {x}))",
   LIST_INDUCT_TAC
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[VES1_DEF;trans_cont]
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[FVvs_DEF;FVes_DEF]
   THEN IMP_RES_TAC FVv_VE1
   THEN POP_ASSUM (ASSUME_TAC o SPEC "h:exp")
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% FVvs_VES1
g "!es ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVvs (FST (VES1 es ss))    = FVes es) /\
        (!x. FVv ((SND (VES1 es ss)) x) = {x}))";;
e(LIST_INDUCT_TAC

  THEN REPEAT GEN_TAC THEN STRIP_TAC
  THEN REWRITE_TAC[VES1_DEF;trans_cont]
  THEN BETA_TAC
  THEN REWRITE_TAC[]);;
e(ASM_REWRITE_TAC[FVvs_DEF;FVes_DEF]);;
e(IMP_RES_TAC FVv_VE1);;
e(POP_ASSUM (ASSUME_TAC o SPEC "h:exp"));;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let FVvs_VES =
 prove_thm
  (`FVvs_VES`,
   "(!es.   FVvs (VES es)          = FVes es) /\
    (!es x. FVv ((VES_state es) x) = {x}  )",
   REWRITE_TAC[VES_DEF;VES_state_DEF]
   THEN SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}"
            (REWRITE_THM o MATCH_MP FVvs_VES1)
   THEN REWRITE_TAC[IDENT_SS;FVv_DEF]
  );;

% FVvs_VES
g "(!es.   FVvs (VES es)          = FVes es) /\
   (!es x. FVv ((VES_state es) x) = {x}  )";;
e(REWRITE_TAC[VES_DEF;VES_state_DEF]);;
e(SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}"
      (REWRITE_THM o MATCH_MP FVvs_VES1));;
e(REWRITE_TAC[IDENT_SS;FVv_DEF]);;
%

let FVa_AB1 =
 prove_thm
  (`FVa_AB1`,
   "!b ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVa  (FST (AB1 b ss))    = FVb b) /\
        (!x. FVv ((SND (AB1 b ss)) x) = {x}))",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN REWRITE_TAC[AB1_DEF;trans_cont]
   THEN BETA_TAC
   THEN REWRITE_TAC[FVa_DEF;FVb_DEF]
   THEN IMP_RES_TAC FVv_VE1
   THEN POP_ASSUM (ASSUME_TAC o SPEC "e0:exp")
   THEN IMP_RES_TAC FVv_VE1
   THEN IMP_RES_TAC FVvs_VES1
   THEN POP_ASSUM (ASSUME_TAC o SPEC "l0:exp list")
   THEN IMP_RES_TAC FVvs_VES1
   THEN RES_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% FVa_AB1
g "!b ss. 
       (!x. FVv (ss x) = {x}) ==>
       ((    FVa  (FST (AB1 b ss))    = FVb b) /\
        (!x. FVv ((SND (AB1 b ss)) x) = {x}))";;
e(INDUCT_THEN bexp_induct STRIP_ASSUME_TAC

  THEN REPEAT GEN_TAC THEN STRIP_TAC
  THEN REWRITE_TAC[AB1_DEF;trans_cont]
  THEN BETA_TAC
  THEN REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(IMP_RES_TAC FVv_VE1);;
e(POP_ASSUM (ASSUME_TAC o SPEC "e0:exp"));;
e(IMP_RES_TAC FVv_VE1);;
e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(IMP_RES_TAC FVv_VE1);;
e(POP_ASSUM (ASSUME_TAC o SPEC "e0:exp"));;
e(IMP_RES_TAC FVv_VE1);;
e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(IMP_RES_TAC FVvs_VES1);;
e(POP_ASSUM (ASSUME_TAC o SPEC "l0:exp list"));;
e(IMP_RES_TAC FVvs_VES1);;
e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(RES_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(RES_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[FVa_DEF;FVb_DEF]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let FVa_AB =
 prove_thm
  (`FVa_AB`,
   "(!b.   FVa  (AB b)          = FVb b) /\
    (!b x. FVv ((AB_state b) x) = {x}  )",
   REWRITE_TAC[AB_DEF;AB_state_DEF]
   THEN SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}"
            (REWRITE_THM o MATCH_MP FVa_AB1)
   THEN REWRITE_TAC[IDENT_SS;FVv_DEF]
  );;

% FVa_AB
g "(!b.   FVa  (AB b)          = FVb b) /\
   (!b x. FVv ((AB_state b) x) = {x}  )";;
e(REWRITE_TAC[AB_DEF;AB_state_DEF]);;
e(SUBGOAL_THEN "!x. FVv (IDENT_SS x) = {x}"
      (REWRITE_THM o MATCH_MP FVa_AB1));;
e(REWRITE_TAC[IDENT_SS;FVv_DEF]);;
%



% ================================================================== %
% Now that we have defined the translations of programming language  %
% expressions into the assertion language, we can define how to find %
% the precondition of executing a programming language expression,   %
% by substituting the effect of a programming language expression    %
% into an assertion language expression.  There are different        %
% operators to compute these preconditions, depending on the exact   %
% type of the programming language expression and the type of the    %
% assertion language expression subject to the substitution.         %
% ================================================================== %

% Preconditions of assertions before prog. lang. expressions.  %

let ve_pre = new_definition
  (`ve_pre`,
   "ve_pre e v = v <|v (VE_state e)");;

let ves_pre = new_definition
  (`ves_pre`,
   "ves_pre es v = v <|v (VES_state es)");;

let vb_pre = new_definition
  (`vb_pre`,
   "vb_pre b v = v <|v (AB_state b)");;

let vse_pre = new_definition
  (`vse_pre`,
   "vse_pre e vs = vs <|vs (VE_state e)");;

let vses_pre = new_definition
  (`vses_pre`,
   "vses_pre es vs = vs <|vs (VES_state es)");;

let vsb_pre = new_definition
  (`vsb_pre`,
   "vsb_pre b vs = vs <|vs (AB_state b)");;

let ae_pre = new_definition
  (`ae_pre`,
   "ae_pre e a = a <|a (VE_state e)");;

let aes_pre = new_definition
  (`aes_pre`,
   "aes_pre es a = a <|a (VES_state es)");;

let ab_pre = new_definition
  (`ab_pre`,
   "ab_pre b a = a <|a (AB_state b)");;


% =========================================================== %
% Substitutions of assertions before prog. lang. expressions. %
% =========================================================== %

let assign =
  new_infix_def
  `assign`
  "$:= x e = BND (VE e) x (VE_state e)";;

let assign_THM =
 prove_thm
  (`assign_THM`,
   "!e s1 n s2 x. E e s1 n s2 ==> (s1 <|s (x := e) = BND n x s2)",
   REWRITE_TAC[VE_THM]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[assign;s_subst_BND]
  );;

%
g "!e s1 n s2 x. E e s1 n s2 ==> (s1 <|s (x := e) = BND n x s2)";;
e(REWRITE_TAC[VE_THM]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[assign;s_subst_BND]);;
%

% ============================================================ %
% With the above ":=" operator, we can express the correctness %
% of substitutions of programming lang. exprs in terms of      %
% the corresponding assertion lang. substitutions, as:         %
%     !e x a s. A(a <|a (x := e))s = A a(s <|s (x := e))       %
% which is a direct instance of the a_substitution lemma.      %
% ============================================================ %




% =========================================================== %
% Substitutions of assertions before lists of programming     %
% language expressions.                                       %
% =========================================================== %

let assign_list =
  new_infix_def
  `assign_list`
  "$:== xs es = BNDsL (VES es) xs (VES_state es)";;

let assign_list_THM =
 prove_thm
  (`assign_list_THM`,
   "!es s1 ns s2 xs. ES es s1 ns s2 ==> (s1 <|s (xs :== es) = BNDsL ns xs s2)",
   REWRITE_TAC[VES_THM]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[assign_list;s_subst_BNDsL]
  );;

%
g "!es s1 ns s2 xs. ES es s1 ns s2 ==> (s1 <|s (xs :== es) = BNDsL ns xs s2)";;
e(REWRITE_TAC[VES_THM]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[assign_list;s_subst_BNDsL]);;
%

let MAP_assign_list_subst_MATCH =
 prove_thm
  (`MAP_assign_list_subst_MATCH`,
   "!es xs s.
      (LENGTH xs = LENGTH es) /\ DL xs ==>
      (MAP (s <|s (xs :== es)) xs = VS(VES es)s)",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[assign_list;s_subst_BNDsL]
   THEN MATCH_MP_TAC (IMP2AND MAP_BNDsL_IDENT)
   THEN ASM_REWRITE_TAC[LENGTH_VS;LENGTH_VES]
  );;

% MAP_assign_list_subst_MATCH
g "!es xs s.
      (LENGTH xs = LENGTH es) /\ DL xs ==>
      (MAP (s <|s (xs :== es)) xs = VS(VES es)s)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[assign_list;s_subst_BNDsL]);;
e(MATCH_MP_TAC (IMP2AND MAP_BNDsL_IDENT));;
e(ASM_REWRITE_TAC[LENGTH_VS;LENGTH_VES]);;
%

let MAP_assign_list_subst_NOT_MATCH =
 prove_thm
  (`MAP_assign_list_subst_NOT_MATCH`,
   "!es xs zs s.
      DISJOINT(SL zs)(SL xs) ==>
      (MAP (s <|s (xs :== es)) zs = MAP (s <|s (VES_state es)) zs)",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[assign_list;s_subst_BNDsL]
   THEN MATCH_MP_TAC MAP_BNDsL_NOT_MATCH
   THEN ASM_REWRITE_TAC[]
  );;

% MAP_assign_list_subst_NOT_MATCH
g "!es xs zs s.
      DISJOINT(SL zs)(SL xs) ==>
      (MAP (s <|s (xs :== es)) zs = MAP (s <|s (VES_state es)) zs)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[assign_list;s_subst_BNDsL]);;
e(MATCH_MP_TAC MAP_BNDsL_NOT_MATCH);;
e(ASM_REWRITE_TAC[]);;
%


close_theory ();;

