% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains definitions of functions to substitute variables for
  variables within assertion-language expressions.
  This is both for one particular variable for another,
  and for a list of such variable-for-variable substitutions.
  First an operator, //v, is defined to create such substitutions from
  two lists of variables.  Then substitution operators are defined to
  apply such substitutions to assertion-language expressions.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `var_substitutions`;;

new_parent `substitutions`;;



% Set up autoloading from ancestor theories of var_substitutions: %

loadf `autoload`;;

autoload_all_defs_and_thms
 `bindings variants finite_sets more_finite_sets
  assert_syntax assert_semantics substitutions`;;

loadf `semantics_tactics`;;




% ============================================================== %
% ============================================================== %
%  For making substitutions on actual code, we need to restrict  %
%  the substitutions to simply variables for variables.          %
% ============================================================== %
% ============================================================== %

new_type_abbrev(`vsubst`, ":var -> var");;


let vsubst_DEF = 
  new_infix_list_rec_definition
  (`vsubst_DEF`,
   "($//v [] xs = (I:vsubst)) /\
    ($//v (CONS y ys) xs =
        ((xs = []) => I |
          (BND y (HD xs)
            (BND (($//v ys (TL xs)) (HD xs)) (@z. ($//v ys (TL xs)) z = y)
              ($//v ys (TL xs))))))"
  );;



let vsubst_THM =
 prove_thm
  (`vsubst_THM`,
   "(!xs. [] //v xs = I) /\
    (!ys. ys //v [] = I) /\
    (!ys xs y x. (CONS y ys) //v (CONS x xs) =
     (BND y x (BND ((ys //v xs) x) (@z. (ys //v xs) z = y) (ys //v xs))))",
   REWRITE_TAC[vsubst_DEF;HD;TL;NOT_CONS_NIL]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_DEF]
  );;

% vsubst_THM:
g  "(!xs. [] //v xs = I) /\
    (!ys. ys //v [] = I) /\
    (!ys xs y x. (CONS y ys) //v (CONS x xs) =
     (BND y x (BND ((ys //v xs) x) (@z. (ys //v xs) z = y) (ys //v xs))))";;

e(REWRITE_TAC[vsubst_DEF;HD;TL;NOT_CONS_NIL]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vsubst_DEF]);;

 e(REWRITE_TAC[vsubst_DEF]);;
%




let IDENTITY_vsubst =
 prove_thm
  (`IDENTITY_vsubst`,
   "!xs. xs //v xs = I",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM]
   THEN ASM_REWRITE_TAC[I_THM;SELECT_REFL;BND_I]
  );;

% IDENTITY_vsubst:
g  "!xs. xs //v xs = I";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vsubst_THM]);;

 e(REWRITE_TAC[vsubst_THM]);;
 e(ASM_REWRITE_TAC[I_THM;SELECT_REFL;BND_I]);;
%


let APPEND_IDENTITY_vsubst =
 prove_thm
  (`APPEND_IDENTITY_vsubst`,
   "!ys xs zs.
        (LENGTH xs = LENGTH ys) ==>
        ((APPEND ys zs //v APPEND xs zs) = (ys //v xs))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;vsubst_THM;IDENTITY_vsubst]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% APPEND_IDENTITY_vsubst:
g  "!ys xs zs.
        (LENGTH xs = LENGTH ys) ==>
        ((APPEND ys zs //v APPEND xs zs) = (ys //v xs))";;

e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;vsubst_THM;IDENTITY_vsubst]);;

 e(REWRITE_TAC[APPEND;vsubst_THM]);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%



let vsubst_ONTO =
 prove_thm
  (`vsubst_ONTO`,
   "!ys xs. ONTO (ys //v xs)",
   LIST_INDUCT_TAC
   THEN GEN_TAC
   THEN REWRITE_TAC[ONTO_DEF]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THENL
     [  GEN_TAC
        THEN REWRITE_TAC[vsubst_THM;I_THM]
        THEN EXISTS_TAC "y:var"
        THEN REFL_TAC;

        LIST_INDUCT_TAC
        THENL
          [  GEN_TAC
             THEN REWRITE_TAC[vsubst_THM;I_THM]
             THEN EXISTS_TAC "y:var"
             THEN REFL_TAC;

             REPEAT GEN_TAC
             THEN REWRITE_TAC[vsubst_THM]
             THEN POP_TAC
             THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
             THEN IMP_RES_TAC ONTO_INVERSE
             THEN ASM_CASES_TAC "(y:var) = h"
             THENL
               [  EXISTS_TAC "h':var"
                  THEN ASM_REWRITE_TAC[BND_MATCH];

                  ASM_CASES_TAC "(y:var) = ((ys //v xs) h')"
                  THENL
                    [  EXISTS_TAC "@z. (ys //v xs)z = h"
                       THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                       THENL
                         [  DISCH_THEN (REWRITE_ALL_THM o SYM)
                            THEN UNDISCH_LAST_TAC
                            THEN ASM_REWRITE_TAC[];

                            ASM_REWRITE_TAC[BND_MATCH]
                         ];

                       EXISTS_TAC "@z. (ys //v xs)z = y"
                       THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                       THENL
                         [  DISCH_THEN (REWRITE_ALL_THM o SYM)
                            THEN UNDISCH_LAST_TAC
                            THEN ASM_REWRITE_TAC[];

                            USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                            THENL
                              [  DISCH_THEN (\th. ASSUME_TAC (AP_TERM "ys //v xs" th))
                                 THEN UNDISCH_LAST_TAC
                                 THEN ASM_REWRITE_TAC[];

                                 ASM_REWRITE_TAC[]
                              ]
                         ]
                    ]
               ]
          ]
     ]
  );;

% vsubst_ONTO:

g "!ys xs. ONTO (ys //v xs)";;
e(LIST_INDUCT_TAC);;

  e(GEN_TAC);;
  e(REWRITE_TAC[ONTO_DEF;vsubst_THM]);;
  e(GEN_TAC);;
  e(REWRITE_TAC[I_THM]);;
  e(EXISTS_TAC "y:var");;
  e(REFL_TAC);;

  e(GEN_TAC);;
  e(REWRITE_TAC[ONTO_DEF]);;
  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(LIST_INDUCT_TAC);;

   e(GEN_TAC);;
   e(REWRITE_TAC[vsubst_THM;I_THM]);;
   e(EXISTS_TAC "y:var");;
   e(REFL_TAC);;

   e(REPEAT GEN_TAC);;
   e(REWRITE_TAC[vsubst_THM]);;
   e(POP_TAC);;
   e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
   e(IMP_RES_TAC ONTO_INVERSE);;
   e(ASM_CASES_TAC "(y:var) = h");;

    e(EXISTS_TAC "h':var");;
    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(ASM_CASES_TAC "(y:var) = ((ys //v xs) h')");;

     e(EXISTS_TAC "@z. (ys //v xs)z = h");;
     e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

      e(DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th]));;
      e(UNDISCH_LAST_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(ASM_REWRITE_TAC[BND_MATCH]);;

     e(EXISTS_TAC "@z. (ys //v xs)z = y");;
     e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

      e(DISCH_THEN (REWRITE_ALL_THM o SYM));;
      e(UNDISCH_LAST_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

       e(DISCH_THEN (\th. ASSUME_TAC (AP_TERM "ys //v xs" th)));;
       e(UNDISCH_LAST_TAC);;
       e(ASM_REWRITE_TAC[]);;

       e(ASM_REWRITE_TAC[]);;
%


% We need some help to deal with the select operator @.  %

let vsubst_INVERSE_SELECT =
 prove_thm
  (`vsubst_INVERSE_SELECT`,
   "!ys xs y. (ys //v xs)(@z. (ys //v xs)z = y) = y",
   LIST_INDUCT_TAC
   THEN GEN_TAC
   THENL
     [  REWRITE_TAC[vsubst_THM]
        THEN GEN_TAC
        THEN CONV_TAC SELECT_CONV
        THEN REWRITE_TAC[I_THM]
        THEN EXISTS_TAC "y:var"
        THEN REFL_TAC;

        LIST_INDUCT_TAC
        THEN REPEAT GEN_TAC
        THEN REWRITE_TAC[vsubst_THM]
        THEN CONV_TAC SELECT_CONV
        THENL
          [  REWRITE_TAC[I_THM]
             THEN EXISTS_TAC "y:var"
             THEN REFL_TAC;

        ASM_CASES_TAC "(y:var) = h"
        THENL
          [  EXISTS_TAC "h':var"
             THEN ASM_REWRITE_TAC[BND_MATCH];

             ASM_CASES_TAC "(y:var) = ((ys //v xs) h')"
             THENL
               [  EXISTS_TAC "@z. (ys //v xs)z = h"
                  THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                  THENL
                    [  DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th])
                       THEN POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th))
                       THEN ASM_REWRITE_TAC[];

                       ASM_REWRITE_TAC[BND_MATCH]
                    ];

                  EXISTS_TAC "@z. (ys //v xs)z = y"
                  THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                  THENL
                    [  DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th])
                       THEN POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th))
                       THEN ASM_REWRITE_TAC[];

                       USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
                       THENL
                         [  DISCH_THEN (\th. ASSUME_TAC (AP_TERM "ys //v xs" th))
                            THEN POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th))
                            THEN ASM_REWRITE_TAC[];

                            ASM_REWRITE_TAC[]
                         ]
                    ]
               ]
          ]
          ]
     ]
  );;

% vsubst_INVERSE_SELECT:

g "!ys xs y. (ys //v xs)(@z. (ys //v xs)z = y) = y";;
e(LIST_INDUCT_TAC);;

  e(GEN_TAC);;
  e(REWRITE_TAC[vsubst_THM]);;
  e(GEN_TAC);;
  e(CONV_TAC SELECT_CONV);;
  e(REWRITE_TAC[I_THM]);;
  e(EXISTS_TAC "y:var");;
  e(REFL_TAC);;

  e(GEN_TAC);;
  e(LIST_INDUCT_TAC);;

   e(REPEAT GEN_TAC);;
   e(REWRITE_TAC[vsubst_THM]);;
   e(REWRITE_TAC[I_THM]);;
   e(CONV_TAC SELECT_CONV);;
   e(EXISTS_TAC "y:var");;
   e(REFL_TAC);;

   e(REPEAT GEN_TAC);;
   e(REWRITE_TAC[vsubst_THM]);;
   e(CONV_TAC SELECT_CONV);;
   e(ASM_CASES_TAC "(y:var) = h");;

    e(EXISTS_TAC "h':var");;
    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(ASM_CASES_TAC "(y:var) = ((ys //v xs) h')");;

     e(EXISTS_TAC "@z. (ys //v xs)z = h");;
     e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

      e(DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th]));;
      e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
      e(ASM_REWRITE_TAC[]);;

      e(ASM_REWRITE_TAC[BND_MATCH]);;

     e(EXISTS_TAC "@z. (ys //v xs)z = y");;
     e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

      e(DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th]));;
      e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
      e(ASM_REWRITE_TAC[]);;

      e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

       e(DISCH_THEN (\th. ASSUME_TAC (AP_TERM "ys //v xs" th)));;
       e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
       e(ASM_REWRITE_TAC[]);;

       e(ASM_REWRITE_TAC[]);;
%





let vsubst_ONE_ONE =
 prove_thm
  (`vsubst_ONE_ONE`,
   "!ys xs. ONE_ONE (ys //v xs)",
   REWRITE_TAC[ONE_ONE]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;I_THM]
   THEN GEN_TAC
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;I_THM]
   THEN POP_TAC
   THEN POP_ASSUM (\th. ASSUME_TAC th THEN
        (ASSUME_TAC o GENL["x1:var";"x2:var"] o
           REWRITE_RULE[vsubst_INVERSE_SELECT] o
           SPECL["xs:(var)list";"@z. (ys //v xs)z = x1";"x2:var"]) th)
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(x1:var) = h'"
   THENL
    [ POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[BND_MATCH]
      THEN ASM_CASES_TAC "(x2:var) = h'"
      THENL
       [ POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[BND_MATCH];

         IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         THEN ASM_CASES_TAC "x2 = (@z. (ys //v xs)z = h)"
         THENL
          [ POP_ASSUM (\th. ASM_REWRITE_TAC[SYM th;BND_MATCH])
            THEN FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM);

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
            THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
          ]
       ];

      IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
      THEN ASM_CASES_TAC "x1 = (@z. (ys //v xs)z = h)"
      THENL
       [ POP_ASSUM (ASSUME_TAC o SYM)
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN ASM_CASES_TAC "(x2:var) = h'"
         THENL
          [ ASM_REWRITE_TAC[BND_MATCH]
            THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_CASES_TAC "(x2:var) = x1"
            THENL
             [ POP_ASSUM (\th. REWRITE_TAC[th;BND_MATCH]);

               IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
               THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
             ]
          ];

         IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         THEN POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM)
         THEN ASM_CASES_TAC "(x2:var) = h'"
         THENL
          [ ASM_REWRITE_TAC[BND_MATCH]
            THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_CASES_TAC "x2 = (@z. (ys //v xs)z = h)"
            THENL
             [ POP_ASSUM (\th. REWRITE_ALL_TAC[SYM th;BND_MATCH])
               THEN ASM_REWRITE_TAC[]
               THEN FIRST_ASSUM (ACCEPT_TAC o NOT_EQ_SYM);

               IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
               THEN POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM)
               THEN ASM_REWRITE_TAC[]
             ]
          ]
       ]
    ]
  );;

%< vsubst_ONE_ONE:

g "!ys xs u v. ((ys //v xs) u = (ys //v xs) v) = (u = v)";;
g "!ys xs. ONE_ONE (ys //v xs)";;
e(REWRITE_TAC[ONE_ONE]);;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[vsubst_THM;I_THM]);;

  e(GEN_TAC THEN LIST_INDUCT_TAC
    THEN REWRITE_TAC[vsubst_THM;I_THM]);;
  e(POP_TAC);;
  e(POP_ASSUM (\th. ASSUME_TAC th THEN
        (ASSUME_TAC o GENL["x1:var";"x2:var"] o
           REWRITE_RULE[vsubst_INVERSE_SELECT] o
           SPECL["xs:(var)list";"@z. (ys //v xs)z = x1";"x2:var"]) th));;
  e(REPEAT GEN_TAC);;
  e(ASM_CASES_TAC "(x1:var) = h'");;
% a: %
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[BND_MATCH]);;
   e(ASM_CASES_TAC "(x2:var) = h'");;
% aa: %
    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[BND_MATCH]);;
% a~a: %
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(ASM_CASES_TAC "x2 = (@z. (ys //v xs)z = h)");;
% ab: %
     e(POP_ASSUM (\th. ASM_REWRITE_TAC[SYM th;BND_MATCH]));;
     e(FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
% ac: %
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_REWRITE_TAC[]);;
     e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
     e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
% ~a: %
   e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
   e(ASM_CASES_TAC "x1 = (@z. (ys //v xs)z = h)");;
% b: %
    e(POP_ASSUM (ASSUME_TAC o SYM));;
    e(ASM_REWRITE_TAC[BND_MATCH]);;
    e(ASM_CASES_TAC "(x2:var) = h'");;
% ba: %
     e(ASM_REWRITE_TAC[BND_MATCH]);;
     e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
     e(ASM_REWRITE_TAC[]);;
% b~a: %
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_CASES_TAC "(x2:var) = x1");;
% bb: %
      e(POP_ASSUM (\th. REWRITE_TAC[th;BND_MATCH]));;
% bc: %
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(ASM_REWRITE_TAC[]);;
      e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
      e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
% c: %
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM));;
    e(ASM_CASES_TAC "(x2:var) = h'");;
% ca: %
     e(ASM_REWRITE_TAC[BND_MATCH]);;
     e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
     e(ASM_REWRITE_TAC[]);;
% c~a: %
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_CASES_TAC "x2 = (@z. (ys //v xs)z = h)");;
% cb: %
      e(POP_ASSUM (\th. REWRITE_ALL_TAC[SYM th;BND_MATCH]));;
      e(ASM_REWRITE_TAC[]);;
      e(FIRST_ASSUM (ACCEPT_TAC o NOT_EQ_SYM));;
% cc: %
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM));;
      e(ASM_REWRITE_TAC[]);;
>%


let vsubst_INVERSE_ONTO =
 prove_thm
  (`vsubst_INVERSE_ONTO`,
   "!ys xs. ONTO (\y. @z. (ys //v xs)z = y)",
   REWRITE_TAC[ONTO_DEF]
   THEN BETA_TAC
   THEN REPEAT GEN_TAC
   THEN EXISTS_TAC "(ys //v xs)y"
   THEN ASSUME_TAC (SPEC_ALL vsubst_ONE_ONE)
   THEN REWRITE_ALL_TAC[ONE_ONE]
   THEN ASM_REWRITE_TAC[SELECT_REFL]
  );;

% vsubst_INVERSE_ONTO
g "!ys xs. ONTO (\y. @z. (ys //v xs)z = y)";;
e(REWRITE_TAC[ONTO_DEF]);;
e(BETA_TAC);;
e(REPEAT GEN_TAC);;
e(EXISTS_TAC "(ys //v xs)y");;
e(ASSUME_TAC (SPEC_ALL vsubst_ONE_ONE));;
e(REWRITE_ALL_TAC[ONE_ONE]);;
e(ASM_REWRITE_TAC[SELECT_REFL]);;
%


let vsubst_INVERSE_ONE_ONE =
 prove_thm
  (`vsubst_INVERSE_ONE_ONE`,
   "!ys xs. ONE_ONE (\y. @z. (ys //v xs)z = y)",
   REWRITE_TAC[ONE_ONE_DEF]
   THEN BETA_TAC
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (ASSUME_TAC o AP_TERM "ys //v xs")
   THEN UNDISCH_ALL_TAC
   THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
  );;

% vsubst_INVERSE_ONE_ONE
g "!ys xs. ONE_ONE (\y. @z. (ys //v xs)z = y)";;
e(REWRITE_TAC[ONE_ONE_DEF]);;
e(BETA_TAC);;
e(REPEAT GEN_TAC);;
e(DISCH_THEN (ASSUME_TAC o AP_TERM "ys //v xs"));;
e(UNDISCH_ALL_TAC);;
e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%


let vsubst_11 =
 prove_thm
  (`vsubst_11`,
   "!ys xs u v. ((ys //v xs) u = (ys //v xs) v) = (u = v)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;I_THM]
   THEN GEN_TAC
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;I_THM]
   THEN POP_TAC
   THEN POP_ASSUM (\th. ASSUME_TAC th THEN
        (ASSUME_TAC o GENL["u:var";"v:var"] o
           REWRITE_RULE[vsubst_INVERSE_SELECT] o
           SPECL["xs:(var)list";"@z. (ys //v xs)z = u";"v:var"]) th)
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(u:var) = h'"
   THENL
    [ POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[BND_MATCH]
      THEN ASM_CASES_TAC "(v:var) = h'"
      THENL
       [ POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[BND_MATCH];

         IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         THEN ASM_CASES_TAC "v = (@z. (ys //v xs)z = h)"
         THENL
          [ POP_ASSUM (\th. ASM_REWRITE_TAC[SYM th;BND_MATCH])
            THEN FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM);

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
            THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
          ]
       ];

      IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
      THEN ASM_CASES_TAC "u = (@z. (ys //v xs)z = h)"
      THENL
       [ POP_ASSUM (ASSUME_TAC o SYM)
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN ASM_CASES_TAC "(v:var) = h'"
         THENL
          [ ASM_REWRITE_TAC[BND_MATCH]
            THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_CASES_TAC "(v:var) = u"
            THENL
             [ POP_ASSUM (\th. REWRITE_TAC[th;BND_MATCH]);

               IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
               THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
             ]
          ];

         IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
         THEN POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM)
         THEN ASM_CASES_TAC "(v:var) = h'"
         THENL
          [ ASM_REWRITE_TAC[BND_MATCH]
            THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
            THEN ASM_CASES_TAC "v = (@z. (ys //v xs)z = h)"
            THENL
             [ POP_ASSUM (\th. REWRITE_ALL_TAC[SYM th;BND_MATCH])
               THEN ASM_REWRITE_TAC[]
               THEN FIRST_ASSUM (ACCEPT_TAC o NOT_EQ_SYM);

               IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
               THEN POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM)
               THEN ASM_REWRITE_TAC[]
             ]
          ]
       ]
    ]
  );;


% vsubst_11:
g "!ys xs u v. ((ys //v xs) u = (ys //v xs) v) = (u = v)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[vsubst_THM;I_THM]);;

  e(GEN_TAC THEN LIST_INDUCT_TAC
    THEN REWRITE_TAC[vsubst_THM;I_THM]);;
  e(POP_TAC);;
  e(POP_ASSUM (\th. ASSUME_TAC th THEN
        (ASSUME_TAC o GENL["u:var";"v:var"] o
           REWRITE_RULE[vsubst_INVERSE_SELECT] o
           SPECL["xs:(var)list";"@z. (ys //v xs)z = u";"v:var"]) th));;
  e(REPEAT GEN_TAC);;
  e(ASM_CASES_TAC "(u:var) = h'");;
a:
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[BND_MATCH]);;
   e(ASM_CASES_TAC "(v:var) = h'");;
aa:
    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[BND_MATCH]);;
a~a
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(ASM_CASES_TAC "v = (@z. (ys //v xs)z = h)");;
ab:
     e(POP_ASSUM (\th. ASM_REWRITE_TAC[SYM th;BND_MATCH]));;
     e(FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
ac:
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_REWRITE_TAC[]);;
     e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
     e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
~a:
   e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
   e(ASM_CASES_TAC "u = (@z. (ys //v xs)z = h)");;
b:
    e(POP_ASSUM (ASSUME_TAC o SYM));;
    e(ASM_REWRITE_TAC[BND_MATCH]);;
    e(ASM_CASES_TAC "(v:var) = h'");;
ba:
     e(ASM_REWRITE_TAC[BND_MATCH]);;
     e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
     e(ASM_REWRITE_TAC[]);;
b~a:
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_CASES_TAC "(v:var) = u");;
bb:
      e(POP_ASSUM (\th. REWRITE_TAC[th;BND_MATCH]));;
bc:
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(ASM_REWRITE_TAC[]);;
      e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
      e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
c:
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM));;
    e(ASM_CASES_TAC "(v:var) = h'");;
ca:
     e(ASM_REWRITE_TAC[BND_MATCH]);;
     e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
     e(ASM_REWRITE_TAC[]);;
c~a:
     e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
     e(ASM_CASES_TAC "v = (@z. (ys //v xs)z = h)");;
cb:
      e(POP_ASSUM (\th. REWRITE_ALL_TAC[SYM th;BND_MATCH]));;
      e(ASM_REWRITE_TAC[]);;
      e(FIRST_ASSUM (ACCEPT_TAC o NOT_EQ_SYM));;
cc:
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(POP_ASSUM (ASSUME_TAC o NOT_EQ_SYM));;
      e(ASM_REWRITE_TAC[]);;
%




let vsubst_INVERSE =
 prove_thm
  (`vsubst_INVERSE`,
   "!ys xs. ((ys //v xs) o (\y. @z. (ys //v xs)z = y) = I) /\
            ((\y. @z. (ys //v xs)z = y) o (ys //v xs) = I) ",
   REPEAT STRIP_TAC
   THEN EXT_TAC "x:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[o_THM;I_THM]
   THEN BETA_TAC
   THENL
     [  REWRITE_TAC[vsubst_INVERSE_SELECT];
        REWRITE_TAC[vsubst_11;SELECT_REFL]
     ]
  );;

% vsubst_INVERSE:
g "!ys xs. ((ys //v xs) o (\y. @z. (ys //v xs)z = y) = I) /\
           ((\y. @z. (ys //v xs)z = y) o (ys //v xs) = I) ";;
e(REPEAT STRIP_TAC);;

 e(EXT_TAC "x:var");;
 e(GEN_TAC);;
 e(REWRITE_TAC[o_THM;I_THM]);;
 e(BETA_TAC);;
 e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;

 e(EXT_TAC "x:var");;
 e(GEN_TAC);;
 e(REWRITE_TAC[o_THM;I_THM]);;
 e(BETA_TAC);;
 e(REWRITE_TAC[vsubst_11;SELECT_REFL]);;
%



let vsubst_EQ_INVERSE =
 prove_thm
  (`vsubst_EQ_INVERSE`,
   "!ys xs y x. ((ys //v xs)x = y) = ((@z. (ys //v xs)z = y) = x)",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN DISCH_THEN (REWRITE_THM o SYM)
   THENL
     [  REWRITE_TAC[vsubst_11;SELECT_REFL];
        REWRITE_TAC[vsubst_INVERSE_SELECT]
     ]
  );;

% vsubst_EQ_INVERSE:
g "!ys xs y x. ((ys //v xs)x = y) = ((@z. (ys //v xs)z = y) = x)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[vsubst_11;SELECT_REFL]);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%

let vsubst_EQ_INVERSE1 =
 prove_thm
  (`vsubst_EQ_INVERSE1`,
   "!ys xs y x. ((ys //v xs)x = y) = (x = (@z. (ys //v xs)z = y))",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [  DISCH_THEN (REWRITE_THM o SYM)
        THEN REWRITE_TAC[vsubst_11;SELECT_REFL];

        DISCH_THEN REWRITE_THM
        THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
     ]
  );;

% vsubst_EQ_INVERSE1:
g "!ys xs y x. ((ys //v xs)x = y) = (x = (@z. (ys //v xs)z = y))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[vsubst_11;SELECT_REFL]);;

 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%


let MAP_vsubst_IDENTITY_INVERSE =
 prove_thm
  (`MAP_vsubst_IDENTITY_INVERSE`,
   "!gs ys xs.
      (MAP (ys //v xs) gs = gs) =
      (MAP (\y. @z. (ys //v xs)z = y) gs = gs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN ASM_REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THENL
     [  (ONCE_REWRITE_THM o SYM o ASSUME) "(ys //v xs)h = h"
        THEN REWRITE_TAC[(SYM o SPEC_ALL o
                 ISPEC "\y. @z. (ys //v xs)z = y") o_THM];

        (ONCE_REWRITE_THM o SYM o ASSUME) "(\y. @z. (ys //v xs)z = y)h = h"
        THEN REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ys //v xs") o_THM]
     ]
   THEN ASM_REWRITE_TAC[vsubst_INVERSE;I_THM]
  );;

% MAP_vsubst_IDENTITY_INVERSE:
g "!gs ys xs.
      (MAP (ys //v xs) gs = gs) =
      (MAP (\y. @z. (ys //v xs)z = y) gs = gs)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[MAP]);;
 e(ASM_REWRITE_TAC[CONS_11]);;
 e(REPEAT STRIP_TAC);;
 e(EQ_TAC THEN STRIP_TAC);;

   e(ASM_REWRITE_TAC[]);;
   e((ONCE_REWRITE_THM o SYM o ASSUME) "(ys //v xs)h = h");;
   e(REWRITE_TAC[(SYM o SPEC_ALL o
                 ISPEC "(\y. @z. (ys //v xs)z = y)") o_THM]);;
   e(ASM_REWRITE_TAC[vsubst_INVERSE;I_THM]);;

   e(ASM_REWRITE_TAC[]);;
   e((ONCE_REWRITE_THM o SYM o ASSUME) "(\y. @z. (ys //v xs)z = y)h = h");;
   e(REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ys //v xs") o_THM]);;
   e(ASM_REWRITE_TAC[vsubst_INVERSE;I_THM]);;
%



let IDENT_vsubst =
 prove_thm
  (`IDENT_vsubst`,
   "!ys xs x. ~x IN (SL ys) /\ ~x IN (SL xs) ==> ((ys //v xs)x = x)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;I_THM]
   THEN REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN POP_TAC
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
   THEN REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
   THEN ASM_REWRITE_TAC[]
  );;

% IDENT_vsubst
g "!ys xs x. ~x IN (SL ys) /\ ~x IN (SL xs) ==> ((ys //v xs)x = x)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vsubst_THM;I_THM]);;

 e(REWRITE_TAC[vsubst_THM;I_THM]);;

 e(REWRITE_TAC[vsubst_THM;I_THM]);;
 e(REWRITE_TAC[SET_LIST;IN;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(POP_TAC);;
 e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
 e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

  e(REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
%

let IDENT_vsubst_INVERSE =
 prove_thm
  (`IDENT_vsubst_INVERSE`,
   "!ys xs x.
      ~x IN (SL ys) /\ ~x IN (SL xs) ==>
      ((\y. @z. (ys //v xs)z = y)x = x)",
   REPEAT STRIP_TAC
   THEN BETA_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]
   THEN IMP_RES_TAC IDENT_vsubst
  );;

% IDENT_vsubst_INVERSE
g "!ys xs x.
     ~x IN (SL ys) /\ ~x IN (SL xs) ==>
     ((\y. @z. (ys //v xs)z = y)x = x)";;
e(REPEAT STRIP_TAC);;
e(BETA_TAC);;
e(REWRITE_TAC[SYM(SPEC_ALL vsubst_EQ_INVERSE)]);;
e(IMP_RES_TAC IDENT_vsubst);;
%



let IDENT_MAP_vsubst =
 prove_thm
  (`IDENT_MAP_vsubst`,
   "!zs ys xs.
      DISJOINT(SL ys)(SL zs) /\
      DISJOINT(SL xs)(SL zs) ==>
         (MAP (ys //v xs) zs = zs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN REWRITE_TAC[CONS_11;SET_LIST;DISJOINT_INSERT2]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC IDENT_vsubst
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% IDENT_MAP_vsubst
g "!zs ys xs.
      DISJOINT(SL ys)(SL zs) /\
      DISJOINT(SL xs)(SL zs) ==>
         (MAP (ys //v xs) zs = zs)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[MAP]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[DISJOINT_INSERT2]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC IDENT_vsubst);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let IDENT_MAP_vsubst_INVERSE =
 prove_thm
  (`IDENT_MAP_vsubst_INVERSE`,
   "!zs ys xs.
      DISJOINT(SL ys)(SL zs) /\
      DISJOINT(SL xs)(SL zs) ==>
         (MAP (\y. @z. (ys //v xs)z = y) zs = zs)",
   REWRITE_TAC[SYM(SPEC_ALL MAP_vsubst_IDENTITY_INVERSE);IDENT_MAP_vsubst]
  );;

% IDENT_MAP_vsubst_INVERSE
g "!zs ys xs.
      DISJOINT(SL ys)(SL zs) /\
      DISJOINT(SL xs)(SL zs) ==>
         (MAP (\y. @z. (ys //v xs)z = y) zs = zs)";;
e(REWRITE_TAC[SYM(SPEC_ALL MAP_vsubst_IDENTITY_INVERSE);IDENT_MAP_vsubst]);;
%


let vsubst_ELEMENTS =
 prove_thm
  (`vsubst_ELEMENTS`,
   "!ys xs. (LENGTH xs = LENGTH ys) ==> (DL xs /\ DL ys) ==>
            (!x. ((ys //v xs)x IN SL ys) = (x IN SL xs))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]
   THEN REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[vsubst_THM;SET_LIST;IN]
   THEN POP_TAC
   THEN REWRITE_TAC[DISJOINT_LIST]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(x:var) = h'"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN ASM_REWRITE_TAC[]
        THEN ASM_CASES_TAC "x = (@z. (ys //v l')z = h)"
        THENL
          [  POP_ASSUM (ASSUME_TAC o SYM)
             THEN ASM_REWRITE_TAC[BND_MATCH]
             THEN ONCE_REWRITE_TAC[vsubst_EQ_INVERSE]
             THEN RES_TAC
             THEN ASM_REWRITE_TAC[]
             THEN POP_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL)
             THEN POP_TAC THEN POP_TAC
             THEN POP_ASSUM (REWRITE_THM o SYM)
             THEN ASM_REWRITE_TAC[vsubst_INVERSE_SELECT];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN ONCE_REWRITE_TAC[vsubst_EQ_INVERSE]
             THEN POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
             THEN RES_TAC
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% vsubst_ELEMENTS:
g "!ys xs. (LENGTH xs = LENGTH ys) ==> (DL xs /\ DL ys) ==>
           (!x. ((ys //v xs)x IN SL ys) = (x IN SL xs))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[vsubst_THM]);;
  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[vsubst_THM]);;
  e(REWRITE_TAC[SET_LIST;IN]);;
  e(POP_TAC);;
  e(REWRITE_TAC[DISJOINT_LIST]);;
  e(REPEAT STRIP_TAC);;
  e(ASM_CASES_TAC "(x:var) = h'");;

   e(ASM_REWRITE_TAC[BND_MATCH]);;

   e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
   e(ASM_REWRITE_TAC[]);;
   e(ASM_CASES_TAC "x = (@z. (ys //v l')z = h)");;

    e(POP_ASSUM (ASSUME_TAC o SYM));;
    e(ASM_REWRITE_TAC[BND_MATCH]);;
    e(ONCE_REWRITE_TAC[vsubst_EQ_INVERSE]);;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(POP_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL));;
    e(POP_TAC THEN POP_TAC);;
    e(POP_ASSUM (REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[vsubst_INVERSE_SELECT]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(ONCE_REWRITE_TAC[vsubst_EQ_INVERSE]);;
    e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let vsubst_ELEMENTS_INVERSE =
 prove_thm
  (`vsubst_ELEMENTS_INVERSE`,
   "!ys xs. (LENGTH xs = LENGTH ys) /\ (DL xs /\ DL ys) ==>
           (!x. ((\y. @z. (ys //v xs)z = y)x IN SL xs) = (x IN SL ys))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_ELEMENTS
   THEN FIRST_ASSUM (REWRITE_THM o SYM o SPEC_ALL)
   THEN BETA_TAC
   THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
  );;

% vsubst_ELEMENTS_INVERSE:
g "!ys xs. (LENGTH xs = LENGTH ys) /\ (DL xs /\ DL ys) ==>
           (!x. ((\y. @z. (ys //v xs)z = y)x IN SL xs) = (x IN SL ys))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_ELEMENTS);;
e(FIRST_ASSUM (REWRITE_THM o SYM o SPEC_ALL));;
e(BETA_TAC);;
e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%




let vsubst_REPLACEMENT =
 prove_thm
  (`vsubst_REPLACEMENT`,
   "!ys xs. (LENGTH xs = LENGTH ys) ==> (DL xs /\ DL ys) ==>
           (MAP (ys //v xs) xs = ys)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]
   THEN REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[vsubst_THM;MAP]
   THEN REWRITE_TAC[DISJOINT_LIST;CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM MAP_BND_ABSORPTION
        THEN USE_IMP_THEN REWRITE_THM MAP_BND_ABSORPTION
        THENL
          [  IMP_RES_TAC vsubst_ELEMENTS
             THEN POP_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL)
             THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
             THEN ASM_REWRITE_TAC[];

             RES_TAC
          ]
     ]
  );;

% vsubst_REPLACEMENT:
g "!ys xs. (LENGTH xs = LENGTH ys) ==> (DL xs /\ DL ys) ==>
           (MAP (ys //v xs) xs = ys)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[vsubst_THM]);;
  e(REWRITE_TAC[MAP]);;

  e(REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[vsubst_THM]);;
  e(REWRITE_TAC[MAP]);;
  e(REWRITE_TAC[DISJOINT_LIST]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REPEAT STRIP_TAC);;

   e(REWRITE_TAC[BND_MATCH]);;

   e(IMP_RES_THEN REWRITE_THM MAP_BND_ABSORPTION);;
   e(USE_IMP_THEN REWRITE_THM MAP_BND_ABSORPTION);;

    e(IMP_RES_TAC vsubst_ELEMENTS);;
    e(POP_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL));;
    e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
    e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;
%


let vsubst_REPLACEMENT_INVERSE =
 prove_thm
  (`vsubst_REPLACEMENT_INVERSE`,
   "!ys xs. (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
            (MAP (\y. @z. (ys //v xs)z = y) ys = xs)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_REPLACEMENT
   THEN POP_ASSUM (ONCE_REWRITE_THM o
        (AP_TERM "MAP (\y. @z. (ys //v xs)z = y)") o SYM)
   THEN REWRITE_TAC[MAP_MAP_o;vsubst_INVERSE;MAP_I]
  );;

% vsubst_REPLACEMENT_INVERSE:
g "!ys xs. (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
           (MAP (\y. @z. (ys //v xs)z = y) ys = xs)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_REPLACEMENT);;
e(POP_ASSUM (ONCE_REWRITE_THM o
        (AP_TERM "MAP (\y. @z. (ys //v xs)z = y)") o SYM));;
e(REWRITE_TAC[MAP_MAP_o]);;
e(REWRITE_TAC[vsubst_INVERSE;MAP_I]);;
%




let vsubst_MAP_ONE_ONE =
 prove_thm
  (`vsubst_MAP_ONE_ONE`,
   "!xs ss z. ONE_ONE ss /\ z IN SL xs ==>
              ((MAP ss xs //v xs) z = ss z)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[MAP;vsubst_THM]
   THEN REPEAT STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        RES_TAC
        THEN ASM_CASES_TAC "(z:var) = h"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH
             THEN ASM_REWRITE_TAC[SYM(SPEC_ALL vsubst_EQ_INVERSE1)]
             THEN REWRITE_ALL_TAC[ONE_ONE]
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% vsubst_MAP_ONE_ONE:
g "!xs ss z. ONE_ONE ss /\ z IN SL xs ==>
              ((MAP ss xs //v xs) z = ss z)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[MAP;vsubst_THM]);;
  e(REPEAT STRIP_TAC);;

   e(ASM_REWRITE_TAC[BND_MATCH]);;

   e(RES_TAC);;
   e(ASM_CASES_TAC "(z:var) = h");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SYM(SPEC_ALL vsubst_EQ_INVERSE1)]);;
    e(REWRITE_ALL_TAC[ONE_ONE]);;
    e(ASM_REWRITE_TAC[]);;
%


let state_vsubst_11 =
 prove_thm
  (`state_vsubst_11`,
   "!ys xs (s1:state) s2. (s1 o (ys //v xs) = s2 o (ys //v xs)) = (s1 = s2)",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [  DISCH_TAC
        THEN EXT_TAC "x:var"
        THEN GEN_TAC
        THEN POP_ASSUM (\th. ASSUME_TAC (AP_THM th "@z. (ys //v xs)z = x"))
        THEN POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th))
        THEN REWRITE_TAC[o_THM;vsubst_INVERSE_SELECT];

        DISCH_THEN REWRITE_THM
     ]
  );;

% state_vsubst_11
g "!ys xs (s1:state) s2. (s1 o (ys //v xs) = s2 o (ys //v xs)) = (s1 = s2)";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(EXT_TAC "x:var");;
 e(GEN_TAC);;
 e(POP_ASSUM (\th. ASSUME_TAC (AP_THM th "@z. (ys //v xs)z = x")));;
 e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
 e(REWRITE_TAC[o_THM;vsubst_INVERSE_SELECT]);;

 e(DISCH_THEN REWRITE_THM);;
%




let BND_o_ONE_ONE_THM =
 prove_thm
  (`BND_o_ONE_ONE_THM`,
   "!(ss1:vsubst) n v (s1:state).
     ONE_ONE ss1 ==>
     ((BND n(ss1 v)s1) o ss1 = BND n v (s1 o ss1))",
   REWRITE_TAC[ONE_ONE]
   THEN REPEAT STRIP_TAC
   THEN EXT_TAC "u:var"
   THEN GEN_TAC
   THEN ASM_REWRITE_TAC[BND;o_THM]
  );;

%< BND_o_ONE_ONE_THM
g "!(ss1:vsubst) n v (s1:state).
    ONE_ONE ss1 ==>
    ((BND n(ss1 v)s1) o ss1 = BND n v (s1 o ss1))";;
e(REWRITE_TAC[ONE_ONE]);;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:var");;
e(GEN_TAC);;
e(ASM_REWRITE_TAC[BND;o_THM]);;
>%


let BND_o_vsubst_THM =
 prove_thm
  (`BND_o_vsubst_THM`,
   "!ys xs n v (s1:state).
   ((BND n ((ys //v xs) v) s1) o (ys //v xs) = BND n v (s1 o (ys //v xs)))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "u:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND;o_THM;vsubst_11]
  );;

% BND_o_vsubst_THM
g "!ys xs n v (s1:state).
   ((BND n ((ys //v xs) v) s1) o (ys //v xs) = BND n v (s1 o (ys //v xs)))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:var");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;o_THM;vsubst_11]);;
%

let BND_o_vsubst_THM1 =
 prove_thm
  (`BND_o_vsubst_THM1`,
   "!ys xs n v (s1:state).
   ((BND n v s1) o (\y. @z. (ys //v xs)z = y) =
     BND n ((ys //v xs)v) (s1 o (\y. @z. (ys //v xs)z = y)))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "u:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND;o_THM]
   THEN BETA_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]
   THEN COND_CASES_TAC
   THENL
     [  POP_ASSUM REWRITE_THM;
        POP_ASSUM (REWRITE_THM o NOT_EQ_SYM)
     ]
  );;

% BND_o_vsubst_THM1
g "!ys xs n v (s1:state).
   ((BND n v s1) o (\y. @z. (ys //v xs)z = y) =
     BND n ((ys //v xs)v) (s1 o (\y. @z. (ys //v xs)z = y)))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:var");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;o_THM]);;
e(BETA_TAC);;
e(REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]);;
e(COND_CASES_TAC);;

 e(POP_ASSUM REWRITE_THM);;

 e(POP_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
%

let BND_o_vsubst_THM2 =
 prove_thm
  (`BND_o_vsubst_THM2`,
   "!ys xs n v (s1:state).
   ((BND n v s1) o (ys //v xs) =
     BND n (@z. (ys //v xs)z = v) (s1 o (ys //v xs)))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "u:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[BND;o_THM]
   THEN REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
  );;

% BND_o_vsubst_THM2
g "!ys xs n v (s1:state).
   ((BND n v s1) o (ys //v xs) =
     BND n (@z. (ys //v xs)z = v) (s1 o (ys //v xs)))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "u:var");;
e(GEN_TAC);;
e(REWRITE_TAC[BND;o_THM]);;
e(REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
%



let BNDsL_o_vsubst_THM =
 prove_thm
  (`BNDsL_o_vsubst_THM`,
   "!vs ns (s1:state) ys xs.
    ((BNDsL ns vs s1) o (ys //v xs) =
      BNDsL ns (MAP (\y. @z. (ys //v xs)z = y) vs) (s1 o (ys //v xs)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN REPEAT GEN_TAC
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) BND_o_vsubst_THM]
   THEN BETA_TAC
   THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
  );;

% BNDsL_o_vsubst_THM
g "!vs ns (s1:state) ys xs.
    ((BNDsL ns vs s1) o (ys //v xs) =
      BNDsL ns (MAP (\y. @z. (ys //v xs)z = y) vs) (s1 o (ys //v xs)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;
 e(REPEAT GEN_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_vsubst_THM]);;
 e(BETA_TAC);;
 e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%

let BNDsL_o_vsubst_THM1 =
 prove_thm
  (`BNDsL_o_vsubst_THM1`,
   "!vs ns (s1:state) ys xs.
    ((BNDsL ns vs s1) o (\y. @z. (ys //v xs)z = y) =
      BNDsL ns (MAP (ys //v xs) vs) (s1 o (\y. @z. (ys //v xs)z = y)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsL;MAP]
   THEN REPEAT GEN_TAC
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) BND_o_vsubst_THM1]
  );;

% BNDsL_o_vsubst_THM1
g "!vs ns (s1:state) ys xs.
    ((BNDsL ns vs s1) o (\y. @z. (ys //v xs)z = y) =
      BNDsL ns (MAP (ys //v xs) vs) (s1 o (\y. @z. (ys //v xs)z = y)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;

 e(REWRITE_TAC[BNDsL;MAP]);;
 e(REPEAT GEN_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_vsubst_THM1]);;
%



let BNDsR_o_vsubst_THM =
 prove_thm
  (`BNDsR_o_vsubst_THM`,
   "!vs ns (s1:state) ys xs.
    ((BNDsR ns vs s1) o (ys //v xs) =
      BNDsR ns (MAP (\y. @z. (ys //v xs)z = y) vs) (s1 o (ys //v xs)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR;MAP]
   THEN REPEAT GEN_TAC
   THEN REWRITE_TAC[BND_o_vsubst_THM2]
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% BNDsR_o_vsubst_THM
g "!vs ns (s1:state) ys xs.
    ((BNDsR ns vs s1) o (ys //v xs) =
      BNDsR ns (MAP (\y. @z. (ys //v xs)z = y) vs) (s1 o (ys //v xs)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;
 e(REPEAT GEN_TAC);;
 e(REWRITE_TAC[BND_o_vsubst_THM2]);;
 e(BETA_TAC);;
 e(ASM_REWRITE_TAC[]);;
%

let BNDsR_o_vsubst_THM1 =
 prove_thm
  (`BNDsR_o_vsubst_THM1`,
   "!vs ns (s1:state) ys xs.
    ((BNDsR ns vs s1) o (\y. @z. (ys //v xs)z = y) =
      BNDsR ns (MAP (ys //v xs) vs) (s1 o (\y. @z. (ys //v xs)z = y)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR;MAP]
   THEN REPEAT GEN_TAC
   THEN ASM_REWRITE_TAC[BND_o_vsubst_THM1]
  );;

% BNDsR_o_vsubst_THM1
g "!vs ns (s1:state) ys xs.
    ((BNDsR ns vs s1) o (\y. @z. (ys //v xs)z = y) =
      BNDsR ns (MAP (ys //v xs) vs) (s1 o (\y. @z. (ys //v xs)z = y)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;

 e(REWRITE_TAC[BNDsR;MAP]);;
 e(REPEAT GEN_TAC);;
 e(REWRITE_TAC[BND_o_vsubst_THM1]);;
 e(ASM_REWRITE_TAC[]);;
%


let BNDsL_MAP_o_vsubst =
 prove_thm
  (`BNDsL_MAP_o_vsubst`,
   "!vs ns ys xs (s:state).
      (BNDsL ns (MAP (ys //v xs) vs) s) o (ys //v xs) =
      BNDsL ns vs (s o (ys //v xs))",
   REWRITE_TAC[BNDsL_o_vsubst_THM]
   THEN REWRITE_TAC[(SYM o SPEC_ALL o
               ISPECL["MAP (f:var->var)";"MAP (g:var->var)"]) o_THM]
   THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_o;vsubst_INVERSE;MAP_I]
  );;

% BNDsL_MAP_o_vsubst
g "!vs ns ys xs (s:state).
     (BNDsL ns (MAP (ys //v xs) vs) s) o (ys //v xs) =
     BNDsL ns vs (s o (ys //v xs))";;
e(REWRITE_TAC[BNDsL_o_vsubst_THM]);;
e(REWRITE_TAC[(SYM o SPEC_ALL o
               ISPECL["MAP (f:var->var)";"MAP (g:var->var)"]) o_THM]);;
e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_o]);;
e(REWRITE_TAC[vsubst_INVERSE;MAP_I]);;
%



let MAP_state_vsubst =
 prove_thm
  (`MAP_state_vsubst`,
   "!vs (s1:state) ys xs.
     MAP (s1 o (ys //v xs)) (MAP (\y. @z. (ys //v xs)z = y) vs) =
     MAP s1 vs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[o_THM;vsubst_INVERSE_SELECT]
  );;

% MAP_state_vsubst
g "!vs (s1:state) ys xs.
    MAP (s1 o (ys //v xs)) (MAP (\y. @z. (ys //v xs)z = y) vs) =
    MAP s1 vs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[MAP]);;
 e(BETA_TAC);;
 e(ASM_REWRITE_TAC[o_THM;vsubst_INVERSE_SELECT]);;
%

let MAP_state_vsubst1 =
 prove_thm
  (`MAP_state_vsubst1`,
   "!vs (s1:state) ys xs.
    MAP (s1 o (\y. @z. (ys //v xs)z = y)) (MAP (ys //v xs) vs) =
    MAP s1 vs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THEN REWRITE_TAC[o_THM]
   THEN BETA_TAC
   THEN ASM_REWRITE_TAC[o_THM;vsubst_11;SELECT_REFL]
  );;

% MAP_state_vsubst1
g "!vs (s1:state) ys xs.
    MAP (s1 o (\y. @z. (ys //v xs)z = y)) (MAP (ys //v xs) vs) =
    MAP s1 vs";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;

 e(REWRITE_TAC[MAP]);;
 e(REWRITE_TAC[o_THM]);;
 e(BETA_TAC);;
 e(ASM_REWRITE_TAC[o_THM;vsubst_11;SELECT_REFL]);;
%


let state_vsubst_INVERSE =
 prove_thm
  (`state_vsubst_INVERSE`,
   "!ys xs (s1:state) s2.
    (s1 o (ys //v xs) = s2) = (s1 = (s2 o (\y. @z. (ys //v xs)z = y)))",
   REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [  DISCH_THEN (REWRITE_THM o SYM);
        DISCH_THEN REWRITE_THM
     ]
   THEN REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;vsubst_INVERSE;I_o_ID]
  );;

%
g "!ys xs (s1:state) s2.
    (s1 o (ys //v xs) = s2) = (s1 = (s2 o (\y. @z. (ys //v xs)z = y)))";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;vsubst_INVERSE;I_o_ID]);;

 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC;vsubst_INVERSE;I_o_ID]);;
%

let logical_vsubst =
 prove_thm
  (`logical_vsubst`,
   "!ys xs x.
    ((logicals ys) //v (logicals xs))(logical x) = logical((ys //v xs)x)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;vsubst_THM;I_THM]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(x:var) = h'"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
        THENL
          [  ASM_REWRITE_TAC[logical_11];

             ASM_CASES_TAC "x = (@z. (ys //v xs)z = h)"
             THENL
               [  POP_ASSUM (ASSUME_TAC o SYM)
                  THEN ASM_REWRITE_TAC[BND_MATCH]
                  THEN REWRITE_ALL_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]
                  THEN POP_ASSUM (\th. FIRST_ASSUM
                           (REWRITE_THM o SYM o REWRITE_RULE[th] o 
                                 SPECL["xs:(var)list";"x:var"]))
                  THEN REWRITE_TAC[vsubst_11;SELECT_REFL;BND_MATCH];

                  IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
                  THEN USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH
                  THEN REWRITE_ALL_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
                  THEN ASM_REWRITE_TAC[logical_11]
               ]
          ]
     ]
  );;

% logical_vsubst
g "!ys xs x.
    ((logicals ys) //v (logicals xs))(logical x) = logical((ys //v xs)x)";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[logicals_THM;vsubst_THM;I_THM]);;

  e(REWRITE_TAC[logicals_THM;vsubst_THM;I_THM]);;

  e(REWRITE_TAC[logicals_THM;vsubst_THM;I_THM]);;
  e(REPEAT GEN_TAC);;
  e(ASM_CASES_TAC "(x:var) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

      e(ASM_REWRITE_TAC[logical_11]);;

      e(ASM_CASES_TAC "x = (@z. (ys //v xs)z = h)");;

        e(POP_ASSUM (ASSUME_TAC o SYM));;
        e(ASM_REWRITE_TAC[BND_MATCH]);;
        e(REWRITE_ALL_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]);;
        e(POP_ASSUM (\th. FIRST_ASSUM
              (REWRITE_THM o SYM o REWRITE_RULE[th] o 
                    SPECL["xs:(var)list";"x:var"])));;
        e(REWRITE_TAC[vsubst_11;SELECT_REFL]);;
        e(REWRITE_TAC[BND_MATCH]);;

        e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
        e(USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
        e(REWRITE_ALL_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
        e(ASM_REWRITE_TAC[logical_11]);;
%



% ============================================================ %
% We now define the application of a variable-substitution     %
% for assertion language expressions.                          %
% ============================================================ %


% ============================================================ %
% Assertion expressions:                                       %
% ============================================================ %

let v_vsubst =
    new_infix_rec_def `v_vsubst` vexp_Axiom
      "($<|vv (ANUM n) ss       = ANUM n)  /\
       ($<|vv (AVAR y) ss       = AVAR (ss y))    /\
       ($<|vv (APLUS e1 e2)  ss = APLUS  ($<|vv e1 ss)
                                         ($<|vv e2 ss))  /\
       ($<|vv (AMINUS e1 e2) ss = AMINUS ($<|vv e1 ss)
                                         ($<|vv e2 ss))  /\
       ($<|vv (AMULT e1 e2)  ss = AMULT  ($<|vv e1 ss)
                                         ($<|vv e2 ss))";;

let vs_vsubst =
    new_infix_list_rec_def `vs_vsubst`
      "($<|vsv (NIL) ss         = [])  /\
       ($<|vsv (CONS e es) ss   = CONS (e <|vv ss) ($<|vsv es ss))";;

let a_vsubst =
    new_infix_rec_def `a_vsubst` aexp_Axiom
      "($<|av (ATRUE) ss          = ATRUE)     /\
       ($<|av (AFALSE) ss         = AFALSE)    /\
       ($<|av (AEQ e1 e2) ss      = AEQ  (e1 <|vv ss)
                                         (e2 <|vv ss))  /\
       ($<|av (ALESS e1 e2) ss    = ALESS (e1 <|vv ss)
                                          (e2 <|vv ss)) /\
       ($<|av (ALLESS es1 es2) ss = ALLESS (es1 <|vsv ss)
                                           (es2 <|vsv ss)) /\
       ($<|av (AAND b1 b2) ss     = AAND ($<|av b1 ss)
                                         ($<|av b2 ss))  /\
       ($<|av (AOR  b1 b2) ss     = AOR  ($<|av b1 ss)
                                         ($<|av b2 ss))  /\
       ($<|av (ANOT b) ss         = ANOT ($<|av b ss))   /\
       ($<|av (AIMP b1 b2) ss     = AIMP ($<|av b1 ss)
                                         ($<|av b2 ss))  /\
       ($<|av (AEQB b1 b2) ss     = AEQB ($<|av b1 ss)
                                         ($<|av b2 ss))  /\
       ($<|av (ACOND b1 b2 b3) ss = ACOND ($<|av b1 ss)
                                          ($<|av b2 ss)
                                          ($<|av b3 ss)) /\
       ($<|av (ACLOSE b) ss       = ACLOSE b)            /\
       ($<|av (AFORALL y b) ss    = AFORALL (ss y) ($<|av b ss))  /\
       ($<|av (AEXISTS y b) ss    = AEXISTS (ss y) ($<|av b ss))
           ";;



% But are these substitutions of variables for variables %
% equivalent to the old substitutions on vexp and aexp?  %

let v_vsubst_EQ_subst =
 prove_thm
  (`v_vsubst_EQ_subst`,
   "!v ss. v <|vv ss = v <|v (AVAR o ss)",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_vsubst;v_subst;o_THM]
  );;

%
g "!v ss. v <|vv ss = v <|v (AVAR o ss)";;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC);;

 e(REWRITE_TAC[v_vsubst;v_subst]);;

 e(REWRITE_TAC[v_vsubst;v_subst;o_THM]);;

 e(ASM_REWRITE_TAC[v_vsubst;v_subst;o_THM]);;

 e(ASM_REWRITE_TAC[v_vsubst;v_subst;o_THM]);;

 e(ASM_REWRITE_TAC[v_vsubst;v_subst;o_THM]);;
%

let vs_vsubst_EQ_subst =
 prove_thm
  (`vs_vsubst_EQ_subst`,
   "!vs ss. vs <|vsv ss = vs <|vs (AVAR o ss)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[vs_vsubst;vs_subst]
   THEN ASM_REWRITE_TAC[v_vsubst_EQ_subst]
  );;

%
g "!vs ss. vs <|vsv ss = vs <|vs (AVAR o ss)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vs_vsubst;vs_subst]);;

 e(REWRITE_TAC[vs_vsubst;vs_subst]);;
 e(ASM_REWRITE_TAC[v_vsubst_EQ_subst]);;
%

let s_subst_vsubst =
 prove_thm
  (`s_subst_vsubst`,
   "!s ss. s <|s (AVAR o ss) = s o ss",
   REWRITE_TAC[s_subst]
   THEN REPEAT GEN_TAC
   THEN EXT_TAC "x:var"
   THEN GEN_TAC
   THEN BETA_TAC
   THEN REWRITE_TAC[o_THM;V_DEF]
  );;

% s_subst_vsubst
g "!s ss. s <|s (AVAR o ss) = s o ss";;
e(REWRITE_TAC[s_subst]);;
e(REPEAT GEN_TAC);;
e(EXT_TAC "x:var");;
e(GEN_TAC);;
e(BETA_TAC);;
e(REWRITE_TAC[o_THM;V_DEF]);;
%

let v_vsubst_THM =
 prove_thm
  (`v_vsubst_THM`,
   "!v ss s. V (v <|vv ss) s = V v (s o ss)",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_vsubst;V_DEF;o_THM]
  );;

% v_vsubst_THM
g "!v ss s. V (v <|vv ss) s = V v (s o ss)";;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC);;

 e(ASM_REWRITE_TAC[v_vsubst;V_DEF;o_THM]);;

or,
e(REWRITE_TAC[v_vsubst_subst;v_subst_THM;s_subst_vsubst]);;
%

let vs_vsubst_THM =
 prove_thm
  (`vs_vsubst_THM`,
   "!vs ss s. VS (vs <|vsv ss) s = VS vs (s o ss)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[vs_vsubst;VS_DEF;v_vsubst_THM]
  );;

% vs_vsubst_THM
g "!vs ss s. VS (vs <|vsv ss) s = VS vs (s o ss)";;
e(LIST_INDUCT_TAC);;

 e(ASM_REWRITE_TAC[vs_vsubst;VS_DEF;v_vsubst_THM]);;

or,
e(REWRITE_TAC[v_vsubst_subst;v_subst_THM;s_subst_vsubst]);;
%

let a_vsubst_ONE_ONE =
 prove_thm
  (`a_vsubst_ONE_ONE`,
   "!a ss s.
     ONE_ONE ss ==>
     (A (a <|av ss) s = A a (s o ss))",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;A_DEF;v_vsubst_THM;vs_vsubst_THM;o_THM]
   THEN IMP_RES_THEN REWRITE_THM BND_o_ONE_ONE
  );;

% a_vsubst_ONE_ONE
g "!a ss s.
     ONE_ONE ss ==>
     (A (a <|av ss) s = A a (s o ss))";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[a_vsubst;A_DEF;v_vsubst_THM;vs_vsubst_THM;o_THM]
  THEN IMP_RES_THEN REWRITE_THM BND_o_ONE_ONE
 );;
%

let a_vsubst_THM =
 prove_thm
  (`a_vsubst_THM`,
   "!a ys xs s. A (a <|av (ys //v xs)) s = A a (s o (ys //v xs))",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;A_DEF;v_vsubst_THM;vs_vsubst_THM;
                        o_THM;BND_o_vsubst_THM]
  );;

% a_vsubst_THM
g "!a ys xs s. A (a <|av (ys //v xs)) s = A a (s o (ys //v xs))";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN ASM_REWRITE_TAC[a_vsubst;A_DEF;v_vsubst_THM;vs_vsubst_THM;
                       o_THM;BND_o_vsubst_THM]
 );;
%

let a_vsubst_EQ_subst =
 prove_thm
  (`a_vsubst_EQ_subst`,
   "!a ys xs s. A (a <|av (ys //v xs)) s = A (a <|a (AVAR o (ys //v xs))) s",
   REWRITE_TAC[a_vsubst_THM;a_subst_THM;s_subst_vsubst]
  );;

% a_vsubst_EQ_subst
g "!a ys xs s. A (a <|av (ys //v xs)) s = A (a <|a (AVAR o (ys //v xs))) s";;
e(REWRITE_TAC[a_vsubst_THM;a_subst_THM;s_subst_vsubst]);;
%


let BNDsR_EQ_vsubst =
 prove_thm
  (`BNDsR_EQ_vsubst`,
   "!ys xs ss x.
      (LENGTH xs = LENGTH ys) ==> (DL xs) /\ (DL ys) /\ (x IN SL xs) ==>
      (BNDsR ys xs ss x = (ys //v xs)x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[DISJOINT_LIST;vsubst_THM;BNDsR]
   THEN REPEAT STRIP_TAC
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_TAC vsubst_ELEMENTS
        THEN IMP_RES_TAC IN_NOT_IN
        THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN UNASSUME_THEN (ASSUME_TAC o ONCE_REWRITE_RULE[vsubst_EQ_INVERSE1])
                   "~((ys //v l')x = h)"
        THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% BNDsR_EQ_vsubst
g "!ys xs ss x.
      (LENGTH xs = LENGTH ys) ==> (DL xs) /\ (DL ys) /\ (x IN SL xs) ==>
      (BNDsR ys xs ss x = (ys //v xs)x)";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REWRITE_TAC[DISJOINT_LIST;vsubst_THM;BNDsR]);;
 e(REPEAT STRIP_TAC);;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_TAC vsubst_ELEMENTS);;
  e(IMP_RES_TAC IN_NOT_IN);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(UNASSUME_THEN (ASSUME_TAC o ONCE_REWRITE_RULE[vsubst_EQ_INVERSE1])
            "~((ys //v l')x = h)");;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let state_vsubst_EQ_BNDsR =
 prove_thm
  (`state_vsubst_EQ_BNDsR`,
   "!ys xs (s:state) x.
      (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) /\ (x IN SL xs) ==>
      ((s o (ys //v xs))x = BNDsR(MAP s ys) xs s x)",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[o_THM]
   THEN IMP_RES_TAC BNDsR_EQ_vsubst
   THEN POP_ASSUM (REWRITE_THM o SYM o SPEC "I:vsubst")
   THEN ONCE_REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "s:state") o_THM]
   THEN REWRITE_TAC[o_BNDsR;I_o_ID]
  );;

% state_vsubst_EQ_BNDsR
g "!ys xs (s:state) x.
      (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) /\ (x IN SL xs) ==>
      ((s o (ys //v xs))x = BNDsR(MAP s ys) xs s x)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[o_THM]);;
e(IMP_RES_TAC BNDsR_EQ_vsubst);;
e(POP_ASSUM (REWRITE_THM o SYM o SPEC "I:vsubst"));;
e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "s:state") o_THM]);;
e(REWRITE_TAC[o_BNDsR;I_o_ID]);;
%

let FVv_vsubst_subst =
 prove_thm
  (`FVv_vsubst_subst`,
   "!v ys xs s. (FVv v SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (V (v <|vv (ys //v xs)) s = V (v <|v (ys // xs)) s)",
   REWRITE_TAC[v_vsubst_EQ_subst;v_subst_THM]
   THEN REWRITE_TAC[s_subst_vsubst;s_subst_var_list]
   THEN INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC state_vsubst_EQ_BNDsR
   THEN ASM_REWRITE_TAC[V_DEF]
  );;

% FVv_vsubst_subst
g "!v ys xs s. (FVv v SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (V (v <|vv (ys //v xs)) s = V (v <|v (ys // xs)) s)";;
e(REWRITE_TAC[v_vsubst_EQ_subst;v_subst_THM]);;
e(REWRITE_TAC[s_subst_vsubst;s_subst_var_list]);;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
);;

 e(ASM_REWRITE_TAC[V_DEF]);;

 e(IMP_RES_TAC state_vsubst_EQ_BNDsL);;
 e(ASM_REWRITE_TAC[V_DEF]);;

 e(RES_TAC);;
 e(IMP_RES_TAC state_vsubst_EQ_BNDsL);;
 e(ASM_REWRITE_TAC[V_DEF]);;
%

let FVvs_vsubst_subst =
 prove_thm
  (`FVvs_vsubst_subst`,
   "!vs ys xs s. (FVvs vs SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (VS (vs <|vsv (ys //v xs)) s = VS (vs <|vs (ys // xs)) s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[vs_vsubst;vs_subst]
   THEN REWRITE_TAC[FVvs_DEF;UNION_SUBSET;VS_DEF;CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
    [ MATCH_MP_TAC FVv_vsubst_subst;

      FIRST_ASSUM MATCH_MP_TAC
    ]
   THEN ASM_REWRITE_TAC[]
  );;

% FVvs_vsubst_subst
g "!vs ys xs s. (FVvs vs SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (VS (vs <|vsv (ys //v xs)) s = VS (vs <|vs (ys // xs)) s)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vs_vsubst;vs_subst]);;

 e(REWRITE_TAC[vs_vsubst;vs_subst]);;
 e(REWRITE_TAC[FVvs_DEF;UNION_SUBSET]);;
 e(REWRITE_TAC[VS_DEF;CONS_11]);;
 e(REPEAT STRIP_TAC);;

  e(MATCH_MP_TAC FVv_vsubst_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;
%



let DELETE_SUBSET_SL =
 TAC_PROOF
  (([], "!s xs (x:*). (s DELETE x) SUBSET (SL xs) = (s SUBSET (SL (CONS x xs)))"),
   REWRITE_TAC[SET_LIST;SUBSET_INSERT_DELETE]
  );;

% DELETE_SUBSET_SL
g "!s xs (x:*). (s DELETE x) SUBSET (SL xs) = (s SUBSET (SL (CONS x xs)))";;
e(REWRITE_TAC[SET_LIST;SUBSET_INSERT_DELETE]);;
%




let BNDsR_vsubst =
 prove_thm
  (`BNDsR_vsubst`,
   "!ys xs x. (LENGTH xs = LENGTH ys) ==>
     (x IN SL xs) /\ (DL xs) /\ (DL ys) ==>
     (BNDsR ys xs I x = (ys //v xs) x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[vsubst_THM;BNDsR]
   THEN REWRITE_TAC[SET_LIST;IN;DISJOINT_LIST]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[BND_MATCH]
   THEN IMP_RES_TAC vsubst_ELEMENTS
   THEN IMP_RES_TAC IN_NOT_IN
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
   THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
  );;

% BNDsR_vsubst
g "!ys xs x. (LENGTH xs = LENGTH ys) ==>
     (x IN SL xs) /\ (DL xs) /\ (DL ys) ==>
     (BNDsR ys xs I x = (ys //v xs) x)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[vsubst_THM;BNDsR]);;

  e(REWRITE_TAC[vsubst_THM;BNDsR]);;
  e(REWRITE_TAC[SET_LIST;IN;DISJOINT_LIST]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_TAC vsubst_ELEMENTS);;
    e(IMP_RES_TAC IN_NOT_IN);;
    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;
    e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
%

let subst_BNDsR =
 prove_thm
  (`subst_BNDsR`,
   "!ys xs. (ys // xs) = (AVAR o (BNDsR ys xs I))",
   REWRITE_TAC[var_list_subst;o_BNDsR;I_o_ID;IDENT_SS]
  );;

% subst_BNDsR
g "!ys xs. (ys // xs) = (AVAR o (BNDsR ys xs I))";;
e(REWRITE_TAC[var_list_subst;o_BNDsR;I_o_ID;IDENT_SS]);;
%

let NULL_CONS_subst =
 prove_thm
  (`NULL_CONS_subst`,
   "!ys xs v.
     (LENGTH xs = LENGTH ys) /\
     (v IN SL xs) /\ (DL xs) /\ (DL ys) ==>
    (((CONS ((ys //v xs) v) ys) // (CONS v xs)) = (ys // xs))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN REWRITE_TAC[subst_BNDsR;o_THM;BNDsR]
   THEN ASM_CASES_TAC "(z:var) = v"
   THENL
     [  IMP_RES_TAC BNDsR_vsubst
        THEN ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
     ]
  );;

% NULL_CONS_subst
g "!ys xs v.
     (LENGTH xs = LENGTH ys) /\
     (v IN SL xs) /\ (DL xs) /\ (DL ys) ==>
    (((CONS ((ys //v xs) v) ys) // (CONS v xs)) = (ys // xs))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(REWRITE_TAC[subst_BNDsR]);;
e(REWRITE_TAC[o_THM;BNDsR]);;
e(ASM_CASES_TAC "(z:var) = v");;

  e(IMP_RES_TAC BNDsR_vsubst);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
%

let NULL_APPEND_subst =
 prove_thm
  (`NULL_APPEND_subst`,
   "!vs ys xs.
     (LENGTH xs = LENGTH ys) /\
     (SL vs SUBSET SL xs) /\ (DL xs) /\ (DL ys) ==>
     (((APPEND (MAP (ys //v xs) vs) ys) // (APPEND vs xs)) = (ys // xs))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;SUBSET;MAP;APPEND]
   THEN REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[var_list_THM;subst_BNDsR]
   THEN ASM_CASES_TAC "(z:var) = h"
   THENL
     [  IMP_RES_TAC BNDsR_vsubst
        THEN ASM_REWRITE_TAC[o_THM;BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
     ]
  );;

% NULL_APPEND_subst
g "!vs ys xs.
     (LENGTH xs = LENGTH ys) /\
     (SL vs SUBSET SL xs) /\ (DL xs) /\ (DL ys) ==>
     (((APPEND (MAP (ys //v xs) vs) ys) // (APPEND vs xs)) = (ys // xs))";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;SUBSET;MAP;APPEND]);;

  e(REWRITE_TAC[SET_LIST;SUBSET;MAP;APPEND]);;
  e(REPEAT STRIP_TAC);;
  e(EXT_TAC "z:var");;
  e(GEN_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[var_list_THM]);;
  e(REWRITE_TAC[subst_BNDsR]);;
  e(ASM_CASES_TAC "(z:var) = h");;

    e(IMP_RES_TAC BNDsR_vsubst);;
    e(ASM_REWRITE_TAC[o_THM;BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
%

let NULL_CONS_vsubst =
 prove_thm
  (`NULL_CONS_vsubst`,
   "!ys xs v. ((CONS ((ys //v xs)v) ys) //v (CONS v xs)) = (ys //v xs)",
   DOUBLE_LIST_INDUCT_TAC
   THENL
     [  REWRITE_TAC[vsubst_THM;I_THM;SELECT_REFL;BND_I];

        REWRITE_TAC[vsubst_THM;I_THM;SELECT_REFL;BND_I];

        ONCE_REWRITE_TAC[vsubst_THM]
        THEN REPEAT GEN_TAC
        THEN EXT_TAC "u:var"
        THEN GEN_TAC
        THEN ASM_CASES_TAC "(u:var) = v"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH;vsubst_THM];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN REWRITE_TAC[vsubst_11;SELECT_REFL]
             THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN REWRITE_TAC[vsubst_THM]
          ]
     ]
  );;

% NULL_CONS_vsubst
g "!ys xs v. ((CONS ((ys //v xs)v) ys) //v (CONS v xs)) = (ys //v xs)";;
e(DOUBLE_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[vsubst_THM;I_THM;SELECT_REFL;BND_I]);;

 e(REWRITE_TAC[vsubst_THM;I_THM;SELECT_REFL;BND_I]);;

 e(ONCE_REWRITE_TAC[vsubst_THM]);;
 e(REPEAT GEN_TAC);;
 e(EXT_TAC "u:var");;
 e(GEN_TAC);;
 e(ASM_CASES_TAC "(u:var) = v");;

  e(ASM_REWRITE_TAC[BND_MATCH;vsubst_THM]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(REWRITE_TAC[vsubst_11;SELECT_REFL]);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(REWRITE_TAC[vsubst_THM]);;
%

let BND_s_subst_var_list =
 prove_thm
  (`BND_s_subst_var_list`,
   "!ys xs (s:state) n x z. 
        (z IN SL xs) /\
        (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (BND n x (s <|s (ys // xs)) z  =
             ((BND n ((ys //v xs)x) s) <|s (ys // xs)) z)",
   REWRITE_TAC[s_subst_o;o_THM]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(z:var) = x"
   THEN IMP_RES_TAC BNDsR_vsubst
   THEN ASM_REWRITE_TAC[BND_MATCH]
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN REWRITE_TAC[o_THM]
   THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
   THEN ASM_REWRITE_TAC[vsubst_11]
  );;

% BND_s_subst_var_list
g "!ys xs (s:state) n x z. 
        (z IN SL xs) /\
        (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (BND n x (s <|s (ys // xs)) z  =
             ((BND n ((ys //v xs)x) s) <|s (ys // xs)) z)";;
e(REWRITE_TAC[s_subst_o;o_THM]);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "(z:var) = x");;

  e(IMP_RES_TAC BNDsR_vsubst);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(REWRITE_TAC[o_THM]);;
  e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;
  e(IMP_RES_TAC BNDsR_vsubst);;
  e(ASM_REWRITE_TAC[vsubst_11]);;
%


let V_BND_s_subst_var_list =
 prove_thm
  (`V_BND_s_subst_var_list`,
   "!v ys xs (s:state) n x. 
        (FVv v SUBSET (SL xs)) /\ (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (V v  (BND n x (s <|s (ys // xs)))  =
             V v ((BND n ((ys //v xs)x) s) <|s (ys // xs)))",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC BND_s_subst_var_list
   THEN ASM_REWRITE_TAC[V_DEF]
  );;

% V_BND_s_subst_var_list
g "!v ys xs (s:state) n x. 
        (FVv v SUBSET (SL xs)) /\ (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (V v  (BND n x (s <|s (ys // xs)))  =
             V v ((BND n ((ys //v xs)x) s) <|s (ys // xs)))";;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[FVv_DEF;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN IMP_RES_TAC BND_s_subst_var_list
  THEN ASM_REWRITE_TAC[V_DEF]
);;
%

let VS_BND_s_subst_var_list =
 prove_thm
  (`VS_BND_s_subst_var_list`,
   "!vs ys xs (s:state) n x. 
        (FVvs vs SUBSET (SL xs)) /\ (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (VS vs  (BND n x (s <|s (ys // xs)))  =
             VS vs ((BND n ((ys //v xs)x) s) <|s (ys // xs)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[VS_DEF]
   THEN REWRITE_TAC[FVvs_DEF;UNION_SUBSET;CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
    [ MATCH_MP_TAC V_BND_s_subst_var_list;

      FIRST_ASSUM MATCH_MP_TAC
    ]
   THEN ASM_REWRITE_TAC[]
  );;

% VS_BND_s_subst_var_list
g "!vs ys xs (s:state) n x. 
        (FVvs vs SUBSET (SL xs)) /\ (x IN SL xs) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (VS vs  (BND n x (s <|s (ys // xs)))  =
             VS vs ((BND n ((ys //v xs)x) s) <|s (ys // xs)))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[VS_DEF]);;

 e(REWRITE_TAC[VS_DEF]);;
 e(REWRITE_TAC[FVvs_DEF;UNION_SUBSET;CONS_11]);;
 e(REPEAT STRIP_TAC);;

  e(MATCH_MP_TAC V_BND_s_subst_var_list);;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let IMAGE_var_list_subst =
 prove_thm
  (`IMAGE_var_list_subst`,
   "!s ys xs x. 
    ((s DELETE x) SUBSET (SL xs)) /\
    (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
      ((UNION_SET(IMAGE (FVv o (ys // xs)) (s DELETE x)))
       SUBSET ((SL ys) DELETE ((ys //v xs)x)))",
   REWRITE_TAC[SUBSET_DEF;IN_UNION_SET;IN_IMAGE;IN_DELETE;var_list_subst]
   THEN ONCE_REWRITE_TAC[(SYM o CONJUNCT2 o ISPEC "IDENT_SS") I_o_ID]
   THEN REWRITE_TAC[IDENT_SS;(SYM o SPEC_ALL) o_BNDsR;o_THM;FVv_DEF]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;IN])
            "si = (BNDsR ys xs I (x'':var)) INSERT (EMPTY:(var)set)"
   THEN RES_TAC
   THENL
     [  IMP_RES_TAC BNDsR_MEMBER
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC BNDsR_vsubst
        THEN ASM_REWRITE_TAC[vsubst_11]
     ]
  );;

% IMAGE_var_list_subst
g "!s ys xs x. 
    ((s DELETE x) SUBSET (SL xs)) /\
    (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
      ((UNION_SET(IMAGE (FVv o (ys // xs)) (s DELETE x)))
       SUBSET ((SL ys) DELETE ((ys //v xs)x)))";;
e(REWRITE_TAC[SUBSET_DEF;IN_UNION_SET;IN_IMAGE]);;
e(REWRITE_TAC[IN_DELETE;var_list_subst]);;
e(ONCE_REWRITE_TAC[(SYM o CONJUNCT2 o ISPEC "IDENT_SS") I_o_ID]);;
e(REWRITE_TAC[IDENT_SS;(SYM o SPEC_ALL) o_BNDsR]);;
e(REWRITE_TAC[o_THM;FVv_DEF]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC);;

  e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;IN])
         "si = (BNDsR ys xs I (x'':var)) INSERT (EMPTY:(var)set)");;
  e(RES_TAC);;
  e(IMP_RES_TAC BNDsR_MEMBER);;
  e(ASM_REWRITE_TAC[]);;

  e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;IN])
         "si = (BNDsR ys xs I (x'':var)) INSERT (EMPTY:(var)set)");;
  e(RES_TAC);;
  e(IMP_RES_TAC BNDsR_vsubst);;
  e(ASM_REWRITE_TAC[vsubst_11]);;
%

let IN_IMAGE_var_list_subst =
 prove_thm
  (`IN_IMAGE_var_list_subst`,
   "!s ys xs y x. 
    ((s DELETE x) SUBSET (SL xs)) /\
     (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) /\
     ~(y IN ((SL ys) DELETE ((ys //v xs)x))) ==>
     ~(y IN (UNION_SET(IMAGE (FVv o (ys // xs)) (s DELETE x))))",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC IMAGE_var_list_subst
   THEN FIRST_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL o
                 REWRITE_RULE[SUBSET_DEF])
   THEN ASM_REWRITE_TAC[]
  );;

% IN_IMAGE_var_list_subst
g "!s ys xs y x. 
    ((s DELETE x) SUBSET (SL xs)) /\
     (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) /\
     ~(y IN ((SL ys) DELETE ((ys //v xs)x))) ==>
     ~(y IN (UNION_SET(IMAGE (FVv o (ys // xs)) (s DELETE x))))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC IMAGE_var_list_subst);;
e(FIRST_ASSUM (MATCH_MP_TAC o CONTRAPOS o SPEC_ALL o
                 REWRITE_RULE[SUBSET_DEF]));;
e(ASM_REWRITE_TAC[]);;
%


let A_BND_s_subst_var_list =
 prove_thm
  (`A_BND_s_subst_var_list`,
   "!a ys xs (s:state) n x. 
        ((FVa a DELETE x) SUBSET (SL xs)) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (A a  (BND n x (s <|s (ys // xs)))  =
             A a ((BND n ((ys //v xs)x) s) <|s
                    ((CONS ((ys //v xs)x) ys) // (CONS x xs))))",
   REWRITE_TAC[var_list_THM]
   THEN REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]
   THEN REWRITE_TAC[BND_s_subst]
   THEN REPEAT STRIP_TAC
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC A_BND_s_subst
   THEN MATCH_MP_TAC IN_IMAGE_FV_DELETE
   THEN MATCH_MP_TAC IN_IMAGE_var_list_subst
   THEN ASM_REWRITE_TAC[IN_DELETE]
  );;

% A_BND_s_subst_var_list
g "!a ys xs (s:state) n x. 
        ((FVa a DELETE x) SUBSET (SL xs)) /\
        (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
            (A a  (BND n x (s <|s (ys // xs)))  =
             A a ((BND n ((ys //v xs)x) s) <|s
                    ((CONS ((ys //v xs)x) ys) // (CONS x xs))))";;
Hmm...

e(REWRITE_TAC[var_list_THM]);;
e(REWRITE_TAC[s_subst_BND;V_DEF;BND_MATCH]);;
e(REWRITE_TAC[BND_s_subst]);;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC A_BND_s_subst);;
e(MATCH_MP_TAC IN_IMAGE_FV_DELETE);;
e(MATCH_MP_TAC IN_IMAGE_var_list_subst);;
e(ASM_REWRITE_TAC[IN_DELETE]);;


#s_subst_BND;;
|- !v x ss s. s <|s (BND v x ss) = BND(V v s)x(s <|s ss)

#BND_s_subst;;
|- !ss n x s. BND n x(s <|s ss) = s <|s (BND(ANUM n)x ss)

#A_BND_s_subst;;
|- !a ss m x s.
    ~x IN (UNION_SET(IMAGE(FVv o ss)(FVa a))) ==>
    (A a((BND m x s) <|s ss) = A a(s <|s ss))

#IN_IMAGE_FV_DELETE;;
|- !n s ss z x.
    ~x IN (UNION_SET(IMAGE(FVv o ss)(s DELETE z))) ==>
    ~x IN (UNION_SET(IMAGE(FVv o (BND(ANUM n)z ss))s))
%

let BND_o_CONS_vsubst =
 prove_thm
  (`BND_o_CONS_vsubst`,
   "!ys xs (s:state) n v.
            ((BND n ((ys //v xs)v) s)  o 
                       ((CONS ((ys //v xs)v) ys) //v (CONS v xs))) =
              BND n v (s o (ys //v xs))",
   REWRITE_TAC[NULL_CONS_vsubst;BND_o_vsubst_THM]
  );;

% BND_o_CONS_vsubst
g "!ys xs (s:state) n v.
            ((BND n ((ys //v xs)v) s)  o 
                       ((CONS ((ys //v xs)v) ys) //v (CONS v xs))) =
              BND n v (s o (ys //v xs))";;
e(REWRITE_TAC[NULL_CONS_vsubst]);;
e(REWRITE_TAC[BND_o_vsubst_THM]);;
%



let FVa_vsubst_subst =
 prove_thm
  (`FVa_vsubst_subst`,
   "!a ys xs s. (FVa a SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (A (a <|av (ys //v xs)) s = A (a <|a (ys // xs)) s)",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[FVa_DEF;SUBSET;UNION_SUBSET]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC FVv_vsubst_subst
   THEN IMP_RES_TAC FVvs_vsubst_subst
   THEN REWRITE_ALL_TAC[a_vsubst_EQ_subst;a_subst_THM;
                        vs_vsubst_EQ_subst;vs_subst_THM;
                        v_vsubst_EQ_subst;v_subst_THM]
   THEN ASM_REWRITE_TAC[A_DEF]
   THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
   THEN (FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC)
   THEN REWRITE_ALL_TAC[s_subst_vsubst]
   THEN (USE_IMP_THEN REWRITE_THM A_BND_s_subst_var_list
         THENL [ ASM_REWRITE_TAC[]; ALL_TAC ])
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) BND_o_CONS_vsubst]
   THEN (ASM_CASES_TAC "(v:var) IN SL xs"
   THENL
     [  USE_IMP_THEN REWRITE_THM NULL_CONS_subst
             THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
        THEN REWRITE_TAC[NULL_CONS_vsubst]
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN IMP_RES_THEN (ONCE_REWRITE_THM o SYM) ABSORPTION
        THEN ASM_REWRITE_TAC[SUBSET_INSERT_DELETE];

        FIRST_ASSUM MATCH_MP_TAC
        THEN REWRITE_TAC[SET_LIST;LENGTH;DISJOINT_LIST;SUBSET_INSERT_DELETE]
        THEN IMP_RES_THEN (IMP_RES_THEN (IMP_RES_THEN
                            (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)))
                  vsubst_ELEMENTS
        THEN ASM_REWRITE_TAC[]
     ]  )
  );;

%< FVa_vsubst_subst
g "!a ys xs s. (FVa a SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (A (a <|av (ys //v xs)) s = A (a <|a (ys // xs)) s)";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN IMP_RES_TAC FVv_vsubst_subst
  THEN IMP_RES_TAC FVvs_vsubst_subst
  THEN REWRITE_ALL_TAC[a_vsubst_EQ_subst;a_subst_THM;
                       vs_vsubst_EQ_subst;vs_subst_THM;
                       v_vsubst_EQ_subst;v_subst_THM]
  THEN ASM_REWRITE_TAC[A_DEF]
);;

let st1 = get_state();;
set_state(st1);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
  );;

 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(REWRITE_ALL_TAC[s_subst_vsubst]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_CONS_vsubst]);;
 e(USE_IMP_THEN REWRITE_THM A_BND_s_subst_var_list
   THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
let st2 = get_state();;
set_state(st2);;
 
 e(ASM_CASES_TAC "(v:var) IN SL xs");;

   e(REWRITE_TAC[NULL_CONS_vsubst]);;
   e(USE_IMP_THEN REWRITE_THM NULL_CONS_subst
     THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_THEN (ONCE_REWRITE_THM o SYM) ABSORPTION);;
   e(ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[SET_LIST;LENGTH;DISJOINT_LIST;SUBSET_INSERT_DELETE]);;
   e(IMP_RES_THEN (IMP_RES_THEN (IMP_RES_THEN
                    (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)))
          vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

let st3 = get_state();;
set_state(st3);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
  );;

 e(REWRITE_ALL_TAC[s_subst_vsubst]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_CONS_vsubst]);;
 e(IMP_RES_TAC A_BND_s_subst_var_list);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC 
  );;
 e(ASM_REWRITE_TAC[]);;
let st4 = get_state();;
set_state(st4);;
 
 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(ASM_CASES_TAC "(v:var) IN SL xs");;

   e(REWRITE_TAC[NULL_CONS_vsubst]);;
   e(IMP_RES_TAC NULL_CONS_subst);;
   e(POP_TAC THEN POP_TAC THEN POP_TAC);;
   e(ASM_REWRITE_TAC[]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_TAC ABSORPTION);;
   e(POP_ASSUM (\th. REWRITE_ALL_TAC[DELETE_SUBSET;th]));;
   e(ASM_REWRITE_TAC[]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[SET_LIST;LENGTH;DISJOINT_LIST]);;
   e(REWRITE_ALL_TAC[DELETE_SUBSET]);;
   e(IMP_RES_THEN (IMP_RES_THEN (IMP_RES_THEN
                    (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)))
          vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

>%

%< FVa_vsubst_subst
g "!a ys xs s. (FVa a SUBSET SL xs) /\
         (LENGTH xs = LENGTH ys) /\ (DL xs) /\ (DL ys) ==>
         (A (a <|av (ys //v xs)) s = A (a <|a (ys // xs)) s)";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[FVa_DEF;SUBSET;UNION_SUBSET]
  THEN REPEAT STRIP_TAC
  THEN REWRITE_TAC[a_vsubst;a_subst;A_DEF]

  THEN ( (USE_IMP_THEN REWRITE_THM FVv_vsubst_subst  THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (USE_IMP_THEN REWRITE_THM FVv_vsubst_subst  THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (USE_IMP_THEN REWRITE_THM FVvs_vsubst_subst THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (USE_IMP_THEN REWRITE_THM FVvs_vsubst_subst THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)     THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)     THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
  THEN ( (FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)     THEN ASM_REWRITE_TAC[]) ORELSE ALL_TAC )
%  THEN RES_TAC
  THEN IMP_RES_TAC FVv_vsubst_subst
  THEN IMP_RES_TAC FVvs_vsubst_subst %
  THEN REWRITE_ALL_TAC[a_vsubst_EQ_subst;a_subst_THM;
                       vs_vsubst_EQ_subst;vs_subst_THM;
                       v_vsubst_EQ_subst;v_subst_THM]
  THEN ASM_REWRITE_TAC[A_DEF]
);;

let st1 = get_state();;
set_state(st1);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
  );;

 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(REWRITE_ALL_TAC[s_subst_vsubst]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_CONS_vsubst]);;
 e(USE_IMP_THEN REWRITE_THM A_BND_s_subst_var_list
   THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
let st2 = get_state();;
set_state(st2);;
 
 e(ASM_CASES_TAC "(v:var) IN SL xs");;

   e(REWRITE_TAC[NULL_CONS_vsubst]);;
   e(USE_IMP_THEN REWRITE_THM NULL_CONS_subst
     THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_THEN (ONCE_REWRITE_THM o SYM) ABSORPTION);;
   e(ASM_REWRITE_TAC[SUBSET_INSERT_DELETE]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[SET_LIST;LENGTH;DISJOINT_LIST;SUBSET_INSERT_DELETE]);;
   e(IMP_RES_THEN (IMP_RES_THEN (IMP_RES_THEN
                    (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)))
          vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

let st3 = get_state();;
set_state(st3);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
  );;

 e(REWRITE_ALL_TAC[s_subst_vsubst]);;
 e(REWRITE_TAC[(SYM o SPEC_ALL) BND_o_CONS_vsubst]);;
 e(IMP_RES_TAC A_BND_s_subst_var_list);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN 
   POP_TAC THEN POP_TAC 
  );;
 e(ASM_REWRITE_TAC[]);;
let st4 = get_state();;
set_state(st4);;
 
 e(FORALL_EQ_TAC ORELSE EXISTS_EQ_TAC);;
 e(ASM_CASES_TAC "(v:var) IN SL xs");;

   e(REWRITE_TAC[NULL_CONS_vsubst]);;
   e(IMP_RES_TAC NULL_CONS_subst);;
   e(POP_TAC THEN POP_TAC THEN POP_TAC);;
   e(ASM_REWRITE_TAC[]);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_TAC ABSORPTION);;
   e(POP_ASSUM (\th. REWRITE_ALL_TAC[DELETE_SUBSET;th]));;
   e(ASM_REWRITE_TAC[]);;

   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[SET_LIST;LENGTH;DISJOINT_LIST]);;
   e(REWRITE_ALL_TAC[DELETE_SUBSET]);;
   e(IMP_RES_THEN (IMP_RES_THEN (IMP_RES_THEN
                    (IMP_RES_TAC o CONTRAPOS o SPEC_ALL)))
          vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

>%


let v_subst_o =
 prove_thm
  (`v_subst_o`,
   "!v ss1 ss2. (v <|v (AVAR o (ss2 o ss1))) =
                ((v <|v (AVAR o ss1)) <|v (AVAR o ss2))",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_subst;o_THM]
  );;

let v_vsubst_o =
 prove_thm
  (`v_vsubst_o`,
   "!v ss1 ss2. (v <|vv (ss2 o ss1)) = ((v <|vv ss1) <|vv ss2)",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_vsubst;o_THM]
  );;


let vs_subst_o =
 prove_thm
  (`vs_subst_o`,
   "!vs ss1 ss2. (vs <|vs (AVAR o (ss2 o ss1))) =
                ((vs <|vs (AVAR o ss1)) <|vs (AVAR o ss2))",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[vs_subst;v_subst_o]
  );;

let vs_vsubst_o =
 prove_thm
  (`vs_vsubst_o`,
   "!vs ss1 ss2. (vs <|vsv (ss2 o ss1)) = ((vs <|vsv ss1) <|vsv ss2)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[vs_vsubst;v_vsubst_o]
  );;


% This is probably not true; it does not prove.
let a_subst_o =
 prove_thm
  (`a_subst_o`,
   "!a ss1 ss2. (a <|a (AVAR o (ss2 o ss1))) =
                ((a <|a (AVAR o ss1)) <|a (AVAR o ss2))",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_subst;o_THM;v_subst_o;vs_subst_o]
  );;
%

let a_vsubst_o =
 prove_thm
  (`a_vsubst_o`,
   "!a ss1 ss2. (a <|av (ss2 o ss1)) = ((a <|av ss1) <|av ss2)",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;o_THM;v_vsubst_o;vs_vsubst_o]
  );;


let v_vsubst_I =
 prove_thm
  (`v_vsubst_I`,
   "!v. (v <|vv I) = v",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_vsubst;I_THM]
  );;

let vs_vsubst_I =
 prove_thm
  (`vs_vsubst_I`,
   "!vs. (vs <|vsv I) = vs",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[vs_vsubst;v_vsubst_I]
  );;

let a_vsubst_I =
 prove_thm
  (`a_vsubst_I`,
   "!a. (a <|av I) = a",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;I_THM;v_vsubst_I;vs_vsubst_I]
  );;

let FVv_vsubst =
 prove_thm
  (`FVv_vsubst`,
   "!v ss. FVv (v <|vv ss) = IMAGE ss (FVv v)",
   INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[v_vsubst;FVv_DEF;IMAGE;IMAGE_UNION]
  );;

% FVv_vsubst
g "!v ss. FVv (v <|vv ss) = IMAGE ss (FVv v)";;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC);;
e(ASM_REWRITE_TAC[v_vsubst;FVv_DEF;IMAGE;IMAGE_UNION]);;
%

let FVvs_vsubst =
 prove_thm
  (`FVvs_vsubst`,
   "!vs ss. FVvs (vs <|vsv ss) = IMAGE ss (FVvs vs)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[vs_vsubst;FVvs_DEF;IMAGE;IMAGE_UNION;FVv_vsubst]
  );;

% FVvs_vsubst
g "!vs ss. FVvs (vs <|vsv ss) = IMAGE ss (FVvs vs)";;
e(LIST_INDUCT_TAC);;
e(ASM_REWRITE_TAC[vs_vsubst;FVvs_DEF;IMAGE;IMAGE_UNION;FVv_vsubst]);;
%



let FVa_ONE_ONE_vsubst =
 prove_thm
  (`FVa_ONE_ONE_vsubst`,
   "!a ss. ONE_ONE ss ==> (FVa (a <|av ss) = IMAGE ss (FVa a))",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;FVa_DEF;FVv_vsubst;IMAGE;IMAGE_UNION;
                        FVv_vsubst;FVvs_vsubst]
   THEN REWRITE_TAC[EXTENSION;IN_DELETE;IN_IMAGE]
   THEN REPEAT GEN_TAC
   THEN ( EQ_TAC
   THEN STRIP_TAC
   THENL
     [  EXISTS_TAC "x':var"
        THEN UNASSUME_THEN REWRITE_ALL_THM
                     "x = (ss:vsubst) x'"
        THEN IMP_RES_THEN REWRITE_ALL_THM ONE_ONE
        THEN ASM_REWRITE_TAC[];

        ASM_REWRITE_TAC[]
        THEN IMP_RES_THEN REWRITE_ALL_THM ONE_ONE
        THEN ASM_REWRITE_TAC[]
        THEN EXISTS_TAC "x':var"
        THEN ASM_REWRITE_TAC[]
     ]  )
  );;

% FVa_ONE_ONE_vsubst:
g "!a ss. ONE_ONE ss ==> (FVa (a <|av ss) = IMAGE ss (FVa a))";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN REPEAT STRIP_TAC
  THEN RES_TAC
  THEN ASM_REWRITE_TAC[a_vsubst;FVa_DEF;FVv_vsubst;IMAGE;IMAGE_UNION;
                       FVv_vsubst;FVvs_vsubst]);;

  e(REWRITE_TAC[EXTENSION]);;
  e(REPEAT GEN_TAC);;
  e(REWRITE_TAC[IN_DELETE;IN_IMAGE]);;
  e(EQ_TAC);;

    e(STRIP_TAC);;
    e(EXISTS_TAC "x':var");;
    e(UNASSUME_THEN REWRITE_ALL_THM
            "x = (ss:vsubst) x'");;
    e(IMP_RES_THEN REWRITE_ALL_THM ONE_ONE);;
    e(ASM_REWRITE_TAC[]);;

    e(STRIP_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
    e(ASM_REWRITE_TAC[]);;
    e(EXISTS_TAC "x':var");;
    e(ASM_REWRITE_TAC[]);;
%

let FVa_vsubst =
 prove_thm
  (`FVa_vsubst`,
   "!a ys xs. FVa (a <|av (ys //v xs)) = IMAGE (ys //v xs) (FVa a)",
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN ASM_REWRITE_TAC[a_vsubst;FVa_DEF;IMAGE;IMAGE_UNION;
                        FVv_vsubst;FVvs_vsubst]
   THEN REWRITE_TAC[EXTENSION;IN_DELETE;IN_IMAGE]
   THEN REPEAT GEN_TAC
   THEN ( EQ_TAC
   THEN STRIP_TAC
   THENL
     [  EXISTS_TAC "x':var"
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;vsubst_11])
                     "x = (ys //v xs)x'"
        THEN ASM_REWRITE_TAC[];

        ASM_REWRITE_TAC[vsubst_11]
        THEN EXISTS_TAC "x':var"
        THEN ASM_REWRITE_TAC[]
     ]  )
  );;

% FVa_vsubst
g "!a ys xs. FVa (a <|av (ys //v xs)) = IMAGE (ys //v xs) (FVa a)";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN ASM_REWRITE_TAC[a_vsubst;FVa_DEF;FVv_vsubst;IMAGE;IMAGE_UNION;
                       FVv_vsubst;FVvs_vsubst]);;

  e(REWRITE_TAC[EXTENSION]);;
  e(REPEAT GEN_TAC);;
  e(REWRITE_TAC[IN_DELETE;IN_IMAGE]);;
  e(EQ_TAC);;

    e(STRIP_TAC);;
    e(EXISTS_TAC "x':var");;
    e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;vsubst_11])
            "x = (ys //v xs)x'");;
    e(ASM_REWRITE_TAC[]);;

    e(STRIP_TAC);;
    e(ASM_REWRITE_TAC[vsubst_11]);;
    e(EXISTS_TAC "x':var");;
    e(ASM_REWRITE_TAC[]);;
%

let FVa_vsubst_EQ_subst =
 prove_thm
  (`FVa_vsubst_EQ_subst`,
   "!a ys xs.
     (FVa a) SUBSET (SL xs) /\ (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     (FVa (a <|av (ys //v xs)) = FVa (a <|a (ys // xs)))",
   REWRITE_TAC[SUBSET_DEF;FVa_vsubst;FVa_a_subst;EXTENSION]
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[IN_IMAGE;IN_UNION_SET;subst_BNDsR;o_THM;FVv_DEF]
   THEN EQ_TAC
   THEN STRIP_TAC
   THENL
     [  EXISTS_TAC "INSERT (x:var) EMPTY"
        THEN REWRITE_TAC[IN]
        THEN EXISTS_TAC "x':var"
        THEN RES_TAC
        THEN IMP_RES_TAC BNDsR_vsubst
        THEN ASM_REWRITE_TAC[];

        EXISTS_TAC "x':var"
        THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;IN])
                  "si = INSERT (BNDsR ys xs (I:var->var) x') EMPTY"
        THEN RES_TAC
        THEN IMP_RES_TAC BNDsR_vsubst
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% FVa_vsubst_EQ_subst
g "!a ys xs.
     (FVa a) SUBSET (SL xs) /\ (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     (FVa (a <|av (ys //v xs)) = FVa (a <|a (ys // xs)))";;
e(REWRITE_TAC[SUBSET_DEF;FVa_vsubst;FVa_a_subst;EXTENSION]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[IN_IMAGE;IN_UNION_SET;subst_BNDsR;o_THM;FVv_DEF]);;
e(EQ_TAC);;

    e(STRIP_TAC);;
    e(EXISTS_TAC "INSERT (x:var) EMPTY");;
    e(REWRITE_TAC[IN]);;
    e(EXISTS_TAC "x':var");;
    e(RES_TAC);;
    e(IMP_RES_TAC BNDsR_vsubst);;
    e(ASM_REWRITE_TAC[]);;

    e(STRIP_TAC);;
    e(EXISTS_TAC "x':var");;
    e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;IN])
                  "si = INSERT (BNDsR ys xs (I:var->var) x') EMPTY");;
    e(RES_TAC);;
    e(IMP_RES_TAC BNDsR_vsubst);;
    e(ASM_REWRITE_TAC[]);;
%

let vsubst_SUBSET =
 prove_thm
  (`vsubst_SUBSET`,
   "!s ys xs.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((IMAGE (ys //v xs) s) SUBSET (SL ys) = (s SUBSET (SL xs)))",
   REWRITE_TAC[SUBSET_DEF;IN_IMAGE]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL
     [  IMP_RES_TAC vsubst_ELEMENTS
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN POP_TAC THEN POP_TAC THEN POP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "x:var"
        THEN ASM_REWRITE_TAC[];

        RES_TAC
        THEN IMP_RES_TAC vsubst_ELEMENTS
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% vsubst_SUBSET
g "!s ys xs.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((IMAGE (ys //v xs) s) SUBSET (SL ys) = (s SUBSET (SL xs)))";;
e(REWRITE_TAC[SUBSET_DEF;IN_IMAGE]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

    e(REPEAT STRIP_TAC);;
    e(IMP_RES_TAC vsubst_ELEMENTS);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(POP_TAC THEN POP_TAC THEN POP_TAC);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(EXISTS_TAC "x:var");;
    e(ASM_REWRITE_TAC[]);;

    e(REPEAT STRIP_TAC);;
    e(RES_TAC);;
    e(IMP_RES_TAC vsubst_ELEMENTS);;
    e(ASM_REWRITE_TAC[]);;
%

let DL_MAP_vsubst =
 prove_thm
  (`DL_MAP_vsubst`,
   "!vs ys xs. DL (MAP (ys //v xs) vs) = DL vs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;DISJOINT_LIST]
   THEN REWRITE_TAC[SL_MAP;IN_IMAGE]
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN ASM_REWRITE_TAC[vsubst_11;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [  STRIP_TAC
        THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o SPEC "h:var")
        THEN ASM_REWRITE_TAC[];

        STRIP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN GEN_TAC
        THEN ASM_CASES_TAC "(h:var) = x"
        THENL
          [  POP_ASSUM (ASM_REWRITE_THM o SYM);

             ASM_REWRITE_TAC[]
          ]
     ]
  );;

% DL_MAP_vsubst
g "!vs ys xs. DL (MAP (ys //v xs) vs) = DL vs";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[MAP;DISJOINT_LIST]);;

e(REWRITE_TAC[MAP;DISJOINT_LIST]);;
e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(ASM_REWRITE_TAC[vsubst_11;DE_MORGAN_THM]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o SPEC "h:var"));;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(h:var) = x");;
    e(POP_ASSUM (ASM_REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;
%

let DISJOINT_MAP_vsubst =
 prove_thm
  (`DISJOINT_MAP_vsubst`,
   "!us vs ys xs.
      DISJOINT (SL (MAP (ys //v xs) us)) (SL (MAP (ys //v xs) vs)) =
      DISJOINT (SL us) (SL vs)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[MAP;SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN REWRITE_TAC[SL_MAP;IN_IMAGE]
   THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
   THEN REWRITE_TAC[vsubst_11;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [  STRIP_TAC
        THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o SPEC "h:var")
        THEN ASM_REWRITE_TAC[];

        STRIP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN GEN_TAC
        THEN ASM_CASES_TAC "(h:var) = x"
        THENL
          [  POP_ASSUM (ASM_REWRITE_THM o SYM);

             ASM_REWRITE_TAC[]
          ]
     ]
  );;

% DISJOINT_MAP_vsubst
g "!us vs ys xs.
      DISJOINT (SL (MAP (ys //v xs) us)) (SL (MAP (ys //v xs) vs)) =
      DISJOINT (SL us) (SL vs)";;
e(LIST_INDUCT_TAC);;
e(ASM_REWRITE_TAC[MAP;SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]);;

e(ASM_REWRITE_TAC[MAP;SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]);;
e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
e(REWRITE_TAC[vsubst_11;DE_MORGAN_THM]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[] o SPEC "h:var"));;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(h:var) = x");;
    e(POP_ASSUM (ASM_REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;
%


let EXISTS_EQUAL =
 TAC_PROOF
  (([], "!(y:*) s. (?x. (y = x) /\ x IN s) = y IN s"),
   REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL [ ALL_TAC; EXISTS_TAC "y:*" ]
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!(y:*) s. (?x. (y = x) /\ x IN s) = y IN s";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(REPEAT STRIP_TAC);;
  e(EXISTS_TAC "y:*");;
  e(ASM_REWRITE_TAC[]);;
%

let IMAGE_vsubst_SUBSET =
 prove_thm
  (`IMAGE_vsubst_SUBSET`,
   "!s t ys xs.
     (IMAGE (ys //v xs) s) SUBSET (IMAGE (ys //v xs) t) =
      s SUBSET t",
   REWRITE_TAC[SUBSET_DEF;IN_IMAGE]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
     [  REPEAT STRIP_TAC
        THEN FIRST_ASSUM (STRIP_ASSUME_TAC o SPEC "(ys //v xs)x")
        THEN REWRITE_ALL_TAC[vsubst_11;EXISTS_EQUAL]
        THEN RES_TAC;

        REPEAT STRIP_TAC
        THEN EXISTS_TAC "x':var"
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% IMAGE_vsubst_SUBSET
g "!s t ys xs.
     (IMAGE (ys //v xs) s) SUBSET (IMAGE (ys //v xs) t) =
      s SUBSET t";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(REWRITE_TAC[IN_IMAGE]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o SPEC "(ys //v xs)x"));;
  e(REWRITE_ALL_TAC[vsubst_11;EXISTS_EQUAL]);;
  e(RES_TAC);;

  e(REPEAT STRIP_TAC);;
  e(EXISTS_TAC "x':var");;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%



let FVa_vsubst_SUBSET =
 prove_thm
  (`FVa_vsubst_SUBSET`,
   "!a ys xs.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((FVa (a <|av (ys //v xs)) SUBSET SL ys) = (FVa a SUBSET SL xs))",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[FVa_vsubst]
   THEN MATCH_MP_TAC vsubst_SUBSET
   THEN ASM_REWRITE_TAC[]
  );;

% FVa_vsubst_SUBSET
g "!a ys xs.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((FVa (a <|av (ys //v xs)) SUBSET SL ys) = (FVa a SUBSET SL xs))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[FVa_vsubst]);;
e(MATCH_MP_TAC vsubst_SUBSET);;
e(ASM_REWRITE_TAC[]);;
%

let FVa_subst_SUBSET =
 prove_thm
  (`FVa_subst_SUBSET`,
   "!a ys xs.
     (FVa a SUBSET SL xs) /\ (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((FVa (a <|a (ys // xs))) SUBSET (SL ys))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC FVa_vsubst_EQ_subst
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN REWRITE_TAC[FVa_vsubst]
   THEN IMP_RES_TAC vsubst_SUBSET
  );;

% FVa_subst_SUBSET
g "!a ys xs.
     (FVa a SUBSET SL xs) /\ (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
     ((FVa (a <|a (ys // xs))) SUBSET (SL ys))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC FVa_vsubst_EQ_subst);;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(REWRITE_TAC[FVa_vsubst]);;
e(IMP_RES_TAC vsubst_SUBSET);;
%




let NULL_APPEND_vsubst =
 prove_thm
  (`NULL_APPEND_vsubst`,
   "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) ==>
     (DL xs) /\ (DL ys) /\ (v IN (SL xs)) ==>
     ((APPEND ys zs //v APPEND xs ws) v = (ys //v xs) v)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REWRITE_TAC[DISJOINT_LIST;APPEND;vsubst_THM]
   THEN ASM_CASES_TAC "(v:var) = h'"
   THEN ASM_REWRITE_TAC[BND_MATCH]
   THEN IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH
   THEN STRIP_TAC
   THEN RES_TAC
   THEN USE_IMP_TAC BND_NOT_MATCH
   THENL
     [  ALL_TAC;

        ASM_REWRITE_TAC[]
        THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
     ]
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
   THEN IMP_RES_TAC vsubst_ELEMENTS
   THEN IMP_RES_TAC IN_NOT_IN
  );;

% NULL_APPEND_vsubst
g "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) ==>
     (DL xs) /\ (DL ys) /\ (v IN (SL xs)) ==>
     ((APPEND ys zs //v APPEND xs ws) v = (ys //v xs) v)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN]);;

  e(REWRITE_TAC[SET_LIST;IN]);;
  e(REWRITE_TAC[DISJOINT_LIST;APPEND;vsubst_THM]);;
  e(ASM_CASES_TAC "(v:var) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
    e(STRIP_TAC);;
    e(RES_TAC);;
    e(USE_IMP_TAC BND_NOT_MATCH);;

      e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
      e(IMP_RES_TAC vsubst_ELEMENTS);;
      e(IMP_RES_TAC IN_NOT_IN);;

      e(ASM_REWRITE_TAC[]);;
      e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;
      e(REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
      e(IMP_RES_TAC vsubst_ELEMENTS);;
      e(IMP_RES_TAC IN_NOT_IN);;
%


let IMAGE_NULL_APPEND_vsubst =
 prove_thm
  (`IMAGE_NULL_APPEND_vsubst`,
   "!s y x z w.
     (LENGTH x = LENGTH y) /\ DL x /\ DL y /\ (s SUBSET (SL x)) ==>
     ((IMAGE (APPEND y z //v APPEND x w) s) =
      (IMAGE (y //v x) s))",
   REWRITE_TAC[SUBSET_DEF;EXTENSION]
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[IN_IMAGE]
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN EXISTS_TAC "x'':var"
   THEN RES_TAC
   THEN IMP_RES_TAC NULL_APPEND_vsubst
   THEN ASM_REWRITE_TAC[]
  );;

% IMAGE_NULL_APPEND_vsubst
g "!s y x z w.
     (LENGTH x = LENGTH y) /\ DL x /\ DL y /\ (s SUBSET (SL x)) ==>
     ((IMAGE (APPEND y z //v APPEND x w) s) =
      (IMAGE (y //v x) s))";;
e(REWRITE_TAC[SUBSET_DEF;EXTENSION]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[IN_IMAGE]);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "x'':var");;
  e(RES_TAC);;
  e(IMP_RES_TAC NULL_APPEND_vsubst);;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "x'':var");;
  e(RES_TAC);;
  e(IMP_RES_TAC NULL_APPEND_vsubst);;
  e(ASM_REWRITE_TAC[]);;
%

let MAP_NULL_APPEND_vsubst =
 prove_thm
  (`MAP_NULL_APPEND_vsubst`,
   "!v y x z w.
     (LENGTH x = LENGTH y) /\ DL x /\ DL y /\ ((SL v) SUBSET (SL x)) ==>
     ((MAP (APPEND y z //v APPEND x w) v) =
      (MAP (y //v x) v))",
   REWRITE_TAC[SUBSET_DEF]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
        THEN ASM_REWRITE_TAC[]
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN REWRITE_TAC[];

        FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% MAP_NULL_APPEND_vsubst
g "!v y x z w.
     (LENGTH x = LENGTH y) /\ DL x /\ DL y /\ ((SL v) SUBSET (SL x)) ==>
     ((MAP (APPEND y z //v APPEND x w) v) =
      (MAP (y //v x) v))";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REPEAT STRIP_TAC);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;
    e(ASM_REWRITE_TAC[]);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(REPEAT STRIP_TAC);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let NULL_APPEND_vsubst2 =
 prove_thm
  (`NULL_APPEND_vsubst2`,
   "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) ==> (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (v IN (SL ws)) ==>
     ((APPEND ys zs //v APPEND xs ws) v = (zs //v ws) v)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND]
   THEN REWRITE_TAC[SET_LIST;DISJOINT_INSERT;DISJOINT_LIST;vsubst_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC IN_NOT_IN
   THEN POP_TAC
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN SUBGOAL_THEN "((APPEND ys zs) //v (APPEND l' ws))v = (zs //v ws)v"
                     ASSUME_TAC
   THENL
     [  FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
        THEN ASM_REWRITE_TAC[];

        USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
        THEN MATCH_MP_TAC IN_NOT_IN
        THEN EXISTS_TAC "SL (zs:(var)list)"
        THEN IMP_RES_TAC vsubst_ELEMENTS
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% NULL_APPEND_vsubst2
g "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) ==> (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (v IN (SL ws)) ==>
     ((APPEND ys zs //v APPEND xs ws) v = (zs //v ws) v)";;
e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND]);;

  e(REWRITE_TAC[APPEND]);;
  e(REWRITE_TAC[SET_LIST;DISJOINT_INSERT]);;
  e(REWRITE_TAC[DISJOINT_LIST;vsubst_THM]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC IN_NOT_IN);;
  e(POP_TAC);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(SUBGOAL_THEN "((APPEND ys zs) //v (APPEND l' ws))v = (zs //v ws)v"
                 ASSUME_TAC);;

    e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
    e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
    e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
    e(MATCH_MP_TAC IN_NOT_IN);;
    e(EXISTS_TAC "SL (zs:(var)list)");;
    e(IMP_RES_TAC vsubst_ELEMENTS);;
    e(ASM_REWRITE_TAC[]);;
%

let MAP_NULL_APPEND_vsubst2 =
 prove_thm
  (`MAP_NULL_APPEND_vsubst2`,
   "!v v y x z w.
     (LENGTH x = LENGTH y) /\ (LENGTH w = LENGTH z) /\
     DISJOINT(SL x)(SL w) /\ DISJOINT(SL y)(SL z) /\
     DL w /\ DL z /\ (SL v SUBSET SL w) ==>
     ((MAP (APPEND y z //v APPEND x w) v) =
      (MAP (z //v w) v))",
   REWRITE_TAC[SUBSET_DEF]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
        THEN ASM_REWRITE_TAC[]
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN REWRITE_TAC[];

        FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% MAP_NULL_APPEND_vsubst2
g "!v v y x z w.
     (LENGTH x = LENGTH y) /\ (LENGTH w = LENGTH z) /\
     DISJOINT(SL x)(SL w) /\ DISJOINT(SL y)(SL z) /\
     DL w /\ DL z /\ (SL v SUBSET SL w) ==>
     ((MAP (APPEND y z //v APPEND x w) v) =
      (MAP (z //v w) v))";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REPEAT STRIP_TAC);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;
    e(ASM_REWRITE_TAC[]);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(REPEAT STRIP_TAC);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%




let NULL_APPEND_vsubst_INVERSE =
 prove_thm
  (`NULL_APPEND_vsubst_INVERSE`,
   "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) /\
     (DL xs) /\ (DL ys) /\ (v IN (SL ys)) ==>
     ((\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) v =
      (\y. @z. (ys //v xs)z = y) v)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_ELEMENTS_INVERSE
   THEN IMP_RES_TAC NULL_APPEND_vsubst
   THEN UNDISCH_LAST_TAC
   THEN BETA_TAC
   THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
   THEN DISCH_THEN (\th. SUBST1_TAC (SYM (SPEC_ALL th)) THEN ASSUME_TAC th)
   THEN REWRITE_TAC[vsubst_11]
   THEN ASM_REWRITE_TAC[SELECT_REFL]
  );;

% NULL_APPEND_vsubst_INVERSE
g "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) /\
     (DL xs) /\ (DL ys) /\ (v IN (SL ys)) ==>
     ((\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) v =
      (\y. @z. (ys //v xs)z = y) v)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_ELEMENTS_INVERSE);;
e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
          THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
e(IMP_RES_TAC NULL_APPEND_vsubst);;
e(UNDISCH_LAST_TAC);;
e(BETA_TAC);;
e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
e(DISCH_THEN (\th. SUBST1_TAC (SYM (SPEC_ALL th)) THEN ASSUME_TAC th));;
e(REWRITE_TAC[vsubst_11]);;
e(ASM_REWRITE_TAC[SELECT_REFL]);;
%


let IMAGE_NULL_APPEND_vsubst_INVERSE =
 prove_thm
  (`IMAGE_NULL_APPEND_vsubst_INVERSE`,
   "!s ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ (s SUBSET (SL ys)) ==>
     ((IMAGE (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) s) =
      (IMAGE (\y. @z. (ys //v xs)z = y) s))",
   REWRITE_TAC[SUBSET_DEF;EXTENSION]
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[IN_IMAGE]
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN EXISTS_TAC "x':var"
   THEN RES_TAC
   THEN IMP_RES_TAC NULL_APPEND_vsubst_INVERSE
   THEN ASM_REWRITE_TAC[]
  );;

% IMAGE_NULL_APPEND_vsubst_INVERSE
g "!s ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ (s SUBSET (SL ys)) ==>
     ((IMAGE (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) s) =
      (IMAGE (\y. @z. (ys //v xs)z = y) s))";;
e(REWRITE_TAC[SUBSET_DEF;EXTENSION]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[IN_IMAGE]);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "x':var");;
  e(RES_TAC);;
  e(IMP_RES_TAC NULL_APPEND_vsubst_INVERSE);;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(EXISTS_TAC "x':var");;
  e(RES_TAC);;
  e(IMP_RES_TAC NULL_APPEND_vsubst_INVERSE);;
  e(ASM_REWRITE_TAC[]);;
%

let MAP_NULL_APPEND_vsubst_INVERSE =
 prove_thm
  (`MAP_NULL_APPEND_vsubst_INVERSE`,
   "!vs ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ ((SL vs) SUBSET (SL ys)) ==>
     ((MAP (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) vs) =
      (MAP (\y. @z. (ys //v xs)z = y) vs))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_REPLACEMENT_INVERSE
   THEN POP_TAC THEN POP_TAC THEN POP_TAC
   THEN SUBGOAL_THEN "(SL (MAP(\y. @z. (ys //v xs)z = y)vs)) SUBSET (SL xs)"
             ASSUME_TAC
   THENL
     [  POP_ASSUM (SUBST1_TAC o SYM o AP_TERM "SL:(var)list->(var)set")
        THEN REWRITE_TAC[SL_MAP]
        THEN IMP_RES_THEN REWRITE_THM IMAGE_SUBSET;

        IMP_RES_TAC MAP_NULL_APPEND_vsubst
        THEN POP_TAC THEN POP_TAC THEN POP_TAC
        THEN UNDISCH_LAST_TAC
        THEN REWRITE_TAC[MAP_MAP_o;vsubst_INVERSE;MAP_I]
        THEN DISCH_THEN (\th.
                SUBST1_TAC (SYM (SPECL["zs:var list";"ws:var list"] th))
                THEN ASSUME_TAC th)
        THEN CONV_TAC (RATOR_CONV (REWRITE_CONV[MAP_MAP_o]))
        THEN ASM_REWRITE_TAC[o_ASSOC;vsubst_INVERSE;I_o_ID]
     ]
  );;

% MAP_NULL_APPEND_vsubst_INVERSE
g "!vs ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ ((SL vs) SUBSET (SL ys)) ==>
     ((MAP (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) vs) =
      (MAP (\y. @z. (ys //v xs)z = y) vs))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_REPLACEMENT_INVERSE);;
e(POP_TAC THEN POP_TAC THEN POP_TAC);;
e(SUBGOAL_THEN "(SL (MAP(\y. @z. (ys //v xs)z = y)vs)) SUBSET (SL xs)"
        ASSUME_TAC);;

 e(POP_ASSUM (SUBST1_TAC o SYM o AP_TERM "SL:(var)list->(var)set"));;
 e(REWRITE_TAC[SL_MAP]);;
 e(IMP_RES_THEN REWRITE_THM IMAGE_SUBSET);;

 e(IMP_RES_TAC MAP_NULL_APPEND_vsubst);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(REWRITE_TAC[MAP_MAP_o;vsubst_INVERSE;MAP_I]);;
 e(DISCH_THEN (\th.
       SUBST1_TAC (SYM (SPECL["zs:var list";"ws:var list"] th))
       THEN ASSUME_TAC th));;
 e(CONV_TAC (RATOR_CONV (REWRITE_CONV[MAP_MAP_o])));;
 e(ASM_REWRITE_TAC[o_ASSOC;vsubst_INVERSE;I_o_ID]);;
%


let NULL_APPEND_vsubst2_INVERSE =
 prove_thm
  (`NULL_APPEND_vsubst2_INVERSE`,
   "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) /\ (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (v IN (SL zs)) ==>
     ((\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) v =
      (\y. @z. (zs //v ws)z = y) v)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_ELEMENTS_INVERSE
   THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN
        POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
   THEN SUBGOAL_THEN
          "((APPEND ys zs) //v (APPEND xs ws))((\y. @z. (zs //v ws)z = y)v) =
            (zs //v ws)((\y. @z. (zs //v ws)z = y)v)"
          ASSUME_TAC
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN BETA_TAC
        THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
        THEN DISCH_THEN (\th. SUBST1_TAC (SYM (SPEC_ALL th)) THEN ASSUME_TAC th)
        THEN REWRITE_TAC[vsubst_11]
        THEN ASM_REWRITE_TAC[SELECT_REFL]
     ]
  );;

% NULL_APPEND_vsubst2_INVERSE
g "!ys xs zs ws v.
     (LENGTH xs = LENGTH ys) /\ (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (v IN (SL zs)) ==>
     ((\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) v =
      (\y. @z. (zs //v ws)z = y) v)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_ELEMENTS_INVERSE);;
e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN
  POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN
  POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
e(SUBGOAL_THEN
    "((APPEND ys zs) //v (APPEND xs ws))((\y. @z. (zs //v ws)z = y)v) =
      (zs //v ws)((\y. @z. (zs //v ws)z = y)v)"
    ASSUME_TAC);;

 e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;
 e(ASM_REWRITE_TAC[]);;

 e(UNDISCH_LAST_TAC);;
 e(BETA_TAC);;
 e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
 e(DISCH_THEN (\th. SUBST1_TAC (SYM (SPEC_ALL th)) THEN ASSUME_TAC th));;
 e(REWRITE_TAC[vsubst_11]);;
 e(ASM_REWRITE_TAC[SELECT_REFL]);;
%

let MAP_NULL_APPEND_vsubst2_INVERSE =
 prove_thm
  (`MAP_NULL_APPEND_vsubst2_INVERSE`,
   "!vs ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (SL vs SUBSET SL zs) ==>
     ((MAP (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) vs) =
      (MAP (\y. @z. (zs //v ws)z = y) vs))",
   REWRITE_TAC[SUBSET_DEF]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC NULL_APPEND_vsubst2_INVERSE
        THEN ASM_REWRITE_TAC[]
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN REWRITE_TAC[];

        FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% MAP_NULL_APPEND_vsubst2_INVERSE
g "!vs ys xs zs ws.
     (LENGTH xs = LENGTH ys) /\ (LENGTH ws = LENGTH zs) /\
     DISJOINT(SL xs)(SL ws) /\ DISJOINT(SL ys)(SL zs) /\
     DL ws /\ DL zs /\ (SL vs SUBSET SL zs) ==>
     ((MAP (\y. @z. (APPEND ys zs //v APPEND xs ws)z = y) vs) =
      (MAP (\y. @z. (zs //v ws)z = y) vs))";;
e(REWRITE_TAC[SUBSET_DEF]);;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;

  e(REWRITE_TAC[SET_LIST;IN;MAP]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REPEAT STRIP_TAC);;

    e(MATCH_MP_TAC NULL_APPEND_vsubst2_INVERSE);;
    e(ASM_REWRITE_TAC[]);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(REWRITE_TAC[]);;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
    e(REPEAT STRIP_TAC);;
    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let vsubst_REVERSE_INVERSE_LEMMA =
 TAC_PROOF
  (([],
    "!xs ys. (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
       (!x.
         (x IN SL xs) ==>
         ((\y. @z. (xs //v ys)z = y)x = (ys //v xs)x))"
   ),
   REWRITE_TAC[SYM(SPEC_ALL MAP_EXTENSION)]
   THEN REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE
   THENL
     [  ALL_TAC;
        IMP_RES_TAC vsubst_REPLACEMENT
     ]
   THEN ASM_REWRITE_TAC[]
  );;

% vsubst_REVERSE_INVERSE_LEMMA
g "!xs ys. (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys ==>
       (!x.
         (x IN SL xs) ==>
         ((\y. @z. (xs //v ys)z = y)x = (ys //v xs)x))";;
e(REWRITE_TAC[SYM(SPEC_ALL MAP_EXTENSION)]);;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM vsubst_REPLACEMENT_INVERSE);;

e(ASM_REWRITE_TAC[]);;

e(IMP_RES_TAC vsubst_REPLACEMENT);;
e(ASM_REWRITE_TAC[]);;
%

let vsubst_REVERSE_INVERSE =
 prove_thm
  (`vsubst_REVERSE_INVERSE`,
   "!xs ys x.
      (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ (x IN SL xs) ==>
      ((\y. @z. (xs //v ys)z = y)x = (ys //v xs)x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_REVERSE_INVERSE_LEMMA
  );;

% vsubst_REVERSE_INVERSE
g "!xs ys x.
      (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\ (x IN SL xs) ==>
      ((\y. @z. (xs //v ys)z = y)x = (ys //v xs)x)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_REVERSE_INVERSE_LEMMA);;
%




let vsubst_TRANS_ELEMENTS_LEMMA =
 TAC_PROOF
  (([],
    "!zs ys xs.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs ==>
      (!x. (x IN SL xs) ==>
           (((zs //v ys) o (ys //v xs)) x = (zs //v xs) x))"
   ),
   REWRITE_TAC[SYM(SPEC_ALL MAP_EXTENSION)]
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[MAP_o;o_THM]
   THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
   THENL
     [  ASM_REWRITE_TAC[];

        USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
        THENL
          [  ASM_REWRITE_TAC[];

             USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% vsubst_TRANS_ELEMENTS_LEMMA
g "!zs ys xs.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs ==>
      (!x. (x IN SL xs) ==>
           (((zs //v ys) o (ys //v xs)) x = (zs //v xs) x))";;
e(REWRITE_TAC[SYM(SPEC_ALL MAP_EXTENSION)]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[MAP_o;o_THM]);;
e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

    e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[]);;
%

let vsubst_TRANS_ELEMENTS =
 prove_thm
  (`vsubst_TRANS_ELEMENTS`,
   "!zs ys xs x.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs /\ (x IN SL xs) ==>
        (((zs //v ys) o (ys //v xs)) x = (zs //v xs) x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vsubst_TRANS_ELEMENTS_LEMMA
  );;

% vsubst_TRANS_ELEMENTS
g "!zs ys xs x.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs /\ (x IN SL xs) ==>
        (((zs //v ys) o (ys //v xs)) x = (zs //v xs) x)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vsubst_TRANS_ELEMENTS_LEMMA);;
%

let vsubst_TRANS_ELEMENTS1 =
 prove_thm
  (`vsubst_TRANS_ELEMENTS1`,
   "!zs ys xs x.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs /\ (x IN SL xs) ==>
        ((zs //v ys)((ys //v xs) x) = (zs //v xs) x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (REWRITE_RULE[o_THM] vsubst_TRANS_ELEMENTS_LEMMA)
  );;

% vsubst_TRANS_ELEMENTS1
g "!zs ys xs x.
      (LENGTH xs = LENGTH ys) /\ (LENGTH ys = LENGTH zs) /\
      DL xs /\ DL ys /\ DL zs /\ (x IN SL xs) ==>
        ((zs //v ys)((ys //v xs) x) = (zs //v xs) x)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (REWRITE_RULE[o_THM] vsubst_TRANS_ELEMENTS_LEMMA));;
%





close_theory();;

