% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains definitions of functions to create variants of
  variables, that do not occur within a specified exclusion set.
  This is necessary for the definition of proper substitution, in the
  theory "substitutions".

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `variants`;;

new_parent `bindings`;;
new_parent `variables`;;


% --------------------------------------------------------------------- %
% Need the ascii and string libraries.                                  %
% --------------------------------------------------------------------- %

loadf `ascii`;;
loadf `stringconv`;;
loadf `string_rules`;;


% Set up autoloading from ancestor theories of variants %

loadf `autoload`;;

autoload_all_defs_and_thms
     `string variables finite_sets more_finite_sets bindings`;;


loadf `semantics_tactics`;;



% =============================================================== %
% Unfortunately, the structure of strings as given in the library %
% is pretty sparse, for example, no concatenation or length       %
% operators.  So we will have to provide them ourselves!          %
% =============================================================== %

let str_cat = 
    new_recursive_definition false string_Axiom `str_cat`
      "(str_cat (``) t = t)  /\
       (str_cat (STRING a s) t = STRING a (str_cat s t))";;

let str_length = 
    new_recursive_definition false string_Axiom `str_length`
      "(str_length (``) = 0)  /\
       (str_length (STRING a s) = SUC (str_length s))";;

let str_length_THM = prove_thm
   (`str_length_THM`,
    "!x y. ~(str_length x = str_length y) ==> ~(x = y)",
    REPEAT GEN_TAC
    THEN CONTRAPOS_TAC
    THEN DISCH_TAC
    THEN ASM_REWRITE_TAC[]
   );;

%
g "!x y. ~(str_length x = str_length y) ==> ~(x = y)";;
e(REPEAT GEN_TAC);;
e(CONTRAPOS_TAC);;
e(DISCH_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let str_length_cat = prove_thm
   (`str_length_cat`,
    "!x y. str_length (str_cat x y) = (str_length x + str_length y)",
    INDUCT_THEN string_Induct ASSUME_TAC
    THEN ASM_REWRITE_TAC[str_cat;str_length;ADD]
   );;

% str_length_cat:
g "!x y. str_length (str_cat x y) = (str_length x + str_length y)";;
e(INDUCT_THEN string_Induct ASSUME_TAC);;

 e(REWRITE_TAC[str_cat;str_length;ADD]);;

 e(ASM_REWRITE_TAC[str_cat;str_length;ADD]);;
%


% =============================================================== %
% We will separate the space of names of variables into two sets, %
% based on the first character.  Variables that begin with the    %
% character LOG_CHAR will be logical variables, and all others    %
% will be program variables.  Program variables derive their      %
% values from the current state, and may change from state to     %
% state during a computation.  Logical variables never change     %
% their value during a computation, and in fact may not even be   %
% mentioned or referred to in program text, except for loop       %
% invariants and other assertion-language expressions.  This      %
% ensures that they never change!                                 %
% =============================================================== %

let LOG_CHAR =
    new_definition
    (`LOG_CHAR`,
     "LOG_CHAR = ASCII F T F T T T T F");;  % the character ^ %

% string_CONV (STRING ((snd o dest_eq o concl) LOG_CHAR) ``);; %

let mk_logical = 
    new_definition
    (`mk_logical`,
     "mk_logical s = STRING LOG_CHAR s");;

let mk_logical_THM = prove_thm
   (`mk_logical_THM`,
    "!x. ~(mk_logical x = x)",
    REWRITE_TAC[mk_logical] THEN
    INDUCT_THEN string_Induct ASSUME_TAC
    THEN REWRITE_TAC[NOT_EMPTY_STRING;STRING_11]
    THEN GEN_TAC
    THEN UNDISCH_ALL_TAC
    THEN CONTRAPOS_TAC
    THEN STRIP_TAC
    THEN ASM_REWRITE_TAC[]
   );;

%
g "!x. ~(mk_logical x = x)";;
e(REWRITE_TAC[mk_logical]);;
e(INDUCT_THEN string_Induct ASSUME_TAC);;

 e(REWRITE_TAC[NOT_EMPTY_STRING]);;

 e(REWRITE_TAC[STRING_11]);;
 e(GEN_TAC);;
 e(UNDISCH_ALL_TAC);;
 e(CONTRAPOS_TAC);;
 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let mk_logical_11 =
 prove_thm
  (`mk_logical_11`,
   "!x y. (mk_logical x = mk_logical y) = (x = y)",
   REWRITE_TAC[mk_logical;STRING_11]
  );;

%
g "!x y. (mk_logical x = mk_logical y) = (x = y)";;
e(REWRITE_TAC[mk_logical]);;
e(REWRITE_TAC[STRING_11]);;
%


let dest_mk_logical =
    new_rec_def `dest_mk_logical` string_Axiom
      "dest_mk_logical (STRING a s) = s";;

let dest_mk_logical_THM =
 prove_thm
  (`dest_mk_logical_THM`,
   "!s. dest_mk_logical (mk_logical s) = s",
   REWRITE_TAC[mk_logical;dest_mk_logical]
  );;

% dest_mk_logical_THM
g "!s. dest_mk_logical (mk_logical s) = s";;
e(REWRITE_TAC[mk_logical]);;
e(REWRITE_TAC[dest_mk_logical]);;
%


% Here we define the functions to convert a normal program variable %
% into a logical variable, and a list into a list of logical vars.  %

let logical = 
    new_recursive_definition false var_Axiom `logical`
      "logical (VAR s n) = (VAR (mk_logical s) n)";;

let logicals = 
    new_definition (`logicals`,
      "logicals = MAP logical");;

let dest_logical = 
    new_rec_def `dest_logical` var_Axiom
      "dest_logical (VAR s n) = (VAR (dest_mk_logical s) n)";;

let dest_logical_THM =
 prove_thm
  (`dest_logical_THM`,
   "!x. dest_logical (logical x) = x",
   INDUCT_THEN var_induct ASSUME_TAC
   THEN REWRITE_TAC[logical;dest_logical;dest_mk_logical_THM]
  );;

% dest_logical_THM
g "!x. dest_logical (logical x) = x";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[logical]);;
e(REWRITE_TAC[dest_logical]);;
e(REWRITE_TAC[dest_mk_logical_THM]);;
%

let logical_11 =
 prove_thm
  (`logical_11`,
   "!x y. (logical x = logical y) = (x = y)",
   INDUCT_THEN var_induct ASSUME_TAC
   THEN GEN_TAC THEN GEN_TAC
   THEN INDUCT_THEN var_induct ASSUME_TAC
   THEN GEN_TAC THEN GEN_TAC
   THEN REWRITE_TAC[logical;var_one_one;mk_logical_11]
  );;

%
g "!x y. (logical x = logical y) = (x = y)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(GEN_TAC THEN GEN_TAC);;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(GEN_TAC THEN GEN_TAC);;
e(REWRITE_TAC[logical]);;
e(REWRITE_TAC[var_one_one]);;
e(REWRITE_TAC[mk_logical_11]);;
%

let logical_NOT_IDENT =
 prove_thm
  (`logical_NOT_IDENT`,
   "!x. ~(x = logical x)",
   INDUCT_THEN var_induct ASSUME_TAC
   THEN GEN_TAC THEN GEN_TAC
   THEN REWRITE_TAC[logical;var_one_one]
   THEN REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) mk_logical_THM]
  );;

%
g "!x. ~(x = logical x)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(GEN_TAC THEN GEN_TAC);;
e(REWRITE_TAC[logical]);;
e(REWRITE_TAC[var_one_one]);;
e(REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) mk_logical_THM]);;
%

let logicals_THM =
 prove_thm
  (`logicals_THM`,
   "(logicals [] = []) /\
    (!x h. logicals (CONS h x) = (CONS (logical h) (logicals x)))",
   REWRITE_TAC[logicals;MAP]
  );;

%
g "(logicals [] = []) /\
    (!x h. logicals (CONS h x) = (CONS (logical h) (logicals x)))";;
e(REWRITE_TAC[logicals]);;
e(REWRITE_TAC[MAP]);;
%


let logicals_APPEND =
 prove_thm
  (`logicals_APPEND`,
   "!x y. logicals (APPEND x y) = APPEND (logicals x) (logicals y)",
   REWRITE_TAC[logicals;MAP_APPEND]
  );;

% logicals_APPEND
g "!x y. logicals (APPEND x y) = APPEND (logicals x) (logicals y)";;
e(REWRITE_TAC[logicals]);;
e(REWRITE_TAC[MAP_APPEND]);;
%

let logicals_11 =
 prove_thm
  (`logicals_11`,
   "!y x. (logicals x = logicals y) = (x = y)",
   REWRITE_TAC[logicals]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP]
   THENL
     [  LIST_INDUCT_TAC
        THEN REWRITE_TAC[MAP;NOT_CONS_NIL];

        GEN_TAC
        THEN LIST_INDUCT_TAC
        THEN REWRITE_TAC[MAP]
        THENL
          [  REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) NOT_CONS_NIL];

             ASM_REWRITE_TAC[CONS_11;logical_11]
          ]
     ]
  );;

%
g "!y x. (logicals x = logicals y) = (x = y)";;
e(REWRITE_TAC[logicals]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP]);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[MAP]);;

  e(REWRITE_TAC[MAP]);;
  e(REWRITE_TAC[NOT_CONS_NIL]);;

 e(REWRITE_TAC[MAP]);;
 e(GEN_TAC);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[MAP]);;
  e(REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) NOT_CONS_NIL]);;

  e(REWRITE_TAC[MAP]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REWRITE_TAC[logical_11]);;
  e(ASM_REWRITE_TAC[]);;
%

let IN_logicals_11 =
 prove_thm
  (`IN_logicals_11`,
   "!xs y. (logical y IN (SL (logicals xs))) = (y IN (SL xs))",
   REWRITE_TAC[logicals]
   THEN LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[MAP;SET_LIST;IN;logical_11]
  );;

%
g "!xs y. (logical y IN (SL (logicals xs))) = (y IN (SL xs))";;
e(REWRITE_TAC[logicals]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[MAP;SET_LIST;IN]);;

 e(REWRITE_TAC[MAP;SET_LIST;IN]);;
 e(ASM_REWRITE_TAC[logical_11]);;
%

let DL_logicals =
 prove_thm
  (`DL_logicals`,
   "!x. DL (logicals x) = DL x",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;DISJOINT_LIST]
   THEN ASM_REWRITE_TAC[IN_logicals_11]
  );;

% DL_logicals
g "!x. DL (logicals x) = DL x";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[logicals_THM;DISJOINT_LIST]);;

 e(REWRITE_TAC[logicals_THM;DISJOINT_LIST]);;
 e(ASM_REWRITE_TAC[IN_logicals_11]);;
%

let LENGTH_logicals =
 prove_thm
  (`LENGTH_logicals`,
   "!x. LENGTH (logicals x) = LENGTH x",
   REWRITE_TAC[logicals;LENGTH_MAP]
  );;

% LENGTH_logicals
g "!x. LENGTH (logicals x) = LENGTH x";;
e(REWRITE_TAC[logicals;LENGTH_MAP]);;
%



% +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ %
% +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ %



% =============================================================== %
% Variants of variables are variables with the same name, but     %
% with different numbers attached.  They are considered different %
% variables in the state.  For a clean definition, a variable     %
% itself is also considered to be a (null) variant of itself.     %
% =============================================================== %

% =============================================================== %
% An easy way to make a variant is to take an existing variable   %
% and add an integer to its "variant" number.                     %
% =============================================================== %

let mk_variant = 
    new_recursive_definition false var_Axiom `mk_variant`
      "mk_variant (VAR s n) m = (VAR s (n+m))";;

let Index_mk_variant = prove_thm
   (`Index_mk_variant`,
    "!x k. Index(mk_variant x k) = Index x + k",
    INDUCT_THEN var_induct ASSUME_TAC  THEN
    REWRITE_TAC[mk_variant;Index]
   );;

%
g "!x k. Index(mk_variant x k) = Index x + k";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant]);;
e(REWRITE_TAC[Index]);;
%

let Base_mk_variant = prove_thm
   (`Base_mk_variant`,
    "!x k. Base(mk_variant x k) = Base x",
    INDUCT_THEN var_induct ASSUME_TAC  THEN
    REWRITE_TAC[mk_variant;Base]
   );;

%
g "!x k. Base(mk_variant x k) = Base x";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant]);;
e(REWRITE_TAC[Base]);;
%


let mk_variant_ident = prove_thm
   (`mk_variant_ident`,
    "!x k. (mk_variant x k = x) = (k = 0)",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[mk_variant;var_one_one;ADD_INV_0_EQ]
   );;

%
g "!x k. (mk_variant x k = x) = (k = 0)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant;var_one_one]);;
e(REWRITE_TAC[ADD_INV_0_EQ]);;
%


let mk_variant_equal = prove_thm
   (`mk_variant_equal`,
    "!x m n. (mk_variant x m = mk_variant x n) = (m = n)",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[mk_variant;var_one_one]
    THEN ONCE_REWRITE_TAC[ADD_SYM]
    THEN REWRITE_TAC[EQ_MONO_ADD_EQ]
   );;

%
g "!x m n. (mk_variant x m = mk_variant x n) = (m = n)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant;var_one_one]);;
e(ONCE_REWRITE_TAC[ADD_SYM]);;
e(REWRITE_TAC[EQ_MONO_ADD_EQ]);;
%

let mk_variant_compose = prove_thm
   (`mk_variant_compose`,
    "!x m n. mk_variant (mk_variant x m) n = (mk_variant x (m+n))",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[mk_variant;var_one_one;ADD_ASSOC]
   );;

%
g "!x m n. mk_variant (mk_variant x m) n = (mk_variant x (m+n))";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant;var_one_one;ADD_ASSOC]);;
%


% =============================================================== %
% We would like to be able to test if a variable is a variant of  %
% another variable.                                               %
% =============================================================== %

let is_variant = 
    new_infix_definition (`is_variant`,
      "$is_variant y x =
       ((Base y = Base x) /\ (Index x <= Index y))");;

let var_is_variant = prove_thm
   (`var_is_variant`,
    "!x. x is_variant x",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[is_variant;Index;LESS_EQ_REFL]
   );;

%
g "!x. x is_variant x";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[is_variant]);;
e(REWRITE_TAC[Index]);;
e(REWRITE_TAC[LESS_EQ_REFL]);;
%

let mk_variant_is_variant = prove_thm
   (`mk_variant_is_variant`,
    "!x k. (mk_variant x k) is_variant x",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[mk_variant;is_variant;Base;Index;LESS_EQ_ADD]
   );;

%
g "!x k. (mk_variant x k) is_variant x";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant]);;
e(REWRITE_TAC[is_variant]);;
e(REWRITE_TAC[Base;Index]);;
e(REWRITE_TAC[LESS_EQ_ADD]);;
%

let is_variant_TRANS = prove_thm
   (`is_variant_TRANS`,
    "!x y z. (z is_variant y) /\ (y is_variant x) ==> (z is_variant x)",
    REWRITE_TAC[is_variant]
    THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]
    THEN IMP_RES_TAC LESS_EQ_TRANS
   );;

%
g "!x y z. (z is_variant y) /\ (y is_variant x) ==> (z is_variant x)";;
e(REWRITE_TAC[is_variant]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(IMP_RES_TAC LESS_EQ_TRANS);;
%

let is_variant_SOME_mk_variant = prove_thm
   (`is_variant_SOME_mk_variant`,
    "!x y. y is_variant x = (?k. y = mk_variant x k)",
    INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[mk_variant]
    THEN GEN_TAC  THEN GEN_TAC
    THEN INDUCT_THEN var_induct ASSUME_TAC
    THEN REWRITE_TAC[is_variant;Base;Index;var_one_one]
    THEN REPEAT GEN_TAC
    THEN EQ_TAC
    THENL
      [  REPEAT STRIP_TAC
         THEN EXISTS_TAC "n' - n"
         THEN IMP_RES_TAC SUB_ADD
         THEN ONCE_REWRITE_TAC[ADD_SYM]
         THEN ASM_REWRITE_TAC[];

         REPEAT STRIP_TAC
         THEN ASM_REWRITE_TAC[LESS_EQ_ADD]
      ]
   );;

%
g "!x y. y is_variant x = (?k. y = mk_variant x k)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[mk_variant]);;
e(GEN_TAC);;
e(GEN_TAC);;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[is_variant]);;
e(REWRITE_TAC[Base;Index]);;
e(REWRITE_TAC[var_one_one]);;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;
 e(EXISTS_TAC "n' - n");;
 e(IMP_RES_TAC SUB_ADD);;
 e(ONCE_REWRITE_TAC[ADD_SYM]);;
 e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;

  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[LESS_EQ_ADD]);;
%


% =============================================================== %
% Once we can make variants of a variable, we can make a whole    %
% set of them, of any size we like, all distinct from each other. %
% =============================================================== %

let variant_set = 
    new_prim_rec_definition
    (`variant_set`,
      "(variant_set x 0       = EMPTY)  /\
       (variant_set x (SUC k) = INSERT (mk_variant x k)
                                       (variant_set x k))");;

let IN_variant_set = prove_thm
   (`IN_variant_set`,
    "!m x y. (y IN variant_set x m)
           = (?n. (n < m) /\ (y = mk_variant x n))",
    INDUCT_TAC
    THEN ASM_REWRITE_TAC[variant_set;IN;NOT_LESS_0;LESS_THM]
    THEN REPEAT STRIP_TAC
    THEN EQ_TAC
    THENL
       [ REPEAT STRIP_TAC  THENL
         [ EXISTS_TAC "m:num"  THEN
           ASM_REWRITE_TAC[];

           EXISTS_TAC "n:num"  THEN
           ASM_REWRITE_TAC[]
         ];

         REPEAT STRIP_TAC  THENL
         [ ASM_REWRITE_TAC[];

           DISJ2_TAC  THEN
           EXISTS_TAC "n:num"  THEN
           ASM_REWRITE_TAC[]
         ]
      ]
   );;

%
g "!m x y. (y IN variant_set x m)
           = (?n. (n < m) /\ (y = mk_variant x n))";;

e(INDUCT_TAC);;
e(REWRITE_TAC[variant_set]);;
e(REWRITE_TAC[IN]);;
e(REWRITE_TAC[NOT_LESS_0]);;

e(REWRITE_TAC[variant_set]);;
e(ASM_REWRITE_TAC[IN]);;
e(REWRITE_TAC[LESS_THM]);;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

e(REPEAT STRIP_TAC);;

 e(EXISTS_TAC "m:num");;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "n:num");;
 e(ASM_REWRITE_TAC[]);;

e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(DISJ2_TAC);;
 e(EXISTS_TAC "n:num");;
 e(ASM_REWRITE_TAC[]);;
%


let CARD_variant_set = prove_thm
   (`CARD_variant_set`,
    "!m x. CARD (variant_set x m) = m",
    INDUCT_TAC
    THEN REWRITE_TAC[variant_set;CARD_DEF]
    THEN REWRITE_TAC[IN_variant_set;mk_variant_equal]
    THEN GEN_TAC
    THEN COND_CASES_TAC
    THENL [POP_ASSUM STRIP_ASSUME_TAC
           THEN IMP_RES_TAC EQ_SYM
           THEN IMP_RES_TAC LESS_NOT_EQ;

           ASM_REWRITE_TAC[]
          ]
   );;

%
g "!m x. CARD (variant_set x m) = m";;
e(INDUCT_TAC);;
e(REWRITE_TAC[variant_set]);;
e(REWRITE_TAC[CARD_DEF]);;

e(REWRITE_TAC[variant_set]);;
e(REWRITE_TAC[CARD_DEF]);;
e(REWRITE_TAC[IN_variant_set]);;
e(REWRITE_TAC[mk_variant_equal]);;
e(GEN_TAC);;
e(COND_CASES_TAC);;

e(POP_ASSUM STRIP_ASSUME_TAC);;
e(IMP_RES_TAC EQ_SYM);;
e(IMP_RES_TAC LESS_NOT_EQ);;

e(ASM_REWRITE_TAC[]);;
%


% =================================================================== %
% We need to be able to detect when a variable is a variant of        %
% another variable, and when it is the *minimum* such variant,        %
% subject to the condition that it not be a member of a given set of  %
% forbidden variables.                                                %
% =================================================================== %


% ======================================================= %
% Here finally is the definition of the variant function. %
% "variant x s" is a string with x as its prefix, but     %
% with some number appended to x.  This variant is        %
% guaranteed to not be in s, and to be the smallest       %
% such variant (with the least index appended).           %
% ======================================================= %

let variant_EXISTS =
    TAC_PROOF
    (([], "!x s. ?y.
              (y IN variant_set x (SUC(CARD s)) /\ ~y IN s)
              /\ !z. z IN variant_set x (SUC(CARD s)) /\ ~z IN s ==>
                     (Index y <= Index z)"),
     REWRITE_TAC[SYM (SPEC_ALL IN_DIFF);SYM (SPEC_ALL SET_MINIMUM)]
     THEN REWRITE_TAC[MEMBER_NOT_EMPTY;SYM (SPEC_ALL CARD_EQ_0)]
     THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
     THEN REPEAT GEN_TAC
     THEN MATCH_MP_TAC LESS_NOT_EQ
     THEN MATCH_MP_TAC LESS_CARD_DIFF
     THEN REWRITE_TAC[CARD_variant_set;LESS_SUC_REFL]
    );;

%
g("!x s. ?y.
              (y IN variant_set x (SUC(CARD s)) /\ ~y IN s)
              /\ !z. z IN variant_set x (SUC(CARD s)) /\ ~z IN s ==>
                     (Index y <= Index z)");;

e(REWRITE_TAC[SYM (SPEC_ALL IN_DIFF)]);;
e(REWRITE_TAC[SYM (SPEC_ALL SET_MINIMUM)]);;
e(REWRITE_TAC[MEMBER_NOT_EMPTY]);;
e(REWRITE_TAC[SYM (SPEC_ALL CARD_EQ_0)]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC LESS_NOT_EQ);;
e(MATCH_MP_TAC LESS_CARD_DIFF);;
e(REWRITE_TAC[CARD_variant_set;LESS_SUC_REFL]);;
%



let variant =
    let th1 = CONV_RULE SKOLEM_CONV variant_EXISTS in
    new_specification `variant` [`constant`,`variant`] th1;;



%
variant = 
|- !x s.
    ((variant x s) IN (variant_set x(SUC(CARD s))) /\
     ~(variant x s) IN s) /\
    (!z.
      z IN (variant_set x(SUC(CARD s))) /\ ~z IN s ==>
      (Index(variant x s)) <= (Index z))
%


% We may want to use any of these three properties specifically. %
% The next three corollaries select each property.               %


let variant_in_variant_set = prove_thm
   (`variant_in_variant_set`,
    "!x s. (variant x s) IN (variant_set x (SUC(CARD s)))",
    REWRITE_TAC[variant]
   );;

%
g "!x s. (variant x s) IN (variant_set x (SUC(CARD s)))";;
e(REWRITE_TAC[variant]);;
%


let variant_not_in_set = prove_thm
   (`variant_not_in_set`,
    "!x s. ~(variant x s IN s)",
    REWRITE_TAC[variant]
   );;

%
variant_not_in_set
g "!x s. ~(variant x s IN s)";;
e(REWRITE_TAC[variant]);;
%

let variant_minimum = prove_thm
   (`variant_minimum`,
    "!x s y. y IN (variant_set x (SUC(CARD s))) /\ ~y IN s ==>
                     (Index(variant x s)) <= (Index y)",
    REWRITE_TAC[variant]
   );;

%
variant_minimum:
g "!x s y. y IN (variant_set x (SUC(CARD s))) /\ ~y IN s ==>
                     (Index(variant x s)) <= (Index y)";;
e(REWRITE_TAC[variant]);;
%




let variant_is_variant = prove_thm
   (`variant_is_variant`,
    "!x s. (variant x s) is_variant x",
    REPEAT GEN_TAC
    THEN STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[IN_variant_set]
                                      variant_in_variant_set))
    THEN ASM_REWRITE_TAC[mk_variant_is_variant]
   );;

%
g("!x s. (variant x s) is_variant x");;
e(REPEAT GEN_TAC);;
e(STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[IN_variant_set]
                               variant_in_variant_set)));;
e(ASM_REWRITE_TAC[mk_variant_is_variant]);;
%



% Now we wish to express the variant definition more simply,   %
% by just saying that the variant selected is just a variant,  %
% without referring to any variant-sets.                       %


let variant_DEF = prove_thm
   (`variant_DEF`,
    "!x s.
              ((variant x s) is_variant x /\ ~(variant x s) IN s)
              /\ !z. z is_variant x /\ ~z IN s ==>
                     (Index (variant x s) <= Index z)",
    REPEAT GEN_TAC
    THEN REWRITE_TAC[variant;variant_is_variant]
    THEN STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[IN_variant_set]
                                      variant_in_variant_set))
    THEN REWRITE_TAC[is_variant_SOME_mk_variant]
    THEN REPEAT STRIP_TAC
    THEN DISJ_CASES_TAC (SPECL ["k:num";"SUC(CARD (s:var set))"] LESS_CASES)
    THENL
      [  MATCH_MP_TAC (REWRITE_RULE[IN_variant_set] variant_minimum)
         THEN ASM_REWRITE_TAC[]
         THEN EXISTS_TAC "k:num"
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[Index_mk_variant]
         THEN ONCE_REWRITE_TAC[ADD_SYM]
         THEN REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]
         THEN IMP_RES_TAC LESS_IMP_LESS_OR_EQ
         THEN IMP_RES_TAC LESS_EQ_TRANS
      ]
   );;

%
g("!x s.
              ((variant x s) is_variant x /\ ~(variant x s) IN s)
              /\ !z. z is_variant x /\ ~z IN s ==>
                     (Index (variant x s) <= Index z)");;

e(REPEAT GEN_TAC);;
e(REWRITE_TAC[variant;variant_is_variant]);;
e(STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[IN_variant_set]
                               variant_in_variant_set)));;
e(REWRITE_TAC[is_variant_SOME_mk_variant]);;
e(REPEAT STRIP_TAC);;
e(DISJ_CASES_TAC (SPECL ["k:num";"SUC(CARD (s:var set))"] LESS_CASES));;

  e(MATCH_MP_TAC (REWRITE_RULE[IN_variant_set] variant_minimum));;
  e(ASM_REWRITE_TAC[]);;
  e(EXISTS_TAC "k:num");;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[Index_mk_variant]);;
  e(ONCE_REWRITE_TAC[ADD_SYM]);;
  e(REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]);;
  e(IMP_RES_TAC LESS_IMP_LESS_OR_EQ);;
  e(IMP_RES_TAC LESS_EQ_TRANS);;
%


let variant_minimum_DEF = prove_thm
   (`variant_minimum_DEF`,
    "!x s y. y is_variant x /\ ~y IN s ==>
                     (Index(variant x s) <= Index y)",
    REWRITE_TAC[variant_DEF]
   );;



let Base_variant = prove_thm
   (`Base_variant`,
    "!x s. Base (variant x s) = Base x",
    REPEAT GEN_TAC
    THEN STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[is_variant]
                                      variant_is_variant))
   );;

%
g("!x s. Base (variant x s) = Base x");;
e(REPEAT GEN_TAC);;
e(STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[is_variant]
                               variant_is_variant)));;
%

let Index_variant = prove_thm
   (`Index_variant`,
    "!x s. Index x <= Index (variant x s)",
    REPEAT GEN_TAC
    THEN STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[is_variant]
                                      variant_is_variant))
   );;

%
g("!x s. Index x <= Index (variant x s)");;
e(REPEAT GEN_TAC);;
e(STRIP_ASSUME_TAC (SPEC_ALL (REWRITE_RULE[is_variant]
                               variant_is_variant)));;
%


let variant_EMPTY = prove_thm
   (`variant_EMPTY`,
    "!x. variant x EMPTY = x",
    GEN_TAC
    THEN STRIP_ASSUME_TAC (SPECL ["x:var";"EMPTY:var set"]
               (REWRITE_RULE[is_variant] variant_is_variant))
    THEN STRIP_ASSUME_TAC (SPEC "x:var" var_is_variant)
    THEN IMP_RES_TAC (REWRITE_RULE[IN]
               (SPECL ["x:var";"EMPTY:var set";"x:var"] variant_minimum_DEF))
    THEN IMP_RES_TAC LESS_EQUAL_ANTISYM
    THEN POP_TAC
    THEN ONCE_REWRITE_TAC[SYM (SPEC_ALL Base_Index)]
    THEN ASM_REWRITE_TAC[]
   );;

%
g("!x. variant x EMPTY = x");;

e(GEN_TAC);;
e(STRIP_ASSUME_TAC (SPECL ["x:var";"EMPTY:var set"] (REWRITE_RULE[is_variant]
                               variant_is_variant)));;
e(STRIP_ASSUME_TAC (SPEC "x:var" var_is_variant));;
e(IMP_RES_TAC (REWRITE_RULE[IN] (SPECL ["x:var";"EMPTY:var set";"x:var"]
                               variant_minimum_DEF)));;
e(IMP_RES_TAC LESS_EQUAL_ANTISYM);;
e(POP_TAC);;
e(ONCE_REWRITE_TAC[SYM (SPEC_ALL Base_Index)]);;
e(ASM_REWRITE_TAC[]);;
%



let LESS_EQ_NOT_EQ = prove_thm
   (`LESS_EQ_NOT_EQ`,
    "!m n. m <= n /\ ~(m = n) ==> (m+1) <= n",
    REWRITE_TAC[SYM(SPEC_ALL ADD1);SYM(SPEC_ALL LESS_EQ)]
    THEN REWRITE_TAC[LESS_OR_EQ]
    THEN REPEAT GEN_TAC
    THEN STRIP_TAC
    THEN RES_TAC
   );;

%
g "!m n. m <= n /\ ~(m = n) ==> (m+1) <= n";;
e(REWRITE_TAC[SYM(SPEC_ALL ADD1);SYM(SPEC_ALL LESS_EQ)]);;
e(REWRITE_TAC[LESS_OR_EQ]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(RES_TAC);;
%

let SET_IN_OUT =
    TAC_PROOF
    (([], "!(x:*) y s. (x IN s) /\ ~(y IN s) ==> ~(x = y)"),
     REPEAT STRIP_TAC
     THEN UNDISCH_TAC "(x:*) IN s"
     THEN ASM_REWRITE_TAC[]
    );;

%
g "!(x:*) y s. (x IN s) /\ ~(y IN s) ==> ~(x = y)";;
e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "(x:*) IN s");;
e(ASM_REWRITE_TAC[]);;
%

let variant_not_ident = prove_thm
   (`variant_not_ident`,
    "!x s. (x IN s) ==> ~(x = variant x s)",
    REPEAT GEN_TAC
    THEN STRIP_TAC
    THEN MATCH_MP_TAC SET_IN_OUT
    THEN EXISTS_TAC "s:var set"
    THEN ASM_REWRITE_TAC[variant_not_in_set]
   );;

%
g "!x s. (x IN s) ==> ~(x = variant x s)";;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(MATCH_MP_TAC SET_IN_OUT);;
e(EXISTS_TAC "s:var set");;
e(ASM_REWRITE_TAC[variant_not_in_set]);;
%


let Index_variant_not_ident = prove_thm
   (`Index_variant_not_ident`,
    "!x s. (x IN s) ==> ~(Index x = Index (variant x s))",
    REPEAT STRIP_TAC
    THEN ASSUME_TAC (SYM(SPEC_ALL Base_variant))
    THEN IMP_RES_TAC VAR_EQ
    THEN IMP_RES_TAC variant_not_ident
   );;

%
g "!x s. (x IN s) ==> ~(Index x = Index (variant x s))";;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (SYM(SPEC_ALL Base_variant)));;
e(IMP_RES_TAC VAR_EQ);;
e(IMP_RES_TAC variant_not_ident);;
%

let variant_mk_variant_is_variant = prove_thm
   (`variant_mk_variant_is_variant`,
    "!x k s. (variant (mk_variant x k) s) is_variant x",
    REPEAT GEN_TAC
    THEN MATCH_MP_TAC is_variant_TRANS
    THEN EXISTS_TAC "mk_variant x k"
    THEN REWRITE_TAC[variant_is_variant;mk_variant_is_variant]
   );;

%
g("!x k s. (variant (mk_variant x k) s) is_variant x");;
e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC is_variant_TRANS);;
e(EXISTS_TAC "mk_variant x k");;
e(REWRITE_TAC[variant_is_variant;mk_variant_is_variant]);;
%

let variant_mk_variant_not_ident = prove_thm
   (`variant_mk_variant_not_ident`,
    "!x s. ~(variant (mk_variant x 1) s = x)",
    REPEAT GEN_TAC
    THEN ONCE_REWRITE_TAC[SYM(SPEC_ALL Base_Index)]
    THEN REWRITE_TAC[var_one_one;Base_variant;Base_mk_variant]
    THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
    THEN MATCH_MP_TAC LESS_NOT_EQ
    THEN REWRITE_TAC[LESS_EQ;ADD1]
    THEN REWRITE_TAC[SYM(SPEC_ALL Index_mk_variant);Index_variant]
   );;

%
g("!x s. ~(variant (mk_variant x 1) s = x)");;
e(REPEAT GEN_TAC);;
e(ONCE_REWRITE_TAC[SYM(SPEC_ALL Base_Index)]);;
e(REWRITE_TAC[var_one_one]);;
e(REWRITE_TAC[Base_variant;Base_mk_variant]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(MATCH_MP_TAC LESS_NOT_EQ);;
e(REWRITE_TAC[LESS_EQ;ADD1]);;
e(REWRITE_TAC[SYM(SPEC_ALL Index_mk_variant)]);;
e(REWRITE_TAC[Index_variant]);;
%


let variant_THM = prove_thm
   (`variant_THM`,
    "!x s. variant x s = (x IN s => variant (mk_variant x 1) s  |  x)",
    REPEAT STRIP_TAC
    THEN COND_CASES_TAC
    THEN MATCH_MP_TAC VAR_EQ
    THEN REWRITE_TAC[Base_variant;Base_mk_variant]
    THEN MATCH_MP_TAC LESS_EQUAL_ANTISYM
    THENL
      [  STRIP_TAC
         THEN MATCH_MP_TAC variant_minimum_DEF
         THEN REWRITE_TAC[variant_not_in_set;is_variant;
                               variant_mk_variant_is_variant]
         THEN REWRITE_TAC[Base_variant;Base_mk_variant;Index_mk_variant]
         THEN MATCH_MP_TAC LESS_EQ_NOT_EQ
         THEN REWRITE_TAC[Index_variant]
         THEN IMP_RES_TAC Index_variant_not_ident;

         STRIP_TAC
         THENL
           [  
              MATCH_MP_TAC variant_minimum_DEF
              THEN ASM_REWRITE_TAC[var_is_variant];

              REWRITE_TAC[Index_variant]
           ]
      ]
   );;

%
g("!x s. variant x s = (x IN s => variant (mk_variant x 1) s  |  x)");;

e(REPEAT STRIP_TAC);;
e(COND_CASES_TAC);;

 e(MATCH_MP_TAC VAR_EQ);;
 e(REWRITE_TAC[Base_variant;Base_mk_variant]);;
 e(MATCH_MP_TAC LESS_EQUAL_ANTISYM);;
 e(STRIP_TAC);;

  e(MATCH_MP_TAC variant_minimum_DEF);;
  e(REWRITE_TAC[variant_not_in_set]);;
  e(REWRITE_TAC[variant_mk_variant_is_variant]);;

  e(MATCH_MP_TAC variant_minimum_DEF);;
  e(REWRITE_TAC[variant_not_in_set]);;
  e(REWRITE_TAC[is_variant]);;
  e(REWRITE_TAC[Base_variant;Base_mk_variant;Index_mk_variant]);;
  e(MATCH_MP_TAC LESS_EQ_NOT_EQ);;
  e(REWRITE_TAC[Index_variant]);;
  e(IMP_RES_TAC Index_variant_not_ident);;

 e(MATCH_MP_TAC VAR_EQ);;
 e(REWRITE_TAC[Base_variant;Base_mk_variant]);;
 e(MATCH_MP_TAC LESS_EQUAL_ANTISYM);;
 e(STRIP_TAC);;

  e(MATCH_MP_TAC variant_minimum_DEF);;
  e(ASM_REWRITE_TAC[var_is_variant]);;

  e(REWRITE_TAC[Index_variant]);;
%


let variant_ident =
 prove_thm
  (`variant_ident`,
   "!x s. ~x IN s ==> (variant x s = x)",
    REPEAT STRIP_TAC THEN
    ONCE_REWRITE_TAC[variant_THM] THEN
    ASM_REWRITE_TAC[]
  );;

%
g "!x s. ~x IN s ==> (variant x s = x)";;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[variant_THM]);;
e(ASM_REWRITE_TAC[]);;
%


let variant_DELETE =
 prove_thm
  (`variant_DELETE`,
   "!x s. variant x (s DELETE x) = x",
    REPEAT GEN_TAC THEN
    MATCH_MP_TAC variant_ident THEN
    REWRITE_TAC[IN_DELETE]
  );;

%
g "!x s. variant x (s DELETE x) = x";;
e(REPEAT GEN_TAC);;
e(MATCH_MP_TAC variant_ident);;
e(REWRITE_TAC[IN_DELETE]);;
%


let variant_logical_NOT_IDENT =
 prove_thm
  (`variant_logical_NOT_IDENT`,
   "!x t. ~(x = variant (logical x) t)",
   INDUCT_THEN var_induct ASSUME_TAC
   THEN REPEAT GEN_TAC
   THEN ONCE_REWRITE_TAC[SYM(SPEC "variant (logical (VAR s n)) t" Base_Index)]
   THEN REWRITE_TAC[logical;Base_variant;Base;var_one_one;DE_MORGAN_THM]
   THEN REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) mk_logical_THM]
  );;

%
g "!x t. ~(x = variant (logical x) t)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REPEAT GEN_TAC);;
e(ONCE_REWRITE_TAC[SYM(SPEC "variant (logical (VAR s n)) t" Base_Index)]);;
e(REWRITE_TAC[logical;Base_variant;Base]);;
e(REWRITE_TAC[var_one_one;DE_MORGAN_THM]);;
e(DISJ1_TAC);;
e(REWRITE_TAC[(NOT_EQ_SYM o SPEC_ALL) mk_logical_THM]);;
%

let variant_increment =
 prove_thm
  (`variant_increment`,
   "!x s.
      (x IN s) ==> (variant x s = variant (mk_variant x 1) s)",
   REPEAT STRIP_TAC
   THEN ASSUME_TAC (SPECL["x:var";"s:(var)set"] variant_THM)
   THEN ASM_REWRITE_TAC[]
  );;

% variant_increment
g "!x s.
      (x IN s) ==> (variant x s = variant (mk_variant x 1) s)";;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (SPECL["x:var";"s:(var)set"] variant_THM));;
e(ASM_REWRITE_TAC[]);;
%



% We define the function 'variants', to take a list of variables  %
% and form a list of variants of the variables in the list, with  %
% the provision that the list produced should have no duplicates. %


let variants = 
    new_list_rec_def `variants`
      "(variants NIL s  =  NIL)  /\
       (variants (CONS x xs) s  =
            let x' = variant x s in
              (CONS x' (variants xs (INSERT x' s))))";;

% Alternative definition of variants:

let variants = 
    new_list_rec_def `variants`
      "(variants NIL s  =  NIL)  /\
       (variants (CONS x xs) s  =
            let xs' = variants xs s in
              (CONS (variant x (SL xs' UNION s)) xs')";;
%

let variants_THM =
 prove_thm
  (`variants_THM`,
   "(variants NIL s  =  NIL)  /\
    (variants (CONS x xs) s  =
         (CONS (variant x s) (variants xs (INSERT (variant x s) s))))",
   REWRITE_TAC[variants]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REFL_TAC
  );;

%
g "(variants NIL s  =  NIL)  /\
   (variants (CONS x xs) s  =
         (CONS (variant x s) (variants xs (INSERT (variant x s) s))))";;
e(REWRITE_TAC[variants]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REFL_TAC);;
%

let NOT_IN_variants_INSERT =
 prove_thm
  (`NOT_IN_variants_INSERT`,
   "!xs y s. ~(y IN SL (variants xs (INSERT y s)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[variants_THM;SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THENL
     [  ASSUME_TAC (SPECL ["h:var";"INSERT (y:var) s"] variant_not_in_set)
        THEN DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th;IN])
        THEN CHECK_TAC;

        ONCE_REWRITE_TAC[INSERT_COMM]
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!xs y s. ~(y IN SL (variants xs (INSERT y s)))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[IN]);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[IN]);;
 e(REWRITE_TAC[DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;

  e(ASSUME_TAC (SPECL ["h:var";"INSERT (y:var) s"] variant_not_in_set));;
  e(DISCH_THEN (\th. REWRITE_ALL_TAC[SYM th;IN]));;
  e(CHECK_TAC);;

  e(ONCE_REWRITE_TAC[INSERT_COMM]);;
  e(ASM_REWRITE_TAC[]);;
%

let variants_APPEND =
 prove_thm
  (`variants_APPEND`,
   "!x y s.
    variants (APPEND x y) s  =
        APPEND (variants x s) (variants y (SL(variants x s) UNION s))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[variants_THM;SET_LIST;APPEND;UNION]
   THEN ASM_REWRITE_TAC[CONS_11;variant_not_in_set]
   THEN ONCE_REWRITE_TAC[UNION_COMM]
   THEN REWRITE_TAC[UNION;NOT_IN_variants_INSERT]
  );;

%
g "!x y s.
    variants (APPEND x y) s  =
        APPEND (variants x s) (variants y (SL(variants x s) UNION s))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[UNION]);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[UNION]);;
 e(REWRITE_TAC[variant_not_in_set]);;
 e(ASM_REWRITE_TAC[]);;
 e(ONCE_REWRITE_TAC[UNION_COMM]);;
 e(REWRITE_TAC[UNION]);;
 e(REWRITE_TAC[NOT_IN_variants_INSERT]);;
%

let DISJOINT_variants =
 prove_thm
  (`DISJOINT_variants`,
   "!x s. (DISJOINT (SL (variants x s)) s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[variants_THM;SET_LIST;DISJOINT_EMPTY]
   THEN REWRITE_TAC[DISJOINT_INSERT;variant_not_in_set]
   THEN REPEAT GEN_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o
              ONCE_REWRITE_RULE[DISJOINT_SYM] o
              REWRITE_RULE[DISJOINT_INSERT] o
              ONCE_REWRITE_RULE[DISJOINT_SYM] o
              SPEC "INSERT (variant h s) s")
  );;

%
g "!x s. (DISJOINT (SL (variants x s)) s)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[DISJOINT_EMPTY]);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[DISJOINT_INSERT]);;
 e(REWRITE_TAC[variant_not_in_set]);;
 e(REPEAT GEN_TAC);;
 e(POP_ASSUM (STRIP_ASSUME_TAC o
              ONCE_REWRITE_RULE[DISJOINT_SYM] o
              REWRITE_RULE[DISJOINT_INSERT] o
              ONCE_REWRITE_RULE[DISJOINT_SYM] o
              SPEC "INSERT (variant h s) s"));;
%



let DL_variants =
 prove_thm
  (`DL_variants`,
   "!x s. DL (variants x s)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[variants_THM;DISJOINT_LIST]
   THEN REWRITE_TAC[NOT_IN_variants_INSERT]
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!x s. DL (variants x s)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[DISJOINT_LIST]);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[DISJOINT_LIST]);;
 e(REWRITE_TAC[NOT_IN_variants_INSERT]);;
 e(ASM_REWRITE_TAC[]);;
%

let LENGTH_variants =
 prove_thm
  (`LENGTH_variants`,
   "!x s. LENGTH (variants x s) = LENGTH x",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[variants_THM;LENGTH]
  );;

%
g "!x s. (LENGTH (variants x s)) = (LENGTH x)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[variants_THM]);;

 e(REWRITE_TAC[variants_THM]);;
 e(REWRITE_TAC[LENGTH]);;
 e(ASM_REWRITE_TAC[]);;
%



let NOT_IN_variants =
 prove_thm
  (`NOT_IN_variants`,
   "!x y s. y IN s ==> ~(y IN SL (variants x s))",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN ASSUME_TAC (SPEC_ALL DISJOINT_variants)
   THEN IMP_RES_TAC IN_DISJOINT_IMP
  );;

% NOT_IN_variants
g "!x y s. y IN s ==> ~(y IN SL (variants x s))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(ASSUME_TAC (SPEC_ALL DISJOINT_variants));;
e(IMP_RES_TAC IN_DISJOINT_IMP);;
%


let DISJOINT_variants_UNION =
 prove_thm
  (`DISJOINT_variants_UNION`,
   "!x s t.
     DISJOINT(SL(variants x (s UNION t)))s /\
     DISJOINT(SL(variants x (s UNION t)))t",
   REPEAT GEN_TAC
   THEN (ASSUME_TAC o SPECL["x:(var)list";"(s:(var)set) UNION t"])
             DISJOINT_variants
   THEN UNDISCH_ALL_TAC
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_UNION]
  );;

% DISJOINT_variants_UNION
g "!x s t.
     DISJOINT(SL(variants x (s UNION t)))s /\
     DISJOINT(SL(variants x (s UNION t)))t";;
e(REPEAT GEN_TAC);;
e((ASSUME_TAC o SPECL["x:(var)list";"(s:(var)set) UNION t"])
        DISJOINT_variants);;
e(UNDISCH_ALL_TAC);;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_UNION]);;
%


let DISJOINT_variants_APPEND =
 prove_thm
  (`DISJOINT_variants_APPEND`,
    "!x a b.
     DISJOINT (SL (variants x (SL (APPEND a b)))) (SL a) /\
     DISJOINT (SL (variants x (SL (APPEND a b)))) (SL b) ",
   REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION]
  );;

% DISJOINT_variants_APPEND
g "!x a b.
     DISJOINT (SL (variants x (SL (APPEND a b)))) (SL a) /\
     DISJOINT (SL (variants x (SL (APPEND a b)))) (SL b) ";;
e(REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION]);;
%



let DISJOINT_variants_UNION3 =
 prove_thm
  (`DISJOINT_variants_UNION3`,
    "!x s t u.
      DISJOINT (SL(variants x (s UNION (t UNION u)))) s /\
      DISJOINT (SL(variants x (s UNION (t UNION u)))) t /\
      DISJOINT (SL(variants x (s UNION (t UNION u)))) u /\
      DISJOINT s (SL(variants x (s UNION (t UNION u)))) /\
      DISJOINT t (SL(variants x (s UNION (t UNION u)))) /\
      DISJOINT u (SL(variants x (s UNION (t UNION u))))",
   REPEAT GEN_TAC
   THEN (ASSUME_TAC o SPECL["x:(var)list";"(s:(var)set) UNION (t UNION u)"])
              DISJOINT_variants
   THEN UNDISCH_ALL_TAC
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_UNION]
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN ASM_REWRITE_TAC[]
  );;

% DISJOINT_variants_UNION3
g "!x s t u.
      DISJOINT (SL(variants x (s UNION (t UNION u)))) s /\
      DISJOINT (SL(variants x (s UNION (t UNION u)))) t /\
      DISJOINT (SL(variants x (s UNION (t UNION u)))) u /\
      DISJOINT s (SL(variants x (s UNION (t UNION u)))) /\
      DISJOINT t (SL(variants x (s UNION (t UNION u)))) /\
      DISJOINT u (SL(variants x (s UNION (t UNION u))))";;

e(REPEAT GEN_TAC);;
e((ASSUME_TAC o SPECL["x:(var)list";"(s:(var)set) UNION (t UNION u)"])
        DISJOINT_variants);;
e(UNDISCH_ALL_TAC);;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_UNION]);;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(ASM_REWRITE_TAC[]);;
%




close_theory();;
