% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definition of the well-formedness predicates
  and the proofs of theorems about the behavior of programming-language
  constructs which sastisfy these well-formedness predicates.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `well_formed`;;

new_parent `progress`;;


% --------------------------------------------------------------------- %
% Need the string, finite_sets, and ind_defs libraries.                 %
% --------------------------------------------------------------------- %
loadt `load_string`;;
load_string();;

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of well_formed %

loadf `autoload`;;

autoload_all_defs_and_thms
  `variables more_finite_sets bindings variants
   assert_syntax assert_semantics prog_syntax prog_semantics
   substitutions var_substitutions prog_substitutions
   free_variables translations progress`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;




% ============================================================ %
% Define an operator in the assertion language to specify that %
% two lists of variables have equal values.                    %
% ============================================================ %


let mk_eqs = 
    new_list_rec_definition (`mk_eqs`,
      "(mk_eqs (NIL) ys = ATRUE)  /\
       (mk_eqs (CONS x xs) ys = 
             ((ys = []) => ATRUE
                        |  (AAND (AEQ (AVAR x) (AVAR (HD ys)))
                                      (mk_eqs xs (TL ys)))))");;


let mk_eqs_THM =
 prove_thm
  (`mk_eqs_THM`,
   "!xs ys y x. (mk_eqs [] ys = ATRUE) /\
                (mk_eqs xs [] = ATRUE) /\
                (mk_eqs (CONS x xs) (CONS y ys) =
                            (AAND (AEQ (AVAR x) (AVAR y)) (mk_eqs xs ys)))",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_eqs;NOT_CONS_NIL;HD;TL]
  );;

% mk_eqs_THM
g "!xs ys y x. (mk_eqs [] ys = ATRUE) /\
               (mk_eqs xs [] = ATRUE) /\
               (mk_eqs (CONS x xs) (CONS y ys) =
                           (AAND (AEQ (AVAR x) (AVAR y)) (mk_eqs xs ys)))";;
e(DOUBLE_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[mk_eqs;NOT_CONS_NIL;HD;TL]);;
%



let mk_eqs_subst_BND =
 prove_thm
  (`mk_eqs_subst_BND`,
   "!y z x e. (LENGTH z = LENGTH y) ==> ~(x IN FVa(mk_eqs y z)) ==>
              ((mk_eqs y z <|a (BND e x ss)) = (mk_eqs y z <|a ss))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_eqs_THM;a_subst;v_subst]
   THEN REWRITE_TAC[FVa_DEF;FVv_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN REWRITE_TAC[aexp_one_one]
   THEN EVERY_ASSUM (\th. (ASSUME_TAC o NOT_EQ_SYM)th ? ALL_TAC)
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN RES_TAC
   THEN POP_ASSUM REWRITE_THM
  );;

% mk_eqs_subst_BND
g "!y z x e. (LENGTH z = LENGTH y) ==> ~(x IN FVa(mk_eqs y z)) ==>
             ((mk_eqs y z <|a (BND e x ss)) = (mk_eqs y z <|a ss))";;

e(LENGTH_LIST_INDUCT_TAC);;

e(REWRITE_TAC[mk_eqs_THM;a_subst]);;

e(REWRITE_TAC[mk_eqs_THM;a_subst]);;
e(REWRITE_TAC[v_subst]);;
e(REWRITE_TAC[FVa_DEF;FVv_DEF]);;
e(REWRITE_TAC[IN;IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[aexp_one_one]);;
e(EVERY_ASSUM (\th. (ASSUME_TAC o NOT_EQ_SYM)th ? ALL_TAC));;
e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
e(RES_TAC);;
e(POP_ASSUM REWRITE_THM);;
%


let FVa_mk_eqs =
 prove_thm
  (`FVa_mk_eqs`,
   "!ys xs. (LENGTH xs = LENGTH ys) ==>
            (FVa (mk_eqs ys xs) = (SL ys UNION SL xs))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_eqs_THM;FVa_DEF;SET_LIST;UNION]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[FVv_DEF;IN;EXTENSION;IN_UNION]
   THEN GEN_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%
g "!ys xs. (LENGTH xs = LENGTH ys) ==>
           (FVa (mk_eqs ys xs) = (SL ys UNION SL xs))";;
e(LENGTH_LIST_INDUCT_TAC);;

e(REWRITE_TAC[mk_eqs_THM;FVa_DEF;SET_LIST;UNION]);;

e(REWRITE_TAC[mk_eqs_THM;FVa_DEF;SET_LIST;UNION]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[FVv_DEF;IN;EXTENSION]);;
e(REWRITE_TAC[IN;IN_UNION]);;
e(GEN_TAC);;
e(EQ_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);;
%

let FVa_mk_eqs_logicals =
 prove_thm
  (`FVa_mk_eqs_logicals`,
   "!x. FVa(mk_eqs(logicals x)x) = (SL x UNION SL (logicals x))",
   GEN_TAC
   THEN USE_IMP_THEN REWRITE_THM FVa_mk_eqs
   THENL [ REWRITE_TAC[logicals;LENGTH_MAP];
           REWRITE_TAC[ISPECL["SL (logicals x)";"SL (x:(var)list)"] UNION_COMM]
         ]
  );;

%
g "!x. FVa(mk_eqs(logicals x)x) =
        (SL x UNION SL (logicals x))";;
e(GEN_TAC);;
e(USE_IMP_THEN REWRITE_THM FVa_mk_eqs);;

 e(REWRITE_TAC[logicals;LENGTH_MAP]);;

 e(REWRITE_TAC[ISPECL["SL (logicals x)";"SL (x:(var)list)"] UNION_COMM]);;
%

let FVa_mk_eqs_variants_logicals =
 prove_thm
  (`FVa_mk_eqs_variants_logicals`,
   "!x s. FVa(mk_eqs(variants (logicals x) s) x) =
          (SL x UNION SL (variants (logicals x) s))",
   REPEAT GEN_TAC
   THEN USE_IMP_THEN REWRITE_THM FVa_mk_eqs
   THENL [ REWRITE_TAC[LENGTH_variants;logicals;LENGTH_MAP];
           REWRITE_TAC[ISPECL["SL (variants (logicals x) s)";
                              "SL (x:(var)list)"] UNION_COMM]
         ]
  );;

%
g "!x s. FVa(mk_eqs(variants (logicals x) s) x) =
          (SL x UNION SL (variants (logicals x) s))";;
e(REPEAT GEN_TAC);;
e(USE_IMP_THEN REWRITE_THM FVa_mk_eqs);;

 e(REWRITE_TAC[LENGTH_variants;logicals;LENGTH_MAP]);;

 e(REWRITE_TAC[ISPECL["SL (variants (logicals x) s)";
                      "SL (x:(var)list)"] UNION_COMM]);;
%



let A_mk_eqs_THM =
 prove_thm
  (`A_mk_eqs_THM`,
   "!y x s. (LENGTH x = LENGTH y) ==>
            (A (mk_eqs y x) s = (MAP s y = MAP s x))",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_eqs_THM;A_DEF]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[MAP;V_DEF;CONS_11]
  );;

%
g "!y x s. (LENGTH x = LENGTH y) ==>
           (A (mk_eqs y x) s = (MAP s y = MAP s x))";;
e(LENGTH_LIST_INDUCT_TAC);;

e(REWRITE_TAC[mk_eqs_THM;A_DEF]);;

e(REWRITE_TAC[mk_eqs_THM;A_DEF]);;
e(REWRITE_TAC[MAP;V_DEF;CONS_11]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%



% ========================================================== %
% We now define the well-formedness predicates for variable  %
% names, expressions, expression lists, boolean expressions, %
% commands, and environments.                                %
% The well-formedness predicate for environments involves    %
% the CSPEC construct, so we introduce it here, after        %
% the theory on specifications.                              %
%                                                            %
%   c is well-formed in environment r iff                    %
%     if c = CALL p xs es,                                   %
%     where r p = vars,vals,glbs,pre,post,calls,rec,c,       %
%     then the lengths of xs and vars are the same, and      %
%     the lengths of es and vals are the same.               %
%   (Since all variables are integers, this is all the       %
%    type-checking needed.)                                  %
% ========================================================== %


let WFs = 
    new_recursive_definition false string_Axiom `WFs`
      "(WFs (``) = T)  /\
       (WFs (STRING a s) = ~(a = LOG_CHAR))";;  % ~(a = `^`) %

let WFx = 
    new_recursive_definition false var_Axiom `WFx`
      "WFx (VAR s n) = WFs s";;

%
let NOT_WFx = 
    new_definition `NOT_WFx`
      "NOT_WFx x = ~WFx x";;
%

let WFxs = 
    new_list_rec_definition (`WFxs`,
      "(WFxs (NIL) = T)  /\
       (WFxs (CONS x xs) = WFx x /\ WFxs xs)");;

let NOT_WFxs = 
    new_list_rec_definition (`NOT_WFxs`,
      "(NOT_WFxs (NIL) = T)  /\
       (NOT_WFxs (CONS x xs) = ~WFx x /\ NOT_WFxs xs)");;

let WFe = 
    new_recursive_definition false exp_Axiom `WFe`
      "(WFe (NUM n)       = T)      /\
       (WFe (PVAR x)      = WFx x)  /\
       (WFe (INC x)       = WFx x)  /\
       (WFe (PLUS e1 e2)  = WFe e1 /\ WFe e2)  /\
       (WFe (MINUS e1 e2) = WFe e1 /\ WFe e2)  /\
       (WFe (MULT e1 e2)  = WFe e1 /\ WFe e2)";;

let WFes = 
    new_list_rec_definition (`WFes`,
      "(WFes (NIL) = T)  /\
       (WFes (CONS e es) = WFe e /\ WFes es)");;

let WFb = 
    new_recursive_definition false bexp_Axiom `WFb`
      "(WFb (EQ   e1 e2)    = WFe e1 /\ WFe e2)  /\
       (WFb (LESS e1 e2)    = WFe e1 /\ WFe e2)  /\
       (WFb (LLESS es1 es2) = WFes es1 /\ WFes es2)  /\
       (WFb (AND  b1 b2)    = WFb b1 /\ WFb b2)  /\
       (WFb (OR   b1 b2)    = WFb b1 /\ WFb b2)  /\
       (WFb (NOT b)         = WFb b)";;

let WFc = 
    new_recursive_definition false cmd_Axiom `WFc`
      "(WFc (SKIP) (g:prog_env) (r:env) = T)  /\
       (WFc (ABORT) g r          = T)  /\
       (WFc (ASSIGN x e) g r     = (WFx x /\ WFe e))  /\
       (WFc (SEQ c1 c2) g r      = (WFc c1 g r /\ WFc c2 g r))  /\
       (WFc (IF b c1 c2) g r     = (WFb b /\ WFc c1 g r /\ WFc c2 g r))  /\
       (WFc (WHILE a pr b c) g r =
             (WFb b /\ WFc c g r /\
              (?v x.
                (pr = ALESS v (AVAR x)) /\ ~WFx x /\
                ~(x IN (FVa a UNION FVv v)) /\
                (!p. ~(x IN (FVa (g p)))))))  /\
       (WFc (CALL p xs es) g r   = (WFxs xs /\ WFes es /\
                                 (LENGTH ((FST o r) p) = LENGTH xs) /\
                                 (LENGTH ((FST o SND o r) p) = LENGTH es) /\
                                 (DL (APPEND xs ((FST o SND o SND o r) p)))))";;

%<  future version: includes the possibility of progress conditions that use <<:

let WFc = 
    new_recursive_definition false cmd_Axiom `WFc`
      "(WFc (SKIP) (r:env)     = T)  /\
       (WFc (ABORT) r          = T)  /\
       (WFc (ASSIGN x e) r     = (WFx x /\ WFe e))  /\
       (WFc (SEQ c1 c2) r      = (WFc c1 r /\ WFc c2 r))  /\
       (WFc (IF b c1 c2) r     = (WFb b /\ WFc c1 r /\ WFc c2 r))  /\
       (WFc (WHILE a pr b c) r = (WFb b /\ WFc c r /\
             ((?v x. (pr = ALESS v (AVAR x)) /\
                     ~(x IN (FVa a UNION FVv v UNION FVb b UNION FVc c r))) \/
              (?vs xs. (pr = ALLESS vs (MAP AVAR xs)) /\
                     DISJOINT (SL xs)
                              (FVa a UNION FVvs vs UNION FVb b UNION FVc c r)))
         ))  /\
       (WFc (CALL p xs es) r   = (WFxs xs /\ WFes es /\
                                 (LENGTH ((FST o r) p) = LENGTH xs) /\
                                 (LENGTH ((FST o SND o r) p) = LENGTH es) /\
                                 (DL (APPEND xs ((FST o SND o SND o r) p)))))";;
>%


let WFcalls = 
    new_definition (`WFcalls`,
      "WFcalls (calls:prog_env) (r:env) =
          (!p.
             let (vars,vals,glbs,pre,post,calls',rec,c) = r p in
             let x = APPEND (APPEND vars vals) glbs in
             (?z. FVa (calls p) SUBSET (SL (APPEND x (logicals z)))))");;


let WFproc_syntax = 
    new_definition (`WFproc_syntax`,
      "!vars vals glbs pre post calls rec c.
          WFproc_syntax (vars,vals,glbs,pre,post,calls,rec,c) (r:env) =
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            ( WFxs x  /\
              DL x  /\
              WFc c calls r  /\
             (GVc c r  SUBSET  SL glbs) /\
             (FVc c r  SUBSET  SL x)    /\
             (FVa pre  SUBSET  SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))  /\
             (!s.
                let (vars',vals',glbs',pre',post',calls',rec',c') = r s in
                let x' = APPEND (APPEND vars' vals') glbs' in
                FVa (calls s) SUBSET (SL (APPEND x' x0)))  /\
             ((rec = AFALSE) \/
              (?v y. (rec = ALESS v (AVAR y)) /\
                     ~WFx y /\ (FVv v SUBSET SL x)))
            )");;

let WFproc_pre = 
    new_definition (`WFproc_pre`,
      "WFproc_pre p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            PRE_PROGRESS(pre, c, r)");;

let WFproc_calls = 
    new_definition (`WFproc_calls`,
      "WFproc_calls p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            BODY_PROGRESS(AAND (mk_eqs x0 x) pre, c, calls, r)");;


let induct_pre =
    new_rec_def `induct_pre` aexp_Axiom
      "(induct_pre AFALSE = ATRUE) /\
       (induct_pre (ALESS e1 e2) = (AEQ e1 e2))";;

let WFproc_rec = 
    new_definition (`WFproc_rec`,
      "WFproc_rec p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r)");;

let WFproc_partial = 
    new_definition (`WFproc_partial`,
      "WFproc_partial p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            PCSPEC(AAND (mk_eqs x0 x) pre, c, post, r)");;

let WFproc_total = 
    new_definition (`WFproc_total`,
      "WFproc_total p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            TCSPEC(AAND (mk_eqs x0 x) pre, c, r)");;

let WFproc_correct = 
    new_definition (`WFproc_correct`,
      "WFproc_correct p r =
            let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
            let x   = APPEND (APPEND vars vals) glbs in
            let x0  = logicals x in
            CSPEC(AAND (mk_eqs x0 x) pre, c, post, r)");;

let WFproc_correct_aspects =
 prove_thm
  (`WFproc_correct_aspects`,
   "!p r. WFproc_correct p r = WFproc_partial p r /\ WFproc_total p r",
   REWRITE_TAC[WFproc_correct;WFproc_partial;WFproc_total]
   THEN REPEAT GEN_TAC
   THEN ONCE_REWRITE_TAC[ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[CSPEC]
  );;

%
g "!p r. WFproc_correct p r = WFproc_partial p r /\ WFproc_total p r";;
e(REWRITE_TAC[WFproc_correct;WFproc_partial;WFproc_total]);;
e(REPEAT GEN_TAC);;
e(ONCE_REWRITE_TAC[ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[CSPEC]);;
%

let WFprocp = 
    new_definition (`WFprocp`,
      "WFprocp p r =
            WFproc_syntax (r p) r  /\
            WFproc_partial p r");;


let WFproc = 
    new_definition (`WFproc`,
      "WFproc p r =
            WFproc_syntax (r p) r  /\
            WFproc_pre     p r  /\
            WFproc_calls   p r  /\
            WFproc_rec     p r  /\
            WFproc_correct p r");;

let WFenv_syntax = 
    new_definition (`WFenv_syntax`,
      "WFenv_syntax r =
         !p. WFproc_syntax (r p) r");;

let WFenv_pre = 
    new_definition (`WFenv_pre`,
      "WFenv_pre r =
         !p. WFproc_pre p r");;

let WFenv_calls = 
    new_definition (`WFenv_calls`,
      "WFenv_calls r =
         !p. WFproc_calls p r");;

let WFenv_rec = 
    new_definition (`WFenv_rec`,
      "WFenv_rec r =
         !p. WFproc_rec p r");;

let WFenv_partial = 
    new_definition (`WFenv_partial`,
      "WFenv_partial r =
         !p. WFproc_partial p r");;

let WFenv_total = 
    new_definition (`WFenv_total`,
      "WFenv_total r =
         !p. WFproc_total p r");;

let WFenv_correct = 
    new_definition (`WFenv_correct`,
      "WFenv_correct r =
         !p. WFproc_correct p r");;

let WFenv_correct_aspects =
 prove_thm
  (`WFenv_correct_aspects`,
   "!r. WFenv_correct r = WFenv_partial r /\ WFenv_total r",
   REWRITE_TAC[WFenv_correct;WFenv_partial;WFenv_total]
   THEN REWRITE_TAC[WFproc_correct_aspects]
   THEN CONV_TAC (DEPTH_CONV FORALL_AND_CONV)
   THEN REWRITE_TAC[]
  );;

%
g "!r. WFenv_correct r = WFenv_partial r /\ WFenv_total r";;
e(REWRITE_TAC[WFenv_correct;WFenv_partial;WFenv_total]);;
e(REWRITE_TAC[WFproc_correct_aspects]);;
e(CONV_TAC (DEPTH_CONV FORALL_AND_CONV));;
e(REWRITE_TAC[]);;
%

let WFenvp = 
    new_definition (`WFenvp`,
      "WFenvp r =
         !p. WFprocp p r");;


let WFenvp_aspects =
 prove_thm
  (`WFenvp_aspects`,
   "!r. WFenvp r = 
          WFenv_syntax r /\
          WFenv_partial r",
   REWRITE_TAC[WFenvp;WFprocp]
   THEN REWRITE_TAC[WFenv_syntax;WFenv_partial]
   THEN CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV)
   THEN REWRITE_TAC[]
  );;

% WFenvp_aspects
g "!r. WFenvp r = 
          WFenv_syntax r /\
          WFenv_partial r";;
e(REWRITE_TAC[WFenvp;WFprocp]);;
e(REWRITE_TAC[WFenv_syntax;WFenv_partial]);;
e(CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV));;
e(REWRITE_TAC[]);;
%


let WFenvp_WFenv_syntax =
 prove_thm
  (`WFenvp_WFenv_syntax`,
   "!r. WFenvp r ==> WFenv_syntax r",
   REWRITE_TAC[WFenvp_aspects]
   THEN REPEAT STRIP_TAC
  );;

%
g "!r. WFenvp r ==> WFenv_syntax r";;
e(REWRITE_TAC[WFenvp_aspects]);;
e(REPEAT STRIP_TAC);;
%


let WFenv = 
    new_definition (`WFenv`,
      "WFenv r =
         !p. WFproc p r");;

let WFenv_aspects =
 prove_thm
  (`WFenv_aspects`,
   "!r. WFenv r = 
          WFenv_syntax r /\
          WFenv_pre r /\
          WFenv_calls r /\
          WFenv_rec r /\
          WFenv_correct r",
   REWRITE_TAC[WFenv;WFproc]
   THEN REWRITE_TAC[WFenv_syntax;WFenv_pre;WFenv_calls;WFenv_rec;WFenv_correct]
   THEN CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV)
   THEN REWRITE_TAC[]
  );;

% WFenv_aspects
g "!r. WFenv r = 
          WFenv_syntax r /\
          WFenv_pre r /\
          WFenv_calls r /\
          WFenv_rec r /\
          WFenv_correct r";;
e(REWRITE_TAC[WFenv;WFproc]);;
e(REWRITE_TAC[WFenv_syntax;WFenv_pre;WFenv_calls;WFenv_rec;WFenv_correct]);;
e(CONV_TAC (TOP_DEPTH_CONV FORALL_AND_CONV));;
e(REWRITE_TAC[]);;
%

let WFenv_WFenv_syntax =
 prove_thm
  (`WFenv_WFenv_syntax`,
   "!r. WFenv r ==> WFenv_syntax r",
   REWRITE_TAC[WFenv_aspects]
   THEN REPEAT STRIP_TAC
  );;

% WFenv_WFenv_syntax
g "!r. WFenv r ==> WFenv_syntax r";;
e(REWRITE_TAC[WFenv_aspects]);;
e(REPEAT STRIP_TAC);;
%

let WFenv_WFenvp =
 prove_thm
  (`WFenv_WFenvp`,
   "!r. WFenv r ==> WFenvp r",
   REWRITE_TAC[WFenv_aspects;WFenv_correct_aspects;WFenvp_aspects]
   THEN GEN_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFenv_WFenvp
g "!r. WFenv r ==> WFenvp r";;
e(REWRITE_TAC[WFenv_aspects]);;
e(REWRITE_TAC[WFenv_correct_aspects]);;
e(REWRITE_TAC[WFenvp_aspects]);;
e(GEN_TAC);;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%



let WFd = 
    new_recursive_definition false decl_Axiom `WFd`
      "(WFd (PROC p vars vals glbs pre post calls rec c) r =
           WFproc_syntax (vars,vals,glbs,pre,post,calls,rec,c) r)  /\
       (WFd (DSEQ d1 d2) r  = (WFd d1 r /\ WFd d2 r))  /\
       (WFd (DEMPTY) r = T)";;

let EMPTY_PROGRESS_ENV =
    new_def `EMPTY_PROGRESS_ENV`
      "g0 = (\p. ATRUE) : prog_env";;

let WFp = 
    new_recursive_definition false prog_Axiom `WFp`
      "WFp (PROG d c) = let r = mkenv d r0 in
                        (WFd d r /\ WFc c g0 r)";;


let WFsubst = 
    new_definition (`WFsubst`,
      "!ss. WFsubst ss =
         (!x. WFx (ss x) = WFx x) /\
         (!x. WFx x ==> (ss (logical x) = logical (ss x)))");;


let WFcsubst = 
    new_definition (`WFcsubst`,
      "!c ss. WFcsubst c r ss =
         (!x. WFx (ss x) = WFx x) /\
         (ONE_ONE ss) /\ (ONTO ss) /\
         (!x. x IN GVc c r ==> (ss x = x))");;



let WFcsubst_IMP_biject =
 prove_thm
  (`WFcsubst_IMP_biject`,
   "!c r ss. WFcsubst c r ss ==> ONE_ONE ss /\ ONTO ss",
   REWRITE_TAC[WFcsubst]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFcsubst_IMP_biject
g "!c r ss. WFcsubst c r ss ==> ONE_ONE ss /\ ONTO ss";;
e(REWRITE_TAC[WFcsubst]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%


let WFenv_EMPTY_ENV =
 prove_thm
  (`WFenv_EMPTY_ENV`,
   "WFenv r0",
   REWRITE_TAC[WFenv;WFproc]
   THEN GEN_TAC
   THEN REPEAT CONJ_TAC
   THEN REWRITE_TAC[EMPTY_ENV]
   THENL
    [ REWRITE_TAC[WFproc_syntax]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[APPEND]
      THEN REWRITE_TAC[logicals_THM]
      THEN REWRITE_TAC[WFxs]
      THEN REWRITE_TAC[DISJOINT_LIST]
      THEN REWRITE_TAC[WFc]
      THEN REWRITE_TAC[GVc;SUBSET]
      THEN REWRITE_TAC[FVc_DEF;SUBSET]
      THEN REWRITE_TAC[FVa_DEF;SUBSET];

      REWRITE_TAC[WFproc_pre]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[PRE_PROGRESS]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[CPROGRESS;A_DEF];

      REWRITE_TAC[WFproc_calls]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[APPEND]
      THEN REWRITE_TAC[logicals_THM]
      THEN REWRITE_TAC[BODY_PROGRESS]
      THEN REWRITE_TAC[CPROGRESS;A_DEF];

      REWRITE_TAC[WFproc_rec]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[REC_PROGRESS]
      THEN REWRITE_TAC[PATH_PROGRESS;A_DEF];

      REWRITE_TAC[WFproc_correct]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[APPEND]
      THEN REWRITE_TAC[logicals_THM]
      THEN REWRITE_TAC[CSPEC]
      THEN REWRITE_TAC[PCSPEC;TCSPEC;A_DEF]
    ]
  );;

% WFenv_EMPTY_ENV
g "WFenv r0";;
e(REWRITE_TAC[WFenv;WFproc]);;
e(GEN_TAC);;
e(REPEAT CONJ_TAC);;

 e(REWRITE_TAC[EMPTY_ENV;WFproc_syntax]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[logicals_THM]);;
 e(REWRITE_TAC[WFxs]);;
 e(REWRITE_TAC[DISJOINT_LIST]);;
 e(REWRITE_TAC[WFc]);;
 e(REWRITE_TAC[GVc;SUBSET]);;
 e(REWRITE_TAC[FVc_DEF;SUBSET]);;
 e(REWRITE_TAC[FVa_DEF;SUBSET]);;

 e(REWRITE_TAC[EMPTY_ENV;WFproc_pre]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[PRE_PROGRESS]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[CPROGRESS;A_DEF]);;

 e(REWRITE_TAC[EMPTY_ENV;WFproc_calls]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[logicals_THM]);;
 e(REWRITE_TAC[BODY_PROGRESS]);;
 e(REWRITE_TAC[CPROGRESS;A_DEF]);;

 e(REWRITE_TAC[EMPTY_ENV;WFproc_rec]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[REC_PROGRESS]);;
 e(REWRITE_TAC[PATH_PROGRESS;A_DEF]);;

 e(REWRITE_TAC[EMPTY_ENV;WFproc_correct]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[logicals_THM]);;
 e(REWRITE_TAC[CSPEC]);;
 e(REWRITE_TAC[PCSPEC;TCSPEC;A_DEF]);;
%


let NOT_WFs_dest_mk_logical =
 prove_thm
  (`NOT_WFs_dest_mk_logical`,
   "!s. ~WFs s ==> (mk_logical (dest_mk_logical s) = s)",
   INDUCT_THEN string_Induct ASSUME_TAC
   THEN REWRITE_TAC[WFs]
   THEN REWRITE_TAC[dest_mk_logical]
   THEN REWRITE_TAC[mk_logical]
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% NOT_WFs_dest_mk_logical
g "!s. ~WFs s ==> (mk_logical (dest_mk_logical s) = s)";;
e(INDUCT_THEN string_Induct ASSUME_TAC);;

 e(REWRITE_TAC[WFs]);;

 e(REWRITE_TAC[WFs]);;
 e(REWRITE_TAC[dest_mk_logical]);;
 e(REWRITE_TAC[mk_logical]);;
 e(REPEAT STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let NOT_WFx_dest_logical =
 prove_thm
  (`NOT_WFx_dest_logical`,
   "!x. ~WFx x ==> (logical (dest_logical x) = x)",
   INDUCT_THEN var_induct ASSUME_TAC
   THEN REWRITE_TAC[WFx]
   THEN REWRITE_TAC[dest_logical]
   THEN REWRITE_TAC[logical]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC NOT_WFs_dest_mk_logical
   THEN ASM_REWRITE_TAC[]
  );;

% NOT_WFx_dest_logical
g "!x. ~WFx x ==> (logical (dest_logical x) = x)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[WFx]);;
e(REWRITE_TAC[dest_logical]);;
e(REWRITE_TAC[logical]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC NOT_WFs_dest_mk_logical);;
e(ASM_REWRITE_TAC[]);;
%


let WFs_NOT_WFs =
 prove_thm
  (`WFs_NOT_WFs`,
   "!x y. WFs x /\ ~WFs y ==> ~(y = x)",
   INDUCT_THEN string_Induct ASSUME_TAC
   THENL
    [ REWRITE_TAC[WFs]
      THEN INDUCT_THEN string_Induct ASSUME_TAC
      THEN REWRITE_TAC[WFs;NOT_EMPTY_STRING];

      GEN_TAC
      THEN INDUCT_THEN string_Induct ASSUME_TAC
      THEN REWRITE_TAC[NOT_STRING_EMPTY]
      THEN REWRITE_TAC[STRING_11;DE_MORGAN_THM]
      THEN REWRITE_TAC[WFs]
      THEN REPEAT STRIP_TAC
      THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% WFs_NOT_WFs:
g "!x y. WFs x /\ ~WFs y ==> ~(y = x)";;
e(INDUCT_THEN string_Induct ASSUME_TAC);;

  e(REWRITE_TAC[WFs]);;
  e(INDUCT_THEN string_Induct ASSUME_TAC);;

    e(REWRITE_TAC[WFs]);;

    e(REWRITE_TAC[NOT_EMPTY_STRING]);;

  e(GEN_TAC);;
  e(INDUCT_THEN string_Induct ASSUME_TAC);;

    e(REWRITE_TAC[NOT_STRING_EMPTY]);;

    e(REWRITE_TAC[STRING_11]);;
    e(REWRITE_TAC[WFs]);;
    e(REWRITE_TAC[DE_MORGAN_THM]);;
    e(REPEAT STRIP_TAC);;
    e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
    e(ASM_REWRITE_TAC[]);;
%

let WFx_NOT_WFx =
 prove_thm
  (`WFx_NOT_WFx`,
   "!x y. WFx x /\ ~WFx y  ==>  ~(x = y)",
   REPEAT STRIP_TAC
   THEN POP_ASSUM REWRITE_ALL_THM
   THEN RES_TAC
  );;

% WFx_NOT_WFx
g "!x y. WFx x /\ ~WFx y  ==>  ~(x = y)";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM REWRITE_ALL_THM);;
e(RES_TAC);;
%

let WFx_NOT_WFxs =
 prove_thm
  (`WFx_NOT_WFxs`,
   "!ys x. WFx x /\ NOT_WFxs ys ==> ~(x IN SL ys)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;SET_LIST;IN]
   THEN REWRITE_TAC[DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFx_NOT_WFx
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFx_NOT_WFxs:
g "!ys x. WFx x /\ NOT_WFxs ys ==> ~(x IN SL ys)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[NOT_WFxs;SET_LIST;IN]);;

  e(REWRITE_TAC[NOT_WFxs;SET_LIST;IN]);;
  e(REWRITE_TAC[DE_MORGAN_THM]);;
  e(REPEAT GEN_TAC THEN STRIP_TAC);;
  e(IMP_RES_TAC WFx_NOT_WFx);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%

let WFxs_NOT_WFx =
 prove_thm
  (`WFxs_NOT_WFx`,
   "!xs y. WFxs xs /\ ~WFx y ==> ~(y IN SL xs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;SET_LIST;IN;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFx_NOT_WFx
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFxs_NOT_WFx:
g "!xs y. WFxs xs /\ ~WFx y ==> ~(y IN SL xs)";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[WFxs;SET_LIST;IN]);;

e(REWRITE_TAC[WFxs;SET_LIST;IN]);;
e(REWRITE_TAC[DE_MORGAN_THM]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WFx_NOT_WFx);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let WFxs_NOT_WFxs =
 prove_thm
  (`WFxs_NOT_WFxs`,
   "!x y. WFxs x /\ NOT_WFxs y ==> DISJOINT(SL x)(SL y)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC WFx_NOT_WFxs
   THEN ASM_REWRITE_TAC[]
  );;

% WFxs_NOT_WFxs
g "!x y. WFxs x /\ NOT_WFxs y ==> DISJOINT(SL x)(SL y)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;DISJOINT_EMPTY]);;

 e(REWRITE_TAC[WFxs;SET_LIST;DISJOINT_INSERT]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(IMP_RES_TAC WFx_NOT_WFxs);;
 e(ASM_REWRITE_TAC[]);;
%

let WFxs_IS_WFx =
 prove_thm
  (`WFxs_IS_WFx`,
   "!xs x. WFxs xs /\ (x IN SL xs) ==> WFx x",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;SET_LIST;IN]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
     [ ASM_REWRITE_TAC[];

       RES_TAC
     ] 
  );;

% WFxs_IS_WFx
g "!xs x. WFxs xs /\ (x IN SL xs) ==> WFx x";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REWRITE_TAC[WFxs]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;

  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
%

let NOT_WFxs_IS_NOT_WFx =
 prove_thm
  (`NOT_WFxs_IS_NOT_WFx`,
   "!xs x. NOT_WFxs xs /\ (x IN SL xs) ==> ~WFx x",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;SET_LIST;IN]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
     [ ASM_REWRITE_TAC[];

       RES_TAC
     ] 
  );;

% ALL_IN_NOT_WFxs
g "!xs y. NOT_WFxs xs /\ (y IN SL xs) ==> ~WFx y";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REWRITE_TAC[NOT_WFxs]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;

  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
%

let SUBSET_WFxs_NOT_WFxs =
 prove_thm
  (`SUBSET_WFxs_NOT_WFxs`,
   "!s xs ys. (s SUBSET SL xs) /\ WFxs xs /\ NOT_WFxs ys ==>
              DISJOINT (SL ys) s",
   ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN SET_INDUCT_TAC
   THEN REWRITE_TAC[SUBSET;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFxs_IS_WFx
   THEN IMP_RES_TAC WFx_NOT_WFxs
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% SUBSET_WFxs_NOT_WFxs
g "!s xs ys. (s SUBSET SL xs) /\ WFxs xs /\ NOT_WFxs ys ==>
              DISJOINT (SL ys) s";;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(SET_INDUCT_TAC);;

 e(REWRITE_TAC[SUBSET]);;
 e(REWRITE_TAC[DISJOINT_EMPTY]);;

 e(REWRITE_TAC[SUBSET]);;
 e(REWRITE_TAC[DISJOINT_INSERT]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFxs_IS_WFx);;
 e(IMP_RES_TAC WFx_NOT_WFxs);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%



let WFe_NOT_WFx =
 prove_thm
  (`WFe_NOT_WFx`,
   "!e y. WFe e /\ ~WFx y ==> ~(y IN FVe e)",
   INDUCT_THEN exp_induct ASSUME_TAC
   THEN REWRITE_TAC[WFe;FVe_DEF;IN]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REWRITE_TAC[WFx_NOT_WFx]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFe_NOT_WFx:
g "!e y. WFe e /\ ~WFx y ==> ~(y IN FVe e)";;
e(INDUCT_THEN exp_induct ASSUME_TAC);;

 e(REWRITE_TAC[WFe;FVe_DEF;IN]);;

 e(REWRITE_TAC[WFe;FVe_DEF;IN]);;
 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(REWRITE_TAC[WFx_NOT_WFx]);;

 e(REWRITE_TAC[WFe;FVe_DEF;IN]);;
 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(REWRITE_TAC[WFx_NOT_WFx]);;

 e(REWRITE_TAC[WFe;FVe_DEF;IN]);;
 e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(IMP_RES_TAC WFx_NOT_WFx);;
 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(ASM_REWRITE_TAC[]);;
%

let WFes_NOT_WFx =
 prove_thm
  (`WFes_NOT_WFx`,
   "!es y. WFes es /\ ~WFx y ==> ~(y IN FVes es)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFes;FVes_DEF;IN]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC WFe_NOT_WFx
   THEN ASM_REWRITE_TAC[]
  );;

% WFes_NOT_WFx:
g "!es y. WFes es /\ ~WFx y ==> ~(y IN FVes es)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[WFes;FVes_DEF;IN]);;

 e(REWRITE_TAC[WFes;FVes_DEF;IN]);;
 e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(IMP_RES_TAC WFe_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;
%

let WFb_NOT_WFx =
 prove_thm
  (`WFb_NOT_WFx`,
   "!b y. WFb b /\ ~WFx y ==> ~(y IN FVb b)",
   INDUCT_THEN bexp_induct ASSUME_TAC
   THEN REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFe_NOT_WFx
   THEN IMP_RES_TAC WFes_NOT_WFx
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFb_NOT_WFx:
g "!b y. WFb b /\ ~WFx y ==> ~(y IN FVb b)";;
e(INDUCT_THEN bexp_induct ASSUME_TAC);;

 e(REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFe_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFes_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFb;FVb_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let WFxs_APPEND =
 prove_thm
  (`WFxs_APPEND`,
   "!xs ys. WFxs (APPEND xs ys) = WFxs xs /\ WFxs ys",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[APPEND;WFxs;CONJ_ASSOC]
  );;

% WFxs_APPEND
g "!xs ys. WFxs (APPEND xs ys) = WFxs xs /\ WFxs ys";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;WFxs]);;

 e(REWRITE_TAC[APPEND;WFxs]);;
 e(ASM_REWRITE_TAC[CONJ_ASSOC]);;
%

let NOT_WFxs_APPEND =
 prove_thm
  (`NOT_WFxs_APPEND`,
   "!xs ys. NOT_WFxs (APPEND xs ys) = NOT_WFxs xs /\ NOT_WFxs ys",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[APPEND;NOT_WFxs;CONJ_ASSOC]
  );;

% NOT_WFxs_APPEND
g "!xs ys. NOT_WFxs (APPEND xs ys) = NOT_WFxs xs /\ NOT_WFxs ys";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;NOT_WFxs]);;

 e(REWRITE_TAC[APPEND;NOT_WFxs]);;
 e(ASM_REWRITE_TAC[CONJ_ASSOC]);;
%



let WFenv_vars_NOT_WFx =
 prove_thm
  (`WFenv_vars_NOT_WFx`,
   "!r p y. WFenv_syntax r /\ ~WFx y ==>
       ~(y IN (SL           (FST (r p))  )) /\
       ~(y IN (SL      (FST (SND (r p))) )) /\
       ~(y IN (SL (FST (SND (SND (r p))))))",
   REWRITE_TAC[WFenv_syntax]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_ALL_TAC
   THEN DISCH_THEN (ASSUME_TAC o SPEC "p:string")
   THEN UNDISCH_ALL_TAC
   THEN ONCE_REWRITE_TAC[ENV_TUPLE]
   THEN REWRITE_TAC[WFproc_syntax]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[WFxs_APPEND]
   THEN STRIP_TAC
   THEN STRIP_TAC
   THEN IMP_RES_TAC WFxs_NOT_WFx
   THEN ASM_REWRITE_TAC[]
  );;

% WFenv_vars_NOT_WFx
g "!r p y. WFenv_syntax r /\ ~WFx y ==>
       ~(y IN (SL           (FST (r p))  )) /\
       ~(y IN (SL      (FST (SND (r p))) )) /\
       ~(y IN (SL (FST (SND (SND (r p))))))";;
e(REWRITE_TAC[WFenv_syntax]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(UNDISCH_ALL_TAC);;
e(DISCH_THEN (ASSUME_TAC o SPEC "p:string"));;
e(UNDISCH_ALL_TAC);;
e(ONCE_REWRITE_TAC[ENV_TUPLE]);;
e(REWRITE_TAC[WFproc_syntax]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[WFxs_APPEND]);;
e(STRIP_TAC);;

 e(STRIP_TAC);;
 e(IMP_RES_TAC WFxs_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(IMP_RES_TAC WFxs_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;
%

let WFc_NOT_WFx =
 prove_thm
  (`WFc_NOT_WFx`,
   "!c g r y. WFc c g r /\ WFenv_syntax r /\ ~WFx y ==> ~(y IN FVc c r)",
   INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[WFc;FVc_DEF;IN]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[o_THM;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFx_NOT_WFx
   THEN IMP_RES_TAC WFxs_NOT_WFx
   THEN IMP_RES_TAC WFe_NOT_WFx
   THEN IMP_RES_TAC WFes_NOT_WFx
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFenv_vars_NOT_WFx
   THEN RES_TAC
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN ASM_REWRITE_TAC[]
   THEN ASM_REWRITE_TAC[FVxs_SL]
  );;

% WFc_NOT_WFx:
g "!c g r y. WFc c g r /\ WFenv_syntax r /\ ~WFx y ==> ~(y IN FVc c r)";;
e(INDUCT_THEN cmd_induct ASSUME_TAC);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFx_NOT_WFx);;
 e(IMP_RES_TAC WFe_NOT_WFx);;
 e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFb_NOT_WFx);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN]);;
 e(REWRITE_TAC[o_THM;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(IMP_RES_TAC WFxs_NOT_WFx);;
 e(IMP_RES_TAC WFes_NOT_WFx);;
 e(IMP_RES_TAC WFenv_vars_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;
 e(ASM_REWRITE_TAC[FVxs_SL]);;
%


let WFc_NOT_WFxs =
 prove_thm
  (`WFc_NOT_WFxs`,
   "!xs c g r .
    WFc c g r /\ WFenv_syntax r /\ NOT_WFxs xs ==>
    (DISJOINT (SL xs) (FVc c r))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;SET_LIST;DISJOINT_EMPTY;DISJOINT_INSERT]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFc_NOT_WFxs:
g "!xs c g r .
    WFc c g r /\ WFenv_syntax r /\ NOT_WFxs xs ==>
    (DISJOINT (SL xs) (FVc c r))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;DISJOINT_EMPTY]);;

 e(REWRITE_TAC[NOT_WFxs;SET_LIST;DISJOINT_INSERT]);;
 e(REPEAT GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(IMP_RES_TAC WFc_NOT_WFx);;
 e(ASM_REWRITE_TAC[]);;
%




let NOT_WFs_mk_logical =
 prove_thm
  (`NOT_WFs_mk_logical`,
   "!x. ~WFs (mk_logical x)",
   REWRITE_TAC[mk_logical;WFs]
  );;

% NOT_WFs_mk_logical
g "!x. ~WFs (mk_logical x)";;
e(REWRITE_TAC[mk_logical;WFs]);;
%

let NOT_WFx_logical =
 prove_thm
  (`NOT_WFx_logical`,
   "!x. ~WFx (logical x)",
   GEN_TAC
   THEN (SUBST1_TAC o SYM o SPEC_ALL) Base_Index
   THEN REWRITE_TAC[logical;WFx]
   THEN REWRITE_TAC[NOT_WFs_mk_logical]
  );;

% NOT_WFx_logical
g "!x. ~WFx (logical x)";;
e(GEN_TAC);;
e((SUBST1_TAC o SYM o SPEC_ALL) Base_Index);;
e(REWRITE_TAC[logical;WFx]);;
e(REWRITE_TAC[NOT_WFs_mk_logical]);;
%

let NOT_WFxs_logicals =
 prove_thm
  (`NOT_WFxs_logicals`,
   "!xs. NOT_WFxs (logicals xs)",
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[logicals_THM;NOT_WFxs]
   THEN REWRITE_TAC[NOT_WFx_logical]
  );;

% NOT_WFxs_logicals
g "!xs. NOT_WFxs (logicals xs)";;
e(LIST_INDUCT_TAC);;

  e(ASM_REWRITE_TAC[logicals_THM;NOT_WFxs]);;

  e(ASM_REWRITE_TAC[logicals_THM;NOT_WFxs]);;
  e(REWRITE_TAC[NOT_WFx_logical]);;
%

let logicals_IS_NOT_WFx =
 prove_thm
  (`logicals_IS_NOT_WFx`,
   "!xs x.
      (x IN SL (logicals xs))  ==>  ~(WFx x)",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN MATCH_MP_TAC NOT_WFxs_IS_NOT_WFx
   THEN EXISTS_TAC "logicals xs"
   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
  );;

% logicals_IS_NOT_WFx
g "!xs x.
      (x IN SL (logicals xs))  ==>  ~(WFx x)";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(MATCH_MP_TAC NOT_WFxs_IS_NOT_WFx);;
e(EXISTS_TAC "logicals xs");;
e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
%

let WFx_NOT_logicals =
 prove_thm
  (`WFx_NOT_logicals`,
   "!x y. WFx x ==> ~(x IN SL(logicals y))",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN MATCH_MP_TAC WFx_NOT_WFxs
   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
  );;

% WFx_NOT_logicals
g "!x y. WFx x ==> ~(x IN SL(logicals y))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(MATCH_MP_TAC WFx_NOT_WFxs);;
e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
%

let WFxs_NOT_logicals =
 prove_thm
  (`WFxs_NOT_logicals`,
   "!x y. WFxs x ==> DISJOINT (SL x) (SL (logicals y))",
   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC WFxs_NOT_WFxs
   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
  );;

% WFxs_NOT_logicals
g "!x y. WFxs x ==> DISJOINT (SL x) (SL (logicals y))";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC WFxs_NOT_WFxs);;
e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
%

let WFx_variant =
 prove_thm
  (`WFx_variant`,
   "!x s. WFx (variant x s) = WFx x",
   ONCE_REWRITE_TAC[(SYM o SPEC_ALL) Base_Index]
   THEN REWRITE_TAC[Base_variant;WFx]
  );;

% WFx_variant
g "!x s. WFx (variant x s) = WFx x";;
e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL) Base_Index]);;
e(REWRITE_TAC[Base_variant;WFx]);;
%

let WFxs_variants =
 prove_thm
  (`WFxs_variants`,
   "!xs s. WFxs (variants xs s) = WFxs xs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;variants_THM]
   THEN ASM_REWRITE_TAC[WFx_variant]
  );;

% WFxs_variants
g "!xs s. WFxs (variants xs s) = WFxs xs";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[WFxs;variants_THM]);;

e(REWRITE_TAC[WFxs;variants_THM]);;
e(ASM_REWRITE_TAC[WFx_variant]);;
%

let NOT_WFxs_variants =
 prove_thm
  (`NOT_WFxs_variants`,
   "!xs s. NOT_WFxs (variants xs s) = NOT_WFxs xs",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;variants_THM]
   THEN ASM_REWRITE_TAC[WFx_variant]
  );;

% NOT_WFxs_variants
g "!xs s. NOT_WFxs (variants xs s) = NOT_WFxs xs";;
e(LIST_INDUCT_TAC);;

e(REWRITE_TAC[NOT_WFxs;variants_THM]);;

e(REWRITE_TAC[NOT_WFxs;variants_THM]);;
e(ASM_REWRITE_TAC[WFx_variant]);;
%

let var_list_logicals_IDENT_APPEND_LEMMA =
 TAC_PROOF
  (([],
    "!zs ys xs.
      (LENGTH xs = LENGTH ys) ==>
      (WFxs xs) /\ (WFxs ys) /\ (WFxs zs) ==>
      (((APPEND (APPEND ys zs) (logicals (APPEND ys zs))) //
        (APPEND (APPEND xs zs) (logicals (APPEND xs zs)))) =
       ((APPEND ys (logicals ys)) // (APPEND xs (logicals xs))))"),
   GEN_TAC
   THEN LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]
   THEN REWRITE_TAC[WFxs;logicals_THM]
   THEN STRIP_TAC
   THEN RES_TAC
   THEN USE_IMP_THEN REWRITE_THM var_list_APPEND_CONS
   THENL
    [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_logicals]
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs;NOT_WFx_logical;NOT_WFxs_logicals];

      REWRITE_TAC[APPEND;var_list_THM]
      THEN ASM_REWRITE_TAC[]
      THEN USE_IMP_THEN REWRITE_THM var_list_APPEND_CONS
      THENL
       [ 
         ASM_REWRITE_TAC[LENGTH_logicals]
         THEN MATCH_MP_TAC WFxs_NOT_WFxs
         THEN ASM_REWRITE_TAC[NOT_WFxs;NOT_WFx_logical;NOT_WFxs_logicals];

         REWRITE_TAC[var_list_THM]
       ]
    ]
  );;

% var_list_logicals_IDENT_APPEND_LEMMA
g "!zs ys xs.
      (LENGTH xs = LENGTH ys) ==>
      (WFxs xs) /\ (WFxs ys) /\ (WFxs zs) ==>
      (((APPEND (APPEND ys zs) (logicals (APPEND ys zs))) //
        (APPEND (APPEND xs zs) (logicals (APPEND xs zs)))) =
       ((APPEND ys (logicals ys)) // (APPEND xs (logicals xs))))";;
e(GEN_TAC);;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]);;

 e(REWRITE_TAC[APPEND;var_list_THM;var_list_IDENT]);;
 e(REWRITE_TAC[WFxs;logicals_THM]);;
 e(STRIP_TAC);;
 e(RES_TAC);;
 e(USE_IMP_THEN REWRITE_THM var_list_APPEND_CONS);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_logicals]);;
  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs;NOT_WFx_logical;NOT_WFxs_logicals]);;

  e(REWRITE_TAC[APPEND;var_list_THM]);;
  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN REWRITE_THM var_list_APPEND_CONS);;

   e(ASM_REWRITE_TAC[LENGTH_logicals]);;
   e(MATCH_MP_TAC WFxs_NOT_WFxs);;
   e(ASM_REWRITE_TAC[NOT_WFxs;NOT_WFx_logical;NOT_WFxs_logicals]);;

   e(REWRITE_TAC[var_list_THM]);;
%

let var_list_logicals_IDENT_APPEND =
  save_thm(`var_list_logicals_IDENT_APPEND`,
      REWRITE_RULE[AND_IMP_INTRO] var_list_logicals_IDENT_APPEND_LEMMA);;


let vsubst_PRESERVES_WFx =
 prove_thm
  (`vsubst_PRESERVES_WFx`,
   "!ys xs. (MAP WFx ys = MAP WFx xs) ==>
         (!x. WFx ((ys //v xs) x) = WFx x)",
   DOUBLE_LIST_INDUCT_TAC
   THEN REWRITE_TAC[MAP;APPEND;vsubst_THM;I_THM]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN POP_TAC THEN POP_TAC
   THEN ASM_CASES_TAC "(x:var) = h'"
   THENL
     [  ASM_REWRITE_TAC[BND_MATCH];

        IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
        THEN ASM_CASES_TAC "x = (@z. (ys //v xs)z = h)"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH]
             THEN FIRST_ASSUM (ONCE_REWRITE_THM o SYM o
                                       SPEC "@z. (ys //v xs)z = h")
             THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
             THEN ASM_REWRITE_TAC[];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

%
g "!ys xs. (MAP WFx ys = MAP WFx xs) ==>
         (!x. WFx ((ys //v xs) x) = WFx x)";;
e(DOUBLE_LIST_INDUCT_TAC);;

e(REWRITE_TAC[MAP;APPEND;vsubst_THM;I_THM]);;

e(REWRITE_TAC[MAP;APPEND;vsubst_THM;I_THM]);;

e(REWRITE_TAC[MAP;APPEND;vsubst_THM;I_THM]);;
e(REWRITE_TAC[CONS_11]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(POP_TAC THEN POP_TAC);;
e(ASM_CASES_TAC "(x:var) = h'");;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(ASM_CASES_TAC "x = (@z. (ys //v xs)z = h)");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;
    e(FIRST_ASSUM (ONCE_REWRITE_THM o SYM o SPEC "@z. (ys //v xs)z = h"));;
    e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
%

let MAP_IMP_ANTIMAP_WFx =
 prove_thm
  (`MAP_IMP_ANTIMAP_WFx`,
   "!ys xs x. (!x. WFx ((ys //v xs) x) = WFx x)  ==>
             (WFx (@z. (ys //v xs)z = x) = WFx x)",
   REPEAT STRIP_TAC
   THEN POP_ASSUM (ONCE_REWRITE_THM o SYM o SPEC "@z. (ys //v xs)z = x")
   THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
  );;

% MAP_IMP_ANTIMAP_WFx
g "!ys xs x. (!x. WFx ((ys //v xs) x) = WFx x)  ==>
             (WFx (@z. (ys //v xs)z = x) = WFx x)";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (ONCE_REWRITE_THM o SYM o SPEC "@z. (ys //v xs)z = x"));;
e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
%

let WFx_biject_INVERSE =
 prove_thm
  (`WFx_biject_INVERSE`,
   "!(ss:vsubst).
       ONE_ONE ss /\ ONTO ss ==>
       ((!x. WFx (@z. ss z = x) = WFx x)  =
        (!x. WFx (ss x) = WFx x))",
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THENL
     [ POP_ASSUM (ONCE_REWRITE_THM o SYM o SPEC "(ss:vsubst)x")
       THEN IMP_RES_THEN REWRITE_THM ONE_ONE
       THEN REWRITE_TAC[SELECT_REFL];

       POP_ASSUM (ONCE_REWRITE_THM o SYM o SPEC "@z. (ss:vsubst)z = x")
       THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE
     ]
  );;

% WFx_biject_INVERSE
g "!(ss:vsubst).
       ONE_ONE ss /\ ONTO ss ==>
       ((!x. WFx (@z. ss z = x) = WFx x)  =
        (!x. WFx (ss x) = WFx x))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (ONCE_REWRITE_THM o pthm o SYM o SPEC "(ss:vsubst)x"));;
 e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
 e(REWRITE_TAC[SELECT_REFL]);;

 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (ONCE_REWRITE_THM o SYM o pthm o SPEC "@z. (ss:vsubst)z = x"));;
 e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;
%

let NOT_WFxs_IMP_MAP_EQ =
 prove_thm
  (`NOT_WFxs_IMP_MAP_EQ`,
   "!zs ws.
           (LENGTH ws = LENGTH zs) ==>
           NOT_WFxs ws /\ NOT_WFxs zs  ==>
           (MAP WFx zs = MAP WFx ws)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;MAP;CONS_11]
   THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% NOT_WFxs_IMP_MAP_EQ
g "!zs ws.
           (LENGTH ws = LENGTH zs) ==>
           NOT_WFxs ws /\ NOT_WFxs zs  ==>
           (MAP WFx zs = MAP WFx ws)";;
e(LENGTH_LIST_INDUCT_TAC);;
e(REWRITE_TAC[NOT_WFxs;MAP;CONS_11]);;
e(STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let WFxs_IMP_MAP_EQ =
 prove_thm
  (`WFxs_IMP_MAP_EQ`,
   "!ys xs.
           (LENGTH xs = LENGTH ys) ==>
           WFxs xs /\ WFxs ys  ==>
           (MAP WFx ys = MAP WFx xs)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;MAP;CONS_11]
   THEN STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFxs_IMP_MAP_EQ
g "!ys xs.
           (LENGTH xs = LENGTH ys) ==>
           WFxs xs /\ WFxs ys  ==>
           (MAP WFx ys = MAP WFx xs)";;
e(LENGTH_LIST_INDUCT_TAC);;
e(REWRITE_TAC[WFxs;MAP;CONS_11]);;
e(STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let NOT_WFxs_vsubst_INDEPENDENCE =
 prove_thm
  (`NOT_WFxs_vsubst_INDEPENDENCE`,
   "!zs ws x.
           (LENGTH ws = LENGTH zs) ==>
           NOT_WFxs ws /\ NOT_WFxs zs /\
           WFx x  ==>
           ((zs //v ws)x = x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[NOT_WFxs;vsubst_THM;I_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC WFx_NOT_WFx
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH
   THEN MATCH_MP_TAC WFx_NOT_WFx
   THEN ASM_REWRITE_TAC[]
   THEN USE_IMP_THEN ASM_REWRITE_THM MAP_IMP_ANTIMAP_WFx
   THEN MATCH_MP_TAC vsubst_PRESERVES_WFx
   THEN IMP_RES_TAC NOT_WFxs_IMP_MAP_EQ
  );;

% NOT_WFxs_vsubst_INDEPENDENCE
g "!zs ws x.
           (LENGTH ws = LENGTH zs) ==>
           NOT_WFxs ws /\ NOT_WFxs zs /\
           WFx x  ==>
           ((zs //v ws)x = x)";;

e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[NOT_WFxs;vsubst_THM;I_THM]);;

  e(REWRITE_TAC[NOT_WFxs;vsubst_THM;I_THM]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(IMP_RES_TAC WFx_NOT_WFx);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
  e(MATCH_MP_TAC WFx_NOT_WFx);;
  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM MAP_IMP_ANTIMAP_WFx);;
  e(MATCH_MP_TAC vsubst_PRESERVES_WFx);;
  e(IMP_RES_TAC NOT_WFxs_IMP_MAP_EQ);;
let st1 = get_state();;
set_state(st1);;
%

let WFxs_vsubst_INDEPENDENCE =
 prove_thm
  (`WFxs_vsubst_INDEPENDENCE`,
   "!ys xs x.
           (LENGTH xs = LENGTH ys) ==>
           WFxs xs /\ WFxs ys /\
           ~WFx x  ==>
           ((ys //v xs)x = x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;vsubst_THM;I_THM]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_THEN (IMP_RES_THEN (STRIP_ASSUME_TAC o NOT_EQ_SYM)) WFx_NOT_WFx
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN MATCH_MP_TAC WFx_NOT_WFx
   THEN ASM_REWRITE_TAC[]
   THEN USE_IMP_THEN ASM_REWRITE_THM MAP_IMP_ANTIMAP_WFx
   THEN MATCH_MP_TAC vsubst_PRESERVES_WFx
   THEN IMP_RES_TAC WFxs_IMP_MAP_EQ
  );;

% WFxs_vsubst_INDEPENDENCE
g "!ys xs x.
           (LENGTH xs = LENGTH ys) ==>
           WFxs xs /\ WFxs ys /\
           ~WFx x  ==>
           ((ys //v xs)x = x)";;

e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[WFxs;vsubst_THM;I_THM]);;

  e(REWRITE_TAC[WFxs;vsubst_THM;I_THM]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(IMP_RES_THEN (IMP_RES_THEN (STRIP_ASSUME_TAC o NOT_EQ_SYM)) WFx_NOT_WFx);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(USE_IMP_THEN ASM_REWRITE_THM BND_NOT_MATCH);;
  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(MATCH_MP_TAC WFx_NOT_WFx);;
  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM MAP_IMP_ANTIMAP_WFx);;
  e(MATCH_MP_TAC vsubst_PRESERVES_WFx);;
  e(IMP_RES_TAC WFxs_IMP_MAP_EQ);;
%

let STANDARD_SUBST_OF_WFx =
 prove_thm
  (`STANDARD_SUBST_OF_WFx`,
   "!ys xs zs ws x.
           (LENGTH xs = LENGTH ys) ==>
           (LENGTH ws = LENGTH zs) /\
           WFxs xs /\ WFxs ys /\
           NOT_WFxs ws /\ NOT_WFxs zs /\
           WFx x  ==>
           ((APPEND ys zs //v APPEND xs ws)x = (ys //v xs)x)",
   LENGTH_LIST_INDUCT_TAC
   THENL
     [  REWRITE_TAC[WFxs;APPEND;vsubst_THM;I_THM]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_TAC NOT_WFxs_vsubst_INDEPENDENCE;

        REWRITE_TAC[WFxs;APPEND;vsubst_THM]
        THEN REPEAT STRIP_TAC
        THEN ASM_CASES_TAC "(x:var) = h'"
        THENL
          [  ASM_REWRITE_TAC[BND_MATCH];

             IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
             THEN FIRST_ASSUM (\th. ASSUME_TAC (MATCH_MP (MATCH_MP th
                     (ASSUME "LENGTH (l':(var)list) = LENGTH (ys:(var)list)"))
                    (LIST_CONJ (map ASSUME
                     [ "LENGTH (ws:(var)list) = LENGTH (zs:(var)list)";
                       "WFxs l'";
                       "WFxs ys";
                       "NOT_WFxs ws";
                       "NOT_WFxs zs";
                       "WFx x"
                     ]))))
             THEN ASM_CASES_TAC "x = (@z. (ys //v l')z = h)"
             THENL
               [  POP_ASSUM (ASSUME_TAC o SYM)
                  THEN ASM_REWRITE_TAC[BND_MATCH]
                  THEN POP_ASSUM (\th1. POP_ASSUM (\th2.
                        REWRITE_TAC[vsubst_11;SELECT_REFL;BND_MATCH;
                         (SYM o TRANS th2 o
                          REWRITE_RULE[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]) th1 ]))
                  THEN FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
                  THEN ASM_REWRITE_TAC[];

                  IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
                  THEN POP_ASSUM (\th1. POP_ASSUM (\th2.
                        (ASSUME_TAC o ONCE_REWRITE_RULE[vsubst_EQ_INVERSE1] o
                         ONCE_REWRITE_RULE[SYM th2] o
                         REWRITE_RULE[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]) th1))
                  THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
                  THEN FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
                  THEN ASM_REWRITE_TAC[]
               ]
          ]
     ]
  );;

% STANDARD_SUBST_OF_WFx
g "!ys xs zs ws x.
           (LENGTH xs = LENGTH ys) ==>
           (LENGTH ws = LENGTH zs) /\
           WFxs xs /\ WFxs ys /\
           NOT_WFxs ws /\ NOT_WFxs zs /\
           WFx x  ==>
           ((APPEND ys zs //v APPEND xs ws)x = (ys //v xs)x)";;

e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[WFxs;APPEND;vsubst_THM;I_THM]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC NOT_WFxs_vsubst_INDEPENDENCE);;

  e(REWRITE_TAC[WFxs;APPEND;vsubst_THM]);;
  e(REPEAT STRIP_TAC);;
  e(ASM_CASES_TAC "(x:var) = h'");;

    e(ASM_REWRITE_TAC[BND_MATCH]);;

    e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
    e(FIRST_ASSUM (\th. ASSUME_TAC (MATCH_MP (MATCH_MP th
                     (ASSUME "LENGTH (l':(var)list) = LENGTH (ys:(var)list)"))
                    (LIST_CONJ (map ASSUME
                     [ "LENGTH (ws:(var)list) = LENGTH (zs:(var)list)";
                       "WFxs l'";
                       "WFxs ys";
                       "NOT_WFxs ws";
                       "NOT_WFxs zs";
                       "WFx x"
                     ])))));;
    e(ASM_CASES_TAC "x = (@z. (ys //v l')z = h)");;

      e(POP_ASSUM (ASSUME_TAC o SYM));;
      e(ASM_REWRITE_TAC[BND_MATCH]);;
      e(POP_ASSUM (\th1. POP_ASSUM (\th2.
         REWRITE_TAC[vsubst_11;SELECT_REFL;BND_MATCH;
          (SYM o TRANS th2 o
           REWRITE_RULE[(SYM o SPEC_ALL) vsubst_EQ_INVERSE]) th1 ])));;
      e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
      e(ASM_REWRITE_TAC[]);;

      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(POP_ASSUM (\th1. POP_ASSUM (\th2.
          (ASSUME_TAC o ONCE_REWRITE_RULE[vsubst_EQ_INVERSE1] o
           ONCE_REWRITE_RULE[SYM th2] o
           REWRITE_RULE[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]) th1)));;
      e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
      e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
      e(ASM_REWRITE_TAC[]);;

%

let STANDARD_SUBST_OF_NOT_WFx =
 prove_thm
  (`STANDARD_SUBST_OF_NOT_WFx`,
   "!ys xs zs ws x.
           (LENGTH xs = LENGTH ys) ==>
           (LENGTH ws = LENGTH zs) /\
           WFxs xs /\ WFxs ys /\
           NOT_WFxs ws /\ NOT_WFxs zs /\
           ~WFx x  ==>
           ((APPEND ys zs //v APPEND xs ws)x = (zs //v ws)x)",
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;APPEND;vsubst_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_THEN (IMP_RES_THEN (STRIP_ASSUME_TAC o NOT_EQ_SYM)) WFx_NOT_WFx
   THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
   THEN USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
   THENL
     [  REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]
        THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM o REWRITE_RULE[AND_IMP_INTRO])
        THENL
          [  ASM_REWRITE_TAC[];

             ONCE_REWRITE_TAC[EQ_SYM_EQ]
             THEN MATCH_MP_TAC WFx_NOT_WFx
             THEN ASM_REWRITE_TAC[]
             THEN IMP_RES_TAC NOT_WFxs_IMP_MAP_EQ
             THEN IMP_RES_THEN ASM_REWRITE_THM vsubst_PRESERVES_WFx
          ];

        FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% STANDARD_SUBST_OF_NOT_WFx
g "!ys xs zs ws x.
           (LENGTH xs = LENGTH ys) ==>
           (LENGTH ws = LENGTH zs) /\
           ALL_EL WFx xs /\ ALL_EL WFx ys /\
           ALL_EL ($~ o WFx) ws /\ ALL_EL ($~ o WFx) zs /\
           ~WFx x  ==>
           ((APPEND ys zs //v APPEND xs ws)x = (zs //v ws)x)";;

e(LENGTH_LIST_INDUCT_TAC);;

  e(REWRITE_TAC[ALL_EL;APPEND;vsubst_THM]);;

  e(REWRITE_TAC[ALL_EL;APPEND;vsubst_THM]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_THEN (IMP_RES_THEN (STRIP_ASSUME_TAC o NOT_EQ_SYM)) WFx_NOT_WFx);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;

    e(REWRITE_TAC[(SYM o SPEC_ALL) vsubst_EQ_INVERSE1]);;
    e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM o REWRITE_RULE[AND_IMP_INTRO]));;

      e(ASM_REWRITE_TAC[]);;

      e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
      e(MATCH_MP_TAC WFx_NOT_WFx);;
      e(ASM_REWRITE_TAC[]);;
      e(IMP_RES_TAC NOT_WFxs_IMP_MAP_EQ);;
      e(IMP_RES_THEN ASM_REWRITE_THM vsubst_PRESERVES_WFx);;

    e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
    e(ASM_REWRITE_TAC[]);;
%


let STANDARD_SUBST_IS_WFsubst =
 prove_thm
  (`STANDARD_SUBST_IS_WFsubst`,
   "!ys xs. (LENGTH xs = LENGTH ys) /\ WFxs xs /\ WFxs ys ==>
            WFsubst (APPEND ys (logicals ys) //v APPEND xs (logicals xs))",
   REWRITE_TAC[WFsubst]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC vsubst_PRESERVES_WFx
        THEN REWRITE_TAC[MAP_APPEND]
        THEN USE_IMP_THEN REWRITE_THM APPEND_11
        THENL
          [  ASM_REWRITE_TAC[LENGTH_MAP;LENGTH;logicals];

             CONJ_TAC
             THENL
               [  IMP_RES_TAC WFxs_IMP_MAP_EQ;

                  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO]
                                       NOT_WFxs_IMP_MAP_EQ)
                  THEN REWRITE_TAC[NOT_WFxs_logicals]
                  THEN ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP]
               ]
          ];

        USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO]
                                  STANDARD_SUBST_OF_NOT_WFx)
        THENL
          [  ASM_REWRITE_TAC[NOT_WFxs_logicals;NOT_WFx_logical]
             THEN ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP];

             USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO]
                                       STANDARD_SUBST_OF_WFx)
             THENL
               [  ASM_REWRITE_TAC[NOT_WFxs_logicals;NOT_WFx_logical]
                  THEN ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP];

                  REWRITE_TAC[logical_vsubst]
               ]
          ]
     ]
  );;

% STANDARD_SUBST_IS_WFsubst
g "!ys xs. (LENGTH xs = LENGTH ys) /\ ALL_EL WFx xs /\ ALL_EL WFx ys ==>
           WFsubst (APPEND ys (logicals ys) //v APPEND xs (logicals xs))";;
e(REWRITE_TAC[WFsubst]);;
e(REPEAT STRIP_TAC);;

  e(MATCH_MP_TAC vsubst_PRESERVES_WFx);;
  e(REWRITE_TAC[MAP_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM APPEND_11);;

    e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH;logicals]);;

    e(CONJ_TAC);;

      e(IMP_RES_TAC WFxs_IMP_MAP_EQ);;

      e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NOT_WFxs_IMP_MAP_EQ));;
      e(REWRITE_TAC[NOT_WFxs_logicals]);;
      e(ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP]);;

  e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO]
                              STANDARD_SUBST_OF_NOT_WFx));;

    e(ASM_REWRITE_TAC[NOT_WFxs_logicals;NOT_WFx_logical]);;
    e(ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP]);;

    e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO]
                                STANDARD_SUBST_OF_WFx));;

      e(ASM_REWRITE_TAC[NOT_WFxs_logicals;NOT_WFx_logical]);;
      e(ASM_REWRITE_TAC[LENGTH;logicals;LENGTH_MAP]);;

      e(REWRITE_TAC[logical_vsubst]);;
%




let MAP_STANDARD_SUBST_OF_WFx =
 prove_thm
  (`MAP_STANDARD_SUBST_OF_WFx`,
   "!z y x. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y /\ WFxs z ==>
         (MAP (APPEND y (logicals y) //v APPEND x (logicals x)) z =
          MAP (y //v x) z)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;DISJOINT_LIST;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] STANDARD_SUBST_OF_WFx)
        THEN ASM_REWRITE_TAC[LENGTH_logicals;NOT_WFxs_logicals];

        RES_TAC
     ]
  );;

% MAP_STANDARD_SUBST_OF_WFx
g "!z y x. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y /\ WFxs z ==>
         (MAP (APPEND y (logicals y) //v APPEND x (logicals x)) z =
          MAP (y //v x) z)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[WFxs;DISJOINT_LIST;MAP]);;

  e(REWRITE_TAC[WFxs;DISJOINT_LIST;MAP]);;
  e(REWRITE_TAC[CONS_11]);;
  e(REPEAT STRIP_TAC);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] STANDARD_SUBST_OF_WFx));;
    e(ASM_REWRITE_TAC[LENGTH_logicals;NOT_WFxs_logicals]);;

    e(RES_TAC);;
%


let MAP_STANDARD_SUBST_REPLACEMENT =
 prove_thm
  (`MAP_STANDARD_SUBST_REPLACEMENT`,
   "!y x. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y ==>
         (MAP (APPEND y (logicals y) //v APPEND x (logicals x)) x = y)",
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_OF_WFx
   THENL
     [  ASM_REWRITE_TAC[];

        IMP_RES_TAC vsubst_REPLACEMENT
     ]
  );;

% MAP_STANDARD_SUBST_REPLACEMENT
g "!y x. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y ==>
         (MAP (APPEND y (logicals y) //v APPEND x (logicals x)) x = y)";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_OF_WFx);;

  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC vsubst_REPLACEMENT);;
%





let WELL_FORMED_PROCEDURE_SYNTAX =
 prove_thm
  (`WELL_FORMED_PROCEDURE_SYNTAX`,
   "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs)))))
%    /\ ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
        ",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_TAC "WFenv_syntax r"
   THEN REWRITE_TAC[WFenv_syntax]
   THEN DISCH_THEN (ASSUME_TAC o SPEC_ALL)
   THEN UNDISCH_TAC "WFproc_syntax (r p) r"
   THEN ASM_REWRITE_TAC[WFproc_syntax]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< WELL_FORMED_PROCEDURE_SYNTAX:
g "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs)))))
%    /\ ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
        ";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(UNDISCH_TAC "WFenv_syntax r");;
e(REWRITE_TAC[WFenv_syntax]);;
e(DISCH_THEN (ASSUME_TAC o SPEC_ALL));;
e(UNDISCH_TAC "WFproc_syntax (r p) r");;
e(ASM_REWRITE_TAC[WFproc_syntax]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[]);;
>%


let WELL_FORMED_PROCEDURE_PARTIAL =
 prove_thm
  (`WELL_FORMED_PROCEDURE_PARTIAL`,
   "!r p vars vals glbs pre post calls rec c.
       WFenvp r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs)))))
%    /\ ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
    /\ PCSPEC (AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                            (APPEND (APPEND vars vals) glbs))
                    pre, c, post, r)
        ",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_TAC "WFenvp r"
   THEN REWRITE_TAC[WFenvp_aspects]
   THEN STRIP_TAC
   THEN POP_ASSUM MP_TAC
   THEN REWRITE_TAC[WFenv_partial]
   THEN DISCH_THEN (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[WFproc_partial]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN ASM_REWRITE_TAC[]
  );;

%< WELL_FORMED_PROCEDURE_PARTIAL:
g "!r p vars vals glbs pre post calls rec c.
       WFenvp r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs)))))
%    /\ ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
    /\ PCSPEC (AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                            (APPEND (APPEND vars vals) glbs))
                    pre, c, post, r)
        ";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(UNDISCH_TAC "WFenvp r");;
e(REWRITE_TAC[WFenvp_aspects]);;
e(STRIP_TAC);;
e(POP_ASSUM MP_TAC);;
e(REWRITE_TAC[WFenv_partial]);;
e(DISCH_THEN (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[WFproc_partial]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(ASM_REWRITE_TAC[]);;
>%


let WELL_FORMED_PROCEDURE_REC =
 prove_thm
  (`WELL_FORMED_PROCEDURE_REC`,
   "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs))))
        ",
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_TAC "WFenv_syntax r"
   THEN REWRITE_TAC[WFenv_syntax]
   THEN DISCH_THEN (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[WFproc_syntax]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THENL
    [ ASM_REWRITE_TAC[];

      DISJ2_TAC
      THEN EXISTS_TAC "v:vexp"
      THEN EXISTS_TAC "y:var"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< WELL_FORMED_PROCEDURE_REC:
g "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs))))
        ";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(UNDISCH_TAC "WFenv_syntax r");;
e(REWRITE_TAC[WFenv_syntax]);;
e(DISCH_THEN (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[WFproc_syntax]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(DISJ2_TAC);;
 e(EXISTS_TAC "v:vexp");;
 e(EXISTS_TAC "y:var");;
 e(ASM_REWRITE_TAC[]);;
>%


let FVa_REC =
 prove_thm
  (`FVa_REC`,
   "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       (?z. (FVa rec) SUBSET (SL(APPEND (APPEND (APPEND vars vals) glbs)
                                        (logicals z))))",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC)
   THENL
    [ ASM_REWRITE_TAC[FVa_DEF;SUBSET];

      EXISTS_TAC "[dest_logical y]"
      THEN IMP_RES_TAC NOT_WFx_dest_logical
      THEN ASM_REWRITE_TAC[FVa_DEF;FVv_DEF;UNION_SUBSET;logicals_THM]
      THEN REWRITE_TAC[SUBSET_DEF]
      THEN ONCE_REWRITE_TAC[SL_APPEND]
      THEN REWRITE_TAC[IN;IN_UNION;SET_LIST]
      THEN REPEAT STRIP_TAC
      THENL
       [ IMP_RES_TAC SUBSET_DEF
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< FVa_REC:
g "!r p vars vals glbs pre post calls rec c.
       WFenv_syntax r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       (?z. (FVa rec) SUBSET (SL(APPEND (APPEND (APPEND vars vals) glbs)
                                        (logicals z))))";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC));;

 e(ASM_REWRITE_TAC[FVa_DEF;SUBSET]);;

 e(EXISTS_TAC "[dest_logical y]");;
 e(IMP_RES_TAC NOT_WFx_dest_logical);;
 e(ASM_REWRITE_TAC[FVa_DEF;FVv_DEF;UNION_SUBSET;logicals_THM]);;
 e(REWRITE_TAC[SUBSET_DEF]);;
 e(ONCE_REWRITE_TAC[SL_APPEND]);;
 e(REWRITE_TAC[IN;IN_UNION;SET_LIST]);;
 e(REPEAT STRIP_TAC);;

  e(IMP_RES_TAC SUBSET_DEF);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
>%



let WELL_FORMED_PROCEDURE =
 prove_thm
  (`WELL_FORMED_PROCEDURE`,
   "!r p vars vals glbs pre post calls rec c.
       WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs))))) /\
%       ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
       PRE_PROGRESS(pre,c,r) /\
       BODY_PROGRESS(AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                                  (APPEND (APPEND vars vals) glbs))
                          pre, c, calls, r) /\
       REC_PROGRESS(AAND pre (induct_pre rec),p,rec,r) /\
       CSPEC (AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                           (APPEND (APPEND vars vals) glbs))
                   pre, c, post, r)",

   REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_TAC "WFenv r"
   THEN REWRITE_TAC[WFenv]
   THEN DISCH_THEN (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[WFproc]
   THEN ASM_REWRITE_TAC[WFproc_syntax;WFproc_pre;WFproc_calls;
                        WFproc_rec;WFproc_correct]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< WELL_FORMED_PROCEDURE:
g "!r p vars vals glbs pre post calls rec c.
       WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) ==>
       WFxs (APPEND (APPEND vars vals) glbs) /\
       DL (APPEND (APPEND vars vals) glbs) /\
       WFc c calls r /\
       ((GVc c r) SUBSET (SL glbs)) /\
       ((FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa pre) SUBSET (SL (APPEND (APPEND vars vals) glbs))) /\
       ((FVa post) SUBSET
               (SL(APPEND (APPEND (APPEND vars vals) glbs)
                          (logicals (APPEND (APPEND vars vals) glbs))))) /\
       (!s. let (vars',vals',glbs',pre',post',calls',rec',c') = r s
            in
             (FVa (calls s)) SUBSET
              (SL (APPEND (APPEND(APPEND vars' vals')glbs')
                          (logicals(APPEND(APPEND vars vals)glbs))))) /\
%       ((rec = AFALSE) \/
        (?v y.
          (rec = ALESS v(AVAR y)) /\
          ~WFx y /\
          (FVv v) SUBSET (SL(APPEND(APPEND vars vals)glbs)))) /\
%
       PRE_PROGRESS(pre,c,r) /\
       BODY_PROGRESS(AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                                  (APPEND (APPEND vars vals) glbs))
                          pre, c, calls, r) /\
       REC_PROGRESS(AAND pre (induct_pre rec),p,rec,r) /\
       CSPEC (AAND (mk_eqs (logicals (APPEND (APPEND vars vals) glbs))
                           (APPEND (APPEND vars vals) glbs))
                   pre, c, post, r)";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(UNDISCH_TAC "WFenv r");;
e(REWRITE_TAC[WFenv]);;
e(DISCH_THEN (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[WFproc]);;
e(ASM_REWRITE_TAC[WFproc_syntax;WFproc_pre;WFproc_calls;
                  WFproc_rec;WFproc_correct]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[]);;
>%






let WFsubst_I =
 prove_thm
  (`WFsubst_I`,
   "WFsubst (I:var->var)",
   REWRITE_TAC[WFsubst;I_THM]
  );;

% WFsubst_I
g "WFsubst (I:var->var)";;
e(REWRITE_TAC[WFsubst;I_THM]);;
%


let WFxs_subst =
 prove_thm
  (`WFxs_subst`,
   "!xs ss. (!x. WFx(ss x) = WFx x) ==> (WFxs (xs <|xs ss) = WFxs xs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[xs_subst;WFxs]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% WFxs_subst
g "!xs ss. (!x. WFx(ss x) = WFx x) ==> (WFxs (xs <|xs ss) = WFxs xs)";;
e(LIST_INDUCT_TAC);;
e(REWRITE_TAC[xs_subst;WFxs]);;

e(REWRITE_TAC[xs_subst;WFxs]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let WFe_subst =
 prove_thm
  (`WFe_subst`,
   "!e ss. (!x. WFx(ss x) = WFx x) ==> (WFe (e <|e ss) = WFe e)",
   INDUCT_THEN exp_induct STRIP_ASSUME_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[e_subst;WFe]
  );;

% WFe_subst
g "!e ss. (!x. WFx(ss x) = WFx x) ==> (WFe (e <|e ss) = WFe e)";;
e(INDUCT_THEN exp_induct STRIP_ASSUME_TAC);;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[e_subst;WFe]);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;
%

let WFes_subst =
 prove_thm
  (`WFes_subst`,
   "!es ss. (!x. WFx(ss x) = WFx x) ==> (WFes (es <|es ss) = WFes es)",
   LIST_INDUCT_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC WFe_subst
   THEN ASM_REWRITE_TAC[es_subst;WFes]
  );;

let WFb_subst =
 prove_thm
  (`WFb_subst`,
   "!b ss. (!x. WFx(ss x) = WFx x) ==> (WFb (b <|b ss) = WFb b)",
   INDUCT_THEN bexp_induct STRIP_ASSUME_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN IMP_RES_TAC WFe_subst
   THEN IMP_RES_TAC WFes_subst
   THEN ASM_REWRITE_TAC[b_subst;WFb]
  );;

% We don't care about the following theorem. 

let WFenv_subst_vars =
 prove_thm
  (`WFenv_subst_vars`,
   "!r ss. WFsubst ss ==>
      (FST o (r <|env ss) = (MAP ss o FST o r)) /\
      (FST o SND o (r <|env ss) = (MAP ss o FST o SND o r))",
   REPEAT STRIP_TAC
   THEN EXT_TAC "p:string"
   THEN GEN_TAC
   THEN REWRITE_TAC[env_subst;o_THM]
   THEN BETA_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;
%

% WFenv_subst_vars
g "!r ss. WFsubst ss ==>
      (FST o (r <|env ss) = (MAP ss o FST o r)) /\
      (FST o SND o (r <|env ss) = (MAP ss o FST o SND o r))";;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "p:string");;
e(GEN_TAC);;
e(REWRITE_TAC[env_subst;o_THM]);;
e(BETA_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
%


let OR_IMPLIES_LEMMA =
 TAC_PROOF
  (([], "!a b c. ((a \/ b) ==> c) = ((a ==> c) /\ (b ==> c))"),
   REPEAT GEN_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% OR_IMPLIES_LEMMA
g "!a b c. ((a \/ b) ==> c) = ((a ==> c) /\ (b ==> c))";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(CONJ_TAC);;

  e(DISCH_TAC);;
  e(RES_TAC);;

  e(DISCH_TAC);;
  e(RES_TAC);;

 e(STRIP_TAC);;
 e(STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%





let EXISTS_vsubst_LEMMA =
 TAC_PROOF
  (([],
    "!a1 ss a0 (g:prog_env).
     (!x. WFx(ss x) = WFx x) /\ ONE_ONE ss ==>
     ((?v x.
        (a1 <|av ss = ALESS v(AVAR x)) /\
        ~WFx x /\
        ~x IN ((FVa(a0 <|av ss)) UNION (FVv v)) /\
        (!p. ~(x IN FVa ((g <|g ss) p)))) =
      (?v x.
        (a1 = ALESS v(AVAR x)) /\
        ~WFx x /\
        ~x IN ((FVa a0) UNION (FVv v)) /\
        (!p. ~(x IN FVa (g p)))))"),
   INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[a_vsubst]
   THEN REWRITE_TAC[aexp_distinct;aexp_one_one]
   THEN GEN_TAC
   THEN INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[v_vsubst]
   THEN REWRITE_TAC[vexp_distinct;vexp_one_one]
   THEN REWRITE_TAC[IN_UNION;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN REWRITE_TAC[g_subst]
   THEN BETA_TAC
   THEN IMP_RES_THEN REWRITE_THM FVa_ONE_ONE_vsubst
   THEN REWRITE_TAC[IN_IMAGE]
   THEN CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV)
   THEN REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]
   THEN EQ_TAC
   THEN STRIP_TAC
   THENL
    [ EXISTS_TAC "v0:vexp"
      THEN EXISTS_TAC "v:var"
      THEN UNDISCH_ALL_TAC
      THEN DISCH_TAC THEN DISCH_TAC
      THEN DISCH_THEN (REWRITE_THM o SYM)
      THEN REWRITE_TAC[FVv_vsubst;IN_IMAGE]
      THEN CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV)
      THEN REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]
      THEN DISCH_THEN (REWRITE_THM o SYM)
      THEN IMP_RES_THEN ASM_REWRITE_THM ONE_ONE
      THEN REPEAT DISCH_TAC
      THEN ASM_REWRITE_TAC[]
      THEN REPEAT CONJ_TAC
      THEN REPEAT GEN_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN REWRITE_TAC[];

      EXISTS_TAC "v0 <|vv ss"
      THEN EXISTS_TAC "(ss:vsubst)v"
      THEN ASM_REWRITE_TAC[FVv_vsubst;IN_IMAGE]
      THEN CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV)
      THEN REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]
      THEN IMP_RES_THEN ASM_REWRITE_THM ONE_ONE
      THEN REPEAT CONJ_TAC
      THEN REPEAT GEN_TAC
      THEN DISCH_THEN (REWRITE_THM o SYM)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< EXISTS_vsubst_LEMMA
g "!a1 ss a0 (g:prog_env).
      (!x. WFx(ss x) = WFx x) /\ ONE_ONE ss ==>
      ((?v x.
         (a1 <|av ss = ALESS v(AVAR x)) /\
         ~WFx x /\
         ~x IN ((FVa(a0 <|av ss)) UNION (FVv v)) /\
         (!p. ~(x IN FVa ((g <|g ss) p)))) =
       (?v x.
         (a1 = ALESS v(AVAR x)) /\
         ~WFx x /\
         ~x IN ((FVa a0) UNION (FVv v)) /\
         (!p. ~(x IN FVa (g p)))))";;
e(INDUCT_THEN aexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[a_vsubst]
  THEN REWRITE_TAC[aexp_distinct;aexp_one_one]);;
e(GEN_TAC);;
e(INDUCT_THEN vexp_induct STRIP_ASSUME_TAC
  THEN REWRITE_TAC[v_vsubst]
  THEN REWRITE_TAC[vexp_distinct;vexp_one_one]);;
e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(REWRITE_TAC[g_subst]);;
e(BETA_TAC);;
e(IMP_RES_THEN REWRITE_THM FVa_ONE_ONE_vsubst);;
e(REWRITE_TAC[IN_IMAGE]);;
e(CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV));;
e(REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]);;
e(EQ_TAC);;

 e(STRIP_TAC);;
 e(EXISTS_TAC "v0:vexp");;
 e(EXISTS_TAC "v:var");;
 e(UNDISCH_ALL_TAC);;
 e(DISCH_TAC THEN DISCH_TAC);;
 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(REWRITE_TAC[FVv_vsubst;IN_IMAGE]);;
 e(CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV));;
 e(REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]);;
 e(DISCH_THEN (REWRITE_THM o SYM));;
 e(IMP_RES_THEN ASM_REWRITE_THM ONE_ONE);;
 e(REPEAT DISCH_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(CONJ_TAC);;

  e(CONJ_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC[]);;

  e(GEN_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(EXISTS_TAC "v0 <|vv ss");;
 e(EXISTS_TAC "(ss:vsubst)v");;
 e(ASM_REWRITE_TAC[FVv_vsubst;IN_IMAGE]);;
 e(CONV_TAC (ONCE_DEPTH_CONV NOT_EXISTS_CONV));;
 e(REWRITE_TAC[DE_MORGAN_THM;(SYM o SPEC_ALL)IMP_DISJ_THM]);;
 e(IMP_RES_THEN ASM_REWRITE_THM ONE_ONE);;
 e(REPEAT CONJ_TAC
   THEN REPEAT GEN_TAC
   THEN DISCH_THEN (REWRITE_THM o SYM)
   THEN ASM_REWRITE_TAC[]);;

>%



let WFc_subst =
 prove_thm
  (`WFc_subst`,
   "!c g r ss. WFcsubst c r ss ==> (WFc (c <|c ss) (g <|g ss) r = WFc c g r)",
   REWRITE_TAC[WFcsubst]
   THEN INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[GVc;IN;IN_UNION;OR_IMPLIES_LEMMA]
   THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
   THEN REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[c_subst;WFc]
   THENL
    [ IMP_RES_THEN REWRITE_THM WFe_subst;

      FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ ASM_REWRITE_TAC[];

         FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_THEN REWRITE_THM WFb_subst
      THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ ASM_REWRITE_TAC[];

         FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_THEN REWRITE_THM WFb_subst
      THEN FIRST_ASSUM (USE_IMP_THEN REWRITE_THM)
      THENL
       [ ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM EXISTS_vsubst_LEMMA
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_THEN REWRITE_THM WFxs_subst
      THEN IMP_RES_THEN REWRITE_THM WFes_subst
      THEN IMP_RES_TAC DL_APPEND_xs_subst
      THEN ASM_REWRITE_TAC[LENGTH_xs_subst;LENGTH_es_subst]
    ]
  );;

% WFc_subst
g "!c g r ss. WFcsubst c r ss ==> (WFc (c <|c ss) (g <|g ss) r = WFc c g r)";;
e(REWRITE_TAC[WFcsubst]);;
e(INDUCT_THEN cmd_induct STRIP_ASSUME_TAC
   THEN REWRITE_TAC[GVc;IN;IN_UNION;OR_IMPLIES_LEMMA]
   THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
   THEN REPEAT STRIP_TAC);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;
e(IMP_RES_THEN REWRITE_THM WFe_subst);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;
e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

 e(ASM_REWRITE_TAC[]);;

 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
 e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;
e(IMP_RES_THEN REWRITE_THM WFb_subst);;
e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

 e(ASM_REWRITE_TAC[]);;

 e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;
 e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;
e(IMP_RES_THEN REWRITE_THM WFb_subst);;
e(FIRST_ASSUM (USE_IMP_THEN REWRITE_THM));;

 e(ASM_REWRITE_TAC[]);;

 e(USE_IMP_THEN REWRITE_THM EXISTS_vsubst_LEMMA);;
 e(ASM_REWRITE_TAC[]);;

e(ASM_REWRITE_TAC[c_subst;WFc]);;
e(IMP_RES_THEN REWRITE_THM WFxs_subst);;
e(IMP_RES_THEN REWRITE_THM WFes_subst);;
e(REWRITE_TAC[LENGTH_xs_subst;LENGTH_es_subst]);;
e(IMP_RES_TAC DL_APPEND_xs_subst);;
e(ASM_REWRITE_TAC[]);;
%



let SUBST_OF_WFx_KEEPS_WFx_LEMMA =
 TAC_PROOF
  (([],
    "!ys xs. (LENGTH xs = LENGTH ys) ==> WFxs xs /\ WFxs ys ==>
             (!x. WFx ((ys //v xs)x) = WFx x)"
   ),
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[WFxs;vsubst_THM;I_THM]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(x:var) = h'"
   THENL
     [ ASM_REWRITE_TAC[BND_MATCH];

       IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
       THEN ASM_CASES_TAC "x = (@z. (ys //v l')z = h)"
       THENL
         [ ASM_REWRITE_TAC[BND_MATCH]
           THEN FIRST_ASSUM (\th1. FIRST_ASSUM (\th2. ASSUME_TAC (MATCH_MP th1 th2)))
           THEN POP_ASSUM (\th. let cs = (conjuncts o fst o dest_imp o concl) th in
                     ASSUME_TAC (MP th (LIST_CONJ (map ASSUME cs))))
           THEN FIRST_ASSUM (REWRITE_THM o SYM o SPEC "@z. (ys //v l')z = h")
           THEN REWRITE_TAC[vsubst_INVERSE_SELECT]
           THEN ASM_REWRITE_TAC[];

           IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
           THEN FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO])
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

%< SUBST_OF_WFx_KEEPS_WFx_LEMMA
g "!ys xs. (LENGTH xs = LENGTH ys) ==> WFxs xs /\ WFxs ys ==>
           (!x. WFx ((ys //v xs)x) = WFx x)";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[WFxs;vsubst_THM;I_THM]);;

 e(REWRITE_TAC[WFxs;vsubst_THM;I_THM]);;
 e(REPEAT STRIP_TAC);;
 e(ASM_CASES_TAC "(x:var) = h'");;

  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
  e(ASM_CASES_TAC "x = (@z. (ys //v l')z = h)");;

   e(ASM_REWRITE_TAC[BND_MATCH]);;
   e(FIRST_ASSUM (\th1. FIRST_ASSUM (\th2. ASSUME_TAC (MATCH_MP th1 th2))));;
   e(POP_ASSUM (\th. let cs = (conjuncts o fst o dest_imp o concl) th in
                     ASSUME_TAC (MP th (LIST_CONJ (map ASSUME cs)))));;
%
   e(POP_ASSUM (\th. ASSUME_TAC (MATCH_MP th
                       (CONJ (ASSUME "WFxs l'") (ASSUME "WFxs ys")))));;
%
   e(FIRST_ASSUM (REWRITE_THM o SYM o SPEC "@z. (ys //v l')z = h"));;
   e(REWRITE_TAC[vsubst_INVERSE_SELECT]);;
   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
   e(FIRST_ASSUM (MATCH_MP_TAC o REWRITE_RULE[AND_IMP_INTRO]));;
   e(ASM_REWRITE_TAC[]);;
>%


let SUBST_OF_WFx_KEEPS_WFx =
 prove_thm
  (`SUBST_OF_WFx_KEEPS_WFx`,
   "!ys xs. (LENGTH xs = LENGTH ys) /\ WFxs xs /\ WFxs ys ==>
            (!x. WFx ((ys //v xs)x) = WFx x)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC SUBST_OF_WFx_KEEPS_WFx_LEMMA
   THEN ASM_REWRITE_TAC[]
  );;


let SUBST_INVERSE_OF_WFx_KEEPS_WFx =
 prove_thm
  (`SUBST_INVERSE_OF_WFx_KEEPS_WFx`,
   "!ys xs. (LENGTH xs = LENGTH ys) /\ WFxs xs /\ WFxs ys ==>
            (!x. WFx ((\y. @z. (ys //v xs)z = y)x) = WFx x)",
   REPEAT STRIP_TAC
   THEN BETA_TAC
   THEN (STRIP_ASSUME_TAC o
         SPEC "x:var" o
         REWRITE_RULE[ONTO_DEF] o
         SPECL["ys: var list";"xs: var list"]) vsubst_ONTO
   THEN IMP_RES_TAC SUBST_OF_WFx_KEEPS_WFx
   THEN ASM_REWRITE_TAC[vsubst_11;SELECT_REFL]
  );;

% SUBST_INVERSE_OF_WFx_KEEPS_WFx
g "!ys xs. (LENGTH xs = LENGTH ys) /\ WFxs xs /\ WFxs ys ==>
            (!x. WFx ((\y. @z. (ys //v xs)z = y)x) = WFx x)";;
e(REPEAT STRIP_TAC);;
e(BETA_TAC);;
e((STRIP_ASSUME_TAC o
   SPEC "x:var" o
   REWRITE_RULE[ONTO_DEF] o
   SPECL["ys: var list";"xs: var list"]) vsubst_ONTO);;
e(IMP_RES_TAC SUBST_OF_WFx_KEEPS_WFx);;
e(ASM_REWRITE_TAC[vsubst_11;SELECT_REFL]);;
%


let o_OF_WFx_KEEPS_WFx =
 prove_thm
  (`o_OF_WFx_KEEPS_WFx`,
   "!f g.
       (!x. WFx (f x) = WFx x) /\ (!x. WFx (g x) = WFx x) ==>
       (!x. WFx ((f o g)x) = WFx x)",
   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[o_THM]
  );;

% o_OF_WFx_KEEPS_WFx
g "!f g.
       (!x. WFx (f x) = WFx x) /\ (!x. WFx (g x) = WFx x) ==>
       (!x. WFx ((f o g)x) = WFx x)";;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[o_THM]);;
%



let WFcsubst_vsubst =
 prove_thm
  (`WFcsubst_vsubst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
     DISJOINT(SL vals')(SL(APPEND xs glbs))
     ==>
     WFcsubst c r 
        ( (APPEND xs  vals') //v (APPEND vars vals) )",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[WFcsubst]
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN CONJ_TAC
   THENL
     [ MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx
       THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN REWRITE_ALL_TAC[WFxs_APPEND]
       THEN ASM_REWRITE_TAC[LENGTH_APPEND];

       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN GEN_TAC
       THEN DISCH_TAC
       THEN MATCH_MP_TAC IDENT_vsubst
       THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
       THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
       THEN SUBGOAL_THEN "WFx x" ASSUME_TAC
       THENL
         [ MATCH_MP_TAC WFxs_IS_WFx
           THEN EXISTS_TAC "glbs:(var)list"
           THEN REWRITE_ALL_TAC[WFxs_APPEND]
           THEN ASM_REWRITE_TAC[];

           IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST
           THEN UNDISCH_TAC "~(x:var) IN (SL(APPEND vars vals))"
           THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN STRIP_TAC
           THEN ASM_REWRITE_TAC[]
           THEN IMP_RES_THEN STRIP_ASSUME_TAC IN_DISJOINT_IMP
           THEN POP_ASSUM MATCH_MP_TAC
           THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
         ]
     ]
  );;

%< WFcsubst_vsubst
g "!r p vars vals glbs pre post calls rec c vals' xs.
    (WFenv_syntax r) /\
    (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
    (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
    (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
    DISJOINT(SL vals')(SL(APPEND xs glbs))
    ==>
    WFcsubst c r 
       ( (APPEND xs  vals') //v (APPEND vars vals) )";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[WFcsubst]);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(GEN_TAC);;
 e(DISCH_TAC);;
 e(MATCH_MP_TAC IDENT_vsubst);;
 e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(SUBGOAL_THEN "WFx x" ASSUME_TAC);;

  e(MATCH_MP_TAC WFxs_IS_WFx);;
  e(EXISTS_TAC "glbs:(var)list");;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST);;
  e(UNDISCH_TAC "~(x:var) IN (SL(APPEND vars vals))");;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(IMP_RES_THEN STRIP_ASSUME_TAC IN_DISJOINT_IMP);;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;
>%



let WFcsubst_proc_subst =
 prove_thm
  (`WFcsubst_proc_subst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
    (WFenv_syntax r) /\
    (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
    (vals' = variants vals (SL(APPEND xs glbs))) /\
    (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) 
    ==>
    WFcsubst c r 
       ( (APPEND xs  vals') //v (APPEND vars vals) )",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[WFcsubst]
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN CONJ_TAC
   THENL
     [ MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx
       THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN REWRITE_ALL_TAC[WFxs_APPEND]
       THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;LENGTH_variants];

       IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN GEN_TAC
       THEN DISCH_TAC
       THEN MATCH_MP_TAC IDENT_vsubst
       THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
       THEN IMP_RES_TAC SUBSET_DEF
       THEN POP_ASSUM (\th. POP_TAC THEN POP_TAC THEN POP_TAC THEN ASSUME_TAC th)
       THEN SUBGOAL_THEN "WFx x" ASSUME_TAC
       THENL
         [ MATCH_MP_TAC WFxs_IS_WFx
           THEN EXISTS_TAC "glbs:(var)list"
           THEN REWRITE_ALL_TAC[WFxs_APPEND]
           THEN ASM_REWRITE_TAC[];

           IMP_RES_TAC IN_DISJOINT_LIST
           THEN UNDISCH_TAC "~(x:var) IN (SL(APPEND vars vals))"
           THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN STRIP_TAC
           THEN ASM_REWRITE_TAC[]
           THEN SUBGOAL_THEN "DL (APPEND (variants vals (SL (APPEND xs glbs)))
                                         (APPEND xs glbs)) " ASSUME_TAC
           THENL
             [ ONCE_REWRITE_TAC[DL_APPEND]
               THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants];

               IMP_RES_TAC IN_DISJOINT_LIST
               THEN POP_TAC
               THEN POP_ASSUM MATCH_MP_TAC
               THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
             ]
         ]
     ]
  );;

%< WFcsubst_proc_subst
g "!r p vars vals glbs pre post calls rec c vals' xs.
    (WFenv_syntax r) /\
    (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
    (vals' = variants vals (SL(APPEND xs glbs))) /\
    (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) 
    ==>
    WFcsubst c r 
       ( (APPEND xs  vals') //v (APPEND vars vals) )";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[WFcsubst]);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC SUBST_OF_WFx_KEEPS_WFx);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_APPEND;LENGTH_variants]);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(GEN_TAC);;
 e(DISCH_TAC);;
 e(MATCH_MP_TAC IDENT_vsubst);;
 e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(IMP_RES_TAC SUBSET_DEF);;
 e(POP_ASSUM (\th. POP_TAC THEN POP_TAC THEN POP_TAC THEN ASSUME_TAC th));;
 e(SUBGOAL_THEN "WFx x" ASSUME_TAC);;

  e(MATCH_MP_TAC WFxs_IS_WFx);;
  e(EXISTS_TAC "glbs:(var)list");;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC IN_DISJOINT_LIST);;
  e(UNDISCH_TAC "~(x:var) IN (SL(APPEND vars vals))");;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(SUBGOAL_THEN "DL (APPEND (variants vals (SL (APPEND xs glbs)))
                             (APPEND xs glbs)) " ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants]);;

   e(IMP_RES_TAC IN_DISJOINT_LIST);;
   e(POP_TAC);;
   e(POP_ASSUM MATCH_MP_TAC);;
   e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;
>%




let WFc_vsubst =
 prove_thm
  (`WFc_vsubst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
     DISJOINT(SL vals')(SL(APPEND xs glbs))
     ==>
     (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
   THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_vsubst)
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
     DISJOINT(SL vals')(SL(APPEND xs glbs))
     ==>
     (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
e(MATCH_MP_TAC (SPEC_ALL WFcsubst_vsubst));;
e(ASM_REWRITE_TAC[]);;
>%

let WFc_proc_subst =
 prove_thm
  (`WFc_proc_subst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs)
     ==>
     (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
   THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs)
     ==>
     (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
          (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
e(ASM_REWRITE_TAC[]);;
>%




let WFcsubst_standard_subst =
 prove_thm
  (`WFcsubst_standard_subst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
     (WFxs xs) /\ (WFxs vals') /\
     DL(APPEND xs vals') /\ DISJOINT (SL(APPEND xs vals')) (SL glbs)
     ==>
     let x = APPEND (APPEND vars vals) glbs in
     let x0 = APPEND (APPEND xs vals') glbs in
     let x' = logicals x in
     let x0' = logicals x0 in
     WFcsubst c r 
        ( (APPEND x0 x0') //v (APPEND x x') )",

   REPEAT STRIP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[WFcsubst]
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN CONJ_TAC
   THENL
    [ MATCH_MP_TAC vsubst_PRESERVES_WFx
      THEN ONCE_REWRITE_TAC[MAP_APPEND]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals];

         CONJ_TAC
         THENL
          [ MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] WFxs_IMP_MAP_EQ)
            THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals]
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN REWRITE_ALL_TAC[WFxs_APPEND]
            THEN ASM_REWRITE_TAC[];

            MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NOT_WFxs_IMP_MAP_EQ)
            THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals]
            THEN REWRITE_TAC[NOT_WFxs_logicals]
          ]
       ];

      IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN GEN_TAC
      THEN DISCH_TAC
      THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
      THEN SUBGOAL_THEN "DL(APPEND(APPEND xs vals')(glbs:(var)list))"
               STRIP_ASSUME_TAC
      THENL
       [ REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
         THENL
          [ ASM_REWRITE_TAC[LENGTH_APPEND]
            THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION];

            USE_IMP_THEN REWRITE_THM
                         (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
            THENL
             [ ONCE_REWRITE_ALL_TAC[DL_APPEND]
               THEN ASM_REWRITE_TAC[LENGTH_APPEND];

               REWRITE_TAC[IDENTITY_vsubst;I_THM]
             ]
          ]
       ]
    ]
  );;

%< WFcsubst_standard_subst
g "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
     (WFxs xs) /\ (WFxs vals') /\
     DL(APPEND xs vals') /\ DISJOINT (SL(APPEND xs vals')) (SL glbs)
     ==>
     let x = APPEND (APPEND vars vals) glbs in
     let x0 = APPEND (APPEND xs vals') glbs in
     let x' = logicals x in
     let x0' = logicals x0 in
     WFcsubst c r 
        ( (APPEND x0 x0') //v (APPEND x x') )";;
e(REPEAT STRIP_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[WFcsubst]);;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC vsubst_PRESERVES_WFx);;
 e(ONCE_REWRITE_TAC[MAP_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM APPEND_11);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals]);;

  e(CONJ_TAC);;

   e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] WFxs_IMP_MAP_EQ));;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals]);;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[]);;

   e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] NOT_WFxs_IMP_MAP_EQ));;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP;LENGTH_logicals]);;
   e(REWRITE_TAC[NOT_WFxs_logicals]);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(GEN_TAC);;
 e(DISCH_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(SUBGOAL_THEN "DL(APPEND(APPEND xs vals')(glbs:(var)list))"
        STRIP_ASSUME_TAC);;

  e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

   e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
   e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

   e(USE_IMP_THEN REWRITE_THM
         (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;

    e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
    e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

    e(REWRITE_TAC[IDENTITY_vsubst;I_THM]);;
>%




let FVc_vsubst =
 prove_thm
  (`FVc_vsubst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
     DISJOINT(SL vals')(SL(APPEND xs glbs))
     ==>
     (FVc (c <|c ((APPEND xs  vals') //v (APPEND vars vals))) r SUBSET
      SL(APPEND (APPEND xs vals') glbs))",

   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM FVc_subst
   THENL
     [ EXISTS_TAC "glbs:(var)list"
       THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN ASM_REWRITE_TAC[]
       THEN MATCH_MP_TAC IDENT_MAP_vsubst
       THEN ONCE_REWRITE_ALL_TAC[DL_APPEND]
       THEN ASM_REWRITE_TAC[SL_APPEND;DISJOINT_UNION]
       THEN ONCE_REWRITE_ALL_TAC[DISJOINT_SYM]
       THEN ONCE_REWRITE_ALL_TAC[DISJOINT_APPEND]
       THEN ASM_REWRITE_TAC[];

       REWRITE_TAC[SUBSET_DEF]
       THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN REWRITE_ALL_TAC[SUBSET_DEF;SL_APPEND;IN_UNION;DE_MORGAN_THM]
       THEN REWRITE_TAC[IN_IMAGE]
       THEN GEN_TAC
       THEN STRIP_TAC
       THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND])
                  "DL(APPEND xs glbs:(var)list)"
       THEN UNASSUME_THEN (STRIP_ASSUME_TAC o
                           REWRITE_RULE[DL_APPEND;SL_APPEND;DISJOINT_UNION])
                   "DL(APPEND(APPEND vars vals)glbs:(var)list)"
       THEN RES_TAC
       THENL
         [ ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM
                       (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
           THENL
             [ ASM_REWRITE_TAC[];

               IMP_RES_TAC vsubst_ELEMENTS
               THEN ASM_REWRITE_TAC[]
             ];

           ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM
                       (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
           THENL
             [ ASM_REWRITE_TAC[]
               THEN REWRITE_ALL_TAC[DISJOINT_UNION2]
               THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
               THEN ASM_REWRITE_TAC[];

               USE_IMP_THEN ASM_REWRITE_THM
                     (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS)
               THEN MATCH_MP_TAC
                     (REWRITE_RULE[AND_IMP_INTRO]
                        (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS))
               THEN ASM_REWRITE_TAC[]
             ];

           ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM IDENT_vsubst
           THEN IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_IMP
           THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN IMP_RES_TAC IN_DISJOINT_IMP
           THEN FIRST_ASSUM MATCH_MP_TAC
           THEN ASM_REWRITE_TAC[IN_UNION]
         ]
     ]
  );;

%< FVc_vsubst
g "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
     (WFxs vals') /\ (LENGTH vals' = LENGTH vals) /\ (DL vals') /\
     DISJOINT(SL vals')(SL(APPEND xs glbs))
     ==>
     (FVc (c <|c ((APPEND xs  vals') //v (APPEND vars vals))) r SUBSET
      SL(APPEND (APPEND xs vals') glbs))";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM FVc_subst);;

 e(EXISTS_TAC "glbs:(var)list");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e(MATCH_MP_TAC IDENT_MAP_vsubst);;
 e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
 e(ASM_REWRITE_TAC[SL_APPEND;DISJOINT_UNION]);;
 e(ONCE_REWRITE_ALL_TAC[DISJOINT_SYM]);;
 e(ONCE_REWRITE_ALL_TAC[DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[SUBSET_DEF]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[SUBSET_DEF;SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(REWRITE_TAC[IN_IMAGE]);;
 e(GEN_TAC);;
 e(STRIP_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND])
         "DL(APPEND xs glbs:(var)list)");;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[DL_APPEND;SL_APPEND;DISJOINT_UNION])
         "DL(APPEND(APPEND vars vals)glbs:(var)list)");;
 e(RES_TAC);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_TAC vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;

   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_ALL_TAC[DISJOINT_UNION2]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(ASM_REWRITE_TAC[]);;

   e(USE_IMP_THEN ASM_REWRITE_THM
           (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS));;
   e(MATCH_MP_TAC
           (REWRITE_RULE[AND_IMP_INTRO]
              (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS)));;
   e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM IDENT_vsubst);;
  e(IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_IMP);;
  e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(IMP_RES_TAC IN_DISJOINT_IMP);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(ASM_REWRITE_TAC[IN_UNION]);;
>%



let FVc_proc_subst =
 prove_thm
  (`FVc_proc_subst`,
   "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) 
     ==>
     (FVc (c <|c ((APPEND xs  vals') //v (APPEND vars vals))) r SUBSET
      SL(APPEND (APPEND xs vals') glbs))",
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM FVc_subst
   THENL
     [ EXISTS_TAC "glbs:(var)list"
       THEN CONJ_TAC
       THENL
         [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX;

           MATCH_MP_TAC IDENT_MAP_vsubst
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN ONCE_REWRITE_ALL_TAC[DL_APPEND]
           THEN ASM_REWRITE_TAC[SL_APPEND;DISJOINT_UNION]
           THEN REWRITE_TAC[DISJOINT_variants_UNION]
         ];

       REWRITE_TAC[SUBSET_DEF]
       THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
       THEN REWRITE_ALL_TAC[SUBSET_DEF;SL_APPEND;IN_UNION;DE_MORGAN_THM]
       THEN REWRITE_TAC[IN_IMAGE]
       THEN GEN_TAC
       THEN STRIP_TAC
       THEN UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND])
                  "DL(APPEND xs glbs:(var)list)"
       THEN UNASSUME_THEN (STRIP_ASSUME_TAC o
                           REWRITE_RULE[DL_APPEND;SL_APPEND;DISJOINT_UNION])
                   "DL(APPEND(APPEND vars vals)glbs:(var)list)"
       THEN RES_TAC
       THENL
         [ ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM
                       (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
           THENL
             [ ASM_REWRITE_TAC[];

               IMP_RES_TAC vsubst_ELEMENTS
               THEN ASM_REWRITE_TAC[]
             ];

           ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM
                       (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
           THENL
             [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
               THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
               THEN REWRITE_TAC[DISJOINT_variants_UNION];

               USE_IMP_THEN ASM_REWRITE_THM
                     (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS)
               THEN MATCH_MP_TAC
                     (REWRITE_RULE[AND_IMP_INTRO]
                        (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS))
               THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
             ];

           ASM_REWRITE_TAC[]
           THEN USE_IMP_THEN ASM_REWRITE_THM IDENT_vsubst
           THEN IMP_RES_TAC IN_DISJOINT_IMP
           THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN STRIP_ASSUME_TAC((SPECL["vals:(var)list";
                                        "(SL xs):(var)set";"(SL glbs):(var)set"])
                                 DISJOINT_variants_UNION)
           THEN IMP_RES_TAC IN_DISJOINT_IMP
         ]
     ]
  );;

%< FVc_proc_subst
g "!r p vars vals glbs pre post calls rec c vals' xs.
     (WFenv_syntax r) /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) 
     ==>
     (FVc (c <|c ((APPEND xs  vals') //v (APPEND vars vals))) r SUBSET
      SL(APPEND (APPEND xs vals') glbs))";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM FVc_subst);;

 e(EXISTS_TAC "glbs:(var)list");;
 e(CONJ_TAC);;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

  e(MATCH_MP_TAC IDENT_MAP_vsubst);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
  e(ASM_REWRITE_TAC[SL_APPEND;DISJOINT_UNION]);;
  e(REWRITE_TAC[DISJOINT_variants_UNION]);;

 e(REWRITE_TAC[SUBSET_DEF]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[SUBSET_DEF;SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
 e(REWRITE_TAC[IN_IMAGE]);;
 e(GEN_TAC);;
 e(STRIP_TAC);;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[DL_APPEND])
         "DL(APPEND xs glbs:(var)list)");;
 e(UNASSUME_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[DL_APPEND;SL_APPEND;DISJOINT_UNION])
         "DL(APPEND(APPEND vars vals)glbs:(var)list)");;
 e(RES_TAC);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

   e(ASM_REWRITE_TAC[]);;

   e(IMP_RES_TAC vsubst_ELEMENTS);;
   e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;

   e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants_UNION]);;

   e(USE_IMP_THEN ASM_REWRITE_THM
           (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS));;
   e(MATCH_MP_TAC
           (REWRITE_RULE[AND_IMP_INTRO]
              (CONV_RULE (DEPTH_CONV RIGHT_IMP_FORALL_CONV) vsubst_ELEMENTS)));;
   e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN ASM_REWRITE_THM IDENT_vsubst);;
  e(IMP_RES_TAC IN_DISJOINT_IMP);;
  e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(STRIP_ASSUME_TAC((SPECL["vals:(var)list";
                            "(SL xs):(var)set";"(SL glbs):(var)set"])
                     DISJOINT_variants_UNION));;
  e(IMP_RES_TAC IN_DISJOINT_IMP);;
>%




let C_PROC_FREE_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' xs (s2:state) s3 z.
      (WFenv_syntax r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL(APPEND xs glbs))) /\
      (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
      (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
           (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r /\
       WFenv_syntax r ==>
        (!z.
          ~z IN (FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r) ==>
          (s3 z = BNDsL ns vals' s2 z))) /\
      ~(z IN (SL xs)) /\ ~(z IN (SL glbs))
      ==>
      (BNDsL (MAP s2 vals') vals' s3 z = s2 z)"
   ),
   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(z:var) IN (SL vals')"
   THENL
     [ IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH;

       IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
       THEN FIRST_ASSUM (SUBST1_TAC o SYM
                         o ISPECL["ns:(num)list";"s2:state"]
                         o MATCH_MP BNDsL_NOT_MATCH)
       THEN SUBGOAL_THEN "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                              (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
                  ASSUME_TAC
       THENL
         [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN USE_IMP_THEN ASM_REWRITE_THM WFc_subst
           THEN MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst)
           THEN ASM_REWRITE_TAC[];

           RES_TAC
           THEN POP_ASSUM MATCH_MP_TAC
           THEN IMP_RES_TAC FVc_proc_subst
           THEN MATCH_MP_TAC NOT_IN_SUBSET
           THEN EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):(var)set"
           THEN POP_ASSUM REWRITE_THM
           THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
           THEN ASM_REWRITE_TAC[]
         ]
     ]
  );;

%< C_PROC_FREE_ONLY_IMPACT_LEMMA
g "!r p vars vals glbs pre post calls rec c vals' xs (s2:state) s3 z.
      (WFenv_syntax r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL(APPEND xs glbs))) /\
      (WFxs xs) /\ (LENGTH vars = LENGTH xs) /\ DL(APPEND xs glbs) /\
      (WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
           (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r /\
       WFenv_syntax r ==>
        (!z.
          ~z IN (FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r) ==>
          (s3 z = BNDsL ns vals' s2 z))) /\
      ~(z IN (SL xs)) /\ ~(z IN (SL glbs))
      ==>
      (BNDsL (MAP s2 vals') vals' s3 z = s2 z)";;
 e(REPEAT STRIP_TAC);;
 e(ASM_CASES_TAC "(z:var) IN (SL vals')");;

  e(IMP_RES_THEN ASM_REWRITE_THM BNDsL_MAP_MATCH);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(FIRST_ASSUM (SUBST1_TAC o SYM
                 o ISPECL["ns:(num)list";"s2:state"]
                 o MATCH_MP BNDsL_NOT_MATCH));;
  e(SUBGOAL_THEN "WFc(c <|c ((APPEND xs vals') //v (APPEND vars vals)))r"
         ASSUME_TAC);;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(USE_IMP_THEN ASM_REWRITE_THM WFc_subst);;
   e(MATCH_MP_TAC (SPEC_ALL WFcsubst_proc_subst));;
   e(ASM_REWRITE_TAC[]);;

   e(RES_TAC);;
   e(POP_ASSUM MATCH_MP_TAC);;
   e(IMP_RES_TAC FVc_proc_subst);;
   e(MATCH_MP_TAC NOT_IN_SUBSET);;
   e(EXISTS_TAC "SL(APPEND(APPEND xs vals')glbs):(var)set");;
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[SL_APPEND]);;
   e(REWRITE_TAC[IN_UNION;DE_MORGAN_THM]);;
   e(ASM_REWRITE_TAC[]);;
>%



% ================================================================== %
% Here is the C_FREE_ONLY_IMPACT theorem we could not prove earlier. %
% ================================================================== %

let C_FREE_ONLY_IMPACT_LEMMA =
 TAC_PROOF
  (([],
    "!c r s1 s2.
      C c r s1 s2 ==>
      (!g. WFc c g r /\ WFenv_syntax r ==>
           (!z. (~z IN FVc c r) ==>
                (s2 z = s1 z)))"
   ),
   CMD_INDUCT_TAC
   THEN REWRITE_ALL_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]
   THEN ASM_REWRITE_TAC[FVxs_SL;o_THM]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC (INST_TYPE [":num",":*"] BND_NOT_MATCH)
      THEN IMP_RES_TAC E_FREE_ONLY_IMPACT
      THEN ASM_REWRITE_TAC[];

      RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      IMP_RES_TAC B_FREE_ONLY_IMPACT
      THEN FIRST_ASSUM (\th. (SPEC "g:prog_env" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC_ALL)
      THEN SUBGOAL_THEN  "?v x. (pr = ALESS v (AVAR x)) /\ ~WFx x /\
                                (~x IN (FVa a) /\ ~x IN (FVv v)) /\
                                (!p. ~x IN (FVa ((g:prog_env) p)))"
               REWRITE_THM
      THENL
       [ EXISTS_TAC "v:vexp"
         THEN EXISTS_TAC "x:var"
         THEN ASM_REWRITE_TAC[aexp_one_one;vexp_one_one];

         ASM_REWRITE_TAC[]
         THEN DISCH_TAC
         THEN RES_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_TAC B_FREE_ONLY_IMPACT;

      REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
      THEN IMP_RES_TAC ES_FREE_ONLY_IMPACT
      THEN POP_ASSUM (REWRITE_THM o SYM)
      THEN MATCH_MP_TAC (SPEC_ALL C_PROC_FREE_ONLY_IMPACT_LEMMA)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< C_FREE_ONLY_IMPACT_LEMMA:
g "!c r s1 s2.
      C c r s1 s2 ==>
      (!g. WFc c g r /\ WFenv_syntax r ==>
           (!z. (~z IN FVc c r) ==>
                (s2 z = s1 z)))";;
e(CMD_INDUCT_TAC);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (INST_TYPE [":num",":*"] BND_NOT_MATCH));;
 e(IMP_RES_TAC E_FREE_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_ALL_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
 e(FIRST_ASSUM (\th. SPEC "g:prog_env" th; UNDISCH_TAC (concl th)));;
 e(DISCH_THEN (MP_TAC o SPEC_ALL));;
 e(SUBGOAL_THEN  "?v x. (pr = ALESS v (AVAR x)) /\ ~WFx x /\
                        (~x IN (FVa a) /\ ~x IN (FVv v)) /\
                        (!p. ~x IN (FVa ((g:prog_env) p)))"
       REWRITE_THM);;

  e(EXISTS_TAC "v:vexp"
     THEN EXISTS_TAC "x:var");;
  e(ASM_REWRITE_TAC[aexp_one_one;vexp_one_one]);;

  e(ASM_REWRITE_TAC[]);;
  e(DISCH_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;

 e(REWRITE_TAC[WFc;FVc_DEF;IN;IN_UNION;DE_MORGAN_THM]);;
 e(ASM_REWRITE_TAC[FVxs_SL;o_THM]);;
 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
 e(IMP_RES_TAC ES_FREE_ONLY_IMPACT);;
 e(POP_ASSUM (REWRITE_THM o SYM));;
 e(MATCH_MP_TAC (SPEC_ALL C_PROC_FREE_ONLY_IMPACT_LEMMA));;
 e(ASM_REWRITE_TAC[]);;
>%


let C_FREE_ONLY_IMPACT =
 prove_thm
  (`C_FREE_ONLY_IMPACT`,
   "!c r s1 s2 g z.
      WFc c g r /\ WFenv_syntax r /\
      C c r s1 s2 /\
      (~z IN FVc c r) ==>
      (s2 z = s1 z)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC C_FREE_ONLY_IMPACT_LEMMA
  );;

%< C_FREE_ONLY_IMPACT:
g "!c r s1 s2 g z.
      WFc c g r /\ WFenv_syntax r /\
      C c r s1 s2 /\
      (~z IN FVc c r) ==>
      (s2 z = s1 z)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC C_FREE_ONLY_IMPACT_LEMMA);;
>%


let C_FREE_ONLY_VARIATION =
 prove_thm
  (`C_FREE_ONLY_VARIATION`,
   "!x c g r s1 s2.
     C c r s1 s2 /\
     WFenv_syntax r /\ WFc c g r /\ FVc c r SUBSET SL x ==>
     (BNDsL (MAP s2 x) x s1 = s2)",
   REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL x"
   THENL
     [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

       IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
       THEN IMP_RES_TAC NOT_IN_SUBSET
       THEN IMP_RES_TAC C_FREE_ONLY_IMPACT
       THEN ASM_REWRITE_TAC[]
     ]
  );;

% C_FREE_ONLY_VARIATION:
g "!x c g r s1 s2.
     C c r s1 s2 /\
     WFenv_syntax r /\ WFc c g r /\ FVc c r SUBSET SL x ==>
     (BNDsL (MAP s2 x) x s1 = s2)";;

e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL x");;

 e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

 e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
 e(IMP_RES_TAC NOT_IN_SUBSET);;
 e(IMP_RES_TAC C_FREE_ONLY_IMPACT);;
 e(ASM_REWRITE_TAC[]);;
%



let BNDsL_MAP_state_SYM =
 prove_thm
  (`BNDsL_MAP_state_SYM`,
   "!vars vals glbs c g r s1 s2.
      WFenv_syntax r /\ WFc c g r /\
      DL (APPEND (APPEND vars vals) glbs) /\
      (FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs)) /\
      C c r s1 s2
      ==>
      (BNDsL (MAP s2 (APPEND vars glbs)) (APPEND vars glbs) s1  =
       BNDsL (MAP s1 vals) vals s2)",
   REWRITE_TAC[SUBSET_DEF]
   THEN REPEAT STRIP_TAC
   THEN EXT_TAC "z:var"
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
   THENL
     [  UNDISCH_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
        THEN STRIP_TAC
        THEN IMP_RES_TAC IN_DISJOINT_LIST3
        THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
        THENL [ USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH;
                USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH;
                USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH ]
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION];

        FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "z:var")
        THEN UNDISCH_TAC "~(z:var) IN SL (APPEND (APPEND vars vals) glbs)"
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
        THEN STRIP_TAC
        THEN IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN IMP_RES_TAC C_FREE_ONLY_IMPACT
        THEN ASM_REWRITE_TAC[]
        THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
        THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION]
     ]
  );;

%<
g "!vars vals glbs c g r s1 s2.
      WFenv_syntax r /\ WFc c g r /\
      DL (APPEND (APPEND vars vals) glbs) /\
      (FVc c r) SUBSET (SL (APPEND (APPEND vars vals) glbs)) /\
      C c r s1 s2
      ==>
      (BNDsL (MAP s2 (APPEND vars glbs)) (APPEND vars glbs) s1  =
       BNDsL (MAP s1 vals) vals s2)";;

e(REWRITE_TAC[SUBSET_DEF]);;
e(REPEAT STRIP_TAC);;
e(EXT_TAC "z:var");;
e(GEN_TAC);;
e(ASM_CASES_TAC "(z:var) IN SL (APPEND (APPEND vars vals) glbs)");;

  e(POP_ASSUM (\th. ASSUME_TAC th THEN UNDISCH_TAC (concl th)));;
  e(REWRITE_TAC[SL_APPEND;IN_UNION]);;
  e(STRIP_TAC);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

    e(IMP_RES_TAC IN_DISJOINT_LIST3);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;

  e(FIRST_ASSUM (IMP_RES_TAC o CONTRAPOS o SPEC "z:var"));;
  e(UNDISCH_TAC "~(z:var) IN SL (APPEND (APPEND vars vals) glbs)");;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
  e(STRIP_TAC);;
  e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(IMP_RES_TAC C_FREE_ONLY_IMPACT);;
  e(ASM_REWRITE_TAC[]);;
  e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
  e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION]);;
>%





close_theory ();;

