% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the proofs of theorems in the form of
  Floyd/Hoare-style programming language axioms and rules of inference.
  Versions are presented for a) partial correctness, b) termination,
  and c) total correctness.

  "You must be perfect, just as your Father in heaven is perfect."

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `hoare_rules`;;

new_parent `cmd_semantics`;;


% --------------------------------------------------------------------- %
% Need the finite_sets and ind_defs libraries.                          %
% --------------------------------------------------------------------- %

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of hoare_rules %

loadf `autoload`;;

autoload_all_defs_and_thms
     `more_finite_sets bindings variables assert_syntax 
      assert_semantics prog_syntax prog_semantics variants 
      free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;




% Program language rewriting and simplification tools. %

let E_REWRITE_RULE = REWRITE_RULE [ exp_distinct; exp_one_one]
                     o CONV_RULE (REWR_CONV E_cases) ;;
let ES_REWRITE_RULE = ONCE_REWRITE_RULE[ES_DEF] ;;
let B_REWRITE_RULE = REWRITE_RULE [bexp_distinct;bexp_one_one]
                     o CONV_RULE (REWR_CONV B_cases) ;;
let C_REWRITE_RULE = REWRITE_RULE [ cmd_distinct; cmd_one_one]
                     o CONV_RULE (REWR_CONV C_cases) ;;

let E_SIMPLIFY_RULE =
    REWRITE_RULE [VE_THM;VE1_DEF;VE_DEF;VE_state_DEF;V_DEF;ETA_AX];;
let ES_SIMPLIFY_RULE =
    REWRITE_RULE [VES_THM;VES1_DEF;VES_DEF;VES_state_DEF;V_DEF;ETA_AX];;
let B_SIMPLIFY_RULE =
    REWRITE_RULE [AB_THM;AB1_DEF;AB_DEF;AB_state_DEF;A_DEF;ETA_AX;V_DEF];;

let imap f = map (\x. f x ? x);;
let E_rewrite = LIST_CONJ o (imap E_REWRITE_RULE) o CONJUNCTS;;
let ES_rewrite = LIST_CONJ o (imap ES_REWRITE_RULE) o CONJUNCTS;;
let B_rewrite = LIST_CONJ o (imap B_REWRITE_RULE) o CONJUNCTS;;
let C_rewrite = LIST_CONJ o (imap C_REWRITE_RULE) o CONJUNCTS;;

let E_simplify = LIST_CONJ o (imap E_SIMPLIFY_RULE) o CONJUNCTS;;
let ES_simplify = LIST_CONJ o (imap ES_SIMPLIFY_RULE) o CONJUNCTS;;
let B_simplify = LIST_CONJ o (imap B_SIMPLIFY_RULE) o CONJUNCTS;;





%=================================%
%=================================%
%   Now for Hoare-style axioms!   %
%=================================%
%=================================%




let ae_pre_THM =
 prove_thm
  (`ae_pre_THM`,
   "!e q. ESPEC(ae_pre e q, e, q)",
   REWRITE_TAC[ESPEC;ae_pre;VE_THM;a_subst_THM]
    THEN REPEAT STRIP_TAC
    THENL
      [ ASM_REWRITE_TAC[];

        EXISTS_TAC "V(VE e)s1"
        THEN EXISTS_TAC "s1 <|s (VE_state e)"
        THEN REWRITE_TAC[]
      ]
  );;

%
g "!e q. ESPEC(ae_pre e q, e, q)";;
e(REWRITE_TAC[ESPEC;ae_pre;VE_THM;a_subst_THM]);;
e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "V(VE e)s1");;
 e(EXISTS_TAC "s1 <|s (VE_state e)");;
 e(REWRITE_TAC[]);;
%


let ab_pre_THM =
 prove_thm
  (`ab_pre_THM`,
   "!b q. BSPEC(ab_pre b q, b, q)",
   REWRITE_TAC[BSPEC;ab_pre;AB_THM;a_subst_THM]
    THEN REPEAT STRIP_TAC
    THENL
      [ ASM_REWRITE_TAC[];

        EXISTS_TAC "A(AB b)s1"
        THEN EXISTS_TAC "s1 <|s (AB_state b)"
        THEN REWRITE_TAC[]
      ]
  );;

%
g "!b q. BSPEC(ab_pre b q, b, q)";;
e(REWRITE_TAC[BSPEC;ab_pre;AB_THM;a_subst_THM]);;
e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "A(AB b)s1");;
 e(EXISTS_TAC "s1 <|s (AB_state b)");;
 e(REWRITE_TAC[]);;
%


% ========================== %
% Precondition strengthening %
% ========================== %


let PRE_STRENGTH_ESPEC =
 prove_thm
  (`PRE_STRENGTH_ESPEC`,
   "!p q r e.
                    CLOSE(AIMP p q) /\ ESPEC(q,e,r)
            %---------------------------------------------------% ==>
                              ESPEC(p,e,r)",

   REWRITE_TAC[CLOSE;ESPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r e.
                    CLOSE(AIMP p q) /\ ESPEC(q,e,r)
            %---------------------------------------------------% ==>
                              ESPEC(p,e,r)";;
e(REWRITE_TAC[CLOSE;ESPEC;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


let PRE_STRENGTH_ES_SPEC =
 prove_thm
  (`PRE_STRENGTH_ES_SPEC`,
   "!p q r es.
                    CLOSE(AIMP p q) /\ ES_SPEC(q,es,r)
            %---------------------------------------------------% ==>
                              ES_SPEC(p,es,r)",

   REWRITE_TAC[CLOSE;ES_SPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;


let PRE_STRENGTH_BSPEC =
 prove_thm
  (`PRE_STRENGTH_BSPEC`,
   "!p q r b.
                    CLOSE(AIMP p q) /\ BSPEC(q,b,r)
            %---------------------------------------------------% ==>
                              BSPEC(p,b,r)",

   REWRITE_TAC[CLOSE;BSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;


let PRE_STRENGTH_PCSPEC =
 prove_thm
  (`PRE_STRENGTH_PCSPEC`,
   "!p q s c r.
                    CLOSE(AIMP p q) /\ PCSPEC(q,c,s,r)
            %---------------------------------------------------% ==>
                              PCSPEC(p,c,s,r)",

   REWRITE_TAC[CLOSE;PCSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;

let PRE_STRENGTH_TCSPEC =
 prove_thm
  (`PRE_STRENGTH_TCSPEC`,
   "!p q c r.
                    CLOSE(AIMP p q) /\ TCSPEC(q,c,r)
            %---------------------------------------------------% ==>
                              TCSPEC(p,c,r)",

   REWRITE_TAC[CLOSE;TCSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

let PRE_STRENGTH_CSPEC =
 prove_thm
  (`PRE_STRENGTH_CSPEC`,
   "!p q s c r.
                    CLOSE(AIMP p q) /\ CSPEC(q,c,s,r)
            %---------------------------------------------------% ==>
                              CSPEC(p,c,s,r)",

   REWRITE_TAC[CLOSE;CSPEC_THM;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q s c r.
                    CLOSE(AIMP p q) /\ CSPEC(q,c,s,r)
            %---------------------------------------------------% ==>
                              CSPEC(p,c,s,r)";;
e(REWRITE_TAC[CLOSE;CSPEC_THM;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


% ======================= %
% Postcondition weakening %
% ======================= %


let POST_WEAK_ESPEC =
 prove_thm
  (`POST_WEAK_ESPEC`,
   "!p q r e.
                   ESPEC(p,e,q)  /\  CLOSE(AIMP q r)
            %---------------------------------------------------% ==>
                             ESPEC(p,e,r)",

   REWRITE_TAC[CLOSE;ESPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r e.
                   ESPEC(p,e,q)  /\  CLOSE(AIMP q r)
            %---------------------------------------------------% ==>
                              ESPEC(p,e,r)";;
e(REWRITE_TAC[CLOSE;ESPEC;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


let POST_WEAK_ES_SPEC =
 prove_thm
  (`POST_WEAK_ES_SPEC`,
   "!p q r es.
                   ES_SPEC(p,es,q)  /\  CLOSE(AIMP q r)
            %---------------------------------------------------% ==>
                             ES_SPEC(p,es,r)",

   REWRITE_TAC[CLOSE;ES_SPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;


let POST_WEAK_BSPEC =
 prove_thm
  (`POST_WEAK_BSPEC`,
   "!p q r b.
                   BSPEC(p,b,q)  /\  CLOSE(AIMP q r)
            %---------------------------------------------------% ==>
                              BSPEC(p,b,r)",

   REWRITE_TAC[CLOSE;BSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;



let POST_WEAK_PCSPEC =
 prove_thm
  (`POST_WEAK_PCSPEC`,
   "!p q s c r.
                  PCSPEC(p,c,q,r)  /\  CLOSE(AIMP q s)
            %---------------------------------------------------% ==>
                             PCSPEC(p,c,s,r)",

   REWRITE_TAC[CLOSE;PCSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
  );;


let POST_WEAK_CSPEC =
 prove_thm
  (`POST_WEAK_CSPEC`,
   "!p q s c r.
                  CSPEC(p,c,q,r)  /\  CLOSE(AIMP q s)
            %---------------------------------------------------% ==>
                             CSPEC(p,c,s,r)",

   REWRITE_TAC[CLOSE;CSPEC_THM;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q s c r.
                  CSPEC(p,c,q,r)  /\  CLOSE(AIMP q s)
            %---------------------------------------------------% ==>
                             CSPEC(p,c,s,r)";;
e(REWRITE_TAC[CLOSE;CSPEC_THM;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


% ===================================== %
% We should have the axiom {false}c{q}. %
% ===================================== %


let FALSE_ESPEC =
 prove_thm
  (`FALSE_ESPEC`,
   "!e q.
                                                                
            %---------------------------------------------------%
                            ESPEC(AFALSE,e,q)",

   REWRITE_TAC[ESPEC;A_DEF]
  );;


let FALSE_ES_SPEC =
 prove_thm
  (`FALSE_ES_SPEC`,
   "!es q.
                                                                
            %---------------------------------------------------%
                            ES_SPEC(AFALSE,es,q)",

   REWRITE_TAC[ES_SPEC;A_DEF]
  );;


let FALSE_BSPEC =
 prove_thm
  (`FALSE_BSPEC`,
   "!b q.
                                                                
            %---------------------------------------------------%
                            BSPEC(AFALSE,b,q)",

   REWRITE_TAC[BSPEC;A_DEF]
  );;


let FALSE_PCSPEC =
 prove_thm
  (`FALSE_PCSPEC`,
   "!c r q.
                                                                
            %---------------------------------------------------%
                            PCSPEC(AFALSE,c,q,r)",

   REWRITE_TAC[PCSPEC;A_DEF]
  );;


let FALSE_CSPEC =
 prove_thm
  (`FALSE_CSPEC`,
   "!c r q.
                                                                
            %---------------------------------------------------%
                            CSPEC(AFALSE,c,q,r)",

   REWRITE_TAC[CSPEC_THM;A_DEF]
  );;

%<
g "!e v q.
                                                                
            %---------------------------------------------------%
                            ESPEC(AFALSE,e,v,q)";;
e(REWRITE_TAC[ESPEC;A_DEF]);;
>%


% ================================================================ %
% Now we can get into the particular axioms and rules of inference %
% for this programming language.                                   %
% ================================================================ %


let NUM_THM =
 prove_thm
  (`NUM_THM`,
   "!q n. 
            %---------------------------------------------------%
                          ESPEC(q, NUM n, q)",

   REWRITE_TAC[ESPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite)
      THEN ASM_REWRITE_TAC[];

      EXISTS_TAC "n:num"
      THEN EXISTS_TAC "s1:state"
      THEN NUM_TAC
    ]
  );;

%<
g "!q n. 
            %---------------------------------------------------%
                          ESPEC(q, NUM n, q)";;

e(REWRITE_TAC[ESPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite));;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "n:num");;
 e(EXISTS_TAC "s1:state");;
 e(NUM_TAC);;
>%


let PVAR_THM =
 prove_thm
  (`PVAR_THM`,
   "!q x. 
            %---------------------------------------------------%
                            ESPEC(q, PVAR x, q)",

   REWRITE_TAC[ESPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite)
      THEN ASM_REWRITE_TAC[];

      EXISTS_TAC "(s1:state)x"
      THEN EXISTS_TAC "s1:state"
      THEN PVAR_TAC
    ]
  );;

%<
g "!q x. 
            %---------------------------------------------------%
                            ESPEC(q, PVAR x, q)";;

e(REWRITE_TAC[ESPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite));;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "(s1:state)x");;
 e(EXISTS_TAC "s1:state");;
 e(PVAR_TAC);;
>%


let INC_THM =
 prove_thm
  (`INC_THM`,
   "!q x. 
            %---------------------------------------------------%
           ESPEC(q <|a (APLUS (AVAR x) (ANUM 1) / x), INC x, q)",

   REWRITE_TAC[ESPEC;single_subst;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IDENT_SS_var] o E_simplify)
      THEN ASM_REWRITE_TAC[];

      EXISTS_TAC "(s1:state)x + 1"
      THEN EXISTS_TAC "BND ((s1:state)x + 1) x (s1:state)"
      THEN INC_TAC
      THEN PVAR_TAC
    ]
  );;

%<
g "!q x. 
            %---------------------------------------------------%
           ESPEC(q <|a (APLUS (AVAR x) (ANUM 1) / x), INC x, q)";;

e(REWRITE_TAC[ESPEC]);;
e(REWRITE_TAC[single_subst;a_subst_THM]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IDENT_SS_var] o E_simplify));;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "(s1:state)x + 1");;
 e(EXISTS_TAC "BND ((s1:state)x + 1) x (s1:state)");;
 e(INC_TAC);;
 e(PVAR_TAC);;
>%



let PLUS_THM =
 prove_thm
  (`PLUS_THM`,
   "!p q r e1 e2. 
                  ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                        ESPEC(p, PLUS e1 e2, q)",

   REWRITE_TAC[ESPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "n + n'"
      THEN EXISTS_TAC "s2':state"
      THEN PLUS_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r e1 e2. 
                  ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                        ESPEC(p, PLUS e1 e2, q)";;

e(REWRITE_TAC[ESPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "n + n'");;
 e(EXISTS_TAC "s2':state");;
 e(PLUS_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let MINUS_THM =
 prove_thm
  (`MINUS_THM`,
   "!p q r e1 e2. 
                  ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                        ESPEC(p, MINUS e1 e2, q)",

   REWRITE_TAC[ESPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "n - n'"
      THEN EXISTS_TAC "s2':state"
      THEN MINUS_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;



let MULT_THM =
 prove_thm
  (`MULT_THM`,
   "!p q r e1 e2. 
                  ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                        ESPEC(p, MULT e1 e2, q)",

   REWRITE_TAC[ESPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o E_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "n * n'"
      THEN EXISTS_TAC "s2':state"
      THEN MULT_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;



let ES_NULL_THM =
 prove_thm
  (`ES_NULL_THM`,
   "!q. 
            
            %---------------------------------------------------%
                          ES_SPEC(q, [], q)",

   REWRITE_TAC[ES_SPEC;ES_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ ASM_REWRITE_TAC[];

      EXISTS_TAC "[]:num list"
      THEN EXISTS_TAC "s1:state"
      THEN REWRITE_TAC[]
    ]
  );;

%<
g "!q. 
            
            %---------------------------------------------------%
                          ES_SPEC(q, [], q)";;

e(REWRITE_TAC[ES_SPEC;ES_DEF]);;
e(REPEAT STRIP_TAC);;

 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "[]:num list");;
 e(EXISTS_TAC "s1:state");;
 e(REWRITE_TAC[]);;
>%



let ES_CONS_THM =
 prove_thm
  (`ES_CONS_THM`,
   "!p q r e es. 
                    ESPEC(p, e, r) /\ ES_SPEC(r, es, q)
            %---------------------------------------------------% ==>
                        ES_SPEC(p, CONS e es, q)",

   REWRITE_TAC[ESPEC;ES_SPEC;ES_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "CONS (n:num) ns"
      THEN EXISTS_TAC "s2':state"
      THEN REWRITE_TAC[NULL]
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[HD;TL]
    ]
  );;

%<
g "!p q r e es. 
                    ESPEC(p, e, r) /\ ES_SPEC(r, es, q)
            %---------------------------------------------------% ==>
                        ES_SPEC(p, CONS e es, q)";;

e(REWRITE_TAC[ESPEC;ES_SPEC;ES_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "CONS (n:num) ns");;
 e(EXISTS_TAC "s2':state");;
 e(REWRITE_TAC[NULL]);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[HD;TL]);;
>%



let EQ_THM =
 prove_thm
  (`EQ_THM`,
   "!p q r e1 e2. 
                   ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                          BSPEC(p, EQ e1 e2, q)",

   REWRITE_TAC[ESPEC;BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "(n:num) = n'"
      THEN EXISTS_TAC "s2':state"
      THEN BEQ_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r e1 e2. 
                   ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                          BSPEC(p, EQ e1 e2, q)";;

e(REWRITE_TAC[ESPEC;BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "(n:num) = n'");;
 e(EXISTS_TAC "s2':state");;
 e(BEQ_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let BLESS_THM =
 prove_thm
  (`BLESS_THM`,
   "!p q r e1 e2. 
                   ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                          BSPEC(p, LESS e1 e2, q)",

   REWRITE_TAC[ESPEC;BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "n < n'"
      THEN EXISTS_TAC "s2':state"
      THEN LESS_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r e1 e2. 
                   ESPEC(p, e1, r) /\ ESPEC(r, e2, q)
            %---------------------------------------------------% ==>
                          BSPEC(p, LESS e1 e2, q)";;

e(REWRITE_TAC[ESPEC;BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e1:exp) = e1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e2:exp) = e2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "n < n'");;
 e(EXISTS_TAC "s2':state");;
 e(LESS_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let LLESS_THM =
 prove_thm
  (`LLESS_THM`,
   "!p q r es1 es2. 
                   ES_SPEC(p, es1, r) /\ ES_SPEC(r, es2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, LLESS es1 es2, q)",

   REWRITE_TAC[ES_SPEC;BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es1:exp list) = es1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es2:exp list) = es2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "ns << ns'"
      THEN EXISTS_TAC "s2':state"
      THEN LLESS_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r es1 es2. 
                   ES_SPEC(p, es1, r) /\ ES_SPEC(r, es2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, LLESS es1 es2, q)";;

e(REWRITE_TAC[ES_SPEC;BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es1:exp list) = es1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(es2:exp list) = es2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "ns << ns'");;
 e(EXISTS_TAC "s2':state");;
 e(LLESS_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let AND_THM =
 prove_thm
  (`AND_THM`,
   "!p q r b1 b2. 
                  BSPEC(p, b1, r) /\ BSPEC(r, b2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, AND b1 b2, q)",

   REWRITE_TAC[BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b1:bexp) = b1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b2:bexp) = b2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "t /\ t'"
      THEN EXISTS_TAC "s2':state"
      THEN AND_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r b1 b2. 
                  BSPEC(p, b1, r) /\ BSPEC(r, b2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, AND b1 b2, q)";;

e(REWRITE_TAC[BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b1:bexp) = b1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b2:bexp) = b2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "t /\ t'");;
 e(EXISTS_TAC "s2':state");;
 e(AND_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let OR_THM =
 prove_thm
  (`OR_THM`,
   "!p q r b1 b2. 
                  BSPEC(p, b1, r) /\ BSPEC(r, b2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, OR b1 b2, q)",

   REWRITE_TAC[BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b1:bexp) = b1'"
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b2:bexp) = b2'"
      THEN RES_TAC
      THEN RES_TAC;

      RES_TAC
      THEN RES_TAC
      THEN EXISTS_TAC "t \/ t'"
      THEN EXISTS_TAC "s2':state"
      THEN OR_TAC
      THEN EXISTS_TAC "s2:state"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!p q r b1 b2. 
                  BSPEC(p, b1, r) /\ BSPEC(r, b2, q)
            %---------------------------------------------------% ==>
                         BSPEC(p, OR b1 b2, q)";;

e(REWRITE_TAC[BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b1:bexp) = b1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b2:bexp) = b2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "t \/ t'");;
 e(EXISTS_TAC "s2':state");;
 e(OR_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let NOT_THM =
 prove_thm
  (`NOT_THM`,
   "!p q b. 
                              BSPEC(p, b, q)
            %---------------------------------------------------% ==>
                            BSPEC(p, NOT b, q)",

   REWRITE_TAC[BSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite)
      THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
      THEN RES_TAC;

      RES_TAC
      THEN EXISTS_TAC "~t"
      THEN EXISTS_TAC "s2:state"
      THEN NOT_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;


%<
g "!p q b. 
                              BSPEC(p, b, q)
            %---------------------------------------------------% ==>
                            BSPEC(p, NOT b, q)";;

e(REWRITE_TAC[BSPEC]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o B_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
 e(RES_TAC);;

 e(RES_TAC);;
 e(EXISTS_TAC "~t");;
 e(EXISTS_TAC "s2:state");;
 e(NOT_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%



let SKIPp_THM =
 prove_thm
  (`SKIPp_THM`,
   "!q r.
                                                                
            %---------------------------------------------------%
                            PCSPEC(q,SKIP,q,r)",

   REWRITE_TAC[PCSPEC]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
   THEN ASM_REWRITE_TAC[]
  );;


let SKIPt_THM =
 prove_thm
  (`SKIPt_THM`,
   "!q r.
                                                                
            %---------------------------------------------------%
                            TCSPEC(q,SKIP,r)",

   REWRITE_TAC[TCSPEC]
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC "s1:state"
   THEN SKIP_TAC
  );;


let SKIP_THM =
 prove_thm
  (`SKIP_THM`,
   "!q r.
                                                                
            %---------------------------------------------------%
                            CSPEC(q,SKIP,q,r)",

   REWRITE_TAC[CSPEC;SKIPp_THM;SKIPt_THM]
  );;

%
for SKIP:
g "!q r. CSPEC(q,SKIP,q,r)";;
e(REWRITE_TAC[CSPEC_THM]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "s1:state");;
 e(SKIP_TAC);;
%



let ABORTp_THM =
 prove_thm
  (`ABORTp_THM`,
   "!a q r.
                                                                
            %---------------------------------------------------%
                            PCSPEC(a,ABORT,q,r)",

   REWRITE_TAC[PCSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
  );;


let ABORTt_THM =
 prove_thm
  (`ABORTt_THM`,
   "!r.
                                                                
            %---------------------------------------------------%
                            TCSPEC(AFALSE,ABORT,r)",

   REWRITE_TAC[TCSPEC;A_DEF]
  );;


let ABORT_THM =
 prove_thm
  (`ABORT_THM`,
   "!q r.
                                                                
            %---------------------------------------------------%
                            CSPEC(AFALSE,ABORT,q,r)",

   REWRITE_TAC[CSPEC_THM;A_DEF]
  );;


%
for ABORT:
g "!q r. CSPEC(AFALSE,ABORT,q,r)";;
e(REWRITE_TAC[CSPEC_THM;A_DEF]);;
%


let ASSIGNp_THM =
 prove_thm
  (`ASSIGNp_THM`,
   "!q x e r.
                                                                
            %---------------------------------------------------%
                 PCSPEC(q <|a (x := e), ASSIGN x e, q, r)",

   REWRITE_TAC[PCSPEC;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(x:var) = x'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e:exp) = e'"
   THEN IMP_RES_THEN REWRITE_ALL_THM assign_THM
   THEN ASM_REWRITE_TAC[]
  );;


let ASSIGNt_THM =
 prove_thm
  (`ASSIGNt_THM`,
   "!a x e r.
                                                                
            %---------------------------------------------------%
                          TCSPEC(a, ASSIGN x e, r)",

   REWRITE_TAC[TCSPEC;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN EXISTS_TAC "s1 <|s (x := e)"
   THEN REWRITE_TAC[assign;s_subst_BND]
   THEN ASSIGN_TAC
   THEN REWRITE_TAC[VE_THM]
  );;


let ASSIGN_THM =
 prove_thm
  (`ASSIGN_THM`,
   "!q x e r.
                                                                
            %---------------------------------------------------%
                 CSPEC(q <|a (x := e), ASSIGN x e, q, r)",

   REWRITE_TAC[CSPEC;ASSIGNp_THM;ASSIGNt_THM]
  );;


%<
g "!q x e r.
                                                                
            %---------------------------------------------------%
                 CSPEC(q <|a (x := e), ASSIGN x e, q, r)";;

e(REWRITE_TAC[CSPEC_THM]);;
e(REWRITE_TAC[a_subst_THM]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(x:var) = x'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(e:exp) = e'");;
 e(IMP_RES_THEN REWRITE_ALL_THM assign_THM);;
 e(ASM_REWRITE_TAC[]);;

 e(EXISTS_TAC "s1 <|s (x := e)");;
 e(REWRITE_TAC[assign]);;
 e(REWRITE_TAC[s_subst_BND]);;
 e(ASSIGN_TAC);;
 e(REWRITE_TAC[VE_THM]);;

E e s1 n s2 ==> (s1 <|s (x := e) = BND n x s2)
>%



let SEQp_THM =
 prove_thm
  (`SEQp_THM`,
   "!p q s c1 c2 r.
                PCSPEC(p, c1, s, r)   /\   PCSPEC(s, c2, q, r)
            %---------------------------------------------------% ==>
                       PCSPEC(p, SEQ c1 c2, q, r)",

   REWRITE_TAC[PCSPEC]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'"
   THEN RES_TAC
   THEN RES_TAC
  );;


let SEQt_THM =
 prove_thm
  (`SEQt_THM`,
   "!p s c1 c2 r.
                        TCSPEC(p, c1, r)      /\
                        PCSPEC(p, c1, s, r)   /\
                        TCSPEC(s, c2, r)
            %---------------------------------------------------% ==>
                        TCSPEC(p, SEQ c1 c2, r)",

   REWRITE_TAC[TCSPEC;PCSPEC]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
   THEN RES_TAC
   THEN POP_TAC
   THEN RES_TAC
   THEN EXISTS_TAC "s2':state"
   THEN SEQ_TAC
   THEN EXISTS_TAC "s2:state"
   THEN ASM_REWRITE_TAC[]
  );;


let SEQ_THM =
 prove_thm
  (`SEQ_THM`,
   "!p q s c1 c2 r.
                 CSPEC(p, c1, s, r)   /\   CSPEC(s, c2, q, r)
            %---------------------------------------------------% ==>
                        CSPEC(p, SEQ c1 c2, q, r)",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC SEQp_THM;

      IMP_RES_TAC SEQt_THM
    ]
  );;

%<
g "!p q s c1 c2 r.
                 CSPEC(p, c1, s, r)   /\   CSPEC(s, c2, q, r)
            %---------------------------------------------------% ==>
                        CSPEC(p, SEQ c1 c2, q, r)";;

e(REWRITE_TAC[CSPEC_THM]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
 e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
 e(RES_TAC);;
 e(RES_TAC);;

 e(RES_TAC);;
 e(RES_TAC);;
 e(EXISTS_TAC "s2':state");;
 e(SEQ_TAC);;
 e(EXISTS_TAC "s2:state");;
 e(ASM_REWRITE_TAC[]);;
>%



let IFp_THM =
 prove_thm
  (`IFp_THM`,
   "!p q r1 r2 b c1 c2 r.
                   BSPEC(AAND p (AB b), b, r1) /\
                   PCSPEC(r1, c1, q, r)      /\
                   BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                   PCSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
                   PCSPEC(p, IF b c1 c2, q, r)",

   REWRITE_TAC[PCSPEC;BSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'"
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM])
   THEN RES_TAC
   THEN RES_TAC
  );;


let IFt_THM =
 prove_thm
  (`IFt_THM`,
   "!p r1 r2 b c1 c2 r.
                   BSPEC(AAND p (AB b), b, r1) /\
                   TCSPEC(r1, c1, r)      /\
                   BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                   TCSPEC(r2, c2, r)
            %---------------------------------------------------% ==>
                   TCSPEC(p, IF b c1 c2, r)",

   REWRITE_TAC[TCSPEC;BSPEC;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "A(AB b)s1"
   THEN RES_TAC
   THEN RES_TAC
   THEN EXISTS_TAC "s2'':state"
   THENL [ IF_T_TAC; IF_F_TAC ]
   THEN EXISTS_TAC "s2:state"
   THEN REWRITE_ALL_TAC[AB_THM]
   THEN ASM_REWRITE_TAC[]
  );;


let IF_THM =
 prove_thm
  (`IF_THM`,
   "!p q r1 r2 b c1 c2 r.
                   BSPEC(AAND p (AB b), b, r1) /\
                   CSPEC(r1, c1, q, r)      /\
                   BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                   CSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
                   CSPEC(p, IF b c1 c2, q, r)",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC (AND2IMP IFp_THM);

      IMP_RES_TAC (AND2IMP IFt_THM)
    ]
  );;


%<
g "!p q r1 r2 b c1 c2 r.
                BSPEC(AAND p (AB b), b, r1) /\
                CSPEC(r1, c1, q, r)      /\
                BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                CSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
                CSPEC(p, IF b c1 c2, q, r)";;

e(REWRITE_TAC[CSPEC_THM;BSPEC]);;
e(REWRITE_TAC[A_DEF]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(CONJ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(RES_TAC);;
  e(RES_TAC);;

  e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'");;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c1:cmd) = c1'");;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c2:cmd) = c2'");;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(RES_TAC);;
  e(RES_TAC);;

 e(REPEAT STRIP_TAC);;
 e(ASM_CASES_TAC "A(AB b)s1");;

  e(RES_TAC);;
  e(RES_TAC);;
  e(EXISTS_TAC "s2'':state");;
  e(IF_T_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(REWRITE_ALL_TAC[AB_THM]);;
  e(ASM_REWRITE_TAC[]);;

  e(RES_TAC);;
  e(RES_TAC);;
  e(EXISTS_TAC "s2'':state");;
  e(IF_F_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(REWRITE_ALL_TAC[AB_THM]);;
  e(ASM_REWRITE_TAC[]);;
>%


let IFp_THM1 =
 prove_thm
  (`IFp_THM1`,
   "!q r1 r2 b c1 c2 r.
               PCSPEC(r1, c1, q, r)      /\
               PCSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
               PCSPEC(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                      IF b c1 c2, q, r)",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL IFp_THM)
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
   THENL [ EXISTS_TAC "ab_pre b r1";
           EXISTS_TAC "ab_pre b r2"
         ]
   THEN REWRITE_TAC[ab_pre_THM]
   THEN REWRITE_TAC[CLOSE;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM REWRITE_ALL_THM
   THEN ASM_REWRITE_TAC[]
  );;


let IFt_THM1 =
 prove_thm
  (`IFt_THM1`,
   "!r1 r2 b c1 c2 r.
               TCSPEC(r1, c1, r)      /\
               TCSPEC(r2, c2, r)
            %---------------------------------------------------% ==>
               TCSPEC(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                      IF b c1 c2, r)",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL IFt_THM)
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
   THENL [ EXISTS_TAC "ab_pre b r1";
           EXISTS_TAC "ab_pre b r2"
         ]
   THEN REWRITE_TAC[ab_pre_THM]
   THEN REWRITE_TAC[CLOSE;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM REWRITE_ALL_THM
   THEN ASM_REWRITE_TAC[]
  );;


let IF_THM1 =
 prove_thm
  (`IF_THM1`,
   "!q r1 r2 b c1 c2 r.
                CSPEC(r1, c1, q, r)      /\
                CSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
                CSPEC(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                      IF b c1 c2, q, r)",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC IFp_THM1;

      IMP_RES_TAC IFt_THM1
    ]
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!q r1 r2 b c1 c2 r.
                CSPEC(r1, c1, q, r)      /\
                CSPEC(r2, c2, q, r)
            %---------------------------------------------------% ==>
                CSPEC(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                      IF b c1 c2, q, r)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC IF_THM);;
e(EXISTS_TAC "r1:aexp");;
e(EXISTS_TAC "r2:aexp");;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b r1");;
 e(REWRITE_TAC[ab_pre_THM]);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(ASM_REWRITE_TAC[]);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b r2");;
 e(REWRITE_TAC[ab_pre_THM]);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(ASM_REWRITE_TAC[]);;
>%



let  CMD_STR_INDUCT_TAC = RULE_INDUCT_THEN C_str_induct
                                           STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;



let NUM_WELL_FOUNDED =
 prove_thm
  (`NUM_WELL_FOUNDED`,
   "!R. (!n. (!m. m < n ==> R m) ==> R n) ==> (!n. R n)",

   GEN_TAC
   THEN CONV_TAC CONTRAPOS_CONV
   THEN CONV_TAC (DEPTH_CONV NOT_FORALL_CONV)
   THEN REWRITE_TAC[NOT_IMP]
   THEN DISCH_THEN (STRIP_ASSUME_TAC o
              MATCH_MP (REWRITE_RULE[o_THM] (SPEC "$~ o (R:num->bool)" WOP)))
   THEN EXISTS_TAC "n:num"
   THEN ASM_REWRITE_TAC[]
  );;

% NUM_WELL_FOUNDED
g "!R. (!n. (!m. m < n ==> R m) ==> R n) ==> (!n. R n)";;

e(GEN_TAC);;
e(CONV_TAC CONTRAPOS_CONV);;
e(CONV_TAC (DEPTH_CONV NOT_FORALL_CONV));;
e(REWRITE_TAC[NOT_IMP]);;
e(DISCH_THEN (STRIP_ASSUME_TAC o
              MATCH_MP (REWRITE_RULE[o_THM] (SPEC "$~ o (R:num->bool)" WOP))));;
e(EXISTS_TAC "n:num");;
e(ASM_REWRITE_TAC[]);;
%





let BND_SAME_COMMUTE =
 TAC_PROOF
  (([],
    "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)"),
   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:**) = x"
   THENL
    [ ASM_REWRITE_TAC[];

      IMP_RES_TAC BND_COMMUTE
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% BND_SAME_COMMUTE
g "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)";;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:**) = x");;

 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC BND_COMMUTE);;
 e(ASM_REWRITE_TAC[]);;
%




let BSPEC_CHANGE_VAR_LEMMA =
 prove_thm
  (`BSPEC_CHANGE_VAR_LEMMA`,
   "!a v x b c g r y p q.
     WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
     (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, p) =
        BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR y)), b, p <|a (AVAR y / x)) )",

   REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (\th.
             (STRIP_ASSUME_TAC o 
              REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
              SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                   UNION FVa p UNION FVa q"])
             variant_not_in_set)
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[BSPEC;A_DEF;V_DEF]
   THEN MATCH_MP_TAC CONJ_11
   THEN CONJ_TAC
   THENL
    [ REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]
      THEN EQ_TAC THEN REPEAT STRIP_TAC
      THENL
       [ FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "BND ((s2:state) y) x (s1:state)"
         THEN EXISTS_TAC "t:bool"
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN REWRITE_TAC[BND_MATCH]
         THEN IMP_RES_TAC B_BND_THM
         THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
         THEN ASM_REWRITE_TAC[];

         IMP_RES_THEN (SUBST1_TAC o SYM o SPECL["(s2:state) x";"s2:state"])
                   A_BND_THM
         THEN FIRST_ASSUM (MP_TAC o
                   SPECL["BND ((s2:state) x) y (s1:state)";"t:bool";
                         "BND (s2 x) y (s2:state)"])
         THEN REWRITE_TAC[BND_MATCH]
         THEN ONCE_REWRITE_TAC[BND_SAME_COMMUTE]
         THEN REWRITE_TAC[IDENT_BND]
         THEN DISCH_THEN MATCH_MP_TAC
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN IMP_RES_TAC B_BND_THM
         THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
         THEN ASM_REWRITE_TAC[]
       ];

      EQ_TAC THEN REPEAT STRIP_TAC
      THENL
       [ FIRST_ASSUM (MP_TAC o SPEC "BND ((s1:state) y) x (s1:state)")
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN STRIP_TAC
         THEN IMP_RES_TAC B_BND_THM
         THEN POP_TAC
         THEN POP_ASSUM (MP_TAC o SPEC "(s1:state)x")
         THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND]
         THEN STRIP_TAC
         THEN EXISTS_TAC "t:bool"
         THEN EXISTS_TAC "BND ((s1:state) x) x (s2:state)"
         THEN ASM_REWRITE_TAC[];

         FIRST_ASSUM (MP_TAC o SPEC "BND ((s1:state) x) y (s1:state)")
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN STRIP_TAC
         THEN IMP_RES_TAC B_BND_THM
         THEN POP_ASSUM (MP_TAC o SPEC "(s1:state)y")
         THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND]
         THEN STRIP_TAC
         THEN EXISTS_TAC "t:bool"
         THEN EXISTS_TAC "BND ((s1:state) y) y (s2:state)"
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< BSPEC_CHANGE_VAR_LEMMA
g "!a v x b c g r y p q.
     WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
     (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR x)), b, p) =
        BSPEC(AAND (AAND a (AB b)) (AEQ v (AVAR y)), b, p <|a (AVAR y / x)) )";;
e(REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa p UNION FVa q"])
  variant_not_in_set));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(REWRITE_TAC[BSPEC;A_DEF;V_DEF]);;
e(MATCH_MP_TAC CONJ_11);;
e(CONJ_TAC);;

 e(REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]);;
 e(EQ_TAC THEN REPEAT STRIP_TAC);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BND ((s2:state) y) x (s1:state)");;
  e(EXISTS_TAC "t:bool");;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(REWRITE_TAC[BND_MATCH]);;
  e(IMP_RES_TAC B_BND_THM);;
  e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(IMP_RES_THEN (SUBST1_TAC o SYM o SPECL["(s2:state) x";"s2:state"])
         A_BND_THM);;
  e(FIRST_ASSUM (MP_TAC o
      SPECL["BND ((s2:state) x) y (s1:state)";"t:bool";
            "BND (s2 x) y (s2:state)"]));;
  e(REWRITE_TAC[BND_MATCH]);;
  e(ONCE_REWRITE_TAC[BND_SAME_COMMUTE]);;
  e(REWRITE_TAC[IDENT_BND]);;
  e(DISCH_THEN MATCH_MP_TAC);;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(IMP_RES_TAC B_BND_THM);;
  e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
  e(ASM_REWRITE_TAC[]);;

 e(EQ_TAC THEN REPEAT STRIP_TAC);;

let st1 = get_state();;
set_state(st1);;
  e(FIRST_ASSUM (MP_TAC o SPEC "BND ((s1:state) y) x (s1:state)"));;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC B_BND_THM);;
  e(POP_TAC);;
  e(POP_ASSUM (MP_TAC o SPEC "(s1:state)x"));;
  e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "t:bool");;
  e(EXISTS_TAC "BND ((s1:state) x) x (s2:state)");;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM (MP_TAC o SPEC "BND ((s1:state) x) y (s1:state)"));;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC B_BND_THM);;
  e(POP_ASSUM (MP_TAC o SPEC "(s1:state)y"));;
  e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "t:bool");;
  e(EXISTS_TAC "BND ((s1:state) y) y (s2:state)");;
  e(ASM_REWRITE_TAC[]);;
>%




let PCSPEC_CHANGE_VAR_LEMMA =
 prove_thm
  (`PCSPEC_CHANGE_VAR_LEMMA`,
   "!a v x b c g r y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( PCSPEC(p,                  c, AAND a (ALESS v (AVAR x)), r) =
        PCSPEC(p <|a (AVAR y / x), c, AAND a (ALESS v (AVAR y)), r) )",

   REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (\th.
             (STRIP_ASSUME_TAC o 
              REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
              SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                   UNION FVa p UNION FVa q"])
             variant_not_in_set)
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN REWRITE_TAC[PCSPEC;A_DEF;V_DEF]
   THEN REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]
   THEN EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
       [ FIRST_ASSUM (MP_TAC o
                      SPECL["BND ((s2:state) y) x (s1:state)";
                            "BND (s2 y) x (s2:state)"])
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN REWRITE_TAC[BND_MATCH]
         THEN DISCH_THEN MATCH_MP_TAC
         THEN IMP_RES_TAC C_BND_THM
         THEN IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  C_FREE_ONLY_IMPACT)
         THEN ASM_REWRITE_TAC[];

         FIRST_ASSUM (MP_TAC o
                   SPECL["BND (s1 x) y (s1:state)";
                         "BND (s2 x) y (s2:state)"])
         THEN REWRITE_TAC[BND_MATCH]
         THEN ONCE_REWRITE_TAC[BND_SAME_COMMUTE]
         THEN REWRITE_TAC[IDENT_BND]
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN DISCH_THEN MATCH_MP_TAC
         THEN IMP_RES_TAC C_BND_THM
         THEN IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  C_FREE_ONLY_IMPACT)
         THEN ASM_REWRITE_TAC[]
       ]
  );;


let TCSPEC_CHANGE_VAR_LEMMA =
 prove_thm
  (`TCSPEC_CHANGE_VAR_LEMMA`,
   "!a v x b c g r y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( TCSPEC(p,                  c, r) =
        TCSPEC(p <|a (AVAR y / x), c, r) )",

   REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (\th.
             (STRIP_ASSUME_TAC o 
              REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
              SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                   UNION FVa p UNION FVa q"])
             variant_not_in_set)
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN REWRITE_TAC[TCSPEC;A_DEF;V_DEF]
   THEN REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]
   THEN EQ_TAC THEN REPEAT STRIP_TAC
   THENL
       [ RES_TAC
         THEN IMP_RES_TAC C_BND_THM
         THEN POP_TAC
         THEN POP_ASSUM (MP_TAC o SPEC "(s1:state)x")
         THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND]
         THEN STRIP_TAC
         THEN EXISTS_TAC "BND ((s1:state) x) x (s2:state)"
         THEN ASM_REWRITE_TAC[];

         FIRST_ASSUM (MP_TAC o SPEC "BND (s1 x) y (s1:state)")
         THEN REWRITE_TAC[BND_MATCH]
         THEN ONCE_REWRITE_TAC[BND_SAME_COMMUTE]
         THEN REWRITE_TAC[IDENT_BND]
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN ASM_REWRITE_TAC[]
         THEN STRIP_TAC
         THEN IMP_RES_TAC C_BND_THM
         THEN POP_ASSUM (MP_TAC o SPEC "(s1:state)y")
         THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND]
         THEN STRIP_TAC
         THEN EXISTS_TAC "BND ((s1:state) y) y (s2:state)"
         THEN ASM_REWRITE_TAC[]
       ]
  );;


let CSPEC_CHANGE_VAR_LEMMA =
 prove_thm
  (`CSPEC_CHANGE_VAR_LEMMA`,
   "!a v x b c g r y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( CSPEC(p,                  c, AAND a (ALESS v (AVAR x)), r) =
        CSPEC(p <|a (AVAR y / x), c, AAND a (ALESS v (AVAR y)), r) )",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP PCSPEC_CHANGE_VAR_LEMMA)
   THEN IMP_RES_TAC (AND2IMP TCSPEC_CHANGE_VAR_LEMMA)
   THEN ASM_REWRITE_TAC[]
  );;

%< CSPEC_CHANGE_VAR_LEMMA
g "!a v x b c g r y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( CSPEC(p,                  c, AAND a (ALESS v (AVAR x)), r) =
        CSPEC(p <|a (AVAR y / x), c, AAND a (ALESS v (AVAR y)), r) )";;
e(REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa p UNION FVa q"])
  variant_not_in_set));;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(REWRITE_TAC[CSPEC_THM;A_DEF;V_DEF]);;
e(REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]);;
e(MATCH_MP_TAC CONJ_11);;
e(CONJ_TAC);;

 e(EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC);;

  e(FIRST_ASSUM (MP_TAC o
                 SPECL["BND ((s2:state) y) x (s1:state)";
                       "BND (s2 y) x (s2:state)"]));;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(REWRITE_TAC[BND_MATCH]);;
  e(DISCH_THEN MATCH_MP_TAC);;
  e(IMP_RES_TAC C_BND_THM);;
  e(IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  C_FREE_ONLY_IMPACT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(FIRST_ASSUM (MP_TAC o
                 SPECL["BND (s1 x) y (s1:state)";
                       "BND (s2 x) y (s2:state)"]));;
  e(REWRITE_TAC[BND_MATCH]);;
  e(ONCE_REWRITE_TAC[BND_SAME_COMMUTE]);;
  e(REWRITE_TAC[IDENT_BND]);;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(DISCH_THEN MATCH_MP_TAC);;
  e(IMP_RES_TAC C_BND_THM);;
  e(IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  C_FREE_ONLY_IMPACT));;
  e(ASM_REWRITE_TAC[]);;

 e(EQ_TAC THEN REPEAT STRIP_TAC);;

let st1 = get_state();;
set_state(st1);;
  e(RES_TAC);;
  e(IMP_RES_TAC C_BND_THM);;
  e(POP_TAC);;
  e(POP_ASSUM (MP_TAC o SPEC "(s1:state)x"));;
  e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "BND ((s1:state) x) x (s2:state)");;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM (MP_TAC o SPEC "BND (s1 x) y (s1:state)"));;
  e(REWRITE_TAC[BND_MATCH]);;
  e(ONCE_REWRITE_TAC[BND_SAME_COMMUTE]);;
  e(REWRITE_TAC[IDENT_BND]);;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(ASM_REWRITE_TAC[]);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC C_BND_THM);;
  e(POP_ASSUM (MP_TAC o SPEC "(s1:state)y"));;
  e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "BND ((s1:state) y) y (s2:state)");;
  e(ASM_REWRITE_TAC[]);;
>%



let IMP_CONJ_EQ =
 TAC_PROOF
  (([],
    "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))"),
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% IMP_CONJ_EQ
g "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%



let WHILEp_LEMMA =
 prove_thm
  (`WHILEp_LEMMA`,
   "!a v x b c g r p q n.
      (\n.
        WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
        PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
        BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
        BSPEC(AAND a (ANOT (AB b)), b, q) ==>
        PCSPEC(AAND a (AEQ v (ANUM n)),
               WHILE a (ALESS v (AVAR x)) b c, q, r)
      ) n",

   REPEAT GEN_TAC
   THEN SPEC_TAC("n:num","n:num")
   THEN MATCH_MP_TAC NUM_WELL_FOUNDED
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th))
   THEN ASM_REWRITE_TAC[]
   THEN DISCH_THEN (ASSUME_TAC o REWRITE_RULE[PCSPEC;IMP_CONJ_EQ])
   THEN UNDISCH_ALL_TAC
   THEN CONV_TAC (DEPTH_CONV FORALL_AND_CONV)
   THEN CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN DEFINE_NEW_VAR
             "y = variant x (FVa a UNION FVv v UNION FVb b UNION
                             FVc c r UNION FVa p UNION FVa q)"
   THEN IMP_RES2_THEN REWRITE_THM BSPEC_CHANGE_VAR_LEMMA
   THEN IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM) PCSPEC_CHANGE_VAR_LEMMA
   THEN UNDISCH_ALL_TAC
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o
                     REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM])
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN DISCH_THEN (\th.
           (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
            SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                 UNION FVa p UNION FVa q"])
           variant_not_in_set)
   THEN SUBGOAL_THEN "~y IN FVc (WHILE a (ALESS v (AVAR x))b c) r" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVc_DEF;IN_UNION]; ALL_TAC ]
   THEN SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[AND_IMP_INTRO]
   THEN REWRITE_TAC[BSPEC;PCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite)
   THENL
       [ UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
         THEN IMP_RES_THEN (SUBST1_TAC o SYM o
                            SPECL["n:num";"s2:state"]) A_BND_THM
         THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM])
         THEN POP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "V v (BND n y s3)"
         THEN EXISTS_TAC "BND n y (s3:state)"
         THEN IMP_RES_TAC C_BND_THM
         THEN ASM_REWRITE_TAC[]
         THEN ONCE_REWRITE_TAC[CONJ_SYM]
         THEN SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n y (s3:state) y)"
                       REWRITE_THM
              THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "BND n y (s2':state)"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "BND n y (s1:state)"
         THEN EXISTS_TAC "T"
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN IMP_RES_TAC B_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH];

         UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
         THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM])
         THEN POP_TAC
         THEN MATCH_MP_TAC (ASSUME
               "!s1 t s2. (A a s1 /\ ~A (AB b) s1) /\ B b s1 t s2 ==> A q s2")
         THEN EXISTS_TAC "s1:state"
         THEN EXISTS_TAC "F"
         THEN ASM_REWRITE_TAC[]
       ]
  );;


let WHILEt_LEMMA =
 prove_thm
  (`WHILEt_LEMMA`,
   "!a v x b c g r p q n.
      (\n.
        WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
        PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
        TCSPEC(p, c, r) /\
        BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
        BSPEC(AAND a (ANOT (AB b)), b, q) ==>
        TCSPEC(AAND a (AEQ v (ANUM n)),
               WHILE a (ALESS v (AVAR x)) b c, r)
      ) n",

   REPEAT GEN_TAC
   THEN SPEC_TAC("n:num","n:num")
   THEN MATCH_MP_TAC NUM_WELL_FOUNDED
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th))
   THEN ASM_REWRITE_TAC[]
   THEN DISCH_THEN (ASSUME_TAC o REWRITE_RULE[PCSPEC;TCSPEC;IMP_CONJ_EQ])
   THEN UNDISCH_ALL_TAC
   THEN CONV_TAC (DEPTH_CONV FORALL_AND_CONV)
   THEN CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN DEFINE_NEW_VAR
             "y = variant x (FVa a UNION FVv v UNION FVb b UNION
                             FVc c r UNION FVa p UNION FVa q)"
   THEN IMP_RES2_THEN REWRITE_THM BSPEC_CHANGE_VAR_LEMMA
   THEN IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM) PCSPEC_CHANGE_VAR_LEMMA
   THEN IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM) TCSPEC_CHANGE_VAR_LEMMA
   THEN UNDISCH_ALL_TAC
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o
                     REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM])
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN DISCH_THEN (\th.
           (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
            SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                 UNION FVa p UNION FVa q"])
           variant_not_in_set)
   THEN SUBGOAL_THEN "~y IN FVc (WHILE a (ALESS v (AVAR x))b c) r" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVc_DEF;IN_UNION]; ALL_TAC ]
   THEN SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[AND_IMP_INTRO]
   THEN REWRITE_TAC[BSPEC;PCSPEC;TCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "A (AB b) s1"
   THENL
       [ UNASSUME_THEN (MP_TAC o SPEC "BND n y (s1:state)")
               "!s1. (A a s1 /\ A(AB b)s1) /\ (V v s1 = s1 y) ==>
                           (?t s2. B b s1 t s2)"
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN STRIP_TAC
         THEN IMP_RES_TAC B_FREE_ONLY_IMPACT
         THEN UNASSUME_THEN (MP_TAC o SPECL["BND n y (s1:state)";
                                            "t:bool";"s2:state"])
                      "!s1 t s2. ((A a s1 /\ A(AB b)s1) /\
                                  (V v s1 = s1 y)) /\ B b s1 t s2 ==>
                                  A (p <|a ((AVAR y) / x)) s2"
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN STRIP_TAC
         THEN IMP_RES_TAC (ASSUME "!s1. A(p <|a ((AVAR y) / x))s1 ==>
                                        (?s2. C c r s1 s2)")
         THEN IMP_RES_TAC
                  (ASSUME "!s1 s2.
                            A(p <|a ((AVAR y) / x))s1 /\ C c r s1 s2 ==>
                            A a s2 /\ (V v s2) < (s2 y)")
         THEN UNDISCH_LAST_TAC
         THEN IMP_RES_TAC C_FREE_ONLY_IMPACT
         THEN ASM_REWRITE_TAC[BND_MATCH]
         THEN STRIP_TAC
         THEN UNASSUME_THEN (MP_TAC o SPECL["V v s2'";"s2':state"])
               "!m s1. m < n /\ A a s1 /\ (V v s1 = m) ==>
                       (?s2. C(WHILE a(ALESS v(AVAR x))b c)r s1 s2)"
         THEN ASM_REWRITE_TAC[]
         THEN STRIP_TAC
         THEN EXISTS_TAC "BND ((s1:state) y) y (s2'':state)"
         THEN WHILE_T_TAC
         THEN EXISTS_TAC "BND ((s1:state) y) y (s2:state)"
         THEN EXISTS_TAC "BND ((s1:state) y) y (s2':state)"
         THEN IMP_RES_TAC C_BND_THM
         THEN POP_TAC THEN POP_ASSUM REWRITE_THM
         THEN POP_TAC THEN POP_ASSUM REWRITE_THM
         THEN IMP_RES_TAC B_BND_THM
         THEN POP_TAC THEN POP_ASSUM (MP_TAC o SPEC "(s1:state) y")
         THEN REWRITE_TAC[BND_OVERRIDE;IDENT_BND]
         THEN REWRITE_TAC[AB_THM]
         THEN STRIP_TAC
         THEN ASM_REWRITE_TAC[];

         UNASSUME_THEN (MP_TAC o SPEC_ALL)
                 "!s1. A a s1 /\ ~A(AB b)s1 ==> (?t s2. B b s1 t s2)"
         THEN ASM_REWRITE_TAC[]
         THEN STRIP_TAC
         THEN EXISTS_TAC "s2:state"
         THEN WHILE_F_TAC
         THEN REWRITE_ALL_TAC[AB_THM]
         THEN ASM_REWRITE_TAC[]
       ]
  );;

%<
g "!a v x b c g r p q n.
      (\n.
        WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
        CSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
        BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
        BSPEC(AAND a (ANOT (AB b)), b, q) ==>
        CSPEC(AAND a (AEQ v (ANUM n)),
              WHILE a (ALESS v (AVAR x)) b c, q, r)
      ) n";;

e(REPEAT GEN_TAC);;
e(SPEC_TAC("n:num","n:num"));;
e(MATCH_MP_TAC NUM_WELL_FOUNDED);;
e(BETA_TAC);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th)));;
e(ASM_REWRITE_TAC[]);;
e(DISCH_THEN (ASSUME_TAC o REWRITE_RULE[CSPEC_THM;IMP_CONJ_EQ]));;
e(UNDISCH_ALL_TAC);;
e(CONV_TAC (DEPTH_CONV FORALL_AND_CONV));;
e(CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(DEFINE_NEW_VAR "y = variant x (FVa a UNION FVv v UNION FVb b UNION
                                 FVc c r UNION FVa p UNION FVa q)");;
e(IMP_RES2_THEN REWRITE_THM BSPEC_CHANGE_VAR_LEMMA);;
e(IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM) CSPEC_CHANGE_VAR_LEMMA);;
e(UNDISCH_ALL_TAC);;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM]));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(DISCH_THEN (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa p UNION FVa q"])
  variant_not_in_set));;
e(SUBGOAL_THEN "~y IN FVc (WHILE a (ALESS v (AVAR x))b c) r" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVc_DEF;IN_UNION]; ALL_TAC ]);;
e(SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(REWRITE_TAC[AND_IMP_INTRO]);;
e(REWRITE_TAC[BSPEC;CSPEC_THM;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o C_rewrite));;

  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
  e(IMP_RES_THEN (SUBST1_TAC o pthm o SYM o SPECL["n:num";"s2:state"]) A_BND_THM);;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(POP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "V v (BND n y s3)");;
  e(EXISTS_TAC "BND n y (s3:state)");;
  e(IMP_RES_TAC C_BND_THM);;
  e(ASM_REWRITE_TAC[]);;
  e(ONCE_REWRITE_TAC[CONJ_SYM]);;
  e(SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n y (s3:state) y)" REWRITE_THM
     THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BND n y (s2':state)");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BND n y (s1:state)");;
  e(EXISTS_TAC "T");;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(IMP_RES_TAC B_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(POP_TAC);;
  e(MATCH_MP_TAC (ASSUME
     "!s1 t s2. (A a s1 /\ ~A (AB b) s1) /\ B b s1 t s2 ==> A q s2"));;
  e(EXISTS_TAC "s1:state");;
  e(EXISTS_TAC "F");;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
 e(ASM_CASES_TAC "A (AB b) s1");;

  e(UNASSUME_THEN (MP_TAC o SPEC "BND n y (s1:state)")
    "!s1. (A a s1 /\ A(AB b)s1) /\ (V v s1 = s1 y) ==> (?t s2. B b s1 t s2)");;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC B_FREE_ONLY_IMPACT);;
  e(UNASSUME_THEN (MP_TAC o SPECL["BND n y (s1:state)";"t:bool";"s2:state"])
    "!s1 t s2. ((A a s1 /\ A(AB b)s1) /\ (V v s1 = s1 y)) /\ B b s1 t s2 ==>
               A (p <|a ((AVAR y) / x)) s2");;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC
       (ASSUME "!s1. A(p <|a ((AVAR y) / x))s1 ==> (?s2. C c r s1 s2)"));;
  e(IMP_RES_TAC (ASSUME "!s1 s2.
                            A(p <|a ((AVAR y) / x))s1 /\ C c r s1 s2 ==>
                            A a s2 /\ (V v s2) < (s2 y)"));;
  e(UNDISCH_LAST_TAC);;
  e(IMP_RES_TAC C_FREE_ONLY_IMPACT);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(STRIP_TAC);;
  e(UNASSUME_THEN (MP_TAC o SPECL["V v s2'";"s2':state"])
    "!m s1. m < n /\ A a s1 /\ (V v s1 = m) ==>
            (?s2. C(WHILE a(ALESS v(AVAR x))b c)r s1 s2)");;
  e(ASM_REWRITE_TAC[]);;
  e(STRIP_TAC);;
let st2 = get_state();;
set_state(st2);;
  e(EXISTS_TAC "BND ((s1:state) y) y (s2'':state)");;
  e(WHILE_T_TAC);;
  e(EXISTS_TAC "BND ((s1:state) y) y (s2:state)");;
  e(EXISTS_TAC "BND ((s1:state) y) y (s2':state)");;
  e(IMP_RES_TAC C_BND_THM);;
  e(POP_TAC THEN POP_ASSUM REWRITE_THM);;
  e(POP_TAC THEN POP_ASSUM REWRITE_THM);;
  e(IMP_RES_TAC B_BND_THM);;
  e(POP_TAC THEN POP_ASSUM (MP_TAC o SPEC "(s1:state) y"));;
  e(REWRITE_TAC[BND_OVERRIDE;IDENT_BND]);;
  e(REWRITE_TAC[AB_THM]);;
  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(UNASSUME_THEN (MP_TAC o SPEC_ALL)
    "!s1. A a s1 /\ ~A(AB b)s1 ==> (?t s2. B b s1 t s2)");;
  e(ASM_REWRITE_TAC[]);;
  e(STRIP_TAC);;
  e(EXISTS_TAC "s2:state");;
  e(WHILE_F_TAC);;
  e(REWRITE_ALL_TAC[AB_THM]);;
  e(ASM_REWRITE_TAC[]);;
>%



let WHILEp_THM =
 prove_thm
  (`WHILEp_THM`,
   "!a v x b c g r p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      PCSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC ((AND2IMP o
                      CONV_RULE (ONCE_DEPTH_CONV BETA_CONV))  WHILEp_LEMMA)
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[PCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC "V v s1")
   THEN EXISTS_TAC "s1:state"
   THEN ASM_REWRITE_TAC[]
  );;


let WHILEt_THM =
 prove_thm
  (`WHILEt_THM`,
   "!a v x b c g r p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
      TCSPEC(p, c, r) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      TCSPEC(a, WHILE a (ALESS v (AVAR x)) b c, r)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC ((AND2IMP o
                      CONV_RULE (ONCE_DEPTH_CONV BETA_CONV))  WHILEt_LEMMA)
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[TCSPEC;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC "V v s1")
   THEN ASM_REWRITE_TAC[]
  );;


let WHILE_THM =
 prove_thm
  (`WHILE_THM`,
   "!a v x b c g r p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      CSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      CSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)",
   REWRITE_TAC[CSPEC]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP WHILEp_THM)
   THEN IMP_RES_TAC (AND2IMP WHILEt_THM)
   THEN ASM_REWRITE_TAC[]
  );;


%< WHILE_THM
g "!a v x b c g r p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      CSPEC(p, c, AAND a (ALESS v (AVAR x)), r) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      CSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC ((REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO] o
                      CONV_RULE (ONCE_DEPTH_CONV BETA_CONV))  WHILE_LEMMA));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[CSPEC_THM;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;

 e(FIRST_ASSUM (MATCH_MP_TAC o CONJUNCT1 o SPEC "V v s1"));;
 e(EXISTS_TAC "s1:state");;
 e(ASM_REWRITE_TAC[]);;

 e(FIRST_ASSUM (MATCH_MP_TAC o CONJUNCT2 o SPEC "V v s1"));;
 e(ASM_REWRITE_TAC[]);;
>%



let WHILEp_THM1 =
 prove_thm
  (`WHILEp_THM1`,
   "!p q a b c g r.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             PCSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)",

    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (SPEC_ALL WHILEp_THM)
    THEN ASM_REWRITE_TAC[]
    THEN CONJ_TAC
    THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
    THENL
      [  EXISTS_TAC "ab_pre b p";
         EXISTS_TAC "ab_pre b q"
      ]
    THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;


let WHILEt_THM1 =
 prove_thm
  (`WHILEt_THM1`,
   "!p q a b c g r.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             PCSPEC(p, c, AAND a (ALESS v (AVAR x)), r)  /\
             TCSPEC(p, c, r)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             TCSPEC(a, WHILE a (ALESS v (AVAR x)) b c, r)",

    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (SPEC_ALL WHILEt_THM)
    THEN ASM_REWRITE_TAC[]
    THEN CONJ_TAC
    THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
    THENL
      [  EXISTS_TAC "ab_pre b p";
         EXISTS_TAC "ab_pre b q"
      ]
    THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;


let WHILE_THM1 =
 prove_thm
  (`WHILE_THM1`,
   "!p q a b c g r.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             CSPEC(p, c, AAND a (ALESS v (AVAR x)), r)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             CSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)",

    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (SPEC_ALL WHILE_THM)
    THEN ASM_REWRITE_TAC[]
    THEN CONJ_TAC
    THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
    THENL
      [  EXISTS_TAC "ab_pre b p";
         EXISTS_TAC "ab_pre b q"
      ]
    THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;

%< WHILE_THM1
g "!p q a b c g r.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             CSPEC(p, c, AAND a (ALESS v (AVAR x)), r)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             CSPEC(a, WHILE a (ALESS v (AVAR x)) b c, q, r)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL WHILE_THM));;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b p");;
 e(ASM_REWRITE_TAC[ab_pre_THM]);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b q");;
 e(ASM_REWRITE_TAC[ab_pre_THM]);;
>%





%===================================%
%===================================%
%  Now for the Rule of Adaptation!  %
%===================================%
%===================================%


% ============================================================== %
% We need one new syntactic operator for the assertion language, %
% to function as the AFORALL universal quantification operator   %
% for a list of variables, as opposed to quantifying over a      %
% single variable.                                               %
% ============================================================== %

let mk_foralls = 
    new_list_rec_definition (`mk_foralls`,
      "(mk_foralls (NIL) b = b)  /\
       (mk_foralls (CONS a as) b = AFORALL a (mk_foralls as b))");;


let A_mk_foralls =
 prove_thm
  (`A_mk_foralls`,
   "!x p s. A (mk_foralls x p) s =
            (!n. (LENGTH n = LENGTH x) ==> A p (BNDsL n x s))",

   LIST_INDUCT_TAC
    THEN REPEAT GEN_TAC
    THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS;mk_foralls;BNDsL;A_DEF]
    THEN EQ_TAC
    THENL
      [ REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[];

        DISCH_THEN (ACCEPT_TAC
                   o REWRITE_RULE[]
                   o (SPEC "[]:(num)list"));

        REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[BNDsL]
        THEN FIRST_ASSUM (ASSUME_TAC o SPEC "h':num")
        THEN RES_TAC;

        STRIP_TAC
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT STRIP_TAC
        THEN FIRST_ASSUM (ASSUME_TAC o SPEC "CONS (n:num) n'")
        THEN RES_TAC
        THEN REWRITE_ALL_TAC[BNDsL]
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "n:num"
        THEN REFL_TAC
      ]
  );;

%
g "!x p s. A (mk_foralls x p) s =
           (!n. (LENGTH n = LENGTH x) ==> A p (BNDsL n x s))";;

e(LIST_INDUCT_TAC);;

 e(REPEAT GEN_TAC);;
 e(REWRITE_TAC[LENGTH;LENGTH_NIL]);;
 e(REWRITE_TAC[mk_foralls;BNDsL]);;
 e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(DISCH_THEN (ACCEPT_TAC
                o REWRITE_RULE[]
                o (SPEC "[]:(num)list")));;

 e(REWRITE_TAC[LENGTH;LENGTH_CONS]);;
 e(REWRITE_TAC[mk_foralls]);;
 e(REWRITE_TAC[A_DEF]);;
 e(REPEAT GEN_TAC);;
 e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[BNDsL]);;
  e(FIRST_ASSUM (ASSUME_TAC o SPEC "h':num"));;
  e(RES_TAC);;

  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;
  e(REPEAT STRIP_TAC);;
  e(FIRST_ASSUM (ASSUME_TAC o SPEC "CONS (n:num) n'"));;
  e(RES_TAC);;
  e(REWRITE_ALL_TAC[BNDsL]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "n:num");;
  e(REFL_TAC);;
%


let FVa_mk_foralls =
 prove_thm
  (`FVa_mk_foralls`,
   "!xs a. FVa (mk_foralls xs a) = (FVa a DIFF SL xs)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_foralls;SET_LIST;DIFF_EMPTY]
   THEN ASM_REWRITE_TAC[FVa_DEF;DIFF_INSERT]
   THEN REWRITE_TAC[EXTENSION;IN_DELETE;IN_DIFF]
   THEN REPEAT GEN_TAC
   THEN EQ_TAC
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< FVa_mk_foralls
g "!xs a. FVa (mk_foralls xs a) = (FVa a DIFF SL xs)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[mk_foralls;SET_LIST;DIFF_EMPTY]);;

 e(REWRITE_TAC[mk_foralls;SET_LIST;DIFF_EMPTY]);;
 e(ASM_REWRITE_TAC[FVa_DEF;DIFF_INSERT]);;
 e(REWRITE_TAC[EXTENSION]);;
 e(REWRITE_TAC[IN_DELETE;IN_DIFF]);;
 e(REPEAT GEN_TAC);;
 e(EQ_TAC);;

  e(STRIP_TAC THEN ASM_REWRITE_TAC[]);;

  e(STRIP_TAC THEN ASM_REWRITE_TAC[]);;
>%


let ROA_POSTCONDITION_INVARIANCE_DISJOINT =
 TAC_PROOF
  (([],
   "!x a c g r s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ C c r s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (logicals x)))) =
         A a (BNDsR (MAP s1 x) (logicals x) s2))"
   ),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC WFxs_NOT_WFxs
   THEN SUBGOAL_THEN "BNDsL (MAP s2 x) x (BNDsR (MAP s1 x) (logicals x) s1) =
                      BNDsR (MAP s1 x) (logicals x) (BNDsL (MAP s2 x) x (s1:state))"
                REWRITE_THM
   THENL
     [  IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE;

        IMP_RES_TAC C_FREE_ONLY_VARIATION
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% ROA_POSTCONDITION_INVARIANCE_DISJOINT
g "!x a c g r s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ C c r s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (logicals x)))) =
         A a (BNDsR (MAP s1 x) (logicals x) s2))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
e(IMP_RES_TAC WFxs_NOT_WFxs);;
e(SUBGOAL_THEN "BNDsL (MAP s2 x) x (BNDsR (MAP s1 x) (logicals x) s1) =
                BNDsR (MAP s1 x) (logicals x) (BNDsL (MAP s2 x) x (s1:state))"
           REWRITE_THM);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;

  e(IMP_RES_TAC C_FREE_ONLY_VARIATION);;
  e(ASM_REWRITE_TAC[]);;
%



let ROA_POSTCONDITION_INVARIANCE =
 TAC_PROOF
  (([],
   "!x s a c g r s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ C c r s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (variants (logicals x) s)))) =
         A a (BNDsR (MAP s1 x) (variants (logicals x) s) s2))"
   ),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE
   THENL
     [  MATCH_MP_TAC WFxs_NOT_WFxs
        THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals];

        IMP_RES_TAC C_FREE_ONLY_VARIATION
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% ROA_POSTCONDITION_INVARIANCE
g "!x s a c g r s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ C c r s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (variants (logicals x) s)))) =
         A a (BNDsR (MAP s1 x) (variants (logicals x) s) s2))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

  e(IMP_RES_TAC C_FREE_ONLY_VARIATION);;
  e(ASM_REWRITE_TAC[]);;
%





% ===================================================== %
% ===================================================== %
%                  RULE_OF_ADAPTATION:                  %
% ===================================================== %
% ===================================================== %


let RULE_OF_ADAPTATIONp_DISJOINT =
 prove_thm
  (`RULE_OF_ADAPTATIONp_DISJOINT`,
   "!q pre post c g r x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             PCSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
      PCSPEC( AAND pre
                   ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
              c, q, r) )",
   REPEAT GEN_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[PCSPEC;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC A_BNDsR_THM
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "MAP (s1:state) x"
   THEN IMP_RES_TAC (AND2IMP ROA_POSTCONDITION_INVARIANCE_DISJOINT)
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o IMP2AND)
   THEN REWRITE_TAC[LENGTH_MAP]
   THEN POP_ASSUM REWRITE_THM
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
   THEN IMP_RES_TAC (AND2IMP C_BNDsR_THM)
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "BNDsR(MAP s1 x)(logicals x)(s1:state)"
   THEN POP_ASSUM REWRITE_THM
   THEN REWRITE_TAC[(SYM o SPEC_ALL) s_subst_var_list]
   THEN CONJ_TAC
   THENL
     [  MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx
        THEN ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals];

        IMP_RES_TAC SUBSET_WFxs_NOT_WFxs
        THEN IMP_RES_THEN ASM_REWRITE_THM A_var_list_subst_THM
     ]
  );;


let RULE_OF_ADAPTATIONt_DISJOINT =
 prove_thm
  (`RULE_OF_ADAPTATIONt_DISJOINT`,
   "!q pre post c g r x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             TCSPEC(AAND (mk_eqs x0 x) pre, c, r) 
           %---------------------------------------------------% ==>
      TCSPEC( AAND pre
                   ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
              c, r) )",
   REPEAT GEN_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[TCSPEC;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MP_TAC o SPEC "s1 <|s (x // logicals x)")
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN USE_THEN REWRITE_THM A_mk_eqs_NOT_WFx_WFx
        THENL [ ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals]; ALL_TAC ] 
   THEN USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
        THENL [ IMP_RES_TAC SUBSET_WFxs_NOT_WFxs; ALL_TAC ] 
   THEN REWRITE_TAC[s_subst_var_list]
   THEN STRIP_TAC
   THEN EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)"
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
   THEN IMP_RES_TAC (AND2IMP C_BNDsR_THM)
   THEN POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)")
   THEN USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION
        THENL [ REWRITE_TAC[LENGTH_MAP]; ALL_TAC ] 
   THEN REWRITE_TAC[IDENT_BNDsR]
  );;


let RULE_OF_ADAPTATION_DISJOINT =
 prove_thm
  (`RULE_OF_ADAPTATION_DISJOINT`,
   "!q pre post c g r x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             CSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
       CSPEC( AAND pre
                   ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
              c, q, r) )",
   REPEAT GEN_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      RULE_OF_ADAPTATIONp_DISJOINT))
        THEN ASM_REWRITE_TAC[];

        MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      RULE_OF_ADAPTATIONt_DISJOINT))
        THEN ASM_REWRITE_TAC[]
     ]
  );;


%< RULE_OF_ADAPTATION_DISJOINT
g "!q pre post c g r x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             CSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
       CSPEC( AAND pre
                   ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
              c, q, r) )";;
e(REPEAT GEN_TAC
    THEN CONV_TAC let_CONV
    THEN REWRITE_TAC[CSPEC_THM;A_DEF;a_subst_THM;SL_APPEND]
    THEN REWRITE_TAC[A_mk_foralls;A_DEF]
    THEN REPEAT STRIP_TAC);;

e(IMP_RES_TAC A_BNDsR_THM);;
e(POP_TAC
  THEN POP_ASSUM MATCH_MP_TAC
  THEN POP_TAC);;
e(EXISTS_TAC "MAP (s1:state) x");;
e(IMP_RES_TAC (AND2IMP ROA_POSTCONDITION_INVARIANCE_DISJOINT));;
e(POP_TAC
  THEN POP_ASSUM MATCH_MP_TAC);;
e(FIRST_ASSUM (MATCH_MP_TAC o IMP2AND));;
e(REWRITE_TAC[LENGTH_MAP]);;
e(POP_ASSUM REWRITE_THM);;
e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
e(IMP_RES_TAC (AND2IMP WFc_NOT_WFxs));;
e(IMP_RES_TAC (AND2IMP C_BNDsR_THM));;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "BNDsR(MAP s1 x)(logicals x)(s1:state)");;
e(POP_ASSUM REWRITE_THM);;
e(REWRITE_TAC[(SYM o SPEC_ALL) s_subst_var_list]);;
e(CONJ_TAC);;

  e(MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx);;
  e(ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals]);;

  e(IMP_RES_TAC SUBSET_WFxs_NOT_WFxs);;
  e(IMP_RES_THEN ASM_REWRITE_THM A_var_list_subst_THM);;


e(FIRST_ASSUM (MP_TAC o SPEC "s1 <|s (x // logicals x)"));;
e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
e(USE_THEN REWRITE_THM A_mk_eqs_NOT_WFx_WFx);;

  e(ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals]);;

e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;

  e(IMP_RES_TAC SUBSET_WFxs_NOT_WFxs);;

e(REWRITE_TAC[s_subst_var_list]);;
e(STRIP_TAC);;
e(EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)");;
e(IMP_RES_TAC (AND2IMP WFc_NOT_WFxs));;
e(IMP_RES_TAC (AND2IMP C_BNDsR_THM));;
e(POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)"));;
e(USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION);;

  e(REWRITE_TAC[LENGTH_MAP]);;

e(REWRITE_TAC[IDENT_BNDsR]);;
>%



let mk_eqs_variant_subst =
 prove_thm
  (`mk_eqs_variant_subst`,
   "!x s.  let x0  = logicals x in
           let x0' = variants x0 s in
           (DL x /\ WFxs x  ==>
           (((mk_eqs x0 x) <|a (x0' // x0)) = (mk_eqs x0' x)))",
   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[var_list_subst]
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;variants_THM;mk_eqs_THM;a_subst;v_subst]
   THEN REWRITE_TAC[DISJOINT_LIST;WFxs;aexp_one_one]
   THEN REWRITE_TAC[MAP;BNDsR;BND_MATCH]
   THEN REPEAT STRIP_TAC
   THENL
     [  USE_IMP_THEN REWRITE_THM BND_NOT_MATCH
        THENL
          [  REWRITE_TAC[logical_NOT_IDENT];

             USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] BNDsR_NOT_MATCH)
             THENL
               [ MATCH_MP_TAC WFx_NOT_WFxs
                 THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

                 REWRITE_TAC[IDENT_SS]
               ]
          ];

        USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] mk_eqs_subst_BND)
        THENL
          [  USE_IMP_THEN REWRITE_THM FVa_mk_eqs
             THENL
               [ REWRITE_TAC[LENGTH_logicals];

                 REWRITE_TAC[LENGTH_logicals;IN_UNION;DE_MORGAN_THM]
                 THEN ASM_REWRITE_TAC[IN_logicals_11]
                 THEN MATCH_MP_TAC WFxs_NOT_WFx
                 THEN ASM_REWRITE_TAC[NOT_WFx_logical]
               ];

             RES_TAC
             THEN ASM_REWRITE_TAC[]
          ];
     ]
  );;

%
g "!x s.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (DL x /\ WFxs x  ==>
           (((mk_eqs x0 x) <|a (x0' // x0)) = (mk_eqs x0' x)))";;

e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[var_list_subst]);;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[logicals_THM;variants_THM;mk_eqs_THM;a_subst]);;

 e(REWRITE_TAC[logicals_THM;variants_THM;mk_eqs_THM;a_subst;v_subst]);;
 e(REWRITE_TAC[DISJOINT_LIST;WFxs]);;
 e(REWRITE_TAC[aexp_one_one]);;
 e(REWRITE_TAC[MAP;BNDsR]);;
 e(REWRITE_TAC[BND_MATCH]);;
 e(REPEAT STRIP_TAC);;

  e(USE_IMP_THEN REWRITE_THM BND_NOT_MATCH);;
    e(REWRITE_TAC[logical_NOT_IDENT]);;
  e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] BNDsR_NOT_MATCH));;
    e(MATCH_MP_TAC WFx_NOT_WFxs);;
    e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;
  e(REWRITE_TAC[IDENT_SS]);;

  e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] mk_eqs_subst_BND));;
    e(USE_IMP_THEN REWRITE_THM FVa_mk_eqs);;

      e(REWRITE_TAC[LENGTH_logicals]);;

      e(REWRITE_TAC[LENGTH_logicals;IN_UNION;DE_MORGAN_THM]);;
      e(ASM_REWRITE_TAC[IN_logicals_11]);;
      e(MATCH_MP_TAC WFxs_NOT_WFx);;
      e(ASM_REWRITE_TAC[NOT_WFx_logical]);;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;
%


let C_state_variants_logicals =
 prove_thm
  (`C_state_variants_logicals`,
   "!x s c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\
     C c r s1 s2 ==>
     ((MAP s2 (variants(logicals x)s)) = (MAP s1 (variants(logicals x)s)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;variants_THM;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT)
        THEN POP_ASSUM MATCH_MP_TAC
        THEN MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFx)
        THEN ASM_REWRITE_TAC[WFx_variant;NOT_WFx_logical];

        RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!x s c g r s1 s2.
     WFc c g r /\ WFenv_syntax r /\
     C c r s1 s2 ==>
     ((MAP s2 (variants(logicals x)s)) = (MAP s1 (variants(logicals x)s)))";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[logicals_THM;variants_THM;MAP]);;

 e(REWRITE_TAC[logicals_THM;variants_THM;MAP]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REPEAT STRIP_TAC);;

  e(IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT));;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFx));;
  e(ASM_REWRITE_TAC[WFx_variant;NOT_WFx_logical]);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%


let PCSPEC_logicals_SUBSTITUTION =
 prove_thm
  (`PCSPEC_logicals_SUBSTITUTION`,
   "!x s c g r pre post.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
           (FVa post SUBSET (SL (APPEND x x0)))   /\
            PCSPEC(AAND (mk_eqs x0  x) pre, c, post, r)  
           ==>
            PCSPEC(AAND (mk_eqs x0' x) pre, c, post <|a (x0' // x0), r))",

   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[PCSPEC;A_DEF;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "(s1:state) <|s ((variants(logicals x)s) // (logicals x))"
   THEN REPEAT CONJ_TAC
   THENL
          [  IMP_RES_TAC (AND2IMP (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst))
             THEN ASM_REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)];

             USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
             THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
             THEN IMP_RES_TAC (AND2IMP SUBSET_WFxs_NOT_WFxs);

             REWRITE_TAC[s_subst_var_list]
             THEN IMP_RES_TAC (AND2IMP C_state_variants_logicals)
             THEN POP_ASSUM REWRITE_THM
             THEN MATCH_MP_TAC (SPEC_ALL C_BNDsR_THM)
             THEN ASM_REWRITE_TAC[]
             THEN MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFxs)
             THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
          ]
  );;


let TCSPEC_logicals_SUBSTITUTION =
 prove_thm
  (`TCSPEC_logicals_SUBSTITUTION`,
   "!x s c g r pre.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
            TCSPEC(AAND (mk_eqs x0  x) pre, c, r)  
           ==>
            TCSPEC(AAND (mk_eqs x0' x) pre, c, r))",

   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[TCSPEC;A_DEF;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MP_TAC o
               SPEC "(s1:state) <|s
                     ((variants(logicals x)s) // (logicals x))")
   THEN REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)]
   THEN USE_IMP_THEN ASM_REWRITE_THM
              (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst)
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN REWRITE_TAC[a_subst_THM]
   THEN USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
   THENL
     [  ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
        THEN IMP_RES_TAC (AND2IMP SUBSET_WFxs_NOT_WFxs);

        ALL_TAC
     ]
   THEN REWRITE_TAC[s_subst_var_list]
   THEN STRIP_TAC
   THEN EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)"
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
   THEN IMP_RES_TAC (AND2IMP C_BNDsR_THM)
   THEN POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)")
   THEN USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION
   THENL
     [  REWRITE_TAC[LENGTH_MAP];
        REWRITE_TAC[IDENT_BNDsR]
     ]
  );;


let CSPEC_logicals_SUBSTITUTION =
 prove_thm
  (`CSPEC_logicals_SUBSTITUTION`,
   "!x s c g r pre post.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
           (FVa post SUBSET (SL (APPEND x x0)))   /\
            CSPEC(AAND (mk_eqs x0  x) pre, c, post, r)  
           ==>
            CSPEC(AAND (mk_eqs x0' x) pre, c, post <|a (x0' // x0), r))",

   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      PCSPEC_logicals_SUBSTITUTION))
        THEN ASM_REWRITE_TAC[];

        MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      TCSPEC_logicals_SUBSTITUTION))
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!x s c g r pre post.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
           (FVa post SUBSET (SL (APPEND x x0)))   /\
            CSPEC(AAND (mk_eqs x0  x) pre, c, post, r)  
           ==>
            CSPEC(AAND (mk_eqs x0' x) pre, c, post <|a (x0' // x0), r))";;

e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[CSPEC_THM]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[a_subst_THM]);;
e(REPEAT STRIP_TAC);;

 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "(s1:state) <|s ((variants(logicals x)s) // (logicals x))");;
 e(REPEAT CONJ_TAC);;

  e(REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)]);;
  e(IMP_RES_TAC (AND2IMP (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst)));;
  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;
  e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
  e(IMP_RES_TAC (AND2IMP SUBSET_WFxs_NOT_WFxs));;

  e(REWRITE_TAC[s_subst_var_list]);;
  e(IMP_RES_TAC (AND2IMP C_state_variants_logicals));;
  e(POP_ASSUM REWRITE_THM);;
  e(MATCH_MP_TAC (SPEC_ALL C_BNDsR_THM));;
  e(ASM_REWRITE_TAC[]);;
  e(MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFxs));;
  e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

 e(FIRST_ASSUM (MP_TAC o
       SPEC "(s1:state) <|s ((variants(logicals x)s) // (logicals x))"));;
 e(REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)]);;
 e(USE_IMP_THEN ASM_REWRITE_THM
       (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst));;

   e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[a_subst_THM]);;
 e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;

   e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
   e(IMP_RES_TAC (AND2IMP SUBSET_WFxs_NOT_WFxs));;

 e(REWRITE_TAC[s_subst_var_list]);;
 e(STRIP_TAC);;
 e(EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)");;
 e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
 e(IMP_RES_TAC (AND2IMP WFc_NOT_WFxs));;
 e(IMP_RES_TAC (AND2IMP C_BNDsR_THM));;
 e(POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)"));;
 e(USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION);;

  e(REWRITE_TAC[LENGTH_MAP]);;

 e(REWRITE_TAC[IDENT_BNDsR]);;
%


% ================================= %
% RULE_OF_ADAPTATION (General Case) %
% ================================= %


let RULE_OF_ADAPTATIONp =
 prove_thm
  (`RULE_OF_ADAPTATIONp`,
   "!q r pre post c g x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             PCSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
       PCSPEC( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, q, r) )",

   REPEAT GEN_TAC
   THEN REPEAT (CONV_TAC let_CONV)
   THEN REWRITE_TAC[PCSPEC;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN ASSUME_TAC (SPECL ["logicals x";"FVa q"] DISJOINT_variants)
   THEN IMP_RES_TAC (AND2IMP A_BNDsR_THM)
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "MAP (s1:state) x"
   THEN USE_IMP_THEN REWRITE_THM
            (SPEC_ALL (ONCE_REWRITE_RULE[EQ_SYM_EQ]
                               ROA_POSTCONDITION_INVARIANCE))
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN FIRST_ASSUM (MATCH_MP_TAC o IMP2AND)
   THEN REWRITE_TAC[LENGTH_MAP]
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL ROA_POSTCONDITION_INVARIANCE)
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN SUBGOAL_THEN
        "PCSPEC(AAND (mk_eqs (variants(logicals x)(FVa q)) x) pre, c,
                post <|a ((variants(logicals x)(FVa q)) // (logicals x)),r)"
           (STRIP_ASSUME_TAC o REWRITE_RULE[PCSPEC;A_DEF;IMP_CONJ_EQ])
   THENL
         [ MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                PCSPEC_logicals_SUBSTITUTION))
           THEN REWRITE_TAC[PCSPEC;A_DEF;SL_APPEND]
           THEN ASM_REWRITE_TAC[];

           POP_ASSUM MATCH_MP_TAC
           THEN EXISTS_TAC "BNDsR (MAP s1 x) (variants(logicals x)(FVa q)) (s1:state)"
           THEN REWRITE_TAC[SYM (SPEC_ALL s_subst_var_list)]
           THEN REPEAT CONJ_TAC
           THENL  % 3 subgoals %
            [ MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx
              THEN ASM_REWRITE_TAC[LENGTH_variants;LENGTH_logicals;DL_variants;
                                   NOT_WFxs_variants;NOT_WFxs_logicals];

              USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
              THEN MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs
              THEN EXISTS_TAC "x:(var)list"
              THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals];

              REWRITE_TAC[s_subst_var_list]
              THEN MATCH_MP_TAC (SPEC_ALL C_BNDsR_THM)
              THEN ASM_REWRITE_TAC[]
              THEN MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs
              THEN EXISTS_TAC "x:(var)list"
              THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
            ]
         ]
  );;


let RULE_OF_ADAPTATIONt =
 prove_thm
  (`RULE_OF_ADAPTATIONt`,
   "!q r pre post c g x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             TCSPEC(AAND (mk_eqs x0 x) pre, c, r) 
           %---------------------------------------------------% ==>
       TCSPEC( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, r) )",

   REPEAT GEN_TAC
   THEN REPEAT (CONV_TAC let_CONV)
   THEN REWRITE_TAC[TCSPEC;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_TAC
   THEN FIRST_ASSUM
           (MP_TAC o SPEC "BNDsR (MAP s1 x) (logicals x) (s1:state)")
   THEN USE_IMP_THEN REWRITE_THM A_mk_eqs_THM
        THENL [ REWRITE_TAC[LENGTH_logicals]; ALL_TAC ]
   THEN USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT)
        THENL [ ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals]; ALL_TAC ]
   THEN USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
   THENL
         [ MATCH_MP_TAC WFxs_NOT_WFxs
           THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

           ALL_TAC
         ]
   THEN USE_IMP_THEN ASM_REWRITE_THM A_BNDsR_THM
   THENL
         [ MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs
           THEN EXISTS_TAC "x:var list"
           THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

           ALL_TAC
         ]
   THEN STRIP_TAC
   THEN EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)"
   THEN ASSUME_TAC (SPEC "x:var list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
   THEN IMP_RES_TAC (AND2IMP C_BNDsR_THM)
   THEN POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)")
   THEN USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION
   THENL
         [ REWRITE_TAC[LENGTH_MAP];

           REWRITE_TAC[IDENT_BNDsR]
         ]
  );;


let RULE_OF_ADAPTATION =
 prove_thm
  (`RULE_OF_ADAPTATION`,
   "!q r pre post c g x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             CSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
        CSPEC( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, q, r) )",

   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      RULE_OF_ADAPTATIONp))
        THEN ASM_REWRITE_TAC[];

        MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      RULE_OF_ADAPTATIONt))
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%<
g "!q r pre post c g x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             CSPEC(AAND (mk_eqs x0 x) pre, c, post, r) 
           %---------------------------------------------------% ==>
        CSPEC( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, q, r) )";;

e(REPEAT GEN_TAC
    THEN REPEAT (CONV_TAC let_CONV)
    THEN REWRITE_TAC[CSPEC_THM;A_DEF;a_subst_THM;SL_APPEND]
    THEN REWRITE_TAC[A_mk_foralls;A_DEF]
    THEN REPEAT STRIP_TAC);;

 e(ASSUME_TAC (SPECL ["logicals x";"FVa q"] DISJOINT_variants));;
 e(IMP_RES_TAC (AND2IMP A_BNDsR_THM));;
 e(POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN POP_TAC);;
 e(EXISTS_TAC "MAP (s1:state) x");;
 e(USE_IMP_THEN REWRITE_THM
      (SPEC_ALL (ONCE_REWRITE_RULE[EQ_SYM_EQ] ROA_POSTCONDITION_INVARIANCE)));;

   e(ASM_REWRITE_TAC[]);;

 e(FIRST_ASSUM (MATCH_MP_TAC o IMP2AND));;
 e(REWRITE_TAC[LENGTH_MAP]);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL ROA_POSTCONDITION_INVARIANCE));;

   e(ASM_REWRITE_TAC[]);;

 e(SUBGOAL_THEN
     "CSPEC(AAND (mk_eqs (variants(logicals x)(FVa q)) x) pre, c,
             post <|a ((variants(logicals x)(FVa q)) // (logicals x)),r)"
        (STRIP_ASSUME_TAC o REWRITE_RULE[CSPEC_THM;A_DEF;IMP_CONJ_EQ]));;

   e(MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          CSPEC_logicals_SUBSTITUTION)));;
   e(REWRITE_TAC[CSPEC_THM;A_DEF;SL_APPEND]);;
   e(ASM_REWRITE_TAC[]);;

   e(POP_TAC THEN POP_ASSUM MATCH_MP_TAC);;
   e(EXISTS_TAC "BNDsR (MAP s1 x) (variants(logicals x)(FVa q)) (s1:state)");;
   e(REWRITE_TAC[SYM (SPEC_ALL s_subst_var_list)]);;
   e(REPEAT CONJ_TAC);;  % 3 subgoals %

     e(MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx);;
     e(ASM_REWRITE_TAC[LENGTH_variants;LENGTH_logicals;DL_variants;
                       NOT_WFxs_variants;NOT_WFxs_logicals]);;

     e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;
     e(MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs);;
     e(EXISTS_TAC "x:(var)list");;
     e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

     e(REWRITE_TAC[s_subst_var_list]);;
     e(MATCH_MP_TAC (SPEC_ALL C_BNDsR_THM));;
     e(ASM_REWRITE_TAC[]);;
     e(MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs);;
     e(EXISTS_TAC "x:(var)list");;
     e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

let st1 = get_state();;
set_state st1;;
     e(POP_TAC);;  %unnecessary for tactic, but shortens printouts%
     e(FIRST_ASSUM
          (MP_TAC o SPEC "BNDsR (MAP s1 x) (logicals x) (s1:state)"));;
     e(USE_IMP_THEN REWRITE_THM A_mk_eqs_THM);;

       e(REWRITE_TAC[LENGTH_logicals]);;

     e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] MAP_BNDsR_IDENT));;

       e(ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals]);;

     e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

       e(MATCH_MP_TAC WFxs_NOT_WFxs);;
       e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

     e(USE_IMP_THEN ASM_REWRITE_THM A_BNDsR_THM);;

       e(MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs);;
       e(EXISTS_TAC "x:var list");;
       e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

     e(STRIP_TAC);;
     e(EXISTS_TAC "BNDsR (MAP s1 (logicals x)) (logicals x) (s2:state)");;
     e(ASSUME_TAC (SPEC "x:var list" NOT_WFxs_logicals));;
     e(IMP_RES_TAC WFc_NOT_WFxs);;
     e(IMP_RES_TAC C_BNDsR_THM);;
     e(POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)"));;
     e(USE_IMP_THEN REWRITE_THM BNDsR_BNDsR_ABSORPTION);;

       e(REWRITE_TAC[LENGTH_MAP]);;

     e(REWRITE_TAC[IDENT_BNDsR]);;

>%




let mk_eqs_logicals_vsubst =
 TAC_PROOF
  (([],
   "!x y. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
               WFxs x /\ WFxs y ==>
   (mk_eqs(logicals y)y =
   (mk_eqs(logicals (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x))
                    (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x)))"
   ),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_REPLACEMENT
   THEN ASM_REWRITE_TAC[]
  );;

% mk_eqs_logicals_vsubst
g "!x y. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y ==>
   (mk_eqs(logicals y)y =
   (mk_eqs(logicals (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x))
                    (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x)))";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_REPLACEMENT);;
e(ASM_REWRITE_TAC[]);;
%




let PCSPEC_pre_post_MUTATION =
 TAC_PROOF
  (([],
   "!x y z. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ WFxs y /\
             (FVa pre  SUBSET SL x) /\
             (FVa post SUBSET SL (APPEND x (logicals x)))  ==>
    let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
    (PCSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                 (pre <|a (y // x)),
             c <|c ss,
             post <|a (APPEND y (logicals y) // APPEND x (logicals x)),
             r) =
     PCSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                 (pre <|av ss),
             c <|c ss,
             post <|av ss,
             r))"
   ),

   REPEAT STRIP_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[PCSPEC;A_DEF]
   THEN IMP_RES_TAC WFxs_NOT_logicals
   THEN  FORALL_EQ_TAC
   THEN FORALL_EQ_TAC
   THEN USE_IMP_THEN REWRITE_THM FVa_vsubst_subst
   THENL
       [ ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                         DL_APPEND;DL_logicals]
         THEN MATCH_MP_TAC SUBSET_TRANS
         THEN EXISTS_TAC "SL (x:(var)list)"
         THEN ASM_REWRITE_TAC[SUBSET_UNION];

         USE_IMP_THEN REWRITE_THM FVa_vsubst_subst
         THENL
          [ ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                            DL_APPEND;DL_logicals];

            IMP_RES_TAC A_NULL_APPEND_subst
            THEN ASM_REWRITE_TAC[]
          ]
       ]
  );;


let TCSPEC_pre_post_MUTATION =
 TAC_PROOF
  (([],
   "!x y z. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ WFxs y /\
             (FVa pre  SUBSET SL x)  ==>
    let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
    (TCSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                 (pre <|a (y // x)),
             c <|c ss,
             r) =
     TCSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                 (pre <|av ss),
             c <|c ss,
             r))"
   ),

   REPEAT STRIP_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[TCSPEC;A_DEF]
   THEN IMP_RES_TAC WFxs_NOT_logicals
   THEN FORALL_EQ_TAC
   THEN USE_IMP_THEN REWRITE_THM FVa_vsubst_subst
   THENL
       [ ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                         DL_APPEND;DL_logicals]
         THEN MATCH_MP_TAC SUBSET_TRANS
         THEN EXISTS_TAC "SL (x:(var)list)"
         THEN ASM_REWRITE_TAC[SUBSET_UNION];

         IMP_RES_TAC A_NULL_APPEND_subst
         THEN ASM_REWRITE_TAC[]
       ]
  );;


let CSPEC_pre_post_MUTATION =
 TAC_PROOF
  (([],
   "!x y z. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ WFxs y /\
             (FVa pre  SUBSET SL x) /\
             (FVa post SUBSET SL (APPEND x (logicals x)))  ==>
    let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
    (CSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                (pre <|a (y // x)),
            c <|c ss,
            post <|a (APPEND y (logicals y) // APPEND x (logicals x)),
            r) =
     CSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                (pre <|av ss),
            c <|c ss,
            post <|av ss,
            r))"
   ),

   REPEAT STRIP_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[CSPEC]
   THEN MATCH_MP_TAC CONJ_11
   THEN CONJ_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      PCSPEC_pre_post_MUTATION))
        THEN ASM_REWRITE_TAC[];

        MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                                      TCSPEC_pre_post_MUTATION))
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% CSPEC_pre_post_MUTATION
g "!x y z. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
           WFxs x /\ WFxs y /\
           (FVa pre  SUBSET SL x) /\
           (FVa post SUBSET SL (APPEND x (logicals x)))  ==>
  let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
  (CSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
              (pre <|a (y // x)),
          c <|c ss,
          post <|a (APPEND y (logicals y) // APPEND x (logicals x)),
          r) =
   CSPEC(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
               (pre <|av ss),
          c <|c ss,
          post <|av ss,
          r))";;
e(REPEAT STRIP_TAC);;
e(CONV_TAC let_CONV);;
e(REWRITE_TAC[CSPEC_THM;A_DEF]);;
e(MATCH_MP_TAC CONJ_11);;
e(IMP_RES_TAC WFxs_NOT_logicals);;
e(CONJ_TAC);;

 e(FORALL_EQ_TAC);;
 e(FORALL_EQ_TAC);;
 e(USE_IMP_THEN (REWRITE_THM o pthm) FVa_vsubst_subst);;

   e(ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                     DL_APPEND;DL_logicals]);;
   e(MATCH_MP_TAC SUBSET_TRANS);;
   e(EXISTS_TAC "SL (x:(var)list)");;
   e(ASM_REWRITE_TAC[SUBSET_UNION]);;

   e(USE_IMP_THEN (REWRITE_THM o pthm) FVa_vsubst_subst);;

     e(ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                       DL_APPEND;DL_logicals]);;

     e(IMP_RES_TAC A_NULL_APPEND_subst);;
     e(ASM_REWRITE_TAC[]);;

 e(FORALL_EQ_TAC);;
 e(USE_IMP_THEN (REWRITE_THM o pthm) FVa_vsubst_subst);;

   e(ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                     DL_APPEND;DL_logicals]);;
   e(MATCH_MP_TAC SUBSET_TRANS);;
   e(EXISTS_TAC "SL (x:(var)list)");;
   e(ASM_REWRITE_TAC[SUBSET_UNION]);;

   e(IMP_RES_TAC A_NULL_APPEND_subst);;
   e(ASM_REWRITE_TAC[]);;
%





let MAP_STANDARD_SUBST_GLOBALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          glbs =
      glbs)"),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
   THENL
     [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION];

        USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2
        THENL
          [  ASM_REWRITE_TAC[SUBSET_REFL];

             MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% MAP_STANDARD_SUBST_GLOBALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          glbs =
      glbs)";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION]);;

  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2);;

    e(ASM_REWRITE_TAC[SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[]);;
%


let MAP_STANDARD_SUBST_LOCALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          vs =
      vs')"),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
   THENL
     [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION];

        USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
        THENL
          [  ASM_REWRITE_TAC[SUBSET_REFL];

             MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% MAP_STANDARD_SUBST_LOCALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          vs =
      vs')";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION]);;

  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[]);;
%



let MAP_STANDARD_SUBST_LOCALS_AND_GLOBALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          (APPEND vs glbs) =
      APPEND vs' glbs)"),
   REPEAT STRIP_TAC
   THEN ONCE_REWRITE_TAC[MAP_APPEND]
   THEN USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_LOCALS
   THENL
     [  ASM_REWRITE_TAC[];

        USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_GLOBALS
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% MAP_STANDARD_SUBST_LOCALS_AND_GLOBALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          (APPEND vs glbs) =
      APPEND vs' glbs)";;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[MAP_APPEND]);;
e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_LOCALS);;

  e(ASM_REWRITE_TAC[]);;

e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_GLOBALS);;
e(ASM_REWRITE_TAC[]);;
%



let PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL =
 prove_thm
  (`PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL`,
   "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenvp r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((GVc c' r) SUBSET (SL glbs)) /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPEC (AAND (mk_eqs x0' x') pre', c', post', r))",

   REPEAT GEN_TAC THEN STRIP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN SUBGOAL_THEN
          "WFcsubst c r
               ((APPEND (APPEND (APPEND xs vals') glbs)
                  (logicals (APPEND (APPEND xs vals') glbs))) //v
                (APPEND (APPEND (APPEND vars vals) glbs)
                  (logicals (APPEND (APPEND vars vals) glbs))))"
             ASSUME_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          WFcsubst_standard_subst))
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC WELL_FORMED_PROCEDURE_PARTIAL
        THEN REWRITE_ALL_TAC[DL_APPEND;WFxs_APPEND]
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC  % 6 subgoals %
        THENL
          [  IMP_RES_THEN ASM_REWRITE_THM WFc_subst;

             ASM_REWRITE_TAC[GVc_subst];

             USE_IMP_THEN REWRITE_THM FVc_subst
             THENL
               [  EXISTS_TAC "glbs:(var)list"
                  THEN ASM_REWRITE_TAC[]
                  THEN MATCH_MP_TAC MAP_STANDARD_SUBST_GLOBALS
                  THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                  USE_IMP_THEN REWRITE_THM IMAGE_NULL_APPEND_vsubst
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM vsubst_SUBSET
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]
                    ]
               ];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;LENGTH_logicals;
                                  DL_logicals;logicals_APPEND;DISJOINT_APPEND]
             THEN REPEAT CONJ_TAC
             THEN MATCH_MP_TAC WFxs_NOT_logicals
             THEN ASM_REWRITE_TAC[];

             USE_IMP_THEN ONCE_REWRITE_THM
                      (SPEC "APPEND (APPEND vars vals) (glbs:(var)list)"
                            mk_eqs_logicals_vsubst)
             THENL
               [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                  USE_IMP_THEN REWRITE_THM
                      (CONV_RULE (DEPTH_CONV let_CONV) PCSPEC_pre_post_MUTATION)
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM PCSPEC_biject
                       THEN EXISTS_TAC "calls:prog_env"
                       THEN ASM_REWRITE_TAC[WFxs_APPEND]
                       THEN MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;WFxs_APPEND]
                    ]
               ]
          ]
     ]
  );;


let PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL_COR =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenvp r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPEC (AAND (mk_eqs x0' x') pre', c', post', r))"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL)
   THEN UNDISCH_LAST_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;


let PROCEDURE_CHANGE_OF_VARIABLES =
 prove_thm
  (`PROCEDURE_CHANGE_OF_VARIABLES`,
   "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((GVc c' r) SUBSET (SL glbs)) /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       CSPEC (AAND (mk_eqs x0' x') pre', c', post', r))",

   REPEAT GEN_TAC THEN STRIP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN SUBGOAL_THEN
          "WFcsubst c r
               ((APPEND (APPEND (APPEND xs vals') glbs)
                  (logicals (APPEND (APPEND xs vals') glbs))) //v
                (APPEND (APPEND (APPEND vars vals) glbs)
                  (logicals (APPEND (APPEND vars vals) glbs))))"
             ASSUME_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          WFcsubst_standard_subst))
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC WELL_FORMED_PROCEDURE
        THEN REWRITE_ALL_TAC[DL_APPEND;WFxs_APPEND]
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC  % 6 subgoals %
        THENL
          [  IMP_RES_THEN ASM_REWRITE_THM WFc_subst;

             ASM_REWRITE_TAC[GVc_subst];

             USE_IMP_THEN REWRITE_THM FVc_subst
             THENL
               [  EXISTS_TAC "glbs:(var)list"
                  THEN ASM_REWRITE_TAC[]
                  THEN MATCH_MP_TAC MAP_STANDARD_SUBST_GLOBALS
                  THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                  USE_IMP_THEN REWRITE_THM IMAGE_NULL_APPEND_vsubst
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM vsubst_SUBSET
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]
                    ]
               ];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;LENGTH_logicals;
                                  DL_logicals;logicals_APPEND;DISJOINT_APPEND]
             THEN REPEAT CONJ_TAC
             THEN MATCH_MP_TAC WFxs_NOT_logicals
             THEN ASM_REWRITE_TAC[];

             USE_IMP_THEN ONCE_REWRITE_THM
                      (SPEC "APPEND (APPEND vars vals) (glbs:(var)list)"
                            mk_eqs_logicals_vsubst)
             THENL
               [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                  USE_IMP_THEN REWRITE_THM
                      (CONV_RULE (DEPTH_CONV let_CONV) CSPEC_pre_post_MUTATION)
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM CSPEC_vsubst
                       THEN EXISTS_TAC "calls:prog_env"
                       THEN ASM_REWRITE_TAC[WFxs_APPEND]
                       THEN MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;WFxs_APPEND]
                    ]
               ]
          ]
     ]
  );;

%< PROCEDURE_CHANGE_OF_VARIABLES
g "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((GVc c' r) SUBSET (SL glbs)) /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       CSPEC (AAND (mk_eqs x0' x') pre', c', post', r))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(SUBGOAL_THEN
    "WFcsubst c r
            ((APPEND (APPEND (APPEND xs vals') glbs)
               (logicals (APPEND (APPEND xs vals') glbs))) //v
             (APPEND (APPEND (APPEND vars vals) glbs)
               (logicals (APPEND (APPEND vars vals) glbs))))"
       ASSUME_TAC);;

 e(MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                        WFcsubst_standard_subst)));;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
 e(REWRITE_ALL_TAC[DL_APPEND;WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT CONJ_TAC);;  % 6 subgoals %

  e(IMP_RES_THEN ASM_REWRITE_THM WFc_subst);;

  e(ASM_REWRITE_TAC[GVc_subst]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:(var)list");;
   e(ASM_REWRITE_TAC[]);;
   e(MATCH_MP_TAC MAP_STANDARD_SUBST_GLOBALS);;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

   e(USE_IMP_THEN REWRITE_THM IMAGE_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

    e(USE_IMP_THEN ASM_REWRITE_THM vsubst_SUBSET);;
    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

  e(MATCH_MP_TAC FVa_subst_SUBSET);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

  e(MATCH_MP_TAC FVa_subst_SUBSET);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;LENGTH_logicals;
                    DL_logicals;logicals_APPEND;DISJOINT_APPEND]);;
  e(REPEAT CONJ_TAC
    THEN MATCH_MP_TAC WFxs_NOT_logicals
    THEN ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN ONCE_REWRITE_THM
         (SPEC "APPEND (APPEND vars vals) (glbs:(var)list)"
               mk_eqs_logicals_vsubst));;

   e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND]);;

   e(USE_IMP_THEN (REWRITE_THM o pthm)
         (CONV_RULE (DEPTH_CONV let_CONV) CSPEC_pre_post_MUTATION));;

    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND]);;

    e(USE_IMP_THEN ASM_REWRITE_THM CSPEC_vsubst);;
    e(EXISTS_TAC "calls:prog_env");;
    e(ASM_REWRITE_TAC[WFxs_APPEND]);;
    e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
    e(ASM_REWRITE_TAC[LENGTH_APPEND;WFxs_APPEND]);;
>%


let PROCEDURE_CHANGE_OF_VARIABLES_COR =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       CSPEC (AAND (mk_eqs x0' x') pre', c', post', r))"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP PROCEDURE_CHANGE_OF_VARIABLES)
   THEN UNDISCH_LAST_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% PROCEDURE_CHANGE_OF_VARIABLES_COR
g "!r p vars vals glbs pre post calls rec c xs vals'.
      WFenv r /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       CSPEC (AAND (mk_eqs x0' x') pre', c', post', r))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP PROCEDURE_CHANGE_OF_VARIABLES));;
e(UNDISCH_LAST_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%





let C_SUBSTITUTION_EQ =
 prove_thm
  (`C_SUBSTITUTION_EQ`,
   "!vals' vals'' r p vars vals glbs pre post calls rec c g xs es s1 s2.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     WFxs vals' /\ WFxs vals'' /\
     (LENGTH vals' = LENGTH vals) /\ (LENGTH vals'' = LENGTH vals) /\
     DL vals' /\ DL vals'' /\
     DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
     DISJOINT (SL vals'') (SL (APPEND xs glbs)) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (C ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r s1 s2  =
      C  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r s1 s2)",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]
   THEN MATCH_MP_TAC c_subst_EQ
   THEN STRIP_ASSUME_TAC
           (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
            "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN SUBGOAL_THEN "WFc c calls r /\ WFxs glbs /\
                      DL (APPEND (APPEND vars vals) glbs) /\
                      DL (APPEND (APPEND xs vals' ) glbs) /\
                      DL (APPEND (APPEND xs vals'') glbs) /\
                      FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
           STRIP_ASSUME_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[]
      THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN ASM_REWRITE_TAC[];

      REPEAT STRIP_TAC
      THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
      THEN REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]
      THEN UNDISCH_LAST_TAC
      THEN SPEC_TAC("x:var","x:var")
      THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]
      THEN REWRITE_TAC[MAP_o;o_THM]
      THEN REWRITE_TAC[MAP_APPEND]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN REPEAT CONJ_TAC
      THENL  % 3 subgoals %
       [ USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN MATCH_MP_TAC IDENT_MAP_vsubst
         THEN ASM_REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals;

         USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
         THEN ASM_REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals
       ]
    ]
  );;

%< C_SUBSTITUTION_EQ

g "!vals' vals'' r p vars vals glbs pre post calls rec c g xs es s1 s2.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     WFxs vals' /\ WFxs vals'' /\
     (LENGTH vals' = LENGTH vals) /\ (LENGTH vals'' = LENGTH vals) /\
     DL vals' /\ DL vals'' /\
     DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
     DISJOINT (SL vals'') (SL (APPEND xs glbs)) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (C ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r s1 s2  =
      C  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r s1 s2)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]);;
e(MATCH_MP_TAC c_subst_EQ);;
e(STRIP_ASSUME_TAC
     (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
       "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(SUBGOAL_THEN "WFc c calls r /\ WFxs glbs /\
                DL (APPEND (APPEND vars vals) glbs) /\
                DL (APPEND (APPEND xs vals' ) glbs) /\
                DL (APPEND (APPEND xs vals'') glbs) /\
                FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
     STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]);;
 e(UNDISCH_LAST_TAC);;
 e(SPEC_TAC("x:var","x:var"));;
 e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]);;
 e(REWRITE_TAC[MAP_o;o_THM]);;
 e(REWRITE_TAC[MAP_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM APPEND_11);;

  e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

  e(USE_IMP_THEN REWRITE_THM APPEND_11);;

   e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

   e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
   e(REPEAT CONJ_TAC);;  % 3 subgoals %

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
            (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

      e(ASM_REWRITE_TAC[SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[]);;

       e(ASM_REWRITE_TAC[]);;
       e(MATCH_MP_TAC IDENT_MAP_vsubst);;
       e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;
       e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
       e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

      e(ASM_REWRITE_TAC[SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[]);;

       e(USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

        e(ASM_REWRITE_TAC[]);;

        e(ASM_REWRITE_TAC[]);;
        e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

         e(ASM_REWRITE_TAC[SUBSET_REFL]);;

         e(USE_IMP_THEN REWRITE_THM
                (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
         e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

     e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

     e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

      e(ASM_REWRITE_TAC[]);;
      e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;
      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;
      e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
      e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;

>%


% For the next section, we prove theorems to substantiate the %
% partial correctness of procedure call.                      %



let INTERPRET_CALL =
 prove_thm
  (`INTERPRET_CALL`,
   "!r p vars vals glbs pre post calls rec c vals' xs es s1 s4.
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                   r (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ",
   REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o C_REWRITE_RULE)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "p = (p':string)"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "xs = (xs':(var)list)"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "es = (es':(exp)list)"
   THEN UNDISCH_TAC "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN UNDISCH_TAC "(r:env) p = vars',vals'',glbs',pre',post',calls',rec',c'"
   THEN DISCH_THEN (\th'. DISCH_THEN (\th.
              (REWRITE_ALL_THM
                      o ONCE_REWRITE_RULE[EQ_SYM_EQ]
                      o REWRITE_RULE[PAIR_EQ]) 
                   (TRANS (SYM th) th')))
   THEN EXISTS_TAC "s2':state"
   THEN EXISTS_TAC "s3:state"
   THEN EXISTS_TAC "ns:(num)list"
   THEN REPEAT CONJ_TAC THEN ASM_REWRITE_TAC[]
  );;

%<
g "!r p vars vals glbs pre post calls rec c vals' xs es s1 s4.
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                   r (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (STRIP_ASSUME_TAC o C_REWRITE_RULE));;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "p = (p':string)");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "xs = (xs':(var)list)");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "es = (es':(exp)list)");;
e(UNDISCH_TAC "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(UNDISCH_TAC "(r:env) p = vars',vals'',glbs',pre',post',calls',rec',c'");;
e(DISCH_THEN (\th'. DISCH_THEN (\th. (REWRITE_ALL_THM
          o ONCE_REWRITE_RULE[EQ_SYM_EQ] o pthm) 
          (REWRITE_RULE[PAIR_EQ](TRANS (SYM th) th')))));;
e(EXISTS_TAC "s2':state");;
e(EXISTS_TAC "s3:state");;
e(EXISTS_TAC "ns:(num)list");;
e(REPEAT CONJ_TAC THEN ASM_REWRITE_TAC[]);;
>%



let proc_subst_ONLY_IMPACT =
 prove_thm
  (`proc_subst_ONLY_IMPACT`,
    "!r p vars vals glbs pre post calls rec c xs vals' ns s2 s3 z.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      DL(APPEND xs glbs) /\
      C (c <|c (APPEND xs vals' //v APPEND vars vals))
        r (BNDsL ns vals' s2) s3        /\
      ~(z IN SL(APPEND(APPEND xs vals')glbs))
      ==>
      (s3 z = s2 z)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP FVc_proc_subst)
   THEN IMP_RES_TAC (AND2IMP WFc_proc_subst)
   THEN IMP_RES_TAC NOT_IN_SUBSET
   THEN IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT)
   THEN POP_ASSUM REWRITE_THM
   THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
   THEN UNDISCH_TAC "~(z:var) IN (SL(APPEND(APPEND xs vals')glbs))"
   THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN DISCH_THEN REWRITE_THM
  );;

%<
g "!r p vars vals glbs pre post calls rec c xs vals' ns s2 s3 z.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      DL(APPEND xs glbs) /\
      C (c <|c (APPEND xs vals' //v APPEND vars vals))
        r (BNDsL ns vals' s2) s3        /\
      ~(z IN SL(APPEND(APPEND xs vals')glbs))
      ==>
      (s3 z = s2 z)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP FVc_proc_subst));;
e(IMP_RES_TAC (AND2IMP WFc_proc_subst));;
e(IMP_RES_TAC NOT_IN_SUBSET);;
e(IMP_RES_TAC (AND2IMP C_FREE_ONLY_IMPACT));;
e(POP_ASSUM REWRITE_THM);;
e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
e(UNDISCH_TAC "~(z:var) IN (SL(APPEND(APPEND xs vals')glbs))");;
e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
e(DISCH_THEN REWRITE_THM);;
>%




let INTERPRET_SUBSTITUTED_CALL =
 prove_thm
  (`INTERPRET_SUBSTITUTED_CALL`,
   "!r p vars vals glbs pre post calls rec c g xs vals' s1 s4 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns.
         (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
          C ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                <|c ss)
             r
             (BNDsL ns (MAP ss vals') s2)
             s3 /\
          ES es s1 ns s2 ",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN STRIP_ASSUME_TAC (MATCH_MP INTERPRET_CALL
      (LIST_CONJ (map ASSUME
        ["(r:env) p = vars,vals,glbs,pre,post,calls,rec,c";
         "vals' = variants vals(SL(APPEND xs glbs))";
         "C (CALL p xs es) r s1 s4"])))
   THEN EXISTS_TAC "s2:state"
   THEN EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                               (MAP ss (APPEND (APPEND xs vals') glbs)))
                          (MAP ss (APPEND (APPEND xs vals') glbs))
                          (BNDsL ns (MAP ss vals') s2)"
   THEN EXISTS_TAC "ns:(num)list"
   THEN POP_ASSUM (\th. REWRITE_THM th THEN ASSUME_TAC th)
   THEN SUBGOAL_THEN
             "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
           STRIP_ASSUME_TAC
        THENL
         [ USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11)
           THENL
            [ REWRITE_TAC[LENGTH_MAP];

              ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
            ];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       r ss"
           ASSUME_TAC
        THENL
         [ REWRITE_TAC[WFcsubst]
           THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss"))
           THEN ASM_REWRITE_TAC[GVc_subst]
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REPEAT STRIP_TAC
           THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
           THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
           THEN RES_TAC;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                  (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r SUBSET
              (SL (APPEND (APPEND xs vals') glbs))"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "C ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
                 r s1 s2 =
              C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                 r (s1 o ss) (s2 o ss)"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC c_subst_biject
           THEN EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND vars vals)glbs: var list)"
           ASSUME_TAC
        THENL
         [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[DL_APPEND]
           THEN ASM_REWRITE_TAC[DL_variants]
           THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
           THEN REWRITE_TAC[DISJOINT_variants];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)(MAP (ss:vsubst) vals'))"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[(SYM o ASSUME)
                  "MAP ss(APPEND xs glbs) = APPEND xs glbs :var list"]
           THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
           THEN IMP_RES_THEN REWRITE_THM DL_MAP_ONE_ONE
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "MAP (ss:vsubst) (APPEND(APPEND xs vals') glbs) =
                      (APPEND(APPEND xs (MAP ss vals')) glbs)"
           ASSUME_TAC
        THENL
         [ ASM_REWRITE_TAC[MAP_APPEND];

           ALL_TAC
         ]
   THEN CONJ_TAC
   THENL
    [ REWRITE_TAC[ASSUME "(s4:state) = BNDsL(MAP s2 vals')vals' s3"]
      THEN EXT_TAC "z:var"
      THEN GEN_TAC
      THEN ASM_CASES_TAC "(z:var) IN SL vals'"
      THENL
       [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
         THEN ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
            THENL
             [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
             ]
          ];

         IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
         THEN ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
            THEN USE_IMP_THEN REWRITE_THM
                              (SPEC_ALL proc_subst_ONLY_IMPACT)
            THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
            THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3
            THEN REWRITE_TAC[(SYM o SPEC_ALL) SL_APPEND;
                    (SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN ASM_CASES_TAC "z IN SL(APPEND(APPEND xs (MAP(ss:vsubst)vals'))glbs)"
            THENL
             [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
               THEN SUBGOAL_THEN "(\y. @z. (ss:vsubst)z = y)z = z"
                              (\th. REWRITE_TAC[o_THM;th])
               THEN IMP_RES_THEN (ONCE_REWRITE_THM o SYM o SPEC_ALL)
                                 ONE_ONE
               THEN BETA_TAC
               THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN POP_ASSUM (STRIP_ASSUME_TAC o
                           REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM])
               THENL
                [ RES_THEN REWRITE_THM;

                  RES_THEN CONTR_TAC;

                  RES_THEN REWRITE_THM
                ];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN USE_IMP_THEN REWRITE_THM
                         (SPEC_ALL proc_subst_ONLY_IMPACT)
               THEN UNASSUME_THEN (ASSUME_TAC o SYM)
                         "vals' = variants vals(SL(APPEND xs glbs))"
               THEN ASM_REWRITE_TAC[]
               THEN REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ]
       ];

      MATCH_MP_TAC C_EXTENSION
      THEN EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss"
      THEN EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)"
      THEN REPEAT CONJ_TAC
      THENL  % 5 subgoals %
       [ IMP_RES_THEN REWRITE_THM WFc_subst
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM FVc_subst
         THENL
          [ EXISTS_TAC "glbs:var list"
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[GVc_subst];

            REWRITE_TAC[SL_MAP]
            THEN MATCH_MP_TAC IMAGE_SUBSET
            THEN ASM_REWRITE_TAC[]
          ];

         REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]
         THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
         THEN REWRITE_TAC[I_o_ID]
         THEN REWRITE_TAC[MAP_EXTENSION]
         THEN REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)"
         THENL
          [ IMP_RES_TAC IN_DISJOINT_LIST
            THEN SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC
            THENL
             [ REWRITE_TAC[SL_MAP;IN_IMAGE]
               THEN IMP_RES_THEN REWRITE_THM ONE_ONE
               THEN STRIP_TAC
               THEN UNDISCH_LAST_TAC
               THEN POP_ASSUM (REWRITE_THM o SYM)
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN RES_THEN REWRITE_THM
             ];

            ASM_CASES_TAC "(x:var) IN SL vals'"
            THENL
             [ USE_IMP_THEN REWRITE_THM
                    (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE)
               THEN ASM_REWRITE_TAC[]
               THEN IMP_RES_TAC ES_LENGTH_THM
               THEN ASM_REWRITE_TAC[LENGTH_variants];

               REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ];

         USE_IMP_THEN REWRITE_THM c_subst_biject
         THENL
          [ EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
            THEN ASM_REWRITE_TAC[];

            REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN REWRITE_TAC[I_o_ID]
            THEN CHECK_TAC
          ]
       ]
    ]
  );;

%< INTERPRET_SUBSTITUTED_CALL
g "!r p vars vals glbs pre post calls rec c g xs vals' s1 s4 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns.
         (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
          C ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                <|c ss)
             r
             (BNDsL ns (MAP ss vals') s2)
             s3 /\
          ES es s1 ns s2 ";;
e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(STRIP_ASSUME_TAC (MATCH_MP INTERPRET_CALL
      (LIST_CONJ (map ASSUME
        ["(r:env) p = vars,vals,glbs,pre,post,calls,rec,c";
         "vals' = variants vals(SL(APPEND xs glbs))";
         "C (CALL p xs es) r s1 s4"]))));;
e(EXISTS_TAC "s2:state");;
e(EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                         (MAP ss (APPEND (APPEND xs vals') glbs)))
                    (MAP ss (APPEND (APPEND xs vals') glbs))
                    (BNDsL ns (MAP ss vals') s2)");;
e(EXISTS_TAC "ns:(num)list");;
e(POP_ASSUM (\th. REWRITE_THM th THEN ASSUME_TAC th));;
let st1 = get_state();;
set_state st1;;

e(SUBGOAL_THEN
    "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
   STRIP_ASSUME_TAC);;

  e(USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11));;

   e(REWRITE_TAC[LENGTH_MAP]);;

   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;

e(SUBGOAL_THEN
    "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r ss"
   ASSUME_TAC);;

 e(REWRITE_TAC[WFcsubst]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss")));;
 e(ASM_REWRITE_TAC[GVc_subst]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
 e(RES_TAC);;

e(SUBGOAL_THEN
    "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
  e(ASM_REWRITE_TAC[]);;

e(SUBGOAL_THEN
    "FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r SUBSET
     (SL (APPEND (APPEND xs vals') glbs))"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
  e(ASM_REWRITE_TAC[]);;

e(SUBGOAL_THEN
    "C ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
        r s1 s2 =
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
        r (s1 o ss) (s2 o ss)"
   ASSUME_TAC);;

 e(MATCH_MP_TAC c_subst_biject);;
 e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
 e(ASM_REWRITE_TAC[]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND vars vals)glbs: var list)"
      ASSUME_TAC);;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[DL_variants]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)(MAP (ss:vsubst) vals'))"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[(SYM o ASSUME)
            "MAP ss(APPEND xs glbs) = APPEND xs glbs :var list"]);;
   e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;
   e(IMP_RES_THEN REWRITE_THM DL_MAP_ONE_ONE);;
   e(ASM_REWRITE_TAC[]);;

 e(SUBGOAL_THEN "MAP (ss:vsubst) (APPEND(APPEND xs vals') glbs) =
                 (APPEND(APPEND xs (MAP ss vals')) glbs)"
      ASSUME_TAC);;

   e(ASM_REWRITE_TAC[MAP_APPEND]);;

let st2 = get_state();;
set_state st2;;
e(CONJ_TAC);;

 e(REWRITE_TAC[ASSUME "(s4:state) = BNDsL(MAP s2 vals')vals' s3"]);;
  e(EXT_TAC "z:var");;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(z:var) IN SL vals'");;

   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
   e(ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')");;

    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

     e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
     e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;
     e(ASM_REWRITE_TAC[]);;

     e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')");;

    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM (SPEC_ALL proc_subst_ONLY_IMPACT));;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;
    e(REWRITE_TAC[(SYM o SPEC_ALL) SL_APPEND;
             (SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
let st3 = get_state();;
set_state st3;;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(ASM_CASES_TAC "z IN SL(APPEND(APPEND xs (MAP(ss:vsubst)vals'))glbs)");;

     e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
     e(SUBGOAL_THEN "(\y. @z. (ss:vsubst)z = y)z = z"
             (\th. REWRITE_TAC[o_THM;th]));;
     e(IMP_RES_THEN (ONCE_REWRITE_THM o pthm o SYM o SPEC_ALL)
         ONE_ONE);;
     e(BETA_TAC);;
     e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;
     e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
     e(POP_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM]));;

      e(RES_THEN REWRITE_THM);;

      e(RES_THEN CONTR_TAC);;

      e(RES_THEN REWRITE_THM);;

     e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
     e(USE_IMP_THEN REWRITE_THM (SPEC_ALL proc_subst_ONLY_IMPACT));;
     e(UNASSUME_THEN (ASSUME_TAC o SYM)
                "vals' = variants vals(SL(APPEND xs glbs))");;
     e(ASM_REWRITE_TAC[]);;
     e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
     e(ASM_REWRITE_TAC[]);;

let st4 = get_state();;
set_state(st4);;

 e(MATCH_MP_TAC C_EXTENSION);;
 e(EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss");;
 e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)");;
 e(REPEAT CONJ_TAC);;  % 5 subgoals %

  e(IMP_RES_THEN REWRITE_THM WFc_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:var list");;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(ASM_REWRITE_TAC[GVc_subst]);;

   e(REWRITE_TAC[SL_MAP]);;
   e(MATCH_MP_TAC IMAGE_SUBSET);;
   e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[I_o_ID]);;
  e(REWRITE_TAC[MAP_EXTENSION]);;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)");;

   e(IMP_RES_TAC IN_DISJOINT_LIST);;
   e(SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC);;

    e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
    e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
    e(STRIP_TAC);;
    e(UNDISCH_LAST_TAC);;
    e(POP_ASSUM (REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM);;
    e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
    e(RES_THEN REWRITE_THM);;

   e(ASM_CASES_TAC "(x:var) IN SL vals'");;
let st5 = get_state();;
set_state(st5);;

    e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE));;
    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_TAC ES_LENGTH_THM);;
    e(ASM_REWRITE_TAC[LENGTH_variants]);;

    e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM c_subst_biject);;

   e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

   e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
   e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
   e(REWRITE_TAC[I_o_ID]);;
   e(CHECK_TAC);;
>%




let INTERPRET_variants_SUBSTITUTION_CALL =
 prove_thm
  (`INTERPRET_variants_SUBSTITUTION_CALL`,
   "!r p vars vals glbs pre post calls rec c g vals' vals'' q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     )) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
                 C ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       <|c ss)
                   r (BNDsL ns (MAP ss vals') s2) s3 /\
                 ES es s1 ns s2 ",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL INTERPRET_SUBSTITUTED_CALL)
   THEN FIRST_ASSUM (REWRITE_THM o SYM o SYM)
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THENL
    [ MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants];

      MATCH_MP_TAC IDENT_MAP_vsubst
      THEN REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]
      THEN REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN STRIP_ASSUME_TAC
              (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
      THEN CONJ_TAC
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]
    ]
  );;

%< INTERPRET_variants_SUBSTITUTION_CALL

The idea is, use
  ss = ((vals'' & (logicals vals'')) //v (vals' & (logicals vals'))) !

g "!r p vars vals glbs pre post calls rec c g vals' vals'' q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     )) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
                 C ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       <|c ss)
                   r (BNDsL ns (MAP ss vals') s2) s3 /\
                 ES es s1 ns s2 ";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL INTERPRET_SUBSTITUTED_CALL));;
e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;

 e(MATCH_MP_TAC IDENT_MAP_vsubst);;
 e(REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]);;
 e(REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

>%



let INTERPRET_greater_variants_CALL =
 prove_thm
  (`INTERPRET_greater_variants_CALL`,
   "!vals' r p vars vals glbs pre post calls rec c g q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    r (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ",

   REPEAT STRIP_TAC
   THEN DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))"
   THEN DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                              (APPEND vals'' (logicals vals''))    )"
   THEN SUBGOAL_THEN
          "?s2 s3 ns.
              (s4 = BNDsL(MAP s2(MAP ss vals''))(MAP ss vals'')s3) /\
              C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                 r (BNDsL ns (MAP ss vals'') s2) s3 /\
              ES es s1 ns s2"
           ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPECL["vals'':var list";"vals':var list"]
                      (GENL["vals':var list";"vals'':var list"]
                         (SPEC_ALL INTERPRET_variants_SUBSTITUTION_CALL)))
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN SUBGOAL_THEN
          "!s1 s2.
            C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
               r s1 s2 =
            C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r s1 s2"
          REWRITE_THM
   THENL
    [ REPEAT GEN_TAC
      THEN MATCH_MP_TAC (SPEC_ALL C_SUBSTITUTION_EQ)
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                           DISJOINT_variants;DISJOINT_variants_UNION];

      ALL_TAC
    ]
   THEN SUBGOAL_THEN  "MAP (ss:vsubst) vals'' = vals'"
          REWRITE_THM
   THEN POP_ASSUM REWRITE_THM
   THEN USE_IMP_THEN REWRITE_THM
            (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
   THENL
    [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];

      MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
      THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
    ]
  );;

%< INTERPRET_greater_variants_CALL

g "!vals' r p vars vals glbs pre post calls rec c g q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     C (CALL p xs es) r s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    r (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ";;

e(REPEAT STRIP_TAC);;
e(DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))");;
e(DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                        (APPEND vals'' (logicals vals''))    )");;
e(SUBGOAL_THEN
     "?s2 s3 ns.
         (s4 = BNDsL(MAP s2(MAP ss vals''))(MAP ss vals'')s3) /\
         C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
            r (BNDsL ns (MAP ss vals'') s2) s3 /\
         ES es s1 ns s2"
      ASSUME_TAC);;

  e(MATCH_MP_TAC (SPECL["vals'':var list";"vals':var list"]
                    (GENL["vals':var list";"vals'':var list"]
                       (SPEC_ALL INTERPRET_variants_SUBSTITUTION_CALL))));;
  e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(SUBGOAL_THEN
     "!s1 s2.
       C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
          r s1 s2 =
       C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r s1 s2"
     REWRITE_THM);;

   e(REPEAT GEN_TAC);;
   e(MATCH_MP_TAC (SPEC_ALL C_SUBSTITUTION_EQ));;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_UNION]);;

   e(SUBGOAL_THEN
         "MAP (ss:vsubst) vals'' = vals'"
       REWRITE_THM);;
   e(POP_ASSUM REWRITE_THM);;
   e(USE_IMP_THEN REWRITE_THM
          (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

>%





let A_APPEND_var_list_TRANS =
 prove_thm
  (`A_APPEND_var_list_TRANS`,
   "!a x x' y y' y''.
     (LENGTH x = LENGTH x') /\
     DISJOINT(SL x')(SL y) /\ DISJOINT(SL x')(SL y') /\
     (LENGTH y = LENGTH y') /\ (LENGTH y' = LENGTH y'') /\
     DL y' /\ ((FVa a) INTER SL y') SUBSET (SL y) ==>
     (A ((a <|a (APPEND x' y' // APPEND x y)) <|a (y'' // y')) s =
      A  (a <|a (APPEND x' y'' // APPEND x y)) s)",
   REPEAT STRIP_TAC
   THEN ONCE_REWRITE_TAC[a_subst_THM]
   THEN USE_IMP_THEN REWRITE_THM a_subst_APPEND_var_list_COMPOSE
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN REWRITE_TAC[(SYM o SPEC_ALL) a_subst_THM]
   THEN USE_IMP_THEN REWRITE_THM a_subst_APPEND_var_list_COMPOSE
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN USE_IMP_THEN REWRITE_THM A_var_list_TRANS
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_ALL_TAC[SUBSET_DEF;IN_INTER]
   THEN REWRITE_TAC[FVa_a_subst;IN_UNION_SET;IN_IMAGE]
   THEN REPEAT STRIP_TAC
   THEN UNDISCH_TAC "si = (FVv o (x' // x))x'''"
   THEN REWRITE_TAC[o_THM;FVv_var_list_subst]
   THEN DISCH_THEN (\th. REWRITE_ALL_TAC[th;IN])
   THEN UNASSUME_THEN REWRITE_ALL_THM "x'' = BNDsR x' x (I:var->var) x'''"
   THEN ASM_CASES_TAC "(x''':var) IN SL x"
   THENL
     [  IMP_RES_TAC BNDsR_MEMBER
        THEN POP_ASSUM (ASSUME_TAC o SPEC "I:var->var")
        THEN UNDISCH_TAC "DISJOINT(SL (x':(var)list))(SL (y':(var)list))"
        THEN REWRITE_TAC[IN_DISJOINT]
        THEN CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV)
        THEN REWRITE_TAC[DE_MORGAN_THM]
        THEN DISCH_THEN (STRIP_ASSUME_TAC o SPEC "BNDsR x' x (I:var->var) x'''")
        THEN RES_TAC;

        IMP_RES_THEN REWRITE_ALL_THM BNDsR_NOT_MATCH
        THEN REWRITE_ALL_TAC[I_THM]
        THEN RES_TAC
     ]
  );;

% A_APPEND_var_list_TRANS
g "!a x x' y y' y''.
     (LENGTH x = LENGTH x') /\
     DISJOINT(SL x')(SL y) /\ DISJOINT(SL x')(SL y') /\
     (LENGTH y = LENGTH y') /\ (LENGTH y' = LENGTH y'') /\
     DL y' /\ ((FVa a) INTER SL y') SUBSET (SL y) ==>
     (A ((a <|a (APPEND x' y' // APPEND x y)) <|a (y'' // y')) s =
      A  (a <|a (APPEND x' y'' // APPEND x y)) s)";;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[a_subst_THM]);;
e(USE_IMP_THEN REWRITE_THM a_subst_APPEND_var_list_COMPOSE);;

  e(ASM_REWRITE_TAC[]);;

e(REWRITE_TAC[(SYM o SPEC_ALL) a_subst_THM]);;
e(USE_IMP_THEN REWRITE_THM a_subst_APPEND_var_list_COMPOSE);;

  e(ASM_REWRITE_TAC[]);;

e(USE_IMP_THEN REWRITE_THM A_var_list_TRANS);;
e(ASM_REWRITE_TAC[]);;
e(REWRITE_ALL_TAC[SUBSET_DEF;IN_INTER]);;
e(REWRITE_TAC[FVa_a_subst;IN_UNION_SET;IN_IMAGE]);;
e(REPEAT STRIP_TAC);;
e(UNDISCH_TAC "si = (FVv o (x' // x))x'''");;
e(REWRITE_TAC[o_THM;FVv_var_list_subst]);;
e(DISCH_THEN (\th. REWRITE_ALL_TAC[th;IN]));;
e(UNASSUME_THEN REWRITE_ALL_THM "x'' = BNDsR x' x (I:var->var) x'''");;
e(ASM_CASES_TAC "(x''':var) IN SL x");;

  e(IMP_RES_TAC BNDsR_MEMBER);;
  e(POP_ASSUM (ASSUME_TAC o SPEC "I:var->var"));;
  e(UNDISCH_TAC "DISJOINT(SL (x':(var)list))(SL (y':(var)list))");;
  e(REWRITE_TAC[IN_DISJOINT]);;
  e(CONV_TAC (DEPTH_CONV NOT_EXISTS_CONV));;
  e(REWRITE_TAC[DE_MORGAN_THM]);;
  e(DISCH_THEN (STRIP_ASSUME_TAC o SPEC "BNDsR x' x (I:var->var) x'''"));;

    e(RES_TAC);;

    e(RES_TAC);;

  e(IMP_RES_THEN REWRITE_ALL_THM BNDsR_NOT_MATCH);;
  e(REWRITE_ALL_TAC[I_THM]);;
  e(RES_TAC);;
%





let CHANGE_VAR_FOR_ADAPTATION_DISJOINT_PARTIAL =
 TAC_PROOF
  (([],
    "!xs vals' r p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenvp r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      DISJOINT(SL x0)(FVa q) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x1  = APPEND x x0) /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      DISJOINT(SL x0)(FVa q) /\
      PCSPEC(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
               post <|a (x1 // y1),r)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
            (WFc c' g' r /\
             DL x' /\
             WFxs x' /\
             (GVc c' r) SUBSET (SL glbs) /\
             (FVc c' r) SUBSET (SL x') /\
             (FVa pre') SUBSET (SL x') /\
             (FVa post') SUBSET (SL(APPEND x' x0')) /\
             PCSPEC(AAND(mk_eqs x0' x')pre',c',post',r))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV))
   THENL
    [ MATCH_MP_TAC (SPEC_ALL PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL)
      THEN ASM_REWRITE_TAC[];

      ASM_REWRITE_TAC[]
    ]
  );;


let CHANGE_VAR_FOR_ADAPTATION_DISJOINT =
 TAC_PROOF
  (([],
    "!xs vals' r p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenv r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      DISJOINT(SL x0)(FVa q) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x1  = APPEND x x0) /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      DISJOINT(SL x0)(FVa q) /\
      TCSPEC(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),r)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN IMP_RES_TAC (AND2IMP (CONV_RULE (DEPTH_CONV let_CONV)
                                 PROCEDURE_CHANGE_OF_VARIABLES))
   THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[CSPEC])
   THEN ASM_REWRITE_TAC[]
  );;

%<
g "!xs vals' r p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenv r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      DISJOINT(SL x0)(FVa q) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x1  = APPEND x x0) /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      DISJOINT(SL x0)(FVa q) /\
      CSPEC(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
              post <|a (x1 // y1),r)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
            (WFc c' g' r /\
             DL x' /\
             WFxs x' /\
             (GVc c' r) SUBSET (SL glbs) /\
             (FVc c' r) SUBSET (SL x') /\
             (FVa pre') SUBSET (SL x') /\
             (FVa post') SUBSET (SL(APPEND x' x0')) /\
             CSPEC(AAND(mk_eqs x0' x')pre',c',post',r))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV)));;

    e(MATCH_MP_TAC (SPEC_ALL PROCEDURE_CHANGE_OF_VARIABLES));;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;
>%


let CHANGE_VAR_FOR_ADAPTATION_PARTIAL =
 TAC_PROOF
  (([],
    "!xs vals' r p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenvp r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      PCSPEC(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
               post <|a (x1 // y1),r)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
            (WFc c' g' r /\
             DL x' /\
             WFxs x' /\
             (GVc c' r) SUBSET (SL glbs) /\
             (FVc c' r) SUBSET (SL x') /\
             (FVa pre') SUBSET (SL x') /\
             (FVa post') SUBSET (SL(APPEND x' x0')) /\
             PCSPEC(AAND(mk_eqs x0' x')pre',c',post',r))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV))
   THENL
    [ MATCH_MP_TAC (SPEC_ALL PROCEDURE_CHANGE_OF_VARIABLES_PARTIAL)
      THEN ASM_REWRITE_TAC[];

      ASM_REWRITE_TAC[]
    ]
  );;


let CHANGE_VAR_FOR_ADAPTATION =
 TAC_PROOF
  (([],
    "!xs vals' r p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenv r) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      TCSPEC(AAND(mk_eqs x0 x)(pre <|a (x // y)), c <|c (x1 //v y1), r)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN IMP_RES_TAC (AND2IMP (CONV_RULE (DEPTH_CONV let_CONV)
                                  PROCEDURE_CHANGE_OF_VARIABLES))
   THEN REWRITE_ALL_TAC[CSPEC]
   THEN ASM_REWRITE_TAC[]
  );;


let WFenv_IMP_DL_DISJOINT =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals(SL(APPEND xs glbs))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs (variants vals(SL(APPEND xs glbs)))) /\
      DISJOINT(SL(APPEND xs (variants vals(SL(APPEND xs glbs)))))(SL glbs)"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_variants_APPEND]
  );;

%< WFenv_IMP_DL_DISJOINT
g "!r p vars vals glbs pre post calls rec c vals' xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals(SL(APPEND xs glbs))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs (variants vals(SL(APPEND xs glbs)))) /\
      DISJOINT(SL(APPEND xs (variants vals(SL(APPEND xs glbs)))))(SL glbs)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]);;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_variants_APPEND]);;
>%


let WFenv_IMP_MAIN_DL_DISJOINT =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' q xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs vals') /\
      DISJOINT(SL(APPEND xs vals'))(SL glbs)"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN ASM_REWRITE_TAC[DL_variants;SL_APPEND;DISJOINT_variants_UNION3]
  );;

%< WFenv_IMP_MAIN_DL_DISJOINT
g "!r p vars vals glbs pre post calls rec c vals' q xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs vals') /\
      DISJOINT(SL(APPEND xs vals'))(SL glbs)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
e(ASM_REWRITE_TAC[DL_variants;SL_APPEND;DISJOINT_variants_UNION3]);;
>%



let pre_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!vars vals xs vals' glbs u v x y.
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals' = LENGTH vals) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs)
     ==>
      ((x // y) = (u // v))"),
   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC var_list_IDENT_APPEND
   THEN ASM_REWRITE_TAC[LENGTH_APPEND]
  );;

%< pre_subst_simplify_LEMMA
g "!vars vals xs vals' glbs u v x y.
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals' = LENGTH vals) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs)
     ==>
      ((x // y) = (u // v))";;

e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(MATCH_MP_TAC var_list_IDENT_APPEND);;
e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
>%


let A_post_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A ((post <|a (x1 // y1)) <|a (x0' // x0)) s =
       A (post <|a (APPEND u x0' // APPEND v y0)) s)"),

   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "(APPEND x x0' // APPEND y y0) = (APPEND u x0' // APPEND v y0)"
               (REWRITE_THM o SYM)
   THENL
    [ USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]
         THEN MATCH_MP_TAC WFxs_NOT_WFxs
         THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
         THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

         ALL_TAC
       ]
      THEN REWRITE_TAC[ASSUME "x = APPEND u glbs :var list";
                       ASSUME "y = APPEND v glbs :var list"]
      THEN REWRITE_TAC[APPEND_ASSOC]
      THEN USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals];
         ALL_TAC
       ]
      THEN USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE
      THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      ALL_TAC
    ]
   THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :var list";
                    ASSUME "y1 = APPEND y y0 :var list"]
   THEN MATCH_MP_TAC A_APPEND_var_list_TRANS
   THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals;DL_logicals]
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REPEAT CONJ_TAC
   THENL  % 4 subgoals %

    [ MATCH_MP_TAC WFxs_NOT_WFxs
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      MATCH_MP_TAC WFxs_NOT_WFxs
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      (ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
               "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))"
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[DL_variants];

      REWRITE_TAC[SUBSET_DEF;IN_INTER]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC logicals_IS_NOT_WFx
      THEN IMP_RES2_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[IN_UNION] o
                  ONCE_REWRITE_RULE[SL_APPEND]) SUBSET_DEF
      THEN IMP_RES_TAC WFxs_IS_WFx
      THEN RES_TAC
    ]
  );;

%< A_post_subst_simplify_LEMMA
g "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A ((post <|a (x1 // y1)) <|a (x0' // x0)) s =
       A (post <|a (APPEND u x0' // APPEND v y0)) s)";;

e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN "(APPEND x x0' // APPEND y y0) = (APPEND u x0' // APPEND v y0)"
       (REWRITE_THM o SYM));;

 e(USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;
  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

  e(REWRITE_TAC[ASSUME "x = APPEND u glbs :var list";
                ASSUME "y = APPEND v glbs :var list"]);;
  e(REWRITE_TAC[APPEND_ASSOC]);;
  e(USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND);;

   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;

   e(USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE);;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(MATCH_MP_TAC WFxs_NOT_WFxs);;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :var list";
              ASSUME "y1 = APPEND y y0 :var list"]);;
e(MATCH_MP_TAC A_APPEND_var_list_TRANS);;
e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals;DL_logicals]);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REPEAT CONJ_TAC);;  % 4 subgoals %

 e(MATCH_MP_TAC WFxs_NOT_WFxs);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

 e(MATCH_MP_TAC WFxs_NOT_WFxs);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

 e((ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
            "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))");;
 e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[DL_variants]);;

 e(REWRITE_TAC[SUBSET_DEF;IN_INTER]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC logicals_IS_NOT_WFx);;
 e(IMP_RES2_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[IN_UNION] o
                  ONCE_REWRITE_RULE[SL_APPEND]) SUBSET_DEF);;
 e(IMP_RES_TAC WFxs_IS_WFx);;
 e(RES_TAC);;
>%





let post_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                    <|a (x // x0')))
          s =
       A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                    <|a (x // x0')))
          s)"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[A_DEF]
   THEN MATCH_MP_TAC CONJ_11
   THEN REWRITE_TAC[a_subst_THM]
   THEN REWRITE_TAC[A_mk_foralls]
   THEN REWRITE_TAC[A_DEF]
   THEN REWRITE_TAC[ASSUME "(x1:var list) = APPEND x x0";
                    ASSUME "(y1:var list) = APPEND y y0"]
   THEN FORALL_EQ_TAC
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL A_post_subst_simplify_LEMMA)
   THEN ASM_REWRITE_TAC[]
  );;

%< post_subst_simplify_LEMMA
g "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                    <|a (x // x0')))
          s =
       A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                    <|a (x // x0')))
          s)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[A_DEF]);;
e(MATCH_MP_TAC CONJ_11);;
e(REWRITE_TAC[a_subst_THM]);;
e(REWRITE_TAC[A_mk_foralls]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[ASSUME "(x1:var list) = APPEND x x0";
              ASSUME "(y1:var list) = APPEND y y0"]);;
e(FORALL_EQ_TAC);;
e(USE_IMP_THEN REWRITE_THM (SPEC_ALL A_post_subst_simplify_LEMMA));;
e(ASM_REWRITE_TAC[]);;

. |- A
     ((post <|a ((APPEND x x0) // (APPEND y y0))) <|a (x0' // x0))
     (BNDsL n x(s <|s (x // x0'))) =
     A
     (post <|a ((APPEND u x0') // (APPEND v y0)))
     (BNDsL n x(s <|s (x // x0')))

>%



let command_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' xs u v x y x0 y0 s1 s2.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      (DL (APPEND xs glbs)) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT(SL vals')(SL(APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)
   ==>
      (C (c <|c (APPEND x x0 //v APPEND y y0)) r s1 s2 =
       C (c <|c (APPEND xs vals' //v APPEND vars vals)) r s1 s2)"),

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC c_subst_EQ
   THEN REPEAT STRIP_TAC
   THEN SUBGOAL_THEN
     "WFc c calls r /\
      DL(APPEND (APPEND vars vals) glbs) /\
      DL(APPEND vars vals) /\
      DL(APPEND (APPEND xs vals') glbs) /\
      DL(APPEND xs vals') /\
      (FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs)))"
           STRIP_ASSUME_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[]
      THEN (ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
                 "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))"
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[];

      IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
      THEN USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND];

         ASM_REWRITE_TAC[]
         THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION]
                                          o ONCE_REWRITE_RULE[SL_APPEND])
         THENL   % 2 subgoals %
          [ USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
            THEN ASM_REWRITE_TAC[LENGTH_APPEND];

            USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
            THENL
             [ ASM_REWRITE_TAC[LENGTH_APPEND]
               THEN ONCE_REWRITE_ALL_TAC[DL_APPEND]
               THEN ASM_REWRITE_TAC[];

               REWRITE_TAC[IDENTITY_vsubst;I_THM]
               THEN USE_IMP_THEN REWRITE_THM IDENT_vsubst
               THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST
             ]
          ]
       ]
    ]
  );;


%< 
g "!r p vars vals glbs pre post calls rec c vals' xs u v x y x0 y0 s1 s2.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      (DL (APPEND xs glbs)) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT(SL vals')(SL(APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)
   ==>
      (C (c <|c (APPEND x x0 //v APPEND y y0)) r s1 s2 =
       C (c <|c (APPEND xs vals' //v APPEND vars vals)) r s1 s2)";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC c_subst_EQ);;
e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN
     "WFc c calls r /\
      DL(APPEND (APPEND vars vals) glbs) /\
      DL(APPEND vars vals) /\
      DL(APPEND (APPEND xs vals') glbs) /\
      DL(APPEND xs vals') /\
      (FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs)))"
   STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e((ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
            "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))");;
 e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

  e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

  e(ASM_REWRITE_TAC[]);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION]
                                o ONCE_REWRITE_RULE[SL_APPEND]));;
   % 2 subgoals %

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;
   e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;

    e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
    e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
    e(ASM_REWRITE_TAC[]);;

    e(REWRITE_TAC[IDENTITY_vsubst;I_THM]);;
    e(USE_IMP_THEN REWRITE_THM IDENT_vsubst);;
    e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST);;
>%



% ===================================================== %
% ===================================================== %
%                  PROCEDURE_CALL_RULE                  %
%                                                       %
%                 (partial correctness)                 %
% ===================================================== %
% ===================================================== %


let CALLp_DISJOINT_DISJOINT_THM =
 prove_thm
  (`CALLp_DISJOINT_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenvp r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      PCSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, q, r) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs"
   THEN DEFINE_NEW_VAR "(y0:(var)list) = logicals y"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN "PCSPEC
                       (AAND (pre <|a (x // y))
                            ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                     <|a (x // x0)),
                       (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATIONp_DISJOINT))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT_PARTIAL)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM
             THENL
              [ ASM_REWRITE_TAC[]
                THEN MATCH_MP_TAC var_list_logicals_IDENT_APPEND
                THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_PARTIAL
                THEN REWRITE_ALL_TAC[WFxs_APPEND]
                THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants];

                ALL_TAC
              ]
        THEN REWRITE_TAC[PCSPEC]
        THEN STRIP_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_TAC (AND2IMP INTERPRET_CALL)
        THEN ASM_REWRITE_TAC[]
        THEN REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
        THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
        THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
        THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                         ASSUME "y1 = APPEND y y0 :(var)list"]
        THEN USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA)
        THENL
             [  IMP_RES_TAC WELL_FORMED_PROCEDURE_PARTIAL
                THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
                THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                                    DISJOINT_variants;DISJOINT_variants_APPEND];

                ASM_REWRITE_TAC[]
             ]
     ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       CSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, q, r) )";;


e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs");;
e(DEFINE_NEW_VAR "(y0:(var)list) = logicals y");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN "CSPEC(AAND (pre <|a (x // y))
                           ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                   <|a (x // x0)),
                  (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION_DISJOINT)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM);;

  e(ASM_REWRITE_TAC[]);;
  e(MATCH_MP_TAC var_list_logicals_IDENT_APPEND);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants]);;

 e(REWRITE_TAC[CSPEC_THM]);;
let st2 = get_state();;
set_state st2;;
 e(STRIP_TAC);;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;

  e(IMP_RES_TAC (AND2IMP INTERPRET_CALL));;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
   e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(ASM_REWRITE_TAC[]);;

  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
   e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(DISCH_TAC);;
   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%


let CALLp_DISJOINT_THM =
 prove_thm
  (`CALLp_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvp r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       PCSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
           "PCSPEC(AAND (pre <|a (x // y))
                        ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                  <|a (x0' // x0))
                                             q))
                                <|a (x // x0')),
                   (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                          SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATIONp))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_PARTIAL)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN REWRITE_TAC[PCSPEC]
        THEN STRIP_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_TAC (AND2IMP INTERPRET_CALL)
        THEN REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]
        THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
        THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
        THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_PARTIAL
        THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA)
        THENL
            [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                           DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND];

              REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                          ASSUME "y1 = APPEND y y0 :(var)list"]
              THEN USE_IMP_THEN ONCE_REWRITE_THM
                        (SPEC_ALL command_subst_simplify_LEMMA)
              THENL
               [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                        DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND];

                 ASM_REWRITE_TAC[]
               ]
            ]
     ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
  "CSPEC(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                               SYM (ASSUME "x0' = variants x0 (FVa q)")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[CSPEC_THM]);;
 e(STRIP_TAC);;

let st2 = get_state();;
set_state st2;;
  e(ONCE_REWRITE_TAC[a_subst_THM]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC (AND2IMP INTERPRET_CALL));;
  e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

   e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                 ASSUME "y1 = APPEND y y0 :(var)list"]);;
   e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

    e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                      DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

    e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_THM assign_list_THM);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL
           (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA)));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

  e(DISCH_TAC);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                     DL_APPEND;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(DISCH_TAC);;
   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%






%================================================================%
%   We wish to prove the general case of procedure call.  This   %
%   involves showing that the partial correctness of the body    %
%   which is known from the well-formedness of the environment   %
%   is maintained under replacement of the formal parameters.    %
%================================================================%



let CALLp_THM =
 prove_thm
  (`CALLp_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvp r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
       PCSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
               "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
             "PCSPEC(AAND (pre <|a (x // y))
                          ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                    <|a (x0' // x0)) q))
                                  <|a (x // x0')),
                     (c <|c (x1 //v y1)),q,r)"
             ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                                 SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                            RULE_OF_ADAPTATIONp))))
      THEN EXISTS_TAC "calls <|g (x1 //v y1)"
      THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_PARTIAL)
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
      THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
        THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
   THEN REWRITE_TAC[PCSPEC]
   THEN STRIP_TAC
   THEN ONCE_REWRITE_TAC[a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN SUBGOAL_THEN ((snd o dest_imp o concl o SPEC "s2:state" o GEN "s4:state"
                         o SPEC_ALL o SPEC "vals':var list")
                    INTERPRET_greater_variants_CALL)
           STRIP_ASSUME_TAC
   THENL
       [ MATCH_MP_TAC (SPEC_ALL INTERPRET_greater_variants_CALL)
         THEN ASM_REWRITE_TAC[];

         ALL_TAC
       ]
   THEN REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]
   THEN SUBGOAL_THEN "DISJOINT (SL vals') (FVa q)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[DISJOINT_variants_UNION]; ALL_TAC ]
   THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
   THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_PARTIAL
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA)
   THENL
       [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                         DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

         REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                     ASSUME "y1 = APPEND y y0 :(var)list"]
         THEN USE_IMP_THEN ONCE_REWRITE_THM
                   (SPEC_ALL command_subst_simplify_LEMMA)
         THENL
          [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                            DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

            ASM_REWRITE_TAC[]
          ]
       ]
  );;


%< CALL_THM
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
  "CSPEC(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                               SYM (ASSUME "x0' = variants x0 (FVa q)")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[CSPEC_THM]);;
 e(STRIP_TAC);;
let st2 = get_state();;
set_state st2;;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;

  e(SUBGOAL_THEN ((snd o dest_imp o concl o SPEC "s2:state" o GEN "s4:state"
                       o SPEC_ALL o SPEC "vals':var list")
                  INTERPRET_greater_variants_CALL)
       STRIP_ASSUME_TAC);;

   e(MATCH_MP_TAC (SPEC_ALL INTERPRET_greater_variants_CALL));;
   e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
  e(SUBGOAL_THEN "DISJOINT (SL vals') (FVa q)" ASSUME_TAC);;

   e(ASM_REWRITE_TAC[DISJOINT_variants_UNION]);;

  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

   e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                 ASSUME "y1 = APPEND y y0 :(var)list"]);;
   e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

    e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                      DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

    e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(UNDISCH_LAST_TAC);;
  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_THM assign_list_THM);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL
           (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA)));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

  e(DISCH_TAC);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

let st2 = get_state();;
set_state st2;;
   e(DISCH_TAC);;
   e(DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))");;
   e(DEFINE_NEW_VAR "ss = (APPEND vals' (logicals vals') //v
                           APPEND vals'' (logicals vals''))");;
   e(SUBGOAL_THEN
        "?s3'. C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
                  r (BNDsL(VS(VES es)s1) vals'' (s1 <|s (VES_state es))) s3'"
        STRIP_ASSUME_TAC);;

    e(MATCH_MP_TAC (GENL ["s2:state";"s3:state";"ns:num list"]
                      (SPEC_ALL (SPECL["vals'':var list";"vals':var list"]
                                  C_greater_variants_BODY))));;
    e(EXISTS_TAC "s2:state");;
    e(ASM_REWRITE_TAC[]);;

   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals'') vals'' s3'");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%




let C_SUBSTITUTED_BODY =
 prove_thm
  (`C_SUBSTITUTED_BODY`,
   "!vals' r p vars vals glbs pre post calls rec c g xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
       r (BNDsL ns vals' s2) s3 /\
     ES es s1 ns s2
     ==>
     (?s3'. C ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
            r (BNDsL ns (MAP ss vals') s2) s3')",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                               (MAP ss (APPEND (APPEND xs vals') glbs)))
                          (MAP ss (APPEND (APPEND xs vals') glbs))
                          (BNDsL ns (MAP ss vals') s2)"
   THEN SUBGOAL_THEN
             "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
           STRIP_ASSUME_TAC
        THENL
         [ USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11)
           THENL
            [ REWRITE_TAC[LENGTH_MAP];

              ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
            ];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       r ss"
           ASSUME_TAC
        THENL
         [ REWRITE_TAC[WFcsubst]
           THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss"))
           THEN ASM_REWRITE_TAC[GVc_subst]
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REPEAT STRIP_TAC
           THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
           THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
           THEN RES_TAC;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                  (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL WFc_vsubst)
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REWRITE_ALL_TAC[WFxs_APPEND]
           THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                                DL_variants;DISJOINT_variants_UNION];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[DL_APPEND]
           THEN ASM_REWRITE_TAC[DL_variants]
           THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
           THEN REWRITE_TAC[DISJOINT_variants_UNION];

           ALL_TAC
         ]
   THEN MATCH_MP_TAC C_EXTENSION
      THEN EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss"
      THEN EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)"
      THEN REPEAT CONJ_TAC
      THENL  % 5 subgoals %
       [ IMP_RES_THEN REWRITE_THM WFc_subst
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM FVc_subst
         THENL
          [ EXISTS_TAC "glbs:var list"
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[GVc_subst];

            REWRITE_TAC[SL_MAP]
            THEN MATCH_MP_TAC IMAGE_SUBSET
            THEN MATCH_MP_TAC (SPEC_ALL FVc_vsubst)
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN REWRITE_ALL_TAC[WFxs_APPEND]
            THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                                 DL_variants;DISJOINT_variants_UNION]
          ];

         REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]
         THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
         THEN REWRITE_TAC[I_o_ID]
         THEN REWRITE_TAC[MAP_EXTENSION]
         THEN REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)"
         THENL
          [ IMP_RES_TAC IN_DISJOINT_LIST
            THEN SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC
            THENL
             [ REWRITE_TAC[SL_MAP;IN_IMAGE]
               THEN IMP_RES_THEN REWRITE_THM ONE_ONE
               THEN STRIP_TAC
               THEN UNDISCH_LAST_TAC
               THEN POP_ASSUM (REWRITE_THM o SYM)
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN RES_THEN REWRITE_THM
             ];

            ASM_CASES_TAC "(x:var) IN SL vals'"
            THENL
             [ USE_IMP_THEN REWRITE_THM
                    (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE)
               THEN IMP_RES_TAC ES_LENGTH_THM
               THEN ASM_REWRITE_TAC[LENGTH_variants];

               REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ];

         USE_IMP_THEN REWRITE_THM c_subst_biject
         THENL
          [ EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
            THEN ASM_REWRITE_TAC[];

            REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN REWRITE_TAC[I_o_ID]
            THEN CHECK_TAC
          ]
       ]
  );;

%< C_SUBSTITUTED_BODY
g "!vals' r p vars vals glbs pre post calls rec c g xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
       r (BNDsL ns vals' s2) s3 /\
     ES es s1 ns s2
     ==>
     (?s3'. C ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
            r (BNDsL ns (MAP ss vals') s2) s3')";;

e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                         (MAP ss (APPEND (APPEND xs vals') glbs)))
                    (MAP ss (APPEND (APPEND xs vals') glbs))
                    (BNDsL ns (MAP ss vals') s2)");;
let st1 = get_state();;
set_state st1;;

e(SUBGOAL_THEN
    "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
   STRIP_ASSUME_TAC);;

  e(USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11));;

   e(REWRITE_TAC[LENGTH_MAP]);;

   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;

e(SUBGOAL_THEN
    "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r ss"
   ASSUME_TAC);;

 e(REWRITE_TAC[WFcsubst]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss")));;
 e(ASM_REWRITE_TAC[GVc_subst]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
 e(RES_TAC);;

let st1 = get_state();;
set_state st1;;
e(SUBGOAL_THEN
    "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_vsubst));;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                    DL_variants;DISJOINT_variants_UNION]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[DL_variants]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants_UNION]);;

 e(MATCH_MP_TAC C_EXTENSION);;
 e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)");;
 e(REPEAT CONJ_TAC);;  % 5 subgoals %

  e(IMP_RES_THEN REWRITE_THM WFc_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:var list");;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(ASM_REWRITE_TAC[GVc_subst]);;

   e(REWRITE_TAC[SL_MAP]);;
   e(MATCH_MP_TAC IMAGE_SUBSET);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_vsubst));;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                     DL_variants;DISJOINT_variants_UNION]);;

  e(REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[I_o_ID]);;
  e(REWRITE_TAC[MAP_EXTENSION]);;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)");;

   e(IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST);;
   e(SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC);;

    e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
    e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
    e(STRIP_TAC);;
    e(UNDISCH_LAST_TAC);;
    e(POP_ASSUM (REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM);;
    e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
    e(RES_THEN REWRITE_THM);;

   e(ASM_CASES_TAC "(x:var) IN SL vals'");;
let st5 = get_state();;
set_state(st5);;

    e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE));;
    e(IMP_RES_TAC ES_LENGTH_THM);;
    e(ASM_REWRITE_TAC[LENGTH_variants]);;

    e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM c_subst_biject);;

   e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

   e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
   e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
   e(REWRITE_TAC[I_o_ID]);;
   e(CHECK_TAC);;
>%




let C_variants_SUBSTITUTION_BODY =
 prove_thm
  (`C_variants_SUBSTITUTION_BODY`,
   "!vals' vals'' r p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'  (logicals vals' ) //v
            APPEND vals'' (logicals vals'')     )) /\
     C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
        r (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
              r (BNDsL ns vals' s2) s3' ",

   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "MAP (ss:vsubst) vals'' = vals'" (REWRITE_THM o SYM)
   THENL
    [ FIRST_ASSUM (REWRITE_THM o SYM o SYM)
      THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
      THENL
       [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];

         USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
       ];

      ALL_TAC
    ]
   THEN MATCH_MP_TAC (SPEC_ALL C_SUBSTITUTED_BODY)
   THEN FIRST_ASSUM (REWRITE_THM o SYM o SYM)
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THENL
    [ MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants];

      MATCH_MP_TAC IDENT_MAP_vsubst
      THEN REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]
      THEN REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN STRIP_ASSUME_TAC
              (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
      THEN CONJ_TAC
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]
    ]
  );;

%< C_variants_SUBSTITUTION_BODY

The idea is, use
  ss = ((vals' & (logicals vals')) //v (vals'' & (logicals vals''))) !

g "!vals' vals'' r p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'  (logicals vals' ) //v
            APPEND vals'' (logicals vals'')     )) /\
     C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
        r (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
              r (BNDsL ns vals' s2) s3' ";;

e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN "MAP (ss:vsubst) vals'' = vals'" (REWRITE_THM o SYM));;

  e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

e(MATCH_MP_TAC (SPEC_ALL C_SUBSTITUTED_BODY));;
e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;

 e(MATCH_MP_TAC IDENT_MAP_vsubst);;
 e(REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]);;
 e(REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

>%



let C_greater_variants_BODY =
 prove_thm
  (`C_greater_variants_BODY`,
   "!vals' vals'' r p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals'  = variants vals (SL (APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
        r (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
              r (BNDsL ns vals' s2) s3'",

   REPEAT STRIP_TAC
   THEN DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                              (APPEND vals'' (logicals vals''))    )"
   THEN SUBGOAL_THEN
          "?s3'.
              C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                 r (BNDsL ns vals' s2) s3'"
           ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPEC_ALL C_variants_SUBSTITUTION_BODY)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN SUBGOAL_THEN
          "!s1 s2.
            C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
               r s1 s2 =
            C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r s1 s2"
          REWRITE_THM
   THEN REPEAT GEN_TAC
   THEN MATCH_MP_TAC (GENL["vals':var list";"vals'':var list"]
                         (SPEC_ALL C_SUBSTITUTION_EQ))
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND]
   THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                        DISJOINT_variants;DISJOINT_variants_UNION]
  );;

%< C_greater_variants_BODY

g "!vals' vals'' r p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals'  = variants vals (SL (APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
        r (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
              r (BNDsL ns vals' s2) s3'";;

e(REPEAT STRIP_TAC);;
e(DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                        (APPEND vals'' (logicals vals''))    )");;
e(SUBGOAL_THEN
     "?s3'.
         C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
            r (BNDsL ns vals' s2) s3'"
      ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL C_variants_SUBSTITUTION_BODY));;
  e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(SUBGOAL_THEN
     "!s1 s2.
       C ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
          r s1 s2 =
       C (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r s1 s2"
     REWRITE_THM);;
  e(REPEAT GEN_TAC);;
  e(MATCH_MP_TAC (GENL["vals':var list";"vals'':var list"]
                    (SPEC_ALL C_SUBSTITUTION_EQ)));;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                    DISJOINT_variants;DISJOINT_variants_UNION]);;

>%



% ===================================================== %
% ===================================================== %
%                  PROCEDURE_CALL_RULE                  %
%                                                       %
%                  (total correctness)                  %
% ===================================================== %
% ===================================================== %


let CALLt_DISJOINT_DISJOINT_THM =
 prove_thm
  (`CALLt_DISJOINT_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      TCSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, r) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs"
   THEN DEFINE_NEW_VAR "(y0:(var)list) = logicals y"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN "TCSPEC
                       (AAND (pre <|a (x // y))
                             ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                     <|a (x // x0)),
                       (c <|c (x1 //v y1)),r)"
           ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATIONt_DISJOINT))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM
             THENL
              [ ASM_REWRITE_TAC[]
                THEN MATCH_MP_TAC var_list_logicals_IDENT_APPEND
                THEN IMP_RES_TAC WELL_FORMED_PROCEDURE
                THEN REWRITE_ALL_TAC[WFxs_APPEND]
                THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants];

                ALL_TAC
              ]
        THEN REWRITE_TAC[TCSPEC]
        THEN STRIP_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS)
        THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
        THEN RES_TAC
        THEN UNDISCH_LAST_TAC
        THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                         ASSUME "y1 = APPEND y y0 :(var)list"]
        THEN USE_IMP_THEN ONCE_REWRITE_THM
                     (SPEC_ALL command_subst_simplify_LEMMA)
        THENL
            [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
              THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
              THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                                   DISJOINT_variants;DISJOINT_variants_APPEND];

              DISCH_TAC
              THEN EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2"
              THEN CALL_TAC
              THEN EXISTS_TAC "c:cmd"
              THEN EXISTS_TAC "vars:var list"
              THEN EXISTS_TAC "vals:var list"
              THEN EXISTS_TAC "VS (VES es) s1"
              THEN EXISTS_TAC "glbs:var list"
              THEN EXISTS_TAC "pre:aexp"
              THEN EXISTS_TAC "post:aexp"
              THEN EXISTS_TAC "calls:prog_env"
              THEN EXISTS_TAC "rec:aexp"
              THEN ASM_REWRITE_TAC[]
            ]
     ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       CSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, q, r) )";;


e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs");;
e(DEFINE_NEW_VAR "(y0:(var)list) = logicals y");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN "CSPEC(AAND (pre <|a (x // y))
                           ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                   <|a (x // x0)),
                  (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION_DISJOINT)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM);;

  e(ASM_REWRITE_TAC[]);;
  e(MATCH_MP_TAC var_list_logicals_IDENT_APPEND);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants]);;

 e(REWRITE_TAC[CSPEC_THM]);;
let st2 = get_state();;
set_state st2;;
 e(STRIP_TAC);;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;

  e(IMP_RES_TAC (AND2IMP INTERPRET_CALL));;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
   e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(ASM_REWRITE_TAC[]);;

  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
   e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(DISCH_TAC);;
   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%


let CALLt_DISJOINT_THM =
 prove_thm
  (`CALLt_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       TCSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, r) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
           "TCSPEC(AAND (pre <|a (x // y))
                        ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                  <|a (x0' // x0))
                                             q))
                                <|a (x // x0')),
                   (c <|c (x1 //v y1)),r)"
           ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                          SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATIONt))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN REWRITE_TAC[TCSPEC]
        THEN STRIP_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN UNDISCH_LAST_TAC
        THEN ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS)
        THEN IMP_RES_THEN REWRITE_THM assign_list_THM
        THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
        THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL
                  (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA))
        THENL
            [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                              DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND];

              DISCH_TAC
              THEN RES_TAC
              THEN UNDISCH_LAST_TAC
              THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                               ASSUME "y1 = APPEND y y0 :(var)list"]
              THEN USE_IMP_THEN ONCE_REWRITE_THM
                        (SPEC_ALL command_subst_simplify_LEMMA)
              THENL
               [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                                 DL_APPEND;DL_variants;
                                 DISJOINT_variants;DISJOINT_variants_APPEND];

                 DISCH_TAC
                 THEN EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2"
                 THEN CALL_TAC
                 THEN EXISTS_TAC "c:cmd"
                 THEN EXISTS_TAC "vars:var list"
                 THEN EXISTS_TAC "vals:var list"
                 THEN EXISTS_TAC "VS (VES es) s1"
                 THEN EXISTS_TAC "glbs:var list"
                 THEN EXISTS_TAC "pre:aexp"
                 THEN EXISTS_TAC "post:aexp"
                 THEN EXISTS_TAC "calls:prog_env"
                 THEN EXISTS_TAC "rec:aexp"
                 THEN ASM_REWRITE_TAC[]
               ]
         ]
     ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
  "CSPEC(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                               SYM (ASSUME "x0' = variants x0 (FVa q)")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[CSPEC_THM]);;
 e(STRIP_TAC);;

let st2 = get_state();;
set_state st2;;
  e(ONCE_REWRITE_TAC[a_subst_THM]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_TAC (AND2IMP INTERPRET_CALL));;
  e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

   e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                 ASSUME "y1 = APPEND y y0 :(var)list"]);;
   e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

    e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                      DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

    e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_THM assign_list_THM);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL
           (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA)));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

  e(DISCH_TAC);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                     DL_APPEND;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_APPEND]);;

   e(DISCH_TAC);;
   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals') vals' s2");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%






%================================================================%
%   We wish to prove the general case of procedure call.  This   %
%   involves showing that the total correctness of the body      %
%   which is known from the well-formedness of the environment   %
%   is maintained under replacement of the formal parameters.    %
%================================================================%



let CALLt_THM =
 prove_thm
  (`CALLt_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals')  /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs)    /\
      (y   = APPEND v glbs)    /\
      (x0  = logicals x)       /\
      (y0  = logicals y)       /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
       TCSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, r) )",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenv_WFenv_syntax
   THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
               "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
             "TCSPEC(AAND (pre <|a (x // y))
                          ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                    <|a (x0' // x0)) q))
                                  <|a (x // x0')),
                     (c <|c (x1 //v y1)),r)"
             ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                                 SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                            RULE_OF_ADAPTATIONt))))
      THEN EXISTS_TAC "calls <|g (x1 //v y1)"
      THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION)
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
      THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
        THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
   THEN REWRITE_TAC[TCSPEC]
   THEN STRIP_TAC
   THEN ONCE_REWRITE_TAC[a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN UNDISCH_LAST_TAC
   THEN ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS)
   THEN IMP_RES_THEN REWRITE_THM assign_list_THM
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL
             (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA))
   THENL
       [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                         DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

         ALL_TAC
       ]
   THEN DISCH_TAC
   THEN RES_TAC
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                    ASSUME "y1 = APPEND y y0 :(var)list"]
   THEN USE_IMP_THEN ONCE_REWRITE_THM
             (SPEC_ALL command_subst_simplify_LEMMA)
   THENL
       [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                         DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

         DISCH_TAC
         THEN DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))"
         THEN DEFINE_NEW_VAR "ss = (APPEND vals'  (logicals vals') //v
                                    APPEND vals'' (logicals vals''))"
         THEN SUBGOAL_THEN
                "?s3'.
                  C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
                    r (BNDsL(VS(VES es)s1) vals'' (s1 <|s (VES_state es))) s3'"
                STRIP_ASSUME_TAC
         THENL
          [ MATCH_MP_TAC (GENL ["s2:state";"s3:state";"ns:num list"]
                           (SPEC_ALL (SPECL["vals'':var list";"vals':var list"]
                                        C_greater_variants_BODY)))
            THEN EXISTS_TAC "s2:state"
            THEN ASM_REWRITE_TAC[];

            EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals'') vals'' s3'"
            THEN CALL_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:var list"
            THEN EXISTS_TAC "vals:var list"
            THEN EXISTS_TAC "VS (VES es) s1"
            THEN EXISTS_TAC "glbs:var list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
  );;


%< CALL_THM
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
  "CSPEC(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                               SYM (ASSUME "x0' = variants x0 (FVa q)")]
            (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          RULE_OF_ADAPTATION)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[CSPEC_THM]);;
 e(STRIP_TAC);;
let st2 = get_state();;
set_state st2;;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;

  e(SUBGOAL_THEN ((snd o dest_imp o concl o SPEC "s2:state" o GEN "s4:state"
                       o SPEC_ALL o SPEC "vals':var list")
                  INTERPRET_greater_variants_CALL)
       STRIP_ASSUME_TAC);;

   e(MATCH_MP_TAC (SPEC_ALL INTERPRET_greater_variants_CALL));;
   e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
  e(SUBGOAL_THEN "DISJOINT (SL vals') (FVa q)" ASSUME_TAC);;

   e(ASM_REWRITE_TAC[DISJOINT_variants_UNION]);;

  e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
  e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

   e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                 ASSUME "y1 = APPEND y y0 :(var)list"]);;
   e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

    e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                      DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

    e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(UNDISCH_LAST_TAC);;
  e(ASSUME_TAC (SPECL["es:exp list";"s1:state"] VES_EXISTS));;
  e(IMP_RES_THEN REWRITE_THM assign_list_THM);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM (SPEC_ALL
           (ONCE_REWRITE_RULE[EQ_SYM_EQ] post_subst_simplify_LEMMA)));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

  e(DISCH_TAC);;
  e(RES_TAC);;
  e(POP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

let st2 = get_state();;
set_state st2;;
   e(DISCH_TAC);;
   e(DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))");;
   e(DEFINE_NEW_VAR "ss = (APPEND vals' (logicals vals') //v
                           APPEND vals'' (logicals vals''))");;
   e(SUBGOAL_THEN
        "?s3'. C (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
                  r (BNDsL(VS(VES es)s1) vals'' (s1 <|s (VES_state es))) s3'"
        STRIP_ASSUME_TAC);;

    e(MATCH_MP_TAC (GENL ["s2:state";"s3:state";"ns:num list"]
                      (SPEC_ALL (SPECL["vals'':var list";"vals':var list"]
                                  C_greater_variants_BODY))));;
    e(EXISTS_TAC "s2:state");;
    e(ASM_REWRITE_TAC[]);;

   e(EXISTS_TAC "BNDsL (MAP (s1 <|s (VES_state es)) vals'') vals'' s3'");;
   e(CALL_TAC);;
   e(EXISTS_TAC "c:cmd"
     THEN EXISTS_TAC "vars:var list"
     THEN EXISTS_TAC "vals:var list"
     THEN EXISTS_TAC "VS (VES es) s1"
     THEN EXISTS_TAC "glbs:var list"
     THEN EXISTS_TAC "pre:aexp"
     THEN EXISTS_TAC "post:aexp"
     THEN EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "rec:aexp");;
   e(ASM_REWRITE_TAC[]);;
>%



% ===================================================== %
% ===================================================== %
%                  PROCEDURE_CALL_RULE                  %
%                                                       %
%            (partial and total correctness)            %
% ===================================================== %
% ===================================================== %


let CALL_DISJOINT_DISJOINT_THM =
 prove_thm
  (`CALL_DISJOINT_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       CSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, q, r) )",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC WFenv_WFenvp
      THEN MATCH_MP_TAC (SPEC_ALL CALLp_DISJOINT_DISJOINT_THM)
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC (SPEC_ALL CALLt_DISJOINT_DISJOINT_THM)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
       CSPEC( (AAND (pre <|a (u // v))
                    ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                          <|a (x // x0)))
                 <|a (vals' :== es)  ,
              CALL p xs es, q, r) )";;

e(REWRITE_TAC[CSPEC]);;
e(REPEAT STRIP_TAC);;

 e(IMP_RES_TAC WFenv_WFenvp);;
 e(MATCH_MP_TAC (SPEC_ALL CALLp_DISJOINT_DISJOINT_THM));;
 e(ASM_REWRITE_TAC[]);;

 e(MATCH_MP_TAC (SPEC_ALL CALLt_DISJOINT_DISJOINT_THM));;
 e(ASM_REWRITE_TAC[]);;
>%


let CALL_DISJOINT_THM =
 prove_thm
  (`CALL_DISJOINT_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC WFenv_WFenvp
      THEN MATCH_MP_TAC (SPEC_ALL CALLp_DISJOINT_THM)
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC (SPEC_ALL CALLt_DISJOINT_THM)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REWRITE_TAC[CSPEC]);;
e(REPEAT STRIP_TAC);;

 e(IMP_RES_TAC WFenv_WFenvp);;
 e(MATCH_MP_TAC (SPEC_ALL CALLp_DISJOINT_THM));;
 e(ASM_REWRITE_TAC[]);;

 e(MATCH_MP_TAC (SPEC_ALL CALLt_DISJOINT_THM));;
 e(ASM_REWRITE_TAC[]);;
>%






%================================================================%
%   We wish to prove the general case of procedure call.  This   %
%   involves showing that the partial correctness of the body    %
%   which is known from the well-formedness of the environment   %
%   is maintained under replacement of the formal parameters.    %
%================================================================%



let CALL_THM =
 prove_thm
  (`CALL_THM`,
   "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )",

   REWRITE_TAC[CSPEC]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_TAC WFenv_WFenvp
      THEN MATCH_MP_TAC (SPEC_ALL CALLp_THM)
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC (SPEC_ALL CALLt_THM)
      THEN ASM_REWRITE_TAC[]
    ]
  );;


%< CALL_THM
g "!r p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenv r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
        CSPEC( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r) )";;

e(REWRITE_TAC[CSPEC]);;
e(REPEAT STRIP_TAC);;

 e(IMP_RES_TAC WFenv_WFenvp);;
 e(MATCH_MP_TAC (SPEC_ALL CALLp_THM));;
 e(ASM_REWRITE_TAC[]);;

 e(MATCH_MP_TAC (SPEC_ALL CALLt_THM));;
 e(ASM_REWRITE_TAC[]);;
>%



let C_subst_vsubst =
 prove_thm
  (`C_subst_vsubst`,
   "!r c g xs ys s1 s2.
     WFenv_syntax r /\
     WFc c g r /\ 
     (FVc c r SUBSET SL xs) /\
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\
     WFcsubst c r (ys //v xs) ==>
     (C c r (s1 <|s (ys // xs)) s2 ==>
      C (c <|c (ys //v xs)) r s1 (BNDsL (MAP s2 xs) ys s1))",
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM c_subst_biject
   THENL
    [ EXISTS_TAC "g:prog_env"
      THEN ASM_REWRITE_TAC[];

      REWRITE_TAC[BNDsL_o_vsubst_THM]
      THEN IMP_RES_TAC (AND2IMP vsubst_REPLACEMENT_INVERSE)
      THEN POP_ASSUM REWRITE_THM
      THEN MATCH_MP_TAC C_EXTENSION
      THEN EXISTS_TAC "g:prog_env"
      THEN EXISTS_TAC "s1 <|s (ys // xs)"
      THEN ASM_REWRITE_TAC[s_subst_o;MAP_o;o_THM]
      THEN USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
      THENL
       [ ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM (IMP2AND vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< C_subst_vsubst
g "!r c g xs ys s1 s2.
     WFenv_syntax r /\
     WFc c g r /\ 
     (FVc c r SUBSET SL xs) /\
     (LENGTH xs = LENGTH ys) /\ DL xs /\ DL ys /\
     WFcsubst c r (ys //v xs) ==>
     (C c r (s1 <|s (ys // xs)) s2 ==>
      C (c <|c (ys //v xs)) r s1 (BNDsL (MAP s2 xs) ys s1))";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM c_subst_biject);;

 e(EXISTS_TAC "g:prog_env");;
 e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[BNDsL_o_vsubst_THM]);;
 e(IMP_RES_TAC (AND2IMP vsubst_REPLACEMENT_INVERSE));;
 e(POP_ASSUM REWRITE_THM);;
 e(MATCH_MP_TAC C_EXTENSION);;
 e(EXISTS_TAC "g:prog_env");;
 e(EXISTS_TAC "s1 <|s (ys // xs)");;
 e(ASM_REWRITE_TAC[s_subst_o;MAP_o;o_THM]);;
 e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT));;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM (IMP2AND vsubst_REPLACEMENT));;
  e(ASM_REWRITE_TAC[]);;
>%


let PROC_BODY_C_subst_vsubst =
 prove_thm
  (`PROC_BODY_C_subst_vsubst`,
   "!r p vars vals glbs pre post calls rec c xs vals' s1 s2.
     WFenv_syntax r /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (LENGTH xs = LENGTH vars) /\
     (LENGTH vals' = LENGTH vals) /\
     WFxs xs /\ WFxs vals' /\
     DL(APPEND xs vals') /\
     DISJOINT (SL(APPEND xs vals')) (SL glbs) /\
     C c r (s1 <|s (APPEND xs vals' // APPEND vars vals)) s2 ==>
     C (c <|c (APPEND xs vals' //v APPEND vars vals)) r s1
       (BNDsL (MAP s2 (APPEND (APPEND vars vals) glbs))
              (APPEND (APPEND xs vals') glbs) s1)",
   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN
        "C c r (s1 <|s (APPEND (APPEND xs vals') glbs //
                        APPEND (APPEND vars vals) glbs)) s2 ==>
         C (c <|c (APPEND (APPEND xs vals') glbs //v
                   APPEND (APPEND vars vals) glbs)) r s1
           (BNDsL (MAP s2 (APPEND (APPEND vars vals) glbs))
                  (APPEND (APPEND xs vals') glbs) s1)"
     MP_TAC
   THENL
    [ MATCH_MP_TAC C_subst_vsubst
      THEN EXISTS_TAC "calls:prog_env"
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[LENGTH_APPEND]
      THEN ONCE_REWRITE_TAC[DL_APPEND]
      THEN ASM_REWRITE_TAC[]
      THEN ONCE_REWRITE_ALL_TAC[DL_APPEND]
      THEN ASM_REWRITE_TAC[]
      THEN USE_IMP_THEN REWRITE_THM APPEND_IDENTITY_vsubst
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND];

         MATCH_MP_TAC (SPEC_ALL WFcsubst_vsubst)
         THEN ASM_REWRITE_TAC[]
         THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
         THEN ASM_REWRITE_TAC[]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN ASM_REWRITE_TAC[]
       ];

      USE_IMP_AND_THEN REWRITE_THM var_list_IDENT_APPEND
             (ASM_REWRITE_TAC[LENGTH_APPEND])
      THEN USE_IMP_AND_THEN REWRITE_THM APPEND_IDENTITY_vsubst
             (ASM_REWRITE_TAC[LENGTH_APPEND])
      THEN DISCH_THEN MATCH_MP_TAC
      THEN POP_ASSUM REWRITE_THM
    ]
  );;

%< PROC_BODY_C_subst_vsubst
g "!r p vars vals glbs pre post calls rec c xs vals' s1 s2.
     WFenv_syntax r /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
     (LENGTH xs = LENGTH vars) /\
     (LENGTH vals' = LENGTH vals) /\
     WFxs xs /\ WFxs vals' /\
     DL(APPEND xs vals') /\
     DISJOINT (SL(APPEND xs vals')) (SL glbs) /\
     C c r (s1 <|s (APPEND xs vals' // APPEND vars vals)) s2 ==>
     C (c <|c (APPEND xs vals' //v APPEND vars vals)) r s1
       (BNDsL (MAP s2 (APPEND (APPEND vars vals) glbs))
              (APPEND (APPEND xs vals') glbs) s1)";;
e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN
        "C c r (s1 <|s (APPEND (APPEND xs vals') glbs //
                        APPEND (APPEND vars vals) glbs)) s2 ==>
         C (c <|c (APPEND (APPEND xs vals') glbs //v
                   APPEND (APPEND vars vals) glbs)) r s1
           (BNDsL (MAP s2 (APPEND (APPEND vars vals) glbs))
                  (APPEND (APPEND xs vals') glbs) s1)"
     MP_TAC);;

 e(MATCH_MP_TAC C_subst_vsubst);;
 e(EXISTS_TAC "calls:prog_env");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
 e(ONCE_REWRITE_TAC[DL_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(USE_IMP_THEN REWRITE_THM APPEND_IDENTITY_vsubst);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

  e(MATCH_MP_TAC (SPEC_ALL WFcsubst_vsubst));;
  e(ASM_REWRITE_TAC[]);;
  e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
  e(ASM_REWRITE_TAC[]);;
  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(ASM_REWRITE_TAC[]);;

 e(USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

  e(USE_IMP_THEN REWRITE_THM APPEND_IDENTITY_vsubst);;

   e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

let st2 = get_state();;
set_state st2;;
 e(DISCH_THEN MATCH_MP_TAC);;
 e(POP_ASSUM REWRITE_THM);;
>%





let C_calls_rules,C_calls_ind =
    let rules,ind = CONJ_PAIR C_calls_DEF in CONJUNCTS rules,ind;;

let [SEQ1_CALLS_TAC;SEQ2_CALLS_TAC;IF_T_CALLS_TAC;IF_F_CALLS_TAC;
     WHILE1_CALLS_TAC;WHILE2_CALLS_TAC;CALL_CALLS_TAC] =
    map RULE_TAC C_calls_rules;;


let C_calls_CALL_THM =
 prove_thm
  (`C_calls_CALL_THM`,
   "!r p g xs es s1 vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (!s3. C_calls (CALL p xs es) r s1 p s3 ==> (?s4. C c r s3 s4))

     %------------------------------------------------------------------% ==>
        (?s2. C (CALL p xs es) r s1 s2)",

   REPEAT STRIP_TAC
   THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN SUBGOAL_THEN "?ns s2. ES es s1 ns s2" STRIP_ASSUME_TAC
   THENL
    [ EXISTS_TAC "VS (VES es) s1"
      THEN EXISTS_TAC "s1 <|s (VES_state es)"
      THEN REWRITE_TAC[VES_EXISTS];

      DEFINE_NEW_VAR "vals' = variants vals (SL (APPEND xs glbs))"
      THEN SUBGOAL_THEN "C_calls (CALL p xs es) r s1 p
                            ((BNDsL ns vals' s2) <|s
                               ((APPEND xs vals') // (APPEND vars vals)))"
               ASSUME_TAC
      THENL
       [ CALL_CALLS_TAC
         THEN EXISTS_TAC "glbs:var list"
         THEN EXISTS_TAC "pre:aexp"
         THEN EXISTS_TAC "post:aexp"
         THEN EXISTS_TAC "calls:prog_env"
         THEN EXISTS_TAC "rec:aexp"
         THEN EXISTS_TAC "c:cmd"
         THEN ASM_REWRITE_TAC[];

         RES_TAC
         THEN SUBGOAL_THEN
                "C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                   r (BNDsL ns vals' s2)
                   (BNDsL (MAP s4 (APPEND(APPEND vars vals)glbs))
                          (APPEND(APPEND xs vals')glbs)
                          (BNDsL ns vals' s2))"
                 ASSUME_TAC
         THENL
          [ MATCH_MP_TAC (SPEC_ALL PROC_BODY_C_subst_vsubst)
            THEN POP_ASSUM REWRITE_THM
            THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
            THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]
            THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
            THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND];

            EXISTS_TAC
                "BNDsL (MAP (s2:state) vals') vals'
                   (BNDsL
                    (MAP (s4:state) (APPEND (APPEND vars vals) glbs))
                    (APPEND (APPEND xs vals') glbs)
                    (BNDsL ns vals' s2))"
            THEN CALL_TAC
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "vars:var list"
            THEN EXISTS_TAC "vals:var list"
            THEN EXISTS_TAC "ns:num list"
            THEN EXISTS_TAC "glbs:var list"
            THEN EXISTS_TAC "pre:aexp"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "rec:aexp"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

%< C_calls_CALL_THM
g "!r p g xs es s1 vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (!s3. C_calls (CALL p xs es) r s1 p s3 ==> (?s4. C c r s3 s4))

     %------------------------------------------------------------------% ==>
        (?s2. C (CALL p xs es) r s1 s2)";;

e(REPEAT STRIP_TAC);;
e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(SUBGOAL_THEN "?ns s2. ES es s1 ns s2" STRIP_ASSUME_TAC);;

 e(EXISTS_TAC "VS (VES es) s1");;
 e(EXISTS_TAC "s1 <|s (VES_state es)");;
 e(REWRITE_TAC[VES_EXISTS]);;

e(DEFINE_NEW_VAR "vals' = variants vals (SL (APPEND xs glbs))");;
e(SUBGOAL_THEN "C_calls (CALL p xs es) r s1 p
                        ((BNDsL ns vals' s2) <|s
                           ((APPEND xs vals') // (APPEND vars vals)))"
    ASSUME_TAC);;

 e(CALL_CALLS_TAC);;
 e(EXISTS_TAC "glbs:var list"
    THEN EXISTS_TAC "pre:aexp"
    THEN EXISTS_TAC "post:aexp"
    THEN EXISTS_TAC "calls:prog_env"
    THEN EXISTS_TAC "rec:aexp"
    THEN EXISTS_TAC "c:cmd");;
 e(ASM_REWRITE_TAC[]);;

e(RES_TAC);;
e(SUBGOAL_THEN
      "C (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r (BNDsL ns vals' s2)
         (BNDsL (MAP s4 (APPEND(APPEND vars vals)glbs))
                (APPEND(APPEND xs vals')glbs)
                (BNDsL ns vals' s2))"
       ASSUME_TAC);;

 e(MATCH_MP_TAC (SPEC_ALL PROC_BODY_C_subst_vsubst));;

  e(POP_ASSUM REWRITE_THM);;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]);;
  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

e(EXISTS_TAC
      "BNDsL (MAP (s2:state) vals') vals'
         (BNDsL
          (MAP (s4:state) (APPEND (APPEND vars vals) glbs))
          (APPEND (APPEND xs vals') glbs)
          (BNDsL ns vals' s2))");;
e(CALL_TAC);;
e(EXISTS_TAC "c:cmd"
   THEN EXISTS_TAC "vars:var list"
   THEN EXISTS_TAC "vals:var list"
   THEN EXISTS_TAC "ns:num list"
   THEN EXISTS_TAC "glbs:var list"
   THEN EXISTS_TAC "pre:aexp"
   THEN EXISTS_TAC "post:aexp"
   THEN EXISTS_TAC "calls:prog_env"
   THEN EXISTS_TAC "rec:aexp");;
e(ASM_REWRITE_TAC[]);;
>%




close_theory ();;

