% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains theorems about the stage semantics of commands.
  They concern the semantic relations defined in the `semantic stages`
  theory, and are staged version of the theorems proven in `cmd_semantics`.
  These are important theorems, and have substantial proofs, especially
  Ck_EXTENSION and c_substk_biject.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `stage_semantics`;;

new_parent `semantic_stages`;;


% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of stage_semantics %

loadf `autoload`;;

autoload_all_defs_and_thms
     `bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules semantic_stages`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `progress_tactics`;;
loadf `new_ind_def`;;


% --------------------------------------------------------------------- %
% Load the programming language semantics rules.                        %
% --------------------------------------------------------------------- %

let Ck_rules,Ck_ind = let rules,ind = CONJ_PAIR Ck_DEF in CONJUNCTS rules,ind;;

% --------------------------------------------------------------------- %
% Tactics for the rules.						%
% --------------------------------------------------------------------- %

let CMDk_TAC = FIRST (map RULE_TAC Ck_rules);;
let [SKIPk_TAC;ASSIGNk_TAC;SEQk_TAC;IFk_T_TAC;IFk_F_TAC;
     WHILEk_T_TAC;WHILEk_F_TAC;CALLk_TAC] = map RULE_TAC Ck_rules;;

% --------------------------------------------------------------------- %
% Rule induction tactics.                                               %
% --------------------------------------------------------------------- %

let  CMDk_INDUCT_TAC = RULE_INDUCT_THEN Ck_ind
                              STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;

let  CMDk_STR_INDUCT_TAC = mk_str_ind_tac Ck_str_induct;;




%<

 We will want to prove, eventually, 

   |- !c q r k. WFc c r /\ WFenvk r k ==>
                let (p,h) = vcg1 c q p in
                (ALL_EL close h ==> PCSPECk(p,c,q,r,k+1))

in order to prove

   |- !p c q r k. WFc c r /\ WFenvk r k ==>
                  (ALL_EL close (vcgc p c q r) ==> PCSPECk(p,c,q,r,k+1))

in order to prove

   |- !d r k. r = mkenv d r0 /\ WFd d r /\ WFenvk r k ==>
              (ALL_EL close (vcgd d r) ==> WFenvk r (k+1))

and then

   |- !d r. r = mkenv d r0 /\ WFd d r ==>
            (ALL_EL close (vcgd d r) ==> WFenv r)


But first we need to prove all of the superstructure needed to deal
with WFenvk, so that we will have the tools to build the needed
"staged" version of the critical procedure call rule.

>%





%=================================%
%=================================%
%   Now for Hoare-style axioms!   %
%=================================%
%=================================%





% Program language rewriting and simplification tools. %

let Ck_REWRITE_RULE = REWRITE_RULE [ cmd_distinct; cmd_one_one]
                     o CONV_RULE (REWR_CONV Ck_cases) ;;

let imap f = map (\x. f x ? x);;
let Ck_rewrite = LIST_CONJ o (imap Ck_REWRITE_RULE) o CONJUNCTS;;




let PRE_STRENGTH_Ck_THM =
 prove_thm
  (`PRE_STRENGTH_Ck_THM`,
   "!p q s c r k.
                    CLOSE(AIMP p q) /\ PCSPECk(q,c,s,r,k)
            %---------------------------------------------------% ==>
                              PCSPECk(p,c,s,r,k)",

   REWRITE_TAC[CLOSE;PCSPECk;A_DEF]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC
    THEN RES_TAC
  );;

%<
g "!p q s c r k.
                    CLOSE(AIMP p q) /\ PCSPECk(q,c,s,r,k)
            %---------------------------------------------------% ==>
                              PCSPECk(p,c,s,r,k)";;
e(REWRITE_TAC[CLOSE;PCSPECk;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
>%



let POST_WEAK_Ck_THM =
 prove_thm
  (`POST_WEAK_Ck_THM`,
   "!p q s c r k.
                  PCSPECk(p,c,q,r,k)  /\  CLOSE(AIMP q s)
            %---------------------------------------------------% ==>
                             PCSPECk(p,c,s,r,k)",

   REWRITE_TAC[CLOSE;PCSPECk;A_DEF]
    THEN REPEAT STRIP_TAC
    THEN RES_TAC
    THEN RES_TAC
  );;

%<
g "!p q s c r k.
                  PCSPECk(p,c,q,r,k)  /\  CLOSE(AIMP q s)
            %---------------------------------------------------% ==>
                             PCSPEC(p,c,s,r,k)";;
e(REWRITE_TAC[CLOSE;PCSPECk;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RES_TAC);;
e(RES_TAC);;
>%


let FALSE_Ck_THM =
 prove_thm
  (`FALSE_Ck_THM`,
   "!c r k q.
                                                                
            %---------------------------------------------------%
                            PCSPECk(AFALSE,c,q,r,k)",

   REWRITE_TAC[PCSPECk;A_DEF]
  );;



let SKIPk_THM =
 prove_thm
  (`SKIPk_THM`,
   "!q r k.
                                                                
            %---------------------------------------------------%
                            PCSPECk(q,SKIP,q,r,k)",

   REWRITE_TAC[PCSPECk]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN ASM_REWRITE_TAC[]
  );;

%
for SKIP:
g "!q r k. PCSPECk(q,SKIP,q,r,k)";;
e(REWRITE_TAC[PCSPECk]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;
e(ASM_REWRITE_TAC[]);;
%



let ABORTk_THM =
 prove_thm
  (`ABORTk_THM`,
   "!q r k.
                                                                
            %---------------------------------------------------%
                            PCSPECk(AFALSE,ABORT,q,r,k)",

   REWRITE_TAC[PCSPECk;A_DEF]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
  );;


%
for ABORT:
g "!q r k. PCSPECk(AFALSE,ABORT,q,r,k)";;
e(REWRITE_TAC[PCSPECk;A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;
%


let ASSIGNk_THM =
 prove_thm
  (`ASSIGNk_THM`,
   "!q x e r k.
                                                                
            %---------------------------------------------------%
                PCSPECk(q <|a (x := e), ASSIGN x e, q, r, k)",

   REWRITE_TAC[PCSPECk;a_subst_THM;assign;s_subst_BND]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN TRANSLATE_TAC "E"
    THEN ASM_REWRITE_TAC[]
  );;


%<
g "!q x e r k.
                                                                
            %---------------------------------------------------%
                 PCSPECk(q <|a (x := e), ASSIGN x e, q, r, k)";;

e(REWRITE_TAC[PCSPECk]);;
e(REWRITE_TAC[a_subst_THM]);;
e(REWRITE_TAC[assign]);;
e(REWRITE_TAC[s_subst_BND]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;
e(TRANSLATE_TAC "E");;
e(ASM_REWRITE_TAC[]);;

E e s1 n s2 ==> (s1 <|s (x := e) = BND n x s2)
>%



let SEQk_THM =
 prove_thm
  (`SEQk_THM`,
   "!p q s c1 c2 r k.
             PCSPECk(p, c1, s, r, k)  /\  PCSPECk(s, c2, q, r, k)
            %---------------------------------------------------% ==>
                        PCSPECk(p, SEQ c1 c2, q, r, k)",

   REWRITE_TAC[PCSPECk]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN IMP_RES_TAC ALL_GREATER_Ck
    THEN RES_TAC
    THEN RES_TAC
  );;


%<
g "!p q s c1 c2 r k.
                 PCSPECk(p, c1, s, r, k)   /\   PCSPEC(s, c2, q, r, k)
            %---------------------------------------------------% ==>
                        PCSPECk(p, SEQ c1 c2, q, r, k)";;

e(REWRITE_TAC[PCSPECk]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;
e(IMP_RES_TAC ALL_GREATER_Ck);;
e(RES_TAC);;
e(RES_TAC);;
>%



let IFk_THM =
 prove_thm
  (`IFk_THM`,
   "!p q r1 r2 b c1 c2 r k.
                BSPEC(AAND p (AB b), b, r1) /\
                PCSPECk(r1, c1, q, r, k)      /\
                BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(p, IF b c1 c2, q, r, k)",

   REWRITE_TAC[PCSPECk;BSPEC;A_DEF]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN TRANSLATE_TAC "B"
    THEN RES_TAC
    THEN RES_TAC
  );;


%<
g "!p q r1 r2 b c1 c2 r k.
                BSPEC(AAND p (AB b), b, r1) /\
                PCSPECk(r1, c1, q, r, k)      /\
                BSPEC(AAND p (ANOT (AB b)), b, r2) /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(p, IF b c1 c2, q, r, k)";;

e(REWRITE_TAC[PCSPECk;BSPEC]);;
e(REWRITE_TAC[A_DEF]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;
e(TRANSLATE_TAC "B");;
e(RES_TAC);;
e(RES_TAC);;

e(TRANSLATE_TAC "B");;
e(RES_TAC);;
e(RES_TAC);;
>%


let IFk_THM1 =
 prove_thm
  (`IFk_THM1`,
   "!q r1 r2 b c1 c2 r k.
                PCSPECk(r1, c1, q, r, k)      /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                        IF b c1 c2, q, r, k)",

   REWRITE_TAC[PCSPECk;A_DEF;ab_pre;a_subst_THM]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN TRANSLATE_TAC "B"
    THEN POP_ASSUM REWRITE_ALL_THM
    THEN POP_ASSUM REWRITE_ALL_THM
    THEN RES_TAC
  );;

%<
g "!q r1 r2 b c1 c2 r k.
                PCSPECk(r1, c1, q, r, k)      /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(ACOND (AB b) (ab_pre b r1) (ab_pre b r2),
                        IF b c1 c2, q, r, k)";;

e(REWRITE_TAC[PCSPECk]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[ab_pre;a_subst_THM]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;

 e(TRANSLATE_TAC "B");;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(RES_TAC);;

 e(TRANSLATE_TAC "B");;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(RES_TAC);;
>%


let IFk_THM2 =
 prove_thm
  (`IFk_THM2`,
   "!q r1 r2 b c1 c2 r k.
                PCSPECk(r1, c1, q, r, k)      /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(AAND (AIMP       (AB b)  (ab_pre b r1))
                             (AIMP (ANOT (AB b)) (ab_pre b r2)),
                        IF b c1 c2, q, r, k)",

   REWRITE_TAC[PCSPECk;A_DEF;ab_pre;a_subst_THM]
    THEN REPEAT STRIP_TAC
    THEN RULE_STRIP_ASSUM_TAC Ck_rewrite
    THEN TRANSLATE_TAC "B"
    THEN POP_ASSUM REWRITE_ALL_THM
    THEN RES_TAC
  );;

%<
g "!q r1 r2 b c1 c2 r k.
                PCSPECk(r1, c1, q, r, k)      /\
                PCSPECk(r2, c2, q, r, k)
            %---------------------------------------------------% ==>
                PCSPECk(AAND (AIMP       (AB b)  (ab_pre b r1))
                             (AIMP (ANOT (AB b)) (ab_pre b r2)),
                        IF b c1 c2, q, r, k)";;

e(REWRITE_TAC[PCSPECk]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[ab_pre;a_subst_THM]);;
e(REPEAT STRIP_TAC);;
e(RULE_STRIP_ASSUM_TAC Ck_rewrite);;

 e(TRANSLATE_TAC "B");;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(RES_TAC);;

 e(TRANSLATE_TAC "B");;
 e(POP_ASSUM REWRITE_ALL_THM);;
 e(RES_TAC);;
>%


let  CMDk_STR_INDUCT_TAC = RULE_INDUCT_THEN Ck_str_induct
                                           STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;


let BND_SAME_COMMUTE =
 TAC_PROOF
  (([],
    "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)"),
   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:**) = x"
   THENL
    [ ASM_REWRITE_TAC[];

      IMP_RES_TAC BND_COMMUTE
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% BND_SAME_COMMUTE
g "!n x y (s:**->*). BND n y (BND n x s) = BND n x (BND n y s)";;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:**) = x");;

 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC BND_COMMUTE);;
 e(ASM_REWRITE_TAC[]);;
%


let PCSPECk_CHANGE_VAR_LEMMA =
 prove_thm
  (`PCSPECk_CHANGE_VAR_LEMMA`,
   "!a v x b c g r k y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( PCSPECk(p,                  c, AAND a (ALESS v (AVAR x)), r, k) =
        PCSPECk(p <|a (AVAR y / x), c, AAND a (ALESS v (AVAR y)), r, k) )",

   REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (\th.
             (STRIP_ASSUME_TAC o 
              REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
              SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                   UNION FVa p UNION FVa q"])
             variant_not_in_set)
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN REWRITE_TAC[PCSPECk;A_DEF;V_DEF]
   THEN REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]
   THEN EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC
   THENL
       [ FIRST_ASSUM (MP_TAC o
                      SPECL["BND ((s2:state) y) x (s1:state)";
                            "BND (s2 y) x (s2:state)"])
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN REWRITE_TAC[BND_MATCH]
         THEN DISCH_THEN MATCH_MP_TAC
         THEN IMP_RES_TAC Ck_BND_THM
         THEN IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  Ck_FREE_ONLY_IMPACT)
         THEN ASM_REWRITE_TAC[];

         FIRST_ASSUM (MP_TAC o
                   SPECL["BND (s1 x) y (s1:state)";
                         "BND (s2 x) y (s2:state)"])
         THEN REWRITE_TAC[BND_MATCH]
         THEN ONCE_REWRITE_TAC[BND_SAME_COMMUTE]
         THEN REWRITE_TAC[IDENT_BND]
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN DISCH_THEN MATCH_MP_TAC
         THEN IMP_RES_TAC Ck_BND_THM
         THEN IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  Ck_FREE_ONLY_IMPACT)
         THEN ASM_REWRITE_TAC[]
       ]
  );;

%< PCSPECk_CHANGE_VAR_LEMMA
g "!a v x b c g r k y p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r  /\
      (y = variant x
         (FVa a UNION FVv v UNION FVb b UNION FVc c r UNION FVa p UNION FVa q))
      ==>
      ( PCSPECk(p,                  c, AAND a (ALESS v (AVAR x)), r, k) =
        PCSPECk(p <|a (AVAR y / x), c, AAND a (ALESS v (AVAR y)), r, k) )";;
e(REWRITE_TAC[WFc_WHILE;IN_UNION;DE_MORGAN_THM]);;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa p UNION FVa q"])
  variant_not_in_set));;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(REWRITE_TAC[PCSPECk;A_DEF;V_DEF]);;
e(REWRITE_TAC[a_subst_THM;single_subst;s_subst_BND;s_subst_IDENT;V_DEF]);;
e(EQ_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC THEN STRIP_TAC);;

  e(FIRST_ASSUM (MP_TAC o
                 SPECL["BND ((s2:state) y) x (s1:state)";
                       "BND (s2 y) x (s2:state)"]));;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(REWRITE_TAC[BND_MATCH]);;
  e(DISCH_THEN MATCH_MP_TAC);;
  e(IMP_RES_TAC Ck_BND_THM);;
  e(IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  Ck_FREE_ONLY_IMPACT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state(st1);;
  e(FIRST_ASSUM (MP_TAC o
                 SPECL["BND (s1 x) y (s1:state)";
                       "BND (s2 x) y (s2:state)"]));;
  e(REWRITE_TAC[BND_MATCH]);;
  e(ONCE_REWRITE_TAC[BND_SAME_COMMUTE]);;
  e(REWRITE_TAC[IDENT_BND]);;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(DISCH_THEN MATCH_MP_TAC);;
  e(IMP_RES_TAC Ck_BND_THM);;
  e(IMP_RES_TAC (REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO]
                                  Ck_FREE_ONLY_IMPACT));;
  e(ASM_REWRITE_TAC[]);;
>%




let IMP_CONJ_EQ =
 TAC_PROOF
  (([],
    "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))"),
   REPEAT STRIP_TAC
   THEN EQ_TAC
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% IMP_CONJ_EQ
g "!a b c. (a ==> (b /\ c)) = ((a ==> b) /\ (a ==> c))";;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;

 e(REPEAT STRIP_TAC);;

  e(RES_TAC);;

  e(RES_TAC);;
%



let WHILEk_LEMMA =
 prove_thm
  (`WHILEk_LEMMA`,
   "!a v x b c g r k p q n.
      (\n.
        WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
        PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k) /\
        BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
        BSPEC(AAND a (ANOT (AB b)), b, q) ==>
        PCSPECk(AAND a (AEQ v (ANUM n)),
                WHILE a (ALESS v (AVAR x)) b c, q, r, k)
      ) n",

   REPEAT GEN_TAC
   THEN SPEC_TAC("n:num","n:num")
   THEN MATCH_MP_TAC NUM_WELL_FOUNDED
   THEN BETA_TAC
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th))
   THEN ASM_REWRITE_TAC[]
   THEN DISCH_THEN (ASSUME_TAC o REWRITE_RULE[PCSPECk])
   THEN UNDISCH_ALL_TAC
   THEN CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN DEFINE_NEW_VAR
             "y = variant x (FVa a UNION FVv v UNION FVb b UNION
                             FVc c r UNION FVa p UNION FVa q)"
   THEN IMP_RES2_THEN REWRITE_THM BSPEC_CHANGE_VAR_LEMMA
   THEN IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM)
              (AND2IMP PCSPECk_CHANGE_VAR_LEMMA)
   THEN UNDISCH_ALL_TAC
   THEN DISCH_TAC
   THEN DISCH_TAC
   THEN FIRST_ASSUM (STRIP_ASSUME_TAC o
                     REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM])
   THEN IMP_RES_TAC WFb_NOT_WFx
   THEN IMP_RES_TAC WFc_NOT_WFx
   THEN SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN DISCH_THEN (\th.
           (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
            SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                                 UNION FVa p UNION FVa q"])
           variant_not_in_set)
   THEN SUBGOAL_THEN "~y IN FVc (WHILE a (ALESS v (AVAR x))b c) r" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVc_DEF;IN_UNION]; ALL_TAC ]
   THEN SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]
   THEN REWRITE_TAC[AND_IMP_INTRO]
   THEN REWRITE_TAC[BSPEC;PCSPECk;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o Ck_rewrite)
   THENL
       [ UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
         THEN IMP_RES_THEN (SUBST1_TAC o SYM o
                            SPECL["n:num";"s2:state"]) A_BND_THM
         THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM])
         THEN POP_TAC
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "V v (BND n y s3)"
         THEN EXISTS_TAC "BND n y (s3:state)"
         THEN IMP_RES_TAC ALL_GREATER_Ck
         THEN IMP_RES_TAC Ck_BND_THM
         THEN ASM_REWRITE_TAC[]
         THEN ONCE_REWRITE_TAC[CONJ_SYM]
         THEN SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n y (s3:state) y)"
                       REWRITE_THM
              THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "BND n y (s2':state)"
         THEN ASM_REWRITE_TAC[]
         THEN FIRST_ASSUM MATCH_MP_TAC
         THEN EXISTS_TAC "BND n y (s1:state)"
         THEN EXISTS_TAC "T"
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN IMP_RES_TAC B_BND_THM
         THEN ASM_REWRITE_TAC[BND_MATCH];

         UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'"
         THEN FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM])
         THEN POP_TAC
         THEN MATCH_MP_TAC (ASSUME
               "!s1 t s2. (A a s1 /\ ~A (AB b) s1) /\ B b s1 t s2 ==> A q s2")
         THEN EXISTS_TAC "s1:state"
         THEN EXISTS_TAC "F"
         THEN ASM_REWRITE_TAC[]
       ]
  );;

%<
g "!a v x b c g r k p q n.
      (\n.
        WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
        PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k) /\
        BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
        BSPEC(AAND a (ANOT (AB b)), b, q) ==>
        PCSPECk(AAND a (AEQ v (ANUM n)),
                WHILE a (ALESS v (AVAR x)) b c, q, r, k)
      ) n";;

e(REPEAT GEN_TAC);;
e(SPEC_TAC("n:num","n:num"));;
e(MATCH_MP_TAC NUM_WELL_FOUNDED);;
e(BETA_TAC);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (\th. SPEC "0" th; UNDISCH_TAC (concl th)));;
e(ASM_REWRITE_TAC[]);;
e(DISCH_THEN (ASSUME_TAC o REWRITE_RULE[PCSPECk]));;
e(UNDISCH_ALL_TAC);;
e(CONV_TAC (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(DEFINE_NEW_VAR "y = variant x (FVa a UNION FVv v UNION FVb b UNION
                                 FVc c r UNION FVa p UNION FVa q)");;
e(IMP_RES2_THEN REWRITE_THM BSPEC_CHANGE_VAR_LEMMA);;
e(IMP_RES2_THEN (IMP_RES_THEN REWRITE_THM) (AND2IMP PCSPECk_CHANGE_VAR_LEMMA));;
e(UNDISCH_ALL_TAC);;
e(DISCH_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[WFc_WHILE;IN_UNION;DE_MORGAN_THM]));;
e(IMP_RES_TAC WFb_NOT_WFx);;
e(IMP_RES_TAC WFc_NOT_WFx);;
e(SUBGOAL_THEN "~x IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(DISCH_THEN (\th.
  (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION;DE_MORGAN_THM;SYM th] o
   SPECL["x:var";"FVa a UNION FVv v UNION FVb b UNION FVc c r
                        UNION FVa p UNION FVa q"])
  variant_not_in_set));;
e(SUBGOAL_THEN "~y IN FVc (WHILE a (ALESS v (AVAR x))b c) r" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVc_DEF;IN_UNION]; ALL_TAC ]);;
e(SUBGOAL_THEN "~y IN FVa (AB b)" ASSUME_TAC
  THENL [ ASM_REWRITE_TAC[FVa_AB]; ALL_TAC ]);;
e(REWRITE_TAC[AND_IMP_INTRO]);;
e(REWRITE_TAC[BSPEC;PCSPECk;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o Ck_rewrite));;

  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
  e(IMP_RES_THEN (SUBST1_TAC o pthm o SYM o SPECL["n:num";"s2:state"]) A_BND_THM);;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(POP_TAC);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "V v (BND n y s3)");;
  e(EXISTS_TAC "BND n y (s3:state)");;
  e(IMP_RES_TAC ALL_GREATER_Ck);;
  e(IMP_RES_TAC Ck_BND_THM);;
  e(ASM_REWRITE_TAC[]);;
  e(ONCE_REWRITE_TAC[CONJ_SYM]);;
  e(SUBGOAL_THEN "!s. (V v s < n) = (V v s < BND n y (s3:state) y)" REWRITE_THM
     THENL [ REWRITE_TAC[BND_MATCH]; ALL_TAC ]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BND n y (s2':state)");;
  e(ASM_REWRITE_TAC[]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BND n y (s1:state)");;
  e(EXISTS_TAC "T");;
  e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
  e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
  e(IMP_RES_TAC B_BND_THM);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(a:aexp) = a'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "ALESS v (AVAR x) = pr"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(b:bexp) = b'"
     THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(c:cmd) = c'");;
  e(FIRST_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[AB_THM]));;
  e(POP_TAC);;
  e(MATCH_MP_TAC (ASSUME
     "!s1 t s2. (A a s1 /\ ~A (AB b) s1) /\ B b s1 t s2 ==> A q s2"));;
  e(EXISTS_TAC "s1:state");;
  e(EXISTS_TAC "F");;
  e(ASM_REWRITE_TAC[]);;
>%



let COND_IMP =
    TAC_PROOF
    (([], "!a b c. (a => b | c) = ((a ==> b) /\ (~a ==> c))"),
     REPEAT STRIP_TAC
     THEN COND_CASES_TAC
     THEN REWRITE_TAC[]
    );;

% COND_IMP:
g "!a b c. (a => b | c) = ((a ==> b) /\ (~a ==> c))";;
e(REPEAT STRIP_TAC);;
e(COND_CASES_TAC);;
e(REWRITE_TAC[]);;
%




let WHILEk_THM =
 prove_thm
  (`WHILEk_THM`,
   "!a v x b c g r k p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      PCSPECk(a, WHILE a (ALESS v (AVAR x)) b c, q, r, k)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC ((REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO] o
                      CONV_RULE (ONCE_DEPTH_CONV BETA_CONV))  WHILEk_LEMMA)
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[PCSPECk;A_DEF;V_DEF]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC "V v s1")
   THEN EXISTS_TAC "s1:state"
   THEN ASM_REWRITE_TAC[]
  );;


%< WHILEk_THM
g "!a v x b c g r k p q.
      WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
      PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k) /\
      BSPEC(AAND(AAND a (AB b)) (AEQ v (AVAR x)), b, p) /\
      BSPEC(AAND a (ANOT (AB b)), b, q) ==>
      PCSPECk(a, WHILE a (ALESS v (AVAR x)) b c, q, r, k)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC ((REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO] o
                      CONV_RULE (ONCE_DEPTH_CONV BETA_CONV))  WHILEk_LEMMA));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[PCSPECk;A_DEF;V_DEF]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM (MATCH_MP_TAC o SPEC "V v s1"));;
e(EXISTS_TAC "s1:state");;
e(ASM_REWRITE_TAC[]);;
>%



let WHILEk_THM1 =
 prove_thm
  (`WHILEk_THM1`,
   "!p q a b c g r k.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             PCSPECk(a, WHILE a (ALESS v (AVAR x)) b c, q, r, k)",

    REPEAT STRIP_TAC
    THEN MATCH_MP_TAC (SPEC_ALL WHILEk_THM)
    THEN ASM_REWRITE_TAC[]
    THEN CONJ_TAC
    THEN MATCH_MP_TAC PRE_STRENGTH_BSPEC
    THENL
      [  EXISTS_TAC "ab_pre b p";
         EXISTS_TAC "ab_pre b q"
      ]
    THEN ASM_REWRITE_TAC[ab_pre_THM]
  );;

%< WHILEk_THM1
g "!p q a b c g r k.
             WFenv_syntax r /\ WFc (WHILE a (ALESS v (AVAR x)) b c) g r /\
             PCSPECk(p, c, AAND a (ALESS v (AVAR x)), r, k)  /\
             CLOSE (AIMP (AAND (AAND a (AB b)) (AEQ v (AVAR x)))
                         (ab_pre b p))  /\
             CLOSE (AIMP (AAND a (ANOT (AB b))) (ab_pre b q))
            %---------------------------------------------------% ==>
             PCSPECk(a, WHILE a (ALESS v (AVAR x)) b c, q, r, k)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL WHILEk_THM));;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b p");;
 e(ASM_REWRITE_TAC[ab_pre_THM]);;

 e(MATCH_MP_TAC PRE_STRENGTH_BSPEC);;
 e(EXISTS_TAC "ab_pre b q");;
 e(ASM_REWRITE_TAC[ab_pre_THM]);;
>%





%===================================%
%===================================%
%  Now for the Rule of Adaptation!  %
%===================================%
%===================================%


let ROA_POSTCONDITION_INVARIANCE_DISJOINT_STAGE =
 TAC_PROOF
  (([],
   "!x a c g r k s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ Ck c r k s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (logicals x)))) =
         A a (BNDsR (MAP s1 x) (logicals x) s2))"
   ),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC WFxs_NOT_WFxs
   THEN SUBGOAL_THEN "BNDsL (MAP s2 x) x (BNDsR (MAP s1 x) (logicals x) s1) =
                      BNDsR (MAP s1 x) (logicals x) (BNDsL (MAP s2 x) x (s1:state))"
                REWRITE_THM
   THENL
     [  IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE;

        IMP_RES_TAC Ck_FREE_ONLY_VARIATION
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% ROA_POSTCONDITION_INVARIANCE_DISJOINT_STAGE
g "!x a c g r k s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ Ck c r k s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (logicals x)))) =
         A a (BNDsR (MAP s1 x) (logicals x) s2))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
e(IMP_RES_TAC WFxs_NOT_WFxs);;
e(SUBGOAL_THEN "BNDsL (MAP s2 x) x (BNDsR (MAP s1 x) (logicals x) s1) =
                BNDsR (MAP s1 x) (logicals x) (BNDsL (MAP s2 x) x (s1:state))"
           REWRITE_THM);;

  e(IMP_RES_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;

  e(IMP_RES_TAC Ck_FREE_ONLY_VARIATION);;
  e(ASM_REWRITE_TAC[]);;
%



let ROA_POSTCONDITION_INVARIANCE_STAGE =
 TAC_PROOF
  (([],
   "!x s a c g r k s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ Ck c r k s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (variants (logicals x) s)))) =
         A a (BNDsR (MAP s1 x) (variants (logicals x) s) s2))"
   ),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE
   THENL
     [  MATCH_MP_TAC WFxs_NOT_WFxs
        THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals];

        IMP_RES_TAC Ck_FREE_ONLY_VARIATION
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% ROA_POSTCONDITION_INVARIANCE_STAGE
g "!x s a c g r k s1 s2.
        (WFenv_syntax r /\ WFc c g r /\ WFxs x /\
        (FVc c r SUBSET SL x) /\ Ck c r k s1 s2) ==>
        (A a (BNDsL (MAP s2 x) x (s1 <|s (x // (variants (logicals x) s)))) =
         A a (BNDsR (MAP s1 x) (variants (logicals x) s) s2))";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

  e(IMP_RES_TAC Ck_FREE_ONLY_VARIATION);;
  e(ASM_REWRITE_TAC[]);;
%





% ===================================================== %
% ===================================================== %
%                  RULE_OF_ADAPTATION:                  %
% ===================================================== %
% ===================================================== %


let RULE_OF_ADAPTATION_DISJOINT_STAGE =
 prove_thm
  (`RULE_OF_ADAPTATION_DISJOINT_STAGE`,
   "!q pre post c g r k x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             PCSPECk(AAND (mk_eqs x0 x) pre, c, post, r, k) 
           %---------------------------------------------------% ==>
      PCSPECk( AAND pre
                    ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
               c, q, r, k) )",
   REPEAT GEN_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[PCSPECk;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC A_BNDsR_THM
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "MAP (s1:state) x"
   THEN IMP_RES_TAC (AND2IMP ROA_POSTCONDITION_INVARIANCE_DISJOINT_STAGE)
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN FIRST_ASSUM (MATCH_MP_TAC o IMP2AND)
   THEN REWRITE_TAC[LENGTH_MAP]
   THEN POP_ASSUM REWRITE_THM
   THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
   THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
   THEN IMP_RES_TAC (AND2IMP Ck_BNDsR_THM)
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "BNDsR(MAP s1 x)(logicals x)(s1:state)"
   THEN POP_ASSUM REWRITE_THM
   THEN REWRITE_TAC[(SYM o SPEC_ALL) s_subst_var_list]
   THEN CONJ_TAC
   THENL
     [  MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx
        THEN ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals];

        IMP_RES_TAC SUBSET_WFxs_NOT_WFxs
        THEN IMP_RES_THEN ASM_REWRITE_THM A_var_list_subst_THM
     ]
  );;


%< RULE_OF_ADAPTATION_DISJOINT_STAGE
g "!q pre post c g r k x.
             let x0 = logicals x in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))    /\
             (DISJOINT (SL x0) (FVa q)) /\

             PCSPECk(AAND (mk_eqs x0 x) pre, c, post, r, k) 
           %---------------------------------------------------% ==>
      PCSPECk( AAND pre
                    ((mk_foralls x (AIMP post q)) <|a (x // x0)) ,
               c, q, r, k) )";;
e(REPEAT GEN_TAC
    THEN CONV_TAC let_CONV
    THEN REWRITE_TAC[PCSPECk;A_DEF;a_subst_THM;SL_APPEND]
    THEN REWRITE_TAC[A_mk_foralls;A_DEF]
    THEN REPEAT STRIP_TAC);;

e(IMP_RES_TAC A_BNDsR_THM);;
e(POP_TAC
  THEN POP_ASSUM MATCH_MP_TAC
  THEN POP_TAC);;
e(EXISTS_TAC "MAP (s1:state) x");;
e(IMP_RES_TAC (AND2IMP ROA_POSTCONDITION_INVARIANCE_DISJOINT_STAGE));;
e(POP_TAC
  THEN POP_ASSUM MATCH_MP_TAC);;
e(FIRST_ASSUM (MATCH_MP_TAC o IMP2AND));;
e(REWRITE_TAC[LENGTH_MAP]);;
e(POP_ASSUM REWRITE_THM);;
e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
e(IMP_RES_TAC (AND2IMP WFc_NOT_WFxs));;
e(IMP_RES_TAC (AND2IMP Ck_BNDsR_THM));;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "BNDsR(MAP s1 x)(logicals x)(s1:state)");;
e(POP_ASSUM REWRITE_THM);;
e(REWRITE_TAC[(SYM o SPEC_ALL) s_subst_var_list]);;
e(CONJ_TAC);;

  e(MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx);;
  e(ASM_REWRITE_TAC[DL_logicals;LENGTH_logicals]);;

  e(IMP_RES_TAC SUBSET_WFxs_NOT_WFxs);;
  e(IMP_RES_THEN ASM_REWRITE_THM A_var_list_subst_THM);;
>%


let Ck_state_variants_logicals =
 prove_thm
  (`Ck_state_variants_logicals`,
   "!x s c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\
     Ck c r k s1 s2 ==>
     ((MAP s2 (variants(logicals x)s)) = (MAP s1 (variants(logicals x)s)))",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[logicals_THM;variants_THM;MAP]
   THEN REWRITE_TAC[CONS_11]
   THEN REPEAT STRIP_TAC
   THENL
     [  IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT)
        THEN POP_ASSUM MATCH_MP_TAC
        THEN MATCH_MP_TAC WFc_NOT_WFx
        THEN EXISTS_TAC "g:prog_env"
        THEN ASM_REWRITE_TAC[WFx_variant;NOT_WFx_logical];

        RES_TAC
        THEN ASM_REWRITE_TAC[]
     ]
  );;

%
g "!x s c g r k s1 s2.
     WFc c g r /\ WFenv_syntax r /\
     Ck c r k s1 s2 ==>
     ((MAP s2 (variants(logicals x)s)) = (MAP s1 (variants(logicals x)s)))";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[logicals_THM;variants_THM;MAP]);;

 e(REWRITE_TAC[logicals_THM;variants_THM;MAP]);;
 e(REWRITE_TAC[CONS_11]);;
 e(REPEAT STRIP_TAC);;

  e(IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT));;
  e(POP_ASSUM MATCH_MP_TAC);;
  e(MATCH_MP_TAC WFc_NOT_WFx);;
  e(EXISTS_TAC "g:prog_env");;
  e(ASM_REWRITE_TAC[WFx_variant;NOT_WFx_logical]);;

  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;
%


let PCSPECk_logicals_SUBSTITUTION =
 prove_thm
  (`PCSPECk_logicals_SUBSTITUTION`,
   "!x s c g r k pre post.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
           (FVa post SUBSET (SL (APPEND x x0)))   /\
            PCSPECk(AAND (mk_eqs x0  x) pre, c, post, r, k)  
           ==>
            PCSPECk(AAND (mk_eqs x0' x) pre, c, post <|a (x0' // x0), r, k))",

   CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[PCSPECk;A_DEF;a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "(s1:state) <|s ((variants(logicals x)s) // (logicals x))"
   THEN REPEAT CONJ_TAC
   THENL
     [  IMP_RES_TAC (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst)
        THEN ASM_REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)];

        USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
        THEN ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals)
        THEN IMP_RES_TAC SUBSET_WFxs_NOT_WFxs;

        REWRITE_TAC[s_subst_var_list]
        THEN IMP_RES_TAC Ck_state_variants_logicals
        THEN POP_ASSUM REWRITE_THM
        THEN MATCH_MP_TAC (SPEC_ALL Ck_BNDsR_THM)
        THEN ASM_REWRITE_TAC[]
        THEN MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFxs)
        THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
     ]
  );;

%<
g "!x s c g r k pre post.
           let x0  = logicals x in
           let x0' = variants x0 s in
           (WFc c g r /\ WFenv_syntax r /\
            DL x /\ WFxs x  /\
           (FVc c r  SUBSET (SL x))  /\
           (FVa pre  SUBSET (SL x))   /\
           (FVa post SUBSET (SL (APPEND x x0)))   /\
            PCSPECk(AAND (mk_eqs x0  x) pre, c, post, r, k)  
           ==>
            PCSPECk(AAND (mk_eqs x0' x) pre, c, post <|a (x0' // x0), r, k))";;

e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[PCSPECk]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[a_subst_THM]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "(s1:state) <|s ((variants(logicals x)s) // (logicals x))");;
e(REPEAT CONJ_TAC);;  % 3 subgoals %

 e(REWRITE_TAC[SYM (SPEC_ALL a_subst_THM)]);;
 e(IMP_RES_TAC (CONV_RULE (DEPTH_CONV let_CONV) mk_eqs_variant_subst));;
 e(ASM_REWRITE_TAC[]);;

 e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;
 e(ASSUME_TAC (SPEC "x:(var)list" NOT_WFxs_logicals));;
 e(IMP_RES_TAC SUBSET_WFxs_NOT_WFxs);;

 e(REWRITE_TAC[s_subst_var_list]);;
 e(IMP_RES_TAC Ck_state_variants_logicals);;
 e(POP_ASSUM REWRITE_THM);;
 e(MATCH_MP_TAC (SPEC_ALL Ck_BNDsR_THM));;
 e(ASM_REWRITE_TAC[]);;
 e(MATCH_MP_TAC (SPEC_ALL WFc_NOT_WFxs));;
 e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

let st1 = get_state();;
set_state st1;;
>%




% RULE_OF_ADAPTATION_STAGE (General Case for one stage) %

let RULE_OF_ADAPTATION_STAGE =
 prove_thm
  (`RULE_OF_ADAPTATION_STAGE`,
   "!q pre post c g r k x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             PCSPECk(AAND (mk_eqs x0 x) pre, c, post, r, k) 
           %---------------------------------------------------% ==>
       PCSPECk( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, q, r, k) )",

   REPEAT GEN_TAC
   THEN REPEAT (CONV_TAC let_CONV)
   THEN REWRITE_TAC[PCSPECk;A_DEF;a_subst_THM;SL_APPEND]
   THEN REWRITE_TAC[A_mk_foralls;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN ASSUME_TAC (SPECL ["logicals x";"FVa q"] DISJOINT_variants)
   THEN IMP_RES_TAC (AND2IMP A_BNDsR_THM)
   THEN POP_TAC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN POP_TAC
   THEN EXISTS_TAC "MAP (s1:state) x"
   THEN USE_IMP_THEN REWRITE_THM
                 (SPEC_ALL (ONCE_REWRITE_RULE[EQ_SYM_EQ]
                                    ROA_POSTCONDITION_INVARIANCE_STAGE))
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN FIRST_ASSUM (MATCH_MP_TAC o IMP2AND)
   THEN REWRITE_TAC[LENGTH_MAP]
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL ROA_POSTCONDITION_INVARIANCE_STAGE)
        THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
   THEN SUBGOAL_THEN
        "PCSPECk(AAND (mk_eqs (variants(logicals x)(FVa q)) x) pre, c,
            post <|a ((variants(logicals x)(FVa q)) // (logicals x)),r,k)"
           (ASSUME_TAC o REWRITE_RULE[PCSPECk;A_DEF])
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                              PCSPECk_logicals_SUBSTITUTION))
        THEN REWRITE_TAC[PCSPECk;A_DEF;SL_APPEND]
        THEN ASM_REWRITE_TAC[];

        POP_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "BNDsR (MAP s1 x) (variants(logicals x)(FVa q)) (s1:state)"
        THEN REWRITE_TAC[SYM (SPEC_ALL s_subst_var_list)]
        THEN REPEAT CONJ_TAC
        THENL
          [  MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx
             THEN ASM_REWRITE_TAC[LENGTH_variants;LENGTH_logicals;DL_variants;
                                  NOT_WFxs_variants;NOT_WFxs_logicals];

             USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM
             THEN MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs
             THEN EXISTS_TAC "x:(var)list"
             THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals];

             REWRITE_TAC[s_subst_var_list]
             THEN MATCH_MP_TAC (SPEC_ALL Ck_BNDsR_THM)
             THEN ASM_REWRITE_TAC[]
             THEN MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs
             THEN EXISTS_TAC "x:(var)list"
             THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
          ]
     ]
  );;

%<
g "!q pre post c g r k x.
             let x0  = logicals x in
             let x0' = variants x0 (FVa q) in
             (WFc c g r /\ WFenv_syntax r /\
              DL x /\ WFxs x  /\
             (FVc c r SUBSET SL x)  /\
             (FVa pre SUBSET SL x)    /\
             (FVa post SUBSET (SL (APPEND x x0)))   /\

             PCSPECk(AAND (mk_eqs x0 x) pre, c, post, r, k) 
           %---------------------------------------------------% ==>
       PCSPECk( AAND pre
                    ((mk_foralls x (AIMP (post <|a (x0' // x0)) q))
                            <|a (x // x0')) ,
               c, q, r, k) )";;

e(REPEAT GEN_TAC
    THEN REPEAT (CONV_TAC let_CONV)
    THEN REWRITE_TAC[PCSPECk;A_DEF;a_subst_THM;SL_APPEND]
    THEN REWRITE_TAC[A_mk_foralls;A_DEF]
    THEN REPEAT STRIP_TAC);;

e(ASSUME_TAC (SPECL ["logicals x";"FVa q"] DISJOINT_variants));;
e(IMP_RES_TAC (AND2IMP A_BNDsR_THM));;
e(POP_TAC
  THEN POP_ASSUM MATCH_MP_TAC
  THEN POP_TAC);;
e(EXISTS_TAC "MAP (s1:state) x");;
e(USE_IMP_THEN REWRITE_THM
      (SPEC_ALL (ONCE_REWRITE_RULE[EQ_SYM_EQ]
                        ROA_POSTCONDITION_INVARIANCE_STAGE)));;

  e(ASM_REWRITE_TAC[]);;

e(FIRST_ASSUM (MATCH_MP_TAC o IMP2AND));;
e(REWRITE_TAC[LENGTH_MAP]);;
e(USE_IMP_THEN REWRITE_THM (SPEC_ALL ROA_POSTCONDITION_INVARIANCE_STAGE));;

  e(ASM_REWRITE_TAC[]);;

e(SUBGOAL_THEN
    "PCSPECk(AAND (mk_eqs (variants(logicals x)(FVa q)) x) pre, c,
            post <|a ((variants(logicals x)(FVa q)) // (logicals x)),r,k)"
       (ASSUME_TAC o REWRITE_RULE[PCSPECk;A_DEF]));;

  e(MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          PCSPECk_logicals_SUBSTITUTION)));;
  e(REWRITE_TAC[PCSPECk;A_DEF;SL_APPEND]);;
  e(ASM_REWRITE_TAC[]);;

  e(POP_ASSUM MATCH_MP_TAC);;
  e(EXISTS_TAC "BNDsR (MAP s1 x) (variants(logicals x)(FVa q)) (s1:state)");;
  e(REWRITE_TAC[SYM (SPEC_ALL s_subst_var_list)]);;
  e(REPEAT CONJ_TAC);;

    e(MATCH_MP_TAC A_mk_eqs_NOT_WFx_WFx);;
    e(ASM_REWRITE_TAC[LENGTH_variants;LENGTH_logicals;DL_variants;
                      NOT_WFxs_variants;NOT_WFxs_logicals]);;

    e(USE_IMP_THEN ASM_REWRITE_THM A_var_list_subst_THM);;
    e(MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs);;
    e(EXISTS_TAC "x:(var)list");;
    e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

    e(REWRITE_TAC[s_subst_var_list]);;
    e(MATCH_MP_TAC (SPEC_ALL Ck_BNDsR_THM));;
    e(ASM_REWRITE_TAC[]);;
    e(MATCH_MP_TAC SUBSET_WFxs_NOT_WFxs);;
    e(EXISTS_TAC "x:(var)list");;
    e(ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;

b();;
let st1 = get_state();;
set_state st1;;
>%




let mk_eqs_logicals_vsubst =
 TAC_PROOF
  (([],
   "!x y. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
               WFxs x /\ WFxs y ==>
   (mk_eqs(logicals y)y =
   (mk_eqs(logicals (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x))
                    (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x)))"
   ),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_REPLACEMENT
   THEN ASM_REWRITE_TAC[]
  );;

% mk_eqs_logicals_vsubst
g "!x y. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
         WFxs x /\ WFxs y ==>
   (mk_eqs(logicals y)y =
   (mk_eqs(logicals (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x))
                    (MAP (APPEND y (logicals y) //v
                          APPEND x (logicals x)) x)))";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_REPLACEMENT);;
e(ASM_REWRITE_TAC[]);;
%




let PCSPECk_pre_post_MUTATION =
 TAC_PROOF
  (([],
   "!x y z k. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ WFxs y /\
             (FVa pre  SUBSET SL x) /\
             (FVa post SUBSET SL (APPEND x (logicals x)))  ==>
    let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
    (PCSPECk(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                  (pre <|a (y // x)),
             c <|c ss,
             post <|a (APPEND y (logicals y) // APPEND x (logicals x)),
             r, k) =
     PCSPECk(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                  (pre <|av ss),
             c <|c ss,
             post <|av ss,
             r, k))"
   ),

   REPEAT STRIP_TAC
   THEN CONV_TAC let_CONV
   THEN REWRITE_TAC[PCSPECk;A_DEF]
   THEN FORALL_EQ_TAC
   THEN FORALL_EQ_TAC
   THEN USE_IMP_THEN REWRITE_THM FVa_vsubst_subst
   THENL
     [  IMP_RES_TAC WFxs_NOT_logicals
        THEN ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                             DL_APPEND;DL_logicals]
        THEN MATCH_MP_TAC SUBSET_TRANS
        THEN EXISTS_TAC "SL (x:(var)list)"
        THEN ASM_REWRITE_TAC[SUBSET_UNION];

        USE_IMP_THEN REWRITE_THM FVa_vsubst_subst
        THENL
          [  IMP_RES_TAC WFxs_NOT_logicals
             THEN ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                                  DL_APPEND;DL_logicals];

             IMP_RES_TAC A_NULL_APPEND_subst
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% PCSPECk_pre_post_MUTATION
g "!x y z k. (LENGTH x = LENGTH y) /\ DL x /\ DL y /\
             WFxs x /\ WFxs y /\
             (FVa pre  SUBSET SL x) /\
             (FVa post SUBSET SL (APPEND x (logicals x)))  ==>
    let ss = (APPEND y (logicals y) //v APPEND x (logicals x)) in
    (PCSPECk(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                  (pre <|a (y // x)),
             c <|c ss,
             post <|a (APPEND y (logicals y) // APPEND x (logicals x)),
             r, k) =
     PCSPECk(AAND (mk_eqs(logicals(MAP ss z))(MAP ss z))
                  (pre <|av ss),
             c <|c ss,
             post <|av ss,
             r, k))";;
e(REPEAT STRIP_TAC);;
e(CONV_TAC let_CONV);;
e(REWRITE_TAC[PCSPECk;A_DEF]);;
e(FORALL_EQ_TAC);;
e(FORALL_EQ_TAC);;
e(USE_IMP_THEN (REWRITE_THM o pthm) FVa_vsubst_subst);;

  e(IMP_RES_TAC WFxs_NOT_logicals);;
  e(ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                    DL_APPEND;DL_logicals]);;
  e(MATCH_MP_TAC SUBSET_TRANS);;
  e(EXISTS_TAC "SL (x:(var)list)");;
  e(ASM_REWRITE_TAC[SUBSET_UNION]);;

  e(USE_IMP_THEN (REWRITE_THM o pthm) FVa_vsubst_subst);;

    e(IMP_RES_TAC WFxs_NOT_logicals);;
    e(ASM_REWRITE_TAC[SL_APPEND;LENGTH_APPEND;LENGTH_logicals;
                      DL_APPEND;DL_logicals]);;

    e(IMP_RES_TAC A_NULL_APPEND_subst);;
    e(ASM_REWRITE_TAC[]);;
%




let MAP_STANDARD_SUBST_GLOBALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          glbs =
      glbs)"),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
   THENL
     [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION];

        USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2
        THENL
          [  ASM_REWRITE_TAC[SUBSET_REFL];

             MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% MAP_STANDARD_SUBST_GLOBALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          glbs =
      glbs)";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION]);;

  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst2);;

    e(ASM_REWRITE_TAC[SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[]);;
%


let MAP_STANDARD_SUBST_LOCALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          vs =
      vs')"),
   REPEAT STRIP_TAC
   THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
   THENL
     [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION];

        USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
        THENL
          [  ASM_REWRITE_TAC[SUBSET_REFL];

             MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
             THEN ASM_REWRITE_TAC[]
          ]
     ]
  );;

% MAP_STANDARD_SUBST_LOCALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          vs =
      vs')";;
e(REPEAT STRIP_TAC);;
e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;SL_APPEND;SUBSET_UNION]);;

  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[]);;
%



let MAP_STANDARD_SUBST_LOCALS_AND_GLOBALS =
 TAC_PROOF
  (([],
   "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          (APPEND vs glbs) =
      APPEND vs' glbs)"),
   REPEAT STRIP_TAC
   THEN ONCE_REWRITE_TAC[MAP_APPEND]
   THEN USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_LOCALS
   THENL
     [  ASM_REWRITE_TAC[];

        USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_GLOBALS
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% MAP_STANDARD_SUBST_LOCALS_AND_GLOBALS
g "!vs vs' glbs:(var)list.
     (LENGTH vs' = LENGTH vs) /\ DL vs /\ DL vs' /\ DL glbs /\
     DISJOINT(SL vs)(SL glbs) /\ DISJOINT(SL vs')(SL glbs)
     ==>
     (MAP ((APPEND(APPEND vs' glbs)(logicals(APPEND vs' glbs))) //v
           (APPEND(APPEND vs glbs)(logicals(APPEND vs glbs))))
          (APPEND vs glbs) =
      APPEND vs' glbs)";;
e(REPEAT STRIP_TAC);;
e(ONCE_REWRITE_TAC[MAP_APPEND]);;
e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_LOCALS);;

  e(ASM_REWRITE_TAC[]);;

e(USE_IMP_THEN REWRITE_THM MAP_STANDARD_SUBST_GLOBALS);;
e(ASM_REWRITE_TAC[]);;
%



let PROCEDURE_CHANGE_OF_VARIABLES_STAGE =
 prove_thm
  (`PROCEDURE_CHANGE_OF_VARIABLES_STAGE`,
   "!r k p vars vals glbs pre post calls rec c xs vals'.
      WFenvk r k /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((GVc c' r) SUBSET (SL glbs)) /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPECk (AAND (mk_eqs x0' x') pre', c', post', r, k))",

   REPEAT GEN_TAC THEN STRIP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN UNDISCH_TAC "WFenvk r k"
   THEN REWRITE_TAC[WFenvk_aspects;WFenvk_partial]
   THEN STRIP_TAC
   THEN POP_ASSUM (MP_TAC o SPEC_ALL)
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[WFprock_partial]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN SUBGOAL_THEN
          "WFcsubst c r
               ((APPEND (APPEND (APPEND xs vals') glbs)
                  (logicals (APPEND (APPEND xs vals') glbs))) //v
                (APPEND (APPEND (APPEND vars vals) glbs)
                  (logicals (APPEND (APPEND vars vals) glbs))))"
             ASSUME_TAC
   THENL
     [  MATCH_MP_TAC (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                          WFcsubst_standard_subst))
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
        THEN REWRITE_ALL_TAC[DL_APPEND;WFxs_APPEND]
        THEN ASM_REWRITE_TAC[]
        THEN REPEAT CONJ_TAC  % 6 subgoals %
        THENL
          [  IMP_RES_THEN ASM_REWRITE_THM WFc_subst;

             ASM_REWRITE_TAC[GVc_subst];

             USE_IMP_THEN REWRITE_THM FVc_subst
             THENL
               [  EXISTS_TAC "glbs:(var)list"
                  THEN ASM_REWRITE_TAC[]
                  THEN MATCH_MP_TAC MAP_STANDARD_SUBST_GLOBALS
                  THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                  USE_IMP_THEN REWRITE_THM IMAGE_NULL_APPEND_vsubst
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM vsubst_SUBSET
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]
                    ]
               ];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND];

             MATCH_MP_TAC FVa_subst_SUBSET
             THEN ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;LENGTH_logicals;
                                  DL_logicals;logicals_APPEND;DISJOINT_APPEND]
             THEN REPEAT CONJ_TAC
             THEN MATCH_MP_TAC WFxs_NOT_logicals
             THEN ASM_REWRITE_TAC[];

             USE_IMP_THEN ONCE_REWRITE_THM
                      (SPEC "APPEND (APPEND vars vals) (glbs:(var)list)"
                            mk_eqs_logicals_vsubst)
             THENL
               [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                  USE_IMP_THEN REWRITE_THM
                      (CONV_RULE (DEPTH_CONV let_CONV) PCSPECk_pre_post_MUTATION)
                  THENL
                    [  ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND];

                       USE_IMP_THEN ASM_REWRITE_THM PCSPECk_vsubst
                       THEN EXISTS_TAC "calls:prog_env"
                       THEN ASM_REWRITE_TAC[WFxs_APPEND]
                       THEN MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
                       THEN ASM_REWRITE_TAC[LENGTH_APPEND;WFxs_APPEND]
                    ]
               ]
          ]
     ]
  );;

%< PROCEDURE_CHANGE_OF_VARIABLES_STAGE
g "!r k p vars vals glbs pre post calls rec c xs vals'.
      WFenvk r k /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((GVc c' r) SUBSET (SL glbs)) /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPECk (AAND (mk_eqs x0' x') pre', c', post', r, k))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(UNDISCH_TAC "WFenvk r k");;
e(REWRITE_TAC[WFenvk_aspects]);;
e(REWRITE_TAC[WFenvk_partial]);;
e(STRIP_TAC);;
e(POP_ASSUM (MP_TAC o SPEC_ALL));;
e(ASM_REWRITE_TAC[]);;
e(REWRITE_TAC[WFprock_partial]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;
e(SUBGOAL_THEN
    "WFcsubst c r
            ((APPEND (APPEND (APPEND xs vals') glbs)
               (logicals (APPEND (APPEND xs vals') glbs))) //v
             (APPEND (APPEND (APPEND vars vals) glbs)
               (logicals (APPEND (APPEND vars vals) glbs))))"
       ASSUME_TAC);;

 e(MATCH_MP_TAC (CONV_RULE (DEPTH_CONV let_CONV)
                        WFcsubst_standard_subst_STAGE));;
 e(EXISTS_TAC "k:num"
   THEN EXISTS_TAC "p:string"
   THEN EXISTS_TAC "pre:aexp"
   THEN EXISTS_TAC "post:aexp");;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[DL_APPEND;WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(REPEAT CONJ_TAC);;  % 6 subgoals %

  e(IMP_RES_THEN ASM_REWRITE_THM WFc_subst);;

  e(ASM_REWRITE_TAC[GVc_subst]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:(var)list");;
   e(ASM_REWRITE_TAC[]);;
   e(MATCH_MP_TAC MAP_STANDARD_SUBST_GLOBALS);;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

   e(USE_IMP_THEN REWRITE_THM IMAGE_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

    e(USE_IMP_THEN ASM_REWRITE_THM vsubst_SUBSET);;
    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

  e(MATCH_MP_TAC FVa_subst_SUBSET);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND]);;

  e(MATCH_MP_TAC FVa_subst_SUBSET);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;LENGTH_logicals;
                    DL_logicals;logicals_APPEND;DISJOINT_APPEND]);;
  e(REPEAT CONJ_TAC
    THEN MATCH_MP_TAC WFxs_NOT_logicals
    THEN ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN ONCE_REWRITE_THM
         (SPEC "APPEND (APPEND vars vals) (glbs:(var)list)"
               mk_eqs_logicals_vsubst));;

   e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND]);;

   e(USE_IMP_THEN (REWRITE_THM o pthm)
         (CONV_RULE (DEPTH_CONV let_CONV) PCSPECk_pre_post_MUTATION));;

    e(ASM_REWRITE_TAC[LENGTH_APPEND;DL_APPEND;WFxs_APPEND]);;

    e(USE_IMP_THEN ASM_REWRITE_THM PCSPECk_vsubst);;
    e(EXISTS_TAC "calls:prog_env");;
    e(ASM_REWRITE_TAC[WFxs_APPEND]);;
    e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
    e(ASM_REWRITE_TAC[LENGTH_APPEND;WFxs_APPEND]);;
>%


let PROCEDURE_CHANGE_OF_VARIABLES_COR_STAGE =
 TAC_PROOF
  (([],
    "!r k p vars vals glbs pre post calls rec c xs vals'.
      WFenvk r k /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPECk (AAND (mk_eqs x0' x') pre', c', post', r, k))"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP PROCEDURE_CHANGE_OF_VARIABLES_STAGE)
   THEN UNDISCH_LAST_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% PROCEDURE_CHANGE_OF_VARIABLES_COR_STAGE
g "!r k p vars vals glbs pre post calls rec c xs vals'.
      WFenvk r k /\ (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT (SL (APPEND xs vals')) (SL glbs) ==>
      let x   = APPEND (APPEND vars vals) glbs in
      let x0  = logicals x in
      let x'  = APPEND (APPEND xs vals') glbs in
      let x0' = logicals x' in
      let ss  = (APPEND x' x0' //v APPEND x x0) in
      let c'  = c <|c ss in
      let g'  = calls <|g ss in
      let pre'  = pre  <|a (x' // x) in
      let post' = post <|a (APPEND x' x0' // APPEND x x0) in
      (WFc c' g' r /\ DL x' /\
       WFxs x' /\
       ((FVc c' r) SUBSET (SL x')) /\
       ((FVa pre')  SUBSET (SL x')) /\
       ((FVa post') SUBSET (SL (APPEND x' x0'))) /\
       PCSPECk (AAND (mk_eqs x0' x') pre', c', post', r, k))";;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP PROCEDURE_CHANGE_OF_VARIABLES));;
e(UNDISCH_LAST_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
%




let INTERPRET_CALL_STAGE =
 prove_thm
  (`INTERPRET_CALL_STAGE`,
   "!r k p vars vals glbs pre post calls rec c vals' xs es s1 s3'.
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     Ck (CALL p xs es) r (k+1) s1 s3'  ==>
     ?s2 s3 ns. (s3' = BNDsL (MAP s2 vals') vals' s3) /\
                 Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    r k (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ",
   REPEAT STRIP_TAC
   THEN POP_ASSUM (STRIP_ASSUME_TAC o Ck_REWRITE_RULE)
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM
                       o REWRITE_RULE[EQ_MONO_ADD_EQ]) "k + 1 = k' + 1"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "p = (p':string)"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "xs = (xs':(var)list)"
   THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "es = (es':(exp)list)"
   THEN UNDISCH_TAC "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN UNDISCH_TAC "(r:env) p = vars',vals'',glbs',pre',post',calls',rec',c'"
   THEN DISCH_THEN (\th'. DISCH_THEN (\th.
              (REWRITE_ALL_THM
                      o ONCE_REWRITE_RULE[EQ_SYM_EQ]
                      o REWRITE_RULE[PAIR_EQ]) 
                   (TRANS (SYM th) th')))
   THEN EXISTS_TAC "s2':state"
   THEN EXISTS_TAC "s3:state"
   THEN EXISTS_TAC "ns:(num)list"
   THEN REPEAT CONJ_TAC THEN ASM_REWRITE_TAC[]
  );;

%
g "!r k p vars vals glbs pre post calls rec c vals' xs es s1 s3'.
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     Ck (CALL p xs es) r (k+1) s1 s3'  ==>
     ?s2 s3 ns. (s3' = BNDsL (MAP s2 vals') vals' s3) /\
                 Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                    r k (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ";;
e(REPEAT STRIP_TAC);;
e(POP_ASSUM (STRIP_ASSUME_TAC o Ck_REWRITE_RULE));;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM
                 o REWRITE_RULE[EQ_MONO_ADD_EQ]) "k + 1 = k' + 1");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "p = (p':string)");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "xs = (xs':(var)list)");;
e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "es = (es':(exp)list)");;
e(UNDISCH_TAC "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(UNDISCH_TAC "(r:env) p = vars',vals'',glbs',pre',post',calls',rec',c'");;
e(DISCH_THEN (\th'. DISCH_THEN (\th. (REWRITE_ALL_THM
          o ONCE_REWRITE_RULE[EQ_SYM_EQ] o pthm) 
          (REWRITE_RULE[PAIR_EQ](TRANS (SYM th) th')))));;
e(EXISTS_TAC "s2':state");;
e(EXISTS_TAC "s3:state");;
e(EXISTS_TAC "ns:(num)list");;
e(REPEAT CONJ_TAC THEN ASM_REWRITE_TAC[]);;
%



let proc_subst_ONLY_IMPACT_STAGE =
 prove_thm
  (`proc_subst_ONLY_IMPACT_STAGE`,
    "!r k p vars vals glbs pre post calls rec c xs vals' ns s2 s3 z.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      DL(APPEND xs glbs) /\
      Ck (c <|c (APPEND xs vals' //v APPEND vars vals))
        r k (BNDsL ns vals' s2) s3        /\
      ~(z IN SL(APPEND(APPEND xs vals')glbs))
      ==>
      (s3 z = s2 z)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP FVc_proc_subst)
   THEN IMP_RES_TAC (AND2IMP WFc_proc_subst)
   THEN IMP_RES_TAC NOT_IN_SUBSET
   THEN IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT)
   THEN POP_ASSUM REWRITE_THM
   THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
   THEN UNDISCH_TAC "~(z:var) IN (SL(APPEND(APPEND xs vals')glbs))"
   THEN REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
   THEN DISCH_THEN REWRITE_THM
  );;

%
g "!r k p vars vals glbs pre post calls rec c xs vals' ns s2 s3 z.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      DL(APPEND xs glbs) /\
      Ck (c <|c (APPEND xs vals' //v APPEND vars vals))
        r k (BNDsL ns vals' s2) s3        /\
      ~(z IN SL(APPEND(APPEND xs vals')glbs))
      ==>
      (s3 z = s2 z)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP FVc_proc_subst));;
e(IMP_RES_TAC (AND2IMP WFc_proc_subst));;
e(IMP_RES_TAC NOT_IN_SUBSET);;
e(IMP_RES_TAC (AND2IMP Ck_FREE_ONLY_IMPACT));;
e(POP_ASSUM REWRITE_THM);;
e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;
e(UNDISCH_TAC "~(z:var) IN (SL(APPEND(APPEND xs vals')glbs))");;
e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
e(DISCH_THEN REWRITE_THM);;
%




let Ck_SUBSTITUTED_BODY =
 prove_thm
  (`Ck_SUBSTITUTED_BODY`,
   "!vals' r k p vars vals glbs pre post calls rec c g xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
       r k (BNDsL ns vals' s2) s3 /\
     ES es s1 ns s2
     ==>
     (?s3'. Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
            r k (BNDsL ns (MAP ss vals') s2) s3')",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                               (MAP ss (APPEND (APPEND xs vals') glbs)))
                          (MAP ss (APPEND (APPEND xs vals') glbs))
                          (BNDsL ns (MAP ss vals') s2)"
   THEN SUBGOAL_THEN
             "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
           STRIP_ASSUME_TAC
        THENL
         [ USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11)
           THENL
            [ REWRITE_TAC[LENGTH_MAP];

              ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
            ];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       r ss"
           ASSUME_TAC
        THENL
         [ REWRITE_TAC[WFcsubst]
           THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss"))
           THEN ASM_REWRITE_TAC[GVc_subst]
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REPEAT STRIP_TAC
           THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
           THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
           THEN RES_TAC;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                  (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL WFc_vsubst)
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REWRITE_ALL_TAC[WFxs_APPEND]
           THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                                DL_variants;DISJOINT_variants_UNION];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[DL_APPEND]
           THEN ASM_REWRITE_TAC[DL_variants]
           THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
           THEN REWRITE_TAC[DISJOINT_variants_UNION];

           ALL_TAC
         ]
   THEN MATCH_MP_TAC Ck_EXTENSION
      THEN EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss"
      THEN EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)"
      THEN REPEAT CONJ_TAC
      THENL  % 5 subgoals %
       [ IMP_RES_THEN REWRITE_THM WFc_subst
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM FVc_subst
         THENL
          [ EXISTS_TAC "glbs:var list"
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[GVc_subst];

            REWRITE_TAC[SL_MAP]
            THEN MATCH_MP_TAC IMAGE_SUBSET
            THEN MATCH_MP_TAC (SPEC_ALL FVc_vsubst)
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN REWRITE_ALL_TAC[WFxs_APPEND]
            THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                                 DL_variants;DISJOINT_variants_UNION]
          ];

         REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]
         THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
         THEN REWRITE_TAC[I_o_ID]
         THEN REWRITE_TAC[MAP_EXTENSION]
         THEN REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)"
         THENL
          [ IMP_RES_TAC IN_DISJOINT_LIST
            THEN SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC
            THENL
             [ REWRITE_TAC[SL_MAP;IN_IMAGE]
               THEN IMP_RES_THEN REWRITE_THM ONE_ONE
               THEN STRIP_TAC
               THEN UNDISCH_LAST_TAC
               THEN POP_ASSUM (REWRITE_THM o SYM)
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN RES_THEN REWRITE_THM
             ];

            ASM_CASES_TAC "(x:var) IN SL vals'"
            THENL
             [ USE_IMP_THEN REWRITE_THM
                    (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE)
               THEN IMP_RES_TAC ES_LENGTH_THM
               THEN ASM_REWRITE_TAC[LENGTH_variants];

               REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ];

         USE_IMP_THEN REWRITE_THM c_substk_biject
         THENL
          [ EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
            THEN ASM_REWRITE_TAC[];

            REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN REWRITE_TAC[I_o_ID]
            THEN CHECK_TAC
          ]
       ]
  );;

%< Ck_SUBSTITUTED_BODY_STAGE
g "!vals' r k p vars vals glbs pre post calls rec c g xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
       r k (BNDsL ns vals' s2) s3 /\
     ES es s1 ns s2
     ==>
     (?s3'. Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
            r k (BNDsL ns (MAP ss vals') s2) s3')";;

e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                         (MAP ss (APPEND (APPEND xs vals') glbs)))
                    (MAP ss (APPEND (APPEND xs vals') glbs))
                    (BNDsL ns (MAP ss vals') s2)");;
let st1 = get_state();;
set_state st1;;

e(SUBGOAL_THEN
    "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
   STRIP_ASSUME_TAC);;

  e(USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11));;

   e(REWRITE_TAC[LENGTH_MAP]);;

   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;

e(SUBGOAL_THEN
    "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r ss"
   ASSUME_TAC);;

 e(REWRITE_TAC[WFcsubst]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss")));;
 e(ASM_REWRITE_TAC[GVc_subst]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
 e(RES_TAC);;

let st1 = get_state();;
set_state st1;;
e(SUBGOAL_THEN
    "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_vsubst));;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                    DL_variants;DISJOINT_variants_UNION]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[DL_variants]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants_UNION]);;

 e(MATCH_MP_TAC Ck_EXTENSION);;
 e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)");;
 e(REPEAT CONJ_TAC);;  % 5 subgoals %

  e(IMP_RES_THEN REWRITE_THM WFc_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:var list");;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(ASM_REWRITE_TAC[GVc_subst]);;

   e(REWRITE_TAC[SL_MAP]);;
   e(MATCH_MP_TAC IMAGE_SUBSET);;
   e(MATCH_MP_TAC (SPEC_ALL FVc_vsubst));;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;
                     DL_variants;DISJOINT_variants_UNION]);;

  e(REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[I_o_ID]);;
  e(REWRITE_TAC[MAP_EXTENSION]);;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)");;

   e(IMP_RES2_THEN ASSUME_TAC IN_DISJOINT_LIST);;
   e(SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC);;

    e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
    e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
    e(STRIP_TAC);;
    e(UNDISCH_LAST_TAC);;
    e(POP_ASSUM (REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM);;
    e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
    e(RES_THEN REWRITE_THM);;

   e(ASM_CASES_TAC "(x:var) IN SL vals'");;
let st5 = get_state();;
set_state(st5);;

    e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE));;
    e(IMP_RES_TAC ES_LENGTH_THM);;
    e(ASM_REWRITE_TAC[LENGTH_variants]);;

    e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM c_substk_biject);;

   e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

   e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
   e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
   e(REWRITE_TAC[I_o_ID]);;
   e(CHECK_TAC);;
>%



let INTERPRET_SUBSTITUTED_CALL_STAGE =
 prove_thm
  (`INTERPRET_SUBSTITUTED_CALL_STAGE`,
   "!r k p vars vals glbs pre post calls rec c g xs vals' s1 s4 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns.
         (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
          Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                 <|c ss)
             r k
             (BNDsL ns (MAP ss vals') s2)
             s3 /\
          ES es s1 ns s2 ",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
                "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN STRIP_ASSUME_TAC (MATCH_MP INTERPRET_CALL_STAGE
      (LIST_CONJ (map ASSUME
        ["(r:env) p = vars,vals,glbs,pre,post,calls,rec,c";
         "vals' = variants vals(SL(APPEND xs glbs))";
         "Ck(CALL p xs es)r(k + 1)s1 s4"])))
   THEN EXISTS_TAC "s2:state"
   THEN EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                               (MAP ss (APPEND (APPEND xs vals') glbs)))
                          (MAP ss (APPEND (APPEND xs vals') glbs))
                          (BNDsL ns (MAP ss vals') s2)"
   THEN EXISTS_TAC "ns:(num)list"
   THEN POP_ASSUM (\th. REWRITE_THM th THEN ASSUME_TAC th)
   THEN SUBGOAL_THEN
             "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
           STRIP_ASSUME_TAC
        THENL
         [ USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11)
           THENL
            [ REWRITE_TAC[LENGTH_MAP];

              ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
            ];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                       r ss"
           ASSUME_TAC
        THENL
         [ REWRITE_TAC[WFcsubst]
           THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss"))
           THEN ASM_REWRITE_TAC[GVc_subst]
           THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
           THEN REPEAT STRIP_TAC
           THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
           THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
           THEN RES_TAC;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                  (calls <|g ((APPEND xs vals') //v (APPEND vars vals))) r"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL WFc_proc_subst)
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r SUBSET
              (SL (APPEND (APPEND xs vals') glbs))"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC (SPEC_ALL FVc_proc_subst)
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN
             "Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
                  r k s1 s2 =
              Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                  r k (s1 o ss) (s2 o ss)"
           ASSUME_TAC
        THENL
         [ MATCH_MP_TAC c_substk_biject
           THEN EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND vars vals)glbs: var list)"
           ASSUME_TAC
        THENL
         [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX;

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[DL_APPEND]
           THEN ASM_REWRITE_TAC[DL_variants]
           THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
           THEN REWRITE_TAC[DISJOINT_variants];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)(MAP (ss:vsubst) vals'))"
           ASSUME_TAC
        THENL
         [ ONCE_REWRITE_TAC[(SYM o ASSUME)
                  "MAP ss(APPEND xs glbs) = APPEND xs glbs :var list"]
           THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]
           THEN IMP_RES_THEN REWRITE_THM DL_MAP_ONE_ONE
           THEN ASM_REWRITE_TAC[];

           ALL_TAC
         ]
   THEN SUBGOAL_THEN "MAP (ss:vsubst) (APPEND(APPEND xs vals') glbs) =
                      (APPEND(APPEND xs (MAP ss vals')) glbs)"
           ASSUME_TAC
        THENL
         [ ASM_REWRITE_TAC[MAP_APPEND];

           ALL_TAC
         ]
   THEN CONJ_TAC
   THENL
    [ REWRITE_TAC[ASSUME "(s4:state) = BNDsL(MAP s2 vals')vals' s3"]
      THEN EXT_TAC "z:var"
      THEN GEN_TAC
      THEN ASM_CASES_TAC "(z:var) IN SL vals'"
      THENL
       [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
         THEN ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH;

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME)
                             "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH
            THENL
             [ REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
             ]
          ];

         IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
         THEN ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')"
         THENL
          [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
            THEN USE_IMP_THEN REWRITE_THM
                              (SPEC_ALL proc_subst_ONLY_IMPACT_STAGE)
            THEN ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
            THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3
            THEN REWRITE_TAC[(SYM o SPEC_ALL) SL_APPEND;
                    (SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN ASM_REWRITE_TAC[];

            IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
            THEN ASM_CASES_TAC "z IN SL(APPEND(APPEND xs (MAP(ss:vsubst)vals'))glbs)"
            THENL
             [ IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH
               THEN SUBGOAL_THEN "(\y. @z. (ss:vsubst)z = y)z = z"
                              (\th. REWRITE_TAC[o_THM;th])
               THEN IMP_RES_THEN (ONCE_REWRITE_THM o SYM o SPEC_ALL)
                                 ONE_ONE
               THEN BETA_TAC
               THEN IMP_RES_THEN REWRITE_THM ONTO_INVERSE
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN POP_ASSUM (STRIP_ASSUME_TAC o
                           REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM])
               THENL
                [ RES_THEN REWRITE_THM;

                  RES_THEN CONTR_TAC;

                  RES_THEN REWRITE_THM
                ];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN USE_IMP_THEN REWRITE_THM
                         (SPEC_ALL proc_subst_ONLY_IMPACT_STAGE)
               THEN UNASSUME_THEN (ASSUME_TAC o SYM)
                         "vals' = variants vals(SL(APPEND xs glbs))"
               THEN ASM_REWRITE_TAC[]
               THEN REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ]
       ];

      MATCH_MP_TAC Ck_EXTENSION
      THEN EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss"
      THEN EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)"
      THEN REPEAT CONJ_TAC
      THENL  % 5 subgoals %
       [ IMP_RES_THEN REWRITE_THM WFc_subst
         THEN ASM_REWRITE_TAC[];

         ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM FVc_subst
         THENL
          [ EXISTS_TAC "glbs:var list"
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN ASM_REWRITE_TAC[GVc_subst];

            REWRITE_TAC[SL_MAP]
            THEN MATCH_MP_TAC IMAGE_SUBSET
            THEN ASM_REWRITE_TAC[]
          ];

         REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]
         THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
         THEN REWRITE_TAC[I_o_ID]
         THEN REWRITE_TAC[MAP_EXTENSION]
         THEN REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]
         THEN GEN_TAC
         THEN ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)"
         THENL
          [ IMP_RES_TAC IN_DISJOINT_LIST
            THEN SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC
            THENL
             [ REWRITE_TAC[SL_MAP;IN_IMAGE]
               THEN IMP_RES_THEN REWRITE_THM ONE_ONE
               THEN STRIP_TAC
               THEN UNDISCH_LAST_TAC
               THEN POP_ASSUM (REWRITE_THM o SYM)
               THEN ASM_REWRITE_TAC[];

               IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH
               THEN ASM_REWRITE_TAC[]
               THEN SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM
               THEN REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]
               THEN RES_THEN REWRITE_THM
             ];

            ASM_CASES_TAC "(x:var) IN SL vals'"
            THENL
             [ USE_IMP_THEN REWRITE_THM
                    (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE)
               THEN ASM_REWRITE_TAC[]
               THEN IMP_RES_TAC ES_LENGTH_THM
               THEN ASM_REWRITE_TAC[LENGTH_variants];

               REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]
               THEN ASM_REWRITE_TAC[]
             ]
          ];

         USE_IMP_THEN REWRITE_THM c_substk_biject
         THENL
          [ EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)"
            THEN ASM_REWRITE_TAC[];

            REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]
            THEN IMP_RES2_THEN REWRITE_THM o_INVERSE
            THEN REWRITE_TAC[I_o_ID]
            THEN CHECK_TAC
          ]
       ]
    ]
  );;

%< INTERPRET_SUBSTITUTED_CALL_STAGE
g "!r k p vars vals glbs pre post calls rec c g xs vals' s1 s4 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL (APPEND xs glbs))) /\
     WFsubst ss /\
     ONE_ONE ss /\ ONTO ss /\
     (MAP ss (APPEND xs glbs) = (APPEND xs glbs)) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns.
         (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
          Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                 <|c ss)
             r k
             (BNDsL ns (MAP ss vals') s2)
             s3 /\
          ES es s1 ns s2 ";;
e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(UNASSUME_THEN (\th. REWRITE_ALL_THM th THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(STRIP_ASSUME_TAC (MATCH_MP INTERPRET_CALL_STAGE
      (LIST_CONJ (map ASSUME
        ["(r:env) p = vars,vals,glbs,pre,post,calls,rec,c";
         "vals' = variants vals(SL(APPEND xs glbs))";
         "Ck(CALL p xs es)r(k + 1)s1 s4"]))));;
e(EXISTS_TAC "s2:state");;
e(EXISTS_TAC "BNDsL (MAP ((s3:state) o (\y. @z. (ss:vsubst) z = y))
                         (MAP ss (APPEND (APPEND xs vals') glbs)))
                    (MAP ss (APPEND (APPEND xs vals') glbs))
                    (BNDsL ns (MAP ss vals') s2)");;
e(EXISTS_TAC "ns:(num)list");;
e(POP_ASSUM (\th. REWRITE_THM th THEN ASSUME_TAC th));;
let st1 = get_state();;
set_state st1;;

e(SUBGOAL_THEN
    "(MAP (ss:vsubst) xs = xs) /\ (MAP ss glbs = glbs)"
   STRIP_ASSUME_TAC);;

  e(USE_IMP_THEN REWRITE_THM (ONCE_REWRITE_RULE[EQ_SYM_EQ] APPEND_11));;

   e(REWRITE_TAC[LENGTH_MAP]);;

   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;

e(SUBGOAL_THEN
    "WFcsubst (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r ss"
   ASSUME_TAC);;

 e(REWRITE_TAC[WFcsubst]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFsubst] (ASSUME "WFsubst ss")));;
 e(ASM_REWRITE_TAC[GVc_subst]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
 e(RES_TAC);;

e(SUBGOAL_THEN
    "WFc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL WFc_proc_subst));;
  e(ASM_REWRITE_TAC[]);;

e(SUBGOAL_THEN
    "FVc (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r SUBSET
     (SL (APPEND (APPEND xs vals') glbs))"
   ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL FVc_proc_subst));;
  e(ASM_REWRITE_TAC[]);;

e(SUBGOAL_THEN
    "Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals))) <|c ss)
         r k s1 s2 =
     Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
         r k (s1 o ss) (s2 o ss)"
   ASSUME_TAC);;

 e(MATCH_MP_TAC c_substk_biject);;
 e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
 e(ASM_REWRITE_TAC[]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND vars vals)glbs: var list)"
      ASSUME_TAC);;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)vals' :var list)"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[DL_APPEND]);;
   e(ASM_REWRITE_TAC[DL_variants]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(REWRITE_TAC[DISJOINT_variants]);;

 e(SUBGOAL_THEN "DL(APPEND(APPEND xs glbs)(MAP (ss:vsubst) vals'))"
      ASSUME_TAC);;

   e(ONCE_REWRITE_TAC[(SYM o ASSUME)
            "MAP ss(APPEND xs glbs) = APPEND xs glbs :var list"]);;
   e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_APPEND]);;
   e(IMP_RES_THEN REWRITE_THM DL_MAP_ONE_ONE);;
   e(ASM_REWRITE_TAC[]);;

 e(SUBGOAL_THEN "MAP (ss:vsubst) (APPEND(APPEND xs vals') glbs) =
                 (APPEND(APPEND xs (MAP ss vals')) glbs)"
      ASSUME_TAC);;

   e(ASM_REWRITE_TAC[MAP_APPEND]);;

let st2 = get_state();;
set_state st2;;
e(CONJ_TAC);;

 e(REWRITE_TAC[ASSUME "(s4:state) = BNDsL(MAP s2 vals')vals' s3"]);;
  e(EXT_TAC "z:var");;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(z:var) IN SL vals'");;

   e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
   e(ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')");;

    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(USE_IMP_THEN REWRITE_THM BNDsL_NOT_MATCH);;

     e(REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
     e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;
     e(ASM_REWRITE_TAC[]);;

     e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;

   e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
   e(ASM_CASES_TAC "(z:var) IN SL (MAP (ss:vsubst) vals')");;

    e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
    e(USE_IMP_THEN REWRITE_THM (SPEC_ALL proc_subst_ONLY_IMPACT_STAGE));;
    e(ASM_REWRITE_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST3);;
    e(REWRITE_TAC[(SYM o SPEC_ALL) SL_APPEND;
             (SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
let st3 = get_state();;
set_state st3;;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
    e(ASM_CASES_TAC "z IN SL(APPEND(APPEND xs (MAP(ss:vsubst)vals'))glbs)");;

     e(IMP_RES_THEN REWRITE_THM BNDsL_MAP_MATCH);;
     e(SUBGOAL_THEN "(\y. @z. (ss:vsubst)z = y)z = z"
             (\th. REWRITE_TAC[o_THM;th]));;
     e(IMP_RES_THEN (ONCE_REWRITE_THM o pthm o SYM o SPEC_ALL)
         ONE_ONE);;
     e(BETA_TAC);;
     e(IMP_RES_THEN REWRITE_THM ONTO_INVERSE);;
     e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
     e(POP_ASSUM (STRIP_ASSUME_TAC o
               REWRITE_RULE[SL_APPEND;IN_UNION;DE_MORGAN_THM]));;

      e(RES_THEN REWRITE_THM);;

      e(RES_THEN CONTR_TAC);;

      e(RES_THEN REWRITE_THM);;

     e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
     e(USE_IMP_THEN REWRITE_THM (SPEC_ALL proc_subst_ONLY_IMPACT_STAGE));;
     e(UNASSUME_THEN (ASSUME_TAC o SYM)
                "vals' = variants vals(SL(APPEND xs glbs))");;
     e(ASM_REWRITE_TAC[]);;
     e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
     e(ASM_REWRITE_TAC[]);;

let st4 = get_state();;
set_state(st4);;

 e(MATCH_MP_TAC Ck_EXTENSION);;
 e(EXISTS_TAC "(calls <|g (APPEND xs vals' //v APPEND vars vals)) <|g ss");;
 e(EXISTS_TAC "(BNDsL ns vals' (s2:state)) o (\y. @z. (ss:vsubst)z = y)");;
 e(REPEAT CONJ_TAC);;  % 5 subgoals %

  e(IMP_RES_THEN REWRITE_THM WFc_subst);;
  e(ASM_REWRITE_TAC[]);;

  e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM FVc_subst);;

   e(EXISTS_TAC "glbs:var list");;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(ASM_REWRITE_TAC[GVc_subst]);;

   e(REWRITE_TAC[SL_MAP]);;
   e(MATCH_MP_TAC IMAGE_SUBSET);;
   e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[MAP_MAP_o;(SYM o SPEC_ALL) o_ASSOC]);;
  e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
  e(REWRITE_TAC[I_o_ID]);;
  e(REWRITE_TAC[MAP_EXTENSION]);;
  e(REWRITE_TAC[SL_APPEND;IN_UNION;o_THM]);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "(x:var) IN SL(APPEND xs glbs)");;

   e(IMP_RES_TAC IN_DISJOINT_LIST);;
   e(SUBGOAL_THEN "~(((ss:vsubst)x) IN (SL (MAP ss vals')))" ASSUME_TAC);;

    e(REWRITE_TAC[SL_MAP;IN_IMAGE]);;
    e(IMP_RES_THEN REWRITE_THM ONE_ONE);;
    e(STRIP_TAC);;
    e(UNDISCH_LAST_TAC);;
    e(POP_ASSUM (REWRITE_THM o SYM));;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_THEN REWRITE_THM BNDsL_NOT_MATCH);;
    e(ASM_REWRITE_TAC[]);;
    e(SUBGOAL_THEN "(ss:vsubst)x = x" REWRITE_THM);;
    e(REWRITE_ALL_TAC[MAP_IDENT_ELEMENTS]);;
    e(RES_THEN REWRITE_THM);;

   e(ASM_CASES_TAC "(x:var) IN SL vals'");;
let st5 = get_state();;
set_state(st5);;

    e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] BNDsL_MAP_ONE_ONE));;
    e(ASM_REWRITE_TAC[]);;
    e(IMP_RES_TAC ES_LENGTH_THM);;
    e(ASM_REWRITE_TAC[LENGTH_variants]);;

    e(REWRITE_ALL_TAC[SL_APPEND;IN_UNION;DE_MORGAN_THM]);;
    e(ASM_REWRITE_TAC[]);;

  e(USE_IMP_THEN REWRITE_THM c_substk_biject);;

   e(EXISTS_TAC "calls <|g (APPEND xs vals' //v APPEND vars vals)");;
   e(ASM_REWRITE_TAC[]);;

   e(REWRITE_TAC[(SYM o SPEC_ALL) o_ASSOC]);;
   e(IMP_RES2_THEN REWRITE_THM o_INVERSE);;
   e(REWRITE_TAC[I_o_ID]);;
   e(CHECK_TAC);;
>%




let Ck_variants_SUBSTITUTION_BODY =
 prove_thm
  (`Ck_variants_SUBSTITUTION_BODY`,
   "!vals' vals'' r k p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'  (logicals vals' ) //v
            APPEND vals'' (logicals vals'')     )) /\
     Ck (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
               r k (BNDsL ns vals' s2) s3' ",

   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "MAP (ss:vsubst) vals'' = vals'" (REWRITE_THM o SYM)
   THENL
    [ FIRST_ASSUM (REWRITE_THM o SYM o SYM)
      THEN USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst
      THENL
       [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];

         USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
       ];

      ALL_TAC
    ]
   THEN MATCH_MP_TAC (SPEC_ALL Ck_SUBSTITUTED_BODY)
   THEN FIRST_ASSUM (REWRITE_THM o SYM o SYM)
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THENL
    [ MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants];

      MATCH_MP_TAC IDENT_MAP_vsubst
      THEN REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]
      THEN REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN STRIP_ASSUME_TAC
              (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
      THEN CONJ_TAC
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]
    ]
  );;

%< Ck_variants_SUBSTITUTION_BODY

The idea is, use
  ss = ((vals' & (logicals vals')) //v (vals'' & (logicals vals''))) !

g "!vals' vals'' r k p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3 ss.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'  (logicals vals' ) //v
            APPEND vals'' (logicals vals'')     )) /\
     Ck (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
               r k (BNDsL ns vals' s2) s3' ";;

e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN "MAP (ss:vsubst) vals'' = vals'" (REWRITE_THM o SYM));;

  e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
  e(USE_IMP_THEN REWRITE_THM MAP_NULL_APPEND_vsubst);;

    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

e(MATCH_MP_TAC (SPEC_ALL Ck_SUBSTITUTED_BODY));;
e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;

 e(MATCH_MP_TAC IDENT_MAP_vsubst);;
 e(REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]);;
 e(REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

>%





let INTERPRET_variants_SUBSTITUTION_CALL_STAGE =
 prove_thm
  (`INTERPRET_variants_SUBSTITUTION_CALL_STAGE`,
   "!r k g p vars vals glbs pre post calls rec c vals' vals'' q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     )) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
                 Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                        <|c ss)
                    r k (BNDsL ns (MAP ss vals') s2) s3 /\
                 ES es s1 ns s2 ",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC (SPEC_ALL INTERPRET_SUBSTITUTED_CALL_STAGE)
   THEN FIRST_ASSUM (REWRITE_THM o SYM o SYM)
   THEN REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]
   THEN ASM_REWRITE_TAC[]
   THEN CONJ_TAC
   THENL
    [ MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants];

      MATCH_MP_TAC IDENT_MAP_vsubst
      THEN REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]
      THEN REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN STRIP_ASSUME_TAC
              (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
      THEN CONJ_TAC
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]
    ]
  );;

%< INTERPRET_variants_SUBSTITUTION_CALL_STAGE

The idea is, use
  ss = ((vals'' & (logicals vals'')) //v (vals' & (logicals vals'))) !

g "!r k g p vars vals glbs pre post calls rec c vals' vals'' q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     )) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 (MAP ss vals')) (MAP ss vals') s3) /\
                 Ck ((c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                        <|c ss)
                    r k (BNDsL ns (MAP ss vals') s2) s3 /\
                 ES es s1 ns s2 ";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC (SPEC_ALL INTERPRET_SUBSTITUTED_CALL_STAGE));;
e(FIRST_ASSUM (REWRITE_THM o SYM o SYM));;
e(REWRITE_TAC[vsubst_ONE_ONE;vsubst_ONTO]);;
e(ASM_REWRITE_TAC[]);;
e(CONJ_TAC);;

 e(MATCH_MP_TAC STANDARD_SUBST_IS_WFsubst);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;

 e(MATCH_MP_TAC IDENT_MAP_vsubst);;
 e(REWRITE_TAC[CONJUNCT1(SPEC_ALL DISJOINT_APPEND)]);;
 e(REWRITE_TAC[DISJOINT_variants;DISJOINT_variants_UNION]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
 e(CONJ_TAC);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[WFxs_APPEND;NOT_WFxs_logicals]);;

>%




let Ck_SUBSTITUTION_EQ =
 prove_thm
  (`Ck_SUBSTITUTION_EQ`,
   "!vals' vals'' r k p vars vals glbs pre post calls rec c g xs es s1 s2.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     WFxs vals' /\ WFxs vals'' /\
     (LENGTH vals' = LENGTH vals) /\ (LENGTH vals'' = LENGTH vals) /\
     DL vals' /\ DL vals'' /\
     DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
     DISJOINT (SL vals'') (SL (APPEND xs glbs)) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (Ck ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r k s1 s2  =
      Ck  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k s1 s2)",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]
   THEN MATCH_MP_TAC c_substk_EQ
   THEN STRIP_ASSUME_TAC
           (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
            "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN SUBGOAL_THEN "WFc c calls r /\ WFxs glbs /\
                      DL (APPEND (APPEND vars vals) glbs) /\
                      DL (APPEND (APPEND xs vals' ) glbs) /\
                      DL (APPEND (APPEND xs vals'') glbs) /\
                      FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
           STRIP_ASSUME_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[]
      THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN ASM_REWRITE_TAC[];

      REPEAT STRIP_TAC
      THEN IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
      THEN REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]
      THEN UNDISCH_LAST_TAC
      THEN SPEC_TAC("x:var","x:var")
      THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]
      THEN REWRITE_TAC[MAP_o;o_THM]
      THEN REWRITE_TAC[MAP_APPEND]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN REPEAT CONJ_TAC
      THENL  % 3 subgoals %
       [ USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN MATCH_MP_TAC IDENT_MAP_vsubst
         THEN ASM_REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals;

         USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN ASM_REWRITE_TAC[]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
         THEN ASM_REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals
       ]
    ]
  );;

%< Ck_SUBSTITUTION_EQ

g "!vals' vals'' r k p vars vals glbs pre post calls rec c g xs es s1 s2.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     WFxs vals' /\ WFxs vals'' /\
     (LENGTH vals' = LENGTH vals) /\ (LENGTH vals'' = LENGTH vals) /\
     DL vals' /\ DL vals'' /\
     DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
     DISJOINT (SL vals'') (SL (APPEND xs glbs)) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (Ck ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r k s1 s2  =
      Ck  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k s1 s2)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]);;
e(MATCH_MP_TAC c_substk_EQ);;
e(STRIP_ASSUME_TAC
     (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
       "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(SUBGOAL_THEN "WFc c calls r /\ WFxs glbs /\
                DL (APPEND (APPEND vars vals) glbs) /\
                DL (APPEND (APPEND xs vals' ) glbs) /\
                DL (APPEND (APPEND xs vals'') glbs) /\
                FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
     STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]);;
 e(UNDISCH_LAST_TAC);;
 e(SPEC_TAC("x:var","x:var"));;
 e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]);;
 e(REWRITE_TAC[MAP_o;o_THM]);;
 e(REWRITE_TAC[MAP_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM APPEND_11);;

  e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

  e(USE_IMP_THEN REWRITE_THM APPEND_11);;

   e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

   e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
   e(REPEAT CONJ_TAC);;  % 3 subgoals %

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
            (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

      e(ASM_REWRITE_TAC[SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[]);;

       e(ASM_REWRITE_TAC[]);;
       e(MATCH_MP_TAC IDENT_MAP_vsubst);;
       e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;
       e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
       e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

      e(ASM_REWRITE_TAC[SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[]);;

       e(USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

        e(ASM_REWRITE_TAC[]);;

        e(ASM_REWRITE_TAC[]);;
        e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

         e(ASM_REWRITE_TAC[SUBSET_REFL]);;

         e(USE_IMP_THEN REWRITE_THM
                (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
         e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

     e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

     e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

      e(ASM_REWRITE_TAC[]);;
      e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;
      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;
      e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
      e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;

>%



%<
let variants_SUBSTITUTION_EQ_STAGE =
 prove_thm
  (`variants_SUBSTITUTION_EQ_STAGE`,
   "!r k p vars vals glbs pre post calls rec c vals' vals'' q xs es s1 s2.
     WFenvk r k /\
     WFc (CALL p xs es) r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (Ck ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r k s1 s2  =
      Ck  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k s1 s2)",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]
   THEN MATCH_MP_TAC c_substk_EQ
   THEN STRIP_ASSUME_TAC
           (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
            "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN SUBGOAL_THEN "WFc c r /\ WFxs glbs /\
                      DL (APPEND (APPEND vars vals) glbs) /\
                      DL (APPEND (APPEND xs vals' ) glbs) /\
                      DL (APPEND (APPEND xs vals'') glbs) /\
                      FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
           STRIP_ASSUME_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[]
      THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;DISJOINT_variants_APPEND]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION;DISJOINT_variants_UNION3];

      REWRITE_TAC[ASSUME "WFenvk r k";ASSUME "WFc c r"]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC SUBSET_DEF
      THEN REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]
      THEN UNDISCH_LAST_TAC
      THEN POP_TAC
      THEN SPEC_TAC("x:var","x:var")
      THEN REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]
      THEN REWRITE_TAC[MAP_o;o_THM]
      THEN REWRITE_TAC[MAP_APPEND]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN USE_IMP_THEN REWRITE_THM APPEND_11
           THENL [ REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]; ALL_TAC ]
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN REPEAT CONJ_TAC
      THENL  % 3 subgoals %
       [ USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
         THEN REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                        APPEND vals'  (logicals vals' )     )"]
         THEN MATCH_MP_TAC IDENT_MAP_vsubst
         THEN REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals
         THEN ASM_REWRITE_TAC[];

         USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[LENGTH_variants;SUBSET_REFL]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2)
              THENL [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
              THENL [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants]; ALL_TAC ]
         THEN REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                        APPEND vals'  (logicals vals' )     )"]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
              THENL [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];
                      ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM
                   (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
         THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants];

         USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
              THENL [ ASM_REWRITE_TAC[DISJOINT_APPEND]; ALL_TAC ]
         THEN REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                        APPEND vals'  (logicals vals' )     )"]
         THEN USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst
         THEN ASM_REWRITE_TAC[DISJOINT_APPEND]
         THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
         THEN IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals
       ]
    ]
  );;
>%

%< variants_SUBSTITUTION_EQ_STAGE

g "!r k p vars vals glbs pre post calls rec c vals' vals'' q xs es s1 s2.
     WFenvk r k /\
     WFc (CALL p xs es) r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals (SL(APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
     (ss = (APPEND vals'' (logicals vals'') //v
            APPEND vals'  (logicals vals' )     ))   ==>
     (Ck ((c <|c ((APPEND xs vals')  //v (APPEND vars vals))) <|c ss)
         r k s1 s2  =
      Ck  (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k s1 s2)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[(SYM o SPEC_ALL) c_subst_o]);;
e(MATCH_MP_TAC c_substk_EQ);;
e(STRIP_ASSUME_TAC
     (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
       "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(SUBGOAL_THEN "WFc c r /\ WFxs glbs /\
                DL (APPEND (APPEND vars vals) glbs) /\
                DL (APPEND (APPEND xs vals' ) glbs) /\
                DL (APPEND (APPEND xs vals'') glbs) /\
                FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs))"
     STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants;DISJOINT_variants_APPEND]);;
 e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
 e(REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION;DISJOINT_variants_UNION3]);;

 e(REWRITE_TAC[ASSUME "WFenvk r k";ASSUME "WFc c r"]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC SUBSET_DEF);;
 e(REWRITE_TAC[(SYM o SPEC_ALL o ISPEC "ss:vsubst") o_THM]);;
 e(UNDISCH_LAST_TAC);;
 e(POP_TAC);;
 e(SPEC_TAC("x:var","x:var"));;
 e(REWRITE_TAC[(SYM o SPEC_ALL) MAP_EXTENSION]);;
 e(REWRITE_TAC[MAP_o;o_THM]);;
 e(REWRITE_TAC[MAP_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM APPEND_11);;

  e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

  e(USE_IMP_THEN REWRITE_THM APPEND_11);;

   e(REWRITE_TAC[LENGTH_APPEND;LENGTH_MAP]);;

   e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
   e(REPEAT CONJ_TAC);;  % 3 subgoals %

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

     e(ASM_REWRITE_TAC[SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
            (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

      e(ASM_REWRITE_TAC[SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[]);;

       e(REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                   APPEND vals'  (logicals vals' )     )"]);;
       e(MATCH_MP_TAC IDENT_MAP_vsubst);;
       e(REWRITE_TAC[DISJOINT_APPEND]);;
       e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
       e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;
       e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

     e(ASM_REWRITE_TAC[LENGTH_variants;SUBSET_REFL]);;

     e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst2));;

      e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

      e(USE_IMP_THEN REWRITE_THM
             (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

       e(ASM_REWRITE_TAC[LENGTH_variants]);;

       e(USE_IMP_THEN REWRITE_THM
              (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;

        e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

        e(REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                    APPEND vals'  (logicals vals' )     )"]);;
        e(USE_IMP_THEN REWRITE_THM
               (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

         e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

         e(USE_IMP_THEN REWRITE_THM
                (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
         e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

    e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

     e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

     e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;

      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;

      e(REWRITE_TAC[ASSUME "ss = (APPEND vals'' (logicals vals'') //v
                                  APPEND vals'  (logicals vals' )     )"]);;
      e(USE_IMP_THEN REWRITE_THM IDENT_MAP_vsubst);;
      e(ASM_REWRITE_TAC[DISJOINT_APPEND]);;
      e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
      e(IMP_RES_THEN REWRITE_THM WFxs_NOT_logicals);;


>%



let Ck_greater_variants_BODY =
 prove_thm
  (`Ck_greater_variants_BODY`,
   "!vals' vals'' r k p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals'  = variants vals (SL (APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     Ck (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
               r k (BNDsL ns vals' s2) s3'",

   REPEAT STRIP_TAC
   THEN DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                              (APPEND vals'' (logicals vals''))    )"
   THEN SUBGOAL_THEN
          "?s3'.
              Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                  r k (BNDsL ns vals' s2) s3'"
           ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPEC_ALL Ck_variants_SUBSTITUTION_BODY)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN SUBGOAL_THEN
          "!s1 s2.
            Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                r k s1 s2 =
            Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k s1 s2"
          REWRITE_THM
   THEN REPEAT GEN_TAC
   THEN MATCH_MP_TAC (GENL["vals':var list";"vals'':var list"]
                         (SPEC_ALL Ck_SUBSTITUTION_EQ))
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND]
   THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                        DISJOINT_variants;DISJOINT_variants_UNION]
  );;

%< Ck_greater_variants_BODY

g "!vals' vals'' r k p vars vals glbs pre post calls rec c g q xs es ns s1 s2 s3.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals'  = variants vals (SL (APPEND xs glbs))) /\
     (vals'' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     Ck (c <|c ((APPEND xs vals'') //v (APPEND vars vals)))
         r k (BNDsL ns vals'' s2) s3 /\
     ES es s1 ns s2  ==>
     ?s3'. Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
               r k (BNDsL ns vals' s2) s3'";;

e(REPEAT STRIP_TAC);;
e(DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                        (APPEND vals'' (logicals vals''))    )");;
e(SUBGOAL_THEN
     "?s3'.
         Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
             r k (BNDsL ns vals' s2) s3'"
      ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL Ck_variants_SUBSTITUTION_BODY));;
  e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(SUBGOAL_THEN
     "!s1 s2.
       Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
           r k s1 s2 =
       Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k s1 s2"
     REWRITE_THM);;
  e(REPEAT GEN_TAC);;
  e(MATCH_MP_TAC (GENL["vals':var list";"vals'':var list"]
                    (SPEC_ALL Ck_SUBSTITUTION_EQ)));;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                    DISJOINT_variants;DISJOINT_variants_UNION]);;

>%



let INTERPRET_greater_variants_CALL_STAGE =
 prove_thm
  (`INTERPRET_greater_variants_CALL_STAGE`,
   "!vals' r k g p vars vals glbs pre post calls rec c q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                     r k (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ",

   REPEAT STRIP_TAC
   THEN DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))"
   THEN DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                              (APPEND vals'' (logicals vals''))    )"
   THEN SUBGOAL_THEN
          "?s2 s3 ns.
              (s4 = BNDsL(MAP s2(MAP ss vals''))(MAP ss vals'')s3) /\
              Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                   r k (BNDsL ns (MAP ss vals'') s2) s3 /\
              ES es s1 ns s2"
           ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPECL["vals'':var list";"vals':var list"]
                      (GENL["vals':var list";"vals'':var list"]
                         (SPEC_ALL INTERPRET_variants_SUBSTITUTION_CALL_STAGE)))
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN SUBGOAL_THEN
          "!s1 s2.
            Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
                r k s1 s2 =
            Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k s1 s2"
          REWRITE_THM
   THENL
    [ REPEAT GEN_TAC
      THEN MATCH_MP_TAC (SPEC_ALL Ck_SUBSTITUTION_EQ)
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                           DISJOINT_variants;DISJOINT_variants_UNION];

      ALL_TAC
    ]
   THEN SUBGOAL_THEN  "MAP (ss:vsubst) vals'' = vals'"
          REWRITE_THM
   THEN POP_ASSUM REWRITE_THM
   THEN USE_IMP_THEN REWRITE_THM
            (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst)
   THENL
    [ ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL];

      MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT)
      THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants]
    ]
  );;

%< INTERPRET_greater_variants_CALL_STAGE

g "!vals' r k g p vars vals glbs pre post calls rec c q xs es s1 s4.
     WFenv_syntax r /\
     WFc (CALL p xs es) g r /\
     (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
     (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
     Ck (CALL p xs es) r (k+1) s1 s4  ==>
     ?s2 s3 ns. (s4 = BNDsL (MAP s2 vals') vals' s3) /\
                 Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals)))
                     r k (BNDsL ns vals' s2) s3 /\
                 ES es s1 ns s2 ";;

e(REPEAT STRIP_TAC);;
e(DEFINE_NEW_VAR "vals'' = variants vals (SL (APPEND xs glbs))");;
e(DEFINE_NEW_VAR "ss = ((APPEND vals'  (logicals vals' )) //v
                        (APPEND vals'' (logicals vals''))    )");;
e(SUBGOAL_THEN
     "?s2 s3 ns.
         (s4 = BNDsL(MAP s2(MAP ss vals''))(MAP ss vals'')s3) /\
         Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
            r k (BNDsL ns (MAP ss vals'') s2) s3 /\
         ES es s1 ns s2"
      ASSUME_TAC);;

  e(MATCH_MP_TAC (SPECL["vals'':var list";"vals':var list"]
                    (GENL["vals':var list";"vals'':var list"]
                       (SPEC_ALL INTERPRET_variants_SUBSTITUTION_CALL_STAGE))));;
  e(ASM_REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(SUBGOAL_THEN
     "!s1 s2.
       Ck ((c <|c ((APPEND xs vals'') //v (APPEND vars vals))) <|c ss)
          r k s1 s2 =
       Ck (c <|c ((APPEND xs vals') //v (APPEND vars vals))) r k s1 s2"
     REWRITE_THM);;

   e(REPEAT GEN_TAC);;
   e(MATCH_MP_TAC (SPEC_ALL Ck_SUBSTITUTION_EQ));;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_variants;DISJOINT_variants_UNION]);;

   e(SUBGOAL_THEN
         "MAP (ss:vsubst) vals'' = vals'"
       REWRITE_THM);;
   e(POP_ASSUM REWRITE_THM);;
   e(USE_IMP_THEN REWRITE_THM
           (REWRITE_RULE[AND_IMP_INTRO] MAP_NULL_APPEND_vsubst));;

    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants;SUBSET_REFL]);;

    e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vsubst_REPLACEMENT));;
    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

>%



let CHANGE_VAR_FOR_ADAPTATION_DISJOINT_STAGE =
 TAC_PROOF
  (([],
    "!xs vals' r k p vars vals glbs pre post calls rec c
      u v x y x0 y0 x1 y1 u1 v1.
      (WFenvk r k) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      DISJOINT(SL x0)(FVa q) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x1  = APPEND x x0) /\
      (y1  = APPEND y y0) /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v))

     %------------------------------------------------------------------% ==>
      WFc (c <|c (x1 //v y1)) (calls <|g (x1 //v y1)) r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      DISJOINT(SL x0)(FVa q) /\
      PCSPECk(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
                post <|a (x1 // y1),r,k)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
             (WFc c' g' r /\
              DL x' /\
              WFxs x' /\
              (GVc c' r) SUBSET (SL glbs) /\
              (FVc c' r) SUBSET (SL x') /\
              (FVa pre') SUBSET (SL x') /\
              (FVa post') SUBSET (SL(APPEND x' x0')) /\
              PCSPECk(AAND(mk_eqs x0' x')pre',c',post',r,k))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV))
   THENL
    [ MATCH_MP_TAC PROCEDURE_CHANGE_OF_VARIABLES_STAGE
      THEN EXISTS_TAC "p:string"
      THEN EXISTS_TAC "rec:aexp"
      THEN ASM_REWRITE_TAC[];

      ASM_REWRITE_TAC[]
    ]
  );;

%<
g "!xs vals' r k p vars vals glbs pre post calls rec c
      u v x y x0 y0 x1 y1 u1 v1.
      (WFenvk r k) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      DISJOINT(SL x0)(FVa q) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x1  = APPEND x x0) /\
      (y1  = APPEND y y0) /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v))

     %------------------------------------------------------------------% ==>
      WFc (c <|c (x1 //v y1)) (calls <|g (x1 //v y1)) r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      DISJOINT(SL x0)(FVa q) /\
      PCSPECk(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
                post <|a (x1 // y1),r,k)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
             (WFc c' g' r /\
              DL x' /\
              WFxs x' /\
              (GVc c' r) SUBSET (SL glbs) /\
              (FVc c' r) SUBSET (SL x') /\
              (FVa pre') SUBSET (SL x') /\
              (FVa post') SUBSET (SL(APPEND x' x0')) /\
              PCSPECk(AAND(mk_eqs x0' x')pre',c',post',r,k))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV)));;

    e(MATCH_MP_TAC PROCEDURE_CHANGE_OF_VARIABLES_STAGE);;
    e(EXISTS_TAC "p:string");;
    e(EXISTS_TAC "rec:aexp");;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;
>%


let CHANGE_VAR_FOR_ADAPTATION_STAGE =
 TAC_PROOF
  (([],
    "!xs vals' r k p vars vals glbs pre post calls rec c u v x y x0 y0 x1 y1.
      (WFenvk r k) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH xs = LENGTH vars) /\ (LENGTH vals' = LENGTH vals) /\
      WFxs xs /\ WFxs vals' /\
      (DL (APPEND xs vals')) /\ DISJOINT(SL(APPEND xs vals'))(SL glbs) /\
      (u  = APPEND xs vals') /\
      (v  = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)

     %------------------------------------------------------------------% ==>
      WFc(c <|c (x1 //v y1))(calls <|g (x1 //v y1))r /\
      WFenv_syntax r /\
      DL x /\
      WFxs x /\
      (FVc(c <|c (x1 //v y1))r) SUBSET (SL x) /\
      (FVa(pre <|a (x // y))) SUBSET (SL x) /\
      (FVa(post <|a (x1 // y1))) SUBSET (SL(APPEND x x0)) /\
      PCSPECk(AAND(mk_eqs x0 x)(pre <|a (x // y)),c <|c (x1 //v y1),
                post <|a (x1 // y1),r,k)"
  ),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN SUBGOAL_THEN
   "let x = APPEND(APPEND vars vals)glbs in
     let x0 = logicals x in
      let x' = APPEND(APPEND xs vals')glbs in
       let x0' = logicals x' in
        let ss = (APPEND x' x0') //v (APPEND x x0) in
         let c' = c <|c ss in
          let g' = calls <|g ss in
           let pre' = pre <|a (x' // x) in
            let post' = post <|a ((APPEND x' x0') // (APPEND x x0)) in
             (WFc c' g' r /\
              DL x' /\
              WFxs x' /\
              (GVc c' r) SUBSET (SL glbs) /\
              (FVc c' r) SUBSET (SL x') /\
              (FVa pre') SUBSET (SL x') /\
              (FVa post') SUBSET (SL(APPEND x' x0')) /\
              PCSPECk(AAND(mk_eqs x0' x')pre',c',post',r,k))"
        (STRIP_ASSUME_TAC o CONV_RULE (DEPTH_CONV let_CONV))
   THENL
    [ MATCH_MP_TAC (SPEC_ALL PROCEDURE_CHANGE_OF_VARIABLES_STAGE)
      THEN ASM_REWRITE_TAC[];

      ASM_REWRITE_TAC[]
    ]
  );;


let WFenv_IMP_DL_DISJOINT =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals(SL(APPEND xs glbs))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs (variants vals(SL(APPEND xs glbs)))) /\
      DISJOINT(SL(APPEND xs (variants vals(SL(APPEND xs glbs)))))(SL glbs)"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]
   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
   THEN REWRITE_TAC[DISJOINT_variants_APPEND]
  );;

%< WFenv_IMP_DL_DISJOINT
g "!r p vars vals glbs pre post calls rec c vals' xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals(SL(APPEND xs glbs))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs (variants vals(SL(APPEND xs glbs)))) /\
      DISJOINT(SL(APPEND xs (variants vals(SL(APPEND xs glbs)))))(SL glbs)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
e(ASM_REWRITE_TAC[DL_variants;DISJOINT_variants_APPEND]);;
e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
e(REWRITE_TAC[DISJOINT_variants_APPEND]);;
>%

let pre_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!vars vals xs vals' glbs u v x y.
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals' = LENGTH vals) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs)
     ==>
      ((x // y) = (u // v))"),
   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC var_list_IDENT_APPEND
   THEN ASM_REWRITE_TAC[LENGTH_APPEND]
  );;

%< pre_subst_simplify_LEMMA
g "!vars vals xs vals' glbs u v x y.
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals' = LENGTH vals) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs)
     ==>
      ((x // y) = (u // v))";;

e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(MATCH_MP_TAC var_list_IDENT_APPEND);;
e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
>%


let A_post_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A ((post <|a (x1 // y1)) <|a (x0' // x0)) s =
       A (post <|a (APPEND u x0' // APPEND v y0)) s)"),

   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN "(APPEND x x0' // APPEND y y0) = (APPEND u x0' // APPEND v y0)"
               (REWRITE_THM o SYM)
   THENL
    [ USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]
         THEN MATCH_MP_TAC WFxs_NOT_WFxs
         THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
         THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

         ALL_TAC
       ]
      THEN REWRITE_TAC[ASSUME "x = APPEND u glbs :var list";
                       ASSUME "y = APPEND v glbs :var list"]
      THEN REWRITE_TAC[APPEND_ASSOC]
      THEN USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals];
         ALL_TAC
       ]
      THEN USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE
      THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      ALL_TAC
    ]
   THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :var list";
                    ASSUME "y1 = APPEND y y0 :var list"]
   THEN MATCH_MP_TAC A_APPEND_var_list_TRANS
   THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals;DL_logicals]
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REPEAT CONJ_TAC
   THENL  % 4 subgoals %

    [ MATCH_MP_TAC WFxs_NOT_WFxs
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      MATCH_MP_TAC WFxs_NOT_WFxs
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[NOT_WFxs_logicals];

      (ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
               "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))"
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[DL_variants];

      REWRITE_TAC[SUBSET_DEF;IN_INTER]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC logicals_IS_NOT_WFx
      THEN IMP_RES2_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[IN_UNION] o
                  ONCE_REWRITE_RULE[SL_APPEND]) SUBSET_DEF
      THEN IMP_RES_TAC WFxs_IS_WFx
      THEN RES_TAC
    ]
  );;

%< A_post_subst_simplify_LEMMA
g "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A ((post <|a (x1 // y1)) <|a (x0' // x0)) s =
       A (post <|a (APPEND u x0' // APPEND v y0)) s)";;

e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN "(APPEND x x0' // APPEND y y0) = (APPEND u x0' // APPEND v y0)"
       (REWRITE_THM o SYM));;

 e(USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE);;

  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;
  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

  e(REWRITE_TAC[ASSUME "x = APPEND u glbs :var list";
                ASSUME "y = APPEND v glbs :var list"]);;
  e(REWRITE_TAC[APPEND_ASSOC]);;
  e(USE_IMP_THEN REWRITE_THM var_list_IDENT_APPEND);;

   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;

   e(USE_IMP_THEN REWRITE_THM APPEND_var_list_COMMUTE);;
   e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals]);;
   e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
   e(MATCH_MP_TAC WFxs_NOT_WFxs);;
   e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
   e(REWRITE_ALL_TAC[WFxs_APPEND]);;
   e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :var list";
              ASSUME "y1 = APPEND y y0 :var list"]);;
e(MATCH_MP_TAC A_APPEND_var_list_TRANS);;
e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;LENGTH_logicals;DL_logicals]);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REPEAT CONJ_TAC);;  % 4 subgoals %

 e(MATCH_MP_TAC WFxs_NOT_WFxs);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

 e(MATCH_MP_TAC WFxs_NOT_WFxs);;
 e(REWRITE_ALL_TAC[WFxs_APPEND]);;
 e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

 e((ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
            "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))");;
 e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[DL_variants]);;

 e(REWRITE_TAC[SUBSET_DEF;IN_INTER]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC logicals_IS_NOT_WFx);;
 e(IMP_RES2_THEN (STRIP_ASSUME_TAC o
                  REWRITE_RULE[IN_UNION] o
                  ONCE_REWRITE_RULE[SL_APPEND]) SUBSET_DEF);;
 e(IMP_RES_TAC WFxs_IS_WFx);;
 e(RES_TAC);;
>%





let post_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                    <|a (x // x0')))
          s =
       A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                    <|a (x // x0')))
          s)"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[A_DEF]
   THEN MATCH_MP_TAC CONJ_11
   THEN REWRITE_TAC[a_subst_THM]
   THEN REWRITE_TAC[A_mk_foralls]
   THEN REWRITE_TAC[A_DEF]
   THEN REWRITE_TAC[ASSUME "(x1:var list) = APPEND x x0";
                    ASSUME "(y1:var list) = APPEND y y0"]
   THEN FORALL_EQ_TAC
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL A_post_subst_simplify_LEMMA)
   THEN ASM_REWRITE_TAC[]
  );;

%< post_subst_simplify_LEMMA
g "!r p vars vals xs vals' glbs u v x y x0 x0' y0 x1 y1 pre post q c s.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT (SL vals') (SL (APPEND xs glbs)) /\
      (DL (APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)    /\
      (y0  = logicals y)    /\
      (x0' = variants x0 (FVa q))   /\
      (x1  = APPEND x x0)   /\
      (y1  = APPEND y y0)
     ==>
      (A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                    <|a (x // x0')))
          s =
       A (AAND (pre <|a (u // v))
               ((mk_foralls x
                       (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                    <|a (x // x0')))
          s)";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[A_DEF]);;
e(MATCH_MP_TAC CONJ_11);;
e(REWRITE_TAC[a_subst_THM]);;
e(REWRITE_TAC[A_mk_foralls]);;
e(REWRITE_TAC[A_DEF]);;
e(REWRITE_TAC[ASSUME "(x1:var list) = APPEND x x0";
              ASSUME "(y1:var list) = APPEND y y0"]);;
e(FORALL_EQ_TAC);;
e(USE_IMP_THEN REWRITE_THM (SPEC_ALL A_post_subst_simplify_LEMMA));;
e(ASM_REWRITE_TAC[]);;

. |- A
     ((post <|a ((APPEND x x0) // (APPEND y y0))) <|a (x0' // x0))
     (BNDsL n x(s <|s (x // x0'))) =
     A
     (post <|a ((APPEND u x0') // (APPEND v y0)))
     (BNDsL n x(s <|s (x // x0')))

>%



let command_subst_simplify_LEMMA =
 TAC_PROOF
  (([],
    "!r k p vars vals glbs pre post calls rec c vals' xs u v x y x0 y0 s1 s2.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      (DL (APPEND xs glbs)) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT(SL vals')(SL(APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)
   ==>
      (Ck (c <|c (APPEND x x0 //v APPEND y y0)) r k s1 s2 =
       Ck (c <|c (APPEND xs vals' //v APPEND vars vals)) r k s1 s2)"),

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC c_substk_EQ
   THEN REPEAT STRIP_TAC
   THEN SUBGOAL_THEN
     "WFc c calls r /\
      DL(APPEND (APPEND vars vals) glbs) /\
      DL(APPEND vars vals) /\
      DL(APPEND (APPEND xs vals') glbs) /\
      DL(APPEND xs vals') /\
      (FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs)))"
           STRIP_ASSUME_TAC
   THENL
    [ IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[]
      THEN (ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
                 "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))"
      THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
      THEN ASM_REWRITE_TAC[];

      IMP_RES2_THEN ASSUME_TAC SUBSET_DEF
      THEN USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
      THENL
       [ ASM_REWRITE_TAC[LENGTH_APPEND];

         ASM_REWRITE_TAC[]
         THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION]
                                          o ONCE_REWRITE_RULE[SL_APPEND])
         THENL   % 2 subgoals %
          [ USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst)
            THEN ASM_REWRITE_TAC[LENGTH_APPEND];

            USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2)
            THENL
             [ ASM_REWRITE_TAC[LENGTH_APPEND]
               THEN ONCE_REWRITE_ALL_TAC[DL_APPEND]
               THEN ASM_REWRITE_TAC[];

               REWRITE_TAC[IDENTITY_vsubst;I_THM]
               THEN USE_IMP_THEN REWRITE_THM IDENT_vsubst
               THEN IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST
             ]
          ]
       ]
    ]
  );;


%< 
g "!r k p vars vals glbs pre post calls rec c vals' xs u v x y x0 y0 s1 s2.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      WFxs xs /\
      (LENGTH vars = LENGTH xs) /\
      (DL (APPEND xs glbs)) /\
      WFxs vals' /\
      (LENGTH vals' = LENGTH vals) /\
      DL vals' /\
      DISJOINT(SL vals')(SL(APPEND xs glbs)) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)
   ==>
      (Ck (c <|c (APPEND x x0 //v APPEND y y0)) r k s1 s2 =
       Ck (c <|c (APPEND xs vals' //v APPEND vars vals)) r k s1 s2)";;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC c_substk_EQ);;
e(REPEAT STRIP_TAC);;
e(SUBGOAL_THEN
     "WFc c calls r /\
      DL(APPEND (APPEND vars vals) glbs) /\
      DL(APPEND vars vals) /\
      DL(APPEND (APPEND xs vals') glbs) /\
      DL(APPEND xs vals') /\
      (FVc c r SUBSET (SL (APPEND (APPEND vars vals) glbs)))"
   STRIP_ASSUME_TAC);;

 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
 e((ASSUME_TAC o ONCE_REWRITE_RULE[DISJOINT_SYM] o ASSUME)
            "DISJOINT(SL (vals': var list))(SL(APPEND xs glbs))");;
 e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
 e(ASM_REWRITE_TAC[]);;

 e(IMP_RES2_THEN ASSUME_TAC SUBSET_DEF);;
 e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;

  e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

  e(ASM_REWRITE_TAC[]);;
  e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[IN_UNION]
                                o ONCE_REWRITE_RULE[SL_APPEND]));;
   % 2 subgoals %

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst));;
   e(ASM_REWRITE_TAC[LENGTH_APPEND]);;

   e(USE_IMP_THEN REWRITE_THM (REWRITE_RULE[AND_IMP_INTRO] NULL_APPEND_vsubst2));;

    e(ASM_REWRITE_TAC[LENGTH_APPEND]);;
    e(ONCE_REWRITE_ALL_TAC[DL_APPEND]);;
    e(ASM_REWRITE_TAC[]);;

    e(REWRITE_TAC[IDENTITY_vsubst;I_THM]);;
    e(USE_IMP_THEN REWRITE_THM IDENT_vsubst);;
    e(IMP_RES2_THEN REWRITE_THM IN_DISJOINT_LIST);;
>%



% ===================================================== %
% ===================================================== %
%                  PROCEDURE_CALL_RULE                  %
% ===================================================== %
% ===================================================== %


let CALLk_DISJOINT_DISJOINT_THM =
 prove_thm
  (`CALLk_DISJOINT_DISJOINT_THM`,
   "!r k g p xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                           <|a (x // x0)))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs"
   THEN DEFINE_NEW_VAR "(y0:(var)list) = logicals y"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN "PCSPECk
                      (AAND (pre <|a (x // y))
                            ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                    <|a (x // x0)),
                       (c <|c (x1 //v y1)),q,r,k)"
           ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATION_DISJOINT_STAGE))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT_STAGE)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM
             THENL
              [ ASM_REWRITE_TAC[]
                THEN MATCH_MP_TAC var_list_logicals_IDENT_APPEND
                THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                THEN REWRITE_ALL_TAC[WFxs_APPEND]
                THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants];

                ALL_TAC
              ]
        THEN REWRITE_TAC[PCSPECk]
        THEN DISCH_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_TAC (AND2IMP INTERPRET_CALL_STAGE)
        THEN ASM_REWRITE_TAC[]
        THEN REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]
        THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
        THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
        THEN REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                         ASSUME "y1 = APPEND y y0 :(var)list"]
        THEN USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA)
        THENL
          [  IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                                 DISJOINT_variants;DISJOINT_variants_APPEND];

             ASM_REWRITE_TAC[]
          ]
     ]
  );;

%<
g "!r k g p xs es q vars vals glbs pre post calls rec c vals' u v x x0 u1 v1.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (x0  = logicals x)     /\
      (u1  = APPEND u (logicals u)) /\
      (v1  = APPEND v (logicals v)) /\
      DISJOINT(SL(x0))(FVa q) /\ DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x (AIMP (post <|a (u1 // v1)) q))
                           <|a (x // x0)))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )";;


e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(y:(var)list)  = APPEND v glbs");;
e(DEFINE_NEW_VAR "(y0:(var)list) = logicals y");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN "PCSPECk(AAND (pre <|a (x // y))
                            ((mk_foralls x (AIMP (post <|a (x1 // y1)) q))
                                    <|a (x // x0)),
                  (c <|c (x1 //v y1)),q,r,k)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0 = logicals x")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATION_DISJOINT_STAGE)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_DISJOINT_STAGE));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(SUBGOAL_THEN "(x1 // y1) = (u1 // v1)" REWRITE_THM);;

  e(ASM_REWRITE_TAC[]);;
  e(MATCH_MP_TAC var_list_logicals_IDENT_APPEND);;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND]);;
  e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH_variants;WFxs_variants]);;

 e(REWRITE_TAC[PCSPECk]);;
let st2 = get_state();;
set_state st2;;
 e(DISCH_TAC);;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP INTERPRET_CALL_STAGE));;
 e(ASM_REWRITE_TAC[]);;
 e(REWRITE_TAC[(SYM o ASSUME) "vals' = variants vals(SL(APPEND xs glbs))"]);;
 e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
 e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
 e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
               ASSUME "y1 = APPEND y y0 :(var)list"]);;
 e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                    DISJOINT_variants;DISJOINT_variants_APPEND]);;

  e(ASM_REWRITE_TAC[]);;
>%


let CALLk_DISJOINT_THM =
 prove_thm
  (`CALLk_DISJOINT_THM`,
   "!r k g p xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )",

   REWRITE_TAC[WFc;o_THM]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
              "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
          "PCSPECk(AAND (pre <|a (x // y))
                        ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                  <|a (x0' // x0))
                                             q))
                                <|a (x // x0')),
                   (c <|c (x1 //v y1)),q,r,k)"
            ASSUME_TAC
   THENL
     [  MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                          SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATION_STAGE))))
        THEN EXISTS_TAC "calls <|g (x1 //v y1)"
        THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_STAGE)
        THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
        THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT)
        THEN ASM_REWRITE_TAC[];

        UNDISCH_LAST_TAC
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
             THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
        THEN REWRITE_TAC[PCSPECk]
        THEN DISCH_TAC
        THEN ONCE_REWRITE_TAC[a_subst_THM]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_TAC (AND2IMP INTERPRET_CALL_STAGE)
        THEN REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]
        THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
        THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
        THEN FIRST_ASSUM MATCH_MP_TAC
        THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
        THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
        THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
        THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA)
        THENL
         [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                        DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND];

           REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                       ASSUME "y1 = APPEND y y0 :(var)list"]
           THEN USE_IMP_THEN ONCE_REWRITE_THM
                     (SPEC_ALL command_subst_simplify_LEMMA)
           THENL
            [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND];

              ASM_REWRITE_TAC[]
            ]
         ]
     ]
  );;

%<
g "!r k g p xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))   /\
      DISJOINT(SL(vals'))(FVa q)

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0))
                                  q))
                             <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )";;

e(REWRITE_TAC[WFc;o_THM]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
 "PCSPECk(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r,k)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC
            (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                          SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                         RULE_OF_ADAPTATION_STAGE)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_STAGE));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[PCSPECk]);;
 e(DISCH_TAC);;
let st2 = get_state();;
set_state st2;;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP INTERPRET_CALL_STAGE));;
 e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
 e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
 e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                    DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_APPEND]);;

   e(ASM_REWRITE_TAC[]);;
>%





let WFenv_IMP_MAIN_DL_DISJOINT =
 TAC_PROOF
  (([],
    "!r p vars vals glbs pre post calls rec c vals' q xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs vals') /\
      DISJOINT(SL(APPEND xs vals'))(SL glbs)"),
   REPEAT GEN_TAC THEN STRIP_TAC
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN ASM_REWRITE_TAC[DL_variants;SL_APPEND;DISJOINT_variants_UNION3]
  );;

%< WFenv_IMP_MAIN_DL_DISJOINT
g "!r p vars vals glbs pre post calls rec c vals' q xs.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
      (vals' = variants vals ((FVa q) UNION (SL(APPEND xs glbs)))) /\
      (DL(APPEND xs glbs)) ==>
      WFxs vals /\
      DL(APPEND xs vals') /\
      DISJOINT(SL(APPEND xs vals'))(SL glbs)";;

e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
e(ASM_REWRITE_TAC[DL_variants;SL_APPEND;DISJOINT_variants_UNION3]);;
>%




%================================================================%
%   We wish to prove the general case of procedure call.  This   %
%   involves showing that the partial correctness of the body    %
%   which is known from the well-formedness of the environment   %
%   is maintained under replacement of the formal parameters.    %
%================================================================%



let CALLk_THM =
 prove_thm
  (`CALLk_THM`,
   "!r k p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r"))
   THEN UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
               "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c"
   THEN DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0"
   THEN DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0"
   THEN SUBGOAL_THEN
            "PCSPECk(AAND (pre <|a (x // y))
                          ((mk_foralls x (AIMP ((post <|a (x1 // y1))
                                                    <|a (x0' // x0)) q))
                                  <|a (x // x0')),
                     (c <|c (x1 //v y1)),q,r,k)"
              ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                                 SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                            RULE_OF_ADAPTATION_STAGE))))
      THEN EXISTS_TAC "calls <|g (x1 //v y1)"
      THEN MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_STAGE)
      THEN ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]
      THEN MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN UNDISCH_LAST_TAC
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
        THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]
   THEN REWRITE_TAC[PCSPECk]
   THEN DISCH_TAC
   THEN ONCE_REWRITE_TAC[a_subst_THM]
   THEN REPEAT STRIP_TAC
   THEN SUBGOAL_THEN ((snd o dest_imp o concl o SPEC "s2:state" o GEN "s4:state"
                           o SPEC_ALL o SPEC "vals':var list")
                      INTERPRET_greater_variants_CALL_STAGE)
           STRIP_ASSUME_TAC
   THENL
    [ MATCH_MP_TAC (SPEC_ALL INTERPRET_greater_variants_CALL_STAGE)
      THEN ASM_REWRITE_TAC[];

      ALL_TAC
    ]
   THEN REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]
   THEN SUBGOAL_THEN "DISJOINT (SL vals') (FVa q)" ASSUME_TAC
        THENL [ ASM_REWRITE_TAC[DISJOINT_variants_UNION]; ALL_TAC ]
   THEN IMP_RES_THEN REWRITE_THM A_BNDsL_THM
   THEN IMP_RES_THEN REWRITE_ALL_THM assign_list_THM
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "BNDsL ns vals' (s2':state)"
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]
   THEN USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA)
   THENL
    [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                      DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

      REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                  ASSUME "y1 = APPEND y y0 :(var)list"]
      THEN USE_IMP_THEN ONCE_REWRITE_THM
                (SPEC_ALL command_subst_simplify_LEMMA)
      THENL
       [ ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                         DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3];

         ASM_REWRITE_TAC[]
       ]
    ]
  );;


%< CALLk_THM
g "!r k p g xs es q vars vals glbs pre post calls rec c vals' u v x y x0 y0 x0'.
      (WFenvk r k /\ WFc (CALL p xs es) g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (vals' = variants vals ((FVa q) UNION (SL (APPEND xs glbs)))) /\
      (u   = APPEND xs vals') /\
      (v   = APPEND vars vals) /\
      (x   = APPEND u glbs) /\
      (y   = APPEND v glbs) /\
      (x0  = logicals x)     /\
      (y0  = logicals y)     /\
      (x0' = variants x0 (FVa q))

     %------------------------------------------------------------------% ==>
      PCSPECk( (AAND (pre <|a (u // v))
                     ((mk_foralls x
                            (AIMP (post <|a (APPEND u x0' // APPEND v y0)) q))
                         <|a (x // x0')))
                  <|a (vals' :== es)  ,
               CALL p xs es, q, r, k+1) )";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(STRIP_ASSUME_TAC (REWRITE_RULE[WFc] (ASSUME "WFc (CALL p xs es) g r")));;
e(UNASSUME_THEN (\th. REWRITE_ALL_TAC[th;o_THM] THEN ASSUME_TAC th)
         "(r:env) p = vars,vals,glbs,pre,post,calls,rec,c");;
e(DEFINE_NEW_VAR "(x1:(var)list) = APPEND x x0");;
e(DEFINE_NEW_VAR "(y1:(var)list) = APPEND y y0");;
e(SUBGOAL_THEN
 "PCSPECk(AAND (pre <|a (x // y))
              ((mk_foralls x (AIMP ((post <|a (x1 // y1)) <|a (x0' // x0)) q))
                      <|a (x // x0')),
         (c <|c (x1 //v y1)),q,r,k)"
           ASSUME_TAC);;

  e(MATCH_MP_TAC (REWRITE_RULE[SYM (ASSUME "x0  = logicals x");
                                 SYM (ASSUME "x0' = variants x0 (FVa q)")]
              (GEN "g:prog_env" (SPEC_ALL (CONV_RULE (DEPTH_CONV let_CONV)
                            RULE_OF_ADAPTATION_STAGE)))));;
  e(EXISTS_TAC "calls <|g (x1 //v y1)");;
  e(MATCH_MP_TAC (SPEC_ALL CHANGE_VAR_FOR_ADAPTATION_STAGE));;
  e(ASM_REWRITE_TAC[LENGTH_variants;WFxs_variants]);;
  e(MATCH_MP_TAC (SPEC_ALL WFenv_IMP_MAIN_DL_DISJOINT));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
 e(UNDISCH_LAST_TAC);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL pre_subst_simplify_LEMMA)
   THENL[ ASM_REWRITE_TAC[LENGTH_variants]; ALL_TAC ]);;
 e(REWRITE_TAC[PCSPECk]);;
 e(DISCH_TAC);;
let st2 = get_state();;
set_state st2;;
 e(ONCE_REWRITE_TAC[a_subst_THM]);;
 e(REPEAT STRIP_TAC);;
 e(SUBGOAL_THEN ((snd o dest_imp o concl o SPEC "s2:state" o GEN "s4:state"
                      o SPEC_ALL o SPEC "vals':var list")
                 INTERPRET_greater_variants_CALL_STAGE)
      STRIP_ASSUME_TAC);;

  e(MATCH_MP_TAC (SPEC_ALL INTERPRET_greater_variants_CALL_STAGE));;
  e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[ASSUME "(s2:state) = BNDsL(MAP s2' vals')vals' s3"]);;
 e(SUBGOAL_THEN "DISJOINT (SL vals') (FVa q)" ASSUME_TAC);;

  e(ASM_REWRITE_TAC[DISJOINT_variants_UNION]);;

 e(IMP_RES_THEN REWRITE_THM A_BNDsL_THM);;
 e(IMP_RES_THEN REWRITE_ALL_THM assign_list_THM);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "BNDsL ns vals' (s2':state)");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(REWRITE_ALL_TAC[WFxs_APPEND;DL_APPEND;DISJOINT_APPEND]);;
 e(USE_IMP_THEN REWRITE_THM (SPEC_ALL post_subst_simplify_LEMMA));;

  e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                    DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

  e(REWRITE_TAC[ASSUME "x1 = APPEND x x0 :(var)list";
                ASSUME "y1 = APPEND y y0 :(var)list"]);;
  e(USE_IMP_THEN ONCE_REWRITE_THM (SPEC_ALL command_subst_simplify_LEMMA));;

   e(ASM_REWRITE_TAC[WFxs_variants;LENGTH_variants;DL_variants;SL_APPEND;
                     DISJOINT_APPEND;DL_APPEND;DISJOINT_variants_UNION3]);;

   e(ASM_REWRITE_TAC[]);;
>%




close_theory ();;

