% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definition of the verification condition
  generator for this simple programming language, and its proof of
  soundness for the total correctness of programs.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `vcg`;;

new_parent `termination`;;



% --------------------------------------------------------------------- %
% Need the ind_defs library.                                            %
% --------------------------------------------------------------------- %

loadt `load_finite_sets`;;
load_finite_sets();;

load_library `ind_defs`;;


% Set up autoloading from ancestor theories of vcg %

loadf `autoload`;;

autoload_all_defs_and_thms
     `string bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules semantic_stages
      stage_semantics termination`;;


loadf `semantics_tactics`;;
loadf `prog_semantics_tactics`;;
loadf `translations_tactics`;;
loadf `progress_tactics`;;



% Define the infix version of APPEND (from theory `list`) %


let APPEND_INFIX =
    new_infix_def
    `APPEND_INFIX`
    "$&:((*)list -> (*)list -> (*)list) = APPEND";;


% ================================================================== %
% Define some destructor functions to break apart assertion-language %
% expressions, such as ALESS v (AVAR x).                             %
% ================================================================== %

% This destructor may not be needed:

let dest_AVAR =
    new_rec_def `dest_AVAR` vexp_Axiom
      "(dest_AVAR (AVAR x) = x)";;
%

let dest_ALESS =
    new_rec_def `dest_ALESS` aexp_Axiom
      "dest_ALESS (ALESS e1 e2) = (e1,e2)";;



% Define the basic function that specifies the VCG. %



let vcg1 =
    new_rec_def `vcg1` cmd_Axiom
      "(vcg1 (SKIP) calls q r =
                q,[]) /\
       (vcg1 (ABORT) calls q r =
                AFALSE,[]) /\
       (vcg1 (ASSIGN x e) calls q r =
                q <|a (x := e),[]) /\
       (vcg1 (SEQ c1 c2) calls q r =
              let (s,h2) = vcg1 c2 calls q r in
              let (p,h1) = vcg1 c1 calls s r in
                p,(h1 & h2)) /\
       (vcg1 (IF b c1 c2) calls q r =
              let (r1,h1) = vcg1 c1 calls q r in
              let (r2,h2) = vcg1 c2 calls q r in
                ACOND (AB b) (ab_pre b r1) (ab_pre b r2),(h1 & h2))  /\
       (vcg1 (WHILE a pr b c) calls q r =
              let (v0,v1) = dest_ALESS pr in
              let (p,h) = vcg1 c calls (AAND a pr) r in
               a, ([AIMP (AAND (AAND a (AB b)) (AEQ v0 v1)) (ab_pre b p);
                    AIMP (AAND a (ANOT (AB b))) (ab_pre b q) ] & h))  /\
       (vcg1 (CALL p xs es) calls q r =
              let (vars,vals,glbs,pre,post,calls',rec,c) = (r:env) p in
              let vals' = variants vals (FVa q UNION SL (APPEND xs glbs)) in
              let u   = xs & vals' in
              let v   = vars & vals in
              let x'  = u & glbs in
              let y   = v & glbs in
              let x0  = logicals x' in
              let y0  = logicals y in
              let x0' = variants x0 (FVa q) in
              ((AAND ((AAND pre (calls p)) <|a (u // v))
                     ((mk_foralls x'
                         (AIMP (post <|a ((u & x0') // (v & y0))) q))
                       <|a (x' // x0')))
                <|a (vals' :== es)), [])   ";;





let vcg1_0_THM =
 prove_thm
  (`vcg1_0_THM`,
   "!c calls q r.
        WFenv_syntax r /\ WFc c calls r ==>
           let (p,h) = vcg1 c calls q r in
           (ALL_EL CLOSE h ==> PCSPECk(p,c,q,r,0))",
   ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
     [ REWRITE_TAC[SKIPk_THM];

       REWRITE_TAC[ABORTk_THM];

       REWRITE_TAC[ASSIGNk_THM];

       RES_TAC
       THEN IMP_RES_TAC SEQk_THM;

       RES_TAC
       THEN IMP_RES_TAC IFk_THM1
       THEN ASM_REWRITE_TAC[];

       POP_ASSUM MP_TAC
       THEN ASM_REWRITE_TAC[dest_ALESS]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN STRIP_TAC
       THEN RES_TAC
       THEN ASSUME_TAC
               (SPECL ["b:bexp";
                       "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                ab_pre_THM)
       THEN ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM)
       THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
       THEN IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM));

       REWRITE_TAC[PCSPECk_CALL_0]
     ]
  );;

% vcg1_0_THM:
g "!c calls q r.
        WFenv_syntax r /\ WFc c calls r ==>
           let (p,h) = vcg1 c calls q r in
           (ALL_EL CLOSE h ==> PCSPECk(p,c,q,r,0))";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIPk_THM]);;

e(REWRITE_TAC[ABORTk_THM]);;

e(REWRITE_TAC[ASSIGNk_THM]);;

e(RES_TAC);;
e(IMP_RES_TAC SEQk_THM);;

e(RES_TAC);;
e(IMP_RES_TAC IFk_THM1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(RES_TAC);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM)));;

e(REWRITE_TAC[CSPECk_CALL_0]);;
%

let vcg1_0_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_0_THM;;

%
vcg1_0_COR = 
|- !c calls q r.
    WFenv_syntax r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    PCSPECk(FST(vcg1 c calls q r),c,q,r,0)
%





let STAGE_CALL_LEMMA =
 TAC_PROOF
  (([],
    "!a b ss1 d ss2 c q r k.
      PCSPECk((AAND (      a    <|a ss1) d) <|a ss2,c,q,r,k) ==>
      PCSPECk((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,q,r,k)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_Ck_THM
   THEN POP_ASSUM MATCH_MP_TAC
   THEN REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]
   THEN GEN_TAC THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;


let vcg1_k_THM =
 prove_thm
  (`vcg1_k_THM`,
   "!c calls q r k.
       WFenvk r k /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> PCSPECk(p,c,q,r,k+1))",
   ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
     [ REWRITE_TAC[SKIPk_THM];

       REWRITE_TAC[ABORTk_THM];

       REWRITE_TAC[ASSIGNk_THM];

       RES_TAC
       THEN IMP_RES_TAC SEQk_THM;

       RES_TAC
       THEN IMP_RES_TAC IFk_THM1
       THEN ASM_REWRITE_TAC[];

       POP_ASSUM MP_TAC
       THEN ASM_REWRITE_TAC[dest_ALESS]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN STRIP_TAC
       THEN IMP_RES_TAC WFenvk_WFenv_syntax
       THEN RES_TAC
       THEN ASSUME_TAC
               (SPECL ["b:bexp";
                       "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                ab_pre_THM)
       THEN ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM)
       THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
       THEN IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM));

       UNDISCH_LAST_TAC
       THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN MATCH_MP_TAC STAGE_CALL_LEMMA
       THEN MATCH_MP_TAC CALLk_THM
       THEN EXISTS_TAC "calls:prog_env"
       THEN EXISTS_TAC "(FST)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                               ((FST o SND o SND)((r:env) s))"
       THEN EXISTS_TAC "logicals (APPEND (APPEND l0 (variants ((FST o SND)(r s))
                         ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                                          ((FST o SND o SND)((r:env) s)))"
       THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
       THEN ASM_REWRITE_TAC[WFc]
     ]
  );;

% vcg1_k_THM:
g "!c calls q r k.
       WFenvk r k /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> PCSPECk(p,c,q,r,k+1))";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIPk_THM]);;

e(REWRITE_TAC[ABORTk_THM]);;

e(REWRITE_TAC[ASSIGNk_THM]);;

e(RES_TAC);;
e(IMP_RES_TAC SEQk_THM);;

e(RES_TAC);;
e(IMP_RES_TAC IFk_THM1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(RES_TAC);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM)));;

e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(MATCH_MP_TAC STAGE_CALL_LEMMA);;
e(MATCH_MP_TAC CALLk_THM);;
e(EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "(FST)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                     ((FST o SND o SND)((r:env) s))" THEN
  EXISTS_TAC "logicals (APPEND (APPEND l0
                                       (variants ((FST o SND)(r s))
              ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                               ((FST o SND o SND)((r:env) s)))"
 );;
e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
e(ASM_REWRITE_TAC[WFc]);;
%

let vcg1_k_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_k_THM;;

%
vcg1_k_COR = 
|- !c calls q r k.
    WFenvk r k /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    PCSPECk(FST(vcg1 c calls q r),c,q,r,k + 1)
%





let PARTIAL_CALL_LEMMA =
 TAC_PROOF
  (([],
    "!a b ss1 d ss2 c q r.
      PCSPEC((AAND (      a    <|a ss1) d) <|a ss2,c,q,r) ==>
      PCSPEC((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,q,r)"),
   REWRITE_TAC[PCSPEC_PCSPECk]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC STAGE_CALL_LEMMA
   THEN ASM_REWRITE_TAC[]
  );;

%< PARTIAL_CALL_LEMMA
g "!a b ss1 d ss2 c q r.
      PCSPEC((AAND (      a    <|a ss1) d) <|a ss2,c,q,r) ==>
      PCSPEC((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,q,r)";;
e(REWRITE_TAC[PCSPEC_PCSPECk]);;
e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC STAGE_CALL_LEMMA);;
e(ASM_REWRITE_TAC[]);;
>%



let vcg1p_THM =
 prove_thm
  (`vcg1p_THM`,
   "!c calls q r.
       WFenvp r /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> PCSPEC(p,c,q,r))",
   ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL

     [ REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN REWRITE_TAC[SKIPk_THM];

       REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN REWRITE_TAC[ABORTk_THM];

       REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN REWRITE_TAC[ASSIGNk_THM];

       RES_TAC
       THEN REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN GEN_TAC
       THEN MATCH_MP_TAC SEQk_THM
       THEN EXISTS_TAC "FST(vcg1 c' calls q r)"
       THEN ASM_REWRITE_TAC[];

       RES_TAC
       THEN REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN GEN_TAC
       THEN MATCH_MP_TAC IFk_THM1
       THEN ASM_REWRITE_TAC[];

       POP_ASSUM MP_TAC
       THEN ASM_REWRITE_TAC[dest_ALESS]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN STRIP_TAC
       THEN IMP_RES_TAC WFenvp_WFenv_syntax
       THEN RES_TAC
       THEN REWRITE_ALL_TAC[PCSPEC_PCSPECk]
       THEN GEN_TAC
       THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
       THEN ASSUME_TAC
               (SPECL ["b:bexp";
                       "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                      ab_pre_THM)
       THEN ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM)
       THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
       THEN IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM));

       UNDISCH_LAST_TAC
       THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN MATCH_MP_TAC PARTIAL_CALL_LEMMA
       THEN MATCH_MP_TAC CALLp_THM
       THEN EXISTS_TAC "calls:prog_env"
       THEN EXISTS_TAC "(FST)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                               ((FST o SND o SND)((r:env) s))"
       THEN EXISTS_TAC "logicals (APPEND (APPEND l0
                                                 (variants ((FST o SND)(r s))
                        ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                                         ((FST o SND o SND)((r:env) s)))"
       THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
       THEN ASM_REWRITE_TAC[WFc]
     ]
  );;

%< vcg1p_THM:
g "!c calls q r.
       WFenvp r /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> PCSPEC(p,c,q,r))";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(REWRITE_TAC[SKIPk_THM]);;

e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(REWRITE_TAC[ABORTk_THM]);;

e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(REWRITE_TAC[ASSIGNk_THM]);;

e(RES_TAC);;
e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(GEN_TAC);;
e(MATCH_MP_TAC SEQk_THM);;
e(EXISTS_TAC "FST(vcg1 c' calls q r)");;
e(ASM_REWRITE_TAC[]);;

e(RES_TAC);;
e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(GEN_TAC);;
e(MATCH_MP_TAC IFk_THM1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(RES_TAC);;
e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(GEN_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILEk_THM)));;

e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(MATCH_MP_TAC PARTIAL_CALL_LEMMA);;
e(MATCH_MP_TAC CALLp_THM);;
e(EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "(FST)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                     ((FST o SND o SND)((r:env) s))" THEN
  EXISTS_TAC "logicals (APPEND (APPEND l0
                                       (variants ((FST o SND)(r s))
              ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                               ((FST o SND o SND)((r:env) s)))"
 );;
e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
e(ASM_REWRITE_TAC[WFc]);;
>%

let vcg1p_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1p_THM;;

%
vcg1p_COR = 
|- !c calls q r.
    WFenvp r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    PCSPEC(FST(vcg1 c calls q r),c,q,r)
%





let CORRECT_CALL_LEMMA =
 TAC_PROOF
  (([],
    "!a b ss1 d ss2 c q r.
      CSPEC((AAND (      a    <|a ss1) d) <|a ss2,c,q,r) ==>
      CSPEC((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,q,r)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_CSPEC
   THEN POP_ASSUM MATCH_MP_TAC
   THEN REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]
   THEN GEN_TAC THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< CORRECT_CALL_LEMMA
g "!a b ss1 d ss2 c q r.
      CSPEC((AAND (      a    <|a ss1) d) <|a ss2,c,q,r) ==>
      CSPEC((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,q,r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC PRE_STRENGTH_CSPEC);;
e(POP_ASSUM MATCH_MP_TAC);;
e(REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]);;
e(GEN_TAC THEN STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
>%


let vcg1_THM =
 prove_thm
  (`vcg1_THM`,
   "!c calls q r.
       WFenv r /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> CSPEC(p,c,q,r))",
   ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
     [ REWRITE_TAC[SKIP_THM];

       REWRITE_TAC[ABORT_THM];

       REWRITE_TAC[ASSIGN_THM];

       RES_TAC
       THEN IMP_RES_TAC SEQ_THM;

       RES_TAC
       THEN IMP_RES_TAC IF_THM1
       THEN ASM_REWRITE_TAC[];

       POP_ASSUM MP_TAC
       THEN ASM_REWRITE_TAC[dest_ALESS]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN STRIP_TAC
       THEN IMP_RES_TAC WFenv_WFenv_syntax
       THEN RES_TAC
       THEN ASSUME_TAC
               (SPECL ["b:bexp";
                       "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                      ab_pre_THM)
       THEN ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM)
       THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
       THEN IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILE_THM));

       UNDISCH_LAST_TAC
       THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
       THEN CONV_TAC (DEPTH_CONV let_CONV)
       THEN REWRITE_TAC[ALL_EL]
       THEN MATCH_MP_TAC CORRECT_CALL_LEMMA
       THEN MATCH_MP_TAC CALL_THM
       THEN EXISTS_TAC "calls:prog_env"
       THEN EXISTS_TAC "(FST)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)"
       THEN EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                               ((FST o SND o SND)((r:env) s))"
       THEN EXISTS_TAC "logicals (APPEND (APPEND l0
                                                 (variants ((FST o SND)(r s))
                        ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                                         ((FST o SND o SND)((r:env) s)))"
       THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
       THEN ASM_REWRITE_TAC[WFc]
     ]
  );;

%< vcg1_THM:
g "!c calls q r.
       WFenv r /\ WFc c calls r ==>
             let (p,h) = vcg1 c calls q r in
             (ALL_EL CLOSE h ==> CSPEC(p,c,q,r))";;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIP_THM]);;

e(REWRITE_TAC[ABORT_THM]);;

e(REWRITE_TAC[ASSIGN_THM]);;

e(RES_TAC);;
e(IMP_RES_TAC SEQ_THM);;

e(RES_TAC);;
e(IMP_RES_TAC IF_THM1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(IMP_RES_TAC WFenv_WFenv_syntax);;
e(RES_TAC);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(ASSUME_TAC (SPECL ["b:bexp";"q:aexp"] ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILE_THM)));;

let st1 = get_state();;
set_state st1;;
e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(MATCH_MP_TAC CORRECT_CALL_LEMMA);;
e(MATCH_MP_TAC CALL_THM);;
e(EXISTS_TAC "calls:prog_env" THEN
  EXISTS_TAC "(FST)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) s)" THEN
  EXISTS_TAC "APPEND (APPEND (FST(r s)) ((FST o SND)(r s)))
                     ((FST o SND o SND)((r:env) s))" THEN
  EXISTS_TAC "logicals (APPEND (APPEND l0
                                       (variants ((FST o SND)(r s))
              ((FVa q) UNION (SL(APPEND l0((FST o (SND o SND))(r s)))))))
                               ((FST o SND o SND)((r:env) s)))"
 );;
e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
e(ASM_REWRITE_TAC[WFc]);;
>%

let vcg1_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_THM;;

%
vcg1_COR = 
|- !c calls q r.
    WFenv r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    CSPEC(FST(vcg1 c calls q r),c,q,r)
%





let CPROGRESS_CALL_PRE_LEMMA =
 TAC_PROOF
  (([],
    "!a b ss1 d ss2 c p q r.
      CPROGRESS(      (      a    <|a ss1)    <|a ss2,c,p,q,r) ==>
      CPROGRESS((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,p,q,r)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_CPROGRESS
   THEN POP_ASSUM MATCH_MP_TAC
   THEN REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]
   THEN GEN_TAC THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< CPROGRESS_CALL_PRE_LEMMA
g "!a b ss1 d ss2 c p q r.
      CPROGRESS(      (      a    <|a ss1)    <|a ss2,c,p,q,r) ==>
      CPROGRESS((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,p,q,r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC PRE_STRENGTH_CPROGRESS);;
e(POP_ASSUM MATCH_MP_TAC);;
e(REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]);;
e(GEN_TAC THEN STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
>%



let CPROGRESS_CHANGE_PRE_LEMMA =
 TAC_PROOF
  (([],
    "!a1 a2 c p q r.
      (!s. A a1 s = A a2 s) /\
      CPROGRESS(a1,c,p,q,r) ==>
      CPROGRESS(a2,c,p,q,r)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_CPROGRESS
   THEN POP_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[CLOSE;A_DEF]
  );;

%< CPROGRESS_CHANGE_PRE_LEMMA
g "!a1 a2 c p q r.
      (!s. A a1 s = A a2 s) /\
      CPROGRESS(a1,c,p,q,r) ==>
      CPROGRESS(a2,c,p,q,r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC PRE_STRENGTH_CPROGRESS);;
e(POP_ASSUM MATCH_MP_TAC);;
e(ASM_REWRITE_TAC[CLOSE;A_DEF]);;
>%



let BNDsR_APPEND =
 TAC_PROOF
  (([],
    "!z x w y (s:**->*).
        (LENGTH x = LENGTH z) ==>
        (BNDsR (APPEND x y) (APPEND z w) s =
         BNDsR x z (BNDsR y w s))"),
   LENGTH_LIST_INDUCT_TAC
   THEN REWRITE_TAC[BNDsR;APPEND]
   THEN RES_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< BNDsR_APPEND
g "!z x w y (s:**->*).
        (LENGTH x = LENGTH z) ==>
        (BNDsR (APPEND x y) (APPEND z w) s =
         BNDsR x z (BNDsR y w s))";;
e(LENGTH_LIST_INDUCT_TAC);;

 e(REWRITE_TAC[BNDsR;APPEND]);;

 e(REWRITE_TAC[BNDsR;APPEND]);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%



let CPROGRESS_CHANGE_VAR_LEMMA =
 prove_thm
  (`CPROGRESS_CHANGE_VAR_LEMMA`,
   "!a r p vars vals glbs pre post calls rec c' vals' vals'' xs es z c p q q'.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c') /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (vals'' = variants vals (FVa q' UNION (SL (APPEND xs glbs)))) /\
      WFxs xs /\ WFes es /\
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals = LENGTH es) /\
      DL(APPEND xs glbs) /\
      (FVa a) SUBSET (SL(APPEND (APPEND (APPEND vars vals) glbs) (logicals z))) /\
      CPROGRESS((a <|a (APPEND xs vals' // APPEND vars vals))
                   <|a (vals' :== es),c,p,q,r)
      ==>
      CPROGRESS((a <|a (APPEND xs vals'' // APPEND vars vals))
                   <|a (vals'' :== es),c,p,q,r)",

   REPEAT STRIP_TAC
   THEN IMP_RES_TAC CPROGRESS_CHANGE_PRE_LEMMA
   THEN POP_ASSUM MATCH_MP_TAC
   THEN GEN_TAC
   THEN REWRITE_TAC[a_subst_THM]
   THEN MATCH_MP_TAC A_EQ_THM
   THEN MATCH_MP_TAC MAP_EXTENSION_SUBSET
   THEN EXISTS_TAC "APPEND (APPEND (APPEND vars vals) glbs) (logicals z)"
   THEN POP_TAC
   THEN FIRST_ASSUM REWRITE_THM
   THEN REWRITE_TAC[s_subst_var_list]
   THEN REWRITE_TAC[MAP_APPEND]
   THEN USE_IMP_AND_THEN REWRITE_THM BNDsR_APPEND
             (ASM_REWRITE_TAC[LENGTH_MAP])
   THEN USE_IMP_AND_THEN REWRITE_THM BNDsR_APPEND
             (ASM_REWRITE_TAC[LENGTH_MAP])
   THEN USE_IMP_AND_THEN REWRITE_THM APPEND_11
             (REWRITE_TAC[LENGTH_MAP;LENGTH_APPEND])
   THEN USE_IMP_AND_THEN REWRITE_THM APPEND_11
             (REWRITE_TAC[LENGTH_MAP;LENGTH_APPEND])
   THEN USE_IMP_AND_THEN REWRITE_THM APPEND_11
             (REWRITE_TAC[LENGTH_MAP])
   THEN REPEAT CONJ_TAC
   THENL  % 4 subgoals %

    [ % 1st subgoal %
      USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
            (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND]
             THEN ASM_REWRITE_TAC[LENGTH_MAP])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
            (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND]
             THEN ASM_REWRITE_TAC[LENGTH_MAP])
      THEN USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
      THEN ASM_REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION3];

      % 2nd subgoal %
      USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
            (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND]
             THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_assign_list_subst_MATCH
            (ASM_REWRITE_TAC[LENGTH_variants;DL_variants])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
            (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND]
             THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants])
      THEN USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_MATCH
      THEN ASM_REWRITE_TAC[LENGTH_variants;DL_variants];

      % 3rd subgoal %
      USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN ASM_REWRITE_TAC[DISJOINT_variants_APPEND])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]
             THEN ASM_REWRITE_TAC[])
      THEN USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
      THEN ASM_REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION3];

      % 4th subgoal %
      USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN MATCH_MP_TAC WFxs_NOT_WFxs
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND]
             THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN MATCH_MP_TAC WFxs_NOT_WFxs
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND]
             THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN MATCH_MP_TAC WFxs_NOT_WFxs
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND]
             THEN ASM_REWRITE_TAC[WFxs_variants;NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN MATCH_MP_TAC WFxs_NOT_WFxs
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND]
             THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
            (ONCE_REWRITE_TAC[DISJOINT_SYM]
             THEN MATCH_MP_TAC WFxs_NOT_WFxs
             THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
             THEN REWRITE_ALL_TAC[WFxs_APPEND]
             THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH
      THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
      THEN MATCH_MP_TAC WFxs_NOT_WFxs
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN REWRITE_ALL_TAC[WFxs_APPEND]
      THEN ASM_REWRITE_TAC[WFxs_variants;NOT_WFxs_logicals]
    ]
  );;

%< CPROGRESS_CHANGE_VAR_LEMMA
g "!a r p vars vals glbs pre post calls rec c' vals' vals'' xs es z c p q q'.
      WFenv_syntax r /\
      (r p = vars,vals,glbs,pre,post,calls,rec,c') /\
      (vals' = variants vals (SL (APPEND xs glbs))) /\
      (vals'' = variants vals (FVa q' UNION (SL (APPEND xs glbs)))) /\
      WFxs xs /\ WFes es /\
      (LENGTH vars = LENGTH xs) /\
      (LENGTH vals = LENGTH es) /\
      DL(APPEND xs glbs) /\
      (FVa a) SUBSET (SL(APPEND (APPEND (APPEND vars vals) glbs) (logicals z))) /\
      CPROGRESS((a <|a (APPEND xs vals' // APPEND vars vals))
                   <|a (vals' :== es),c,p,q,r)
      ==>
      CPROGRESS((a <|a (APPEND xs vals'' // APPEND vars vals))
                   <|a (vals'' :== es),c,p,q,r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC CPROGRESS_CHANGE_PRE_LEMMA);;
e(POP_ASSUM MATCH_MP_TAC);;
e(GEN_TAC);;
e(REWRITE_TAC[a_subst_THM]);;
e(MATCH_MP_TAC A_EQ_THM);;
e(MATCH_MP_TAC MAP_EXTENSION_SUBSET);;
e(EXISTS_TAC "APPEND (APPEND (APPEND vars vals) glbs) (logicals z)");;
e(POP_TAC);;
e(FIRST_ASSUM REWRITE_THM);;
%e(REWRITE_TAC[assign_list;s_subst_BNDsL]);;%
e(REWRITE_TAC[s_subst_var_list]);;
e(REWRITE_TAC[MAP_APPEND]);;
e(USE_IMP_AND_THEN REWRITE_THM BNDsR_APPEND
     (ASM_REWRITE_TAC[LENGTH_MAP]));;
e(USE_IMP_AND_THEN REWRITE_THM BNDsR_APPEND
     (ASM_REWRITE_TAC[LENGTH_MAP]));;
e(USE_IMP_AND_THEN REWRITE_THM APPEND_11
     (REWRITE_TAC[LENGTH_MAP;LENGTH_APPEND]));;
e(USE_IMP_AND_THEN REWRITE_THM APPEND_11
     (REWRITE_TAC[LENGTH_MAP;LENGTH_APPEND]));;
e(USE_IMP_AND_THEN REWRITE_THM APPEND_11
     (REWRITE_TAC[LENGTH_MAP]));;
e(REPEAT CONJ_TAC);;  % 4 subgoals %

% 1st subgoal %
    e(ALL_TAC);;
let st1 = get_state();;
set_state st1;;
    e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT));;

     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_MAP]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

    e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT));;

     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_MAP]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION3]);;

% 2nd subgoal %
let st1 = get_state();;
set_state st1;;
    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT));;

     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_MATCH);;

     e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT));;

     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND]);;
     e(ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_MATCH);;
    e(ASM_REWRITE_TAC[LENGTH_variants;DL_variants]);;

% 3rd subgoal %
let st1 = get_state();;
set_state st1;;
    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION3]);;

% 4th subgoal %
let st1 = get_state();;
set_state st1;;
    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(MATCH_MP_TAC WFxs_NOT_WFxs);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[WFxs_APPEND]);;
     e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(ASM_REWRITE_TAC[DISJOINT_variants_APPEND]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH);;

     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
     e(REWRITE_ALL_TAC[DL_APPEND;DISJOINT_APPEND]);;
     e(ASM_REWRITE_TAC[]);;

    e(USE_IMP_THEN REWRITE_THM MAP_assign_list_subst_NOT_MATCH);;
    e(ASM_REWRITE_TAC[SL_APPEND;DISJOINT_variants_UNION3]);;

>%



let vcg1_PRE_PROGRESS =
 prove_thm
  (`vcg1_PRE_PROGRESS`,
   "!c calls q r.
       WFenvp r /\ WFc c calls r ==>
       let (p,h) = vcg1 c calls q r in
       (ALL_EL CLOSE h ==> PRE_PROGRESS(p,c,r))",
   REWRITE_TAC[PRE_PROGRESS]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
    [ REWRITE_TAC[SKIP_PROGRESS];

      REWRITE_TAC[ABORT_PROGRESS];

      REWRITE_TAC[ASSIGN_PROGRESS];

      MATCH_MP_TAC PRE_STRENGTH_CPROGRESS
      THEN EXISTS_TAC "AAND (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))
                            (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))"
      THEN REWRITE_TAC[CLOSE;A_DEF]
      THEN MATCH_MP_TAC SEQ_PROGRESS
      THEN EXISTS_TAC "FST(vcg1 c' calls q r)"
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASM_REWRITE_TAC[];

      RES_TAC
      THEN MATCH_MP_TAC IF_PROGRESS1
      THEN ASM_REWRITE_TAC[];

      POP_ASSUM MP_TAC
      THEN ASM_REWRITE_TAC[dest_ALESS]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[ALL_EL]
      THEN STRIP_TAC
      THEN MATCH_MP_TAC WHILE_PROGRESS1
      THEN EXISTS_TAC "FST(vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"
      THEN EXISTS_TAC "calls:prog_env"
      THEN IMP_RES_TAC WFenvp_WFenv_syntax
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASSUME_TAC (SPECL ["b:bexp";
                              "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                       ab_pre_THM)
      THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
      THEN ASM_REWRITE_TAC[WFc_WHILE]
      THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN MATCH_MP_TAC NOT_IN_SUBSET
      THEN EXISTS_TAC "SL (APPEND (APPEND (FST((r:env) p)) ((FST o SND)(r p)))
                                  ((FST o SND o SND)(r p)))"
      THEN UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p" ENV_TUPLE))
      THEN ASM_REWRITE_TAC[]
      THEN MATCH_MP_TAC WFxs_NOT_WFx
      THEN ASM_REWRITE_TAC[];

      UNDISCH_LAST_TAC
      THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
      THEN REWRITE_TAC[ALL_EL]
      THEN MATCH_MP_TAC CPROGRESS_CALL_PRE_LEMMA
      THEN ASM_CASES_TAC "(s:string) = p"
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN MATCH_MP_TAC CPROGRESS_CHANGE_VAR_LEMMA
         THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "variants ((FST o SND)((r:env) p))
                                   (SL (APPEND l0 ((FST o SND o SND)(r p))))"
         THEN EXISTS_TAC "[]:var list"
         THEN EXISTS_TAC "q:aexp"
         THEN IMP_RES_TAC WFenvp_WFenv_syntax
         THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
         THEN REWRITE_TAC[CONJ_ASSOC]
         THEN CONJ_TAC
         THENL
          [ ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
            THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
            THEN UNASSUME_THEN (\th.ALL_TAC)
                          (concl (ISPEC "(r:env) p" ENV_TUPLE))
            THEN REWRITE_ALL_TAC[o_THM]
            THEN ASM_REWRITE_TAC[logicals_THM;APPEND_NIL];

            MATCH_MP_TAC CALL_SAME_PROGRESS
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
            THEN ASM_REWRITE_TAC[WFc]
          ];

         IMP_RES_TAC CALL_NOT_SAME_PROGRESS
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< vcg1_PRE_PROGRESS:
g "!c calls q r.
       WFenvp r /\ WFc c calls r ==>
       let (p,h) = vcg1 c calls q r in
       (ALL_EL CLOSE h ==> PRE_PROGRESS(p,c,r))";;
e(REWRITE_TAC[PRE_PROGRESS]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIP_PROGRESS]);;

e(REWRITE_TAC[ABORT_PROGRESS]);;

e(REWRITE_TAC[ASSIGN_PROGRESS]);;

e(MATCH_MP_TAC PRE_STRENGTH_CPROGRESS);;
e(EXISTS_TAC "AAND (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))
                   (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))");;
e(REWRITE_TAC[CLOSE;A_DEF]);;
e(MATCH_MP_TAC SEQ_PROGRESS);;
e(EXISTS_TAC "FST(vcg1 c' calls q r)");;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASM_REWRITE_TAC[]);;

e(RES_TAC);;
e(MATCH_MP_TAC IF_PROGRESS1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(MATCH_MP_TAC WHILE_PROGRESS1);;
e(EXISTS_TAC "FST(vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)");;
e(EXISTS_TAC "calls:prog_env");;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(ASM_REWRITE_TAC[WFc_WHILE]);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(MATCH_MP_TAC NOT_IN_SUBSET);;
e(EXISTS_TAC "SL (APPEND (APPEND (FST((r:env) p)) ((FST o SND)(r p)))
                         ((FST o SND o SND)(r p)))");;
e(UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p" ENV_TUPLE)));;
e(ASM_REWRITE_TAC[]);;
e(MATCH_MP_TAC WFxs_NOT_WFx);;
e(ASM_REWRITE_TAC[]);;

e(UNDISCH_LAST_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
e(REWRITE_TAC[ALL_EL]);;
e(MATCH_MP_TAC CPROGRESS_CALL_PRE_LEMMA);;
e(ASM_CASES_TAC "(s:string) = p");;

 e(POP_ASSUM REWRITE_ALL_THM);;
 e(MATCH_MP_TAC CPROGRESS_CHANGE_VAR_LEMMA);;
 e(EXISTS_TAC "(FST o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "variants ((FST o SND)((r:env) p))
                                   (SL (APPEND l0 ((FST o SND o SND)(r p))))"
         THEN EXISTS_TAC "[]:var list"
         THEN EXISTS_TAC "q:aexp");;
 e(IMP_RES_TAC WFenvp_WFenv_syntax);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
 e(REWRITE_TAC[CONJ_ASSOC]);;
 e(CONJ_TAC);;

  e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p" ENV_TUPLE)));;
  e(REWRITE_ALL_TAC[o_THM]);;
  e(ASM_REWRITE_TAC[logicals_THM;APPEND_NIL]);;

  e(MATCH_MP_TAC CALL_SAME_PROGRESS);;
  e(EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)");;
  e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
  e(ASM_REWRITE_TAC[WFc]);;

 e(IMP_RES_TAC CALL_NOT_SAME_PROGRESS);;
 e(ASM_REWRITE_TAC[]);;
>%

let vcg1_PRE_PROGRESS_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_PRE_PROGRESS;;

%
vcg1_PRE_PROGRESS_COR = 
|- !c calls q r.
    WFenvp r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    PRE_PROGRESS(FST(vcg1 c calls q r),c,r)
%





let CPROGRESS_CALL_CALLS_LEMMA =
 TAC_PROOF
  (([],
    "!a b ss1 d ss2 c p q r.
      CPROGRESS(      (        b  <|a ss1)    <|a ss2,c,p,q,r) ==>
      CPROGRESS((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,p,q,r)"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC PRE_STRENGTH_CPROGRESS
   THEN POP_ASSUM MATCH_MP_TAC
   THEN REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]
   THEN GEN_TAC THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

%< CPROGRESS_CALL_CALLS_LEMMA
g "!a b ss1 d ss2 c p q r.
      CPROGRESS(      (        b  <|a ss1)    <|a ss2,c,p,q,r) ==>
      CPROGRESS((AAND ((AAND a b) <|a ss1) d) <|a ss2,c,p,q,r)";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC PRE_STRENGTH_CPROGRESS);;
e(POP_ASSUM MATCH_MP_TAC);;
e(REWRITE_TAC[CLOSE;A_DEF;a_subst_THM]);;
e(GEN_TAC THEN STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
>%





let vcg1_BODY_PROGRESS =
 prove_thm
  (`vcg1_BODY_PROGRESS`,
   "!c calls q r.
       WFenvp r /\ WFcalls calls r /\ WFc c calls r ==>
       let (p,h) = vcg1 c calls q r in
       (ALL_EL CLOSE h ==> BODY_PROGRESS(p,c,calls,r))",
   REWRITE_TAC[BODY_PROGRESS]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
    [ REWRITE_TAC[SKIP_PROGRESS];

      REWRITE_TAC[ABORT_PROGRESS];

      REWRITE_TAC[ASSIGN_PROGRESS];

      MATCH_MP_TAC PRE_STRENGTH_CPROGRESS
      THEN EXISTS_TAC "AAND (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))
                            (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))"
      THEN REWRITE_TAC[CLOSE;A_DEF]
      THEN MATCH_MP_TAC SEQ_PROGRESS
      THEN EXISTS_TAC "FST(vcg1 c' calls q r)"
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASM_REWRITE_TAC[];

      RES_TAC
      THEN MATCH_MP_TAC IF_PROGRESS1
      THEN ASM_REWRITE_TAC[];

      POP_ASSUM MP_TAC
      THEN ASM_REWRITE_TAC[dest_ALESS]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[ALL_EL]
      THEN STRIP_TAC
      THEN MATCH_MP_TAC WHILE_PROGRESS1
      THEN EXISTS_TAC "FST(vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"
      THEN EXISTS_TAC "calls:prog_env"
      THEN IMP_RES_TAC WFenvp_WFenv_syntax
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASSUME_TAC (SPECL ["b:bexp";
                              "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                       ab_pre_THM)
      THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
      THEN ASM_REWRITE_TAC[WFc_WHILE];

      UNDISCH_LAST_TAC
      THEN UNDISCH_TAC "WFcalls calls r"
      THEN REWRITE_TAC[WFcalls]
      THEN DISCH_THEN (MP_TAC o SPEC "s:string")
      THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN STRIP_TAC
      THEN REWRITE_TAC[ALL_EL]
      THEN MATCH_MP_TAC CPROGRESS_CALL_CALLS_LEMMA
      THEN ASM_CASES_TAC "(s:string) = p"
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN MATCH_MP_TAC CPROGRESS_CHANGE_VAR_LEMMA
         THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "variants ((FST o SND)((r:env) p))
                                   (SL (APPEND l0 ((FST o SND o SND)(r p))))"
         THEN EXISTS_TAC "z:var list"
         THEN EXISTS_TAC "q:aexp"
         THEN IMP_RES_TAC WFenvp_WFenv_syntax
         THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
         THEN REWRITE_TAC[CONJ_ASSOC]
         THEN CONJ_TAC
         THENL
          [ REWRITE_ALL_TAC[o_THM]
            THEN ASM_REWRITE_TAC[];

            MATCH_MP_TAC CALL_SAME_PROGRESS
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]
            THEN ASM_REWRITE_TAC[WFc]
          ];

         IMP_RES_TAC CALL_NOT_SAME_PROGRESS
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< vcg1_BODY_PROGRESS:
g "!c calls q r.
       WFenvp r /\ WFcalls calls r /\ WFc c calls r ==>
       let (p,h) = vcg1 c calls q r in
       (ALL_EL CLOSE h ==> BODY_PROGRESS(p,c,calls,r))";;
e(REWRITE_TAC[BODY_PROGRESS]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIP_PROGRESS]);;

e(REWRITE_TAC[ABORT_PROGRESS]);;

e(REWRITE_TAC[ASSIGN_PROGRESS]);;

e(MATCH_MP_TAC PRE_STRENGTH_CPROGRESS);;
e(EXISTS_TAC "AAND (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))
                   (FST(vcg1 c calls(FST(vcg1 c' calls q r))r))");;
e(REWRITE_TAC[CLOSE;A_DEF]);;
e(MATCH_MP_TAC SEQ_PROGRESS);;
e(EXISTS_TAC "FST(vcg1 c' calls q r)");;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASM_REWRITE_TAC[]);;

e(RES_TAC);;
e(MATCH_MP_TAC IF_PROGRESS1);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(MATCH_MP_TAC WHILE_PROGRESS1);;
e(EXISTS_TAC "FST(vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)");;
e(EXISTS_TAC "calls:prog_env");;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(ASM_REWRITE_TAC[WFc_WHILE]);;

let st1 = get_state();;
set_state st1;;

e(UNDISCH_LAST_TAC);;
e(UNDISCH_TAC "WFcalls calls r");;
e(REWRITE_TAC[WFcalls]);;
e(DISCH_THEN (MP_TAC o SPEC "s:string"));;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(STRIP_TAC);;
e(REWRITE_TAC[ALL_EL]);;
e(MATCH_MP_TAC CPROGRESS_CALL_CALLS_LEMMA);;
e(ASM_CASES_TAC "(s:string) = p");;

 e(POP_ASSUM REWRITE_ALL_THM);;
 e(MATCH_MP_TAC CPROGRESS_CHANGE_VAR_LEMMA);;
 e(EXISTS_TAC "(FST o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)"
         THEN EXISTS_TAC "variants ((FST o SND)((r:env) p))
                                   (SL (APPEND l0 ((FST o SND o SND)(r p))))"
         THEN EXISTS_TAC "z:var list"
         THEN EXISTS_TAC "q:aexp");;
 e(IMP_RES_TAC WFenvp_WFenv_syntax);;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
 e(REWRITE_TAC[CONJ_ASSOC]);;
 e(CONJ_TAC);;

  e(REWRITE_ALL_TAC[o_THM]);;
  e(ASM_REWRITE_TAC[]);;

  e(MATCH_MP_TAC CALL_SAME_PROGRESS);;
  e(EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "(FST o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p)"
            THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p)");;
  e(REWRITE_TAC[(SYM o SPEC_ALL) ENV_TUPLE]);;
  e(ASM_REWRITE_TAC[WFc]);;

 e(IMP_RES_TAC CALL_NOT_SAME_PROGRESS);;
 e(ASM_REWRITE_TAC[]);;
>%

let vcg1_BODY_PROGRESS_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_BODY_PROGRESS;;

%
vcg1_BODY_PROGRESS_COR = 
|- !c calls q r.
    WFenvp r /\ WFcalls calls r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    BODY_PROGRESS(FST(vcg1 c calls q r),c,calls,r)
%




let vcg1_TERM =
 prove_thm
  (`vcg1_TERM`,
   "!c r calls q.
     WFenvp r /\ WFc c calls r ==>
     let (p,h) = vcg1 c calls q r in
     (ALL_EL CLOSE h ==> C_calls_TERM(p,c,r))",

   ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN cmd_induct ASSUME_TAC
   THEN REWRITE_TAC[vcg1;WFc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THENL
    [ REWRITE_TAC[SKIP_TERM];

      REWRITE_TAC[ABORT_TERM];

      REWRITE_TAC[ASSIGN_TERM];

      MATCH_MP_TAC SEQ_TERM
      THEN EXISTS_TAC "FST(vcg1 c' calls q r)"
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASM_REWRITE_TAC[];

      MATCH_MP_TAC IF_TERM1
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      POP_ASSUM MP_TAC
      THEN ASM_REWRITE_TAC[dest_ALESS]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[ALL_EL]
      THEN STRIP_TAC
      THEN IMP_RES_TAC WFenvp_WFenv_syntax
      THEN RES_TAC
      THEN IMP_RES_TAC vcg1p_COR
      THEN ASSUME_TAC (SPECL ["b:bexp";
                              "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
                             ab_pre_THM)
      THEN IMP_RES_TAC PRE_STRENGTH_BSPEC
      THEN IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILE_TERM));

      MATCH_MP_TAC (SPEC_ALL CALL_TERM)
      THEN IMP_RES_TAC WFenvp_WFenv_syntax
      THEN ASM_REWRITE_TAC[WFc]
    ]
  );;

%<
g "!c r calls q.
     WFenvp r /\ WFc c calls r ==>
     let (p,h) = vcg1 c calls q r in
     (ALL_EL CLOSE h ==> C_calls_TERM(p,c,r))";;

e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(INDUCT_THEN cmd_induct ASSUME_TAC
  THEN  REWRITE_TAC[vcg1;WFc]
  THEN  ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
  THEN  CONV_TAC (DEPTH_CONV let_CONV)
  THEN  REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
  THEN  REPEAT STRIP_TAC
 );;

e(REWRITE_TAC[SKIP_TERM]);;

e(REWRITE_TAC[ABORT_TERM]);;

e(REWRITE_TAC[ASSIGN_TERM]);;

e(MATCH_MP_TAC SEQ_TERM);;
e(EXISTS_TAC "FST(vcg1 c' calls q r)");;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASM_REWRITE_TAC[]);;

e(MATCH_MP_TAC IF_TERM1);;
e(RES_TAC);;
e(ASM_REWRITE_TAC[]);;

e(POP_ASSUM MP_TAC);;
e(ASM_REWRITE_TAC[dest_ALESS]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[ALL_EL]);;
e(STRIP_TAC);;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(RES_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(ASSUME_TAC (SPECL ["b:bexp";
                     "FST (vcg1 c calls (AAND a0 (ALESS v (AVAR x))) r)"]
              ab_pre_THM));;
e(IMP_RES_TAC PRE_STRENGTH_BSPEC);;
e(IMP_RES_TAC (AND2IMP (REWRITE_RULE[WFc_WHILE] WHILE_TERM)));;

e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(MATCH_MP_TAC (SPEC_ALL CALL_TERM));;
e(ASM_REWRITE_TAC[WFc]);;
>%

let vcg1_TERM_COR = 
    ((CONV_RULE (DEPTH_CONV let_CONV))
      o ONCE_REWRITE_RULE [SYM (ISPEC "vcg1 c calls q r" PAIR)])
                vcg1_TERM;;

%
vcg1_TERM_COR = 
|- !c r calls q.
    WFenvp r /\ WFc c calls r ==>
    ALL_EL CLOSE(SND(vcg1 c calls q r)) ==>
    C_calls_TERM(FST(vcg1 c calls q r),c,r)
%



let vcgc = new_def
  `vcgc`
  "vcgc p c calls q r =
       let (a,h) = vcg1 c calls q r in
       [AIMP p a] & h";;


% =================================================== %
% The theorem of the command vcg, vcgc, is that the   %
% truth of the vc's implies the truth of the program: %
%                                                     %
%  (ALL_EL CLOSE (vcgc p c q r)) ==> CSPEC(p,c,q,r)   %
%                                                     %
% =================================================== %


let vcgc_0_THM =
 prove_thm
  (`vcgc_0_THM`,
   "!c p calls q r.
         WFenv_syntax r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPECk(p,c,q,r,0)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_0_COR
   THEN IMP_RES_TAC PRE_STRENGTH_Ck_THM
  );;

%
g "!c p calls q r.
         WFenv_syntax /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPECk(p,c,q,r,0)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_0_COR);;
e(IMP_RES_TAC PRE_STRENGTH_Ck_THM);;
%


let vcgc_k_THM =
 prove_thm
  (`vcgc_k_THM`,
   "!c p calls q r k.
         WFenvk r k /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPECk(p,c,q,r,k+1)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_k_COR
   THEN IMP_RES_TAC PRE_STRENGTH_Ck_THM
  );;

%
g "!c p calls q r k.
         WFenvk r k /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPECk(p,c,q,r,k+1)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_k_COR);;
e(IMP_RES_TAC PRE_STRENGTH_Ck_THM);;
%


let vcgcp_THM =
 prove_thm
  (`vcgcp_THM`,
   "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPEC(p,c,q,r)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1p_COR
   THEN REWRITE_ALL_TAC[PCSPEC_PCSPECk]
   THEN GEN_TAC
   THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
   THEN IMP_RES_TAC PRE_STRENGTH_Ck_THM
  );;

%
g "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PCSPEC(p,c,q,r)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1p_COR);;
e(REWRITE_ALL_TAC[PCSPEC_PCSPECk]);;
e(GEN_TAC);;
e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
e(IMP_RES_TAC PRE_STRENGTH_Ck_THM);;
%


let vcgc_THM =
 prove_thm
  (`vcgc_THM`,
   "!c p calls q r.
         WFenv r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> CSPEC(p,c,q,r)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_COR
   THEN IMP_RES_TAC PRE_STRENGTH_CSPEC
  );;

% vcgc_THM
g "!c p calls q r.
         WFenv r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> CSPEC(p,c,q,r)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_COR);;
e(IMP_RES_TAC PRE_STRENGTH_CSPEC);;
%


let vcgc_PRE_PROGRESS =
 prove_thm
  (`vcgc_PRE_PROGRESS`,
   "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PRE_PROGRESS(p,c,r)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_PRE_PROGRESS_COR
   THEN IMP_RES_TAC PRE_STRENGTH_PRE_PROGRESS
  );;

% vcgc_PRE_PROGRESS
g "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> PRE_PROGRESS(p,c,r)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_PRE_PROGRESS_COR);;
e(IMP_RES_TAC PRE_STRENGTH_PRE_PROGRESS);;
%


let vcgc_BODY_PROGRESS =
 prove_thm
  (`vcgc_BODY_PROGRESS`,
   "!c p calls q r.
         WFenvp r /\ WFcalls calls r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> BODY_PROGRESS(p,c,calls,r)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_BODY_PROGRESS_COR
   THEN IMP_RES_TAC PRE_STRENGTH_BODY_PROGRESS
  );;

% vcgc_BODY_PROGRESS
g "!c p calls q r.
         WFenvp r /\ WFcalls calls r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> BODY_PROGRESS(p,c,calls,r)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_BODY_PROGRESS_COR);;
e(IMP_RES_TAC PRE_STRENGTH_BODY_PROGRESS);;
%


let vcgc_TERM =
 prove_thm
  (`vcgc_TERM`,
   "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> C_calls_TERM(p,c,r)",
   REWRITE_TAC[vcgc]
   THEN ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcg1_TERM_COR
   THEN IMP_RES_TAC PRE_STRENGTH_C_calls_TERM
  );;

% vcgc_TERM
g "!c p calls q r.
         WFenvp r /\ WFc c calls r ==>
         (ALL_EL CLOSE (vcgc p c calls q r)) ==> C_calls_TERM(p,c,r)";;
e(REWRITE_TAC[vcgc]);;
e(ONCE_REWRITE_TAC[SYM (ISPEC "vcg1 c calls q r" PAIR)]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL_APPEND;ALL_EL]);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcg1_TERM_COR);;
e(IMP_RES_TAC PRE_STRENGTH_C_calls_TERM);;
%



% ==================================================== %
%                                                      %
% Define the vcg function that processes declarations. %
%                                                      %
% ==================================================== %


let vcgd =
    new_rec_def `vcgd` decl_Axiom
      "(vcgd (PROC p vars vals glbs pre post calls rec c) r =
              let x  = (vars & vals) & glbs in
              let x0 = logicals x in
                vcgc (AAND (mk_eqs x0 x) pre) c calls post r) /\
       (vcgd (DSEQ d1 d2) r =
              let h1 = vcgd d1 r in
              let h2 = vcgd d2 r in
                (h1 & h2)) /\
       (vcgd (DEMPTY) r = [])   ";;


% ==================================================== %
% To aid in the proof of its correctness, define a new %
% syntactic function to scan a compound declaration    %
% and return a list of the procedures defined within.  %
% ==================================================== %

let proc_names =
    new_rec_def `proc_names` decl_Axiom
      "(proc_names (PROC p vars vals glbs pre post calls rec c) = [p]) /\
       (proc_names (DSEQ d1 d2) = proc_names d1 & proc_names d2) /\
       (proc_names (DEMPTY) = []) ";;



let UNDECLARED_PROCEDURE =
 prove_thm
  (`UNDECLARED_PROCEDURE`,
   "!d p r. ~(p IN (SL (proc_names d))) ==> (mkenv d r p = r p)",
   INDUCT_THEN decl_induct ASSUME_TAC
   THENL
     [  REWRITE_TAC[proc_names;SET_LIST;IN;mkenv]
        THEN REPEAT STRIP_TAC
        THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH;

        REWRITE_TAC[proc_names;APPEND_INFIX;SL_APPEND;IN_UNION;mkenv]
        THEN REWRITE_TAC[DE_MORGAN_THM]
        THEN REPEAT STRIP_TAC
        THEN RES_TAC
        THEN ASM_REWRITE_TAC[];

        REWRITE_TAC[proc_names;SET_LIST;IN;mkenv]
     ]
  );;

% UNDECLARED_PROCEDURE
g "!d p r. ~(p IN (SL (proc_names d))) ==> (mkenv d r p = r p)";;
e(INDUCT_THEN decl_induct ASSUME_TAC);;

  e(REWRITE_TAC[proc_names;SET_LIST;IN;mkenv]);;
  e(REPEAT STRIP_TAC);;
  e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;

  e(REWRITE_TAC[proc_names;APPEND_INFIX;SL_APPEND;IN_UNION;mkenv]);;
  e(REWRITE_TAC[DE_MORGAN_THM]);;
  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[proc_names;SET_LIST;IN;mkenv]);;
%


let vcgd_IMP_ALL_WFproc_syntax =
 prove_thm
  (`vcgd_IMP_ALL_WFproc_syntax`,
   "!d r r'.
      WFd d r /\ ALL_EL CLOSE (vcgd d r) ==>
      (!p. (p IN (SL (proc_names d))) ==> WFproc_syntax (mkenv d r' p) r) ",
   INDUCT_THEN decl_induct ASSUME_TAC
   THENL
     [  REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[BND_MATCH];

        REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[ALL_EL_APPEND;SL_APPEND;IN_UNION]
        THEN REPEAT STRIP_TAC
        THENL
          [  ASM_CASES_TAC "p IN (SL (proc_names d'))"
             THEN RES_TAC
             THEN IMP_RES_TAC UNDECLARED_PROCEDURE
             THEN ASM_REWRITE_TAC[];

             RES_TAC
             THEN ASM_REWRITE_TAC[]
          ];

        REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]
     ]
  );;

% vcgd_IMP_ALL_WFproc_syntax:
g "!d r r'.
      WFd d r /\ ALL_EL CLOSE (vcgd d r) ==>
      (!p. (p IN (SL (proc_names d))) ==> WFproc_syntax (mkenv d r' p) r) ";;
e(INDUCT_THEN decl_induct ASSUME_TAC);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[ALL_EL_APPEND;SL_APPEND;IN_UNION]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_CASES_TAC "p IN (SL (proc_names d'))");;

      e(RES_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(RES_TAC);;
      e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
      e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]);;
%


let vcgd_IMP_ALL_WFproc0 =
 prove_thm
  (`vcgd_IMP_ALL_WFproc0`,
   "!d r r'. WFenv_syntax r /\ WFd d r /\ ALL_EL CLOSE (vcgd d r) ==>
             (!p. (p IN (SL (proc_names d))) ==> WFprock (mkenv d r' p) r 0)",
   INDUCT_THEN decl_induct ASSUME_TAC
   THENL
     [  REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REPEAT STRIP_TAC
        THEN POP_ASSUM REWRITE_THM
        THEN ASM_REWRITE_TAC[BND_MATCH]
        THEN REWRITE_TAC[WFprock;WFprock_partial]
        THEN ASM_REWRITE_TAC[]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN MATCH_MP_TAC (IMP2AND vcgc_0_THM)
        THEN EXISTS_TAC "f:prog_env"
        THEN ASM_REWRITE_TAC[]
        THEN UNDISCH_TAC "WFproc_syntax (l0,l1,l2,a0,a1,f,a2,c) r"
        THEN REWRITE_TAC[WFproc_syntax]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN STRIP_TAC;

        REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;SL_APPEND;IN_UNION]
        THEN REPEAT STRIP_TAC
        THENL
          [  ASM_CASES_TAC "p IN (SL (proc_names d'))"
             THEN RES_TAC
             THEN IMP_RES_TAC UNDECLARED_PROCEDURE
             THEN ASM_REWRITE_TAC[];

             RES_TAC
             THEN ASM_REWRITE_TAC[]
          ];

        REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv]
     ]
  );;

% vcgd_IMP_ALL_WFproc0:
g "!d r r'. WFenv_syntax r /\ WFd d r /\ ALL_EL CLOSE (vcgd d r) ==>
            (!p. (p IN (SL (proc_names d))) ==> WFprock (mkenv d r' p) r 0)";;
e(INDUCT_THEN decl_induct ASSUME_TAC);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv;APPEND_INFIX]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REPEAT STRIP_TAC);;
  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[BND_MATCH]);;
  e(REWRITE_TAC[WFprock;WFprock_partial]);;
  e(ASM_REWRITE_TAC[]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(MATCH_MP_TAC (IMP2AND vcgc_0_THM));;
  e(EXISTS_TAC "f:prog_env");;
  e(ASM_REWRITE_TAC[]);;
  e(UNDISCH_TAC "WFproc_syntax (l0,l1,l2,a0,a1,f,a2,c) r");;
  e(REWRITE_TAC[WFproc_syntax]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(STRIP_TAC);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;SL_APPEND;IN_UNION]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_CASES_TAC "p IN (SL (proc_names d'))");;

      e(RES_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(RES_TAC);;
      e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
      e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[WFd;vcgd;proc_names;SET_LIST;IN;mkenv]);;
%

let WFproc_syntax_r0 =
 prove_thm
  (`WFproc_syntax_r0`,
   "!p r. WFproc_syntax(r0 p)r",
   REWRITE_TAC[EMPTY_ENV]
   THEN REWRITE_TAC[WFproc_syntax;APPEND]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN GEN_TAC
   THEN REWRITE_TAC[WFxs;DISJOINT_LIST]
   THEN REWRITE_TAC[WFc;GVc;FVc_DEF;FVa_DEF;SUBSET]
   THEN GEN_TAC
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
  );;

% WFproc_syntax_r0:
g "!p r. WFproc_syntax(r0 p)r";;
e(REWRITE_TAC[EMPTY_ENV]);;
e(REWRITE_TAC[WFproc_syntax]);;
e(REWRITE_TAC[APPEND]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(GEN_TAC);;
e(REWRITE_TAC[WFxs;DISJOINT_LIST]);;
e(REWRITE_TAC[WFc]);;
e(REWRITE_TAC[GVc;FVc_DEF;FVa_DEF;SUBSET]);;
e(GEN_TAC);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) s" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
%


let WFenv_syntax_IMP_ALL_WFproc_syntax =
 prove_thm
  (`WFenv_syntax_IMP_ALL_WFproc_syntax`,
   "!d r. (r = mkenv d r0) ==>
            (WFenv_syntax r = 
             (!p. (p IN (SL (proc_names d))) ==> WFproc_syntax (r p) r))",
   REWRITE_TAC[WFenv_syntax]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [  STRIP_TAC
        THEN GEN_TAC
        THEN STRIP_TAC
        THEN ASM_REWRITE_TAC[];

        STRIP_TAC
        THEN GEN_TAC
        THEN ASM_CASES_TAC "p IN (SL (proc_names d))"
        THENL
          [  RES_TAC;

             IMP_RES_TAC UNDECLARED_PROCEDURE
             THEN ASM_REWRITE_TAC[]
             THEN REWRITE_TAC[WFproc_syntax_r0]
          ]
     ]
  );;

% WFenv_syntax_IMP_ALL_WFproc_syntax:
g "!d r. (r = mkenv d r0) ==>
            (WFenv_syntax r = 
             (!p. (p IN (SL (proc_names d))) ==> WFproc_syntax (r p) r))";;
e(REWRITE_TAC[WFenv_syntax]);;
    e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
    e(REWRITE_TAC[WFproc_syntax]);;
    e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(GEN_TAC);;
  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "p IN (SL (proc_names d))");;

    e(RES_TAC);;

    e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[WFproc_syntax_r0]);;
%


let WFenv0_IMP_ALL_WFproc0 =
 prove_thm
  (`WFenv0_IMP_ALL_WFproc0`,
   "!d r. (r = mkenv d r0) ==>
          (WFenvk r 0 = 
             (!p. (p IN (SL (proc_names d))) ==> WFprock (r p) r 0))",
   REWRITE_TAC[WFenvk;WFprock]
   THEN REPEAT STRIP_TAC
   THEN EQ_TAC
   THENL
     [  STRIP_TAC
        THEN GEN_TAC
        THEN STRIP_TAC
        THEN ASM_REWRITE_TAC[];

        STRIP_TAC
        THEN GEN_TAC
        THEN ASM_CASES_TAC "p IN (SL (proc_names d))"
        THENL
          [  FIRST_ASSUM MATCH_MP_TAC
             THEN ASM_REWRITE_TAC[];

             IMP_RES_TAC UNDECLARED_PROCEDURE
             THEN ASM_REWRITE_TAC[]
             THEN REWRITE_TAC[WFproc_syntax_r0]
             THEN REWRITE_TAC[EMPTY_ENV]
             THEN REWRITE_TAC[SYM EMPTY_ENV]
             THEN REWRITE_TAC[WFprock_partial]
             THEN REWRITE_TAC[APPEND]
             THEN CONV_TAC (DEPTH_CONV let_CONV)
             THEN REWRITE_TAC[mk_eqs_THM]
             THEN MATCH_MP_TAC PRE_STRENGTH_Ck_THM
             THEN EXISTS_TAC "AFALSE"
             THEN REWRITE_TAC[ABORTk_THM;CLOSE;A_DEF]
          ]
     ]
  );;

% WFenv0_IMP_ALL_WFproc0:
g "!d r. (r = mkenv d r0) ==>
            (WFenvk r 0 = 
             (!p. (p IN (SL (proc_names d))) ==> WFprock (r p) r 0))";;
e(REWRITE_TAC[WFenvk]);;
     e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(REWRITE_TAC[WFprock]);;
     e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT STRIP_TAC);;
e(EQ_TAC);;

  e(STRIP_TAC);;
  e(GEN_TAC);;
  e(STRIP_TAC);;
  e(ASM_REWRITE_TAC[]);;

  e(STRIP_TAC);;
  e(GEN_TAC);;
  e(ASM_CASES_TAC "p IN (SL (proc_names d))");;

    e(FIRST_ASSUM MATCH_MP_TAC);;
    e(ASM_REWRITE_TAC[]);;

    e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
    e(ASM_REWRITE_TAC[]);;
    e(REWRITE_TAC[WFproc_syntax_r0]);;
    e(REWRITE_TAC[EMPTY_ENV]);;
    e(REWRITE_TAC[SYM EMPTY_ENV]);;
    e(REWRITE_TAC[WFprock_partial]);;
    e(REWRITE_TAC[APPEND]);;
    e(CONV_TAC (DEPTH_CONV let_CONV));;
    e(REWRITE_TAC[mk_eqs_THM]);;
    e(MATCH_MP_TAC PRE_STRENGTH_Ck_THM);;
    e(EXISTS_TAC "AFALSE");;
    e(REWRITE_TAC[ABORTk_THM;CLOSE;A_DEF]);;
%


let vcgd_syntax_THM =
 prove_thm
  (`vcgd_syntax_THM`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>  WFenv_syntax r",
   REPEAT STRIP_TAC
   THEN IMP_RES_THEN REWRITE_THM WFenv_syntax_IMP_ALL_WFproc_syntax
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC vcgd_IMP_ALL_WFproc_syntax
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN ASM_REWRITE_TAC[]
  );;

% vcgd_syntax_THM:
g "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>  WFenv_syntax r";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_THEN REWRITE_THM WFenv_syntax_IMP_ALL_WFproc_syntax);;
e(ASM_REWRITE_TAC[]);;
e(MATCH_MP_TAC vcgd_IMP_ALL_WFproc_syntax);;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(ASM_REWRITE_TAC[]);;
%


let vcgd_0_THM =
 prove_thm
  (`vcgd_0_THM`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>  WFenvk r 0",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcgd_syntax_THM
   THEN IMP_RES_THEN REWRITE_THM WFenv0_IMP_ALL_WFproc0
   THEN ASM_REWRITE_TAC[]
   THEN MATCH_MP_TAC vcgd_IMP_ALL_WFproc0
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN ASM_REWRITE_TAC[]
  );;

% vcgd_0_THM:
g "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>  WFenvk r 0";;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcgd_syntax_THM);;
e(IMP_RES_THEN REWRITE_THM WFenv0_IMP_ALL_WFproc0);;
e(ASM_REWRITE_TAC[]);;
e(MATCH_MP_TAC vcgd_IMP_ALL_WFproc0);;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(ASM_REWRITE_TAC[]);;
%


let vcgd_IMP_ALL_vcgc_LEMMA =
 prove_thm
  (`vcgd_IMP_ALL_vcgc_LEMMA`,
   "!d r r'. ALL_EL CLOSE (vcgd d r)  ==>
             !p. (p IN SL (proc_names d)) ==>
                let (vars,vals,glbs,pre,post,calls,rec,c) = mkenv d r' p in
                let x  = (vars & vals) & glbs in
                let x0 = logicals x in
                ALL_EL CLOSE (vcgc (AAND (mk_eqs x0 x) pre) c calls post r)",
   REWRITE_TAC[APPEND_INFIX]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN INDUCT_THEN decl_induct ASSUME_TAC
   THENL
     [  REWRITE_TAC[vcgd;proc_names;SET_LIST;IN;mkenv]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[APPEND_INFIX]
        THEN REPEAT STRIP_TAC
        THEN ASM_REWRITE_TAC[BND_MATCH]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN ASM_REWRITE_TAC[];

        REWRITE_TAC[vcgd;proc_names;mkenv]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;SL_APPEND;IN_UNION]
        THEN REPEAT STRIP_TAC
        THENL
          [  ASM_CASES_TAC "p IN (SL (proc_names d'))"
             THEN RES_TAC
             THEN IMP_RES_TAC UNDECLARED_PROCEDURE
             THEN ASM_REWRITE_TAC[];

             RES_TAC
             THEN ASM_REWRITE_TAC[]
          ];

        REWRITE_TAC[vcgd;proc_names;mkenv;ALL_EL;SET_LIST;IN]
     ]
  );;

% vcgd_IMP_ALL_vcgc_LEMMA:
g "!d r r'. ALL_EL CLOSE (vcgd d r)  ==>
            !p. (p IN SL (proc_names d)) ==>
               let (vars,vals,glbs,pre,post,calls,rec,c) = mkenv d r' p in
               let x  = (vars & vals) & glbs in
               let x0 = logicals x in
               ALL_EL CLOSE (vcgc (AAND (mk_eqs x0 x) pre) c calls post r)";;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX]);;
e(INDUCT_THEN decl_induct ASSUME_TAC);;

  e(REWRITE_TAC[vcgd;proc_names;SET_LIST;IN;mkenv]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[APPEND_INFIX]);;
  e(REPEAT STRIP_TAC);;
  e(ASM_REWRITE_TAC[BND_MATCH]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[vcgd;proc_names;mkenv]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;SL_APPEND;IN_UNION]);;
  e(REPEAT STRIP_TAC);;

    e(ASM_CASES_TAC "p IN (SL (proc_names d'))");;

      e(RES_TAC);;
      e(ASM_REWRITE_TAC[]);;

      e(RES_TAC);;
      e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
      e(ASM_REWRITE_TAC[]);;

    e(RES_TAC);;
    e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[vcgd;proc_names;mkenv;ALL_EL;SET_LIST;IN]);;
%


let vcgd_IMP_ALL_vcgc =
 prove_thm
  (`vcgd_IMP_ALL_vcgc`,
   "!d r. (r = mkenv d r0) /\
          ALL_EL CLOSE (vcgd d r)  ==>
           !p. let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
               let x  = (vars & vals) & glbs in
               let x0 = logicals x in
               ALL_EL CLOSE (vcgc (AAND (mk_eqs x0 x) pre) c calls post r)",
   REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "p IN (SL (proc_names d))"
   THENL
     [  IMP_RES_TAC vcgd_IMP_ALL_vcgc_LEMMA
        THEN UNASSUME_THEN REWRITE_ALL_THM "r = mkenv d r0"
        THEN ASM_REWRITE_TAC[];

        IMP_RES_TAC UNDECLARED_PROCEDURE
        THEN ASM_REWRITE_TAC[EMPTY_ENV]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[APPEND_INFIX;APPEND;mk_eqs_THM]
        THEN REWRITE_TAC[vcgc;vcg1]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;ALL_EL;CLOSE;A_DEF]
     ]
  );;

% vcgd_IMP_ALL_vcgc:
g "!d r. (r = mkenv d r0) /\
          ALL_EL CLOSE (vcgd d r)  ==>
           !p. let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
               let x  = (vars & vals) & glbs in
               let x0 = logicals x in
               ALL_EL CLOSE (vcgc (AAND (mk_eqs x0 x) pre) c calls post r)";;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "p IN (SL (proc_names d))");;

  e(IMP_RES_TAC vcgd_IMP_ALL_vcgc_LEMMA);;
  e(UNASSUME_THEN REWRITE_ALL_THM "r = mkenv d r0");;
  e(ASM_REWRITE_TAC[]);;

  e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
  e(ASM_REWRITE_TAC[EMPTY_ENV]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[APPEND_INFIX;APPEND;mk_eqs_THM]);;
  e(REWRITE_TAC[vcgc;vcg1]);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND;ALL_EL;CLOSE;A_DEF]);;
%


let vcgd_k_THM =
 prove_thm
  (`vcgd_k_THM`,
   "!d r k.
      (r = mkenv d r0) /\ WFd d r /\ ALL_EL CLOSE (vcgd d r) /\
      WFenvk r k  ==>
      WFenvk r (k+1)",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC WFenvk_WFenv_syntax
   THEN REWRITE_TAC[WFenvk_aspects]
   THEN FIRST_ASSUM REWRITE_THM
   THEN REWRITE_TAC[WFenvk_partial]
   THEN GEN_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN REWRITE_TAC[WFprock_partial]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN IMP_RES_TAC vcgd_IMP_ALL_vcgc
   THEN POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX])
   THEN IMP_RES_TAC vcgc_k_THM
  );;

% vcgd_k_THM:
g "!d r k.
      (r = mkenv d r0) /\ WFd d r /\ ALL_EL CLOSE (vcgd d r) /\
      WFenvk r k  ==>
      WFenvk r (k+1)";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC WFenvk_WFenv_syntax);;
e(REWRITE_TAC[WFenvk_aspects]);;
e(FIRST_ASSUM REWRITE_THM);;
e(REWRITE_TAC[WFenvk_partial]);;
e(GEN_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(REWRITE_TAC[WFprock_partial]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(IMP_RES_TAC vcgd_IMP_ALL_vcgc);;
e(POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX]));;
e(IMP_RES_TAC vcgc_k_THM);;
%


let vcgd_THM =
 prove_thm
  (`vcgd_THM`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
          WFenvp r",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[WFenvp_WFenvk]
   THEN INDUCT_TAC
   THENL
     [  IMP_RES_TAC vcgd_0_THM;

        REWRITE_TAC[ADD1]
        THEN IMP_RES_TAC vcgd_k_THM
     ]
  );;

% vcgd_THM:
g "!d r. (r = mkenv d r0) /\
         WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
         WFenvp r";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[WFenvp_WFenvk]);;
e(INDUCT_TAC);;

  e(IMP_RES_TAC vcgd_0_THM);;

  e(REWRITE_TAC[ADD1]);;
  e(IMP_RES_TAC vcgd_k_THM);;
%


let CPROGRESS_mk_eqs_logicals_pre =
 TAC_PROOF
  (([],
    "!p r x a c g vars vals glbs pre post calls rec c'.
      WFenv_syntax r /\ WFc c g r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c')) /\
      (FVa a SUBSET SL x) /\ WFxs x /\ DL x ==>
      (CPROGRESS(AAND (mk_eqs (logicals x) x) a, c, p, pre, r) =
       CPROGRESS(                             a, c, p, pre, r))"),

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[CPROGRESS;A_DEF]
   THEN EQ_TAC
   THENL
    [ REPEAT STRIP_TAC
      THEN USE_IMP_AND_THEN SUBST1_TAC
              ((ONCE_REWRITE_RULE[EQ_SYM_EQ] o
                   SPEC_ALL o SPECL["MAP (s1:state) x";"logicals x"])
               A_BNDsL_THM)
                  (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                   THEN IMP_RES_TAC DISJOINT_SUBSET
                   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
                   THEN FIRST_ASSUM MATCH_MP_TAC
                   THEN MATCH_MP_TAC WFxs_NOT_WFxs
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN FIRST_ASSUM (MATCH_MP_TAC o
                 SPEC "BNDsL (MAP (s1:state) x) (logicals x) s1")
      THEN USE_IMP_AND_THEN REWRITE_THM A_mk_eqs_THM
                  (REWRITE_TAC[LENGTH_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                  (ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                  (MATCH_MP_TAC WFxs_NOT_WFxs
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM A_BNDsL_THM
                  (IMP_RES_TAC DISJOINT_SUBSET
                   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
                   THEN FIRST_ASSUM MATCH_MP_TAC
                   THEN MATCH_MP_TAC WFxs_NOT_WFxs
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_AND_THEN REWRITE_THM C_calls_BNDsL_THM
                  (EXISTS_TAC "g:prog_env"
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN ASM_REWRITE_TAC[];

      REPEAT STRIP_TAC
      THEN RES_TAC
    ]
  );;

% CPROGRESS_mk_eqs_logicals_pre
g "!p r x a c g vars vals glbs pre post calls rec c'.
     WFenv_syntax r /\ WFc c g r /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c')) /\
     (FVa a SUBSET SL x) /\ WFxs x /\ DL x ==>
     (CPROGRESS(AAND (mk_eqs (logicals x) x) a, c, p, pre, r) =
      CPROGRESS(                             a, c, p, pre, r))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[CPROGRESS;A_DEF]);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;
 e((USE_IMP_THEN SUBST1_TAC o ONCE_REWRITE_RULE[EQ_SYM_EQ] o
      SPEC_ALL o SPECL["MAP (s1:state) x";"logicals x"])
         A_BNDsL_THM);;

  e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
  e(IMP_RES_TAC DISJOINT_SUBSET);;
  e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(MATCH_MP_TAC WFxs_NOT_WFxs);;
  e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

  e(FIRST_ASSUM (MATCH_MP_TAC o
                 SPEC "BNDsL (MAP (s1:state) x) (logicals x) s1"));;
  e(USE_IMP_THEN REWRITE_THM A_mk_eqs_THM);;

   e(REWRITE_TAC[LENGTH_logicals]);;

   e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT));;

    e(ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals]);;

    e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

     e(MATCH_MP_TAC WFxs_NOT_WFxs);;
     e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

     e(USE_IMP_THEN REWRITE_THM A_BNDsL_THM);;

      e(IMP_RES_TAC DISJOINT_SUBSET);;
      e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
      e(FIRST_ASSUM MATCH_MP_TAC);;
      e(MATCH_MP_TAC WFxs_NOT_WFxs);;
      e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

      e(USE_THEN REWRITE_THM C_calls_BNDsL_THM);;

       e(EXISTS_TAC "g:prog_env");;
       e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

       e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(RES_TAC);;
%


let WFproc_pre_logicals_LEMMA =
 TAC_PROOF
  (([],
    "!p r vars vals glbs pre post calls rec c.
     WFenv_syntax r /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
     (WFproc_pre p r =
        let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        PRE_PROGRESS(AAND (mk_eqs x0 x) pre, c, r))"),

   REPEAT STRIP_TAC
   THEN ASM_REWRITE_TAC[WFproc_pre]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN EQ_TAC
   THENL
     [  DISCH_TAC
        THEN MATCH_MP_TAC PRE_STRENGTH_PRE_PROGRESS
        THEN EXISTS_TAC "pre:aexp"
        THEN ASM_REWRITE_TAC[CLOSE;A_DEF]
        THEN REPEAT STRIP_TAC;

        REWRITE_TAC[PRE_PROGRESS]
        THEN REPEAT STRIP_TAC
        THEN POP_ASSUM (MP_TAC o SPEC "p':string")
        THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]
        THEN CONV_TAC (DEPTH_CONV let_CONV)
        THEN USE_IMP_THEN REWRITE_THM CPROGRESS_mk_eqs_logicals_pre
        THEN EXISTS_TAC "calls:prog_env"
        THEN EXISTS_TAC "(FST)((r:env) p')"
        THEN EXISTS_TAC "(FST o SND)((r:env) p')"
        THEN EXISTS_TAC "(FST o SND o SND)((r:env) p')"
        THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p')"
        THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p')"
        THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')"
        THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')"
        THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
        THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
        THEN ASM_REWRITE_TAC[]
     ]
  );;

% WFproc_pre_logicals_LEMMA
g "!p r vars vals glbs pre post calls rec c.
     WFenv_syntax r /\
     (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
     (WFproc_pre p r =
        let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        PRE_PROGRESS(AAND (mk_eqs x0 x) pre, c, r))";;

e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[WFproc_pre]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(MATCH_MP_TAC PRE_STRENGTH_PRE_PROGRESS);;
 e(EXISTS_TAC "pre:aexp");;
 e(ASM_REWRITE_TAC[CLOSE;A_DEF]);;
 e(REPEAT STRIP_TAC);;

 e(REWRITE_TAC[PRE_PROGRESS]);;
 e(REPEAT STRIP_TAC);;
 e(POP_ASSUM (MP_TAC o SPEC "p':string"));;
 e(ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(USE_IMP_THEN REWRITE_THM CPROGRESS_mk_eqs_logicals_pre);;
 e(EXISTS_TAC "calls:prog_env"
     THEN EXISTS_TAC "(FST)((r:env) p')"
     THEN EXISTS_TAC "(FST o SND)((r:env) p')"
     THEN EXISTS_TAC "(FST o SND o SND)((r:env) p')"
     THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p')"
     THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p')"
     THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')"
     THEN EXISTS_TAC "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')");;
 e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;
%


let WFproc_pre_logicals =
 prove_thm
  (`WFproc_pre_logicals`,
   "!p r.
     WFenv_syntax r ==>
     (WFproc_pre p r =
        let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
        let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        PRE_PROGRESS(AAND (mk_eqs x0 x) pre, c, r))",
   REPEAT STRIP_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN IMP_RES2_THEN REWRITE_THM (AND2IMP WFproc_pre_logicals_LEMMA)
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

% WFproc_pre_logicals
g "!p r.
     WFenv_syntax r ==>
     (WFproc_pre p r =
        let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
        let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        PRE_PROGRESS(AAND (mk_eqs x0 x) pre, c, r))";;

e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(IMP_RES2_THEN REWRITE_THM (AND2IMP WFproc_pre_logicals_LEMMA));;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[]);;
%




let vcgd_PRE_PROGRESS =
 prove_thm
  (`vcgd_PRE_PROGRESS`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
          WFenv_pre r",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcgd_THM
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN REWRITE_TAC[WFenv_pre]
   THEN IMP_RES_THEN REWRITE_THM WFproc_pre_logicals
   THEN GEN_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC vcgd_IMP_ALL_vcgc
   THEN POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX])
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN IMP_RES_TAC vcgc_PRE_PROGRESS
  );;

% vcgd_PRE_PROGRESS:
g "!d r. (r = mkenv d r0) /\
         WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
         WFenv_pre r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcgd_THM);;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(REWRITE_TAC[WFenv_pre]);;
e(IMP_RES_THEN REWRITE_THM WFproc_pre_logicals);;
e(GEN_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_TAC vcgd_IMP_ALL_vcgc);;
e(POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX]));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(IMP_RES_TAC vcgc_PRE_PROGRESS);;
%



let vcgd_BODY_PROGRESS =
 prove_thm
  (`vcgd_BODY_PROGRESS`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
          WFenv_calls r",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcgd_THM
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN REWRITE_TAC[WFenv_calls;WFproc_calls]
   THEN GEN_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC vcgd_IMP_ALL_vcgc
   THEN POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX])
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN MATCH_MP_TAC (IMP2AND vcgc_BODY_PROGRESS)
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
   THEN UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p" ENV_TUPLE))
   THEN ASM_REWRITE_TAC[]
   THEN UNASSUME_THEN (REWRITE_THM o SYM) "r = mkenv d r0"
   THEN REWRITE_TAC[WFcalls]
   THEN GEN_TAC
   THEN POP_ASSUM (MP_TAC o SPEC "p':string")
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
   THEN DISCH_TAC
   THEN EXISTS_TAC "APPEND (APPEND (FST(r p)) ((FST o SND)(r p)))
                           ((FST o SND o SND)((r:env) p))"
   THEN FIRST_ASSUM REWRITE_THM
  );;

% vcgd_BODY_PROGRESS:
g "!d r. (r = mkenv d r0) /\
         WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
         WFenv_calls r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcgd_THM);;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(REWRITE_TAC[WFenv_calls]);;
e(REWRITE_TAC[WFproc_calls]);;
e(GEN_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_TAC vcgd_IMP_ALL_vcgc);;
e(POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX]));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(MATCH_MP_TAC (IMP2AND vcgc_BODY_PROGRESS));;
e(EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)");;
e(UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p" ENV_TUPLE)));;
e(ASM_REWRITE_TAC[]);;
e(UNASSUME_THEN (REWRITE_THM o SYM) "r = mkenv d r0");;
e(REWRITE_TAC[WFcalls]);;
e(GEN_TAC);;
e(POP_ASSUM (MP_TAC o SPEC "p':string"));;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p'" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
e(DISCH_TAC);;
e(EXISTS_TAC "APPEND (APPEND (FST(r p)) ((FST o SND)(r p)))
                     ((FST o SND o SND)((r:env) p))");;
e(FIRST_ASSUM REWRITE_THM);;
%





let C_calls_TERM_mk_eqs_logicals_pre =
 TAC_PROOF
  (([],
    "!p r x a vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ 
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (FVa a SUBSET SL x) /\ WFxs x /\ DL x ==>
      (C_calls_TERM(AAND (mk_eqs (logicals x) x) a, c, r) =
       C_calls_TERM(                             a, c, r))"),

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[C_calls_TERM;A_DEF]
   THEN EQ_TAC
   THENL
    [ REPEAT STRIP_TAC
      THEN FIRST_ASSUM (MP_TAC o
                 SPEC "BNDsL (MAP (s1:state) x) (logicals x) s1")
      THEN USE_IMP_AND_THEN REWRITE_THM A_mk_eqs_THM
                  (REWRITE_TAC[LENGTH_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                  (ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                  (MATCH_MP_TAC WFxs_NOT_WFxs
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM A_BNDsL_THM
                  (IMP_RES_TAC DISJOINT_SUBSET
                   THEN ONCE_REWRITE_TAC[DISJOINT_SYM]
                   THEN FIRST_ASSUM MATCH_MP_TAC
                   THEN MATCH_MP_TAC WFxs_NOT_WFxs
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN USE_AND_THEN REWRITE_THM
                     (GEN "g:prog_env" (SPEC_ALL C_calls_terminate_BNDsL_THM))
                  (EXISTS_TAC "calls:prog_env"
                   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                   THEN ASM_REWRITE_TAC[NOT_WFxs_logicals])
      THEN ASM_REWRITE_TAC[]
      THEN STRIP_TAC
      THEN ASSUME_TAC (SPEC "x:var list" NOT_WFxs_logicals)
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN IMP_RES_TAC (AND2IMP WFc_NOT_WFxs)
      THEN IMP_RES_TAC (AND2IMP C_BNDsL_THM)
      THEN POP_ASSUM (MP_TAC o SPEC "MAP (s1:state) (logicals x)")
      THEN USE_IMP_AND_THEN REWRITE_THM BNDsL_BNDsL_ABSORPTION
                  (REWRITE_TAC[LENGTH_MAP])
      THEN REWRITE_TAC[IDENT_BNDsL]
      THEN DISCH_TAC
      THEN EXISTS_TAC "BNDsL(MAP(s1:state)(logicals x))(logicals x)(s2:state)"
      THEN ASM_REWRITE_TAC[];

      REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< C_calls_TERM_mk_eqs_logicals_pre
g "!p r x a vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ 
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (FVa a SUBSET SL x) /\ WFxs x /\ DL x ==>
      (C_calls_TERM(AAND (mk_eqs (logicals x) x) a, c, r) =
       C_calls_TERM(                             a, c, r))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[C_calls_TERM;A_DEF]);;
e(EQ_TAC);;

 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM (MP_TAC o SPEC "BNDsL (MAP (s1:state) x) (logicals x) s1"));;
 e(USE_IMP_THEN REWRITE_THM A_mk_eqs_THM);;

  e(REWRITE_TAC[LENGTH_logicals]);;

  e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT));;

   e(ASM_REWRITE_TAC[LENGTH_logicals;LENGTH_MAP;DL_logicals]);;

   e(USE_IMP_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH);;

    e(MATCH_MP_TAC WFxs_NOT_WFxs);;
    e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

    e(USE_IMP_THEN REWRITE_THM A_BNDsL_THM);;

     e(IMP_RES_TAC DISJOINT_SUBSET);;
     e(ONCE_REWRITE_TAC[DISJOINT_SYM]);;
     e(FIRST_ASSUM MATCH_MP_TAC);;
     e(MATCH_MP_TAC WFxs_NOT_WFxs);;
     e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

let st1 = get_state();;
set_state st1;;
     e(USE_THEN REWRITE_THM
                  (GEN "g:prog_env" (SPEC_ALL C_calls_terminate_BNDsL_THM)));;

      e(EXISTS_TAC "calls:prog_env");;
      e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
      e(ASM_REWRITE_TAC[NOT_WFxs_logicals]);;

      e(ASM_REWRITE_TAC[]);;
      e(STRIP_TAC);;
      e(ASSUME_TAC (SPEC "x:var list" NOT_WFxs_logicals));;
      e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
      e(IMP_RES_TAC (AND2IMP WFc_NOT_WFxs));;
      e(IMP_RES_TAC (AND2IMP C_BNDsL_THM));;
      e(POP_ASSUM (MP_TAC o 
                   SPEC "MAP (s1:state) (logicals x)"));;
      e(USE_IMP_THEN REWRITE_THM BNDsL_BNDsL_ABSORPTION);;

       e(REWRITE_TAC[LENGTH_MAP]);;

      e(REWRITE_TAC[IDENT_BNDsL]);;
      e(DISCH_TAC);;
      e(EXISTS_TAC "BNDsL(MAP(s1:state)(logicals x))(logicals x)(s2:state)");;
      e(ASM_REWRITE_TAC[]);;

 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


let C_calls_TERM_logicals_LEMMA =
 TAC_PROOF
  (([],
    "!p r vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      ((let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        C_calls_TERM(AAND (mk_eqs x0 x) pre, c, r)) =
        C_calls_TERM(pre, c, r))"),

   REPEAT STRIP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN MATCH_MP_TAC (GEN "g:prog_env"
             (SPEC_ALL C_calls_TERM_mk_eqs_logicals_pre))
   THEN EXISTS_TAC "calls:prog_env"
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN ASM_REWRITE_TAC[]
  );;

%< C_calls_TERM_logicals_LEMMA
g "!p r vars vals glbs pre post calls rec c.
      WFenv_syntax r /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) ==>
      ((let x = APPEND (APPEND vars vals) glbs in
        let x0 = logicals x in
        C_calls_TERM(AAND (mk_eqs x0 x) pre, c, r)) =
        C_calls_TERM(pre, c, r))";;

e(REPEAT STRIP_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(MATCH_MP_TAC (GEN "g:prog_env" (SPEC_ALL C_calls_TERM_mk_eqs_logicals_pre)));;
e(EXISTS_TAC "calls:prog_env");;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(ASM_REWRITE_TAC[]);;
>%



let vcgd_TERM =
 prove_thm
  (`vcgd_TERM`,
   "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
          WFenv_term r",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC vcgd_THM
   THEN IMP_RES_TAC WFenvp_WFenv_syntax
   THEN REWRITE_TAC[WFenv_term;WFproc_term]
   THEN GEN_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN IMP_RES_TAC vcgd_IMP_ALL_vcgc
   THEN POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX])
   THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
   THEN IMP_RES_TAC C_calls_TERM_logicals_LEMMA
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN MATCH_MP_TAC (IMP2AND vcgc_TERM)
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)"
   THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)"
   THEN ASM_REWRITE_TAC[]
  );;

%< vcgd_TERM:
g "!d r. (r = mkenv d r0) /\
          WFd d r /\ ALL_EL CLOSE (vcgd d r)  ==>
          WFenv_term r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC vcgd_THM);;
e(IMP_RES_TAC WFenvp_WFenv_syntax);;
e(REWRITE_TAC[WFenv_term]);;
e(REWRITE_TAC[WFproc_term]);;
e(GEN_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(IMP_RES_TAC vcgd_IMP_ALL_vcgc);;
e(POP_ASSUM (ASSUME_TAC o SPEC "p:string"
                     o CONV_RULE (DEPTH_CONV let_CONV)
                     o ONCE_REWRITE_RULE[ISPEC "(r:env) p" ENV_TUPLE]
                     o REWRITE_RULE[APPEND_INFIX]));;
e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
e(IMP_RES_TAC C_calls_TERM_logicals_LEMMA);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(MATCH_MP_TAC (IMP2AND vcgc_TERM));;
e(EXISTS_TAC "(FST o SND o SND o SND o SND o SND)((r:env) p)");;
e(EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p)");;
e(ASM_REWRITE_TAC[]);;
>%





% ==================================================== %
%                                                      %
% Define the vcg function that explores the procedure  %
% call graph to discover paths of recursive call.      %
%                                                      %
%                                                      %
%  fan_out_graph_vcs parameters:                       %
%      p - procedure where q is true                   %
%     ps - list of procedure names, path from p to p0  %
%     p0 - original "target" procedure name, where     %
%             p0's "rec" condition is true             %
%      q - postcondition, true within p, sufficient    %
%             to ensure that within p0, "rec" is true  %
%    pcs - path conditions, (pcs p') is the condition  %
%             held in the prior q', when p was p'      %
%      r - environment, where WFenv_calls r            %
% all_ps - list of all defined procedures in r         %
%      n - maximum remaining expansions;               %
%             n = |all_ps| - |ps|                      %
%                                                      %
%  return value:                                       %
%          list of verification conditions (aexp)      %
%          required to prove that the condition q      %
%          is true upon every recursive call of p0     %
%                                                      %
% ==================================================== %


let fan_out_graph_vcs_DEF =
 new_prim_rec_def
  `fan_out_graph_vcs_DEF`
  "(fan_out_graph_vcs p ps p0 q pcs r all_ps 0 = []) /\
   (fan_out_graph_vcs p ps p0 q pcs r all_ps (SUC n) =
       FLAT (MAP
              (\p'.
                 let q1 = call_progress p' p q r in
                 ((q1 = ATRUE) => []
                    | ((p' = p0) =>
                            let (vars,vals,glbs,pre,post,calls,rec,c) = r p0 in
                              [ AIMP (AAND pre (induct_pre rec)) q1 ]
                        | (p' IN SL (CONS p ps) =>
                              [ AIMP (pcs p') q1 ]
                            | fan_out_graph_vcs p' (CONS p ps) p0 q1
                                                (BND q1 p' pcs) r all_ps n
                          )
                      )
                 )
              )
              all_ps
            ))";;


let extend_graph_vcs =
 new_def
  `extend_graph_vcs`
  "extend_graph_vcs p ps p0 q pcs r all_ps n p' =
                 let q1 = call_progress p' p q r in
                 ((q1 = ATRUE) => []
                    | ((p' = p0) =>
                            let (vars,vals,glbs,pre,post,calls,rec,c) = r p0 in
                              [ AIMP (AAND pre (induct_pre rec)) q1 ]
                        | (p' IN SL (CONS p ps) =>
                              [ AIMP (pcs p') q1 ]
                            | fan_out_graph_vcs p' (CONS p ps) p0 q1
                                                (BND q1 p' pcs) r all_ps n
                          )
                      )
                 )";;


let fan_out_graph_vcs =
 prove_thm
  (`fan_out_graph_vcs`,
   "!p ps p0 q r all_ps n.
     fan_out_graph_vcs p ps p0 q pcs r all_ps (SUC n) =
       FLAT (MAP
               (extend_graph_vcs p ps p0 q pcs r all_ps n)
               all_ps
            )",
   ONCE_REWRITE_TAC[SYM
      (ETA_CONV "\p'. extend_graph_vcs p ps p0 q pcs r all_ps n p'")]
   THEN REWRITE_TAC[fan_out_graph_vcs_DEF;extend_graph_vcs]
  );;

%< fan_out_graph_vcs
g "!p ps p0 q pcs r all_ps n.
     fan_out_graph_vcs p ps p0 q pcs r all_ps (SUC n) =
       FLAT (MAP
               (extend_graph_vcs p ps p0 q pcs r all_ps n)
               all_ps
            )";;
e(ONCE_REWRITE_TAC[SYM
     (ETA_CONV "\p'. extend_graph_vcs p ps p0 q pcs r all_ps n p'")]);;
e(REWRITE_TAC[fan_out_graph_vcs_DEF;extend_graph_vcs]);;
>%





let CLOSE_AIMP_AAND =
 TAC_PROOF
  (([],
    "!p q r. CLOSE(AIMP p q) /\ CLOSE(AIMP p r) ==> CLOSE(AIMP p (AAND q r))"),
   REWRITE_TAC[CLOSE;A_DEF]
   THEN REPEAT STRIP_TAC
   THEN RES_TAC
  );;

% CLOSE_AIMP_AAND
g "!p q r. CLOSE(AIMP p q) /\ CLOSE(AIMP p r) ==> CLOSE(AIMP p (AAND q r))";;
e(REWRITE_TAC[CLOSE;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(RES_TAC);;

 e(RES_TAC);;
%

let APPEND_EQ_NIL =
 TAC_PROOF
  (([],
    "!l1 l2:* list.
       (APPEND l1 l2 = []) =
       (l1 = []) /\ (l2 = [])"),
   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[APPEND_NIL];
      REWRITE_TAC[APPEND;NOT_CONS_NIL]
    ]
  );;

% APPEND_EQ_NIL
g "!l1 l2:* list.
       (APPEND l1 l2 = []) =
       (l1 = []) /\ (l2 = [])";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND_NIL]);;

 e(REWRITE_TAC[APPEND;NOT_CONS_NIL]);;
%

let CONS_EQ_APPEND_LESSER =
 TAC_PROOF
  (([],
    "!ps2 ps (p:*) p' ps1 p0.
       (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
       ?ps'. CONS p ps' = APPEND ps2 (CONS p' ps1)"),
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;CONS_11]
   THEN REPEAT STRIP_TAC
   THENL [ EXISTS_TAC "ps1:* list";
           EXISTS_TAC "APPEND ps2 (CONS (p':*) ps1)"
         ]
   THEN ASM_REWRITE_TAC[]
  );;

% CONS_EQ_APPEND_LESSER
g "!ps2 ps (p:*) p' ps1 p0.
       (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
       ?ps'. CONS p ps' = APPEND ps2 (CONS p' ps1)";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REPEAT STRIP_TAC);;
  e(EXISTS_TAC "ps1:* list");;
  e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REPEAT STRIP_TAC);;
  e(EXISTS_TAC "APPEND ps2 (CONS (p':*) ps1)");;
  e(ASM_REWRITE_TAC[]);;
%


let APPEND_CONS_CONS_EQ_APPEND_LESSER =
 TAC_PROOF
  (([],
    "!ps' ps2 ps1 (p':string) p ps p1 p0.
      (APPEND ps' (CONS p' (CONS p ps)) =
       APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
      (?ps''. APPEND ps' (CONS p' (CONS p ps'')) = APPEND ps2 (CONS p1 ps1))
       \/ ((p = p0) /\ (ps = []) /\
           (APPEND ps'[p'] = APPEND ps2 (CONS p1 ps1)))"),
   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[APPEND]
      THEN LIST_INDUCT_TAC
      THENL
       [ REWRITE_TAC[APPEND;CONS_11]
         THEN LIST_INDUCT_TAC
         THENL
          [ REWRITE_TAC[APPEND;CONS_11]
            THEN REWRITE_TAC[NOT_CONS_NIL]
            THEN REPEAT GEN_TAC
            THEN DISCH_THEN REWRITE_THM;

            REWRITE_TAC[APPEND;CONS_11]
            THEN REWRITE_TAC[NOT_NIL_CONS]
            THEN REPEAT STRIP_TAC
            THEN EXISTS_TAC "ps1:string list"
            THEN ASM_REWRITE_TAC[]
          ];

         REWRITE_TAC[APPEND;CONS_11]
         THEN STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES)
         THEN POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[APPEND;CONS_11]
         THEN REWRITE_TAC[NOT_NIL_CONS]
         THEN REPEAT STRIP_TAC
         THENL
          [ EXISTS_TAC "ps1:string list";
            EXISTS_TAC "APPEND t (CONS (p1:string) ps1)"
          ]
         THEN ASM_REWRITE_TAC[]
       ];

      REWRITE_TAC[APPEND]
      THEN GEN_TAC
      THEN LIST_INDUCT_TAC
      THENL
       [ REWRITE_TAC[APPEND;CONS_11]
         THEN LIST_INDUCT_TAC
         THENL
          [ REWRITE_TAC[APPEND]
            THEN STRIP_ASSUME_TAC (ISPEC "ps':string list" list_CASES)
            THEN POP_ASSUM REWRITE_THM
            THEN REWRITE_TAC[APPEND;CONS_11]
            THEN REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL];

            REWRITE_TAC[APPEND]
            THEN REPEAT GEN_TAC THEN STRIP_TAC
            THEN FIRST_ASSUM (MP_TAC o SPEC "[]:string list")
            THEN REWRITE_TAC[APPEND]
            THEN DISCH_THEN IMP_RES_TAC
            THENL
             [ DISJ1_TAC
               THEN EXISTS_TAC "ps'':string list"
               THEN ASM_REWRITE_TAC[];

               ASM_REWRITE_TAC[]
             ]
          ];

         POP_TAC
         THEN REWRITE_TAC[APPEND;CONS_11]
         THEN REPEAT GEN_TAC
         THEN STRIP_TAC
         THEN RES_TAC
         THENL
          [ DISJ1_TAC
            THEN EXISTS_TAC "ps'':string list"
            THEN ASM_REWRITE_TAC[];

            ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;

% APPEND_CONS_CONS_EQ_APPEND_LESSER
g "!ps' ps2 ps1 (p':string) p ps p1 p0.
      (APPEND ps' (CONS p' (CONS p ps)) =
       APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
      (?ps''. APPEND ps' (CONS p' (CONS p ps'')) = APPEND ps2 (CONS p1 ps1))
       \/ ((p = p0) /\ (ps = []) /\
           (APPEND ps'[p'] = APPEND ps2 (CONS p1 ps1)))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND]);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(LIST_INDUCT_TAC);;

   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(REWRITE_TAC[NOT_CONS_NIL]);;
   e(REPEAT GEN_TAC);;
   e(DISCH_THEN REWRITE_THM);;

   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(REWRITE_TAC[NOT_NIL_CONS]);;
   e(REPEAT STRIP_TAC);;
   e(EXISTS_TAC "ps1:string list");;
   e(ASM_REWRITE_TAC[]);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES));;

   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(REWRITE_TAC[NOT_NIL_CONS]);;
   e(REPEAT STRIP_TAC);;
   e(EXISTS_TAC "ps1:string list");;
   e(ASM_REWRITE_TAC[]);;

   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(REWRITE_TAC[NOT_NIL_CONS]);;
   e(REPEAT STRIP_TAC);;
   e(EXISTS_TAC "APPEND t (CONS (p1:string) ps1)");;
   e(ASM_REWRITE_TAC[]);;

 e(REWRITE_TAC[APPEND]);;
 e(GEN_TAC);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(LIST_INDUCT_TAC);;

   e(REWRITE_TAC[APPEND]);;
   e(STRIP_ASSUME_TAC (ISPEC "ps':string list" list_CASES));;

    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[APPEND;CONS_11]);;
    e(REWRITE_TAC[NOT_CONS_NIL]);;

    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[APPEND;CONS_11]);;
    e(REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL]);;

   e(REWRITE_TAC[APPEND]);;
   e(REPEAT GEN_TAC THEN STRIP_TAC);;
   e(FIRST_ASSUM (MP_TAC o SPEC "[]:string list"));;
   e(REWRITE_TAC[APPEND]);;
   e(DISCH_THEN IMP_RES_TAC);;

    e(DISJ1_TAC);;
    e(EXISTS_TAC "ps'':string list");;
    e(ASM_REWRITE_TAC[]);;

    e(ASM_REWRITE_TAC[]);;

  e(POP_TAC);;
  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REPEAT GEN_TAC);;
  e(STRIP_TAC);;
  e(RES_TAC);;

   e(DISJ1_TAC);;
   e(EXISTS_TAC "ps'':string list");;
   e(ASM_REWRITE_TAC[]);;

   e(ASM_REWRITE_TAC[]);;
%


let APPEND_CONS_APPEND_CONS =
 TAC_PROOF
  (([],
    "!ps2 (p':*) ps1 p0.
       APPEND ps2 (CONS p' (APPEND ps1 [p0])) =
       APPEND (APPEND ps2 (CONS p' ps1)) [p0]"),
   LIST_INDUCT_TAC
   THEN ASM_REWRITE_TAC[APPEND;CONS_11]
  );;

% APPEND_CONS_APPEND_CONS
g "!ps2 (p':*) ps1 p0.
       APPEND ps2 (CONS p' (APPEND ps1 [p0])) =
       APPEND (APPEND ps2 (CONS p' ps1)) [p0]";;
e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;

  e(ASM_REWRITE_TAC[APPEND;CONS_11]);;
%


let FVa_call_progress1 =
 TAC_PROOF
  (([],
    "!r vars vals glbs pre post calls rec c p1 p2 q y z.
      WFenv_syntax r /\
      (r p2 = vars,vals,glbs,pre,post,calls,rec,c) /\
      (y = APPEND(APPEND vars vals)glbs) /\
      (FVa q) SUBSET (SL(APPEND y(logicals z))) ==>
      (FVa(call_progress p1 p2 q r)) SUBSET
      (SL(APPEND (APPEND (APPEND ((FST)(r p1)) ((FST o SND)(r p1)))
                         ((FST o SND o SND)(r p1)))
                 (logicals z)))"),
   REPEAT STRIP_TAC
   THEN ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE)
   THEN IMP_RES_TAC (AND2IMP FVa_call_progress)
   THEN POP_TAC THEN POP_TAC
   THEN POP_ASSUM (MATCH_MP_TAC o SPEC_ALL o IMP2AND o
                   CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV))
   THEN POP_TAC THEN POP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% FVa_call_progress1
g "!r vars vals glbs pre post calls rec c p1 p2 q y z.
      WFenv_syntax r /\
      (r p2 = vars,vals,glbs,pre,post,calls,rec,c) /\
      (y = APPEND(APPEND vars vals)glbs) /\
      (FVa q) SUBSET (SL(APPEND y(logicals z))) ==>
      (FVa(call_progress p1 p2 q r)) SUBSET
      (SL(APPEND (APPEND (APPEND ((FST)(r p1)) ((FST o SND)(r p1)))
                         ((FST o SND o SND)(r p1)))
                 (logicals z)))";;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (ISPEC "(r:env) p1" ENV_TUPLE));;
e(IMP_RES_TAC (AND2IMP FVa_call_progress));;
e(POP_TAC THEN POP_TAC);;
e(POP_ASSUM (MATCH_MP_TAC o SPEC_ALL o IMP2AND o
             CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV)));;
e(POP_TAC THEN POP_TAC);;
e(ASM_REWRITE_TAC[]);;
%


let CURRENT_PATH_CONDITION =
 TAC_PROOF
  (([],
    "!ps p p' p0 q rec r.
      ((ps = []) ==> (q = rec)) /\
      (!ps'.
        (ps = APPEND ps'[p0]) ==>
        (q = call_path_progress p ps' p0 rec r)) /\
      ~(p' IN SL(CONS p ps)) ==>
      (!ps'. (CONS p ps = APPEND ps' [p0]) ==>
             (call_progress p' p q r = call_path_progress p' ps' p0 rec r))"),
   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[SET_LIST;IN]
      THEN REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN ASM_REWRITE_TAC[]
      THEN LIST_INDUCT_TAC
      THENL
       [ REWRITE_TAC[APPEND;CONS_11]
         THEN REWRITE_TAC[call_path_progress]
         THEN DISCH_THEN REWRITE_THM;

         REWRITE_TAC[APPEND;CONS_11]
         THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
         THEN REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL]
       ];

      ONCE_REWRITE_TAC[SET_LIST]
      THEN REWRITE_TAC[IN;DE_MORGAN_THM]
      THEN REWRITE_TAC[NOT_CONS_NIL]
      THEN REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN LIST_INDUCT_TAC
      THENL
       [ REWRITE_TAC[APPEND;CONS_11;NOT_CONS_NIL];

         REWRITE_TAC[APPEND;CONS_11]
         THEN REPEAT STRIP_TAC
         THEN REWRITE_TAC[call_path_progress]
         THEN UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = h'"
         THEN FIRST_ASSUM (IMP_RES_THEN (REWRITE_THM o SYM))
       ]
    ]
  );;

% CURRENT_PATH_CONDITION
g "!ps p p' p0 q rec r.
      ((ps = []) ==> (q = rec)) /\
      (!ps'.
        (ps = APPEND ps'[p0]) ==>
        (q = call_path_progress p ps' p0 rec r)) /\
      ~(p' IN SL(CONS p ps)) ==>
      (!ps'. (CONS p ps = APPEND ps' [p0]) ==>
             (call_progress p' p q r = call_path_progress p' ps' p0 rec r))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REWRITE_TAC[call_path_progress]);;
  e(DISCH_THEN REWRITE_THM);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
  e(REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL]);;

 e(ONCE_REWRITE_TAC[SET_LIST]);;
 e(REWRITE_TAC[IN;DE_MORGAN_THM]);;
 e(REWRITE_TAC[NOT_CONS_NIL]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(LIST_INDUCT_TAC);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REWRITE_TAC[NOT_CONS_NIL]);;

  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(REPEAT STRIP_TAC);;
  e(REWRITE_TAC[call_path_progress]);;
  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p:string) = h'");;
  e(FIRST_ASSUM (IMP_RES_THEN (REWRITE_THM o SYM)));;
%


let CURRENT_PATH_CONDITION_ENV =
 TAC_PROOF
  (([],
    "!ps p p' p0 q rec r.
      ((ps = []) ==> (q = rec)) /\
      (!ps'.
        (ps = APPEND ps'[p0]) ==>
        (q = call_path_progress p ps' p0 rec r)) /\
      (!p'.
        p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
        (!ps2 ps1.
          (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
          (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      ~(p' = p0) /\
      ~(p' IN SL(CONS p ps)) ==>
      (!p''.
        ((p'' = p') \/ p'' IN (SL(CONS p ps))) /\ ~(p'' = p0) ==>
         (!ps2 ps1.
           (CONS p' (CONS p ps) = APPEND ps2 (CONS p'' (APPEND ps1 [p0]))) ==>
           (BND (call_progress p' p q r) p' pcs p'' =
            call_path_progress p'' ps1 p0 rec r)))"),
   REPEAT STRIP_TAC
   THENL
    [ UNDISCH_LAST_TAC
      THEN UNDISCH_LAST_TAC
      THEN POP_ASSUM REWRITE_THM
      THEN ASM_REWRITE_TAC[BND_MATCH]
      THEN STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES)
      THENL
       [ POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[APPEND;CONS_11]
         THEN STRIP_ASSUME_TAC (ISPEC "ps1:string list" list_CASES)
         THEN POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[APPEND;CONS_11]
         THEN STRIP_TAC
         THEN RES_TAC
         THEN ASM_REWRITE_TAC[call_path_progress];

         POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[APPEND;CONS_11]
         THEN STRIP_TAC
         THEN UNDISCH_TAC "~((p':string) IN SL(CONS p ps))"
         THEN POP_ASSUM REWRITE_THM
         THEN REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]
       ];

      IMP_RES_TAC IN_NOT_IN
      THEN IMP_RES_THEN REWRITE_THM BND_NOT_MATCH
      THEN POP_TAC
      THEN UNDISCH_LAST_TAC
      THEN STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES)
      THEN POP_ASSUM REWRITE_THM
      THEN REWRITE_TAC[APPEND;CONS_11]
      THENL
       [ DISCH_THEN REWRITE_ALL_THM
         THEN RES_TAC;

         STRIP_TAC
         THEN RES_TAC
       ]
    ]
  );;

% CURRENT_PATH_CONDITION_ENV
g "!ps p p' p0 q rec r.
      ((ps = []) ==> (q = rec)) /\
      (!ps'.
        (ps = APPEND ps'[p0]) ==>
        (q = call_path_progress p ps' p0 rec r)) /\
      (!p'.
        p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
        (!ps2 ps1.
          (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
          (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      ~(p' = p0) /\
      ~(p' IN SL(CONS p ps)) ==>
      (!p''.
        ((p'' = p') \/ p'' IN (SL(CONS p ps))) /\ ~(p'' = p0) ==>
         (!ps2 ps1.
           (CONS p' (CONS p ps) = APPEND ps2 (CONS p'' (APPEND ps1 [p0]))) ==>
           (BND (call_progress p' p q r) p' pcs p'' =
            call_path_progress p'' ps1 p0 rec r)))";;

e(REPEAT STRIP_TAC);;

 e(UNDISCH_LAST_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(POP_ASSUM REWRITE_THM);;
 e(ASM_REWRITE_TAC[BND_MATCH]);;
 e(STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES));;

  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(STRIP_ASSUME_TAC (ISPEC "ps1:string list" list_CASES));;

   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(STRIP_TAC);;
   e(RES_TAC);;
   e(ASM_REWRITE_TAC[call_path_progress]);;

   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(STRIP_TAC);;
   e(RES_TAC);;
   e(ASM_REWRITE_TAC[call_path_progress]);;

  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(STRIP_TAC);;
  e(UNDISCH_TAC "~((p':string) IN SL(CONS p ps))");;
  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]);;

 e(IMP_RES_TAC IN_NOT_IN);;
 e(IMP_RES_THEN REWRITE_THM BND_NOT_MATCH);;
 e(POP_TAC);;
 e(UNDISCH_LAST_TAC);;
 e(STRIP_ASSUME_TAC (ISPEC "ps2:string list" list_CASES));;

  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(DISCH_THEN REWRITE_ALL_THM);;
  e(RES_TAC);;

  e(POP_ASSUM REWRITE_THM);;
  e(REWRITE_TAC[APPEND;CONS_11]);;
  e(STRIP_TAC);;
  e(RES_TAC);;
%



let call_progress_TRUE =
 prove_thm
  (`call_progress_TRUE`,
   "!p' p r q.
       CLOSE q ==>
       CLOSE(call_progress p' p q r)",
   REWRITE_TAC[call_progress]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT GEN_TAC
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[CLOSE;a_subst_THM;A_mk_foralls;A_DEF]
   THEN DISCH_THEN REWRITE_THM
  );;

% call_progress_TRUE
g "!p' p r q.
       CLOSE q ==>
       CLOSE(call_progress p' p q r)";;
e(REWRITE_TAC[call_progress]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT GEN_TAC);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[CLOSE;A_DEF]);;

 e(REWRITE_TAC[CLOSE]);;
 e(REWRITE_TAC[a_subst_THM]);;
 e(REWRITE_TAC[A_mk_foralls]);;
 e(REWRITE_TAC[A_DEF]);;
 e(DISCH_THEN REWRITE_THM);;
%


let call_path_progress_TRUE =
 prove_thm
  (`call_path_progress_TRUE`,
   "!ps p' p r q.
       CLOSE q ==>
       CLOSE(call_path_progress p' ps p q r)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[call_path_progress]
   THEN REWRITE_TAC[call_progress_TRUE]
   THEN REPEAT STRIP_TAC
   THEN MATCH_MP_TAC call_progress_TRUE
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% call_path_progress_TRUE
g "!ps p' p r q.
       CLOSE q ==>
       CLOSE(call_path_progress p' ps p q r)";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[call_path_progress]);;
 e(REWRITE_TAC[call_progress_TRUE]);;

 e(REWRITE_TAC[call_path_progress]);;
 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC call_progress_TRUE);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
%


let A_call_progress_TRUE =
 prove_thm
  (`A_call_progress_TRUE`,
   "!p' p r q s.
       (!s. A q s) ==>
       A (call_progress p' p q r) s",
   CONV_TAC (DEPTH_CONV FORALL_IMP_CONV)
   THEN REWRITE_TAC[REWRITE_RULE[CLOSE]call_progress_TRUE]
  );;

% A_call_progress_TRUE
g "!p' p r q s.
       (!s. A q s) ==>
       A (call_progress p' p q r) s";;

e(CONV_TAC (DEPTH_CONV FORALL_IMP_CONV));;
e(REWRITE_TAC[REWRITE_RULE[CLOSE]call_progress_TRUE]);;
%


let A_call_path_progress_TRUE =
 prove_thm
  (`A_call_path_progress_TRUE`,
   "!ps p' p r q s.
       (!s. A q s) ==>
       A (call_path_progress p' ps p q r) s",
   CONV_TAC (DEPTH_CONV FORALL_IMP_CONV)
   THEN REWRITE_TAC[REWRITE_RULE[CLOSE]call_path_progress_TRUE]
  );;

% A_call_path_progress_TRUE
g "!ps p' p r q s.
       (!s. A q s) ==>
       A (call_path_progress p' ps p q r) s";;

e(CONV_TAC (DEPTH_CONV FORALL_IMP_CONV));;
e(REWRITE_TAC[REWRITE_RULE[CLOSE]call_path_progress_TRUE]);;
%


let mk_foralls_NOT_ATRUE =
 TAC_PROOF
  (([],
    "!y a.
      ~(y = []) ==>
      ~(mk_foralls y a = ATRUE)"),
   REPEAT GEN_TAC
   THEN CONTRAPOS_TAC
   THEN SPEC_TAC("y:var list","y:var list")
   THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[mk_foralls]
   THEN REWRITE_TAC[aexp_distinct]
  );;

% mk_foralls_NOT_ATRUE
g "!y a.
      ~(y = []) ==>
      ~(mk_foralls y a = ATRUE)";;

e(REPEAT GEN_TAC);;
e(CONTRAPOS_TAC);;
e(SPEC_TAC("y:var list","y:var list"));;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[mk_foralls]);;
 e(REWRITE_TAC[aexp_distinct]);;
%


let mk_foralls_AIMP_NOT_ATRUE =
 TAC_PROOF
  (([],
    "!y a b.
      ~(mk_foralls y (AIMP a b) = ATRUE)"),
   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(y:var list) = []"
   THENL
    [ ASM_REWRITE_TAC[mk_foralls]
      THEN REWRITE_TAC[aexp_distinct];

      MATCH_MP_TAC mk_foralls_NOT_ATRUE
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% mk_foralls_AIMP_NOT_ATRUE
g "!y a b.
      ~(mk_foralls y (AIMP a b) = ATRUE)";;

e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "(y:var list) = []");;

 e(ASM_REWRITE_TAC[mk_foralls]);;
 e(REWRITE_TAC[aexp_distinct]);;

 e(MATCH_MP_TAC mk_foralls_NOT_ATRUE);;
 e(ASM_REWRITE_TAC[]);;
%


let a_subst_EQ_ATRUE =
 TAC_PROOF
  (([],
    "!a ss.
      ((a <|a ss) = ATRUE) = (a = ATRUE)"),
   INDUCT_THEN aexp_induct ASSUME_TAC
   THEN REWRITE_TAC[a_subst;aexp_distinct]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[a_subst;aexp_distinct]
  );;

% a_subst_EQ_ATRUE
g "!a ss.
      ((a <|a ss) = ATRUE) = (a = ATRUE)";;

e(INDUCT_THEN aexp_induct ASSUME_TAC
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[a_subst;aexp_distinct]);;
%


let call_progress_ALL_TRUE =
 prove_thm
  (`call_progress_ALL_TRUE`,
   "!p2 p3 q r.
     (call_progress p2 p3 q r = ATRUE) ==>
     (!a. call_progress p2 p3 a r = ATRUE)",
   REWRITE_TAC[call_progress]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT GEN_TAC
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[a_subst_EQ_ATRUE]
   THEN REWRITE_TAC[mk_foralls_AIMP_NOT_ATRUE]
  );;

% call_progress_ALL_TRUE
g "!p2 p3 q r.
     (call_progress p2 p3 q r = ATRUE) ==>
     (!a. call_progress p2 p3 a r = ATRUE)";;

e(REWRITE_TAC[call_progress]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT GEN_TAC);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[a_subst_EQ_ATRUE]);;
 e(REWRITE_TAC[mk_foralls_AIMP_NOT_ATRUE]);;
%


let call_progress_BLOCKED_TRUE =
 prove_thm
  (`call_progress_BLOCKED_TRUE`,
   "!ps ps' p1 p2 p3 p4 q a r.
     (call_progress p2 p3 q r = ATRUE) ==>
     CLOSE(call_path_progress p1 (APPEND ps (CONS p2 (CONS p3 ps'))) p4 a r)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;call_path_progress]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE
      THEN MATCH_MP_TAC call_progress_TRUE
      THEN REWRITE_TAC[CLOSE;A_DEF];

      MATCH_MP_TAC call_progress_TRUE
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "q:aexp"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% call_progress_BLOCKED_TRUE
g "!ps ps' p1 p2 p3 p4 q a r.
     (call_progress p2 p3 q r = ATRUE) ==>
     CLOSE(call_path_progress p1 (APPEND ps (CONS p2 (CONS p3 ps'))) p4 a r)";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[call_path_progress]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE);;
 e(MATCH_MP_TAC call_progress_TRUE);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[call_path_progress]);;
 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC call_progress_TRUE);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "q:aexp");;
 e(ASM_REWRITE_TAC[]);;
%


let call_progress_BLOCKED_BASE_TRUE =
 prove_thm
  (`call_progress_BLOCKED_BASE_TRUE`,
   "!ps p1 p2 p3 q a r.
     (call_progress p2 p3 q r = ATRUE) ==>
     CLOSE(call_path_progress p1 (APPEND ps [p2]) p3 a r)",
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[APPEND;call_path_progress]
   THEN REPEAT STRIP_TAC
   THENL
    [ IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE
      THEN MATCH_MP_TAC call_progress_TRUE
      THEN REWRITE_TAC[CLOSE;A_DEF];

      MATCH_MP_TAC call_progress_TRUE
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "q:aexp"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% call_progress_BLOCKED_BASE_TRUE
g "!ps p1 p2 p3 q a r.
     (call_progress p2 p3 q r = ATRUE) ==>
     CLOSE(call_path_progress p1 (APPEND ps [p2]) p3 a r)";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[call_path_progress]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE);;
 e(MATCH_MP_TAC call_progress_TRUE);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;

 e(REWRITE_TAC[APPEND]);;
 e(REWRITE_TAC[call_path_progress]);;
 e(REPEAT STRIP_TAC);;
 e(MATCH_MP_TAC call_progress_TRUE);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "q:aexp");;
 e(ASM_REWRITE_TAC[]);;
%



let call_path_progress_APPEND_SINGLE =
 prove_thm
  (`call_path_progress_APPEND_SINGLE`,
   "!p0 ps' p' p q r.
      call_path_progress p0 (APPEND ps' [p']) p q r =
      call_path_progress p0 ps' p' (call_progress p' p q r) r",

   REPEAT GEN_TAC
   THEN REWRITE_TAC[REWRITE_RULE[APPEND_NIL]
                       (SPECL["ps':string list";"[]:string list"]
                           call_path_progress_APPEND)]
   THEN REWRITE_TAC[call_path_progress]
  );;

%< call_path_progress_APPEND_SINGLE
g "!p0 ps' p' p q r.
      call_path_progress p0 (APPEND ps' [p']) p q r =
      call_path_progress p0 ps' p' (call_progress p' p q r) r";;

  e(REPEAT GEN_TAC);;
  e(REWRITE_TAC[REWRITE_RULE[APPEND_NIL]
                   (SPECL["ps':string list";"[]:string list"]
                       call_path_progress_APPEND)]);;
  e(REWRITE_TAC[call_path_progress]);;
>%


let call_path_progress_APPEND1 =
 TAC_PROOF
  (([],
    "!ps1 ps2 p1 p2 p3 q r.
       call_path_progress p1 (APPEND ps1 (CONS p2 ps2)) p3 q r =
       call_path_progress p1 ps1 p2 (call_path_progress p2 ps2 p3 q r) r"),
   REWRITE_TAC[REWRITE_RULE[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]
                   call_path_progress_APPEND]
  );;

%< call_path_progress_APPEND1
g "!ps1 ps2 p1 p2 p3 q r.
       call_path_progress p1 (APPEND ps1 (CONS p2 ps2)) p3 q r =
       call_path_progress p1 ps1 p2 (call_path_progress p2 ps2 p3 q r) r";;

  e(REWRITE_TAC[REWRITE_RULE[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]
                   call_path_progress_APPEND]);;
>%




let extend_graph_vcs_INDUCT_COVER_PATHS =
 TAC_PROOF
  (([],
    "!n p' p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      (p' IN SL all_ps) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (extend_graph_vcs p ps p0 q pcs r all_ps n p') ==>
        ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0 (APPEND ps'[p']) p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        ))"),

   REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN REWRITE_TAC[extend_graph_vcs]
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN COND_CASES_TAC
   THENL
    [ REWRITE_TAC[ALL_EL]
      THEN COND_CASES_TAC
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN ASM_REWRITE_TAC[call_path_progress]
         THEN REWRITE_TAC[CLOSE;A_DEF];

         COND_CASES_TAC
         THENL
          [ REWRITE_TAC[CLOSE;A_DEF]
            THEN REPEAT STRIP_TAC
            THEN IMP_RES_TAC CONS_EQ_APPEND_LESSER
            THEN FIRST_ASSUM (REWRITE_THM o SYM)
            THEN REWRITE_TAC[call_path_progress]
            THEN IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE
            THEN REWRITE_TAC[A_DEF];

            CONJ_TAC
            THENL
             [ REPEAT STRIP_TAC
               THEN REWRITE_TAC[call_path_progress_APPEND_SINGLE]
               THEN ASM_REWRITE_TAC[CLOSE;A_DEF]
               THEN REPEAT STRIP_TAC
               THEN MATCH_MP_TAC A_call_path_progress_TRUE
               THEN REWRITE_TAC[A_DEF];

               REPEAT STRIP_TAC
               THEN IMP_RES_TAC APPEND_CONS_CONS_EQ_APPEND_LESSER
               THEN POP_ASSUM (REWRITE_THM o SYM)
               THENL
                [ IMP_RES_THEN MP_TAC call_progress_BLOCKED_TRUE;
                  IMP_RES_THEN MP_TAC call_progress_BLOCKED_BASE_TRUE
                ]
               THEN ASM_REWRITE_TAC[CLOSE;A_DEF]
               THEN DISCH_THEN REWRITE_THM
             ]
          ]
       ];

      POP_TAC
      THEN COND_CASES_TAC
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN REWRITE_TAC[ALL_EL]
         THEN REWRITE_TAC[call_path_progress];

         COND_CASES_TAC
         THENL
          [ ASM_REWRITE_TAC[ALL_EL]
            THEN FIRST_ASSUM (IMP_RES2_THEN ASSUME_TAC o AND2IMP o SPEC "p':string")
            THEN REPEAT STRIP_TAC
            THEN FIRST_ASSUM (IMP_RES_THEN ASSUME_TAC)
            THEN POP_ASSUM (REWRITE_THM o SYM)
            THEN IMP_RES_TAC CONS_EQ_APPEND_LESSER
            THEN FIRST_ASSUM (REWRITE_THM o SYM)
            THEN REWRITE_TAC[call_path_progress]
            THEN UNDISCH_LAST_TAC
            THEN UNDISCH_LAST_TAC
            THEN DISCH_THEN (\th1. DISCH_THEN (\th2.
                       ASSUME_TAC th2 THEN MP_TAC th1))
            THEN REWRITE_TAC[APPEND_CONS_APPEND_CONS]
            THEN FIRST_ASSUM (REWRITE_THM o SYM)
            THEN REWRITE_TAC[APPEND;CONS_11]
            THEN DISCH_TAC
            THEN IMP_RES_THEN (REWRITE_THM o SYM)
                    (ASSUME "!ps'. (ps = APPEND ps'[p0]) ==>
                                   (q = call_path_progress p ps' p0 rec r)")
            THEN ASM_REWRITE_TAC[];

            REWRITE_TAC[]
            THEN DISCH_TAC
            THEN FIRST_ASSUM (\th. (SPECL["p':string";"CONS (p:string) ps"]th;
                                    UNDISCH_TAC (concl th)))
            THEN DISCH_THEN (MP_TAC o
                 SPECL["p':string";"CONS (p:string) ps";
                       "call_progress p' p q r";
                       "BND (call_progress p' p q r) p' pcs";
                       "(FST)((r:env) p')";"(FST o SND)((r:env) p')";
                       "(FST o SND o SND)((r:env) p')";
                       "(FST o SND o SND o SND)((r:env) p')";
                       "(FST o SND o SND o SND o SND)((r:env) p')";
                       "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                       "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                       "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                       "APPEND (APPEND ((FST)((r:env) p')) ((FST o SND)(r p')))
                                   ((FST o SND o SND)(r p'))";
                       "z:var list"])
            THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
            THEN REWRITE_TAC[LENGTH;(SYM o SPEC_ALL)ADD1;ADD_CLAUSES]
            THEN ONCE_REWRITE_TAC[DISJOINT_LIST;SET_LIST]
            THEN ASM_REWRITE_TAC[IN;SUBSET]
            THEN IMP_RES_TAC (AND2IMP FVa_call_progress1)
            THEN POP_TAC
            THEN POP_ASSUM REWRITE_THM
            THEN REWRITE_TAC[NOT_CONS_NIL]
            THEN IMP_RES_TAC (AND2IMP CURRENT_PATH_CONDITION)
            THEN POP_ASSUM REWRITE_THM
            THEN USE_THEN REWRITE_THM CURRENT_PATH_CONDITION_ENV
            THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]
            THEN STRIP_TAC
            THEN POP_ASSUM REWRITE_THM
            THEN REWRITE_TAC[call_path_progress_APPEND_SINGLE]
            THEN POP_ASSUM REWRITE_THM
          ]
       ]
    ]
  );;

%< extend_graph_vcs_INDUCT_COVER_PATHS

g "!n p' p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      (p' IN SL all_ps) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (extend_graph_vcs p ps p0 q pcs r all_ps n p') ==>
        ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0 (APPEND ps'[p']) p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        ))";;

e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(REWRITE_TAC[extend_graph_vcs]);;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[ALL_EL]);;
 e(COND_CASES_TAC);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(ASM_REWRITE_TAC[call_path_progress]);;
  e(REWRITE_TAC[CLOSE;A_DEF]);;

  e(COND_CASES_TAC);;

let st2 = get_state();;
set_state st2;;
   e(REWRITE_TAC[CLOSE;A_DEF]);;
   e(REPEAT STRIP_TAC);;
   e(IMP_RES_TAC CONS_EQ_APPEND_LESSER);;
   e(FIRST_ASSUM (REWRITE_THM o SYM));;
   e(REWRITE_TAC[call_path_progress]);;
   e(IMP_RES_THEN REWRITE_THM call_progress_ALL_TRUE);;
   e(REWRITE_TAC[A_DEF]);;

   e(CONJ_TAC);;

    e(REPEAT STRIP_TAC);;
    e(REWRITE_TAC[call_path_progress_APPEND_SINGLE]);;
    e(ASM_REWRITE_TAC[CLOSE;A_DEF]);;
    e(REPEAT STRIP_TAC);;
let st3 = get_state();;
set_state st3;;
    e(MATCH_MP_TAC A_call_path_progress_TRUE);;
    e(REWRITE_TAC[A_DEF]);;

    e(REPEAT STRIP_TAC);;
    e(IMP_RES_TAC APPEND_CONS_CONS_EQ_APPEND_LESSER);;

     e(POP_ASSUM (REWRITE_THM o SYM));;
     e(IMP_RES_THEN MP_TAC call_progress_BLOCKED_TRUE);;
     e(REWRITE_TAC[CLOSE;A_DEF]);;
     e(DISCH_THEN REWRITE_THM);;

     e(POP_ASSUM (REWRITE_THM o SYM));;
     e(IMP_RES_THEN MP_TAC call_progress_BLOCKED_BASE_TRUE);;
     e(ASM_REWRITE_TAC[CLOSE;A_DEF]);;
     e(DISCH_THEN REWRITE_THM);;

 e(POP_TAC);;
 e(COND_CASES_TAC);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(REWRITE_TAC[ALL_EL]);;
  e(REWRITE_TAC[call_path_progress]);;

let st1 = get_state();;
set_state st1;;
  e(COND_CASES_TAC);;

   e(REWRITE_TAC[ALL_EL]);;
 % Nope, it does not actually solve this case! %
   e(FIRST_ASSUM (IMP_RES2_THEN ASSUME_TAC o AND2IMP o SPEC "p':string"));;
   e(REPEAT STRIP_TAC);;
   e(FIRST_ASSUM (IMP_RES_THEN ASSUME_TAC));;
   e(POP_ASSUM (REWRITE_THM o SYM));;
   e(IMP_RES_TAC CONS_EQ_APPEND_LESSER);;
   e(FIRST_ASSUM (REWRITE_THM o SYM));;
   e(REWRITE_TAC[call_path_progress]);;
   e(UNDISCH_LAST_TAC);;
   e(UNDISCH_LAST_TAC);;
   e(DISCH_THEN (\th1. DISCH_THEN (\th2. ASSUME_TAC th2 THEN MP_TAC th1)));;
   e(REWRITE_TAC[APPEND_CONS_APPEND_CONS]);;
   e(FIRST_ASSUM (REWRITE_THM o SYM));;
   e(REWRITE_TAC[APPEND;CONS_11]);;
   e(DISCH_TAC);;
   e(IMP_RES_THEN (REWRITE_THM o SYM)
        (ASSUME "!ps'. (ps = APPEND ps'[p0]) ==>
                       (q = call_path_progress p ps' p0 rec r)"));;
   e(ASM_REWRITE_TAC[]);;

   e(REWRITE_TAC[]);;
   e(DISCH_TAC);;
   e(CONJ_TAC);;

    e(REPEAT STRIP_TAC);;
    e(FIRST_ASSUM (\th. (SPECL["p':string";"CONS (p:string) ps"]th;
                         UNDISCH_TAC (concl th))));;
    e(DISCH_THEN (MP_TAC o
         SPECL["p':string";"CONS (p:string) ps";"call_progress p' p q r";
               "BND (call_progress p' p q r) p' pcs";
               "(FST)((r:env) p')";"(FST o SND)((r:env) p')";
               "(FST o SND o SND)((r:env) p')";
               "(FST o SND o SND o SND)((r:env) p')";
               "(FST o SND o SND o SND o SND)((r:env) p')";
               "(FST o SND o SND o SND o SND o SND)((r:env) p')";
               "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
               "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
               "APPEND (APPEND ((FST)((r:env) p')) ((FST o SND)(r p')))
                       ((FST o SND o SND)(r p'))";
               "z:var list"]));;
    e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;
    e(REWRITE_TAC[LENGTH;(SYM o SPEC_ALL)ADD1;ADD_CLAUSES]);;
    e(ONCE_REWRITE_TAC[DISJOINT_LIST;SET_LIST]);;
    e(ASM_REWRITE_TAC[IN;SUBSET]);;
let st2 = get_state();;
set_state st2;;
    e(IMP_RES_TAC (AND2IMP FVa_call_progress1));;
    e(POP_TAC);;
    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[NOT_CONS_NIL]);;
    e(IMP_RES_TAC (AND2IMP CURRENT_PATH_CONDITION));;
    e(POP_ASSUM REWRITE_THM);;
    e(USE_THEN REWRITE_THM CURRENT_PATH_CONDITION_ENV
            THENL [ ASM_REWRITE_TAC[]; ALL_TAC ]);;
let st3 = get_state();;
    e(STRIP_TAC);;
    e(POP_ASSUM REWRITE_THM);;
    e(REWRITE_TAC[call_path_progress_APPEND_SINGLE]);;
    e(POP_ASSUM REWRITE_THM);;

let extend_graph_vcs_INDUCT_COVER_PATHS = top_thm();;
>%






let SOME_MAP_extend_graph_vcs_INDUCT_COVER_PATHS =
 TAC_PROOF
  (([],
    "!some_ps n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (FLAT (MAP (extend_graph_vcs p ps p0 q pcs r all_ps n)
                               some_ps)) ==>
        (!p'.
          (p' IN SL some_ps) /\ (p' IN SL all_ps) ==>
             ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0(APPEND ps'[p'])p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        )))"),

   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[SET_LIST;IN];

      REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN REWRITE_TAC[MAP;FLAT;ALL_EL_APPEND]
      THEN STRIP_TAC THEN GEN_TAC
      THEN DISCH_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[SET_LIST;IN])
      THENL
       [ UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p':string) = h"
         THEN MATCH_MP_TAC (SPEC_ALL (ONCE_REWRITE_RULE[AND_IMP_INTRO]
                      extend_graph_vcs_INDUCT_COVER_PATHS))
         THEN ASM_REWRITE_TAC[];

         FIRST_ASSUM (MATCH_MP_TAC o
                 ONCE_REWRITE_RULE[AND_IMP_INTRO] o
                 ONCE_REWRITE_RULE[AND_IMP_INTRO] o
                 CONV_RULE (ONCE_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                 CONV_RULE (RAND_CONV RIGHT_IMP_FORALL_CONV) o
                 SPEC_ALL o
                 SPEC "n:num")
         THEN ASM_REWRITE_TAC[]
       ]
    ]
  );;

%< SOME_MAP_extend_graph_vcs_INDUCT_COVER_PATHS

g "!some_ps n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (FLAT (MAP (extend_graph_vcs p ps p0 q pcs r all_ps n)
                               some_ps)) ==>
        (!p'.
          (p' IN SL some_ps) /\ (p' IN SL all_ps) ==>
             ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0(APPEND ps'[p'])p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        )))";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REPEAT GEN_TAC
    THEN STRIP_TAC);;
 e(REWRITE_TAC[MAP;FLAT;ALL_EL_APPEND]);;
 e(STRIP_TAC THEN GEN_TAC);;
 e(DISCH_THEN (STRIP_ASSUME_TAC o REWRITE_RULE[SET_LIST;IN]));;

  e(UNASSUME_THEN (REWRITE_ALL_THM o SYM) "(p':string) = h");;
  e(MATCH_MP_TAC (SPEC_ALL (ONCE_REWRITE_RULE[AND_IMP_INTRO]
                    extend_graph_vcs_INDUCT_COVER_PATHS)));;
  e(ASM_REWRITE_TAC[]);;

let st1 = get_state();;
set_state st1;;
  e(FIRST_ASSUM (MATCH_MP_TAC o
                 ONCE_REWRITE_RULE[AND_IMP_INTRO] o
                 ONCE_REWRITE_RULE[AND_IMP_INTRO] o
                 CONV_RULE (ONCE_DEPTH_CONV RIGHT_IMP_FORALL_CONV) o
                 CONV_RULE (RAND_CONV RIGHT_IMP_FORALL_CONV) o
                 SPEC_ALL o
                 SPEC "n:num"));;
  e(ASM_REWRITE_TAC[]);;

let SOME_MAP_extend_graph_vcs_INDUCT_COVER_PATHS = top_thm();;
>%





let MAP_extend_graph_vcs_INDUCT_COVER_PATHS =
 TAC_PROOF
  (([],
    "!n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (FLAT (MAP (extend_graph_vcs p ps p0 q pcs r all_ps n)
                               all_ps)) ==>
        (!p'.
          (p' IN SL all_ps) ==>
             ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0 (APPEND ps' [p']) p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        )))"),

   REWRITE_TAC[(REWRITE_RULE[(SYM o SPEC_ALL)CONJ_ASSOC] o
                  REWRITE_RULE[CONJ_ASSOC] o
                  SPEC "all_ps:string list" o
                  GEN "some_ps:string list" o SPEC_ALL)
                SOME_MAP_extend_graph_vcs_INDUCT_COVER_PATHS]
  );;

%< MAP_extend_graph_vcs_INDUCT_COVER_PATHS

g "!n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r))) /\
      (!p ps q pcs vars' vals' glbs' pre' post' calls' rec' c' y z.
        (LENGTH all_ps = (LENGTH ps) + n) /\
        p0 IN (SL(CONS p ps)) /\
        (SL(CONS p ps)) SUBSET (SL all_ps) /\
        DL(CONS p ps) /\
        (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
        (y = APPEND(APPEND vars' vals')glbs') /\
        (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
        ((ps = []) ==> (q = rec)) /\
        (!ps'.
          (ps = APPEND ps'[p0]) ==>
          (q = call_path_progress p ps' p0 rec r)) /\
        (!p'.
          p' IN (SL(CONS p ps)) /\ ~(p' = p0) ==>
          (!ps2 ps1.
            (CONS p ps = APPEND ps2(CONS p'(APPEND ps1[p0]))) ==>
            (pcs p' = call_path_progress p' ps1 p0 rec r))) ==>
        ALL_EL CLOSE(fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
          DL ps' /\ DISJOINT(SL ps')(SL(CONS p ps)) ==>
          CLOSE(AIMP(AAND pre(induct_pre rec))
                    (call_path_progress p0 ps' p q r))) /\
        (!p1 ps' ps1 ps2.
          ~(p1 = p0) /\
          DL ps' /\
          DISJOINT(SL ps')(SL(CONS p ps)) /\
          (APPEND ps'(CONS p ps) = APPEND ps2(CONS p1(APPEND ps1[p0]))) ==>
          CLOSE
          (AIMP
           (call_path_progress p1 ps1 p0 rec r)
           (call_path_progress p1(APPEND ps2(CONS p1 ps1))p0 rec r))))
      ==>
      (ALL_EL CLOSE (FLAT (MAP (extend_graph_vcs p ps p0 q pcs r all_ps n)
                               all_ps)) ==>
        (!p'.
          (p' IN SL all_ps) ==>
             ((p' = p0) =>
              CLOSE(AIMP(AAND pre (induct_pre rec))
                        (call_path_progress p0 [] p q r))
           | ((p' IN SL (CONS p ps)) =>
              (!ps2 ps1.
                (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                CLOSE(AIMP(call_path_progress p' ps1 p0 rec r)
                          (call_path_progress p' 
                                    (APPEND ps2 (CONS p' ps1)) p0 rec r)))
           |
             ((!ps'.
                DL ps' /\ DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) ==>
                CLOSE(AIMP(AAND pre (induct_pre rec))
                          (call_path_progress p0 (APPEND ps' [p']) p q r))) /\
              (!p1 ps' ps1 ps2.
                ~(p1 = p0) /\
                DL ps' /\
                DISJOINT(SL ps')(SL(CONS p' (CONS p ps))) /\
                (APPEND ps' (CONS p' (CONS p ps)) =
                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
                CLOSE
                (AIMP
                 (call_path_progress p1 ps1 p0 rec r)
                 (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0 rec r)))))
        )))";;

e(REWRITE_TAC[(REWRITE_RULE[(SYM o SPEC_ALL)CONJ_ASSOC] o
                 REWRITE_RULE[CONJ_ASSOC] o
                 SPEC "all_ps:string list" o
                 GEN "some_ps:string list" o SPEC_ALL)
              SOME_MAP_extend_graph_vcs_INDUCT_COVER_PATHS]);;
>%




let LIST_END_DECOMPOSITION =
 TAC_PROOF
  (([],
    "!t (h:*). ?b e. CONS h t = APPEND b [e]"),
   LIST_INDUCT_TAC
   THEN REPEAT GEN_TAC
   THENL
    [ EXISTS_TAC "[]:(*)list"
      THEN EXISTS_TAC "h:*"
      THEN REWRITE_TAC[APPEND];

      POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL)
      THEN EXISTS_TAC "CONS h' b:(*)list"
      THEN EXISTS_TAC "e:*"
      THEN ASM_REWRITE_TAC[APPEND]
    ]
  );;

% LIST_END_DECOMPOSITION
g "!t (h:*). ?b e. CONS h t = APPEND b [e]";;
e(LIST_INDUCT_TAC);;

 e(GEN_TAC);;
 e(EXISTS_TAC "[]:(*)list");;
 e(EXISTS_TAC "h:*");;
 e(REWRITE_TAC[APPEND]);;

 e(REPEAT GEN_TAC);;
 e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC_ALL));;
 e(EXISTS_TAC "CONS h' b:(*)list");;
 e(EXISTS_TAC "e:*");;
 e(ASM_REWRITE_TAC[APPEND]);;
%



let fan_out_graph_vcs_COVER_PATHS =
 prove_thm
  (`fan_out_graph_vcs_COVER_PATHS`,
   "!n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (p0 IN SL all_ps) /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (LENGTH all_ps = LENGTH ps + n) /\
      (p0 IN SL (CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
      (y = APPEND(APPEND vars' vals')glbs') /\
      (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r)))
      ==>
      (ALL_EL CLOSE (fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
            DL ps' /\
            DISJOINT (SL ps') (SL (CONS p ps)) ==>
            CLOSE(AIMP(AAND pre (induct_pre rec))
                      (call_path_progress p0 ps' p q r)))
        /\
        (!p1 ps' ps1 ps2.
            ~(p1 = p0) /\
            DL ps' /\
            DISJOINT (SL ps') (SL (CONS p ps)) /\
            (APPEND ps' (CONS p ps) =
             APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
            CLOSE(AIMP(call_path_progress p1 ps1 p0 rec r)
                      (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0
                                          rec r))))",

   INDUCT_TAC
   THENL
    [ REWRITE_TAC[ADD_CLAUSES]
      THEN REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN STRIP_TAC
      THEN IMP_RES_TAC DL_SUBSET_LESS_EQ_LENGTH
      THEN UNDISCH_LAST_TAC
      THEN ASM_REWRITE_TAC[LENGTH]
      THEN ONCE_REWRITE_TAC[(SYM o SPEC_ALL)NOT_SUC_LESS_EQ]
      THEN REWRITE_TAC[ADD1;LESS_EQ_MONO;LESS_EQ_REFL];

      REWRITE_TAC[fan_out_graph_vcs]
      THEN REWRITE_TAC[ADD1]
      THEN REPEAT GEN_TAC
      THEN STRIP_TAC
      THEN DISCH_TAC
      THEN FIRST_ASSUM (\th. (SPECL["p:string";"ps:string list"] th;
                              UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o
               GENL[ "p:string";"ps:string list";"q:aexp";"pcs:string->aexp"] o
               SPECL["p:string";"ps:string list";"p0:string";"q:aexp";
                     "pcs:string->aexp";"r:env";"all_ps:string list";
                     "vars:var list";"vals:var list";"glbs:var list";
                     "pre:aexp";"post:aexp";"calls:prog_env";
                     "rec:aexp";"c:cmd"])
      THEN REWRITE_TAC
               (map ASSUME ["WFenv_syntax r";
                            "WFenv_pre r";
                            "WFenv_calls r";
                            "(r:env) p0 = vars,vals,glbs,pre,post,calls,rec,c";
                            "(p0:string) IN (SL all_ps)";
                            "!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')"])
      THEN DISCH_TAC
      THEN MP_TAC (SPEC_ALL MAP_extend_graph_vcs_INDUCT_COVER_PATHS)
      THEN POP_ASSUM REWRITE_THM
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_TAC
      THEN CONJ_TAC
      THENL
       [ LIST_INDUCT_TAC
         THENL
          [ FIRST_ASSUM (MP_TAC o SPEC "p0:string")
            THEN REWRITE_TAC[ASSUME "(p0:string) IN SL all_ps"]
            THEN DISCH_THEN REWRITE_THM;

            X_GEN_TAC "p':string"
            THEN POP_TAC
            THEN STRIP_ASSUME_TAC (ISPECL["ps':string list";"p':string"]
                                     LIST_END_DECOMPOSITION)
            THEN FIRST_ASSUM REWRITE_THM
            THEN REWRITE_TAC[SL_APPEND;UNION_SUBSET;DISJOINT_LIST;
                             DL_APPEND;DISJOINT_UNION]
            THEN REWRITE_TAC[(ISPECL["e:string";"[]:string list"] o
                                 CONJUNCT2)SET_LIST;
                             CONJUNCT1 SET_LIST;IN]
            THEN REWRITE_TAC[SUBSET;DISJOINT_EMPTY;
                             DISJOINT_INSERT;DISJOINT_INSERT2]
            THEN STRIP_TAC
            THEN ASM_CASES_TAC "(e:string) IN SL all_ps"
            THENL
             [ FIRST_ASSUM (MP_TAC o SPEC "e:string")
               THEN FIRST_ASSUM REWRITE_THM
               THEN COND_CASES_TAC
               THENL
                [ POP_ASSUM REWRITE_ALL_THM
                  THEN RES_TAC;

                  ASM_REWRITE_TAC[]
                  THEN DISCH_THEN (MATCH_MP_TAC o CONJUNCT1)
                  THEN ONCE_REWRITE_TAC[SET_LIST]
                  THEN REWRITE_TAC[DISJOINT_INSERT2]
                  THEN ASM_REWRITE_TAC[]
                ];

               IMP_RES_TAC (ASSUME "!p'. ~p' IN (SL all_ps) ==>
                                         ((r:env) p' = r0 p')")
               THEN REWRITE_TAC[call_path_progress_APPEND_SINGLE]
               THEN ASM_REWRITE_TAC[call_progress;EMPTY_ENV]
               THEN CONV_TAC (DEPTH_CONV let_CONV)
               THEN BETA_TAC
               THEN REWRITE_TAC[CLOSE;A_DEF]
               THEN REPEAT STRIP_TAC
               THEN MATCH_MP_TAC A_call_path_progress_TRUE
               THEN REWRITE_TAC[A_DEF]
             ]
          ];

         CONV_TAC FORALL_SYM_CONV
         THEN LIST_INDUCT_TAC
         THENL
          [ REWRITE_TAC[DISJOINT_LIST;CONJUNCT1 SET_LIST;DISJOINT_EMPTY;APPEND]
            THEN REPEAT STRIP_TAC
            THEN ASM_CASES_TAC "(p1:string) IN SL all_ps"
            THENL
             [ FIRST_ASSUM (MP_TAC o SPEC "p1:string")
               THEN POP_ASSUM REWRITE_THM
               THEN REWRITE_TAC[ASSUME "~((p1:string) = p0)"]
               THEN FIRST_ASSUM REWRITE_THM
               THEN REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]
               THEN DISCH_THEN MATCH_MP_TAC
               THEN REFL_TAC;

               POP_ASSUM (\th1. FIRST_ASSUM (\th2. ASSUME_TAC(MATCH_MP th2 th1)))
               THEN REWRITE_TAC[CLOSE;A_DEF]
               THEN REPEAT STRIP_TAC
               THEN SPEC_TAC("ps2:string list","ps2:string list")
               THEN LIST_INDUCT_TAC
               THEN REWRITE_TAC[APPEND;call_path_progress]
               THEN REWRITE_TAC[call_progress]
               THEN REWRITE_TAC[ASSUME "(r:env) p1 = r0 p1"]
               THEN REWRITE_TAC[EMPTY_ENV]
               THEN CONV_TAC (DEPTH_CONV let_CONV)
               THEN BETA_TAC
               THEN REWRITE_TAC[A_DEF]
               THEN ONCE_REWRITE_TAC[ISPEC "(r:env) h" ENV_TUPLE]
               THEN CONV_TAC (DEPTH_CONV let_CONV)
               THEN REWRITE_TAC[A_DEF]
             ];

            X_GEN_TAC "p':string"
            THEN POP_TAC
            THEN STRIP_ASSUME_TAC (ISPECL["ps':string list";"p':string"]
                                    LIST_END_DECOMPOSITION)
            THEN FIRST_ASSUM REWRITE_THM
            THEN REWRITE_TAC[SL_APPEND;UNION_SUBSET;DISJOINT_LIST;
                             DL_APPEND;DISJOINT_UNION]
            THEN REWRITE_TAC[(ISPECL["e:string";"[]:string list"] o
                                CONJUNCT2)SET_LIST;
                             CONJUNCT1 SET_LIST;IN]
            THEN REWRITE_TAC[SUBSET;DISJOINT_EMPTY;
                             DISJOINT_INSERT;DISJOINT_INSERT2]
            THEN REPEAT STRIP_TAC
            THEN ASM_CASES_TAC "(e:string) IN SL all_ps"
            THENL
             [ FIRST_ASSUM (MP_TAC o SPEC "e:string")
               THEN FIRST_ASSUM REWRITE_THM
               THEN COND_CASES_TAC
               THENL
                [ POP_ASSUM REWRITE_ALL_THM
                  THEN RES_TAC;

                  ASM_REWRITE_TAC[]
                  THEN DISCH_THEN (MATCH_MP_TAC o CONJUNCT2)
                  THEN EXISTS_TAC "b:string list"
                  THEN ONCE_REWRITE_TAC[SET_LIST]
                  THEN REWRITE_TAC[DISJOINT_INSERT2]
                  THEN ASM_REWRITE_TAC[]
                  THEN UNDISCH_TAC "APPEND (APPEND b[e:string]) (CONS p ps) =
                                    APPEND ps2 (CONS p1 (APPEND ps1[p0]))"
                  THEN REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC]
                  THEN REWRITE_TAC[APPEND]
                ];

               IMP_RES_TAC (ASSUME "!p'. ~p' IN (SL all_ps) ==>
                                         ((r:env) p' = r0 p')")
               THEN UNDISCH_TAC "APPEND (APPEND b[e:string]) (CONS p ps) =
                                 APPEND ps2 (CONS p1 (APPEND ps1 [p0]))"
               THEN REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC]
               THEN REWRITE_TAC[APPEND]
               THEN DISCH_TAC
               THEN IMP_RES_TAC APPEND_CONS_CONS_EQ_APPEND_LESSER
               THENL
                [ FIRST_ASSUM (REWRITE_THM o SYM)
                  THEN REWRITE_TAC[REWRITE_RULE[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]
                                        call_path_progress_APPEND];

                  UNASSUME_THEN REWRITE_ALL_THM "(p:string) = p0"
                  THEN UNASSUME_THEN REWRITE_ALL_THM "(ps:string list) = []"
                  THEN FIRST_ASSUM (REWRITE_THM o SYM)
                  THEN REWRITE_TAC[REWRITE_RULE[APPEND_NIL]
                             (SPECL["b:string list";"[]:string list"]
                                  call_path_progress_APPEND)]
                ]
               THEN REWRITE_TAC[call_path_progress]
               THEN REWRITE_TAC[call_progress]
               THEN ASM_REWRITE_TAC[EMPTY_ENV]
               THEN CONV_TAC (DEPTH_CONV let_CONV)
               THEN BETA_TAC
               THEN REWRITE_TAC[CLOSE;A_DEF]
               THEN REPEAT STRIP_TAC
               THEN MP_TAC call_path_progress_TRUE
               THEN REWRITE_TAC[CLOSE]
               THEN CONV_TAC (DEPTH_CONV RIGHT_IMP_FORALL_CONV)
               THEN DISCH_THEN MATCH_MP_TAC
               THEN REWRITE_TAC[A_DEF]
             ]
          ]
       ]
    ]
  );;

%< fan_out_graph_vcs_COVER_PATHS
g "!n p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      (p0 IN SL all_ps) /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (LENGTH all_ps = LENGTH ps + n) /\
      (p0 IN SL (CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
      (y = APPEND(APPEND vars' vals')glbs') /\
      (FVa q) SUBSET (SL(APPEND y(logicals z))) /\
      ((ps = []) ==> (q = rec)) /\
      (!ps'. (ps = APPEND ps' [p0]) ==>
             (q = call_path_progress p ps' p0 rec r)) /\
      (!p'. (p' IN SL (CONS p ps)) /\ ~(p' = p0) ==>
            (!ps2 ps1. (CONS p ps = APPEND ps2 (CONS p' (APPEND ps1 [p0]))) ==>
                       (pcs p' = call_path_progress p' ps1 p0 rec r)))
      ==>
      (ALL_EL CLOSE (fan_out_graph_vcs p ps p0 q pcs r all_ps n) ==>
        (!ps'.
            DL ps' /\
            DISJOINT (SL ps') (SL (CONS p ps)) ==>
            CLOSE(AIMP(AAND pre (induct_pre rec))
                      (call_path_progress p0 ps' p q r)))
        /\
        (!p1 ps' ps1 ps2.
            ~(p1 = p0) /\
            DL ps' /\
            DISJOINT (SL ps') (SL (CONS p ps)) /\
            (APPEND ps' (CONS p ps) =
             APPEND ps2 (CONS p1 (APPEND ps1 [p0]))) ==>
            CLOSE(AIMP(call_path_progress p1 ps1 p0 rec r)
                      (call_path_progress p1 (APPEND ps2 (CONS p1 ps1)) p0
                                          rec r))))";;

e(INDUCT_TAC);;

 e(REWRITE_TAC[ADD_CLAUSES]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(STRIP_TAC);;
 e(IMP_RES_TAC DL_SUBSET_LESS_EQ_LENGTH);;
 e(UNDISCH_LAST_TAC);;
 e(ASM_REWRITE_TAC[LENGTH]);;
 e(ONCE_REWRITE_TAC[(SYM o SPEC_ALL)NOT_SUC_LESS_EQ]);;
 e(REWRITE_TAC[ADD1;LESS_EQ_MONO;LESS_EQ_REFL]);;

 e(REWRITE_TAC[fan_out_graph_vcs]);;
 e(REWRITE_TAC[ADD1]);;
 e(REPEAT GEN_TAC);;
 e(STRIP_TAC);;
 e(DISCH_TAC);;
 e(FIRST_ASSUM (\th. (SPECL["p:string";"ps:string list"] th;
                      UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o
               GENL[ "p:string";"ps:string list";"q:aexp";"pcs:string->aexp"] o
               SPECL["p:string";"ps:string list";"p0:string";"q:aexp";
                     "pcs:string->aexp";"r:env";"all_ps:string list";
                     "vars:var list";"vals:var list";"glbs:var list";
                     "pre:aexp";"post:aexp";"calls:prog_env";
                     "rec:aexp";"c:cmd"]));;
 e(REWRITE_TAC (map ASSUME ["WFenv_syntax r";
                            "WFenv_pre r";
                            "WFenv_calls r";
                            "(r:env) p0 = vars,vals,glbs,pre,post,calls,rec,c";
                            "(p0:string) IN (SL all_ps)";
                            "!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')"]));;
 e(DISCH_TAC);;
let st1 = get_state();;
set_state st1;;
 e(MP_TAC (SPEC_ALL MAP_extend_graph_vcs_INDUCT_COVER_PATHS));;
 e(POP_ASSUM REWRITE_THM);;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_TAC);;

let st2 = get_state();;
set_state st2;;
 e(CONJ_TAC);;

  e(LIST_INDUCT_TAC);;

   e(FIRST_ASSUM (MP_TAC o SPEC "p0:string"));;
   e(REWRITE_TAC[ASSUME "(p0:string) IN SL all_ps"]);;
   e(DISCH_THEN REWRITE_THM);;

   e(X_GEN_TAC "p':string");;
   e(POP_TAC);;
let st3 = get_state();;
set_state st3;;
   e(STRIP_ASSUME_TAC (ISPECL["ps':string list";"p':string"]
                        LIST_END_DECOMPOSITION));;
   e(FIRST_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[SL_APPEND;UNION_SUBSET;DISJOINT_LIST;
                 DL_APPEND;DISJOINT_UNION]);;
   e(REWRITE_TAC[(ISPECL["e:string";"[]:string list"] o CONJUNCT2)SET_LIST;
                 CONJUNCT1 SET_LIST;IN]);;
   e(REWRITE_TAC[SUBSET;DISJOINT_EMPTY;
                 DISJOINT_INSERT;DISJOINT_INSERT2]);;
   e(STRIP_TAC);;
   e(ASM_CASES_TAC "(e:string) IN SL all_ps");;

    e(FIRST_ASSUM (MP_TAC o SPEC "e:string"));;
    e(FIRST_ASSUM REWRITE_THM);;
    e(COND_CASES_TAC);;

     e(POP_ASSUM REWRITE_ALL_THM);;
     e(RES_TAC);;

     e(ASM_REWRITE_TAC[]);;
     e(DISCH_THEN (MATCH_MP_TAC o CONJUNCT1));;
     e(ONCE_REWRITE_TAC[SET_LIST]);;
     e(REWRITE_TAC[DISJOINT_INSERT2]);;
     e(ASM_REWRITE_TAC[]);;
perhaps instead:
                  REWRITE_TAC[(SYM o SPEC_ALL o CONJUNCT2)SET_LIST]
                  THEN ASM_REWRITE_TAC[]

    e(IMP_RES_TAC (ASSUME "!p'. ~p' IN (SL all_ps) ==> ((r:env) p' = r0 p')"));;
    e(REWRITE_TAC[call_path_progress_APPEND_SINGLE]);;
    e(REWRITE_TAC[call_progress]);;
    e(ASM_REWRITE_TAC[EMPTY_ENV]);;
    e(CONV_TAC (DEPTH_CONV let_CONV));;
    e(BETA_TAC);;
    e(REWRITE_TAC[CLOSE;A_DEF]);;
    e(REPEAT STRIP_TAC);;
    e(MATCH_MP_TAC A_call_path_progress_TRUE);;
    e(REWRITE_TAC[A_DEF]);;

let st4 = get_state();;
set_state st4;;
  e(CONV_TAC FORALL_SYM_CONV);;
  e(LIST_INDUCT_TAC);;

   e(REWRITE_TAC[DISJOINT_LIST;CONJUNCT1 SET_LIST;DISJOINT_EMPTY;APPEND]);;
   e(REPEAT STRIP_TAC);;
   e(ASM_CASES_TAC "(p1:string) IN SL all_ps");;

     e(FIRST_ASSUM (MP_TAC o SPEC "p1:string"));;
     e(POP_ASSUM REWRITE_THM);;
     e(REWRITE_TAC[ASSUME "~((p1:string) = p0)"]);;
     e(FIRST_ASSUM REWRITE_THM);;
     e(REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]);;
     e(DISCH_THEN MATCH_MP_TAC);;
     e(REFL_TAC);;

     e(POP_ASSUM (\th1. FIRST_ASSUM (\th2. ASSUME_TAC(MATCH_MP th2 th1))));;
     e(REWRITE_TAC[CLOSE;A_DEF]);;
     e(REPEAT STRIP_TAC);;
     e(SPEC_TAC("ps2:string list","ps2:string list"));;
     e(LIST_INDUCT_TAC);;

      e(REWRITE_TAC[APPEND;call_path_progress]);;
      e(REWRITE_TAC[call_progress]);;
      e(REWRITE_TAC[ASSUME "(r:env) p1 = r0 p1"]);;
      e(REWRITE_TAC[EMPTY_ENV]);;
      e(CONV_TAC (DEPTH_CONV let_CONV));;
      e(BETA_TAC);;
      e(REWRITE_TAC[A_DEF]);;

      e(REWRITE_TAC[APPEND;call_path_progress]);;
      e(REWRITE_TAC[call_progress]);;
      e(REWRITE_TAC[ASSUME "(r:env) p1 = r0 p1"]);;
      e(REWRITE_TAC[EMPTY_ENV]);;
      e(CONV_TAC (DEPTH_CONV let_CONV));;
      e(BETA_TAC);;
      e(REWRITE_TAC[A_DEF]);;
      e(ONCE_REWRITE_TAC[ISPEC "(r:env) h" ENV_TUPLE]);;
      e(CONV_TAC (DEPTH_CONV let_CONV));;
      e(REWRITE_TAC[A_DEF]);;

   e(X_GEN_TAC "p':string");;
   e(POP_TAC);;
   e(STRIP_ASSUME_TAC (ISPECL["ps':string list";"p':string"]
                        LIST_END_DECOMPOSITION));;
   e(FIRST_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[SL_APPEND;UNION_SUBSET;DISJOINT_LIST;
                 DL_APPEND;DISJOINT_UNION]);;
   e(REWRITE_TAC[(ISPECL["e:string";"[]:string list"] o CONJUNCT2)SET_LIST;
                 CONJUNCT1 SET_LIST;IN]);;
   e(REWRITE_TAC[SUBSET;DISJOINT_EMPTY;
                 DISJOINT_INSERT;DISJOINT_INSERT2]);;
   e(REPEAT STRIP_TAC);;
   e(ASM_CASES_TAC "(e:string) IN SL all_ps");;

    e(FIRST_ASSUM (MP_TAC o SPEC "e:string"));;
    e(FIRST_ASSUM REWRITE_THM);;
    e(COND_CASES_TAC);;

     e(POP_ASSUM REWRITE_ALL_THM);;
     e(RES_TAC);;

     e(ASM_REWRITE_TAC[]);;
     e(DISCH_THEN (MATCH_MP_TAC o CONJUNCT2));;
     e(EXISTS_TAC "b:string list");;
     e(ONCE_REWRITE_TAC[SET_LIST]);;
     e(REWRITE_TAC[DISJOINT_INSERT2]);;
     e(ASM_REWRITE_TAC[]);;
     e(UNDISCH_TAC "APPEND (APPEND b[e:string]) (CONS p ps) =
                    APPEND ps2 (CONS p1 (APPEND ps1[p0]))");;
     e(REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC]);;
     e(REWRITE_TAC[APPEND]);;

    e(IMP_RES_TAC (ASSUME "!p'. ~p' IN (SL all_ps) ==> ((r:env) p' = r0 p')"));;
    e(UNDISCH_TAC "APPEND (APPEND b[e:string]) (CONS p ps) =
                   APPEND ps2 (CONS p1 (APPEND ps1 [p0]))");;
    e(REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC]);;
    e(REWRITE_TAC[APPEND]);;
    e(DISCH_TAC);;
    e(IMP_RES_TAC APPEND_CONS_CONS_EQ_APPEND_LESSER);;

     e(FIRST_ASSUM (REWRITE_THM o SYM));;
     e(REWRITE_TAC[REWRITE_RULE[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]
                          call_path_progress_APPEND]);;
     e(REWRITE_TAC[call_path_progress]);;
     e(REWRITE_TAC[call_progress]);;
     e(ASM_REWRITE_TAC[EMPTY_ENV]);;
     e(CONV_TAC (DEPTH_CONV let_CONV));;
     e(BETA_TAC);;
     e(REWRITE_TAC[CLOSE;A_DEF]);;
     e(REPEAT STRIP_TAC);;
     e(MP_TAC call_path_progress_TRUE);;
     e(REWRITE_TAC[CLOSE]);;
     e(CONV_TAC (DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
     e(DISCH_THEN MATCH_MP_TAC);;
     e(REWRITE_TAC[A_DEF]);;

     e(UNASSUME_THEN REWRITE_ALL_THM "(p:string) = p0");;
     e(UNASSUME_THEN REWRITE_ALL_THM "(ps:string list) = []");;
     e(FIRST_ASSUM (REWRITE_THM o SYM));;
     e(REWRITE_TAC[REWRITE_RULE[APPEND_NIL]
                     (SPECL["b:string list";"[]:string list"]
                          call_path_progress_APPEND)]);;
     e(REWRITE_TAC[call_path_progress]);;
     e(REWRITE_TAC[call_progress]);;
     e(ASM_REWRITE_TAC[EMPTY_ENV]);;
     e(CONV_TAC (DEPTH_CONV let_CONV));;
     e(BETA_TAC);;
     e(REWRITE_TAC[CLOSE;A_DEF]);;
     e(REPEAT STRIP_TAC);;
     e(MP_TAC call_path_progress_TRUE);;
     e(REWRITE_TAC[CLOSE]);;
     e(CONV_TAC (DEPTH_CONV RIGHT_IMP_FORALL_CONV));;
     e(DISCH_THEN MATCH_MP_TAC);;
     e(REWRITE_TAC[A_DEF]);;
>%


%<
let extend_graph_vcs_COVER_PATHS =
 prove_thm
  (`extend_graph_vcs_COVER_PATHS`,
   "!n p' p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      p0 IN SL all_ps /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      (p' IN SL all_ps)
      ==>
      (ALL_EL CLOSE (extend_graph_vcs p ps p0 q pcs r all_ps n p') ==>
        (!ps'.
            DL ps' /\ DISJOINT (SL ps') (SL (CONS p' (CONS p ps))) ==>
            ((p' = p0) =>
                 PATH_PROGRESS(AAND pre (induct_pre rec), p0,
                               [], p, q, r)
              | (~(p' IN SL (CONS p ps)) ==>
                 PATH_PROGRESS(AAND pre (induct_pre rec), p0,
                               APPEND ps' [p'], p, q, r)))))",

   REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN REWRITE_TAC[extend_graph_vcs]
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN COND_CASES_TAC
   THENL
    [ REWRITE_TAC[ALL_EL]
      THEN REPEAT STRIP_TAC
      THEN COND_CASES_TAC
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS
         THEN EXISTS_TAC "AAND pre (call_progress p0 p q r)"
         THEN CONJ_TAC
         THENL
          [ ASM_REWRITE_TAC[CLOSE;A_DEF]
            THEN REPEAT STRIP_TAC;

            IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress)
          ];

         DISCH_TAC
         THEN MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_APPEND_SINGLE)
         THEN ASM_REWRITE_TAC[PATH_PROGRESS;A_DEF]
       ];

      POP_TAC
      THEN COND_CASES_TAC
      THENL
       [ POP_ASSUM REWRITE_ALL_THM
         THEN REWRITE_TAC[ALL_EL]
         THEN REPEAT STRIP_TAC
         THEN MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS
         THEN EXISTS_TAC "AAND pre (call_progress p0 p q r)"
         THEN CONJ_TAC
         THENL
          [ MATCH_MP_TAC CLOSE_AIMP_AAND
            THEN ASM_REWRITE_TAC[]
            THEN REWRITE_TAC[CLOSE;A_DEF]
            THEN REPEAT STRIP_TAC;

            IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress)
          ];

         COND_CASES_TAC
         THENL
          [ ASM_REWRITE_TAC[ALL_EL;SET_LIST;IN];

            REWRITE_TAC[ALL_EL]
            THEN REPEAT STRIP_TAC
            THEN MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_APPEND_SINGLE)
            THEN ASM_REWRITE_TAC[]
            THEN MATCH_MP_TAC ((REWRITE_RULE[(SYM o SPEC_ALL)CONJ_ASSOC] o
                                IMP2AND o
                                CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV
                                           THENC (DEPTH_CONV let_CONV)))
                               fan_out_graph_vcs_COVER_PATHS)
            THEN EXISTS_TAC "n:num"
            THEN EXISTS_TAC "CONS (p:string) ps"
            THEN EXISTS_TAC "all_ps:string list"
            THEN EXISTS_TAC "vars:var list"
            THEN EXISTS_TAC "vals:var list"
            THEN EXISTS_TAC "glbs:var list"
            THEN EXISTS_TAC "post:aexp"
            THEN EXISTS_TAC "calls:prog_env"
            THEN EXISTS_TAC "c:cmd"
            THEN EXISTS_TAC "(FST)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND o SND o
                                    SND o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "(FST o SND o SND o SND o
                                    SND o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "(SND o SND o SND o SND o
                                    SND o SND o SND)((r:env) p')"
            THEN EXISTS_TAC "APPEND (APPEND ((FST)((r:env) p'))
                                            ((FST o SND)(r p')))
                                    ((FST o SND o SND)(r p')): var list"
            THEN EXISTS_TAC "z:var list"
            THEN ASM_REWRITE_TAC[LENGTH;(SYM o SPEC_ALL)ADD1;
                                 ADD_CLAUSES;(SYM o SPEC_ALL)ENV_TUPLE]
            THEN ONCE_REWRITE_TAC[DISJOINT_LIST;SET_LIST]
            THEN ASM_REWRITE_TAC[IN;SUBSET]
            THEN UNDISCH_TAC "(r:env) p0 =
                              vars,vals,glbs,pre,post,calls,rec,c"
            THEN ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE)
            THEN DEFINE_NEW_VAR
                  "x = APPEND (APPEND ((FST)((r:env) p'))
                                      ((FST o SND)(r p')))
                              ((FST o SND o SND)(r p')): var list"
            THEN IMP_RES_TAC (AND2IMP FVa_call_progress)
            THEN UNASSUME_THEN (\th.ALL_TAC)
                  (concl (ISPEC "(r:env) p'" ENV_TUPLE))
            THEN UNASSUME_THEN REWRITE_ALL_THM
                  "x = APPEND (APPEND ((FST)((r:env) p'))
                                      ((FST o SND)(r p')))
                              ((FST o SND o SND)(r p')): var list"
            THEN ASM_REWRITE_TAC[]
          ]
       ]
    ]
  );;
>%

%< extend_graph_vcs_COVER_PATHS

g "!n p' p ps p0 q pcs r all_ps vars vals glbs pre post calls rec c
         vars' vals' glbs' pre' post' calls' rec' c' y z.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (r p0 = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      p0 IN SL all_ps /\
      (!p. ~(p IN SL all_ps) ==> (r p = r0 p)) /\
      (LENGTH all_ps = LENGTH ps + n + 1) /\
      p0 IN (SL(CONS p ps)) /\
      (SL (CONS p ps) SUBSET SL all_ps) /\ DL (CONS p ps) /\
      (r p = (vars',vals',glbs',pre',post',calls',rec',c')) /\
      (y = APPEND (APPEND vars' vals') glbs') /\
      (FVa q SUBSET SL(APPEND y (logicals z))) /\
      (p' IN SL all_ps)
      ==>
      (ALL_EL CLOSE (extend_graph_vcs p ps p0 q pcs r all_ps n p') ==>
        (!ps'.
            DL ps' /\ DISJOINT (SL ps') (SL (CONS p' (CONS p ps))) ==>
            ((p' = p0) =>
                 PATH_PROGRESS(AAND pre (induct_pre rec), p0,
                               [], p, q, r)
              | (~(p' IN SL (CONS p ps)) ==>
                 PATH_PROGRESS(AAND pre (induct_pre rec), p0,
                               APPEND ps' [p'], p, q, r)))))";;

e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(REWRITE_TAC[extend_graph_vcs]);;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[ALL_EL]);;
 e(REPEAT STRIP_TAC);;
 e(COND_CASES_TAC);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS);;
  e(EXISTS_TAC "AAND pre (call_progress p0 p q r)");;
  e(CONJ_TAC);;

   e(ASM_REWRITE_TAC[CLOSE;A_DEF]);;
   e(REPEAT STRIP_TAC);;

   e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress));;

  e(DISCH_TAC);;
  e(MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_APPEND_SINGLE));;
  e(ASM_REWRITE_TAC[PATH_PROGRESS;A_DEF]);;

let st1 = get_state();;
set_state st1;;
 e(POP_TAC);;
 e(COND_CASES_TAC);;

  e(POP_ASSUM REWRITE_ALL_THM);;
  e(REWRITE_TAC[ALL_EL]);;
  e(REPEAT STRIP_TAC);;
  e(MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS);;
  e(EXISTS_TAC "AAND pre (call_progress p0 p q r)");;
  e(CONJ_TAC);;

   e(MATCH_MP_TAC CLOSE_AIMP_AAND);;
   e(ASM_REWRITE_TAC[]);;
   e(REWRITE_TAC[CLOSE;A_DEF]);;
   e(REPEAT STRIP_TAC);;

   e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_progress));;

  e(COND_CASES_TAC);;

   e(REWRITE_TAC[ALL_EL]);; % yes, it actually solves this case! %

   e(REWRITE_TAC[ALL_EL]);;
   e(REPEAT STRIP_TAC);;
   e(MATCH_MP_TAC (SPEC_ALL PATH_PROGRESS_APPEND_SINGLE));;
   e(ASM_REWRITE_TAC[]);;
   e(MATCH_MP_TAC ((REWRITE_RULE[(SYM o SPEC_ALL)CONJ_ASSOC] o IMP2AND o
                    CONV_RULE (TOP_DEPTH_CONV RIGHT_IMP_FORALL_CONV
                               THENC (DEPTH_CONV let_CONV)))
                   fan_out_graph_vcs_COVER_PATHS));;
   e(EXISTS_TAC "n:num"
                     THEN EXISTS_TAC "CONS (p:string) ps"
                     THEN EXISTS_TAC "all_ps:string list"
                     THEN EXISTS_TAC "vars:var list"
                     THEN EXISTS_TAC "vals:var list"
                     THEN EXISTS_TAC "glbs:var list"
                     THEN EXISTS_TAC "post:aexp"
                     THEN EXISTS_TAC "calls:prog_env"
                     THEN EXISTS_TAC "c:cmd"
                     THEN EXISTS_TAC "(FST)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND o SND o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND o SND o
                                             SND o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "(FST o SND o SND o SND o
                                             SND o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "(SND o SND o SND o SND o
                                             SND o SND o SND)((r:env) p')"
                     THEN EXISTS_TAC "APPEND (APPEND ((FST)((r:env) p'))
                                                     ((FST o SND)(r p')))
                                             ((FST o SND o SND)(r p')): var list"
                     THEN EXISTS_TAC "z:var list");;
   e(ASM_REWRITE_TAC[LENGTH;(SYM o SPEC_ALL)ADD1;ADD_CLAUSES;
                     (SYM o SPEC_ALL)ENV_TUPLE]);;
   e(ONCE_REWRITE_TAC[DISJOINT_LIST;SET_LIST]);;
   e(ASM_REWRITE_TAC[IN;SUBSET]);;
let st1 = get_state();;
set_state st1;;
   e(UNDISCH_TAC "(r:env) p0 = vars,vals,glbs,pre,post,calls,rec,c");;
   e(ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE));;
   e(DEFINE_NEW_VAR "x = APPEND (APPEND ((FST)((r:env) p'))
                                        ((FST o SND)(r p')))
                                ((FST o SND o SND)(r p')): var list");;
   e(IMP_RES_TAC (AND2IMP FVa_call_progress));;
   e(UNASSUME_THEN (\th.ALL_TAC) (concl (ISPEC "(r:env) p'" ENV_TUPLE)));;
   e(UNASSUME_THEN REWRITE_ALL_THM
                    "x = APPEND (APPEND ((FST)((r:env) p'))
                                        ((FST o SND)(r p')))
                                ((FST o SND o SND)(r p')): var list");;
   e(ASM_REWRITE_TAC[]);;
>%




% ==================================================== %
%                                                      %
% Define the vcg function that returns the set of      %
% verification conditions needed to justify the        %
% "rec" claims in procedure headers.                   %
%                                                      %
% ==================================================== %

let graph_vcs =
 new_def
  `graph_vcs`
  "graph_vcs all_ps r p =
      let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
      fan_out_graph_vcs p [] p rec (\p.ATRUE) r all_ps (LENGTH all_ps)";;



let graph_vcs_THM =
 prove_thm
  (`graph_vcs_THM`,
   "!p r all_ps vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (p IN SL all_ps) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      ALL_EL CLOSE (graph_vcs all_ps r p) ==>
        (!ps.
            DL (APPEND ps [p]) ==>
            CLOSE(AIMP(AAND pre (induct_pre rec))
                      (call_path_progress p ps p rec r)))
        /\
        (!p' ps ps1 ps2.
            ~(p' = p) /\
            DL ps /\
            ~(p IN SL ps) /\
            (ps = APPEND ps2 (CONS p' ps1)) ==>
            CLOSE(AIMP(call_path_progress p' ps1 p rec r)
                      (call_path_progress p' (APPEND ps2 (CONS p' ps1)) p
                                          rec r)))",

   REWRITE_TAC[graph_vcs]
   THEN REPEAT GEN_TAC THEN STRIP_TAC
   THEN UNDISCH_LAST_TAC
   THEN ASM_REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN;
                        DISJOINT_INSERT2;DISJOINT_EMPTY]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN DISCH_TAC
   THEN IMP_RES_TAC (AND2IMP FVa_REC)
   THEN MP_TAC (SPECL ["LENGTH (all_ps:string list)";"p:string";
                       "[]:string list";"p:string";"rec:aexp";
                       "\p:string.ATRUE";"r:env";"all_ps:string list";
                       "vars:var list";"vals:var list";"glbs:var list";
                       "pre:aexp";"post:aexp";"calls:prog_env";
                       "rec:aexp";"c:cmd";
                       "vars:var list";"vals:var list";"glbs:var list";
                       "pre:aexp";"post:aexp";"calls:prog_env";
                       "rec:aexp";"c:cmd";
                       "APPEND (APPEND vars vals) glbs :var list";"z:var list"]
                 fan_out_graph_vcs_COVER_PATHS)
   THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[LENGTH;ADD_CLAUSES;SET_LIST;IN;SUBSET;DISJOINT_LIST]
   THEN ASM_REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL]
   THEN ONCE_REWRITE_TAC[EQ_SYM_EQ]
   THEN REWRITE_TAC[NOT_AND]
   THEN STRIP_TAC
   THEN ASM_REWRITE_TAC[]
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN EXISTS_TAC "ps:string list"
   THEN ASM_REWRITE_TAC[APPEND_CONS_APPEND_CONS]
  );;

%< graph_vcs_THM

g "!p r all_ps vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (p IN SL all_ps) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      ALL_EL CLOSE (graph_vcs all_ps r p) ==>
        (!ps.
            DL (APPEND ps [p]) ==>
            CLOSE(AIMP(AAND pre (induct_pre rec))
                      (call_path_progress p ps p rec r)))
        /\
        (!p' ps ps1 ps2.
            ~(p' = p) /\
            DL ps /\
            ~(p IN SL ps) /\
            (ps = APPEND ps2 (CONS p' ps1)) ==>
            CLOSE(AIMP(call_path_progress p' ps1 p rec r)
                      (call_path_progress p' (APPEND ps2 (CONS p' ps1)) p
                                          rec r)))";;

e(REWRITE_TAC[graph_vcs]);;
e(REPEAT GEN_TAC THEN STRIP_TAC);;
e(UNDISCH_LAST_TAC);;
e(ASM_REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN;
                  DISJOINT_INSERT2;DISJOINT_EMPTY]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(DISCH_TAC);;
e(IMP_RES_TAC (AND2IMP FVa_REC));;
e(MP_TAC (SPECL["LENGTH (all_ps:string list)";"p:string";
                "[]:string list";"p:string";"rec:aexp";
                "\p:string.ATRUE";"r:env";"all_ps:string list";
                "vars:var list";"vals:var list";"glbs:var list";
                "pre:aexp";"post:aexp";"calls:prog_env";
                "rec:aexp";"c:cmd";
                "vars:var list";"vals:var list";"glbs:var list";
                "pre:aexp";"post:aexp";"calls:prog_env";
                "rec:aexp";"c:cmd";
                "APPEND (APPEND vars vals) glbs :var list";"z:var list"]
          fan_out_graph_vcs_COVER_PATHS));;
e(ASM_REWRITE_TAC[]);;
e(REWRITE_TAC[LENGTH;ADD_CLAUSES;SET_LIST;IN;SUBSET;DISJOINT_LIST]);;
e(ASM_REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(REWRITE_TAC[APPEND_EQ_NIL;NOT_CONS_NIL]);;
e(ONCE_REWRITE_TAC[EQ_SYM_EQ]);;
e(REWRITE_TAC[NOT_AND]);;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(EXISTS_TAC "ps:string list");;
e(ASM_REWRITE_TAC[APPEND_CONS_APPEND_CONS]);;
>%



let call_progress_IMP_LEMMA1 =
 TAC_PROOF
  (([],
    "!p' p q (r:env) n x y z
      vars  vals  glbs  pre  post  calls  rec  c
      vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p' = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       (r p  = (vars',vals',glbs',pre',post',calls',rec',c')) /\
       (x = APPEND (APPEND vars vals) glbs) /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (FVa q) SUBSET (SL (APPEND y (logicals z))) /\
       (LENGTH n = LENGTH y) ==>
       (A q (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q))))) =
        A q (BNDsL n y s))"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[s_subst_var_list]
   THEN USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE
   THENL
    [ MATCH_MP_TAC WFxs_NOT_WFxs
      THEN REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
      THEN UNDISCH_TAC "(r:env) p' = vars,vals,glbs,pre,post,calls,rec,c"
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN ASM_REWRITE_TAC[];

      USE_IMP_THEN REWRITE_THM A_BNDsR_THM
      THEN REWRITE_TAC[DISJOINT_variants]
    ]
  );;

%<
g "!p' p q (r:env) n x y z
      vars  vals  glbs  pre  post  calls  rec  c
      vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p' = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       (r p  = (vars',vals',glbs',pre',post',calls',rec',c')) /\
       (x = APPEND (APPEND vars vals) glbs) /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (FVa q) SUBSET (SL (APPEND y (logicals z))) /\
       (LENGTH n = LENGTH y) ==>
       (A q (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q))))) =
        A q (BNDsL n y s))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[s_subst_var_list]);;
e(USE_IMP_THEN REWRITE_THM BNDsL_BNDsR_COMMUTE);;

 e(MATCH_MP_TAC WFxs_NOT_WFxs);;
 e(REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]);;
 e(UNDISCH_TAC "(r:env) p' = vars,vals,glbs,pre,post,calls,rec,c");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(ASM_REWRITE_TAC[]);;

 e(USE_IMP_THEN REWRITE_THM A_BNDsR_THM);;
 e(REWRITE_TAC[DISJOINT_variants]);;
>%


let call_progress_IMP_LEMMA2 =
 TAC_PROOF
  (([],
    "!p' p q1 q2 (r:env) n x y
      vars  vals  glbs  pre  post  calls  rec  c
      vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p' = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       (r p  = (vars',vals',glbs',pre',post',calls',rec',c')) /\
       (x = APPEND (APPEND vars vals) glbs) /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (LENGTH n = LENGTH y) ==>
       (A ((calls p) <|a ((variants (logicals x) (FVa q2)) // (logicals x)))
          (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q2))))) =
        A ((calls p) <|a ((variants (logicals x) (FVa q1)) // (logicals x)))
          (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q1))))))"),

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[a_subst_THM]
   THEN REWRITE_TAC[s_subst_var_list]
   THEN MATCH_MP_TAC A_EQ_THM
   THEN MATCH_MP_TAC MAP_EXTENSION_SUBSET
   THEN EXISTS_TAC "APPEND y (logicals x):var list"
   THEN CONJ_TAC
   THENL
    [ REWRITE_TAC[MAP_APPEND]
      THEN USE_IMP_AND_THEN REWRITE_THM APPEND_11
                  (REWRITE_TAC[LENGTH_MAP])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                     THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                          LENGTH_logicals;DL_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (ONCE_REWRITE_TAC[DISJOINT_SYM]
                     THEN MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                     LENGTH_logicals;DL_variants])
      THEN USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                     THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                          LENGTH_logicals;DL_logicals])
      THEN USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (ONCE_REWRITE_TAC[DISJOINT_SYM]
                     THEN MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX)
      THEN USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
      THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                           LENGTH_logicals;DL_variants];

      UNDISCH_TAC "(r:env) p = vars',vals',glbs',pre',post',calls',rec',c'"
      THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
      THEN DISCH_THEN (\th. UNDISCH_LAST_TAC THEN ASSUME_TAC th)
      THEN DISCH_THEN (MP_TAC o SPEC "p:string")
      THEN ASM_REWRITE_TAC[]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[]
    ]
  );;

%<
g "!p' p q1 q2 (r:env) n x y
      vars  vals  glbs  pre  post  calls  rec  c
      vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p' = (vars,vals,glbs,pre,post,calls,rec,c)) /\
       (r p  = (vars',vals',glbs',pre',post',calls',rec',c')) /\
       (x = APPEND (APPEND vars vals) glbs) /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (LENGTH n = LENGTH y) ==>
       (A ((calls p) <|a ((variants (logicals x) (FVa q2)) // (logicals x)))
          (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q2))))) =
        A ((calls p) <|a ((variants (logicals x) (FVa q1)) // (logicals x)))
          (BNDsL n y (s <|s (x // (variants (logicals x) (FVa q1))))))";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[a_subst_THM]);;
e(REWRITE_TAC[s_subst_var_list]);;
e(MATCH_MP_TAC A_EQ_THM);;
e(MATCH_MP_TAC MAP_EXTENSION_SUBSET);;
e(EXISTS_TAC "APPEND y (logicals x):var list");;
e(CONJ_TAC);;

  e(REWRITE_TAC[MAP_APPEND]);;
  e(USE_IMP_THEN REWRITE_THM APPEND_11);;

   e(REWRITE_TAC[LENGTH_MAP]);;

   e(USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_AND_THEN REWRITE_THM MAP_BNDsR_NOT_MATCH
                    (MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsL_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                     THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                          LENGTH_logicals;DL_logicals]));;
   e(USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (ONCE_REWRITE_TAC[DISJOINT_SYM]
                     THEN MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                     LENGTH_logicals;DL_variants]));;
   e(USE_IMP_AND_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
                    (IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX
                     THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                                          LENGTH_logicals;DL_logicals]));;
   e(USE_IMP_AND_THEN REWRITE_THM MAP_BNDsL_NOT_MATCH
                    (ONCE_REWRITE_TAC[DISJOINT_SYM]
                     THEN MATCH_MP_TAC WFxs_NOT_WFxs
                     THEN ASM_REWRITE_TAC[NOT_WFxs_variants;NOT_WFxs_logicals]
                     THEN IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX));;
   e(USE_IMP_THEN REWRITE_THM (IMP2AND MAP_BNDsR_IDENT)
      THEN ASM_REWRITE_TAC[LENGTH_MAP;LENGTH_variants;
                           LENGTH_logicals;DL_variants]);;

 e(UNDISCH_TAC "(r:env) p = vars',vals',glbs',pre',post',calls',rec',c'");;
 e(IMP_RES_TAC WELL_FORMED_PROCEDURE_SYNTAX);;
 e(DISCH_THEN (\th. UNDISCH_LAST_TAC THEN ASSUME_TAC th));;
 e(DISCH_THEN (MP_TAC o SPEC "p:string"));;
 e(ASM_REWRITE_TAC[]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[]);;
>%



let call_progress_AIMP =
 prove_thm
  (`call_progress_AIMP`,
   "!p' p q1 q2 r y z vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (FVa q1 SUBSET SL(APPEND y (logicals z))) /\
       (FVa q2 SUBSET SL(APPEND y (logicals z))) /\
       CLOSE(AIMP q1 q2) ==>
       CLOSE(AIMP (call_progress p' p q1 r)
                  (call_progress p' p q2 r))",

   REWRITE_TAC[CLOSE;A_DEF]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN GEN_TAC
   THEN REWRITE_TAC[call_progress]
   THEN ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE)
   THEN POP_ASSUM ONCE_REWRITE_THM
   THEN ASM_REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN DEFINE_NEW_VAR
           "x = APPEND (APPEND ((FST)(r p')) ((FST o SND)(r p')))
                       ((FST o SND o SND)((r:env) p'))"
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN REWRITE_TAC[(SYM o ASSUME)
           "y = APPEND(APPEND vars' vals')glbs' :var list"]
   THEN COND_CASES_TAC
   THENL
    [ REWRITE_TAC[A_DEF];

      POP_TAC
      THEN REWRITE_TAC[a_subst_THM]
      THEN REWRITE_TAC[A_mk_foralls]
      THEN REWRITE_TAC[A_DEF]
      THEN REPEAT STRIP_TAC
      THEN USE_IMP_AND_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q2:aexp";"r:env";"n:num list";
                  "x:var list";"y:var list";"z:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA1)
           (ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE])
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_LAST_TAC
      THEN USE_IMP_AND_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q1:aexp";"q2:aexp";"r:env";
                  "n:num list";"x:var list";"y:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA2)
           (ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE])
      THEN DISCH_TAC
      THEN RES_TAC
      THEN UNDISCH_LAST_TAC
      THEN USE_IMP_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q1:aexp";"r:env";"n:num list";
                  "x:var list";"y:var list";"z:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA1)
      THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]
    ]
  );;

%<
g "!p' p q1 q2 r y z vars' vals' glbs' pre' post' calls' rec' c'.
       WFenv_syntax r /\
       (r p = vars',vals',glbs',pre',post',calls',rec',c') /\
       (y = APPEND (APPEND vars' vals') glbs') /\
       (FVa q1 SUBSET SL(APPEND y (logicals z))) /\
       (FVa q2 SUBSET SL(APPEND y (logicals z))) /\
       CLOSE(AIMP q1 q2) ==>
       CLOSE(AIMP (call_progress p' p q1 r)
                  (call_progress p' p q2 r))";;
e(REWRITE_TAC[CLOSE;A_DEF]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(GEN_TAC);;
e(REWRITE_TAC[call_progress]);;
e(ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE));;
e(POP_ASSUM ONCE_REWRITE_THM);;
e(ASM_REWRITE_TAC[]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(DEFINE_NEW_VAR "x = APPEND (APPEND ((FST)(r p')) ((FST o SND)(r p')))
                             ((FST o SND o SND)((r:env) p'))");;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(REWRITE_TAC[(SYM o ASSUME)"y = APPEND(APPEND vars' vals')glbs' :var list"]);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[A_DEF]);;

 e(POP_TAC);;
 e(REWRITE_TAC[a_subst_THM]);;
 e(REWRITE_TAC[A_mk_foralls]);;
 e(REWRITE_TAC[A_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(USE_IMP_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q2:aexp";"r:env";"n:num list";
                  "x:var list";"y:var list";"z:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA1));;

  e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;

  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(UNDISCH_LAST_TAC);;
  e(USE_IMP_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q1:aexp";"q2:aexp";"r:env";
                  "n:num list";"x:var list";"y:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA2));;

   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;

   e(DISCH_TAC);;
   e(RES_TAC);;
   e(UNDISCH_LAST_TAC);;
   e(USE_IMP_THEN REWRITE_THM
           (SPECL["p':string";"p:string";"q1:aexp";"r:env";"n:num list";
                  "x:var list";"y:var list";"z:var list";
                  "(FST)((r:env) p')";
                  "(FST o SND)((r:env) p')";
                  "(FST o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND)((r:env) p')";
                  "(FST o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "(SND o SND o SND o SND o SND o SND o SND)((r:env) p')";
                  "vars':var list";"vals':var list";"glbs':var list";
                  "pre':aexp";"post':aexp";"calls':prog_env";
                  "rec':aexp";"c':cmd"]
            call_progress_IMP_LEMMA1));;
   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)ENV_TUPLE]);;

>%


let call_path_progress_AIMP =
 prove_thm
  (`call_path_progress_AIMP`,
   "!ps p' p q1 q2 r y z vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       (y = APPEND (APPEND vars vals) glbs) /\
       (FVa q1 SUBSET SL(APPEND y (logicals z))) /\
       (FVa q2 SUBSET SL(APPEND y (logicals z))) /\
       CLOSE(AIMP q1 q2) ==>
       CLOSE(AIMP (call_path_progress p' ps p q1 r)
                  (call_path_progress p' ps p q2 r))",

   LIST_INDUCT_TAC
   THENL
    [ REWRITE_TAC[call_path_progress]
      THEN REWRITE_TAC[call_progress_AIMP];

      REPEAT STRIP_TAC
      THEN REWRITE_TAC[call_path_progress]
      THEN ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE)
      THEN DEFINE_NEW_VAR
               "x = APPEND (APPEND ((FST)(r h)) ((FST o SND)(r h)))
                           ((FST o SND o SND)((r:env) h))"
      THEN IMP_RES_TAC (AND2IMP FVa_call_path_progress)
      THEN POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_ASSUM (MP_TAC o SPEC_ALL)
      THEN POP_ASSUM (MP_TAC o SPEC_ALL)
      THEN POP_TAC
      THEN REPEAT DISCH_TAC
      THEN IMP_RES_TAC (AND2IMP call_progress_AIMP)
      THEN POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_ASSUM MATCH_MP_TAC
      THEN POP_TAC
      THEN FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< call_path_progress_AIMP
g "!ps p' p q1 q2 r y z vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       (y = APPEND (APPEND vars vals) glbs) /\
       (FVa q1 SUBSET SL(APPEND y (logicals z))) /\
       (FVa q2 SUBSET SL(APPEND y (logicals z))) /\
       CLOSE(AIMP q1 q2) ==>
       CLOSE(AIMP (call_path_progress p' ps p q1 r)
                  (call_path_progress p' ps p q2 r))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[call_path_progress]);;
 e(REWRITE_TAC[call_progress_AIMP]);;

 e(REPEAT STRIP_TAC);;
 e(REWRITE_TAC[call_path_progress]);;
 e(ASSUME_TAC (ISPEC "(r:env) h" ENV_TUPLE));;
 e(DEFINE_NEW_VAR "x = APPEND (APPEND ((FST)(r h)) ((FST o SND)(r h)))
                              ((FST o SND o SND)((r:env) h))");;
 e(IMP_RES_TAC (AND2IMP FVa_call_path_progress));;
 e(POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(POP_ASSUM (MP_TAC o SPEC_ALL));;
 e(POP_ASSUM (MP_TAC o SPEC_ALL));;
 e(POP_TAC);;
 e(REPEAT DISCH_TAC);;
 e(IMP_RES_TAC (AND2IMP call_progress_AIMP));;
 e(POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(POP_ASSUM MATCH_MP_TAC);;
 e(POP_TAC);;
 e(FIRST_ASSUM (MATCH_MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;
>%




let SPLIT_LIST =
 prove_thm
  (`SPLIT_LIST`,
   "!ps (p:string).
       (p IN SL ps) ==>
       ?ps2 ps1.
           (ps = APPEND ps2 (CONS p ps1)) /\
           ~(p IN SL ps1)",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "(p:string) IN SL ps"
   THENL
    [ FIRST_ASSUM REWRITE_THM
      THEN RES_TAC
      THEN EXISTS_TAC "CONS h ps2:string list"
      THEN EXISTS_TAC "ps1:string list"
      THEN ASM_REWRITE_TAC[APPEND];

      FIRST_ASSUM REWRITE_THM
      THEN DISCH_THEN (REWRITE_THM o SYM)
      THEN EXISTS_TAC "[]:string list"
      THEN EXISTS_TAC "ps:string list"
      THEN ASM_REWRITE_TAC[APPEND]
    ]
  );;

%< SPLIT_LIST
g "!ps (p:string).
       (p IN SL ps) ==>
       ?ps2 ps1.
           (ps = APPEND ps2 (CONS p ps1)) /\
           ~(p IN SL ps1)";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN]);;
 e(REPEAT GEN_TAC);;
 e(ASM_CASES_TAC "(p:string) IN SL ps");;

  e(FIRST_ASSUM REWRITE_THM);;
  e(RES_TAC);;
  e(EXISTS_TAC "CONS h ps2:string list");;
  e(EXISTS_TAC "ps1:string list");;
  e(ASM_REWRITE_TAC[APPEND]);;

  e(FIRST_ASSUM REWRITE_THM);;
  e(DISCH_THEN (REWRITE_THM o SYM));;
  e(EXISTS_TAC "[]:string list");;
  e(EXISTS_TAC "ps:string list");;
  e(ASM_REWRITE_TAC[APPEND]);;
>%


let SPLIT_NOT_DL_LIST =
 prove_thm
  (`SPLIT_NOT_DL_LIST`,
   "!ps:string list.
       ~DL ps ==>
       (?p' ps3 ps2 ps1.
            (ps = APPEND ps3 (CONS p' (APPEND ps2 (CONS p' ps1)))) /\
            DL (APPEND ps2 (CONS p' ps1)))",

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[DISJOINT_LIST;DE_MORGAN_THM]
   THEN REPEAT GEN_TAC
   THEN ASM_CASES_TAC "DL (ps:string list)"
   THENL
    [ FIRST_ASSUM REWRITE_THM
      THEN STRIP_TAC
      THEN IMP_RES_TAC SPLIT_LIST
      THEN EXISTS_TAC "h:string"
      THEN EXISTS_TAC "[]:string list"
      THEN EXISTS_TAC "ps2:string list"
      THEN EXISTS_TAC "ps1:string list"
      THEN ASM_REWRITE_TAC[APPEND]
      THEN UNDISCH_TAC "DL (ps:string list)"
      THEN ASM_REWRITE_TAC[];

      FIRST_ASSUM REWRITE_THM
      THEN RES_TAC
      THEN EXISTS_TAC "p':string"
      THEN EXISTS_TAC "CONS (h:string) ps3"
      THEN EXISTS_TAC "ps2:string list"
      THEN EXISTS_TAC "ps1:string list"
      THEN ASM_REWRITE_TAC[APPEND]
    ]
  );;

%< SPLIT_NOT_DL_LIST
g "!ps:string list.
       ~DL ps ==>
       (?p' ps3 ps2 ps1.
            (ps = APPEND ps3 (CONS p' (APPEND ps2 (CONS p' ps1)))) /\
            DL (APPEND ps2 (CONS p' ps1)))";;
e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[DISJOINT_LIST]);;

 e(REWRITE_TAC[DISJOINT_LIST;DE_MORGAN_THM]);;
 e(REPEAT GEN_TAC);;
 e(ASM_CASES_TAC "DL (ps:string list)");;

  e(FIRST_ASSUM REWRITE_THM);;
  e(STRIP_TAC);;
  e(IMP_RES_TAC SPLIT_LIST);;
  e(EXISTS_TAC "h:string"
       THEN EXISTS_TAC "[]:string list"
       THEN EXISTS_TAC "ps2:string list"
       THEN EXISTS_TAC "ps1:string list");;
  e(ASM_REWRITE_TAC[APPEND]);;
  e(UNDISCH_TAC "DL (ps:string list)");;
  e(ASM_REWRITE_TAC[]);;

  e(FIRST_ASSUM REWRITE_THM);;
  e(RES_TAC);;
  e(EXISTS_TAC "p':string"
       THEN EXISTS_TAC "CONS (h:string) ps3"
       THEN EXISTS_TAC "ps2:string list"
       THEN EXISTS_TAC "ps1:string list");;
  e(ASM_REWRITE_TAC[APPEND]);;
>%


let CALL_PATH_CASES =
 prove_thm
  (`CALL_PATH_CASES`,
   "!ps (p:string).
       (DL (APPEND ps [p])) \/
       (?ps2 ps1.
            (ps = APPEND ps2 (CONS p ps1)) /\
            DL (APPEND ps1 [p])) \/
       (?p' ps3 ps2 ps1.
            ~(p' = p) /\
            (ps = APPEND ps3 (CONS p' (APPEND ps2 (CONS p' ps1)))) /\
            DL (APPEND ps2 (CONS p' ps1)) /\
            ~(p IN SL(APPEND ps2 (CONS p' ps1))))",

   REPEAT GEN_TAC
   THEN ASM_CASES_TAC "DL (APPEND ps [p:string])"
   THEN ASM_REWRITE_TAC[]
   THEN FIRST_ASSUM (MP_TAC o REWRITE_RULE[DL_APPEND;DISJOINT_LIST;SET_LIST;IN])
   THEN DISCH_THEN (MP_TAC o
              REWRITE_RULE[DISJOINT_INSERT2;DISJOINT_EMPTY;DE_MORGAN_THM])
   THEN ASM_CASES_TAC "DL (ps:string list)"
   THENL
    [ FIRST_ASSUM REWRITE_THM
      THEN DISCH_TAC
      THEN DISJ1_TAC
      THEN IMP_RES_TAC SPLIT_LIST
      THEN EXISTS_TAC "ps2:string list"
      THEN EXISTS_TAC "ps1:string list"
      THEN ASM_REWRITE_TAC[]
      THEN UNDISCH_TAC "DL (ps:string list)"
      THEN ASM_REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN]
      THEN ASM_REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]
      THEN STRIP_TAC;

      FIRST_ASSUM REWRITE_THM
      THEN UNDISCH_TAC "~DL(APPEND ps [p:string])"
      THEN IMP_RES_TAC SPLIT_NOT_DL_LIST
      THEN DISCH_TAC
      THEN ASM_CASES_TAC "(p:string) IN SL(APPEND ps2 (CONS p' ps1))"
      THENL
       [ DISJ1_TAC
         THEN IMP_RES_TAC SPLIT_LIST
         THEN EXISTS_TAC "APPEND ps3 (CONS p' ps2'):string list"
         THEN EXISTS_TAC "ps1':string list"
         THEN UNDISCH_TAC "DL(APPEND ps2 (CONS (p':string) ps1))"
         THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]
         THEN REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN]
         THEN REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]
         THEN STRIP_TAC
         THEN ASM_REWRITE_TAC[];

         DISJ2_TAC
         THEN EXISTS_TAC "p':string"
         THEN EXISTS_TAC "ps3:string list"
         THEN EXISTS_TAC "ps2:string list"
         THEN EXISTS_TAC "ps1:string list"
         THEN ASM_REWRITE_TAC[]
         THEN UNDISCH_LAST_TAC
         THEN REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]
         THEN REWRITE_TAC[DE_MORGAN_THM]
         THEN STRIP_TAC
         THEN FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM)
       ]
    ]
  );;

%< CALL_PATH_CASES
g "!ps (p:string).
       (DL (APPEND ps [p])) \/
       (?ps2 ps1.
            (ps = APPEND ps2 (CONS p ps1)) /\
            DL (APPEND ps1 [p])) \/
       (?p' ps3 ps2 ps1.
            ~(p' = p) /\
            (ps = APPEND ps3 (CONS p' (APPEND ps2 (CONS p' ps1)))) /\
            DL (APPEND ps2 (CONS p' ps1)) /\
            ~(p IN SL(APPEND ps2 (CONS p' ps1))))";;
e(REPEAT GEN_TAC);;
e(ASM_CASES_TAC "DL (APPEND ps [p:string])");;

 e(ASM_REWRITE_TAC[]);;

 e(ASM_REWRITE_TAC[]);;
 e(FIRST_ASSUM (MP_TAC o REWRITE_RULE[DL_APPEND;DISJOINT_LIST;SET_LIST;IN]));;
 e(DISCH_THEN (MP_TAC o
       REWRITE_RULE[DISJOINT_INSERT2;DISJOINT_EMPTY;DE_MORGAN_THM]));;
 e(ASM_CASES_TAC "DL (ps:string list)");;

  e(FIRST_ASSUM REWRITE_THM);;
  e(DISCH_TAC);;
  e(DISJ1_TAC);;
  e(IMP_RES_TAC SPLIT_LIST);;
  e(EXISTS_TAC "ps2:string list"
       THEN EXISTS_TAC "ps1:string list");;
  e(ASM_REWRITE_TAC[]);;
  e(UNDISCH_TAC "DL (ps:string list)");;
  e(ASM_REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN]);;
  e(ASM_REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]);;
  e(STRIP_TAC);;

  e(FIRST_ASSUM REWRITE_THM);;
  e(UNDISCH_TAC "~DL(APPEND ps [p:string])");;
  e(IMP_RES_TAC SPLIT_NOT_DL_LIST);;
  e(DISCH_TAC);;
  e(ASM_CASES_TAC "(p:string) IN SL(APPEND ps2 (CONS p' ps1))");;

   e(DISJ1_TAC);;
   e(IMP_RES_TAC SPLIT_LIST);;
   e(EXISTS_TAC "APPEND ps3 (CONS p' ps2'):string list"
       THEN EXISTS_TAC "ps1':string list");;
   e(UNDISCH_TAC "DL(APPEND ps2 (CONS (p':string) ps1))");;
   e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)APPEND_ASSOC;APPEND]);;
   e(REWRITE_TAC[DL_APPEND;DISJOINT_LIST;SET_LIST;IN]);;
   e(REWRITE_TAC[DISJOINT_INSERT2;DISJOINT_EMPTY]);;
   e(STRIP_TAC);;
   e(ASM_REWRITE_TAC[]);;

   e(DISJ2_TAC);;
   e(EXISTS_TAC "p':string"
       THEN EXISTS_TAC "ps3:string list"
       THEN EXISTS_TAC "ps2:string list"
       THEN EXISTS_TAC "ps1:string list");;
   e(ASM_REWRITE_TAC[]);;
   e(UNDISCH_LAST_TAC);;
   e(REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]);;
   e(REWRITE_TAC[DE_MORGAN_THM]);;
   e(STRIP_TAC);;
   e(FIRST_ASSUM (REWRITE_THM o NOT_EQ_SYM));;
>%





let WELL_FOUNDED_NUM_TAC :tactic = \(asl,gl).
     (let n = (fst o dest_forall) gl in
        (GEN_TAC
         THEN UNBETA_TAC n
         THEN SPEC_TAC(n,n)
         THEN MATCH_MP_TAC NUM_WELL_FOUNDED
         THEN BETA_TAC
         THEN GEN_TAC
         THEN DISCH_TAC)) (asl,gl)
      ? failwith `WELL_FOUNDED_NUM_TAC`;;






let SINGLY_RECURSIVE_call_path_progress =
 prove_thm
  (`SINGLY_RECURSIVE_call_path_progress`,
   "!n ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       (LENGTH ps = n) /\
       p IN (SL all_ps) /\
       ~(p IN SL ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       CLOSE(AIMP(AAND pre (induct_pre rec))
                 (call_path_progress p ps p rec r))",

   WELL_FOUNDED_NUM_TAC
   THEN REPEAT STRIP_TAC
   THEN IMP_RES_TAC (AND2IMP graph_vcs_THM)
   THEN STRIP_ASSUME_TAC (SPEC_ALL CALL_PATH_CASES)
   THENL  % 3 subgoals %
    [ FIRST_ASSUM (IMP_RES_TAC o SPEC "ps:string list");

      UNDISCH_TAC "~((p:string) IN SL ps)"
      THEN ASM_REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN];

      FIRST_ASSUM (\th. (SPEC "p':string" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_THEN (MP_TAC o SPEC "APPEND ps2 (CONS p' ps1) :string list")
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_THEN (MP_TAC o SPECL["ps2:string list";"ps1:string list"])
      THEN REWRITE_TAC[]
      THEN DISCH_TAC
      THEN DEFINE_NEW_VAR "y = APPEND (APPEND vars vals) glbs :var list"
      THEN IMP_RES_TAC (AND2IMP FVa_REC)
      THEN UNDISCH_LAST_TAC
      THEN FIRST_ASSUM (REWRITE_THM o SYM)
      THEN DISCH_TAC
      THEN DEFINE_NEW_VAR
                "x = APPEND (APPEND ((FST)(r p')) ((FST o SND)(r p')))
                                    ((FST o SND o SND)((r:env) p'))"
      THEN ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE)
      THEN IMP_RES_TAC (AND2IMP FVa_call_path_progress)
      THEN POP_TAC THEN POP_TAC THEN UNDISCH_LAST_TAC THEN POP_TAC
      THEN DISCH_THEN
               (\th. ASSUME_TAC (SPEC "ps1:string list" th)
                     THEN ASSUME_TAC
                              (SPEC "APPEND ps2 (CONS (p':string) ps1)" th))
      THEN IMP_RES_TAC (AND2IMP call_path_progress_AIMP)
      THEN POP_TAC THEN POP_TAC THEN UNDISCH_LAST_TAC
      THEN POP_TAC THEN POP_TAC
      THEN DISCH_THEN (MP_TAC o SPECL["p:string";"ps3:string list"])
      THEN REWRITE_TAC[(SYM o SPEC_ALL)call_path_progress_APPEND1]
      THEN POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
      THEN POP_TAC THEN POP_TAC THEN POP_TAC
      THEN FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o
               SPEC "LENGTH (APPEND ps3 (CONS p' ps1) :string list)")
      THEN REWRITE_TAC[(SYM o ASSUME)"LENGTH (ps:string list) = n"]
      THEN REWRITE_TAC[ASSUME
               "ps = APPEND ps3 (CONS (p':string) (APPEND ps2 (CONS p' ps1)))"]
      THEN REWRITE_TAC[LENGTH_APPEND;LENGTH;ADD_CLAUSES]
      THEN SUBST1_TAC (SPECL["LENGTH (ps2:string list)";
                             "LENGTH (ps1:string list)"] ADD_SYM)
      THEN REWRITE_TAC[ADD_ASSOC;LESS_MONO_EQ]
      THEN REWRITE_TAC[ADD_SUC;LESS_ADD_SUC]
      THEN DISCH_THEN (MP_TAC o
            SPECL["APPEND ps3 (CONS p' ps1):string list";"p:string";
                  "r:env";"all_ps:string list";
                  "vars:var list";"vals:var list";"glbs:var list";
                  "pre:aexp";"post:aexp";"calls:prog_env";
                  "rec:aexp";"c:cmd"])
      THEN ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH]
      THEN UNDISCH_TAC "~((p:string) IN SL ps)"
      THEN ASM_REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN;DE_MORGAN_THM]
      THEN DISCH_THEN REWRITE_THM
      THEN REWRITE_TAC[CLOSE;A_DEF]
      THEN REPEAT STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< SINGLY_RECURSIVE_call_path_progress
g "!n ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       (LENGTH ps = n) /\
       p IN (SL all_ps) /\
       ~(p IN SL ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       CLOSE(AIMP(AAND pre (induct_pre rec))
                 (call_path_progress p ps p rec r))";;

e(WELL_FOUNDED_NUM_TAC);;
e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC (AND2IMP graph_vcs_THM));;
e(STRIP_ASSUME_TAC (SPEC_ALL CALL_PATH_CASES));;
% 3 subgoals %

 e(FIRST_ASSUM (IMP_RES_TAC o SPEC "ps:string list"));;

 e(UNDISCH_TAC "~((p:string) IN SL ps)");;
 e(ASM_REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN]);;

let st1 = get_state();;
set_state st1;;
e(FIRST_ASSUM (\th. (SPEC "p':string" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC_ALL)
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_THEN (MP_TAC o SPEC "APPEND ps2 (CONS p' ps1) :string list")
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_THEN (MP_TAC o SPECL["ps2:string list";"ps1:string list"])
      THEN REWRITE_TAC[]
      THEN DISCH_TAC);;

 e(FIRST_ASSUM (\th. (SPEC "p':string" th; UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o SPEC_ALL));;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_THEN (MP_TAC o SPEC "APPEND ps2 (CONS p' ps1) :string list"));;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_THEN (MP_TAC o SPECL["ps2:string list";"ps1:string list"]));;
 e(REWRITE_TAC[]);;
 e(DISCH_TAC);;
 e(DEFINE_NEW_VAR "y = APPEND (APPEND vars vals) glbs :var list");;
 e(IMP_RES_TAC (AND2IMP FVa_REC));;
 e(UNDISCH_LAST_TAC
     THEN FIRST_ASSUM (REWRITE_THM o SYM)
     THEN DISCH_TAC);;
 e(DEFINE_NEW_VAR "x = APPEND (APPEND ((FST)(r p')) ((FST o SND)(r p')))
                              ((FST o SND o SND)((r:env) p'))");;
 e(ASSUME_TAC (ISPEC "(r:env) p'" ENV_TUPLE));;
 e(IMP_RES_TAC (AND2IMP FVa_call_path_progress));;
 e(POP_TAC THEN POP_TAC THEN UNDISCH_LAST_TAC THEN POP_TAC);;
 e(DISCH_THEN
      (\th. ASSUME_TAC (SPEC "ps1:string list" th)
            THEN ASSUME_TAC (SPEC "APPEND ps2 (CONS (p':string) ps1)" th)));;
 e(IMP_RES_TAC (AND2IMP call_path_progress_AIMP));;
 e(POP_TAC THEN POP_TAC THEN UNDISCH_LAST_TAC THEN POP_TAC THEN POP_TAC);;
 e(DISCH_THEN (MP_TAC o SPECL["p:string";"ps3:string list"]));;
 e(REWRITE_TAC[(SYM o SPEC_ALL)call_path_progress_APPEND1]);;
 e(POP_TAC THEN POP_TAC THEN POP_TAC THEN POP_TAC
           THEN POP_TAC THEN POP_TAC THEN POP_TAC);;
 e(FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o
         SPEC "LENGTH (APPEND ps3 (CONS p' ps1) :string list)"));;
 e(REWRITE_TAC[(SYM o ASSUME)"LENGTH (ps:string list) = n"]);;
 e(REWRITE_TAC[ASSUME
         "ps = APPEND ps3 (CONS (p':string) (APPEND ps2 (CONS p' ps1)))"]);;
 e(REWRITE_TAC[LENGTH_APPEND;LENGTH;ADD_CLAUSES]);;
 e(SUBST1_TAC (SPECL["LENGTH (ps2:string list)";
                     "LENGTH (ps1:string list)"] ADD_SYM));;
 e(REWRITE_TAC[ADD_ASSOC;LESS_MONO_EQ]);;
 e(REWRITE_TAC[ADD_SUC;LESS_ADD_SUC]);;
 e(DISCH_THEN (MP_TAC o
            SPECL["APPEND ps3 (CONS p' ps1):string list";"p:string";
                  "r:env";"all_ps:string list";
                  "vars:var list";"vals:var list";"glbs:var list";
                  "pre:aexp";"post:aexp";"calls:prog_env";
                  "rec:aexp";"c:cmd"]));;
 e(ASM_REWRITE_TAC[LENGTH_APPEND;LENGTH]);;
 e(UNDISCH_TAC "~((p:string) IN SL ps)");;
 e(ASM_REWRITE_TAC[SL_APPEND;SET_LIST;IN_UNION;IN;DE_MORGAN_THM]);;
 e(DISCH_THEN REWRITE_THM);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;
 e(REPEAT STRIP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(ASM_REWRITE_TAC[]);;
>%


let SINGLY_RECURSIVE_PROGRESS =
 prove_thm
  (`SINGLY_RECURSIVE_PROGRESS`,
   "!ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       p IN (SL all_ps) /\
       ~(p IN SL ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       PATH_PROGRESS(AAND pre (induct_pre rec), p, ps, p, rec, r)",

   REPEAT STRIP_TAC
   THEN MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS
   THEN EXISTS_TAC "AAND pre (call_path_progress p ps p rec r)"
   THEN CONJ_TAC
   THENL
    [ ASSUME_TAC (REFL "LENGTH (ps:string list)")
      THEN IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_call_path_progress)
      THEN UNDISCH_LAST_TAC
      THEN POP_TAC
      THEN REWRITE_TAC[CLOSE;A_DEF]
      THEN DISCH_TAC THEN GEN_TAC THEN STRIP_TAC
      THEN RES_TAC
      THEN ASM_REWRITE_TAC[];

      DEFINE_NEW_VAR "y = APPEND (APPEND vars vals) glbs :var list"
      THEN IMP_RES_TAC (AND2IMP FVa_REC)
      THEN UNDISCH_LAST_TAC
      THEN FIRST_ASSUM (REWRITE_THM o SYM)
      THEN DISCH_TAC
      THEN IMP_RES_TAC (AND2IMP FVa_call_path_progress)
      THEN POP_ASSUM (ASSUME_TAC o SPEC_ALL)
      THEN IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_path_progress)
      THEN ASM_REWRITE_TAC[]
    ]
  );;

%< SINGLY_RECURSIVE_PROGRESS
g "!ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       p IN (SL all_ps) /\
       ~(p IN SL ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       PATH_PROGRESS(AAND pre (induct_pre rec), p, ps, p, rec, r)";;

e(REPEAT STRIP_TAC);;
e(MATCH_MP_TAC PRE_STRENGTH_PATH_PROGRESS);;
e(EXISTS_TAC "AAND pre (call_path_progress p ps p rec r)");;
e(CONJ_TAC);;

 e(ASSUME_TAC (REFL "LENGTH (ps:string list)"));; 
 e(IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_call_path_progress));;
 e(UNDISCH_LAST_TAC);;
 e(POP_TAC);;
 e(REWRITE_TAC[CLOSE;A_DEF]);;
 e(DISCH_TAC THEN GEN_TAC THEN STRIP_TAC);;
 e(RES_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(DEFINE_NEW_VAR "y = APPEND (APPEND vars vals) glbs :var list");;
 e(IMP_RES_TAC (AND2IMP FVa_REC));;
 e(UNDISCH_LAST_TAC
     THEN FIRST_ASSUM (REWRITE_THM o SYM)
     THEN DISCH_TAC);;
 e(IMP_RES_TAC (AND2IMP FVa_call_path_progress));;
 e(POP_ASSUM (ASSUME_TAC o SPEC_ALL));;
 e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_call_path_progress));;
 e(ASM_REWRITE_TAC[]);;
>%




let RECURSIVE_PROGRESS =
 prove_thm
  (`RECURSIVE_PROGRESS`,
   "!n ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       (LENGTH ps = n) /\
       p IN (SL all_ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       PATH_PROGRESS(AAND pre (induct_pre rec),p,ps,p,rec,r)",

   WELL_FOUNDED_NUM_TAC
   THEN REPEAT STRIP_TAC
   THEN ASM_CASES_TAC "(p:string) IN SL ps"
   THENL
    [ UNDISCH_TAC "(p:string) IN SL all_ps"
      THEN IMP_RES_TAC SPLIT_LIST
      THEN DISCH_TAC
      THEN IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_PROGRESS)
      THEN FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th)))
      THEN DISCH_THEN (MP_TAC o SPEC "LENGTH (ps2:string list)")
      THEN REWRITE_TAC[(SYM o ASSUME)"LENGTH (ps:string list) = n"]
      THEN REWRITE_TAC[ASSUME "ps = APPEND ps2 (CONS (p:string) ps1)"]
      THEN REWRITE_TAC[LENGTH_APPEND;LENGTH;LESS_ADD_SUC]
      THEN DISCH_THEN (MP_TAC o
            SPECL["ps2:string list";"p:string";"r:env";"all_ps:string list";
                  "vars:var list";"vals:var list";"glbs:var list";
                  "pre:aexp";"post:aexp";"calls:prog_env";
                  "rec:aexp";"c:cmd"])
      THEN ASM_REWRITE_TAC[]
      THEN DISCH_TAC
      THEN REWRITE_TAC[PATH_PROGRESS;A_DEF]
      THEN REWRITE_TAC[M_calls_APPEND]
      THEN REPEAT STRIP_TAC
      THEN IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC)
      THENL
       [ UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps2,p,rec,r)"
         THEN REWRITE_TAC[PATH_PROGRESS]
         THEN DISCH_THEN (MP_TAC o SPECL["s1:state";"s3:state"])
         THEN ASM_REWRITE_TAC[A_DEF;induct_pre];

         UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps2,p,rec,r)"
         THEN REWRITE_TAC[PATH_PROGRESS;A_DEF]
         THEN DISCH_THEN (IMP_RES_TAC o AND2IMP)
         THEN UNDISCH_LAST_TAC
         THEN ASM_REWRITE_TAC[A_DEF;V_DEF]
         THEN DISCH_TAC
         THEN IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre)
         THEN POP_ASSUM (MP_TAC o SPEC "ps2:string list")
         THEN REWRITE_TAC[PATH_PROGRESS]
         THEN DISCH_THEN (IMP_RES_TAC o AND2IMP)
         THEN UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps1,p,rec,r)"
         THEN ASM_REWRITE_TAC[PATH_PROGRESS;A_DEF;V_DEF;induct_pre]
         THEN DISCH_THEN (MP_TAC o
                 SPECL["BND (V v s3) y (s3:state)";"BND (V v s3) y (s2:state)"])
         THEN IMP_RES_TAC (AND2IMP REC_VAR_NON_INTERFERENCE)
         THEN IMP_RES_THEN REWRITE_THM A_BND_THM
         THEN IMP_RES_THEN REWRITE_THM V_BND_THM
         THEN REWRITE_TAC[BND_MATCH]
         THEN IMP_RES_TAC (AND2IMP M_calls_BND_THM)
         THEN ASM_REWRITE_TAC[]
         THEN STRIP_TAC
         THEN IMP_RES_TAC (AND2IMP LESS_TRANS)
         THEN IMP_RES_TAC (AND2IMP M_calls_PROG_ONLY_IMPACT)
         THEN POP_TAC
         THEN ASM_REWRITE_TAC[]
       ];

      IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_PROGRESS)
    ]
  );;

%< RECURSIVE_PROGRESS
g "!n ps p r all_ps vars vals glbs pre post calls rec c.
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_calls r /\
       (!p'. ~p' IN (SL all_ps) ==> (r p' = r0 p')) /\
       (LENGTH ps = n) /\
       p IN (SL all_ps) /\
       (r p = vars,vals,glbs,pre,post,calls,rec,c) /\
       ALL_EL CLOSE(graph_vcs all_ps r p) ==>
       PATH_PROGRESS(AAND pre (induct_pre rec),p,ps,p,rec,r)";;

e(WELL_FOUNDED_NUM_TAC);;
e(REPEAT STRIP_TAC);;
e(ASM_CASES_TAC "(p:string) IN SL ps");;
% 3 subgoals %

 e(UNDISCH_TAC "(p:string) IN SL all_ps");;
 e(IMP_RES_TAC SPLIT_LIST);;
 e(DISCH_TAC);;
 e(IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_PROGRESS));;
 e(FIRST_ASSUM (\th. (SPEC "0" th; UNDISCH_TAC (concl th))));;
 e(DISCH_THEN (MP_TAC o SPEC "LENGTH (ps2:string list)"));;
 e(REWRITE_TAC[(SYM o ASSUME)"LENGTH (ps:string list) = n"]);;
 e(REWRITE_TAC[ASSUME "ps = APPEND ps2 (CONS (p:string) ps1)"]);;
 e(REWRITE_TAC[LENGTH_APPEND;LENGTH;LESS_ADD_SUC]);;
 e(DISCH_THEN (MP_TAC o
            SPECL["ps2:string list";"p:string";"r:env";"all_ps:string list";
                  "vars:var list";"vals:var list";"glbs:var list";
                  "pre:aexp";"post:aexp";"calls:prog_env";
                  "rec:aexp";"c:cmd"]));;
 e(ASM_REWRITE_TAC[]);;
 e(DISCH_TAC);;
 e(REWRITE_TAC[PATH_PROGRESS;A_DEF]);;
 e(REWRITE_TAC[M_calls_APPEND]);;
 e(REPEAT STRIP_TAC);;
 e(IMP_RES_TAC (AND2IMP WELL_FORMED_PROCEDURE_REC));;

   e(UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps2,p,rec,r)");;
   e(REWRITE_TAC[PATH_PROGRESS]);;
   e(DISCH_THEN (MP_TAC o SPECL["s1:state";"s3:state"]));;
   e(ASM_REWRITE_TAC[A_DEF;induct_pre]);;

   e(UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps2,p,rec,r)");;
   e(REWRITE_TAC[PATH_PROGRESS;A_DEF]);;
   e(DISCH_THEN (IMP_RES_TAC o AND2IMP));;
   e(UNDISCH_LAST_TAC);;
   e(ASM_REWRITE_TAC[A_DEF;V_DEF]);;
   e(DISCH_TAC);;
   e(IMP_RES_TAC (AND2IMP PATH_PROGRESS_pre));;
   e(POP_ASSUM (MP_TAC o SPEC "ps2:string list"));;
   e(REWRITE_TAC[PATH_PROGRESS]);;
   e(DISCH_THEN (IMP_RES_TAC o AND2IMP));;
   e(UNDISCH_TAC "PATH_PROGRESS(AAND pre(induct_pre rec),p,ps1,p,rec,r)");;
   e(ASM_REWRITE_TAC[PATH_PROGRESS;A_DEF;V_DEF;induct_pre]);;
   e(DISCH_THEN (MP_TAC o
          SPECL["BND (V v s3) y (s3:state)";"BND (V v s3) y (s2:state)"]));;
   e(IMP_RES_TAC (AND2IMP REC_VAR_NON_INTERFERENCE));;
   e(IMP_RES_THEN REWRITE_THM A_BND_THM);;
   e(IMP_RES_THEN REWRITE_THM V_BND_THM);;
   e(REWRITE_TAC[BND_MATCH]);;
   e(IMP_RES_TAC (AND2IMP M_calls_BND_THM));;
   e(ASM_REWRITE_TAC[]);;
   e(STRIP_TAC);;
   e(IMP_RES_TAC (AND2IMP LESS_TRANS));;
   e(IMP_RES_TAC (AND2IMP M_calls_PROG_ONLY_IMPACT));;
   e(POP_TAC);;
   e(ASM_REWRITE_TAC[]);;

 e(IMP_RES_TAC (AND2IMP SINGLY_RECURSIVE_PROGRESS));;
>%





let graph_vcs_REC_PROGRESS =
 prove_thm
  (`graph_vcs_REC_PROGRESS`,
   "!p r all_ps vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (p IN SL all_ps) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      ALL_EL CLOSE (graph_vcs all_ps r p) ==>
      REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r)",

   REPEAT STRIP_TAC
   THEN REWRITE_TAC[REC_PROGRESS]
   THEN GEN_TAC
   THEN ASSUME_TAC (REFL "LENGTH (ps:string list)")
   THEN IMP_RES_TAC (AND2IMP RECURSIVE_PROGRESS)
  );;

%< graph_vcs_REC_PROGRESS
g "!p r all_ps vars vals glbs pre post calls rec c.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (p IN SL all_ps) /\
      (r p = (vars,vals,glbs,pre,post,calls,rec,c)) /\
      ALL_EL CLOSE (graph_vcs all_ps r p) ==>
      REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r)";;

e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[REC_PROGRESS]);;
e(GEN_TAC);;
e(ASSUME_TAC (REFL "LENGTH (ps:string list)"));;
e(IMP_RES_TAC (AND2IMP RECURSIVE_PROGRESS));;
>%



let vcgg =
 new_def
  `vcgg`
  "vcgg all_ps r =
      FLAT (MAP (graph_vcs all_ps r) all_ps)";;



let SOME_MAP_graph_vcs_REC_PROGRESS =
 TAC_PROOF
  (([],
    "!some_ps r all_ps.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (SL some_ps SUBSET SL all_ps) /\
      ALL_EL CLOSE (FLAT (MAP (graph_vcs all_ps r) some_ps)) ==>
      (!p. p IN SL some_ps ==>
         let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
         REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r))"),

   LIST_INDUCT_TAC
   THEN REWRITE_TAC[SET_LIST;IN;SUBSET]
   THEN REWRITE_TAC[MAP;FLAT;ALL_EL_APPEND]
   THEN REPEAT STRIP_TAC
   THENL
    [ POP_ASSUM (REWRITE_ALL_THM o SYM)
      THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
      THEN FIRST_ASSUM ONCE_REWRITE_THM
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN IMP_RES_TAC (AND2IMP graph_vcs_REC_PROGRESS);

      FIRST_ASSUM (\th. (SPEC "r:env" th; IMP_RES_TAC (AND2IMP th)))
    ]
  );;

%< SOME_MAP_graph_vcs_REC_PROGRESS
g "!some_ps r all_ps.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      (SL some_ps SUBSET SL all_ps) /\
      ALL_EL CLOSE (FLAT (MAP (graph_vcs all_ps r) some_ps)) ==>
      (!p. p IN SL some_ps ==>
         let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
         REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r))";;

e(LIST_INDUCT_TAC);;

 e(REWRITE_TAC[SET_LIST;IN]);;

 e(REWRITE_TAC[SET_LIST;IN;SUBSET]);;
 e(REWRITE_TAC[MAP;FLAT;ALL_EL_APPEND]);;
 e(REPEAT STRIP_TAC);;

  e(POP_ASSUM (REWRITE_ALL_THM o SYM));;
  e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
  e(FIRST_ASSUM ONCE_REWRITE_THM);;
  e(CONV_TAC (DEPTH_CONV let_CONV));;
  e(IMP_RES_TAC (AND2IMP graph_vcs_REC_PROGRESS));;

  e(FIRST_ASSUM (\th. SPEC "r:env" th; IMP_RES_TAC (AND2IMP th)));;
>%


let vcgg_REC_PROGRESS =
 prove_thm
  (`vcgg_REC_PROGRESS`,
   "!r all_ps.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      ALL_EL CLOSE (vcgg all_ps r) ==>
      (!p.
         let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
         REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r))",
   REWRITE_TAC[vcgg]
   THEN REPEAT GEN_TAC
   THEN STRIP_TAC
   THEN GEN_TAC
   THEN ASM_CASES_TAC "(p:string) IN SL all_ps"
   THENL
    [ IMP_RES_TAC (AND2IMP SOME_MAP_graph_vcs_REC_PROGRESS)
      THEN FIRST_ASSUM (IMP_RES_TAC o REWRITE_RULE[SUBSET_REFL] o
                        SPEC "all_ps:string list");

      RES_TAC
      THEN ASM_REWRITE_TAC[EMPTY_ENV]
      THEN CONV_TAC (DEPTH_CONV let_CONV)
      THEN REWRITE_TAC[REC_PROGRESS;PATH_PROGRESS;A_DEF]
    ]
  );;

%< vcgg_REC_PROGRESS
g "!r all_ps.
      WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      (!p'. ~(p' IN SL all_ps) ==> (r p' = r0 p')) /\
      ALL_EL CLOSE (vcgg all_ps r) ==>
      (!p.
         let (vars,vals,glbs,pre,post,calls,rec,c) = r p in
         REC_PROGRESS(AAND pre (induct_pre rec), p, rec, r))";;

e(REWRITE_TAC[vcgg]);;
e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(GEN_TAC);;
e(ASM_CASES_TAC "(p:string) IN SL all_ps");;

 e(IMP_RES_TAC (AND2IMP SOME_MAP_graph_vcs_REC_PROGRESS));;
 e(FIRST_ASSUM (IMP_RES_TAC o REWRITE_RULE[SUBSET_REFL] o
                SPEC "all_ps:string list"));;

 e(RES_TAC);;
 e(ASM_REWRITE_TAC[EMPTY_ENV]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[REC_PROGRESS;PATH_PROGRESS;A_DEF]);;
>%


let ALL_PROC_NAMES =
 TAC_PROOF
  (([],
    "!d r.
      (r = mkenv d r0) ==>
      (!p'. ~(p' IN SL (proc_names d)) ==> (r p' = r0 p'))"),
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC UNDECLARED_PROCEDURE
   THEN ASM_REWRITE_TAC[]
  );;

%< ALL_PROC_NAMES
g "!d r.
      (r = mkenv d r0) ==>
      (!p'. ~(p' IN SL (proc_names d)) ==> (r p' = r0 p'))";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC UNDECLARED_PROCEDURE);;
e(ASM_REWRITE_TAC[]);;
>%


let vcgg_THM =
 prove_thm
  (`vcgg_THM`,
   "!d r.
      (r = mkenv d r0) /\ WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      ALL_EL CLOSE (vcgg (proc_names d) r) ==>
      WFenv_rec r",
   REPEAT STRIP_TAC
   THEN IMP_RES_TAC ALL_PROC_NAMES
   THEN REWRITE_TAC[WFenv_rec;WFproc_rec]
   THEN GEN_TAC
   THEN IMP_RES_TAC (AND2IMP vcgg_REC_PROGRESS)
   THEN POP_ASSUM (MP_TAC o SPEC_ALL)
   THEN REWRITE_TAC[]
  );;

%< vcgg_THM
g "!d r.
      (r = mkenv d r0) /\ WFenv_syntax r /\ WFenv_pre r /\ WFenv_calls r /\
      ALL_EL CLOSE (vcgg (proc_names d) r) ==>
      WFenv_rec r";;

e(REPEAT STRIP_TAC);;
e(IMP_RES_TAC ALL_PROC_NAMES);;
e(REWRITE_TAC[WFenv_rec;WFproc_rec]);;
e(GEN_TAC);;
e(IMP_RES_TAC (AND2IMP vcgg_REC_PROGRESS));;
e(POP_ASSUM (MP_TAC o SPEC_ALL));;
e(REWRITE_TAC[]);;
>%



let TERMINATION =
 prove_thm
  (`TERMINATION`,
   "!d r.
       (r = mkenv d r0) /\
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_rec r /\
       WFenv_term r  ==>
       WFenv_total r",
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[WFenv_total;WFproc_total]
   THEN GEN_TAC
   THEN IMP_RES_TAC ALL_PROC_NAMES
   THEN ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE)
   THEN IMP_RES_TAC (AND2IMP PROCEDURE_TERMINATION)
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[terminates;TCSPEC;A_DEF]
   THEN FIRST_ASSUM ONCE_REWRITE_THM
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN FIRST_ASSUM (REWRITE_THM o SYM)
   THEN REPEAT STRIP_TAC
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  );;

% TERMINATION
g "!d r.
       (r = mkenv d r0) /\
       WFenv_syntax r /\
       WFenv_pre r /\
       WFenv_rec r /\
       WFenv_term r  ==>
       WFenv_total r";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[WFenv_total;WFproc_total]);;
e(GEN_TAC);;
e(IMP_RES_TAC ALL_PROC_NAMES);;
e(ASSUME_TAC (ISPEC "(r:env) p" ENV_TUPLE));;
e(IMP_RES_TAC (AND2IMP PROCEDURE_TERMINATION));;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[terminates;TCSPEC;A_DEF]);;
e(FIRST_ASSUM ONCE_REWRITE_THM);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(FIRST_ASSUM (REWRITE_THM o SYM));;
e(REPEAT STRIP_TAC);;
e(FIRST_ASSUM MATCH_MP_TAC);;
e(ASM_REWRITE_TAC[]);;
%




% ===================================== %
%    M A I N   V C G   T H E O R E M    %
% ===================================== %


let vcg =
    new_rec_def `vcg` prog_Axiom
      "(vcg (PROG d c) q =
              let r  = mkenv d r0 in
              let h1 = vcgd d r in
              let h2 = vcgg (proc_names d) r in
              let h3 = vcgc ATRUE c g0 q r in
                (h1 & h2 & h3))   ";;



let vcg_THM =
 prove_thm
  (`vcg_THM`,
   "!pi q. WFp pi /\ ALL_EL CLOSE (vcg pi q)  ==>
           PSPEC(pi,q)",
   INDUCT_THEN prog_induct ASSUME_TAC
   THEN REWRITE_TAC[WFp;vcg]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND]
   THEN REPEAT STRIP_TAC
   THEN ASSUME_TAC (REFL "mkenv d r0")
   THEN IMP_RES_TAC vcgd_THM
   THEN IMP_RES_TAC WFenvp_aspects
   THEN IMP_RES_TAC vcgd_PRE_PROGRESS

   THEN IMP_RES_TAC vcgd_BODY_PROGRESS
   THEN IMP_RES_TAC vcgd_TERM
   THEN IMP_RES_TAC (AND2IMP vcgg_THM)
   THEN IMP_RES_TAC (AND2IMP TERMINATION)
   THEN IMP_RES_TAC WFenv_correct_aspects
   THEN IMP_RES_TAC WFenv_aspects
   THEN FIRST_ASSUM (UNASSUME_THEN (\th.ALL_TAC) o concl o SYM)
   THEN REWRITE_TAC[PSPEC]
   THEN REWRITE_TAC[P_cases]
   THEN REWRITE_TAC[prog_one_one]
   THEN IMP_RES_TAC vcgc_THM
   THEN UNDISCH_LAST_TAC
   THEN REWRITE_TAC[CSPEC_THM;A_DEF]
   THEN REPEAT STRIP_TAC
   THENL
    [ FIRST_ASSUM (UNASSUME_THEN (REWRITE_ALL_THM o SYM) o concl)
      THEN FIRST_ASSUM (UNASSUME_THEN (REWRITE_ALL_THM o SYM) o concl)
      THEN IMP_RES_TAC D_IMP_mkenv
      THEN POP_ASSUM (REWRITE_ALL_THM o SYM)
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC "s0"
      THEN ASM_REWRITE_TAC[];

      POP_ASSUM (STRIP_ASSUME_TAC o SPEC "s0")
      THEN EXISTS_TAC "s2:state"
      THEN EXISTS_TAC "d:decl"
      THEN EXISTS_TAC "c:cmd"
      THEN EXISTS_TAC "mkenv d r0"
      THEN ASM_REWRITE_TAC[D_mkenv]
    ]
  );;

% vcg_THM:
First proven 10:22 am, June 20, 1994.  (Pass 3)
First proven for Pass 4.1 at 3:13 pm, July 16, 1994.
First proven for Pass 4.2 at 4:56 pm, November 2, 1994.
First proven for Pass 4.3 at 4:23 pm, March 18, 1995.

g "!pi q. WFp pi /\ ALL_EL CLOSE (vcg pi q)  ==>
          PSPEC(pi,q)";;
e(INDUCT_THEN prog_induct ASSUME_TAC);;
e(REWRITE_TAC[WFp;vcg]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_INFIX;ALL_EL_APPEND]);;
e(REPEAT STRIP_TAC);;
e(ASSUME_TAC (REFL "mkenv d r0"));;
e(IMP_RES_TAC vcgd_THM);;
e(IMP_RES_TAC WFenvp_aspects);;
e(IMP_RES_TAC vcgd_PRE_PROGRESS);;
e(IMP_RES_TAC vcgd_BODY_PROGRESS);;
e(IMP_RES_TAC vcgd_TERM);;
e(IMP_RES_TAC (AND2IMP vcgg_THM));;
e(IMP_RES_TAC (AND2IMP TERMINATION));;
e(IMP_RES_TAC WFenv_correct_aspects);;
e(IMP_RES_TAC WFenv_aspects);;
e(FIRST_ASSUM (UNASSUME_THEN (\th.ALL_TAC) o concl o SYM));;
e(REWRITE_TAC[PSPEC]);;
e(REWRITE_TAC[P_cases]);;
e(REWRITE_TAC[prog_one_one]);;
e(IMP_RES_TAC vcgc_THM);;
e(UNDISCH_LAST_TAC);;
e(REWRITE_TAC[CSPEC_THM;A_DEF]);;
e(REPEAT STRIP_TAC);;

 e(FIRST_ASSUM (UNASSUME_THEN (REWRITE_ALL_THM o SYM) o concl));;
 e(FIRST_ASSUM (UNASSUME_THEN (REWRITE_ALL_THM o SYM) o concl));;
 e(IMP_RES_TAC D_IMP_mkenv);;
 e(POP_ASSUM (REWRITE_ALL_THM o SYM));;
 e(FIRST_ASSUM MATCH_MP_TAC);;
 e(EXISTS_TAC "s0");;
 e(ASM_REWRITE_TAC[]);;

 e(POP_ASSUM (STRIP_ASSUME_TAC o SPEC "s0"));;
 e(EXISTS_TAC "s2:state");;
 e(EXISTS_TAC "d:decl");;
 e(EXISTS_TAC "c:cmd");;
 e(EXISTS_TAC "mkenv d r0");;
 e(ASM_REWRITE_TAC[D_mkenv]);;
%



close_theory ();;

